/*
 * $Id: //depot/prod/newport_news/main/modules/jndmp/src/main/java/com/netapp/ndmp/NdmpConnection.java#6 $
 *
 * Copyright (c) 2010 NetApp, Inc.
 * All rights reserved.
 */

package com.netapp.ndmp;

import static com.netapp.ndmp.rpcgen.ndmp_message.*;

import java.io.IOException;
import java.net.InetSocketAddress;
import java.net.Socket;
import java.net.SocketAddress;
import java.util.HashMap;
import java.util.Map;

import org.acplt.oncrpc.OncRpcException;
import org.acplt.oncrpc.XdrAble;
import org.acplt.oncrpc.XdrTcpDecodingStream;
import org.acplt.oncrpc.XdrTcpEncodingStream;

import com.netapp.ndmp.rpcgen.ndmp_connection_status_reason;
import com.netapp.ndmp.rpcgen.ndmp_header;
import com.netapp.ndmp.rpcgen.ndmp_header_message_type;
import com.netapp.ndmp.rpcgen.ndmp_log_message_post;
import com.netapp.ndmp.rpcgen.ndmp_message;
import com.netapp.ndmp.rpcgen.ndmp_notify_connection_status_post;
import com.netapp.ndmp.rpcgen.ndmp_notify_data_halted_post;
import com.netapp.ndmp.rpcgen.ndmp_notify_data_read_post;

/**
 * This class manages the details of the underlying socket connection, as well as the encoding/decoding of the various
 * XdrAble classes which represent NDMP request/responses.  Mostly here to make NdmpSession unit testable.
 *
 * @author thomaswb
 * @version $Revision: #6 $
 *          Date: $Date: 2012/08/05 $
 */
class NdmpConnection {
    private static enum ConnectState {
        NEW,
        CONNECTED,
        DISCONNECTED
    }

    private final SocketAddress address;
    private final int connectTimeout;
    private final int readTimeout;

    private Socket socket;
    private XdrTcpEncodingStream out;
    private XdrTcpDecodingStream in;
    private int clientSequence;
    private int serverSequence;
    private ConnectState state;
    private int protocolVersion;
    private Notification lastNotification;
    private final Map<Integer, NotificationCallback<?>> notificationCallbacks;


    NdmpConnection(String host, int port, int connectTimeout, int readTimeout) {
        address = new InetSocketAddress(host, port);
        this.connectTimeout = connectTimeout;
        this.readTimeout = readTimeout;
        notificationCallbacks = new HashMap<Integer, NotificationCallback<?>>();
        //Register a callback to update our connection status on receipt of NDMP_NOTIFY_CONNECTION_STATUS
        registerNotificationCallback(NDMP_NOTIFY_CONNECTION_STATUS,
                new NotificationCallback<ndmp_notify_connection_status_post>() {
                    @Override
                    public Class<ndmp_notify_connection_status_post> getNotificationClass() {
                        return ndmp_notify_connection_status_post.class;
                    }

                    @Override
                    public void doCallback(ndmp_notify_connection_status_post notification) throws NdmpException {
                        updateConnectionStatus(notification);
                    }
                });

        //Register a callback for NDMP_NOTIFY_DATA_READ
        registerNotificationCallback(NDMP_NOTIFY_DATA_READ,
                new NotificationCallback<ndmp_notify_data_read_post>() {
                    @Override
                    public Class<ndmp_notify_data_read_post> getNotificationClass() {
                        return ndmp_notify_data_read_post.class;
                    }

                    @Override
                    public void doCallback(ndmp_notify_data_read_post notification) throws NdmpException {
                    }
                });

        //Register a callback for NDMP_NOTIFY_DATA_HALTED
        registerNotificationCallback(NDMP_NOTIFY_DATA_HALTED,
                new NotificationCallback<ndmp_notify_data_halted_post>() {
                    @Override
                    public Class<ndmp_notify_data_halted_post> getNotificationClass() {
                        return ndmp_notify_data_halted_post.class;
                    }

                    @Override
                    public void doCallback(ndmp_notify_data_halted_post notification) throws NdmpException {
                    }
                });

        //Register a callback for NDMP_LOG_MESSAGE
        registerNotificationCallback(NDMP_LOG_MESSAGE,
                new NotificationCallback<ndmp_log_message_post>() {
                    @Override
                    public Class<ndmp_log_message_post> getNotificationClass() {
                        return ndmp_log_message_post.class;
                    }

                    @Override
                    public void doCallback(ndmp_log_message_post notification) throws NdmpException {
                    }
                });
    }

    void connect() throws NdmpConnectException {
        try {
            socket = new Socket();
            socket.setSoTimeout(readTimeout);
            socket.connect(address, connectTimeout);
            in = new XdrTcpDecodingStream(socket, 1024);
            out = new XdrTcpEncodingStream(socket, 1024);
            state = ConnectState.NEW;
            waitForAndProcessNotification(NDMP_NOTIFY_CONNECTION_STATUS, ndmp_notify_connection_status_post.class);
            if (state != ConnectState.CONNECTED) {
                close();
                throw new NdmpException("Connection failed");
            }
        } catch (Exception e) {
            throw new NdmpConnectException(e);
        }
    }

    /**
     * Sends a request to the remote NDMP host and returns the response.  If reply is null,
     * this call returns immediately.
     */
    <T extends XdrAble> T sendMessage(int messageCode, XdrAble request, T reply) throws NdmpException {
        if (state != ConnectState.CONNECTED) {
            throw new NdmpException("Connection not open");
        }

        ndmp_header requestHeader = new ndmp_header();
        requestHeader.sequence = ++clientSequence;
        requestHeader.time_stamp = (int) (System.currentTimeMillis() / 1000);
        requestHeader.message_type = ndmp_header_message_type.NDMP_MESSAGE_REQUEST;
        requestHeader.message_code = messageCode;
        requestHeader.reply_sequence = serverSequence;
        requestHeader.error_code = 0;

        try {
            out.beginEncoding(null, 0);
            requestHeader.xdrEncode(out);
            if (request != null) {
                request.xdrEncode(out);
            }
            out.endEncoding(true);
            if (reply != null) {
                waitForReply(messageCode, reply);
            }
            return reply;
        } catch (OncRpcException e) {
            throw new NdmpException(e);
        } catch (IOException e) {
            throw new NdmpCommunicationException(e);
        }
    }

    void close() {
        if (state == ConnectState.CONNECTED) {
            try {
                sendMessage(NDMP_CONNECT_CLOSE, null, null);
                waitForAndProcessNotification(NDMP_NOTIFY_CONNECTION_STATUS, ndmp_notify_connection_status_post.class);
                if (state != ConnectState.DISCONNECTED) {
                    throw new NdmpException("Disconnect failed");
                }
            } catch (NdmpException e) {
                //Ignored
            }
        }
        try {
            if (socket != null) {
                socket.close();
            }
        } catch (IOException e) {
            //Ignored
        } finally {
            socket = null;
        }
    }

    void registerNotificationCallback(int notificationCode, NotificationCallback<? extends XdrAble> notificationCallback) {
        notificationCallbacks.put(notificationCode, notificationCallback);
    }

    boolean isConnected() {
        return state == NdmpConnection.ConnectState.CONNECTED;
    }

    int getProtocolVersion() {
        return protocolVersion;
    }

    private XdrAble waitForReply(int messageCode, XdrAble reply) throws NdmpException {
        try {
            in.beginDecoding();
            ndmp_header header = getHeader();

            // Ignore the NDMP_LOG_MESSAGE: burt 250387
            // Since filers can send notifications randomly, we should filter them out
            // to get a correct reply.
            while (header.message_code == ndmp_message.NDMP_LOG_MESSAGE) {
                processNotification(header);
                in.endDecoding();
                in.beginDecoding();
                header = getHeader();
            }

            if (header.message_type == ndmp_header_message_type.NDMP_MESSAGE_REPLY) {
                if (header.message_code != messageCode) {
                    if (header.message_code != ndmp_message.NDMP_NOTIFY_DATA_HALTED) {
                        throw new NdmpProtocolException("Unexpected reply code: %d (expecting %d)", header.message_code, messageCode);
                    } else {
                        ndmp_notify_data_halted_post dataHaltedReply = new ndmp_notify_data_halted_post();
                        dataHaltedReply.xdrDecode(in);
                        throw new NdmpDataHaltedException(dataHaltedReply.reason);
                    }
                }
                reply.xdrDecode(in);
                int errorCode = reply.getClass().getField("error").getInt(reply);
                if (errorCode != 0) {
                    throw new NdmpMessageException(messageCode, errorCode);
                }
                return reply;
            } else {
                throw new NdmpProtocolException("Unknown message type: %d, message_code: 0x%x", header.message_type, header.message_code);
            }
        } catch (IOException e) {
            throw new NdmpCommunicationException(e);
        } catch (Exception e) {
            throw new NdmpException(e);
        } finally {
            try {
                in.endDecoding();
            } catch (IOException e) {
                throw new NdmpCommunicationException(e);
            } catch (OncRpcException e) {
                throw new NdmpException(e);
            }
        }
    }

    <T extends XdrAble> T waitForAndProcessNotification(int messageCode, Class<T> clazz) throws NdmpException {
        return waitForAndProcessNotificationWithLog(messageCode, clazz, null);
    }

    <T extends XdrAble> T waitForAndProcessNotificationWithLog(int messageCode, Class<T> clazz, ndmp_log_message_post logMsg) throws NdmpException {
        boolean withLog = (logMsg != null);
        if (!notificationCallbacks.containsKey(messageCode)) {
            throw new IllegalArgumentException("Unknown notify code");
        }
        if (!notificationCallbacks.get(messageCode).getNotificationClass().equals(clazz)) {
            throw new IllegalArgumentException("Wrong class for message code");
        }

        // Ignore the NDMP_LOG_MESSAGE: burt 250387
        // Since filers can send notifications randomly, we should filter them out
        // to get a correct notification.
        // To get log messages, the input logMsg should not be null. If a log message is received,
        // this function returns 'null' and assigns the ndmp log message to the logMsg.
        // The 'messageCode' cannot be 'ndmp_message.NDMP_LOG_MESSAGE'.
        if (lastNotification != null
            && logMsg == null
            && lastNotification.header.message_code == ndmp_message.NDMP_LOG_MESSAGE) {
            lastNotification = null;
        }

        if (lastNotification == null) {
            try {
                in.beginDecoding();
                ndmp_header header = getHeader();
                if (!withLog) {
                    // ignore the NDMP_LOG_MESSAGE: burt 250387
                    while (header.message_code == ndmp_message.NDMP_LOG_MESSAGE) {
                        processNotification(header);
                        in.endDecoding();
                        in.beginDecoding();
                        header = getHeader();
                    }
                }
                processNotification(header);
            } catch (IOException e) {
                throw new NdmpCommunicationException(e);
            } catch (OncRpcException e) {
                throw new NdmpException(e);
            } finally {
                try {
                    in.endDecoding();
                } catch (IOException e) {
                    throw new NdmpCommunicationException(e);
                } catch (OncRpcException e) {
                    throw new NdmpException(e);
                }
            }
        }
        if (lastNotification.header.message_code == ndmp_message.NDMP_LOG_MESSAGE) {
            XdrAble body = lastNotification.body;
            lastNotification = null;
            ndmp_log_message_post receivedLogMsg = ndmp_log_message_post.class.cast(body);
            logMsg.log_type = receivedLogMsg.log_type;
            logMsg.message_id = receivedLogMsg.message_id;
            logMsg.entry = receivedLogMsg.entry;
            return null;
        } else {
            if (lastNotification.header.message_code != messageCode) {
                if (lastNotification.header.message_code != ndmp_message.NDMP_NOTIFY_DATA_HALTED) {
                    throw new NdmpProtocolException("Unexpected notification message: " + lastNotification.header.message_code);
                } else {
                    ndmp_notify_data_halted_post dataHaltedReply = ndmp_notify_data_halted_post.class.cast(lastNotification.body);
                    lastNotification = null;
                    throw new NdmpDataHaltedException(dataHaltedReply.reason);
                }
            }
            XdrAble body = lastNotification.body;
            lastNotification = null;
            return clazz.cast(body);
        }
    }

    private void updateConnectionStatus(ndmp_notify_connection_status_post notification) throws NdmpException {
        protocolVersion = notification.protocol_version;
        if (notification.reason == ndmp_connection_status_reason.NDMP_SHUTDOWN) {
            state = ConnectState.DISCONNECTED;
        } else if (notification.reason == ndmp_connection_status_reason.NDMP_CONNECTED) {
            state = ConnectState.CONNECTED;
        } else {
            throw new NdmpException("Connection refused: " + notification.reason);
        }
    }

    private ndmp_header getHeader() throws NdmpException, OncRpcException, IOException {
        if (state == ConnectState.DISCONNECTED) {
            throw new NdmpException("Connection closed");
        }
        ndmp_header header = new ndmp_header();
        header.xdrDecode(in);
        serverSequence = header.sequence;
        if (header.error_code != 0) {
            throw new NdmpHeaderException(header);
        }
        return header;
    }

    private void processNotification(ndmp_header header) throws OncRpcException, NdmpException {
        if (!notificationCallbacks.containsKey(header.message_code)) {
            throw new NdmpProtocolException("Unknown notify code: %d", header.message_code);
        }

        try {
            NotificationCallback callback = notificationCallbacks.get(header.message_code);
            Class<? extends XdrAble> postClass = callback.getNotificationClass();
            XdrAble notification = postClass.newInstance();
            notification.xdrDecode(in);
            callback.doCallback(notification);
            lastNotification = new Notification(header, notification);
        } catch (InstantiationException e) {
            throw new RuntimeException(e);
        } catch (IllegalAccessException e) {
            throw new RuntimeException(e);
        } catch (IOException e) {
            throw new NdmpCommunicationException(e);
        }
    }


    private static class Notification {
        private final ndmp_header header;
        private final XdrAble body;

        private Notification(ndmp_header header, XdrAble body) {
            this.header = header;
            this.body = body;
        }
    }
}
