/*
 * $Id: //depot/prod/newport_news/main/modules/jndmp/src/main/java/com/netapp/ndmp/NdmpHandlerTable.java#4 $
 *
 * Copyright (c) 2008 Network Appliance, Inc.
 * All rights reserved.
 */

package com.netapp.ndmp;

//CHECKSTYLE:OFF AvoidStarImport importing generically from generated classes
import static com.netapp.ndmp.rpcgen.ndmp_Snapmirror_message.*;
import static com.netapp.ndmp.rpcgen.ndmp_Snapvault_primary_message.*;
import static com.netapp.ndmp.rpcgen.ndmp_Snapvault_secondary_message.*;
import static com.netapp.ndmp.rpcgen.ndmp_message.*;
import static com.netapp.ndmp.rpcgen.ndmp_snap_message.*;

import java.util.HashMap;
import java.util.Map;

import com.google.common.collect.ImmutableMap;
import org.acplt.oncrpc.XdrAble;

import com.netapp.ndmp.rpcgen.*;

//import com.netapp.ndmp.rpcgen.*;

//CHECKSTYLE:ON AvoidStarImport

/**
 * This class maps NDMP message codes to the generated request/response classes.
 *
 */
class NdmpHandlerTable {
    /**
     * Some entries change depending on what version of various extensions we negotiate with the server.  This map
     * is a prototype for the base entries.
     */
    private static final Map<Integer, HandlerEntry> HANDLER_MAP_PROTOTYPE;

    private Map<Integer, HandlerEntry> handlerTable;


    public NdmpHandlerTable() {
        handlerTable = new HashMap<Integer, HandlerEntry>(HANDLER_MAP_PROTOTYPE);
    }

    public boolean isValidRequest(int messageCode) {
        return handlerTable.containsKey(messageCode);
    }

    public Class<? extends XdrAble> getRequestClass(int messageCode) {
        HandlerEntry entry = handlerTable.get(messageCode);
        if (entry == null) {
            throw new IllegalArgumentException(String.format("No request class for messageCode %d", messageCode));
        }
        return entry.requestClass;
    }

    public Class<? extends XdrAble> getReplyClass(int messageCode) {
        HandlerEntry entry = handlerTable.get(messageCode);
        if (entry == null) {
            throw new IllegalArgumentException(String.format("No reply class for messageCode %d", messageCode));
        }
        return entry.replyClass;
    }

    private static class HandlerEntry {
        private final Class<? extends XdrAble> requestClass;
        private final Class<? extends XdrAble> replyClass;

        private HandlerEntry(Class<? extends XdrAble> requestClass, Class<? extends XdrAble> replyClass) {
            this.requestClass = requestClass;
            this.replyClass = replyClass;
        }
    }

    public void setExtensionVersions(Map<Short, Short> extensionVersions) {
        Short snapExtVersion = extensionVersions.get(NdmpSession.SNAP_EXT_CLASS);
        if (snapExtVersion == null) {
            snapExtVersion = 0;
        }

        if (snapExtVersion >= 4) {
            handlerTable.put(NDMP_SVS_GET_STATUS, new HandlerEntry(ndmp_svs_get_status_request.class, ndmp_svs_get_status_reply_v4.class));
            handlerTable.put(NDMP_SNAP_GET_QTREE_CONTENTS, new HandlerEntry(ndmp_snap_qtree_contents_request_v4.class, ndmp_snap_qtree_contents_reply_v4.class));
            handlerTable.put(NDMP_SVP_GET_STATUS, new HandlerEntry(ndmp_svp_get_status_request.class, ndmp_svp_get_status_reply_v4.class));
        }
        if (snapExtVersion >= 6) {
            handlerTable.put(NDMP_SVS_RESTART_RELATIONSHIP, new HandlerEntry(ndmp_svs_restart_relationship_request_v6.class, ndmp_svs_restart_relationship_reply.class));
            handlerTable.put(NDMP_SVS_CREATE_RELATIONSHIP, new HandlerEntry(ndmp_svs_create_relationship_request_v6.class, ndmp_svs_create_relationship_reply.class));
            handlerTable.put(NDMP_SVS_TRANSFER_BACKUP, new HandlerEntry(ndmp_svs_transfer_backup_request_v6.class, ndmp_svs_transfer_backup_reply.class));
        }
        if (snapExtVersion >= 7) {
            handlerTable.put(NDMP_SVS_CREATE_RELATIONSHIP, new HandlerEntry(ndmp_svs_create_relationship_request_v7.class, ndmp_svs_create_relationship_reply.class));
            handlerTable.put(NDMP_SVS_TRANSFER_BACKUP, new HandlerEntry(ndmp_svs_transfer_backup_request_v7.class, ndmp_svs_transfer_backup_reply.class));
            handlerTable.put(NDMP_SVP_RECOVER_BACKUP, new HandlerEntry(ndmp_svp_recover_backup_request_v7.class, ndmp_svp_recover_backup_reply.class));
        }
        if (snapExtVersion >= 9) {
            handlerTable.put(NDMP_SVP_RECOVER_BACKUP, new HandlerEntry(ndmp_svp_recover_backup_request_v9.class, ndmp_svp_recover_backup_reply.class));
            handlerTable.put(NDMP_SVS_CREATE_RELATIONSHIP, new HandlerEntry(ndmp_svs_create_relationship_request_v9.class, ndmp_svs_create_relationship_reply.class));
        }

        Short cabExtVersion = extensionVersions.get(NdmpSession.CAB_EXT_CLASS);
        if (cabExtVersion != null && cabExtVersion != 0) {
            handlerTable.put(NDMP_CAB_DATA_CONN_PREPARE, new HandlerEntry(ndmp_data_conn_prepare_request.class, ndmp_data_conn_prepare_reply.class));
        }
    }

    static {
        ImmutableMap.Builder<Integer, HandlerEntry> builder = ImmutableMap.builder();
        builder.put(NDMP_CONNECT_OPEN, new HandlerEntry(ndmp_connect_open_request.class, ndmp_connect_open_reply.class));
        builder.put(NDMP_CONNECT_CLIENT_AUTH, new HandlerEntry(ndmp_connect_client_auth_request.class, ndmp_connect_client_auth_reply.class));
        builder.put(NDMP_CONNECT_CLOSE, new HandlerEntry(null, null));
        builder.put(NDMP_CONNECT_SERVER_AUTH, new HandlerEntry(ndmp_connect_server_auth_request.class, ndmp_connect_server_auth_reply.class));
        builder.put(NDMP_CONFIG_GET_HOST_INFO, new HandlerEntry(null, ndmp_config_get_host_info_reply.class));
        builder.put(NDMP_CONFIG_GET_CONNECTION_TYPE, new HandlerEntry(null, ndmp_config_get_connection_type_reply.class));
        builder.put(NDMP_CONFIG_GET_AUTH_ATTR, new HandlerEntry(ndmp_config_get_auth_attr_request.class, ndmp_config_get_auth_attr_reply.class));
        builder.put(NDMP_CONFIG_GET_BUTYPE_INFO, new HandlerEntry(null, ndmp_config_get_butype_attr_reply.class));
        builder.put(NDMP_CONFIG_GET_FS_INFO, new HandlerEntry(null, ndmp_config_get_fs_info_reply.class));
        builder.put(NDMP_CONFIG_GET_TAPE_INFO, new HandlerEntry(null, ndmp_config_get_tape_info_reply.class));
        builder.put(NDMP_CONFIG_GET_SCSI_INFO, new HandlerEntry(null, ndmp_config_get_scsi_info_reply.class));
        builder.put(NDMP_CONFIG_GET_SERVER_INFO, new HandlerEntry(null, ndmp_config_get_server_info_reply.class));
        builder.put(NDMP_CONFIG_SET_EXT_LIST, new HandlerEntry(ndmp_config_set_ext_list_request.class, ndmp_config_set_ext_list_reply.class));
        builder.put(NDMP_CONFIG_GET_EXT_LIST, new HandlerEntry(null, ndmp_config_get_ext_list_reply.class));
        builder.put(NDMP_SCSI_OPEN, new HandlerEntry(ndmp_scsi_open_request.class, ndmp_scsi_open_reply.class));
        builder.put(NDMP_SCSI_CLOSE, new HandlerEntry(null, ndmp_scsi_close_reply.class));
        builder.put(NDMP_SCSI_GET_STATE, new HandlerEntry(null, ndmp_scsi_get_state_reply.class));
        builder.put(NDMP_SCSI_RESET_DEVICE, new HandlerEntry(null, ndmp_scsi_reset_device_reply.class));
        builder.put(NDMP_SCSI_EXECUTE_CDB, new HandlerEntry(ndmp_scsi_execute_cdb_request.class, ndmp_scsi_execute_cdb_reply.class));
        builder.put(NDMP_TAPE_OPEN, new HandlerEntry(ndmp_tape_open_request.class, ndmp_tape_open_reply.class));
        builder.put(NDMP_TAPE_CLOSE, new HandlerEntry(null, ndmp_tape_close_reply.class));
        builder.put(NDMP_TAPE_GET_STATE, new HandlerEntry(null, ndmp_tape_get_state_reply.class));
        builder.put(NDMP_TAPE_MTIO, new HandlerEntry(ndmp_tape_mtio_request.class, ndmp_tape_mtio_reply.class));
        builder.put(NDMP_TAPE_WRITE, new HandlerEntry(ndmp_tape_write_request.class, ndmp_tape_write_reply.class));
        builder.put(NDMP_TAPE_READ, new HandlerEntry(ndmp_tape_read_request.class, ndmp_tape_read_reply.class));
        builder.put(NDMP_TAPE_EXECUTE_CDB, new HandlerEntry(ndmp_tape_execute_cdb_request.class, ndmp_tape_execute_cdb_reply.class));
        builder.put(NDMP_DATA_GET_STATE, new HandlerEntry(null, ndmp_data_get_state_reply.class));
        builder.put(NDMP_DATA_START_BACKUP, new HandlerEntry(ndmp_data_start_backup_request.class, ndmp_data_start_backup_reply.class));
        builder.put(NDMP_DATA_START_RECOVER, new HandlerEntry(ndmp_data_start_recover_request.class, ndmp_data_start_recover_reply.class));
        builder.put(NDMP_DATA_ABORT, new HandlerEntry(null, ndmp_data_abort_reply.class));
        builder.put(NDMP_DATA_GET_ENV, new HandlerEntry(null, ndmp_data_get_env_reply.class));
        builder.put(NDMP_DATA_STOP, new HandlerEntry(null, ndmp_data_stop_reply.class));
        builder.put(NDMP_DATA_LISTEN, new HandlerEntry(ndmp_data_listen_request.class, ndmp_data_listen_reply.class));
        builder.put(NDMP_DATA_CONNECT, new HandlerEntry(ndmp_data_connect_request.class, ndmp_data_connect_reply.class));

        builder.put(NDMP_MOVER_GET_STATE, new HandlerEntry(null, ndmp_mover_get_state_reply.class));
        builder.put(NDMP_MOVER_LISTEN, new HandlerEntry(ndmp_mover_listen_request.class, ndmp_mover_listen_reply.class));
        builder.put(NDMP_MOVER_CONTINUE, new HandlerEntry(null, ndmp_mover_continue_reply.class));
        builder.put(NDMP_MOVER_ABORT, new HandlerEntry(null, ndmp_mover_abort_reply.class));
        builder.put(NDMP_MOVER_STOP, new HandlerEntry(null, ndmp_mover_stop_reply.class));
        builder.put(NDMP_MOVER_SET_WINDOW, new HandlerEntry(ndmp_mover_set_window_request.class, ndmp_mover_set_window_reply.class));
        builder.put(NDMP_MOVER_READ, new HandlerEntry(ndmp_mover_read_request.class, ndmp_mover_read_reply.class));
        builder.put(NDMP_MOVER_CLOSE, new HandlerEntry(null, ndmp_mover_close_reply.class));
        builder.put(NDMP_MOVER_SET_RECORD_SIZE, new HandlerEntry(ndmp_mover_set_record_size_request.class, ndmp_mover_set_record_size_reply.class));
        builder.put(NDMP_MOVER_CONNECT, new HandlerEntry(ndmp_mover_connect_request.class, ndmp_mover_connect_reply.class));

        builder.put(NDMP_SNAP_CREATE, new HandlerEntry(ndmp_snap_create_request.class, ndmp_snap_create_reply.class));
        builder.put(NDMP_SNAP_DELETE, new HandlerEntry(ndmp_snap_delete_request.class, ndmp_snap_delete_reply.class));
        builder.put(NDMP_SNAP_RECOVER, new HandlerEntry(ndmp_snap_recover_request.class, ndmp_snap_recover_reply.class));
        builder.put(NDMP_SNAP_RENAME, new HandlerEntry(ndmp_snap_rename_request.class, ndmp_snap_rename_reply.class));
        builder.put(NDMP_SNAP_GET_SERVER_INFO, new HandlerEntry(null, ndmp_snap_get_server_info_reply.class));
        builder.put(NDMP_SNAP_GET_SNAPSHOT_NAMES, new HandlerEntry(ndmp_snap_get_snapshot_names_request.class, ndmp_snap_get_snapshot_names_reply.class));
        builder.put(NDMP_SNAP_GET_SNAPSHOT_STATE, new HandlerEntry(ndmp_snap_get_snapshot_state_request.class, ndmp_snap_get_snapshot_state_reply.class));
        builder.put(NDMP_SNAP_GET_CDR_OP_STATE, new HandlerEntry(ndmp_snap_get_cdr_op_state_request.class, ndmp_snap_get_cdr_op_state_reply.class));
        builder.put(NDMP_SNAP_GET_SNAPSHOT_PATH, new HandlerEntry(ndmp_snap_get_snapshot_path_request.class, ndmp_snap_get_snapshot_path_reply.class));
        builder.put(NDMP_SNAP_SET_RESERVE, new HandlerEntry(ndmp_snap_set_reserve_request.class, ndmp_snap_set_reserve_reply.class));
        builder.put(NDMP_SNAP_GET_RESERVE_INFO, new HandlerEntry(ndmp_snap_get_reserve_info_request.class, ndmp_snap_get_reserve_info_reply.class));
        builder.put(NDMP_SNAP_SET_SCHEDULE, new HandlerEntry(ndmp_snap_set_schedule_request.class, ndmp_snap_set_schedule_reply.class));
        builder.put(NDMP_SNAP_GET_SCHEDULE_INFO, new HandlerEntry(ndmp_snap_get_schedule_info_request.class, ndmp_snap_get_schedule_info_reply.class));
        builder.put(NDMP_SNAP_DIR_LIST, new HandlerEntry(ndmp_snap_dir_list_request.class, ndmp_snap_dir_list_reply.class));

        builder.put(NDMP_SVS_MODIFY_RELATIONSHIP, new HandlerEntry(ndmp_svs_modify_relationship_request.class, ndmp_svs_modify_relationship_reply.class));
        builder.put(NDMP_SVS_ABORT_TRANSFER, new HandlerEntry(ndmp_svs_abort_transfer_request.class, ndmp_svs_abort_transfer_reply.class));
        builder.put(NDMP_SVS_DELETE_RELATIONSHIP, new HandlerEntry(ndmp_svs_delete_relationship_request.class, ndmp_svs_delete_relationship_reply.class));
        builder.put(NDMP_SVS_SET_SNAP_SCHED, new HandlerEntry(ndmp_svs_set_snap_sched_request.class, ndmp_svs_set_snap_sched_reply.class));
        builder.put(NDMP_SVS_GET_SNAP_SCHED, new HandlerEntry(ndmp_svs_get_snap_sched_request.class, ndmp_svs_get_snap_sched_reply.class));
        builder.put(NDMP_SVS_DELETE_SNAP_SCHED, new HandlerEntry(ndmp_svs_delete_snap_sched_request.class, ndmp_svs_delete_snap_sched_reply.class));
        builder.put(NDMP_SVS_CREATE_SNAP, new HandlerEntry(ndmp_svs_create_snap_request.class, ndmp_svs_create_snap_reply.class));
        builder.put(NDMP_SVS_GET_SNAP_STATUS, new HandlerEntry(ndmp_svs_get_snap_status_request.class, ndmp_svs_get_snap_status_reply.class));
        builder.put(NDMP_SVS_GET_CLIENT_STATUS, new HandlerEntry(ndmp_svs_get_client_status_request.class, ndmp_svs_get_client_status_reply.class));
        builder.put(NDMP_SVS_GET_ACCESS_INFO, new HandlerEntry(null, ndmp_svs_get_access_info_reply.class));
        builder.put(NDMP_SVS_SET_ACCESS_INFO, new HandlerEntry(ndmp_svs_set_access_info_request.class, ndmp_svs_set_access_info_reply.class));
        builder.put(NDMP_SVS_GET_INFO, new HandlerEntry(null, ndmp_svs_get_info_reply.class));
        builder.put(NDMP_SVS_ABORT_SNAP, new HandlerEntry(ndmp_svs_abort_snap_request.class, ndmp_svs_abort_snap_reply.class));
        builder.put(NDMP_SVS_RELEASE_RELATIONSHIP, new HandlerEntry(ndmp_svs_release_relationship_request.class, ndmp_svs_release_relationship_reply.class));
        builder.put(NDMP_SVS_GET_DATA_ADDR, new HandlerEntry(ndmp_svs_get_data_address_request.class, ndmp_sv_get_data_address_reply.class));

        builder.put(NDMP_SVP_SET_SNAP_SCHED, new HandlerEntry(ndmp_svp_set_snap_sched_request.class, ndmp_svp_set_snap_sched_reply.class));
        builder.put(NDMP_SVP_GET_SNAP_SCHED, new HandlerEntry(ndmp_svp_get_snap_sched_request.class, ndmp_svp_get_snap_sched_reply.class));
        builder.put(NDMP_SVP_GET_SNAP_STATUS, new HandlerEntry(ndmp_svp_get_snap_status_request.class, ndmp_svp_get_snap_status_reply.class));
        builder.put(NDMP_SVP_DELETE_SNAP_SCHED, new HandlerEntry(ndmp_svp_delete_snap_sched_request.class, ndmp_svp_delete_snap_sched_reply.class));
        builder.put(NDMP_SVP_CREATE_SNAP, new HandlerEntry(ndmp_svp_create_snap_request.class, ndmp_svp_create_snap_reply.class));
        builder.put(NDMP_SVP_ABORT_TRANSFER, new HandlerEntry(ndmp_svp_abort_transfer_request.class, ndmp_svp_abort_transfer_reply.class));
        builder.put(NDMP_SVP_RELEASE_RELATIONSHIP, new HandlerEntry(ndmp_svp_release_relationship_request.class, ndmp_svp_release_relationship_reply.class));
        builder.put(NDMP_SVP_GET_ACCESS_INFO, new HandlerEntry(null, ndmp_svp_get_access_info_reply.class));
        builder.put(NDMP_SVP_SET_ACCESS_INFO, new HandlerEntry(ndmp_svp_set_access_info_request.class, ndmp_svp_set_access_info_reply.class));
        builder.put(NDMP_SVP_GET_INFO, new HandlerEntry(null, ndmp_svp_get_info_reply.class));
        builder.put(NDMP_SVP_GET_DATA_ADDR, new HandlerEntry(ndmp_svp_get_data_address_request.class, ndmp_sv_get_data_address_reply.class));

        builder.put(NDMP_SM_DEFINE_RELATIONSHIP, new HandlerEntry(ndmp_sm_define_relationship_request.class, ndmp_sm_reply.class));
        builder.put(NDMP_SM_DELETE_RELATIONSHIP, new HandlerEntry(ndmp_sm_delete_relationship_request.class, ndmp_sm_reply.class));
        builder.put(NDMP_SM_START_RELATIONSHIP, new HandlerEntry(ndmp_sm_start_relationship_request.class, ndmp_sm_reply.class));
        builder.put(NDMP_SM_TRANSFER_UPDATE, new HandlerEntry(ndmp_sm_transfer_update_request.class, ndmp_sm_reply.class));
        builder.put(NDMP_SM_ABORT_TRANSFER, new HandlerEntry(ndmp_sm_abort_transfer_request.class, ndmp_sm_reply.class));
        builder.put(NDMP_SM_QUIESCE_RELATIONSHIP, new HandlerEntry(ndmp_sm_quiesce_relationship_request.class, ndmp_sm_reply.class));
        builder.put(NDMP_SM_RESUME_RELATIONSHIP, new HandlerEntry(ndmp_sm_resume_relationship_request.class, ndmp_sm_reply.class));
        builder.put(NDMP_SM_BREAK_RELATIONSHIP, new HandlerEntry(ndmp_sm_break_relationship_request.class, ndmp_sm_reply.class));
        builder.put(NDMP_SM_RESTART_RELATIONSHIP, new HandlerEntry(ndmp_sm_restart_relationship_request.class, ndmp_sm_reply.class));
        builder.put(NDMP_SM_RELEASE_RELATIONSHIP, new HandlerEntry(ndmp_sm_release_relationship_request.class, ndmp_sm_reply.class));
        builder.put(NDMP_SM_DEFINE_CONNECTION, new HandlerEntry(ndmp_sm_define_connection_request.class, ndmp_sm_reply.class));
        builder.put(NDMP_SM_DELETE_CONNECTION, new HandlerEntry(ndmp_sm_delete_connection_request.class, ndmp_sm_reply.class));
        builder.put(NDMP_SM_GET_CONNECTION_STATUS, new HandlerEntry(ndmp_sm_get_connection_status_request.class, ndmp_sm_get_connection_status_reply.class));
        builder.put(NDMP_SM_GET_STATUS, new HandlerEntry(ndmp_sm_get_status_request.class, ndmp_sm_get_status_reply.class));
        builder.put(NDMP_SM_GET_RELATIONSHIP_STATUS, new HandlerEntry(ndmp_sm_get_relationship_status_request.class, ndmp_sm_get_relationship_status_reply.class));
        builder.put(NDMP_SM_GET_INFO, new HandlerEntry(null, ndmp_sm_get_info_reply.class));

        builder.put(NDMP_SVS_GET_STATUS, new HandlerEntry(ndmp_svs_get_status_request.class, ndmp_svs_get_status_reply.class));
        builder.put(NDMP_SVS_RESTART_RELATIONSHIP, new HandlerEntry(ndmp_svs_restart_relationship_request.class, ndmp_svs_restart_relationship_reply.class));
        builder.put(NDMP_SVS_CREATE_RELATIONSHIP, new HandlerEntry(ndmp_svs_create_relationship_request.class, ndmp_svs_create_relationship_reply.class));
        builder.put(NDMP_SVS_TRANSFER_BACKUP, new HandlerEntry(ndmp_svs_transfer_backup_request.class, ndmp_svs_transfer_backup_reply.class));
        builder.put(NDMP_SNAP_GET_QTREE_CONTENTS, new HandlerEntry(ndmp_snap_qtree_contents_request.class, ndmp_snap_qtree_contents_reply.class));
        builder.put(NDMP_SVP_GET_STATUS, new HandlerEntry(ndmp_svp_get_status_request.class, ndmp_svp_get_status_reply.class));
        builder.put(NDMP_SVP_RECOVER_BACKUP, new HandlerEntry(ndmp_svp_recover_backup_request.class, ndmp_svp_recover_backup_reply.class));
        HANDLER_MAP_PROTOTYPE = builder.build();
    }
}
