/*
 * $Id: //depot/prod/newport_news/main/modules/jndmp/src/main/java/com/netapp/ndmp/NdmpSession.java#6 $
 *
 * Copyright (c) 2008 Network Appliance, Inc.
 * All rights reserved.
 */

package com.netapp.ndmp;

import static com.netapp.ndmp.rpcgen.ndmp_message.*;

import java.nio.charset.Charset;
import java.security.MessageDigest;
import java.security.NoSuchAlgorithmException;
import java.util.HashMap;
import java.util.Map;

import com.google.common.collect.ImmutableSet;

import org.acplt.oncrpc.XdrAble;

import com.netapp.ndmp.rpcgen.ndmp_auth_data;
import com.netapp.ndmp.rpcgen.ndmp_auth_md5;
import com.netapp.ndmp.rpcgen.ndmp_auth_text;
import com.netapp.ndmp.rpcgen.ndmp_auth_type;
import com.netapp.ndmp.rpcgen.ndmp_class_list;
import com.netapp.ndmp.rpcgen.ndmp_class_version;
import com.netapp.ndmp.rpcgen.ndmp_config_get_auth_attr_reply;
import com.netapp.ndmp.rpcgen.ndmp_config_get_auth_attr_request;
import com.netapp.ndmp.rpcgen.ndmp_config_get_ext_list_reply;
import com.netapp.ndmp.rpcgen.ndmp_config_get_server_info_reply;
import com.netapp.ndmp.rpcgen.ndmp_config_set_ext_list_reply;
import com.netapp.ndmp.rpcgen.ndmp_config_set_ext_list_request;
import com.netapp.ndmp.rpcgen.ndmp_connect_client_auth_reply;
import com.netapp.ndmp.rpcgen.ndmp_connect_client_auth_request;
import com.netapp.ndmp.rpcgen.ndmp_connect_open_reply;
import com.netapp.ndmp.rpcgen.ndmp_connect_open_request;
import com.netapp.ndmp.rpcgen.ndmp_log_message_post;


public class NdmpSession {
    public static final short SNAP_EXT_CLASS = 0x2050;
    public static final short CAB_EXT_CLASS = 0x2052;

    private static final int NDMP_PORT = 10000;
    private static final int MINIMUM_NDMP_PROTO_VERSION = 4;
    private static final short SNAP_EXT_MIN_VERSION = 2;
    private static final short SNAP_EXT_MAX_VERSION = 9;

    private final Map<Short, Short> extensionVersions;
    private NdmpHandlerTable handlerTable;
    private NdmpConnection connection;


    public NdmpSession(String hostname, int port, int connectTimeout, int readTimeout) {
        extensionVersions = new HashMap<Short, Short>();
        handlerTable = new NdmpHandlerTable();
        connection = new NdmpConnection(hostname, port, connectTimeout, readTimeout);
    }

    public NdmpSession(String hostname, int connectTimeout, int readTimeout) {
        this(hostname, NDMP_PORT, connectTimeout, readTimeout);
    }

    public short getExtensionVersion(short extension) {
        Short version = extensionVersions.get(extension);
        if (version == null) {
            return 0;
        }
        return version;
    }

    /**
     * Establishes the NDMP connection without any extensions
     *
     * @param login NDMP login
     * @param password NDMP password
     * @throws NdmpException
     */
    public void connect(String login, String password) throws NdmpException {
        connect(login, password, null);
    }

    /**
     * Establishes the NDMP connection with an extension list
     *
     * @param login NDMP login
     * @param password NDMP password
     * @param extensionList Extension list to be used
     * @throws NdmpException
     */
    public void connect(String login, String password, ImmutableSet<Short> extensionList) throws NdmpException {
        connection.connect();
        ndmp_connect_open_request request = new ndmp_connect_open_request();
        request.protocol_version = MINIMUM_NDMP_PROTO_VERSION;
        sendMessage(NDMP_CONNECT_OPEN, request, new ndmp_connect_open_reply());
        login(login, password);
        negotiateExtensions(extensionList);

        if (connection.getProtocolVersion() < MINIMUM_NDMP_PROTO_VERSION) {
            close();
            throw new NdmpProtocolException("Server NDMP version too old");
        }
    }

    /**
     * Closes the connection with the remote host.  Once closed, this session cannot be reconnected.  This operation
     * is idempotent.
     */
    public void close() {
        connection.close();
    }

    /**
     * Send a request to the remote NDMP host and return the response.
     * TODO Overload this method with variants which do not take a request/reply to avoid these nulls
     *
     * @param messageCode message code corresponding to the NDMP operation.
     * @param request The NDMP request.
     * @param reply The NDMP reply.
     * @return The NDMP reply
     * @throws NdmpException
     */
    public <T extends XdrAble> T sendMessage(int messageCode, XdrAble request, T reply) throws NdmpException {
        if (!handlerTable.isValidRequest(messageCode)) {
            throw new NdmpProtocolException("Unknown message code: %d", messageCode);
        }

        Class<? extends XdrAble> requestClass = handlerTable.getRequestClass(messageCode);
        if (requestClass != null) {
            if (request != null) {
                if (!requestClass.isInstance(request)) {
                    throw new IllegalArgumentException(String.format("Expected reply of type %s, got %s", requestClass.getName(), request.getClass().getName()));
                }
            } else {
                throw new NullPointerException(String.format("Expected request of type %s, got null", requestClass.getName()));
            }
        } else if (null != request) {
            throw new IllegalArgumentException("Request must be null");
        }
        Class<? extends XdrAble> replyClass = handlerTable.getReplyClass(messageCode);
        if (replyClass != null) {
            if (reply != null) {
                if (!replyClass.isInstance(reply)) {
                    throw new IllegalArgumentException(String.format("Expected reply of type %s, got %s", replyClass.getName(), reply.getClass().getName()));
                }
            } else {
                throw new NullPointerException(String.format("Expected reply of type %s, got null", replyClass.getName()));
            }
        } else if (reply != null) {
            throw new IllegalArgumentException("Reply must be null");
        }

        connection.sendMessage(messageCode, request, reply);
        return reply;
    }

    /**
     * Wait for a specific notification.
     *
     * @param messageCode Expecting notification type.
     * @param clazz Expecting notification message class type
     * @return Received notification message
     * @throws NdmpException
     */
    public <T extends XdrAble> T waitForAndProcessNotification(int messageCode, Class<T> clazz) throws NdmpException {
        return connection.waitForAndProcessNotification(messageCode, clazz);
    }

    /**
     * Wait for a specific notification while processing ndmp log messages.
     *
     * @param messageCode Expecting notification type.
     * @param clazz Notification message class type
     * @param logMessage Ndmp log message received from the ndmp server.
     * @return Received notification message. When a log message was received, 'null' is returned.
     * @throws NdmpException
     */
    public <T extends XdrAble> T waitForAndProcessNotification(int messageCode, Class<T> clazz, ndmp_log_message_post logMessage) throws NdmpException {
        return connection.waitForAndProcessNotificationWithLog(messageCode, clazz, logMessage);
    }

    private void login(String login, String password) throws NdmpException {
        ndmp_config_get_server_info_reply serverInfoReply = new ndmp_config_get_server_info_reply();
        sendMessage(NDMP_CONFIG_GET_SERVER_INFO, null, serverInfoReply);
        boolean md5Auth = false;
        if (serverInfoReply.auth_type != null) {
            for (int type : serverInfoReply.auth_type) {
                if (type == ndmp_auth_type.NDMP_AUTH_MD5) {
                    md5Auth = true;
                }
            }
        }

        ndmp_connect_client_auth_request authRequest = new ndmp_connect_client_auth_request();
        authRequest.auth_data = new ndmp_auth_data();

        if (md5Auth) {
            ndmp_config_get_auth_attr_request authAttrRequest = new ndmp_config_get_auth_attr_request();
            ndmp_config_get_auth_attr_reply authAttrReply = new ndmp_config_get_auth_attr_reply();

            authAttrRequest.auth_type = ndmp_auth_type.NDMP_AUTH_MD5;
            sendMessage(NDMP_CONFIG_GET_AUTH_ATTR, authAttrRequest, authAttrReply);
            if (authAttrReply.server_attr.auth_type != ndmp_auth_type.NDMP_AUTH_MD5) {
                throw new NdmpException(String.format("Unexpected auth type: %d", authAttrReply.server_attr.auth_type));
            }
            authRequest.auth_data.auth_type = ndmp_auth_type.NDMP_AUTH_MD5;
            authRequest.auth_data.auth_md5 = new ndmp_auth_md5();
            authRequest.auth_data.auth_md5.auth_id = login;

            byte[] pwBytes = password.getBytes(Charset.forName("UTF-8"));
            int pwLen = Math.min(pwBytes.length, 32);
            byte[] buf = new byte[128];
            System.arraycopy(pwBytes, 0, buf, 0, pwLen);
            System.arraycopy(authAttrReply.server_attr.auth_md5_attr.challenge, 0, buf, 64 - pwLen, 64);
            System.arraycopy(pwBytes, 0, buf, 128 - pwLen, pwLen);
            MessageDigest md5;
            try {
                md5 = MessageDigest.getInstance("MD5");
            } catch (NoSuchAlgorithmException e) {
                throw new RuntimeException(e);
            }
            authRequest.auth_data.auth_md5.auth_digest = md5.digest(buf);
        } else {
            authRequest.auth_data.auth_type = ndmp_auth_type.NDMP_AUTH_TEXT;
            authRequest.auth_data.auth_text = new ndmp_auth_text();
            authRequest.auth_data.auth_text.auth_id = login;
            authRequest.auth_data.auth_text.auth_password = password;
        }

        sendMessage(NDMP_CONNECT_CLIENT_AUTH, authRequest, new ndmp_connect_client_auth_reply());
    }

    private void negotiateExtensions(ImmutableSet<Short> extensionList) throws NdmpException {

        if (extensionList == null || extensionList.size() == 0) {
            return;
        }

        ndmp_config_get_ext_list_reply getReply = new ndmp_config_get_ext_list_reply();
        getReply = connection.sendMessage(NDMP_CONFIG_GET_EXT_LIST, null, getReply);

        for (ndmp_class_list l : getReply.class_list) {
            if (extensionList.contains(l.ext_class_id)) {
                short highest = 0;
                for (short version : l.ext_version) {
                    if (l.ext_class_id == SNAP_EXT_CLASS && version > NdmpSession.SNAP_EXT_MAX_VERSION) {
                        continue;
                    }
                    if (version > highest) {
                        highest = version;
                    }
                }
                if (l.ext_class_id == SNAP_EXT_CLASS && highest < SNAP_EXT_MIN_VERSION) {
                    throw new NdmpProtocolException("Snap extension not found");
                }
                extensionVersions.put(l.ext_class_id, highest);
            }

        }

        if (extensionVersions.size() > 0) {
            ndmp_config_set_ext_list_request setRequest = new ndmp_config_set_ext_list_request();
            setRequest.ndmp_selected_ext = new ndmp_class_version[extensionVersions.size()];
            int i = 0;
            for (short extension : extensionVersions.keySet()) {
                setRequest.ndmp_selected_ext[i] = new ndmp_class_version();
                setRequest.ndmp_selected_ext[i].ext_class_id = extension;
                setRequest.ndmp_selected_ext[i].ext_version = extensionVersions.get(extension);
                i++;
            }

            sendMessage(NDMP_CONFIG_SET_EXT_LIST, setRequest, new ndmp_config_set_ext_list_reply());
            handlerTable.setExtensionVersions(extensionVersions);
        }
    }

    /**
     * For testing purposes only.
     * @param connection
     */
    void setConnection(NdmpConnection connection) {
        this.connection = connection;
    }
}
