package com.netapp.offtap.ndmpcopy.debuglog;

import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.PrintWriter;
import java.io.RandomAccessFile;
import java.text.SimpleDateFormat;
import java.util.Date;

import com.netapp.offtap.ndmpcopy.util.NdmpCons;
import com.netapp.offtap.ndmpcopy.util.NdmpPseudoStream;


/**
 * All log mesages which the server prints out are sent through this class
 */
public class NdmpDebug implements NdmpDebugLog {

	/**
	 * The verbose level indicates the amount of debug information to 
	 * generate.  The general guidelines are:
	 *
	 * 	 0 : All major error messages that are fatal to the server
	 *  10 : High level information such as connection opened, closed, etc...
	 *  20 : 
	 *  30 : Detailed information such as the messages received and their args
	 *  40 :
	 *  50 : Function calls and stuff
	 */
	protected int verbose = 50;

	public	NdmpPseudoStream	out;

	/* Default log names and the number of days to keep them around */
	protected static String logfilename = "ndmpcopy.";
	protected static int logoldestday = 15;

	/* Because of the way the java output files work, we need some funky
	 * synchronization on a single output stream to make sure we catch
	 * all of the debug messages.  Otherwise, multiple threads running
	 * at the same time can trample each other's log messages.  So, if we
	 * dump with debug on, things will be verrrry slow.
	 *
	 * The hashtable is used to store the number of sessions that have a 
	 * particular filehandle open.
	 */
	private static String curlogfilename = null;
	private static PrintWriter curlogfile = null;
	private static Object	fileMutex = new Object();
	private static RandomAccessFile curlogfile_rf;

	private static long 	lastlog = 0;
	private static final int timeout = 1000;	/* 1 second */

	private int sessionid;
	private boolean cleaned;
        private static int LogFileOpenCount = 0;

	/**
	 * Basic constructor.  Each NdmpDebug is associated with a corresponding
	 * Ndmpd object and since all of the NdmpDebug objects print information
	 * out to the same file, the id is used to separate the lines.
	 * The constructor is protected so that it cannot be invoked outside of
	 * the package.
	 *
	 * @param filepath The directory in which to create the log file.  If the
	 *         directory is invalid, then the output will not be stored
	 *         in a log file, but will still be printed out to the console
	 * @param id The session id of the Ndmpd session
	 */
	protected NdmpDebug(String filepath, int id, int verbose) {
		getlogfile(filepath);
		this.sessionid = id;
		this.cleaned = false;
		this.verbose = verbose;
	}

	protected NdmpDebug(String filepath, int id, int verbose, Object out) {
		getlogfile(filepath);
		this.sessionid = id;
		this.cleaned = false;
		this.verbose = verbose;
		this.out = (NdmpPseudoStream) out;
	}

	protected void finalize () throws Throwable {
		super.finalize();
	}

	/**
	 * Cleanup function.  This should be called before the object gets
	 * garbage collected because the underlying file is a RandomAccessFile,
	 * which doesn't close its file descriptor when it closes. This
         * will be called during session cleanup.
	 */
	public void cleanup () {
           closeLogFile(false);
	}

	/**
	 * The file names are generated on a per-day basis.  This function 
	 * returns 'next' log file name and deletes any old logs
	 */
	private String getNextFileName (String path) {
		Date d;
		String filename;
		String[] filelist;
		File dirFile, tempFile;
		int i, k;

		if (path == null)
			return null;

		d = new Date ();
		k = Integer.parseInt((new SimpleDateFormat("yyyyMMdd")).format(d));
		filename = path + logfilename + (new Integer(k)).toString();

		dirFile = new File(path);

		if (dirFile.isDirectory() == false) 
			return null;
		
		filelist = dirFile.list();

		/* Delete logs older than logoldestday logs */
		k = k - logoldestday;
		if (k % 100 > 31)
			k -= 68;
		if (k % 10000 < 100)
			k -= 8800;

		for (i = 0; i < filelist.length; i++) {
			try {
				if ((filelist[i].substring(0, logfilename.length())).equals(
						logfilename)) {
					if (Integer.parseInt(filelist[i].substring(logfilename.length(), logfilename.length() + 8)) < k) {
						tempFile = new File(path + filelist[i]);
						tempFile.delete();
					}
				}
			} catch (Exception e) {
			}
		}
		return filename;
	}

	/**
	 * Opens up a given file and sets it up for writing
	 */
	private void  getlogfile (String filepath) {
		String filename = null;
		synchronized (fileMutex) {
			try {
				RandomAccessFile rf;

				filename = getNextFileName(filepath);
				PrintWriter  pwriter = null; //temp ref.
				
				if (filename == null) 
					return ;

                                /*
                                 * Check if the next filename matches the current. If it does
                                 * not match then close the existing and open a new one.
                                 * Otherwise, check if the current filename actually exists
                                 * and recreate it if necessary.
                                 */
                                if (curlogfilename == null) {
                                        openLogfile(filename);
                                } else if (!filename.equals(curlogfilename)) {
                                        /*
                                         * close the current file and don't decrement the open counter.
                                         */
                                        closeLogFile(true);
                                        /*
                                         * Open the new file.
                                         */
                                        openLogfile(filename);
                                } else {
                                        // curlogfilename exists, and does not require to be changed
                                        // Check whether file actually exists. (burt 46874)
                                        File file = new File(curlogfilename);
                                        if(!file.exists()) {
                                                try {
                                                        closeLogFile(true); // Free up previous one
                                                } catch (Exception ignored) {}
                                                openLogfile(filename);
                                        } else {
                                                LogFileOpenCount++;
                                        }
                                }
				return ;
			} catch (IOException e) {
				if (verbose >= 100)
					e.printStackTrace(System.out);
				/*
				 * silent failure when we can't open the log file?
				 * Stick to the old file.
				 */
				println (50, "Failed to create the log file "+filename);	
			}
			return ;
		}
	}

        /**
         * Opens an existing log file for writing.
         * Creates new file, printwriter if file does not exist.
         * @param filename Full pathname of file to create.
         */
        private void openLogfile(String filename) throws IOException
        {
                PrintWriter  pwriter = null; //temp ref.
                RandomAccessFile rf = new RandomAccessFile (filename, "rw");
                rf.seek(rf.length());
                pwriter = new PrintWriter (new FileOutputStream (rf.getFD()));
                if (null != pwriter) {
                        curlogfilename = filename;
                        curlogfile = pwriter;
                        curlogfile_rf = rf;
                        pwriter = null;
                }
                synchronized (fileMutex) {
                   LogFileOpenCount++;
                }
        }

        /**
         * Closes a logfile if the use count is 0.
         */
        private void closeLogFile(boolean forceClose)
        {
                if ((curlogfile == null) || (LogFileOpenCount ==0)) {
                   return;
                }
                synchronized (fileMutex) {
                   try {
                      if (forceClose == false) {
                         LogFileOpenCount--;
                      }
                      if ((LogFileOpenCount <= 0) || (forceClose == true)) {
                           /*
                           ** Flush the existing stream.
                           */
                           curlogfile.flush();
                           /*
                           ** Close the existing stream.
                           */
                           curlogfile.close();
                           /*
                           ** Close the existing underlying random access file.
                           */
                           curlogfile_rf.close();
                           curlogfilename = null;
                           curlogfile     = null;
                           curlogfile_rf  = null;
                      }
                   } catch (IOException ioe) {
                      System.out.println("Error closing logfile stream: " + ioe);
                   }
                }
        }

	/**
	 * Cause an explicit flush of the log file.
	 */
	public void flush () {
		synchronized (fileMutex) {
			if (curlogfile != null) {
				curlogfile.flush();
                         }
		}
	}

	/**
	 * Prints out a line to the debug log.  This defaults the level of 
	 * line to 0 (which means it always gets printed).
	 */
	public void println (String s) {
		println(0, s);
	}

	/**
	 * The more commonly used interface to print to the debug log.  
	 *
	 * @param level If the level is lower than the verbose level, then print
	 *         out the line, otherwise, ignore it
	 * @param s The string to be printed out
	 */
	public void println (int level, String s) {
		Date d;
		long t;
		String ts;
		SimpleDateFormat sdf;

		if (level > verbose)
			return;

		d = new Date ();
		t = d.getTime();
		sdf = new SimpleDateFormat("MMM dd HH:mm:ss z");
		ts = sdf.format(d) + " [ndmpc:" + sessionid + "]: " + s;

		synchronized (fileMutex) {
			if (curlogfile == null)
				return;
			curlogfile.println(ts); 
			if (t > lastlog + timeout) {
				curlogfile.flush();
				lastlog = t;
			}
		}
	}

	/**
	 * Prints out the information in the Exception or Error.  This interface
	 * defaults the level to 0 (which means it always gets printed).
	 *
	 * @param e The throwable object to be printed
	 * @param s A string prefix which gets printed before the throwable 
	 *         object's internal message is printed out.
	 */
	public void println (Throwable e, String s) {
		println(0, e, s);
	}

	/**
	 * Prints out the information in the Exception or Error.  The level 
	 * indicates the importance of this message.
	 *
	 * @param level If the level is lower than the verbose level, then print
	 *         out the line, otherwise, ignore it
	 * @param e The throwable object to be printed
	 * @param s A string prefix which gets printed before the throwable 
	 *         object's internal message is printed out.
	 */
	public void println (int level, Throwable e, String s) {
		
		if (level > verbose)
			return;

		println(s);
		if (e == null) {
			return;
		}

		println(e.getMessage());
		
		synchronized (fileMutex) {
			if (curlogfile == null)
				return;
			e.printStackTrace(curlogfile);	
		}
		return;
	}

	public static void dbgPrintln (String logmsg) {
		dbgPrintln(logmsg, "");
	}

	// prints to console only for DEBUG builds //
	public static void dbgPrintln (String logmsg, String val) {
		String printStr = "ndmpcopy_DEBUG : ";

		if ((NdmpCons.NDMPCOPY_DEBUG == true) 
				/* &&	ONTAP.isDebugKernel()*/ ) {
			if (null != logmsg) {
				if (null == val) {
					val = "null";
				}
				printStr = printStr + logmsg + val;
			}
			System.out.println(printStr);
		}
		System.out.flush();
	}
}
