package com.netapp.offtap.ndmpcopy.messagereader;

import com.netapp.offtap.ndmpcopy.messages.NdmpMessage;
import com.netapp.offtap.ndmpcopy.ndmpcopycmd.NdmpCopyCmdArgs;
import com.netapp.offtap.ndmpcopy.util.NdmpCons;
import com.netapp.offtap.ndmpcopy.util.NdmpTimeOutException;

//import netapp.cmds.ndmpc.messagereader.*;


/**
 * Internal queue structure used to pipe multiple messages from several ndmpd
 * sessions to a single control thread.
 */
public class NdmpMessageQueue implements NdmpMessageQueueIn, 
						NdmpMessageQueueOut {
	
	private NdmpMsgQueue	msgQueue[];
	private	volatile int	qState;

	private	static final	int		INVALID = 0;
	private	static	final	int		VALID	= 1;
	
	private	static	final	int		QTIME_OUT	= 1000000;

	/**
	 * Constructs a message queue
	 * <p>
	 * @param len The maximum size of the given queue.  
	 */
	public NdmpMessageQueue (int len) {
		msgQueue = new NdmpMsgQueue [2];
		for (int i = 0; i < 2; i++)
			msgQueue [i] = new NdmpMsgQueue (len);
		qState = VALID;
	}

	/**
	 * Gets the next object from the message queue and doesn't block
	 * <p>
	 * @return Returns the next object in the queue.
	 */
	public QueueNode getNoSleep () {
		QueueNode data = null;
		QNode	qNode = null;
		int		id = NdmpCopyCmdArgs.srcFiler;


		qNode = readNoSleep (NdmpCopyCmdArgs.srcFiler);

		if (null == qNode) {
			qNode = readNoSleep (NdmpCopyCmdArgs.dstFiler);
			id	= NdmpCopyCmdArgs.dstFiler;
		}

		if (null == qNode || null == qNode.msg)
			return null;
		
		data = new QueueNode (qNode.msg, id);
		
		synchronized (this) {
			this.notifyAll ();
		}
		
		return data;
	}

	/**
	 * Gets the next object from the message queue and doesn't block
	 * <p>
	 * @param filer  The filer ID (source/Destination)
	 * @return Returns the next object in the queue.
	 */
	public NdmpMessage getNoSleep (int filer) {
		QNode qNode = readNoSleep (filer);
	
		if (null == qNode)
			return null;
		
		NdmpMessage data = qNode.msg;

		synchronized (this) {
			this.notifyAll ();
		}

		return data;
	}

	protected  QNode readNoSleep (int filer) {
		QNode 		qNode = null;

		qNode = msgQueue [filer].getNoSleep ();

		if (null == qNode)
			return null;
			
		return qNode;
	}


	/**
	 * Puts in an object into the queue and doesn't block
	 * <p>
	 * @param data The data element to put into the queue
	 * @return Whether the put operation succeeded
	 */
	public boolean putNoSleep (QueueNode qNode) {

		if (!msgQueue [qNode.readerID].putNoSleep (qNode.msg))
			return false;

		synchronized (this) {
			this.notifyAll ();
		}
		
		return true;
	}

	/**
	 * Puts in an object into the queue and doesn't block
	 * <p>
	 * @param data The data element to put into the queue
	 * @param filer filer index
	 * @return Whether the put operation succeeded
	 */
	public boolean putNoSleep (NdmpMessage msg, int filer) {

		if (!msgQueue [filer].putNoSleep (msg))
			return false;

		synchronized (this) {
			this.notifyAll ();
		}
		
		return true;
	}
	/**
	 * Gets an object from the queue and blocks the thread if queue is empty.
	 * The thread is woken up when an object has been placed into the queue.
	 * <p>
	 * @return The next object in the queue.
	 */
	public QueueNode get () throws NdmpTimeOutException {
		QueueNode	data = null;
		int		retry = 0;
		
		while (VALID == qState && (null == (data = getNoSleep ()))) {
			synchronized (this) {
				try {
					this.wait (QTIME_OUT);	// 1000 sec
				}catch (InterruptedException ignored) {
				}
			}
			
			if (++retry >= NdmpCons.MAX_RETRY_COUNT)
				throw new NdmpTimeOutException (" No response from server," +
					" timed out (Empty Queue)");
		}		
		return data;
	}

	/**
	 * Gets an object from the queue and blocks the thread if queue is empty.
	 * The thread is woken up when an object has been placed into the queue.
	 * <p>
	 * @param	filer The filer index, source/destination
	 * @return The next object in the queue.
	 */
	public NdmpMessage get (int filer) throws NdmpTimeOutException {
		NdmpMessage	data = null;
		int		retry = 0;
		
		while (VALID == qState && (null == (data = getNoSleep (filer)))) {
			synchronized (this) {
				try {
					this.wait (QTIME_OUT);	// 10 sec
				}catch (InterruptedException ignored) {
				}
			}
			if (++retry >= NdmpCons.MAX_RETRY_COUNT)
				throw new NdmpTimeOutException (" No response from server," +
					" timed out (Empty Queue)");
					
		}
		return data;
	}		
	/**
	 * Puts an object into the queue.  If the queue is already at full 
	 * capacity, then the thread blocks until there is room in the queue.
	 * <p>
	 * @param data The queueNode data to insert into the queue.
	 */
	public void put (QueueNode  data) throws NdmpTimeOutException {
		int	retry = 0;

		if (null == data)
			return ;

		while (VALID == qState && (false == putNoSleep (data))) {
			synchronized (this) {
				try {
					this.wait (QTIME_OUT);
				}catch (InterruptedException ignored) {
				}
			}
			
			if (++retry >= NdmpCons.MAX_RETRY_COUNT)
				throw new NdmpTimeOutException (" No response from server, " +
					"time out");
		}
	}
	
	/**
	 * Puts an object into the queue.  If the queue is already at full 
	 * capacity, then the thread blocks until there is room in the queue.
	 * <p>
	 * @param data The NdmpMessage data into the queue.
	 * @param filer filer index
	 */
	public void put (NdmpMessage  data, int filer) throws NdmpTimeOutException {
		int	retry = 0;

		if (null == data || !(NdmpCopyCmdArgs.isValidFilerID (filer)))
			return ;

		while (VALID == qState && (false == putNoSleep (data, filer))) {
			synchronized (this) {
				try {
					this.wait (QTIME_OUT);
				}catch (InterruptedException ignored) {
				}
			}
			
			if (++retry >= NdmpCons.MAX_RETRY_COUNT)
				throw new NdmpTimeOutException (" No response from server, " +
					"time out");
		}
		
	}

	public void invalidateQueue () {
		qState = INVALID;
	}
}
