
package com.netapp.offtap.ndmpcopy.messagereader;

import java.io.IOException;

import com.netapp.offtap.ndmpcopy.debuglog.NdmpDebugLog;
import com.netapp.offtap.ndmpcopy.messages.NdmpMessage;
import com.netapp.offtap.ndmpcopy.messages.NdmpMessageFactory;
import com.netapp.offtap.ndmpcopy.ndmpcopycmd.NdmpCopyCmdArgs;
import com.netapp.offtap.ndmpcopy.streams.NdmpInput;
import com.netapp.offtap.ndmpcopy.util.NdmpTimeOutException;



public class NdmpMessageReader extends Thread {
	protected int readerID; // 0 = source, 1 = destination

	protected NdmpMessageQueueIn msgQueue;
	protected NdmpInput input;
	protected NdmpDebugLog debugLog;

	protected volatile NdmpMessageFactory msgFactory;

	protected volatile int threadState;
	protected static final int InitState = 0;
	protected static final int RunState = 1;
	protected static final int DeadState = 2;
	protected static final int IdleState = 3;
	
	public NdmpMessageReader (NdmpInput input, 
			NdmpMessageFactory msgFactory, 
			NdmpMessageQueueIn msgQueue, // Input Queue only
			NdmpDebugLog debugLog,
			int readerID) throws NdmpQueueException {

		super("ndmpmessagereader");

		if (null == msgQueue)
			throw new NdmpQueueException ("Null message queue");
			
		this. msgQueue = msgQueue;
		this.threadState = InitState;
		this.input = input;
		this.msgFactory = msgFactory;
		this.debugLog = debugLog;
		this.readerID = readerID;
	}

	public void changeMsgFactory (NdmpMessageFactory msgFactory) {
		this.msgFactory = msgFactory;
	}

	
	public void killThread () {
		synchronized (this) {
			threadState = DeadState;
		}
		debugLog.println(50, "Killing message reader thread for " +
			NdmpCopyCmdArgs.filer(readerID));

		this.interrupt ();
	}

	public void idleThread () {
		synchronized (this) {
		    /* Don not set the state to idle if the state is 
		     * already 'Dead' 
		     */
		    if( DeadState != threadState) {
			threadState = IdleState;
		    }
		}
		debugLog.println(50, "Setting message reader thread state to IdleState for " +
			NdmpCopyCmdArgs.filer(readerID));

		this.interrupt ();
	}

	/**
	 * readMessage is called only to get the NOTIFY_CONNECTED messages,
	 * in response to the socket connections to the ndmp servers.
	 */
	public synchronized NdmpMessage	readMessage () throws IOException {
		if (InitState != threadState) {
			debugLog.println(20, "readMessage called with Invalid State" +
				NdmpCopyCmdArgs.filer(readerID));
			throw new IOException ("readMessage called with Invalid State");
		}
		return readMessageReader ();
	}
	
	/**
	 * This is the method that actually reads messages from the inputstream.
	 */
	protected NdmpMessage readMessageReader () throws IOException {
		NdmpMessage msg = null;
		msg = new NdmpMessage ();
		input.resetStream ();
		debugLog.println(50, "Waiting to read message from " +
			NdmpCopyCmdArgs.filer(readerID));
		msg.read (input, msgFactory);
		debugLog.println(50, "Successfully read message from " +
			NdmpCopyCmdArgs.filer(readerID));
		return msg;
	}

	/**
	 * The run method is called only after the NOTIFY_CONNECTED messages are read
	 * by the NdmpCopyThread. If this thread dies for some reason, before the
	 * copy is completed, then the get* methods from the NdmpMessageQueues will 
	 * throw an NdmpTimeOutException. Till then ndmpcopy will look like it has hung.
	 */
	public void run () {
		NdmpMessage msg = null;
		
		try {
			debugLog.println(50, "Message reader for " +
				NdmpCopyCmdArgs.filer(readerID) + " started");
			while (DeadState != threadState) {
				synchronized (this) {
					if (InitState == threadState)
						threadState = RunState;
				}

				/* BURT 241035. Stop reading if the thread 
				 * is in the idle state
				 */
				if (IdleState == threadState) {
					Thread.sleep(50);
					continue;
				}

				try {
					msg = readMessageReader ();
				} catch (IOException e) {
					// If the threadstate is still not dead or idle, that means
					// we have a genuine problem while reading a message.
					// Only then do we log this to the debuglog file in
					// nondebug mode (20). 
					if ((threadState != DeadState) &&
					    (threadState != IdleState)) {
						debugLog.println(20, e, "Exception while " +
							"trying to read message from " +
							NdmpCopyCmdArgs.filer(readerID));
					} 
					if (IdleState != threadState) {
					    synchronized (this) {
							threadState = DeadState;
					    }
					    return ;
					}
				}

				if ((null == msg) || (IdleState == threadState))
					continue ;
				
				QueueNode	node = new QueueNode (msg, readerID);
				
				boolean	msgadded = false;
				while (RunState == threadState && !msgadded) {
					try {
						msgQueue.put (node);
						msgadded = true;
						synchronized (msgQueue) {
							msgQueue.notifyAll ();
						}	
					}catch (NdmpTimeOutException e) {
						// XXX: Change this throw statement if we want to
						// accept the timeout error and try again...
						synchronized (this) {
							threadState = DeadState;
						}
						debugLog.println(20, e, "Timed out waiting " +
							"to queue message received from " +
							NdmpCopyCmdArgs.filer(readerID));
						return ;
						//throw new NdmpFatalException (e.getMessage ());
					}
				}
			}	
			debugLog.println(50, "Message reader for " +
				NdmpCopyCmdArgs.filer(readerID) + " done");
			return ;
		} catch (Exception e) {
			debugLog.println(20, e, "Exception in message reader for " +
				NdmpCopyCmdArgs.filer(readerID));
		}
	}
}	
