package com.netapp.offtap.ndmpcopy.messages;

import java.io.*;
import java.util.*;

import com.netapp.offtap.ndmpcopy.*;
import com.netapp.offtap.ndmpcopy.debuglog.NdmpDebugLog;
import com.netapp.offtap.ndmpcopy.messages.*;
import com.netapp.offtap.ndmpcopy.streams.*;
import com.netapp.offtap.ndmpcopy.util.NdmpCons;

/**
 * The only variables in this class deal with the
 * header of the message. 
 */
public class NdmpMessage {
	
	/* 
	 * The header data structure  
	 */
	protected int sequence;			// Sequence number begins at 0 
	protected int timestamp;		// Time the message was sent 
	protected int msgtype;			// Request or Reply message type 
	protected int method;			// What this message does 
	protected int replysequence;	// Sequence of request we reply to 
	protected int error;			// Errors in reading message 

	protected NdmpMessageData	msgData;	// Message Body 


	/**
	 * Default constructor for creating a <code>NDMP_MESSAGE_REQUEST</code>
	 * message.  The constructor is normally invoked by its sub public classes.
	 */
	public NdmpMessage () {
		msgtype = NdmpCons.NDMP_MESSAGE_REQUEST;
		replysequence = 0;
		error = NdmpCons.NDMP_NO_ERR;

		msgData = null;
	}

	/**
	 * Const'r for creating a new request/reply message to post.
	 * Used by Factory only.
	 */
	public NdmpMessage (int method, int sequence, int replysequence,
			NdmpMessageData msgData, int msgtype) { 

		this.msgtype = msgtype;
		this.method = method;
		this.sequence = sequence;
		this.replysequence = replysequence;
		this.error = NdmpCons.NDMP_NO_ERR;	// since its a creation 

		this.msgData = msgData;	
	}

	public NdmpMessage (NdmpMessage msg, NdmpMessageData msgData) {
		this.msgtype	= NdmpCons.NDMP_MESSAGE_REPLY;
		this.method		= msg.method;
		this.error		= msg.error;

		this.replysequence	= msg.sequence;
		this.msgData		= msgData;
	}

	public NdmpMessage (NdmpMessage msg, NdmpMessageFactory msgFactory,
						Class[] argsType, Object[] args) {
		this.msgtype	= NdmpCons.NDMP_MESSAGE_REPLY;
		this.method		= msg.method;
		this.error		= msg.error;

		this.replysequence	= msg.replysequence;
		this.msgData		= msgFactory.createMessageData (method, 
						argsType, args, NdmpCons.NDMP_MESSAGE_REPLY);
		//XXX: this could possibly be null
	}

	/**
	 *  Read in the header, which is constant across all NDMP messages
	 *  A lot of the fields in the header have fancy names, but
	 *  they're just enumerated types for unsigned_longs. So, we
	 *  read them in as such, but in the MessageHandler, we'll treat
	 *  them as the appropriate enumerated types. This saves us some
	 *  unnecessary complexity in the XDR code, I think.
	 * <p>
	 * Message structure:
	 * <pre>
	 *  u_long							sequence
	 *  u_long							time_stamp
	 *  ndmp_header_message_type (int)	message_type
	 *  ndmp_message (int)				method
	 *  u_long							reply_sequence
	 *  ndmp_error (int)				error
	 * </pre>
	 * @param in The control input stream from which to read the rest
	 *         of the message
	 */
	public void read (NdmpInput input, NdmpMessageFactory msgFactory ) 
			throws IOException {

		sequence = input.getLong();
		timestamp = input.getLong();
		msgtype = input.getLong();
		method 	= input.getLong();
		replysequence = input.getLong();
		error = input.getLong();

		msgData = msgFactory.createMessageData (method, msgtype);
		// XXX: this could possibly be null
	
		if(msgData == null){
			throw new NullPointerException("msgData is null.");
		}
		msgData.read (input);
	}

	/**
	 *  Writes out the NdmpMessage header to the output stream.
	 *  Synchronization on the output stream is assumed to be handled
	 *  by the caller of the function.
	 */
	public void write (NdmpOutput output) throws IOException {
		Date now = new Date();

		timestamp = (new Long(now.getTime() / 1000)).intValue();

		synchronized (output) {
			output.markStartRec();

			output.putLong(sequence);
			output.putLong(timestamp);
			output.putLong(msgtype);
			output.putLong(method);
			output.putLong(replysequence);
			output.putLong(error);

			if (null != msgData)
				msgData.write (output);
				
			output.markEndRec();
		}
	}

	/**
	 * Printing the header is used for debugging purposes.
	 * <p>
	 * @param debugLog The log to write to
	 */
	public void writeLog (NdmpDebugLog debugLog) {
		debugLog.println(50, 
			msgName() + " " + msgTypeName () +" : Message Header\r\n" +
			"\tSequence: "	+ sequence + " , " +  
			"ReplySequence: " + replysequence + " , " + 
			"Timestamp: "	+ timestamp + "\r\n" +
			"\tMethod: " 	+ method + " , " + 
			"Msgtype: " 	+ msgtype + " , " +
			"Error: " + NdmpCons.ErrStrings[error]);

		if (null != msgData)
			msgData.writeLog (debugLog);
	}

	public NdmpMessageData getMsgData() {
		return msgData;
	}

	public String msgError () { return NdmpCons.msgError(error); }

	public String msgName () { return NdmpCons.msgName(method, error); }

	public String msgClassName () { return NdmpCons.msgClassName (method); }

	public String msgTypeName () { return NdmpCons.msgClassTypeName (msgtype); }

	public String msgInterfaceName () { 
		return NdmpCons.msgInterfaceName (method); 
	}

	public boolean isHeaderError () { return (NdmpCons.NDMP_NO_ERR == error);}

	public boolean isHeaderError (int error) {
		return (error == this.error);
	}

	public boolean isMethod (int method) { return (method == this.method); }

}
