
package com.netapp.offtap.ndmpcopy.messages;

//import netapp.cmds.ndmpc.*;
import java.lang.reflect.Constructor;

import com.netapp.offtap.ndmpcopy.debuglog.NdmpDebugLog;
import com.netapp.offtap.ndmpcopy.messages.connect.NdmpMessageConnectOpenRequest;
import com.netapp.offtap.ndmpcopy.util.NdmpCons;

/**
 * NdmpMessageFactory  The message class factory.
 * Responsibility:  This class will create Ndmp Messages. All messages
 * on this earth and beyond are created vis this class only!
 *
 * The interface is divided into two main parts:
 * a> Version Independent interface
 * b> Version dependent interface.
 *
 * For (a): 
 *		createConnectOpenMsg();  for negotiating protocol version.
 *		createNdmpMsgFactory (version ndmpversion); create the factory 
 *							for a given version.
 *
 * For (b):
 *		This depends on 'the' NdmpMsgFactory for a specific protocol version.
 * 		
 * 		createNdmpRequestMsg (int method);
 *		createNdmpReplyMsg (int method);
 *		createNdmpReplyMsg (NdmpMessage msg, int error);
 */
public class NdmpMessageFactory  {
	protected NdmpDebugLog debugLog = null;
	protected int version;		// NOt sure if this be ever used
	protected int sequence = 0;

	public final static int BaseFactVersion		= 3;

	public NdmpMessageFactory (NdmpMessageFactory msgFactory, int version, 
			NdmpDebugLog log) {
		this.version = version;
		this.sequence = msgFactory.sequence;
		this.debugLog = log;
	}

	protected NdmpMessageFactory (NdmpDebugLog log) {
		version = 3;
		this.debugLog = log;
	}
	/**
	 * The static interface..version independent..
	 *	will use the default version 3.
	 */
	
	public NdmpMessage createConnectOpenMsg (short version) {
		NdmpMessageData  msgData = new 
				NdmpMessageConnectOpenRequest (version);

		return new NdmpMessage (NdmpCons.NDMP_CONNECT_OPEN, 
				getNextSequence (), 0, // 0 = reply sequence
				msgData, NdmpCons.NDMP_MESSAGE_REQUEST);
	}

	// to read its reply
	public NdmpMessage createConnectOpenMsg (NdmpMessage msg) {
		return new NdmpMessage (msg, new NdmpMessageDataReply());
	}

	public static NdmpMessageFactory createMsgFactory (NdmpDebugLog log) {
		return new NdmpMessageFactory (log);
	}
	
	/**
	 * This method was designed to build version 4 classes. But is
	 * not used anywhere right now.
	 */
	public static NdmpMessageFactory createMessageFactory (int version,
			NdmpMessageFactory msgFactory) {

		String classname = "NdmpMessageFactory";
		Class fact = null;

		for (int i = version; (i > BaseFactVersion) && (null == fact); --i) {
			try {
				fact = Class.forName (NdmpMessageFactory.class.getClass().getPackage().getName() + "." + 
						"v" + i + "." + classname + "V" + i);
			} catch (ClassNotFoundException e) {
				fact = null;
			}
		}
		
		if (null == fact) {
			msgFactory.version = version;
			return msgFactory;
		}	
		
		Constructor			msgFactConst;
		NdmpMessageFactory	newmsgFact;

		Class[] argsType	= { NdmpMessageFactory.class, int.class };
		Object[] args		= { msgFactory, new Integer (version) };
		
		try {
			msgFactConst	= fact.getConstructor(argsType);
			if (null == msgFactConst)
				return msgFactory;

			newmsgFact		= (NdmpMessageFactory) 
								msgFactConst.newInstance (args);
		}catch (Exception e) {
			System.out.println ("Ndmpcopy: Warning Cant Instantiate " +
				" Java Class" +
				e.getMessage() );
			newmsgFact = null;
			return msgFactory;	// use the old one
		}

		return newmsgFact;
	}
	
	/**
	 * The the interface for NdmpMessageFactory, it covers up the NDMP
	 * inability to classify messages as request, reply, Log.
	 *
	 * The Dynamic or Version dependent interface begins Now.
	 */

	public  NdmpMessage createMessage (NdmpMessage msg) {
		return new NdmpMessage (msg, null);
	}

	public NdmpMessage createRequestMessage (int method, 
			Class[] argsType, Object[] args) {

		NdmpMessageData msgData = createMessageData (method, 
			argsType, args, NdmpCons.NDMP_MESSAGE_REQUEST);
		
		return new NdmpMessage (method, getNextSequence(), 0,
			msgData, NdmpCons.NDMP_MESSAGE_REQUEST);
	}

	public NdmpMessage createReplyMessage (NdmpMessage msg, 
			Class[] argsType, Object[] args) {

		Class dataClass = createMessageDataClass (
				(this.getClass().getPackage().getName() + "." +
				msg.msgInterfaceName().toLowerCase() + "."), 
				(msg.msgClassName () + 
				NdmpCons.msgClassTypeName(NdmpCons.NDMP_MESSAGE_REPLY)),
				NdmpCons.NDMP_MESSAGE_REPLY);
				
		NdmpMessageData msgData = createMessageData (dataClass,
						argsType, args);
		
		return new NdmpMessage (msg, msgData);
	}
	
	public NdmpMessageData createMessageData (int method, int msgtype) {
		return createMessageData (createMessageDataClass (method, msgtype), 
						null, null);
	}
	
	public NdmpMessageData createMessageData (int method, 
			Class[] argsType, Object[] args, int msgtype) {
		
		Constructor msgDatCons;
		NdmpMessageData	msgData;
		Class dataClass = createMessageDataClass (method, msgtype);
		
		return createMessageData (dataClass, argsType, args);
	}

	// This method is used by other createMessageData methods.
	protected NdmpMessageData createMessageData (Class dataClass,
			Class[] argsType, Object[] args) {

		Constructor msgDataCons;
		
		if (null != dataClass) {
			try {
				msgDataCons = dataClass.getConstructor (argsType);
				return (NdmpMessageData) (msgDataCons.newInstance (args));
			} catch (Exception e) {
				debugLog.println (20, "Ndmpcopy: Internal error" +
					e.getMessage());
				
				// dont print this here.... rather
				// put this in the debug log at a higher level
				// tbd: dont catch Exception - its too general - so 
				// catch it at the topmost level of the code
			}
		}
		return null;
	}
	
	public NdmpMessage createLogMsg (int method, Class[] argsType, 
			Object[] args) { 

		NdmpMessageData msgData = createMessageData (method, 
			argsType, args, NdmpCons.NDMP_MESSAGE_REQUEST);

		return new NdmpMessage (method, sequence, 0, msgData, 
			NdmpCons.NDMP_MESSAGE_REQUEST);
	}

	protected int getNextSequence () { 
		++sequence;
		
		return sequence;
	}	

	protected String createMessageDataPackageName (int method) {
		return  (this.getClass().getPackage().getName() + "." +
			NdmpCons.msgInterfaceName (method).toLowerCase() + ".");
	}
	
	protected String createMessageDataName (int method, int msgtype) {			
		return (NdmpCons.msgClassName(method, msgtype));
	}			
		
	protected Class createMessageDataClass (int method, int msgtype) {
		String msgDataName = createMessageDataName (method, msgtype);
		String msgDataPackageName = createMessageDataPackageName (method);
		
		return  createMessageDataClass (msgDataPackageName, 
			msgDataName, msgtype);
	}
	
	protected Class createMessageDataClass (String msgDataPackageName, 
			String msgDataName, int msgtype) {

 		Class dataClass = null;

		for (int i = version; i > BaseFactVersion && null == dataClass; i--) {
			try {
				dataClass = Class.forName (msgDataPackageName + "v" + i + 
								"." + msgDataName + "V" + i); 
			} catch (ClassNotFoundException e) {
				dataClass = null;
			}
		}
	
		if (null != dataClass)
			return dataClass;

		try {
			dataClass = Class.forName (msgDataPackageName + msgDataName);
		}
		catch (ClassNotFoundException e) {
			dataClass = null;
		}

		if (null != dataClass)
			return dataClass;
		
		switch (msgtype) {
			case NdmpCons.NDMP_MESSAGE_REPLY:
				try {
					dataClass = Class.forName (this.getClass().getPackage().getName() +
						".NdmpMessageDataReply");
				} catch (Exception ee) {
					dataClass = null;
				}	
				break;
				
			case NdmpCons.NDMP_MESSAGE_REQUEST:
				dataClass = null;	//No super Request data class exist.
				break;
		}
		return dataClass;
	}	
}
