/*
 * $Id: //depot/prod/ontap/main/prod/common/java/netapp/onc/rpc/XDRRMInputStream.java#4 $
 *
 * Copyright (c) 1998 Network Appliance, Inc. All Rights Reserved.
 */

package com.netapp.offtap.ndmpcopy.xdr.rpc;
 
import java.io.IOException;
import java.io.InputStream;

import com.netapp.offtap.ndmpcopy.xdr.xdr.XDRCons;
import com.netapp.offtap.ndmpcopy.xdr.xdr.XDRConvert;
import com.netapp.offtap.ndmpcopy.xdr.xdr.XDRInputStream;



/**
 * XDRRecInputStream
 * Given an InputStream, this class creates and XDRInputStream,
 * and proceeds to extract basic XDR defined data structures, 
 * with the addition of Record Marking (RM) described on
 * Section 10 of RFC1831 (ONC RPC standard).
 * 
 * The Record Marking leads to the dual calls to the XDRInputStream --
 * One call reads XDR structures. 
 * The other reads in data from different fragments, and calls 
 * XDRConvert to translate the bytes into the requisite data structure.
 */
public class XDRRMInputStream extends XDRInputStream
{
	private int			frag_unread;
	private boolean			last_frag;		/* Indicates last record fragment */

	public XDRRMInputStream (InputStream dataSrc) 
	{
		super(dataSrc);
		this.frag_unread = 0;
		this.last_frag = true;
	}

	public XDRRMInputStream (InputStream dataSrc, int bufsize)
	{
		super(dataSrc, bufsize);
		this.frag_unread = 0;
		this.last_frag = true;
	}
		
 
	/**
	 * Only when reading the frag header, should we assume
	 * we don't need to worry about overlapping a fragment.
	 * Since we haven't read any bites of the fragment, and
	 * XDR transmits everything on 4 byte boundaries we know
	 * that this read SHOULD ALWAYS be successful.
	 */
	private void readFragHdr() throws IOException {
		int	fraghdr = super.getLong();
 
		frag_unread = fraghdr & 0x7fffffff;
		if ((fraghdr&0x80000000)!=0)
			last_frag = true;
		else
			last_frag = false;
	}

	private byte[] readFragData(int num_bytes) throws IOException
	{
		int	tmp_num_bytes, bytes_to_fetch;
		byte[]	tmp_bytes;
		byte[]	data;

		if (num_bytes == 0)
			throw new XDRRMException(
				"Cannot fetch 0 bytes from record.\n");

		data = new byte[num_bytes];
		/* 
		 * Keep reading fragments until we have
		 * the number of bytes we need. Make sure to update 
		 * the amount of data left int he fragments.
		 * Also, copy the data into a big buffer to be returned
		 */
		bytes_to_fetch = num_bytes;
		while (bytes_to_fetch != 0) {
			// We've got more fragments
			if (frag_unread == 0)
				if (last_frag == true)
					throw new XDRRMException ("No more bytes in record");
				else
					readFragHdr();

			// Read what you need from the fragment 

			// A MIN function would be nice...
			if (frag_unread < bytes_to_fetch) 
				tmp_num_bytes = frag_unread;
			else 
				tmp_num_bytes = bytes_to_fetch;

			// Technically, we can get a 0 length fragment...
			if (bytes_to_fetch != 0) {
				super.readData(data, 
				    (num_bytes - bytes_to_fetch),
				    tmp_num_bytes);
				bytes_to_fetch -= tmp_num_bytes;
				frag_unread -= tmp_num_bytes;
			}

			// Uh-oh pardner, this was the last round-up
			if (last_frag == true) { 
				if (bytes_to_fetch != 0)  
					throw new XDRRMException(
					    "No more bytes in record");

				else 
					return (data);
			}
		}
		return (data);
	}
 
	/**
	 * Read in however many bytes are remaining in the
	 * fragment, and ignore them.
	 * This function should be used for restting the input stream.
	 */
	private void tossBytes() throws IOException {
		byte[]	toss_buf = new byte[frag_unread];

		super.readData(toss_buf, 0, frag_unread);
		frag_unread = 0;
	}

	/**
	 * Skip whatever bytes are left from the last (presumably failed)
	 * message.
	 * XXX - Should we read in the header for the next record?
	 */
	public void resetStream() throws IOException {
		
		// Throw away whatever number of bytes remains in the record
		if (frag_unread != 0)            // Toss rest of this fragment
			tossBytes();
 
		// Now, toss the rest of the fragments in the record 
		while (last_frag != true) {
			readFragHdr();
			if (frag_unread != 0)
				tossBytes();
		}
		last_frag = false;
	}

	public byte[] getBytes() throws IOException {
		int	byte_len = getLong();

		return getBytes(byte_len);
	}
		
	public byte[] getBytes(int byte_len) throws IOException {
		byte[]	ret_buf, toss_buf;
		int	toss_bytes;

		if (byte_len == 0)
			return (null);
		ret_buf = readFragData(byte_len);

		// Make sure we're on a XDR UNIT
		toss_bytes = XDRConvert.toAlignBytes(byte_len);
		if (toss_bytes != 0) 
			readFragData(toss_bytes);

		return (ret_buf);
	}

	public int getLong() throws IOException {
		byte[]	tmp_bytes;

		tmp_bytes = readFragData(XDRCons.XDR_LONG_BYTES);
		return (XDRConvert.toLong(tmp_bytes));
	}


	public long getHyper() throws IOException {
		byte[]	tmp_bytes;

		tmp_bytes = readFragData(XDRCons.XDR_HYPER_BYTES);
		return (XDRConvert.toHyper(tmp_bytes));
	}

	public short getShort() throws IOException {
		byte[]	tmp_bytes;

		tmp_bytes = readFragData(XDRCons.XDR_SHORT_BYTES);
		return (XDRConvert.toShort(tmp_bytes));
	}

	public byte getByte() throws IOException {
		byte[]	tmp_bytes;

		tmp_bytes = readFragData(XDRCons.XDR_BYTE_BYTES);
		return (XDRConvert.toByte(tmp_bytes));
	}

	public String getString() throws IOException {
		byte[]	byte_arr;
		
		byte_arr = getBytes();
		if (byte_arr == null)
			return (new String(""));

		return (new String(byte_arr));
	}

        public int getFragUnread() {
                return frag_unread;
        }

}
