/*
 * $Id: //depot/prod/ontap/main/prod/common/java/netapp/onc/xdr/XDRConvert.java#2 $
 *
 * Copyright (c) 1998 Network Appliance, Inc. All Rights Reserved.
 */

package com.netapp.offtap.ndmpcopy.xdr.xdr;

import java.io.IOException;

/**
 * Convert a number of bytes into viable XDR data structures
 */
public class XDRConvert
{

	final static public int toNumXDRBytes(int num_bytes)
	{
		int	align_bytes, overhang;

		align_bytes = 0;
		overhang = num_bytes % XDRCons.XDR_ALIGN;
		if (overhang > 0)
			align_bytes = XDRCons.XDR_ALIGN - overhang;

		return (num_bytes + align_bytes);  
	}

	final static public int toAvailXDRBytes(int num_bytes)
	{
		int	overhang; 

		overhang = num_bytes % XDRCons.XDR_ALIGN;
		return (num_bytes - overhang);
	}

	final static public int toNonAlignedBytes(int num_bytes)
	{
		return (num_bytes % XDRCons.XDR_ALIGN);
	}

	final static public int toAlignBytes(int num_bytes)
	{
		int 	overhang;

		overhang = num_bytes % XDRCons.XDR_ALIGN;
		if (overhang != 0)
			return (XDRCons.XDR_ALIGN - overhang);
		return (0);
	}
	/**
	 * Put a byte array into the buffer. This basically
	 * means we need to align it on a XDR boundary.
	 */
	final static public int fromBytes(byte[] val, int in_idx, int amt_data, 
	    byte[] buffer, int out_idx) throws IOException
	{
		int	i, total_len;

		total_len = toNumXDRBytes(amt_data);

		if ((out_idx + total_len) > buffer.length) 
			throw new IOException (
			    "Buffer space not large enough for data.");

		System.arraycopy(val, in_idx, buffer, out_idx, amt_data);

		for (i = (out_idx + amt_data); i < total_len; i++)
			buffer[i] = 0;

		return (out_idx + total_len);
	}

	final static public int fromBytes(byte[] val, byte[] buffer, 
	    int out_idx) throws IOException
	{
		return (fromBytes(val, 0, val.length, buffer, out_idx));
	} 
		
	final static public int fromBytesPart(byte[] val, byte[] buffer, 
	    int out_idx) throws IOException
	{
		return (fromBytesPart(val, 0, val.length, buffer, out_idx));
	}

	final static public int fromBytesPart(byte[] val, int in_idx,
	    int data_amt, byte[] buffer, int out_idx) throws IOException
	{
		int	max_avail_bytes, max_bytes_xdr;
		int	num_bytes, idx;

		max_avail_bytes = buffer.length - out_idx;
		max_bytes_xdr = toNumXDRBytes(data_amt);

		if (max_avail_bytes < max_bytes_xdr)
			num_bytes = toAvailXDRBytes(max_avail_bytes);
		else
			num_bytes = data_amt;

		idx = fromBytes(val, in_idx, num_bytes, buffer, out_idx);
		return (idx - out_idx);

	}
 
	
	/**
	 * Put an XDR hyper (64 bits) to the buffer
	 */
	final static public int fromHyper(long val, byte[] buffer, int offset) 
	    throws IOException
	{
		if ((offset + XDRCons.XDR_HYPER_BYTES) > buffer.length) 
			throw new IOException (
			    "Buffer space not large enough for hyper.");

		buffer[offset++] = (byte)((val >> 56) & 0xff);
		buffer[offset++] = (byte)((val >> 48) & 0xff);
		buffer[offset++] = (byte)((val >> 40) & 0xff);
		buffer[offset++] = (byte)((val >> 32) & 0xff);
		buffer[offset++] = (byte)((val >> 24) & 0xff);
		buffer[offset++] = (byte)((val >> 16) & 0xff);
		buffer[offset++] = (byte)((val >>  8) & 0xff);
		buffer[offset++] = (byte)((val >>  0) & 0xff);

		return (offset);
	}

	/**
	 * Put an XDR long to the buffer
	 */
	final static public int fromLong(int val, byte[] buffer, int offset) 
	    throws IOException 
	{
		if ((offset + XDRCons.XDR_LONG_BYTES) > buffer.length) 
			throw new IOException (
			    "Buffer space not large enough for long.");

		buffer[offset++] = (byte)((val >> 24) & 0xff);
		buffer[offset++] = (byte)((val >> 16) & 0xff);
		buffer[offset++] = (byte)((val >>  8) & 0xff);
		buffer[offset++] = (byte)((val >>  0) & 0xff);

		return (offset);
	}

	/**
	 * Put an XDR short to the buffer
	 */
	final static public int fromShort(short val, byte[] buffer, int offset) 
	    throws IOException 
	{
		if ((offset + XDRCons.XDR_SHORT_BYTES) > buffer.length) 
			throw new IOException (
			    "Buffer space not large enough for short.");

		buffer[offset++] = 0;
		buffer[offset++] = 0;
		buffer[offset++] = (byte)((val >>  8) & 0xff);
		buffer[offset++] = (byte)((val >>  0) & 0xff);

		return (offset);
	}

	/**
	 * Put a byte out to the marshal buffer
	 */
	final static public int fromByte(byte val, byte[] buffer, int offset) 
	    throws IOException 
	{
		if (offset + XDRCons.XDR_BYTE_BYTES > buffer.length)
			throw new IOException (
			    "Buffer space not large enough for byte.");

		buffer[offset++] = 0;
		buffer[offset++] = 0;
		buffer[offset++] = 0;
		buffer[offset++] = val;

		return (offset);
	}

	/**
	 * Put a XDR string to the marshal buffer
	 */
	final static public int fromString(String s, byte[] buffer, int offset) 
	    throws IOException 
	{
		return fromBytes(s.getBytes(), buffer, offset);
	}


	/**
	 * Really, we can't do a whole lot here. 
	 * Basically, we're a glorified copy. They tell us how many
	 * bytes out of the buffer, at which offset, and we
	 * send back a standalone buffer with just that data in it...
	 */
	final static public byte[] toBytes(byte[] buffer, int datalen, 
	    int offset) throws IOException 
	{
		byte[]	retval = new byte[datalen];

		if ((datalen + offset) > buffer.length) 
			throw new IOException(
			    "Cannot extract " +datalen+ " bytes from buffer.");

		System.arraycopy(buffer, offset, retval, 0, datalen);
		return (retval);
	}
		
	/**
	 * Translate a hyper from the buffer
	 */ 
	final static public long toHyper(byte[] buffer, int offset) 
	    throws IOException 
	{
		long 	retval;
	
		if ((offset + XDRCons.XDR_HYPER_BYTES) > buffer.length)
			throw new IOException(
			    "Cannot extract a hyper from buffer.");

		retval = 0;
		// Java doesn't like long constants, so we use bit shifting
		retval |= (buffer[offset + 0] << 56) & ((long)0xff000000 << 32);
		retval |= (buffer[offset + 1] << 48) & ((long)0x00ff0000 << 32);
		retval |= (buffer[offset + 2] << 40) & ((long)0x0000ff00 << 32);
		retval |= (buffer[offset + 3] << 32) & ((long)0x000000ff << 32);

		retval |= (buffer[offset + 4] << 24) & (long)0x00000000ff000000;
		retval |= (buffer[offset + 5] << 16) & (long)0x0000000000ff0000;
		retval |= (buffer[offset + 6] <<  8) & (long)0x000000000000ff00;
		retval |= (buffer[offset + 7] <<  0) & (long)0x00000000000000ff;
		return (retval);
	}

	final static public long toHyper(byte[] buffer)
	    throws IOException
	{
		return (toHyper(buffer, 0));
	}
			
	/**
	 * Turn a buffer into a long
	 */ 
	final static public int toLong(byte[] buffer, int offset) 
	    throws IOException 
	{
		int	retval;

		if ((offset + XDRCons.XDR_LONG_BYTES) > buffer.length)
			throw new IOException(
			    "Cannot extract a long from buffer.");

		retval = 0;
		retval |= ((buffer[offset + 0] << 24) & 0xff000000);
		retval |= ((buffer[offset + 1] << 16) & 0x00ff0000);
		retval |= ((buffer[offset + 2] <<  8) & 0x0000ff00);
		retval |= ((buffer[offset + 3] <<  0) & 0x000000ff);

		return (retval);
	}

	final static public int toLong(byte[] buffer)
	    throws IOException
	{
		return (toLong(buffer, 0));
	}

	/**
	 * Turn a buffer into a short
	 */ 
	final static public short toShort(byte[] buffer, int offset) 
	    throws IOException 
	{
		short	retval;

		if ((offset + XDRCons.XDR_SHORT_BYTES) > buffer.length)
			throw new IOException(
			    "Cannot extract a short from buffer.");

		retval = 0;
		// Ignore top two bytes
		retval |= ((buffer[offset + 2] << 8) & 0xff00);
		retval |= ((buffer[offset + 3] << 0) & 0x00ff);

		return (retval);
	}

	final static public short toShort(byte[] buffer)
	    throws IOException
	{
		return (toShort(buffer, 0));
	}

	/**
	 * Turn a buffer into a byte
	 */ 
	final static public byte toByte(byte[] buffer, int offset) 
	    throws IOException 
	{
		if((offset + XDRCons.XDR_BYTE_BYTES) > buffer.length)
			throw new IOException(
			    "Cannot extract a byte from the buffer.");

		return (buffer[offset + 3]);
	}
	final static public byte toByte(byte[] buffer)
	    throws IOException
	{
		return (toByte(buffer, 0));
	}
}
