package com.netapp.offtap3;

import com.netapp.offtap3.internal.ontap.support.ZapiValidator;
import com.netapp.offtap3.ontap.DedupeResource;
import com.netapp.offtap3.utils.StringUtils;
import com.netapp.offtap3.utils.Utils;
import org.apache.log4j.Logger;

import java.util.*;

/**
 * Used to configure the Offtap objects.  Reasonable defaults are provided, so this object
 * can be built with just the values that should be overridden.
 */
public class OfftapPreferences {
    private static final Logger LOGGER = Logger.getLogger(OfftapPreferences.class);
    private static final Map<ProxyType, Set<? extends Resource>> DEFAULT_RESOURCES = new HashMap<ProxyType, Set<? extends Resource>>();
    
    private static final String KEY_OFFTAP_SECURITY_SSLV3_ENABLED =  "offtap.security.sslv3.enabled";

    public static final Set<ControllerResource> AEONFLUX_2_X_CONTROLLER_DISCOVERY = EnumSet.of(
            ControllerResource.system_info, ControllerResource.licenses, ControllerResource.aggregates,
            ControllerResource.interfaces, ControllerResource.volumes, ControllerResource.luns,
            ControllerResource.protocols, ControllerResource.iscsi, ControllerResource.igroups, ControllerResource.fcp,
            ControllerResource.vfilers, ControllerResource.svms, ControllerResource.cf_info, ControllerResource.vfiler_status, ControllerResource.space_usage);

    static {
        DEFAULT_RESOURCES.put(ProxyType.CONTROLLER, AEONFLUX_2_X_CONTROLLER_DISCOVERY);
        DEFAULT_RESOURCES.put(ProxyType.VOLUME, EnumSet.allOf(VolumeResource.class));
        DEFAULT_RESOURCES.put(ProxyType.LUN, Collections.<Resource>emptySet());
        DEFAULT_RESOURCES.put(ProxyType.FILE, Collections.<Resource>emptySet());
        DEFAULT_RESOURCES.put(ProxyType.DEDUPE, EnumSet.of(DedupeResource.status));
        DEFAULT_RESOURCES.put(ProxyType.SNAPMIRROR, Collections.<Resource>emptySet());
        DEFAULT_RESOURCES.put(ProxyType.EXPORT_POLICY, Collections.<Resource>emptySet());
    }

    static final String FILE_PERSISTOR = "file";
    static final String NO_PERSISTOR = "none";

    private static final int DEFAULT_NDMP_PROGRESS_TIMEOUT = 120;
    private static final int DEFAULT_NDMP_BEGIN_TIMEOUT = 45;
    private static final int DEFAULT_NDMP_DESTINATION_CHECK_INTERVAL = 10;

    private Properties props = new Properties();
    // at some point, we can use properties to drive what kind of validation we want
    private ZapiValidator zapiValidator = new ZapiValidator();

    public OfftapPreferences() {
    }

    public OfftapPreferences(Map properties) {
        if (properties != null) {
            props.putAll(properties);
        }
    }
    
    public static String getInitialResourcesController() {
    	List<String> initialResources = new ArrayList<String>();
    	for (ControllerResource curr : AEONFLUX_2_X_CONTROLLER_DISCOVERY) {
    		initialResources.add(curr.toString());
    	}
    	return Utils.join(initialResources, ",");
    }

    /**
     * @return A copy of the offtap preferences as a Properties object.
     */
    public Properties asProperties() {
        return (Properties) props.clone();
    }
    
    public void putProperties(Properties incoming) {
    	props.putAll(incoming);
    }

    public OfftapPreferences(Properties preferences) {
        props = preferences;
    }

    public OfftapPreferences(Properties preferences, String keyPrefix) {
        for (String key : preferences.stringPropertyNames()) {
            if (key.startsWith(keyPrefix)) {
                String actualKey = key.substring(keyPrefix.length() + 1);
                props.setProperty(actualKey, preferences.getProperty(key));
            }
        }
    }

    public ZapiValidator getZapiValidator() {
        return zapiValidator;
    }

    public void resetProperties(Properties preferences) {
        props = preferences;
    }

    public int getManagerMaxRunningControllerThreads() {
        return getIntProperty("offtap.controllerManager.maxThreads", 4);
    }

    public int getZapiTimeout() {
        return getIntProperty("offtap.zapiTimeout", 60);
    }
    
    //10 min
    public int getZapiErrorTimeoutInOntapConnection() {
        return getIntProperty("offtap.zapi.error.ontap.connection.timeout", 600000);
    }

    public int getZapiConnectionTimeout() {
        return getIntProperty("offtap.zapi.connection.timeout", 10);
    }
    
    public boolean getWantsUnmodifiableCollections() {
    	return getBooleanProperty("wants.unmodifiable.collections", true);
    }

    public int getMaxRunningControllerThreads() {
        return getIntProperty("offtap.controller.maxThreads", 20);
    }

    public boolean getWantsZapiValidation() {
        return getBooleanProperty("offtap.zapi.validation", false);
    }

    public int getMaxRunningVolumeThreads() {
        return getIntProperty("offtap.volume.maxThreads", 20);
    }

    public int getMaxVolumeListInfoRecords() {
        return getIntProperty("offtap.volume.listInfo.maxRecords", 50);
    }

    /**
     * How many seconds to wait for the root volume LS mirror to be updated.
     */
    public int getRootVolMirrorUpdateTimeout() {
        return getIntProperty("offtap.volume.rootVolMirror.updateTimeout", 30);
    }

    public int getMaxRunningExportTaskThreads() {
        return getIntProperty("offtap.exportTask.maxThreads", 20);
    }

    public int getMaxRunningLunTaskThreads() {
        return getIntProperty("offtap.lunTask.maxThreads", 20);
    }

    /**
     * The value to use for max-records in an *-iter ZAPI request.
     * Specifying a negative value will result in not sending a max-records
     * tag in the request, and yield the ONTAP default (which is currently 20).
     */
    public long getIteratorSize() {
        return getLongProperty("offtap.iterator.size", 500L);
    }

    public long getOperationTimeoutPeriod() {
        return 1000 * getLongProperty("offtap.operation.timeout.period.seconds", -1L);
    }

    public String getOperationTimeoutPeriodExemptions() {
        return props.getProperty("offtap.operation.timeout.exemptions");
    }

    public boolean getWantsTraceOutput() {
        return getBooleanProperty("offtap.zapi.trace.output", false);
    }

    public boolean getWantsRawHttpCommunication() {
        return getBooleanProperty("offtap.wants.raw.http.communication", true);
    }

    public String getPersistorFileName() {
        return props.getProperty("offtap.persistor.fileName", "info.x");
    }

    public boolean isFilePersisted() {
        return FILE_PERSISTOR.equals(props.getProperty("offtap.persistor.type", FILE_PERSISTOR));
    }

    public boolean getWantsCloneValidation() {
        return getBooleanProperty("offtap.wants.clone.validation", false);
    }

    public long getZapiSleepTime() {
        return getLongProperty("hidden.offtap.pref.zapi.sleep.time.milliseconds", 0);
    }

    public boolean getWantsRetryOnRecoverableZapiFailure() {
        return getBooleanProperty("offtap.wants.retry.on.recoverable.zapi.failure", true);
    }

    public int getZapiThrottleSize() {
    	return getIntProperty("offtap.zapi.throttle.size", 20);
    }

    /*
     * ZAPI Prioritization is used in ZAPIInvoker to say what operations are high priority
     * requestors, mediums, and lows.  The operation is stored in the ThreadLocal
     * ZAPIStatsCounter.setRecordedKey.  Talk to dsarisky if you want to know how all
     * this stuff really works.
     */
    public String getDefaultZapiPriority() {
        return props.getProperty("priority.default.priority", "MEDIUM");
    }

    public String getLowPriorityZapiRequestors() {
        return props.getProperty("priority.low.priority.callers", "");
    }

    public String getMediumPriorityZapiRequestors() {
        return props.getProperty("priority.medium.priority.callers", "");
    }

    public String getHighPriorityZapiRequestors() {
        return props.getProperty("priority.high.priority.callers", "");
    }
    
    public int getMccSleepTimeInSeconds() {
    	return this.getIntProperty("mcc.polling.time.in.seconds", 10);
    }
    
    public int getMccOldZapiTimeInSeconds() {
    	return this.getIntProperty("mcc.old.zapi.time.in.seconds", 10);
    }
    
    public int getZapiInvocationMinimalTimeout() {
    	return this.getIntProperty("zapi.invocation.minimal.timeout", 15);
    }
    
    public double getZapiThrottleFactor() {
        return getDoubleProperty("offtap.zapi.throttle.factor", 0.75);       
    }
    
    private double getDoubleProperty(String name, double defaultValue) {
   	 String value = props.getProperty(name);
        if (value == null) {
            return defaultValue;
        }
        else {
            return Double.parseDouble(value);
        }
   }
    
   
    public boolean getNewCmodeZapiThrottle() {
        return getBooleanProperty("offtap.wants.new.cmode.zapi.throttle", true);
    }

    public int getSiscloneRetryTimesOnEbusy() {
        return getIntProperty("offtap.clone.retry.times", 5);
    }

    public int getSiscloneIntervalBetweenRetryInSecs() {
        return getIntProperty("offtap.clone.retry.interval", 5);  // WAFL avg CP time
    }

    /**
     * After issuing a snapmirror-abort ZAPI, how long should we keep retrying before
     * giving up.
     */
    public int getSnapMirrorAbortRetryTimeout() {
        return getIntProperty("offtap.snapmirror.abort.timeout", 10);
    }

    public void setSnapMirrorAbortRetryTimeout(String value) {
        props.setProperty("offtap.snapmirror.abort.timeout", value);
    }

    /**
     * After issuing a snapmirror-break ZAPI, how long should we keep retrying before
     * giving up.
     */
    public int getSnapMirrorBreakRetryTimeout() {
        return getIntProperty("offtap.snapmirror.break.timeout", 10);
    }

    public void setSnapMirrorBreakRetryTimeout(String value) {
        props.setProperty("offtap.snapmirror.break.timeout", value);
    }
    
    public int getMccCheckThresholdInSeconds() {
        return this.getIntProperty("mcc.check.threshold.in.seconds", 60);
    }

    /**
     * After issuing a snapmirror-quiesce ZAPI, how long should we keep retrying before
     * giving up.
     */
    public int getSnapMirrorQuiesceRetryTimeout() {
        return getIntProperty("offtap.snapmirror.quiesce.timeout", 10);
    }

    public void setSnapMirrorQuiesceRetryTimeout(String value) {
        props.setProperty("offtap.snapmirror.quiesce.timeout", value);
    }
    
    /**
     * After issuing a snapmirror-resume ZAPI, how long should we keep retrying before
     * giving up.
     */
    public int getSnapMirrorResumeRetryTimeout() {
        return getIntProperty("offtap.snapmirror.resume.timeout", 10);
    }

    public void setSnapMirrorResumeRetryTimeout(String value) {
        props.setProperty("offtap.snapmirror.resume.timeout", value);
    }

    public boolean getSnmpLoadThrowsExceptionWhenDisabled() {
        return getBooleanProperty("offtap.snmp.loadException", false);
    }
    
    /**
     * Fix for CVE-2014-3566: SSLv3 Nondeterministic CBC Padding Vulnerability
     * 
     * @param enabled <code>true</code> for SSLv3 enabled compatibility mode.
     */
    public void setSslv3Enabled(boolean enabled) {
        props.setProperty(KEY_OFFTAP_SECURITY_SSLV3_ENABLED, "" + enabled);
    }
    
    /**
     * Fix for CVE-2014-3566: SSLv3 Nondeterministic CBC Padding Vulnerability
     * 
     * @return <code>true</code> to enable SSLv3 for compatibility mode.
     */
    public boolean getSslv3Enabled() {
        boolean ret = getBooleanProperty(KEY_OFFTAP_SECURITY_SSLV3_ENABLED, false);
        
        /*
         * Since we could break a customer with some really old ONTAP load,
         * we want to make sure there is a way to enable this SSLv3 compatibility mode.
         * Therefore, you can enable via normal means or set the value via a Java
         * System property.
         */
        if (!ret && props.getProperty(KEY_OFFTAP_SECURITY_SSLV3_ENABLED) == null) {
            // Also check system properties
            String sProp = System.getProperty(KEY_OFFTAP_SECURITY_SSLV3_ENABLED);
            if (StringUtils.isReal(sProp)) {
                ret = Boolean.parseBoolean(sProp);
            }
        }
        
        return ret;
    }

    /**
     * Ontap SN early builds have a lot of bugs, but we need to be able to verify
     * them with vvol code.  So we're introducing a hack flag so that we can workaround some of
     * the various things that do not work.
     */
    public boolean getWantsSNCompabilityMode() {
        return getBooleanProperty("sn.compatibility.mode", false);
    }

    /**
     * Returns the initial resources for the specified storage systems type and proxy type. These are the resources
     * that must be loaded when the controller is created.
     * @param systems The Set of storage systems for which these resources must be compatible.
     * @param proxyType The type of proxy that will be loading the resources.
     * @return A Set of Resource enums that must be loaded.
     */
    public Set<Resource> getInitialResources(Set<StorageSystem> systems, ProxyType proxyType) {
        String initialResourcesName = props.getProperty("initial.resources." + proxyType.name().toLowerCase());
        if (initialResourcesName != null) {
            Set<Resource> initialResources = new LinkedHashSet<Resource>();
            StringTokenizer tokenizer = new StringTokenizer(initialResourcesName, ",");
            while (tokenizer.hasMoreTokens()) {
                String token = tokenizer.nextToken();
                try {
                    Resource resource = proxyType.valueOfResource(token);
                    if (resource.supports(systems)) {
                        initialResources.add(resource);
                    } else {
                        LOGGER.trace(
                                "Ignoring resource " + resource + " since it is not in the list of supported systems: " + systems);
                    }
                } catch (IllegalArgumentException e) {
                    LOGGER.error("An unrecognized resource type was requested '" + token + "'", e);
                }
            }
            return initialResources;
        } else {
            Set<Resource> initialResources = new HashSet<Resource>(DEFAULT_RESOURCES.get(proxyType));
            for (Iterator<Resource> iter = initialResources.iterator(); iter.hasNext(); ) {
                Resource resource = iter.next();
                if (!resource.supports(systems)) {
                    iter.remove();
                }
            }
            return initialResources;
        }
    }

    /**
     * The NDMP copy process checks that progress is being made. If the destination file has not changed size
     * after this amount of time, it is considered to have failed.
     * @return The progress timeout for the NDMP copy process in seconds.
     */
    public int getNdmpProgressTimeout() {
        return getIntProperty("offtap.ndmp.progress.timeout", DEFAULT_NDMP_PROGRESS_TIMEOUT);
    }

    /**
     * The NDMP copy process checks that the copy process begins by verifying the destination file exists. If the
     * destination file has not been created after this amount of time, it is considered to have failed.
     * @return The begin timeout for the NDMP copy process in seconds.
     */
    public int getNdmpBeginTimeout() {
        return getIntProperty("offtap.ndmp.being.timeout", DEFAULT_NDMP_BEGIN_TIMEOUT);
    }

    /**
     * The interval in seconds between checks for status of the destination file during an NDMP copy process.
     */
    public int getNdmpDestinationCheckInterval() { return getIntProperty("offtap.ndmp.destinationcheck.interval", DEFAULT_NDMP_DESTINATION_CHECK_INTERVAL); }

    private long getLongProperty(String name, long defaultValue) {
        String value = props.getProperty(name);
        if (value == null) {
            return defaultValue;
        } else {
            return Long.parseLong(value);
        }
    }

    private int getIntProperty(String name, int defaultValue) {
        String value = props.getProperty(name);
        if (value == null) {
            return defaultValue;
        } else {
            return Integer.parseInt(value);
        }
    }

    private boolean getBooleanProperty(String name, boolean defaultValue) {
        String value = props.getProperty(name);
        if (value == null) {
            return defaultValue;
        } else {
            return Boolean.parseBoolean(value);
        }
    }
}
