package com.netapp.offtap3;

import com.netapp.offtap3.utils.StringUtils;

/**
 * Characterizes the storage space/size similar to TimeUnit in JDK.
 */
public class Size implements Comparable<Size> {

    private final long value;
    private final Unit unit;

    public static final Size ZERO = Size.createB(0L);

    public Size(long value, Unit unit) {
        this.value = value;
        this.unit = unit;
    }

    public static Size parse(String ontapFormat) {
        // expect to be some numbers with a letter on the end
        ontapFormat = ontapFormat.toLowerCase();
        if (StringUtils.isNumeric(ontapFormat)) {
            return createKB(Long.parseLong(ontapFormat));
        }

        Long val = Long.parseLong(ontapFormat.substring(0, ontapFormat.length()-1));
        char suffix = ontapFormat.charAt(ontapFormat.length()-1);
        switch (suffix) {
            case 'k' : return createKB(val);
            case 'm' : return createMB(val);
            case 'g' : return createGB(val);
            case 't' : return createTB(val);
            default :
                throw new IllegalArgumentException("Unable to parse this size measurement: " + ontapFormat);
        }
    }

    public static Size createKB(long value) {
        return new Size(value, Unit.KB);
    }

    public static Size createMB(long value) {
        return new Size(value, Unit.MB);
    }

    public static Size createGB(long value) {
        return new Size(value, Unit.GB);
    }

    public static Size createTB(long value) { return new Size(value, Unit.TB); }

    public static Size createB(long value) {
        return new Size(value, Unit.B);
    }

    public Size plus(Size size) {
        return Size.createB(toBytes() + size.toBytes());
    }

    public Size minus(Size size) {
        return Size.createB(toBytes() - size.toBytes());
    }

    public boolean equals(Object obj) {
        if (obj instanceof Size) {
            Size size = (Size) obj;
            if (unit == size.unit) {
                return value == size.value;
            }
            else {
                return toBytes() == size.toBytes();
            }

        }
        return false;
    }

    public int hashCode() {
        return (int) toBytes();
    }

    public long toMB() {
        return unit.convert(value, Unit.MB);
    }

    public long toGB() {
        return unit.convert(value, Unit.GB);
    }

    public long toTB() { return unit.convert(value, Unit.TB); }

    public long toBytes() {
        return unit.convert(value, Unit.B);
    }

    public long convert(Unit target) {
        return unit.convert(value, target);
    }

    public String toString() {
        return value + " " + unit;
    }

    @Override
    public int compareTo(Size size) {
        return Long.valueOf(toBytes()).compareTo(size.toBytes());
    }

    public enum Unit {
        B(0),
        KB(1),
        MB(2),
        GB(3),
        TB(4),
        PB(5);

        private final long factor;

        Unit(int exp) {
            factor = (long) Math.pow(1024, exp);
        }

        public long convert(long size, Unit newUnit) {
            return size * factor / newUnit.factor;
        }
    }
}
