package com.netapp.offtap3.factories;

import com.netapp.offtap3.ConnectionException;
import com.netapp.offtap3.Controller;
import com.netapp.offtap3.ControllerConnection;
import com.netapp.offtap3.HasChildControllers;
import com.netapp.offtap3.HasParentController;
import com.netapp.offtap3.factories.platform.PlatformDataLoader;
import com.netapp.offtap3.internal.ontap.MCCManager;
import com.netapp.offtap3.internal.ontap.controller.HasChildOntapControllers;
import com.netapp.offtap3.ontap.AdminSvm;
import com.netapp.offtap3.ontap.DataSvm;
import com.netapp.offtap3.utils.SafeConcurrentHashMap;
import com.netapp.offtap3.utils.Utils.VscThreadFactory;

import org.apache.log4j.Logger;

import java.io.IOException;
import java.util.Collection;
import java.util.Collections;
import java.util.Map;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.concurrent.TimeUnit;

public class ControllerFactory {

    private static final Logger LOGGER = Logger.getLogger(ControllerFactory.class);

	private final Map<String, Controller> controllers = new SafeConcurrentHashMap<String, Controller>();
    private final ConnectionPersistor persistor;

    private static final ControllerFactory singleton = new ControllerFactory();
    private static final Map<ControllerConnection, ControllerConnection> loadingControllers = new SafeConcurrentHashMap<ControllerConnection, ControllerConnection>();
    private static final int MAX_REFRESH_THREADS = 20;
    private static final long MAX_REFRESH_TIMEOUT = 10; // Minutes

    private ControllerFactory() {
        try {
            persistor = new PropertiesFileConnectionPersistor("aeonflux_connections.properties");
        }
        catch (IOException ioe) {
            throw new RuntimeException("Unable to create a persistor", ioe);
        }

    }

    public static ControllerFactory getInstance() {
        return singleton;
    }

	public Controller addConnection(ControllerConnection controllerConnection) throws ConnectionException, DataLoaderException {
        ControllerConnection connectionSync = controllerConnection;
        // If a connection with the same information is already being loaded, find it.
        synchronized (loadingControllers) {
            if (loadingControllers.values().contains(controllerConnection)) {
                connectionSync = loadingControllers.get(controllerConnection);
            } else {
                loadingControllers.put(controllerConnection, controllerConnection);
            }
        }
        // Synchronize on the connection so that any one connection can only be loaded once.
        synchronized (connectionSync) {
            Controller controller = null;
            if (persistor.listAll().contains(controllerConnection)) {
                // This connection has already been added. Let's refresh it and return it so we don't get multiple instances of the top level controller.
            	try {
            		controller = getControllerbyIp(controllerConnection.getIpAddress());
                    if (controller != null) {
                        controller.refresh();
                    }
            	} catch (Exception e) {
                    LOGGER.debug("Caught exception when trying to refresh controller that is being added again.", e);
            		// lets go for a new one instead
            	}
            }
            try {
                // Persist connection info
                persistor.add(controllerConnection);
                controller = addController(controllerConnection);
            } finally {
                loadingControllers.remove(controllerConnection);
            }
            return controller;
        }
	}
	
    /**
     * Updates the controller information without having to reload the controller. If the controller was already loaded, its connection information will be
     * updated so that future calls use the new information.
     * @param controllerConnection The connection information to update.
     */
    public void updateConnection(ControllerConnection controllerConnection) {
        LOGGER.debug("Updating connection information for controller with IP address: " + controllerConnection.getIpAddress());
        Controller controller = getControllerbyIp(controllerConnection.getIpAddress());
        persistor.remove(controller.getConnection());
        persistor.add(controllerConnection);
        controller.setConnection(controllerConnection);
        if (controller instanceof HasChildControllers) {
            for (Controller child : ((HasChildControllers)controller).getChildControllers().values()) {
                child.setConnection(controllerConnection);
            }
        }
    }

    private Controller addController(ControllerConnection controllerConnection) throws DataLoaderException, ConnectionException {
        LOGGER.debug("Adding new controller, ip: " + controllerConnection.getIpAddress());
        PlatformDataLoader loader = PlatformDataLoaderFactory.buildPlatformDataLoader(controllerConnection);
        Controller controller = loader.load(controllerConnection);
        controllers.put(controller.getId(), controller);
        if (controller instanceof HasChildOntapControllers) {
            for (Controller child : ((HasChildControllers) controller).getChildControllers().values()) {
                controllers.put(child.getId(), child);
            }
        }
        return controller;
    }

	public Controller getController(String controllerId) {
		return controllers.get(controllerId);
	}

    /**
     * Get a controller by connect Ip (not useful for data svms that are loaded through an admin svm)
     *
     * @return The controller if it can be found or <code>null</code> if it cannot.
     */
    public Controller getControllerbyIp(String ipAddress){
        if (ipAddress == null) return null;
        for (Controller controller : controllers.values()){
            if (ipAddress.equals(controller.getConnection().getIpAddress())){
                if (controller instanceof HasParentController) {
                    return ((HasParentController)controller).getParentController();
                } else {
                    return controller;
                }
            }
        }
        return null;
    }

    /**
     * Get a controller by name
     *
     * @return The controller if it can be found or <code>null</code> if it cannot.
     */
    public Controller getControllerbyName(String name){
        if (name == null) return null;
        for (Controller controller : controllers.values()){
            if (name.equals(controller.getName())){
                return controller;
            }
        }
        return null;
    }

    /**
     * This takes the string name of a data svm or a vfiler and the id of either the admin svm or physical vfiler
     * and returns the Controller object from that information.
     *
     * @param name - the name of the data svm or the vfiler
     * @param id - the id of the admin svm or the pfiler the vfiler lives on
     * @return Controller - a data svm with the given name whose parent admin svm matches the given id
     *                        - a vfiler with the given name whose parent pfiler matches the given id
     *                        - null if it cannot be found
     */
    public Controller getChildControllerbyName(String name, String id){
        if (name == null) return null;
        if (id == null) return null;

        Controller controller = getController(id);

        if(controller instanceof HasChildControllers){
            return ((HasChildControllers)controller).getChildControllers().get(name);
        }
        return null;
    }

    public Collection<ControllerConnection> getPersistedConnections() {
        return Collections.unmodifiableCollection(persistor.listAll());
    }

	public Collection<Controller> getControllers() {
		return Collections.unmodifiableCollection(controllers.values());
	}

    public void removeController(Controller controller) {
        removeController(controller.getId());
    }

	public void removeController(String controllerId) {
        // Updated persisted connections
		Controller controller = controllers.remove(controllerId);
        if (controller != null) {
            if (controller instanceof HasChildOntapControllers) {
                // remove children
                for (Controller child : ((HasChildOntapControllers)controller).getChildControllers().values()) {
                    removeController(child);
                }
            }
            persistor.remove(controller.getConnection());
        }
	}

    public void removeAll() {
        controllers.clear();
        persistor.removeAll();
    }

    public void load() throws ConnectionException, DataLoaderException {
        controllers.clear();
        for (ControllerConnection connection : persistor.listAll()) {
            addController(connection);
        }
    }

    public void refreshAll(boolean waitForCompletion) throws DataLoaderException {
        if (controllers.size() > 0) {
            final ExecutorService pool = Executors.newFixedThreadPool(MAX_REFRESH_THREADS, new VscThreadFactory("controllerFactoryPool"));
            for (Controller controller : controllers.values()) {
                final Controller fController = controller;
                pool.execute(new Runnable() {
                    @Override
                    public void run() {
                        fController.refresh();
                    }
                });
            }
            pool.shutdown();
            if (waitForCompletion) {
                try {
                    pool.awaitTermination(MAX_REFRESH_TIMEOUT, TimeUnit.MINUTES);
                }
                catch (InterruptedException e) {
                    LOGGER.warn("refresh for all controllers was interrupted", e);
                }
            }
        }
    }
}
