package com.netapp.offtap3.factories;

import com.google.common.io.BaseEncoding;
import com.netapp.offtap3.ControllerConnection;
import com.netapp.offtap3.OfftapPreferences;
import com.netapp.offtap3.utils.Utils;
import org.apache.log4j.Logger;

import javax.crypto.Cipher;
import javax.crypto.spec.IvParameterSpec;
import javax.crypto.spec.SecretKeySpec;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.nio.charset.Charset;
import java.security.GeneralSecurityException;
import java.security.MessageDigest;
import java.util.Arrays;
import java.util.Collection;
import java.util.Collections;
import java.util.Enumeration;
import java.util.LinkedHashMap;
import java.util.Map;
import java.util.Properties;
import java.util.TreeSet;

public class PropertiesFileConnectionPersistor implements ConnectionPersistor {

    private static final Logger logger = Logger.getLogger(PropertiesFileConnectionPersistor.class);

    private static final String COUNT_PROP = "count";
    private static final String PREFIX = "controller";
    private static final String IP_KEY = "ip";
    private static final String USERNAME_KEY = "username";
    private static final String PASSWORD_KEY = "password";
    private static final String PORT_KEY = "port";
    private static final String SSL_KEY = "ssl";

    private final String fileName;
    private Map<String, ControllerConnection> connections = new LinkedHashMap<String, ControllerConnection>();

    public PropertiesFileConnectionPersistor(String fileName) throws IOException {
        this.fileName = fileName;
        load();
    }


    @Override
    public synchronized void add(ControllerConnection conn) {
        if (!connections.containsKey(conn.getIpAddress())) {
            connections.put(conn.getIpAddress(), conn);
            save();
        }
    }

    protected void load() throws IOException {
        Properties savedConnections = saneOrderProps();
        File propsFile = new File(fileName);
        if (propsFile.exists()) {
            FileInputStream fis = new FileInputStream(propsFile);
            try {
                savedConnections.load(fis);
                int size = Integer.parseInt(savedConnections.getProperty(COUNT_PROP, "0"));
                for (int i = 0; i < size; i++) {
                    ControllerConnection conn = readProperties(PREFIX + i, savedConnections);
                    connections.put(conn.getIpAddress(), conn);
                }
            }
            finally {
                Utils.safeClose(fis);
            }
        }
    }

    protected void save() {
        FileOutputStream fos = null;
        try {
            Properties savedConnections = saneOrderProps();
            savedConnections.setProperty(COUNT_PROP, String.valueOf(connections.size()));
            int i=0;
            for (ControllerConnection conn : connections.values()) {
                writeProperties(conn, PREFIX + i, savedConnections);
                i++;
            }
            fos = new FileOutputStream(fileName);
            savedConnections.store(fos, "AeonFlux persisted connection information");
        }
        catch (IOException e) {
           logger.error("Problem saving file: " + e, e);
        }
        finally {
            Utils.safeFlush(fos);
            Utils.safeClose(fos);
        }
    }

    private void writeProperties(ControllerConnection conn, String prefix, Properties props) {
        props.setProperty(prefixKey(prefix, IP_KEY), conn.getIpAddress());
        props.setProperty(prefixKey(prefix, USERNAME_KEY), conn.getUserName());
        props.setProperty(prefixKey(prefix, PASSWORD_KEY), encryptPassword(conn.getPassword()));
        props.setProperty(prefixKey(prefix, PORT_KEY), String.valueOf(conn.getPort()));
        props.setProperty(prefixKey(prefix, SSL_KEY), String.valueOf(conn.isSSL()));
        Properties prefs = conn.getOfftapPreferences().asProperties();
        for (String key : prefs.stringPropertyNames()) {
            props.setProperty(prefixKey(prefix, key), prefs.getProperty(key));
        }
    }

    private ControllerConnection readProperties(String prefix, Properties props) {
        OfftapPreferences prefs = new OfftapPreferences(props, prefix);
        ControllerConnection conn = ControllerConnectionFactory.getInstance().createConnection(prefs);
        conn.setIpAddress(props.getProperty(prefixKey(prefix, IP_KEY)));
        conn.setUserName(props.getProperty(prefixKey(prefix, USERNAME_KEY)));
        conn.setPassword(decryptPassword(props.getProperty(prefixKey(prefix, PASSWORD_KEY))));
        conn.setSSL(Boolean.parseBoolean(props.getProperty(prefixKey(prefix, SSL_KEY), "false")));
        conn.setPort(Integer.parseInt(props.getProperty(prefixKey(prefix, PORT_KEY), "80")));
        return conn;
    }

    private String prefixKey(String prefix, String key) {
        return prefix + "." + key;
    }

    @Override
    public synchronized boolean remove(ControllerConnection conn) {
        if (connections.containsKey(conn.getIpAddress())) {
            connections.remove(conn.getIpAddress());
            save();
            return true;
        }
        return false;
    }

    @Override
    public synchronized void removeAll() {
        connections.clear();
        save();
    }

    @Override
    public synchronized Collection<ControllerConnection> listAll() {
        return Collections.unmodifiableCollection(connections.values());
    }


    private Cipher getPasswordCipher(int mode) {
        final String TRANSFORMATION = "AES/CBC/PKCS5Padding";
        final String KEY_ALGORITHM = "AES";
        final int KEY_BYTES = 16;
        final int IV_BYTES = 16;
        final Charset UTF8 = Charset.forName("UTF-8");
        try {
            byte[] keyBytes = new byte[KEY_BYTES];
            byte[] ivBytes = new byte[IV_BYTES];
            MessageDigest md5 = MessageDigest.getInstance("MD5");
            // generate the key by running the class name through MD5
            byte[] buffer = md5.digest(getClass().getName().getBytes(UTF8));
            System.arraycopy(buffer, 0, keyBytes, 0, Math.min(buffer.length, KEY_BYTES));
            Arrays.fill(buffer, (byte) 0);
            md5.reset();
            byte[] buffer2 = md5.digest(buffer);
            Arrays.fill(buffer, (byte) 0);
            // generate the IV by running the key back through MD5
            System.arraycopy(buffer2, 0, ivBytes, 0, Math.min(buffer2.length, IV_BYTES));
            Arrays.fill(buffer2, (byte) 0);
            Cipher cipher = Cipher.getInstance(TRANSFORMATION);
            cipher.init(mode, new SecretKeySpec(keyBytes, KEY_ALGORITHM), new IvParameterSpec(ivBytes));
            return cipher;
        }
        catch (GeneralSecurityException e) {
            throw new RuntimeException("Never should happen", e);
        }
    }

    private String decryptPassword(String cipherText) {
        Cipher cipher = getPasswordCipher(Cipher.DECRYPT_MODE);
        try {
            byte[] encrypted = BaseEncoding.base64().decode(cipherText);
            byte[] decrypted = cipher.doFinal(encrypted);
            return new String(decrypted);
        }
        catch (GeneralSecurityException e) {
            throw new RuntimeException("unexpected failure during decryption", e);
        }
    }

    private String encryptPassword(String clearText) {
        Cipher cipher = getPasswordCipher(Cipher.ENCRYPT_MODE);
        byte[] input = clearText.getBytes();

        try {
            byte[] encrypted = cipher.doFinal(input);
            return BaseEncoding.base64().encode(encrypted);
        }
        catch (GeneralSecurityException e) {
            throw new RuntimeException("unexpected failure during encryption", e);
        }
    }

    private Properties saneOrderProps() {
        return new Properties() {
            @Override
            public synchronized Enumeration<Object> keys() {
                return Collections.enumeration(new TreeSet<Object>(super.keySet()));
            }
        };
    }
}
