/*
 * $Id: //depot/prod/aeonflux/R3.0.1/src/main/java/com/netapp/offtap3/internal/AbstractCachingDataLoader.java#1 $
 *
 * Copyright (c) 2014 NetApp, Inc.
 * All rights reserved.
 */
package com.netapp.offtap3.internal;


import com.netapp.offtap3.Refreshable;
import com.netapp.offtap3.Resource;
import com.netapp.offtap3.factories.DataLoaderException;
import com.netapp.offtap3.internal.ontap.dataloader.volume.CModeVolumeDataLoader;
import com.netapp.offtap3.internal.ontap.utils.StatsCollector;

import java.util.Collections;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;

import org.apache.log4j.Logger;

public abstract class AbstractCachingDataLoader<T, S, Z> extends AbstractDataLoader<T> {
	private static final Logger logger = Logger.getLogger(AbstractCachingDataLoader.class);
    private final String cacheKey = getClass().getCanonicalName();
    private final DataCache<S, Z> dataCache;

    public AbstractCachingDataLoader(DataCache<S, Z> dataCache) {
        this.dataCache = dataCache;
    }

    protected DataCache<S, Z> getDataCache() {
        return dataCache;
    }

    @Override
    protected final boolean doLoad(T object, LoadType loadType, Set<Resource> resources) throws DataLoaderException {
        // When refreshing, invalidate all resources.
        Object outerLock = dataCache;
        // per discussion with ChrisM and Dan - initial loads can be broader - but refreshes have a race condition
        // with all the volumes using the same data cache and need heavier synchronization
        if (outerLock == null || LoadType.initial == loadType) {
        	outerLock = new Object();
        }
        long askTime = System.currentTimeMillis();
    	synchronized(outerLock) {
    		StatsCollector.getInstance().addCounter("lock-acdl-outer", System.currentTimeMillis() - askTime);
	        if (loadType == LoadType.refresh) {
	            invalidateDataCache(resources);
	        }
	
	        // Only retrieve resources that are invalid.
	        Set<Resource> invalidResources = getInvalidResources(resources);
	        if (invalidResources.size() > 0) {
	            try {
	                Map<Resource, S> dataMap = retrieve(object, loadType, invalidResources);
	                updateDataCache(dataMap);
	                // its known to be good here
	            } finally {
	                validateDataCache(invalidResources);
	            }
	        }
	
	        // Perform parsing for all resources.
	        boolean result = true;
	        for (Resource resource : resources) {
	        	long askTime2 = System.currentTimeMillis();
	            synchronized (dataCache) {
	            	StatsCollector.getInstance().addCounter("lock-acdl-datacache", System.currentTimeMillis() - askTime2);
	            	S data = dataCache.getData(getCacheKey(resource));
	                // No sense in trying to parse null data.
	                if (data != null) {
	                    if (!parse(object, data, resource)) {
	                        result = false;
	                    }
	                }
	            }
	        }
	        return result;
    	}
    }

    @Override
    protected Refreshable doRefresh(T loadTarget, Resource resource, Refreshable refreshTarget) {
        // When refreshing a specific target, don't mess with the cache. Just retrieve and parse the data directly.
        S data = retrieve(loadTarget, resource, refreshTarget);
        return parse(loadTarget, data, resource, refreshTarget);
    }

    protected Object getCacheKey(Resource resource) {
        return cacheKey + resource.name();
    }

    private void invalidateDataCache(Set<Resource> resources) {
        // Note that the data cache is only synchronized in small blocks so that multithreaded retrievals of data
        // can run without blocking each other.
    	long askTime = System.currentTimeMillis();
        synchronized (dataCache) {
        	StatsCollector.getInstance().addCounter("lock-acdl-datacache", System.currentTimeMillis() - askTime);
            for (Resource resource : resources) {
                dataCache.setValid(getCacheKey(resource), false);
            }
        }
    }

    private void updateDataCache(Map<Resource, S> dataMap) {
    	long askTime = System.currentTimeMillis();
        synchronized (dataCache) {
        	StatsCollector.getInstance().addCounter("lock-acdl-datacache", System.currentTimeMillis() - askTime);
            for (Map.Entry<Resource, S> entry : dataMap.entrySet()) {
                Object resourceCacheKey = getCacheKey(entry.getKey());
                if (entry.getValue() != null) {
                    dataCache.setData(resourceCacheKey, entry.getValue());
                }
            }
        }
    }

    private void validateDataCache(Set<Resource> resources) {
    	long askTime = System.currentTimeMillis();
        synchronized (dataCache) {
        	StatsCollector.getInstance().addCounter("lock-acdl-datacache", System.currentTimeMillis() - askTime);
            for (Resource resource : resources) {
                // Even if we were unable to load the resource, set the validity to true so we don't try to load it
                // over and over.
                Object resourceCacheKey = getCacheKey(resource);
                dataCache.setValid(resourceCacheKey, true);
            }
        }
    }

    private Set<Resource> getInvalidResources(Set<Resource> resources) {
        Set<Resource> invalidResources = new HashSet<Resource>();
        for (Resource resource : resources) {
            if (!dataCache.isValid(getCacheKey(resource))) {
                invalidResources.add(resource);
            }
        }
        return invalidResources;
    }

    /**
     * @throws com.netapp.offtap3.factories.DataLoaderException;
     */
    protected abstract Map<Resource, S> retrieve(T object, LoadType loadType, Set<Resource> resources);

    /**
     * This should be overriden for any data loader that creates Loadable objects and needs to do anything special.
     * @throws com.netapp.offtap3.factories.DataLoaderException;
     */
    protected S retrieve(T loadTarget, Resource resource, Refreshable refreshable) {
        Map<Resource, S> dataMap = retrieve(loadTarget, LoadType.refresh, Collections.singleton(resource));
        return dataMap.get(resource);
    }

    /**
     * @return if the data was successfully parsed (determines if "isLoaded" is now true)
     */
    protected abstract boolean parse(T loadTarget, S data, Resource resource);

    /**
     * This should be overriden for any data loader that creates Loadable objects and is not updating the existing
     * object instance, which is likely all of them.
     */
    protected Refreshable parse(T loadTarget, S data, Resource resource, Refreshable reloadTarget) {
        parse(loadTarget, data, resource);
        return reloadTarget;
    }
}