package com.netapp.offtap3.internal;

import java.util.Date;
import java.util.EnumSet;
import java.util.HashSet;
import java.util.Set;

import org.apache.log4j.Logger;

import com.google.common.base.Objects;
import com.netapp.offtap3.Controller;
import com.netapp.offtap3.ControllerConnection;
import com.netapp.offtap3.ControllerResource;
import com.netapp.offtap3.OfftapPreferences;
import com.netapp.offtap3.Refreshable;
import com.netapp.offtap3.Resource;
import com.netapp.offtap3.StorageProtocol;
import com.netapp.offtap3.Version;
import com.netapp.offtap3.utils.StringUtils;
import com.netapp.offtap3.utils.Utils;

public abstract class AbstractController<D extends Controller> extends
        AbstractRefreshable implements Controller, HasProxy<D> {

    private static final Logger logger = Logger.getLogger(AbstractController.class);

    private String id = "";
    private String systemId = "";
    private String model = "";
    private Version version;
    private ControllerConnection controllerConnection;
	private EnumSet<StorageProtocol> supportedProtocols = EnumSet.noneOf(StorageProtocol.class);
    private Set<String> licenses = new HashSet<String>();
    private Set<String> originalLicensesShouldBeOverridden = licenses;
    private SpaceUsageImpl spaceUsage = new SpaceUsageImpl();

    public AbstractController(ControllerConnection controllerConnection) {
        super(null);
        this.controllerConnection = controllerConnection;
        
    }

    @Override
    public String getId () {
        return id;
    }

    public void setId(String id) {
        this.id = id;
    }

    public void setId(String id,String id2) {
        this.id = id +"-" +id2;
    }

    public String getSystemId() {
        return systemId;
    }

    public void setSystemId(String systemId) {
        this.systemId = systemId;
    }

    public String getModel () {
        return model;
    }

    public void setModel(String model) {
        this.model = model;
    }

    @Override
    public ControllerConnection getConnection() {
    	if (controllerConnection != null &&  
    			StringUtils.isReal(this.getId()) && (controllerConnection instanceof ControllerConnectionImpl)) {
    		((ControllerConnectionImpl) controllerConnection).setFilerIdOfTarget(this.getId());
    	}
        return controllerConnection;
    }

    protected void setInternalConnection(ControllerConnection controllerConnection) {
        this.controllerConnection = controllerConnection;
    }

    @Override
    public SpaceUsageImpl getSpaceUsage() {
        return spaceUsage;
    }

	@Override
	public boolean supportsProtocol(StorageProtocol protocol) {
		return supportedProtocols != null && supportedProtocols.contains(protocol);
	}

    public void addSupportedProtocol(StorageProtocol storageProtocol) {
        supportedProtocols.add(storageProtocol);
    }

    public void removeSupportedProtocol(StorageProtocol storageProtocol) {
        supportedProtocols.remove(storageProtocol);
    }

    @Override
    public boolean equals(Object obj) {
        // WARNING: if you want to override this implementation
        // think about the fact that the parameter (obj) could be a proxy
        // and is NOT an instanceof any of the impl classes, only the
        // interfaces, and this.getClass() will never equal proxy.getClass()
        return obj instanceof Controller && getId().equals(((Controller) obj).getId());
    }


    @Override
    public int hashCode() {
        return Objects.hashCode(getId());
    }

    @Override
    public String toString() {
        return Objects.toStringHelper(this)
                .addValue(id)
                .addValue(getName())
                .addValue(getConnection().getIpAddress())
                .toString();
    }

    @Override
    public Set<String> getLicenses() {
    	if (licenses != null && licenses.size() == 0 && originalLicensesShouldBeOverridden == licenses) {
    		// someone is getting back licenses but nothing was ever set for them - that's bad.
    		logger.debug("Returning original empty licenses as they were never set");
    	}
        return licenses;
    }

    public void setLicenses(Set<String> licenses) {    	
        this.licenses = licenses;
    }

    @Override
    public boolean isLicensed(String feature) {
        Set<String> licenses = getLicenses();
        if (licenses != null){
            if (licenses.contains(feature)) {
                return true;
            }

            // ok - we wanted snapmirror, and the filer had snapmirror_dp.  So lets see if
            // anything in the list startsWithIgnoreCase what we're looking for.
            // per conversation with Brian mcgiverin, this is a ford taurus level solution,
            // better than a ford fiesta, but not a caddillac
            String originalFeature = feature;
            for (String curr : licenses) {
                String originalCurr = curr;
                curr = curr.replaceAll("-", "_");
                curr = curr.toUpperCase();
                feature = feature.replaceAll("-", "_");
                feature = feature.toUpperCase();
                if (curr.startsWith(feature)) {
                    logger.debug("Considering " + originalCurr + " to be acceptable match for " + originalFeature);
                    return true;
                }
            }
        }
        return false;
    }

    @Override
    public Version getVersion() { return version;}

    public void setVersion(Version version) {
        this.version = version;
    }

    public OfftapPreferences getOfftapPreferences() {
        return controllerConnection.getOfftapPreferences();
    }

    @Override
    public void refresh(Resource... resources) {
        // Handled by the proxy.
    }

    @Override
    public Refreshable refresh(Refreshable refreshable) {
        // Handled by the proxy
        return null;
    }

    /**
     * Special case for Controller objects. Instead of reload() pass the call to reload(this) against the loader, it
     * just reloads anything that has already been loaded on the controller.
     */
    @Override
    public Refreshable refresh() {
        getProxy().refresh(ControllerResource.values());
        setLastLoad(new Date());
        return this;
    }
}
