/*
 * $Id: //depot/prod/aeonflux/R3.0.1/src/main/java/com/netapp/offtap3/internal/AbstractDataLoader.java#1 $
 *
 * Copyright (c) 2014 NetApp, Inc.
 * All rights reserved.
 */
package com.netapp.offtap3.internal;

import com.netapp.offtap3.Refreshable;
import com.netapp.offtap3.Resource;
import com.netapp.offtap3.factories.DataLoader;
import com.netapp.offtap3.factories.DataLoaderException;
import com.netapp.offtap3.internal.ontap.utils.StatsCollector;

import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

public abstract class AbstractDataLoader<T> implements DataLoader<T> {

    private final Map<Resource, Boolean> loaded = Collections.synchronizedMap(new HashMap<Resource, Boolean>());
    private final Object loadingLock = new Object();
    private final String loadingLockString = "lock-adl-loading-" + this.getClass().getName();

    public final void load(T object, LoadType loadType, Set<Resource> resources) throws DataLoaderException {
        // only one thread can invoke the data loader at a time
    	long askTime = System.currentTimeMillis();
        synchronized(loadingLock) {
        	long now = System.currentTimeMillis() - askTime;
        	StatsCollector.getInstance().addCounter(loadingLockString, now);
            if (loadType == LoadType.refresh) {
                invalidate();
            }
           // check to be sure that the data was not loaded while we were blocked
            Set<Resource> unloadedResources = getUnloadedResources(resources);
            if (unloadedResources.size() > 0) {
                boolean result = doLoad(object, loadType, unloadedResources);
                for (Resource resource : unloadedResources) {
                    loaded.put(resource, result);
                }
            }
        }
    }

    @Override
    public final Refreshable refresh(T loadTarget, Resource resource, Refreshable refreshTarget) {    	
        Refreshable result = doRefresh(loadTarget, resource, refreshTarget);
        loaded.put(resource, true);
        return result;
    }

    protected abstract boolean doLoad(T object, LoadType loadType, Set<Resource> resources) throws DataLoaderException;

    /**
     * Should be overriden by any data loader that creates Loadable objects.
     */
    protected Refreshable doRefresh(T loadTarget, Resource resource, Refreshable refreshTarget) {
        doLoad(loadTarget, LoadType.refresh, Collections.singleton(resource));
        return refreshTarget;
    }

    @Override
    public final boolean isLoaded(Set<Resource> resources) {
        for (Resource resource : resources) {
            if (!isLoaded(resource)) {
                return false;
            }
        }
        return true;
    }

    private Set<Resource> getUnloadedResources(Set<Resource> resources) {
        Set<Resource> unloadedResources = new HashSet<Resource>();
        for (Resource resource : resources) {
            if (!isLoaded(resource)) {
                unloadedResources.add(resource);
            }
        }
        return unloadedResources;
    }

    private boolean isLoaded(Resource resource) {
        Boolean result = loaded.get(resource);
        return result == null ? false : result;
    }

    @Override
    public void invalidate() {
    	long askTime = System.currentTimeMillis();
        synchronized (loaded) {
        	StatsCollector.getInstance().addCounter("lock-adl-loaded", System.currentTimeMillis() - askTime);
            for (Resource resource : loaded.keySet()) {
                loaded.put(resource, false);
            }
        }
    }

    public Refreshable getFromList(List<? extends Refreshable> loadables, String name) {
        for (Refreshable refreshable : loadables) {
            if (refreshable.getName().equals(name)) {
                return refreshable;
            }
        }
        return null;
    }

    protected void setLoaded(Resource resource, boolean isLoaded) {
        loaded.put(resource, isLoaded);
    }
}
