package com.netapp.offtap3.internal;

import java.lang.reflect.InvocationHandler;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.lang.reflect.Proxy;
import java.util.Arrays;
import java.util.Collection;
import java.util.Collections;
import java.util.HashSet;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.SortedMap;
import java.util.SortedSet;
import java.util.concurrent.CompletionService;
import java.util.concurrent.ExecutionException;
import java.util.concurrent.ExecutorCompletionService;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.concurrent.Future;

import org.apache.log4j.Logger;

import com.google.common.collect.HashMultimap;
import com.netapp.offtap3.DiscoveryProgressListener;
import com.netapp.offtap3.HasName;
import com.netapp.offtap3.OfftapPreferences;
import com.netapp.offtap3.ProxyType;
import com.netapp.offtap3.Refreshable;
import com.netapp.offtap3.Resource;
import com.netapp.offtap3.StorageSystem;
import com.netapp.offtap3.factories.DataLoader;
import com.netapp.offtap3.factories.DataLoaderException;
import com.netapp.offtap3.internal.ontap.dataloader.HasResourceDependencies;
import com.netapp.offtap3.internal.ontap.support.ZAPIStatsCounter;
import com.netapp.offtap3.ontap.OntapController;
import com.netapp.offtap3.ontap.utils.OntapCallable;
import com.netapp.offtap3.utils.WhatsHappeningMonitor;

public abstract class OfftapProxy<I, T extends HasProxy<I>> implements InvocationHandler {

    private final static Logger LOGGER = Logger.getLogger(OfftapProxy.class);
    private static Method getNameMethod;

    static {
        try {
            getNameMethod = HasName.class.getDeclaredMethod("getName");
        } catch (NoSuchMethodException e) {
            throw new RuntimeException("YOU HAVE BIG PROBLEMS.  getName() is not on HasName.class");
        }
    }

    private final ProxyType proxyType;
    private T proxiedObject;
    private Set<StorageSystem> storageSystems;

    private Map<Resource, DataLoader<? super T>> loaderMap = new LinkedHashMap<Resource, DataLoader<? super T>>();

    public OfftapProxy(ProxyType proxyType) {
        this.proxyType = proxyType;
    }

    @SuppressWarnings("unchecked")
    public static <S extends OfftapProxy<I, T>, I, T extends HasProxy<I>> I create(Class<I> proxyClass, Class<S> clazz, T proxiedObject) {
        InvocationHandler proxyObj = instantiate(clazz, proxiedObject);
        return (I) Proxy.newProxyInstance(proxyClass.getClassLoader(), new Class[]{proxyClass}, proxyObj);
    }

    @SuppressWarnings("unchecked")
    public static <S extends OfftapProxy<I, T>, I, T extends HasProxy<I>> I create(Class[] proxyClasses, Class<S> clazz, T proxiedObject) {
        InvocationHandler proxyObj = instantiate(clazz, proxiedObject);
        return (I) Proxy.newProxyInstance(clazz.getClassLoader(), proxyClasses, proxyObj);
    }

    protected static <S extends OfftapProxy<I, T>, I, T extends HasProxy<I>> S instantiate(Class<S> clazz, T proxiedObject) {
        S proxy = build(clazz, proxiedObject);
        proxy.initialize();
        return proxy;
    }

    protected static <S extends OfftapProxy<I, T>, I, T extends HasProxy<I>> S build(Class<S> clazz, T proxiedObject) {
        try {
            S proxy = clazz.newInstance();
            proxy.setProxiedObject(proxiedObject);

            return proxy;
        } catch (Exception e) {
            throw new RuntimeException(e);
        }
    }

    protected void initialize() {
        storageSystems = proxiedObject.getStorageSystems();
        addDataLoaders();
        try {
            initialLoad();
        } catch (DataLoaderException e) {
            LOGGER.error("Error occurred while performing initial load: " + e.getMessage(), e);
        }
    }

    public T getProxiedObject() {
        return proxiedObject;
    }

    public void setProxiedObject(T proxiedObject) {
        this.proxiedObject = proxiedObject;
    }

    protected OfftapPreferences getOfftapPreferences() {
        return proxiedObject.getConnection().getOfftapPreferences();
    }

    protected abstract void addDataLoaders();

    protected Set<Resource> getInitialResources() {
        return getOfftapPreferences().getInitialResources(storageSystems, proxyType);
    }

    protected ExecutorService getExecutorService() {
        return Executors.newSingleThreadExecutor();
    }

    protected void initialLoad() throws DataLoaderException {
        Set<Resource> resourcesToLoad = getInitialResources();

        ExecutorService pool = getExecutorService();
        try {
	        CompletionService<Set<Resource>> completionService = new ExecutorCompletionService<Set<Resource>>(pool);
	        Set<Set<Resource>> reducedResources = reduceByDataLoader(resourcesToLoad);
	
	        Collection<DiscoveryProgressListener> listeners = getListeners();
	
	        DiscoveryProgressJobImpl job = null;
	        if (!listeners.isEmpty()) {
	            job = new DiscoveryProgressJobImpl(reducedResources.size(), listeners);
	            insureName(job);
	            job.notifyStart();
	        }
	
	        for (final Set<Resource> resourceSet : reducedResources) {
	            completionService.submit(new OntapCallable<Set<Resource>>() {
	            	public Set<Resource> callLocal() throws DataLoaderException {
	                	String requestor = "";
	                    if (!ZAPIStatsCounter.getInstance().getRecordedKey().isEmpty()) {
	                        requestor = ZAPIStatsCounter.getInstance().getRecordedKey() + "";
	                    }
	                    String parallelCallbacks = "parallel-callbacks";
	                    WhatsHappeningMonitor.getMonitor(parallelCallbacks).setMaxAllowableSize(10000);
	                    WhatsHappeningMonitor.getMonitor(parallelCallbacks)
	                                         .registerStartCall(requestor, resourceSet + "");
	                    try {
		                    executeDataLoader(DataLoader.LoadType.initial, resourceSet);
		                    return resourceSet;
	                    } finally {
	                    	WhatsHappeningMonitor.getMonitor(parallelCallbacks).registerEndCall();
	                    }
	                }
	
	            });
	        }
	
	        // Wait until all data
	
	        for (int r = 0; r < reducedResources.size(); r++) {
	            try {
	                Future<Set<Resource>> future = completionService.take();
	                Set<Resource> completedResource = future.get();
	                if (job != null) {
	                    job.setStatus(completedResource.toString());
	                    job.notifyUpdate();
	                }
	                LOGGER.trace(proxiedObject + " has successfully loaded: " + completedResource);
	            } catch (InterruptedException e) {
	                LOGGER.error(proxiedObject + " failed when waiting for resources to load", e);
	            } catch (ExecutionException e) {
	                if (proxiedObject instanceof OntapController) {
	                    ((OntapController) proxiedObject).getConnection()
	                                                     .addDiscoveryError("Failure loading resource", e.getCause());
	                }
	                LOGGER.error(proxiedObject + " failed loading resource", e.getCause());
	            }
	        }
	
	        if (job != null) {
	            job.notifyFinished();
	        }
        } finally {
        	pool.shutdown();
        }
    }

    private Set<Set<Resource>> reduceByDataLoader(Set<Resource> resources) {
        HashMultimap<DataLoader, Resource> resourceMap = HashMultimap.create();
        for (Resource resource : resources) {
            DataLoader loader = getDataLoader(resource);
            if (loader != null) {
                resourceMap.put(loader, resource);
            } else {
                LOGGER.debug("No loader available for resource " + resource + " on proxy object " + proxiedObject);
            }
        }
        Set<Set<Resource>> reducedSet = new HashSet<Set<Resource>>();
        for (DataLoader loader : resourceMap.keySet()) {
            reducedSet.add(resourceMap.get(loader));
        }
        return reducedSet;
    }

    protected abstract Resource getResourceAnnotation(Method method);

    @Override
    public Object invoke(Object proxy, Method method, Object[] args) throws Throwable {
        if (isRefreshMethod(method, args)) {
            Set<Resource> resourcesToRefresh = getResourcesToRefresh(args);
            Refreshable refreshable = getRefreshable(args);
            // Reduce the resources to sets that have the same dataloader for efficiency.
            Set<Set<Resource>> resourceSets = reduceByDataLoader(resourcesToRefresh);
            for (Set<Resource> resourceSet : resourceSets) {
                refreshable = executeDataLoader(DataLoader.LoadType.refresh, resourceSet, refreshable);
                // If we got back a refreshable instance, go ahead and return it. Otherwise, continue on so that the
                // lastRefresh date can be set on the object.
                if (refreshable != null) {
                    return refreshable;
                }
            }
        } else {
            Resource resource = getResourceAnnotation(method);
            if (resource != null) {
                executeDataLoader(DataLoader.LoadType.lazy, Collections.singleton(resource));
            }
        }
        Object toReturn;

        try {
            toReturn = method.invoke(proxiedObject, args);
        } catch (InvocationTargetException e) {
            // while a Java Proxy/InvocationHandler understands to not wrap exceptions,
            // java.lang.reflect.Method is not as smart.
            // so, we need to unpack it here, to insure the expected exception is thrown.
            throw e.getTargetException();
        }

        return makeCollectionUnmodifiable(toReturn);
    }

    private boolean isRefreshMethod(Method method, Object[] args) {
        return method.getName().equals("refresh") &&
                (args != null && args.length == 1 && (args[0] instanceof Resource[] || args[0] instanceof Refreshable));
    }

    /**
     * @param args Precondition is that the args object is either a Refreshable or an array of Resources.
     * @return The set of resources to refresh.
     */
    protected Set<Resource> getResourcesToRefresh(Object[] args) {
        Set<Resource> resourcesToRefresh = new HashSet<Resource>();
        if (args[0] instanceof Refreshable) {
            Resource resource = getResource((Refreshable)args[0]);
            if (resource != null) {
                resourcesToRefresh.add(resource);
            } else {
                LOGGER.debug("Tried to refresh " + args[0] + ", but no resource located for it.");
            }
        } else {
            // in this case, caller specifically requested a refresh of certain resources
            resourcesToRefresh.addAll(Arrays.asList((Resource[]) args[0]));
            // We only want to refresh the already loaded resources.
            Set<Resource> loadedResources = getLoadedResources();
            if (!loadedResources.equals(resourcesToRefresh)) {
                Set<Resource> resourcesNotToRefresh = new HashSet<Resource>(resourcesToRefresh);
                resourcesNotToRefresh.removeAll(loadedResources);
                LOGGER.debug("Did not refresh the following resources on object \"" + proxiedObject + "\" because they do not appear to be loaded: " + resourcesNotToRefresh);
            }
            resourcesToRefresh.retainAll(loadedResources);
        }
        return resourcesToRefresh;
    }

    private Refreshable getRefreshable(Object[] args) {
        if (args != null && args.length == 1 && args[0] instanceof Refreshable) {
            return (Refreshable) args[0];
        } else {
            return null;
        }
    }

    /**
     *@throws com.netapp.offtap3.factories.DataLoaderException
     */
    protected void executeDataLoader(DataLoader.LoadType loadType, Set<Resource> resources) {
        executeDataLoader(loadType, resources, null);
    }

    /**
     *@throws com.netapp.offtap3.factories.DataLoaderException
     */
    protected Refreshable executeDataLoader(DataLoader.LoadType loadType, Set<Resource> resources, Refreshable refreshable) {
        DataLoader<? super T> dataLoader = getDataLoader(resources);

        if (dataLoader == null) {
            // TODO aeonflux-next some resources only exist for certain types of controllers,
            // and this should never happen, but let's not throw an exception until we are
            // further along in development, when we have more of the resource and data loaders implemented
            LOGGER.warn(proxiedObject + " does not have a data loader for type(s) '" + resources + "'");
            return null;
        }

        boolean refresh = loadType == DataLoader.LoadType.refresh;
        Refreshable refreshedRefreshable = null;
        if (!dataLoader.isLoaded(resources) || refresh) {
            if (dataLoader instanceof HasResourceDependencies) {
                for (Resource dependency : ((HasResourceDependencies) dataLoader).getResources(refreshable)) {
                    LOGGER.debug(proxiedObject + " will now check that " + dependency +
                                         " has been loaded before starting " + resources);
                    executeDataLoader(loadType, Collections.singleton(dependency));
                }
            }

            LOGGER.debug(
                    String.format("%s is going to start %s: %s ", proxiedObject, refresh ? "refreshing" : "loading",
                                  resources));

            if (refresh && refreshable != null && resources.size() == 1) {
                refreshedRefreshable = dataLoader.refresh(proxiedObject, resources.iterator().next(), refreshable);
            } else {
                dataLoader.load(proxiedObject, loadType, resources);
            }

            LOGGER.debug(
                    String.format("%s finished %s : %s", proxiedObject, refresh ? "refreshing" : "loading", resources));
        }
        return refreshedRefreshable;
    }

    protected void addDataLoader(Resource type, DataLoader<? super T> dataLoader) {
        loaderMap.put(type, dataLoader);
    }

    /**
     * This should be overriden for any proxy that creates loadable object for which refresh can be called.
     */
    protected Resource getResource(Refreshable refreshable) {
        return null;
    }

    /**
     * Returns the data loader for the resource.
     *
     * @param resource The resource.
     */
    public DataLoader<? super T> getDataLoader(Resource resource) {
        return loaderMap.get(resource);
    }

    /**
     * Returns the data loader for the first resource. It is assumed that if multiple resources are specified that they all use the same data loader.
     *
     * @param resources At least one resource.
     */
    public DataLoader<? super T> getDataLoader(Set<Resource> resources) {
        return getDataLoader(resources.iterator().next());
    }

    /**
     * Retrieves a collection of discovery progress listeners from the connection on the proxied object.
     */
    private Collection<DiscoveryProgressListener> getListeners() {
        return proxiedObject.getConnection().getDiscoveryProgressListeners(proxyType);
    }

    private Object makeCollectionUnmodifiable(Object toReturn) {
    	if (this.getProxiedObject().getConnection().getOfftapPreferences().getWantsUnmodifiableCollections()) {
	        if (toReturn instanceof Collection) {
	            if (toReturn instanceof List) {
	                return Collections.unmodifiableList((List) toReturn);
	            } else if (toReturn instanceof SortedSet) {
	                return Collections.unmodifiableSortedSet((SortedSet) toReturn);
	            } else if (toReturn instanceof Set) {
	                return Collections.unmodifiableSet((Set) toReturn);
	            } else {
	                return Collections.unmodifiableCollection((Collection) toReturn);
	            }
	        } else if (toReturn instanceof Map) {
	            if (toReturn instanceof SortedMap) {
	                return Collections.unmodifiableSortedMap((SortedMap) toReturn);
	            } else {
	                return Collections.unmodifiableMap((Map) toReturn);
	            }
	        }
    	}
        return toReturn;
    }

    /**
     * Uses the name of the proxied object for the job if the job does not already have a name.
     */
    private void insureName(DiscoveryProgressJobImpl job) {
        if (job.getName() == null) {
            try {
                job.setName((String) getNameMethod.invoke(getProxiedObject()));
            } catch (Exception e) {
                LOGGER.warn("Unable to get a name for this proxy: " + getProxiedObject());
            }
        }
    }

    protected Set<Resource> getLoadedResources() {
        Set<Resource> loadedResources = new HashSet<Resource>();
        for (Resource resource : loaderMap.keySet()) {
            DataLoader loader = loaderMap.get(resource);
            if (loader.isLoaded(Collections.singleton(resource))) {
                loadedResources.add(resource);
            }
        }
        return loadedResources;
    }

}
