/*
 * $Id: //depot/prod/aeonflux/R3.0.1/src/main/java/com/netapp/offtap3/internal/UnixUserImpl.java#1 $
 *
 * Copyright (c) 2014 NetApp, Inc.
 * All rights reserved.
 */
package com.netapp.offtap3.internal;

import com.netapp.offtap3.UnixGroup;
import com.netapp.offtap3.UnixUser;
import com.netapp.offtap3.UserAndGroupOperationException;
import com.netapp.offtap3.internal.ontap.support.ZapiCache;
import com.netapp.offtap3.ontap.ApiFailedException;
import com.netapp.offtap3.ontap.NaElement;
import org.apache.log4j.Logger;

import java.util.Collections;
import java.util.HashSet;
import java.util.Set;

public class UnixUserImpl implements UnixUser {
    private static final Logger LOGGER = Logger.getLogger(UnixUserImpl.class);

    private int id;
    private String name;
    private UnixGroupImpl primaryGroup;
    private Set<UnixGroup> secondaryGroups = new HashSet<UnixGroup>();

    public UnixUserImpl(UnixGroupImpl primaryGroup, String name, int id) {
        this.name = name;
        this.id = id;
        this.primaryGroup = primaryGroup;
    }

    @Override
    public int getId() {
        return id;
    }

    @Override
    public UnixGroup getPrimaryGroup() {
        return primaryGroup;
    }

    @Override
    public Set<UnixGroup> getSecondaryGroups() {
        return Collections.unmodifiableSet(secondaryGroups);
    }

    public void setSecondaryGroups(Set<UnixGroup> groups) {
        this.secondaryGroups = groups;
    }

    @Override
    public void addSecondaryGroup(UnixGroup group) throws UserAndGroupOperationException {
        LOGGER.debug("addSecondaryGroup: user=" + getName() + ", group=" + group.getName());
        NaElement input = new NaElement("name-mapping-unix-group-add-user");
        input.addNewChild("group-name", group.getName());
        input.addNewChild("user-name", getName());
        try {
            NaElement output = primaryGroup.getControllerImpl().invokeZapi(input, ZapiCache.NO_CACHE);
            if ("passed".equals(output.getAttr("status"))) {
                secondaryGroups.add(group);
                ((UnixGroupImpl)group).addUser(this);
            }
        } catch (ApiFailedException e) {
            String message = "Unable to add unix group " + group.getName() + " to secondary groups for user " + getName() + ".";
            LOGGER.error("addSecondaryGroup: " + message);
            throw new UserAndGroupOperationException(message, e);
        }
    }

    @Override
    public void removeSecondaryGroup(UnixGroup group) throws UserAndGroupOperationException {
        LOGGER.debug("removeSecondaryGroup: user=" + getName() + ", group=" + group.getName());
        NaElement input = new NaElement("name-mapping-unix-group-delete-user");
        input.addNewChild("group-name", group.getName());
        input.addNewChild("user-name", getName());
        try {
            NaElement output = primaryGroup.getControllerImpl().invokeZapi(input, ZapiCache.NO_CACHE);
            if ("passed".equals(output.getAttr("status"))) {
                secondaryGroups.remove(group);
                ((UnixGroupImpl)group).removeUser(getName());
            }
        } catch (ApiFailedException e) {
            String message = "Unable to remove unix group " + group.getName() + " from secondary groups for user " + getName() + ".";
            LOGGER.error("removeSecondaryGroup: " + message);
            throw new UserAndGroupOperationException(message, e);
        }
    }

    public void setPrimaryGroup(UnixGroupImpl group) {
        this.primaryGroup = group;
    }

    @Override
    public String getName() {
        return name;
    }

    @Override
    public void destroy() throws UserAndGroupOperationException {
        LOGGER.debug("destroy: " + getName());

        boolean destroyedGroup = destroyPrivateGroup();

        NaElement input = new NaElement("name-mapping-unix-user-destroy");
        input.addNewChild("user-name", getName());
        try {
            NaElement output = primaryGroup.getControllerImpl().invokeZapi(input, ZapiCache.NO_CACHE);
            if ("passed".equals(output.getAttr("status"))) {
                primaryGroup.getControllerImpl().getUnixUsers().remove(name);
                if (!destroyedGroup) {
                    primaryGroup.removeUser(name);
                }
                // Remove this user from its secondary groups.
                for (UnixGroup group : secondaryGroups) {
                    removeSecondaryGroup(group);
                }
            }
            LOGGER.info("Destroyed " + getName() + ":" + output.getAttr("status") + ".");
        } catch (ApiFailedException e) {
            String message = "Unable to destroy unix group " + getName() + " on data svm " + primaryGroup.getControllerImpl().getName() + ".";
            LOGGER.error("destroy: " + message);
            throw new UserAndGroupOperationException(message, e);
        }
    }

    private boolean destroyPrivateGroup() throws UserAndGroupOperationException {
        // Check if the primary group has the same name and this user as the only user.
        if (primaryGroup.getName().equals(getName()) && primaryGroup.getUsers().size() == 1 && primaryGroup.getUsers().containsKey(getName())) {
            // We have a private group. Remove this user and destroy it.
            primaryGroup.removeUser(getName());
            primaryGroup.destroy();
            return true;
        } else {
            return false;
        }
    }
}
