/*
 * $Id$
 *
 * Copyright (c) 2015 NetApp, Inc.
 * All rights reserved.
 */
package com.netapp.offtap3.internal.mars;

import java.io.IOException;
import java.util.Map;

import org.apache.log4j.Logger;

import com.fasterxml.jackson.core.JsonProcessingException;
import com.fasterxml.jackson.core.type.TypeReference;
import com.fasterxml.jackson.databind.ObjectMapper;
import com.google.common.io.BaseEncoding;
import com.netapp.offtap3.ControllerConnection;
import com.netapp.offtap3.OfftapPreferences;
import com.netapp.offtap3.factories.ControllerType;
import com.netapp.offtap3.internal.ControllerConnectionImpl;
import com.netapp.offtap3.internal.mars.webmapi.model.common.ResponseContentType;
import com.netapp.offtap3.internal.mars.webmapi.util.MarsError;
import com.netapp.offtap3.internal.mars.webmapi.util.WebMapiClient;
import com.netapp.offtap3.internal.mars.webmapi.util.WebMapiClient.HttpConnectionType;
import com.netapp.offtap3.mars.MarsApiFailedException;
import com.netapp.offtap3.utils.WhatsHappeningMonitor;

public class MarsConnection extends ControllerConnectionImpl {
    private static final Logger logger = Logger.getLogger(MarsConnection.class);
    public static final String MARS_API_VERSION = "1.1";
    private static final String REQUEST_ID = "request1234";

    private String sessionId;
    private WebMapiClient client;
    private ObjectMapper mapper = new ObjectMapper();

    public MarsConnection(OfftapPreferences preferences) {
        super(preferences);
        setControllerType(ControllerType.MARS);
    }

    public MarsConnection(ControllerConnection controllerConnection) {
        super(controllerConnection);
        setControllerType(ControllerType.MARS);
        if (controllerConnection instanceof MarsConnection) {
            MarsConnection marsConnection = (MarsConnection) controllerConnection;
            sessionId = marsConnection.sessionId;
            client = marsConnection.client;
        }
    }

    public String login() throws MarsApiFailedException {
        logger.debug("Initiating login on Mars controller " + getIpAddress());
        WebMapiClient loginClient = new WebMapiClient(getIpAddress(), MARS_API_VERSION, "", getPort(), isSSL());

        String loginBody = "{\"username\":\"" + getUserName() + "\", \"password\":\"" + BaseEncoding.base64()
                                                                                                    .encode(getPassword()
                                                                                                                    .getBytes())
                + "\"}";
        WhatsHappeningMonitor.getMonitor("mars-login").registerStartCall("mars-connection", getIpAddress());
        String sessionInfo = loginClient.connect(getRequestId(), "sessions", null, HttpConnectionType.POST, loginBody);
        logger.debug("mars session info = " + sessionInfo);
        WhatsHappeningMonitor.getMonitor("mars-login").registerEndCall();
        if (sessionInfo != null) {
            ObjectMapper objectMapper = new ObjectMapper();

            Map<String, Object> sessionMap;
            try {
                sessionMap = objectMapper.readValue(sessionInfo.getBytes(), new TypeReference<Map<String, Object>>() {
                });
            } catch (IOException e) {
                throw new MarsApiFailedException(MarsError.RESPONSE_DECODE_FAILED, e);
            }
            sessionId = (String) sessionMap.get("result");
            client = new WebMapiClient(getIpAddress(), MARS_API_VERSION, sessionId, getPort(), isSSL());
            return sessionId;
        } else {
            return null;
        }
    }

    public String get(String path) throws MarsApiFailedException {
        logger.debug("Initiating Mars REST GET request against path " + path);
        String result = get(path, (String)null);
        logger.debug("Received result: " + result);
        return result;
    }

    public String get(String path, String query) throws MarsApiFailedException {
        logger.debug("Initiating Mars REST GET request against path " + path + " with query " + query);
        String result = connect(getRequestId(), path, query, HttpConnectionType.GET, null);
        logger.debug("Received result: " + result);
        return result;
    }

    public <T extends ResponseContentType> T get(String path, TypeReference<T> reference) throws MarsApiFailedException {
        String result = get(path);
        return convertToType(result, reference);
    }

    public <T extends ResponseContentType> T post(String path, Object bodyObject, TypeReference<T> typeReference) throws MarsApiFailedException {
        String body = convertToString(bodyObject);
        return convertToType(post(path, body), typeReference);
    }

    public String post(String path, String body) throws MarsApiFailedException {
        logger.debug("Initiating Mars REST POST request against path " + path + " with body " + body);
        String result = connect(getRequestId(), path, null, HttpConnectionType.POST, body);
        logger.debug("Received result: " + result);
        return result;
    }

    public <T extends ResponseContentType> T put(String path, Object bodyObject, TypeReference<T> typeReference) throws MarsApiFailedException {
        String body = convertToString(bodyObject);
        return convertToType(put(path, body), typeReference);
    }

    public String put(String path, String body) throws MarsApiFailedException {
        logger.debug("Initiating Mars REST PUT request against path " + path + " with body " + body);
        String result = connect(getRequestId(), path, null, HttpConnectionType.PUT, body);
        logger.debug("Received result: " + result);
        return result;
    }

    public <T extends ResponseContentType> T delete(String path, TypeReference<T> typeReference) throws MarsApiFailedException {
        return convertToType(delete(path), typeReference);
    }

    public String delete(String path) throws MarsApiFailedException {
        return delete(path, (String) null);
    }

    public <T extends ResponseContentType> T delete(String path, String query, TypeReference<T> typeReference) throws MarsApiFailedException {
        return convertToType(delete(path, query), typeReference);
    }

    public String delete(String path, String query) throws MarsApiFailedException {
        logger.debug("Initiating Mars REST DELETE request against path " + path + " with query " + query);
        String result = connect(getRequestId(), path, query, HttpConnectionType.DELETE, null);
        logger.debug("Received result: " + result);
        return result;
    }

    private String connect(String requestId, String path, String query, HttpConnectionType connectionType, String body) throws MarsApiFailedException {
        try {
            return client.connect(requestId, path, query, connectionType, body);
        } catch (MarsApiFailedException e) {
            if (e.getError() == MarsError.UNAUTHORIZED) {
                logger.debug("Retrying command after received and unauthorized exception...");
                // Session may have timed out. Try to login again and if successful, retry the command.
                try {
                    login();
                    return client.connect(requestId, path, query, connectionType, body);
                } catch(MarsApiFailedException e2) {
                    // If we fail again, throw the original exception to avoid confusion over some exception during login that may be unrelated
                    // to the original but likely more pertinent exception.
                    throw e;
                }
            } else {
                throw e;
            }
        }
    }

    private <T extends ResponseContentType> T convertToType(String value, TypeReference<T> reference) throws MarsApiFailedException {
        T response;
        try {
            response = mapper.readValue(value.getBytes(), reference);
        } catch (IOException e) {
            throw new MarsApiFailedException(MarsError.RESPONSE_DECODE_FAILED, e);
        }
        if (response.getError() != null) {
            throw new MarsApiFailedException(response.getError());
        }
        return response;
    }

    private String convertToString(Object value) throws MarsApiFailedException {
        try {
            return mapper.writeValueAsString(value);
        } catch (JsonProcessingException e) {
            throw new MarsApiFailedException(MarsError.BODY_ENCODE_FAILED, e);
        }
    }

    public String getRequestId() {
        return REQUEST_ID;
    }
}
