/*
 * $Id$
 *
 * Copyright (c) 2015 NetApp, Inc.
 * All rights reserved.
 */
package com.netapp.offtap3.internal.mars;

import java.util.Collections;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.concurrent.CopyOnWriteArrayList;

import org.apache.log4j.Logger;

import com.fasterxml.jackson.core.type.TypeReference;
import com.netapp.offtap3.ControllerConnection;
import com.netapp.offtap3.FcInterface;
import com.netapp.offtap3.FcpAdapter;
import com.netapp.offtap3.IGroup;
import com.netapp.offtap3.IGroupSpec;
import com.netapp.offtap3.IGroupSpec.Builder;
import com.netapp.offtap3.Lun;
import com.netapp.offtap3.LunSpec;
import com.netapp.offtap3.Size;
import com.netapp.offtap3.StorageProtocol;
import com.netapp.offtap3.StorageSystem;
import com.netapp.offtap3.internal.AbstractController;
import com.netapp.offtap3.internal.IGroupSpecImpl.BuilderImpl;
import com.netapp.offtap3.internal.OfftapProxy;
import com.netapp.offtap3.internal.mars.proxy.MarsControllerProxy;
import com.netapp.offtap3.internal.mars.webmapi.model.common.ResponseContentType;
import com.netapp.offtap3.internal.mars.webmapi.model.domain.IGroupCreateInputType;
import com.netapp.offtap3.internal.mars.webmapi.model.domain.LunCreateInputType;
import com.netapp.offtap3.internal.mars.webmapi.model.domain.OsType;
import com.netapp.offtap3.internal.mars.webmapi.model.domain.UuidType;
import com.netapp.offtap3.mars.LunMaxExtentSize;
import com.netapp.offtap3.mars.MarsApiFailedException;
import com.netapp.offtap3.mars.MarsController;
import com.netapp.offtap3.mars.MarsLun;
import com.netapp.offtap3.mars.MarsLunSpec;
import com.netapp.offtap3.ontap.IGroupOperationException;
import com.netapp.offtap3.ontap.LunOperationException;
import com.netapp.offtap3.utils.SafeConcurrentHashMap;

public class MarsControllerImpl extends AbstractController<MarsController> implements MarsController {
    private static final Logger LOGGER = Logger.getLogger(MarsControllerImpl.class);
    private static final Set<StorageSystem> STORAGE_SYSTEMS = Collections.singleton(StorageSystem.mars);

    private static final long DEFAULT_LUN_BLOCK_SIZE = 512;

    private MarsController proxy;
    private Map<String, MarsLun> luns = new SafeConcurrentHashMap<String, MarsLun>();
    private List<FcpAdapter> fcpAdapters = new CopyOnWriteArrayList<>();
    private List<FcInterface> fcInterfaces = new CopyOnWriteArrayList<>();
    private List<IGroup> igroups;

    public MarsControllerImpl(ControllerConnection controllerConnection) {
        super(new MarsConnection(controllerConnection));
        // Mars only supports the fiber channel protocol. Will need to update
        // this if it changes in the future.
        addSupportedProtocol(StorageProtocol.FCP);

    }

    @Override
    public void setConnection(ControllerConnection connection) {
        super.setInternalConnection(new MarsConnection(connection));
    }

    @Override
    public MarsConnection getConnection() {
        return (MarsConnection) super.getConnection();
    }

    @Override
    public MarsLunSpec.Builder getLunSpecBuilder(String name, Size size) {
        return new MarsLunSpecImpl.BuilderImpl(name, size);
    }

    public void setLuns(Map<String, MarsLun> luns) {
        this.luns = luns;
    }

    @Override
    public Map<String, MarsLun> getLuns() {
        return luns;
    }

    @Override
    public MarsLun getLun(String path) {
        return luns.get(path);
    }

    @Override
    public MarsLun createLun(LunSpec lunSpec) {
        LunCreateInputType input = new LunCreateInputType();
        LunMaxExtentSize maxExtentSize = LunMaxExtentSize.maxDefault;
        if (lunSpec instanceof MarsLunSpec) {
            input.setMaxExtentSize(((MarsLunSpec)lunSpec).getMaxExtentSize());
        }
        input.setName(lunSpec.getLunName());
        input.setSize(String.valueOf(lunSpec.getSize().toBytes()));
        try {
            ResponseContentType<UuidType> response = getConnection().post("luns", input, new TypeReference<ResponseContentType<UuidType>>() {});
            MarsLunImpl marsLun = new MarsLunImpl(this);
            marsLun.setMaxExtentSize(maxExtentSize);
            marsLun.setBlockSize(DEFAULT_LUN_BLOCK_SIZE);
            marsLun.setTotalSize(lunSpec.getSize().toBytes());
            marsLun.setUuid(response.getResult().getUuid());
            marsLun.setName(input.getName());
            marsLun.setPath(input.getName());
            marsLun.setOnline(true);
            luns.put(marsLun.getName(), marsLun);
            return marsLun;
        } catch (MarsApiFailedException e) {
           throw new LunOperationException("An error occurred while creating the lun " + lunSpec.getLunName() + ": " + e.getMessage(), e);
        }
    }

    @Override
    public List<FcInterface> getFcInterfaces() {
        return fcInterfaces;
    }

    public void setFcInterfaces(List<FcInterface> fcInterfaces) {
        this.fcInterfaces = fcInterfaces;
    }

    @Override
    public synchronized MarsController getProxy() {
        if (proxy == null) {
            proxy = OfftapProxy.create(MarsController.class, MarsControllerProxy.class, this);
        }
        return proxy;
    }

    @Override
    public Set<StorageSystem> getStorageSystems() {
        return STORAGE_SYSTEMS;
    }

    public void setIGroups(List<IGroup> igroups) {
        this.igroups = igroups;
    }

    @Override
    public List<IGroup> getIGroups() {
        return igroups;
    }

    @Override
    public IGroup getIGroup(String name) {
        for (IGroup group : getIGroups()) {
            if (group.getName().equals(name)) {
                return group;
            }
        }
        return null;
    }

    @Override
    public Builder getIGroupSpecBuilder(String iGroupName, String osType) {
        return new BuilderImpl(iGroupName, osType);
    }

    @Override
    public IGroup createIGroup(IGroupSpec iGroupSpec) throws IGroupOperationException {
        IGroupCreateInputType groupInput = new IGroupCreateInputType();
        groupInput.setName(iGroupSpec.getName());
        groupInput.setOstype(OsType.getType(iGroupSpec.getOsType()));

        try {
            getConnection().post("igroups", groupInput, new TypeReference<ResponseContentType<UuidType>>() {});
            MarsIGroupImpl iGroup = new MarsIGroupImpl(this);
            iGroup.setName(iGroupSpec.getName());
            return iGroup.refresh();
        } catch (MarsApiFailedException e) {
            LOGGER.error("Got an error while creating the IGroup " + getName() + ": " + e.getMessage(), e);
            throw new IGroupOperationException("An error occurred while creating the IGroup " + groupInput.getName() + ": " + e.getMessage(), e);
        }
    }



    @Override
    public Set<Lun> getLunsMappedToInitiator(String initiator) {
        Set<Lun> mappedLuns = new HashSet<>();
        for (Lun lun : getProxy().getLuns().values()) {
            for (IGroup igroup : lun.getIgroups()) {
                if (igroup.getInitiatorIds().contains(initiator)) {
                    mappedLuns.add(lun);
                }
            }
        }
        return mappedLuns;
    }

    @Override
    public List<FcpAdapter> getFcpAdapters() {
        return fcpAdapters;
    }

    public void setFcpAdapters(List<FcpAdapter> fcpAdapters) {
        this.fcpAdapters = fcpAdapters;
    }
}
