/*
 * $Id$
 *
 * Copyright (c) 2015 NetApp, Inc.
 * All rights reserved.
 */
package com.netapp.offtap3.internal.mars;

import java.util.Collections;
import java.util.Iterator;
import java.util.Set;

import org.apache.log4j.Logger;

import com.fasterxml.jackson.core.type.TypeReference;
import com.netapp.offtap3.IGroup;
import com.netapp.offtap3.internal.IGroupImpl;
import com.netapp.offtap3.internal.mars.webmapi.model.common.ResponseContentType;
import com.netapp.offtap3.internal.mars.webmapi.model.domain.IGroupAddOutputType;
import com.netapp.offtap3.internal.mars.webmapi.model.domain.IGroupModifyInputType;
import com.netapp.offtap3.internal.mars.webmapi.model.domain.IGroupRemoveOutputType;
import com.netapp.offtap3.internal.mars.webmapi.model.domain.ResourceRenameInputType;
import com.netapp.offtap3.internal.mars.webmapi.util.MarsError;
import com.netapp.offtap3.mars.MarsApiFailedException;
import com.netapp.offtap3.mars.MarsController;
import com.netapp.offtap3.ontap.ApiFailedException;
import com.netapp.offtap3.ontap.IGroupOperationException;

public class MarsIGroupImpl extends IGroupImpl<MarsControllerImpl, MarsController> implements IGroup {
    private static final Logger LOGGER = Logger.getLogger(MarsIGroupImpl.class);

    public MarsIGroupImpl(MarsControllerImpl controller) {
        super(controller);
    }

    @Override
    public void addInitiator(String initiator) {
        addInitiators(Collections.singleton(initiator));
    }

    @Override
    public void addInitiators(Set<String> initiators) throws IGroupOperationException {
        LOGGER.debug("Adding initiators " + initiators + " to IGroup " + getName());
        MarsConnection connection = getControllerImpl().getConnection();
        IGroupModifyInputType initiatorsInput = new IGroupModifyInputType();
        String[] initiatorArray = new String[initiators.size()];
        Iterator<String> initiatorIterator = initiators.iterator();
        for (int i = 0; i < initiators.size(); i++) {
            initiatorArray[i] = initiatorIterator.next();
        }
        initiatorsInput.setInitiators(initiatorArray);
        initiatorsInput.setName(getName());
        try {
            connection.put("igroups/action/add", initiatorsInput,
                           new TypeReference<ResponseContentType<IGroupAddOutputType>>() {
                           });
        } catch (MarsApiFailedException e) {
            if (MarsError.IGROUP_ADD_DUPLICATE_INPUTS != e.getError()) {
                throw new IGroupOperationException(
                        "An error ocurred while trying to add the initiators " + initiators + " to the group "
                                + getName() + ": " + e.getMessage(), e);
            }
        }
    }

    @Override
    public void removeInitiators(Set<String> initiators, boolean force) {
        LOGGER.debug("Removing initiators " + initiators + " from IGroup " + getName());
        MarsConnection connection = getControllerImpl().getConnection();
        IGroupModifyInputType initiatorsInput = new IGroupModifyInputType();
        String[] initiatorArray = new String[initiators.size()];
        Iterator<String> initiatorIterator = initiators.iterator();
        for (int i = 0; i < initiators.size(); i++) {
            initiatorArray[i] = initiatorIterator.next();
        }
        initiatorsInput.setInitiators(initiatorArray);
        initiatorsInput.setName(getName());
        try {
            connection.put("igroups/action/remove", initiatorsInput,
                           new TypeReference<ResponseContentType<IGroupRemoveOutputType>>() {
                           });

        } catch (MarsApiFailedException e) {
            if (MarsError.IGROUP_REMOVE_INITIATOR_NOT_FOUND != e.getError()
                    && MarsError.IGROUP_REMOVE_DUPLICATE_INPUTS != e.getError()) {
                throw new IGroupOperationException(
                        "An error ocurred while trying to remove the initiators " + initiators + " from the group "
                                + getName() + ": " + e.getMessage(), e);
            } else {
                LOGGER.debug("While removing initiators, one or more was either not found or was a duplicate: " + e
                        .getMessage());
            }
        }
    }

    @Override
    public void rename(String newName) {
        LOGGER.debug("Renaming IGroup " + getName() + " to " + newName);
        MarsConnection connection = getControllerImpl().getConnection();
        ResourceRenameInputType inputType = new ResourceRenameInputType();
        inputType.setName(getName());
        inputType.setNewName(newName);
        try {
            connection.put("igroups/action/rename", inputType, new TypeReference<ResponseContentType>() {
            });
            setName(newName);
        } catch (MarsApiFailedException e) {
            throw new IGroupOperationException(
                    "An error occurred while trying to rename the igroup " + getName() + " to " + newName + ": " + e
                            .getMessage(), e);
        }
    }

    @Override
    public void destroy(boolean force) throws IGroupOperationException {
        LOGGER.debug("Destroying IGroup " + getName());
        MarsConnection connection = getControllerImpl().getConnection();
        try {
            connection.delete("igroups/" + getName(), new TypeReference<ResponseContentType>() {
            });
        } catch (MarsApiFailedException e) {
            throw new IGroupOperationException(
                    "An error occurred while trying to destroy the igroup " + getName() + ": " + e.getMessage(), e);
        }
    }

    @Override
    public void destroy() throws IGroupOperationException {
        destroy(false);
    }
}
