/*
 * $Id$
 *
 * Copyright (c) 2015 NetApp, Inc.
 * All rights reserved.
 */
package com.netapp.offtap3.internal.mars.dataloader;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.List;
import java.util.Set;
import java.util.concurrent.CopyOnWriteArrayList;

import org.apache.log4j.Logger;

import com.fasterxml.jackson.core.type.TypeReference;
import com.netapp.offtap3.FcInterface;
import com.netapp.offtap3.FcpAdapter;
import com.netapp.offtap3.Refreshable;
import com.netapp.offtap3.Resource;
import com.netapp.offtap3.factories.DataLoaderException;
import com.netapp.offtap3.internal.AbstractDataLoader;
import com.netapp.offtap3.internal.FcpAdapterImpl;
import com.netapp.offtap3.internal.mars.MarsControllerImpl;
import com.netapp.offtap3.internal.mars.webmapi.model.common.ContainerType;
import com.netapp.offtap3.internal.mars.webmapi.model.common.ResponseContentType;
import com.netapp.offtap3.internal.mars.webmapi.model.domain.LifInfoType;
import com.netapp.offtap3.internal.mars.webmapi.model.domain.LifStateType;
import com.netapp.offtap3.internal.ontap.intf.FcInterfaceImpl;
import com.netapp.offtap3.mars.MarsApiFailedException;

public class MarsFcInterfaceDataLoader extends AbstractDataLoader<MarsControllerImpl> {
    private static final Logger LOGGER = Logger.getLogger(MarsFcInterfaceDataLoader.class);

    @Override
    protected Refreshable doRefresh(MarsControllerImpl controller, Resource resource, Refreshable refreshTarget) {
        LOGGER.debug("Entering refresh for fc interface " + refreshTarget.getName());
        doLoad(controller, LoadType.refresh, Collections.singleton(resource), (FcInterfaceImpl)refreshTarget);
        for (FcInterface adapter : controller.getFcInterfaces()) {
            if (adapter.getName().equals(refreshTarget.getName())) {
                return adapter;
            }
        }
        LOGGER.debug("Could not find fc interface " + refreshTarget.getName() + " after refresh!");
        return null;
    }

    @Override
    protected boolean doLoad(MarsControllerImpl controller, LoadType loadType, Set<Resource> resources) throws DataLoaderException {
        return doLoad(controller, loadType, resources, null);
    }

    protected boolean doLoad(MarsControllerImpl controller, LoadType loadType, Set<Resource> resources, FcInterfaceImpl existingInterface) throws DataLoaderException {
        LOGGER.debug("Beginning load of Fibre Channel interfaces for mars controller " + controller);
        List<FcInterface> fcInterfaces = new CopyOnWriteArrayList<>();
        List<LifInfoType> lifInfoTypes = Collections.emptyList();
        try {
            if (existingInterface != null) {
                fcInterfaces.addAll(controller.getFcInterfaces());
                fcInterfaces.remove(existingInterface);
                ResponseContentType<LifInfoType> response = controller.getConnection().get("net/interfaces/fc/" + existingInterface.getName(), new TypeReference<ResponseContentType<LifInfoType>>() {});
                if (response.getResult() != null) {
                    lifInfoTypes = Collections.singletonList(response.getResult());
                }
            } else {
                ResponseContentType<ContainerType<LifInfoType>> response = controller.getConnection()
                                                                                     .get("net/interfaces/fc",
                                                                                          new TypeReference<ResponseContentType<ContainerType<LifInfoType>>>() {
                                                                                          });
                if (response.getResult() != null && response.getResult().getInstancesCount() > 0) {
                    lifInfoTypes = response.getResult().getInstances();
                }
            }
            for (LifInfoType lifInfo : lifInfoTypes) {
                FcInterfaceImpl fcInterface = new FcInterfaceImpl(controller);
                fcInterface.setName(lifInfo.getName());
                fcInterface.setCurrentPort(lifInfo.getPortName());
                fcInterface.setHomePort(lifInfo.getPortName());
                fcInterface.setHomeNode(lifInfo.getNodeName());
                fcInterface.setCurrentNode(lifInfo.getNodeName());
                fcInterface.setUuid(lifInfo.getName());
                fcInterface.setWwpn(lifInfo.getFcLifInfo().getWwpn());
                fcInterface.setNodeName(lifInfo.getFcLifInfo().getWwnn());
                fcInterface.setIsHome(true);
                fcInterface.setIsUp(lifInfo.getState() == LifStateType.online);
                fcInterface.setOwningVserver(controller.getName());
                fcInterface.setRole("data");
                fcInterface.setProtocols(Collections.singletonList("fcp"));
                fcInterfaces.add(fcInterface);
            }
            controller.setFcInterfaces(fcInterfaces);
            return true;
        } catch (MarsApiFailedException e) {
            throw new DataLoaderException(e);
        }
    }
}
