/*
 * $Id$
 *
 * Copyright (c) 2015 NetApp, Inc.
 * All rights reserved.
 */
package com.netapp.offtap3.internal.mars.dataloader;

import java.util.Collections;
import java.util.List;
import java.util.Set;
import java.util.concurrent.CopyOnWriteArrayList;

import org.apache.log4j.Logger;

import com.fasterxml.jackson.core.type.TypeReference;
import com.netapp.offtap3.FcpAdapter;
import com.netapp.offtap3.Refreshable;
import com.netapp.offtap3.Resource;
import com.netapp.offtap3.factories.DataLoaderException;
import com.netapp.offtap3.internal.AbstractDataLoader;
import com.netapp.offtap3.internal.FcpAdapterImpl;
import com.netapp.offtap3.internal.mars.MarsControllerImpl;
import com.netapp.offtap3.internal.mars.webmapi.model.common.ContainerType;
import com.netapp.offtap3.internal.mars.webmapi.model.common.ResponseContentType;
import com.netapp.offtap3.internal.mars.webmapi.model.domain.FcPortInfoType;
import com.netapp.offtap3.internal.mars.webmapi.model.domain.LoggedInInitiatorInfoType;
import com.netapp.offtap3.internal.mars.webmapi.model.domain.PortContainerType;
import com.netapp.offtap3.mars.MarsApiFailedException;

public class MarsFcpDataLoader extends AbstractDataLoader<MarsControllerImpl> {
    private static final Logger LOGGER = Logger.getLogger(MarsFcpDataLoader.class);

    @Override
    protected Refreshable doRefresh(MarsControllerImpl controller, Resource resource, Refreshable refreshTarget) {
        LOGGER.debug("Entering refresh for fcp adapter " + refreshTarget.getName());
        doLoad(controller, LoadType.refresh, Collections.singleton(resource), (FcpAdapterImpl)refreshTarget);
        for (FcpAdapter adapter : controller.getFcpAdapters()) {
            if (adapter.getName().equals(refreshTarget.getName())) {
                return adapter;
            }
        }
        LOGGER.debug("Could not find fcp adapter " + refreshTarget.getName() + " after refresh!");
        return null;
    }

    @Override
    protected boolean doLoad(MarsControllerImpl controller, LoadType loadType, Set<Resource> resources) throws DataLoaderException {
        return doLoad(controller, loadType, resources, null);
    }

    protected boolean doLoad(MarsControllerImpl controller, LoadType loadType, Set<Resource> resources, FcpAdapterImpl existingAdapter) throws DataLoaderException {
        LOGGER.debug("Starting data load of Mars fibre channel adapters for controller " + controller);
        try {
            List<FcpAdapter> fcpAdapters = new CopyOnWriteArrayList<>();
            // TODO: In the future when the Mars REST API supports query parameters, try to query only for the port we need to refresh instead of rebuilding the
            // entire list.
            ResponseContentType<PortContainerType> portResponse = controller.getConnection().get("net/ports", new TypeReference<ResponseContentType<PortContainerType>>() {});
            if (portResponse.getResult() != null && portResponse.getResult().getFcPorts() != null) {
                for (FcPortInfoType fcPort : portResponse.getResult().getFcPorts()) {
                    FcpAdapterImpl fcpAdapter = new FcpAdapterImpl(controller);
                    fcpAdapter.setName(fcPort.getName());
                    fcpAdapter.setPort(fcPort.getHardwareName());
                    fcpAdapter.setState(fcPort.getState().toString());
                    fcpAdapters.add(fcpAdapter);
                }

                // Now that we have the fcp adapters, let's get all the initiators and set them in each adapter.
                ResponseContentType<ContainerType<LoggedInInitiatorInfoType>> initiatorsResponse = controller.getConnection().get("fcp/initiators", new TypeReference<ResponseContentType<ContainerType<LoggedInInitiatorInfoType>>>() {});
                if (initiatorsResponse.getResult() != null && initiatorsResponse.getResult().getInstancesCount() > 0) {
                    for (LoggedInInitiatorInfoType initiatorInfo : initiatorsResponse.getResult().getInstances()) {
                        for (FcpAdapter fcpAdapter : fcpAdapters) {
                            if (initiatorInfo.getPortName().equals(fcpAdapter.getPort())) {
                                fcpAdapter.getInitiators().add(initiatorInfo.getWwpn());
                            }
                        }
                    }
                }
            }

            controller.setFcpAdapters(fcpAdapters);
        } catch (MarsApiFailedException e) {
            LOGGER.error("Error occurred while loading fiber channel adapters for Mars controller " + controller, e);
            throw new DataLoaderException(e);
        }
        return true;
    }
}
