/*
 * $Id$
 *
 * Copyright (c) 2015 NetApp, Inc.
 * All rights reserved.
 */
package com.netapp.offtap3.internal.mars.dataloader;

import java.util.Collections;
import java.util.Date;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.apache.log4j.Logger;

import com.fasterxml.jackson.core.type.TypeReference;
import com.netapp.offtap3.internal.mars.webmapi.util.MarsError;
import com.netapp.offtap3.mars.MarsApiFailedException;
import com.netapp.offtap3.ontap.ApiFailedException;
import com.netapp.offtap3.Refreshable;
import com.netapp.offtap3.Resource;
import com.netapp.offtap3.factories.DataLoaderException;
import com.netapp.offtap3.internal.AbstractDataLoader;
import com.netapp.offtap3.internal.mars.MarsControllerImpl;
import com.netapp.offtap3.internal.mars.MarsLunImpl;
import com.netapp.offtap3.internal.mars.MarsPlatformDataLoader;
import com.netapp.offtap3.internal.mars.webmapi.model.common.ContainerType;
import com.netapp.offtap3.internal.mars.webmapi.model.common.ResponseContentType;
import com.netapp.offtap3.internal.mars.webmapi.model.domain.LunInfoType;
import com.netapp.offtap3.internal.mars.webmapi.model.domain.LunStateType;
import com.netapp.offtap3.internal.mars.webmapi.model.domain.LunStatusType;
import com.netapp.offtap3.mars.MarsLun;
import com.netapp.offtap3.utils.SafeConcurrentHashMap;

public class MarsLunDataLoader extends AbstractDataLoader<MarsControllerImpl> {
    private static final Logger logger = Logger.getLogger(MarsPlatformDataLoader.class);

    @Override
    protected boolean doLoad(MarsControllerImpl controller, LoadType loadType, Set<Resource> resources) throws DataLoaderException {
        return doLoad(controller, loadType, resources, null);
    }

    @Override
    protected Refreshable doRefresh(MarsControllerImpl controller, Resource resource, Refreshable refreshTarget) {
        MarsLunImpl marsLun = (MarsLunImpl) refreshTarget;
        logger.debug("Mars lun refresh for lun " + marsLun.getName() + " with uuid " + marsLun.getUuid());
        doLoad(controller, LoadType.refresh, Collections.singleton(resource), marsLun);
        MarsLun lun = controller.getLun(marsLun.getPath());
        logger.debug("Returning refreshed mars lun :" + lun);
        return lun;
    }

    protected boolean doLoad(MarsControllerImpl marsController, LoadType loadType, Set<Resource> resources, MarsLunImpl existingLun) throws DataLoaderException {
        try {
            List<LunInfoType> luns = Collections.emptyList();
            Map<String, MarsLun> lunMap;
            if (existingLun != null) {
                lunMap = new HashMap<>(marsController.getLuns());
                lunMap.remove(existingLun.getName());
                try {
                    ResponseContentType<LunInfoType> result = marsController.getConnection()
                                                                            .get("luns/" + existingLun.getName(),
                                                                                 new TypeReference<ResponseContentType<LunInfoType>>() {
                                                                                 });
                    luns = Collections.singletonList(result.getResult());
                } catch (MarsApiFailedException e) {
                    if (MarsError.LUN_SHOW_LUN_NOT_FOUND == e.getError()) {
                        logger.debug("Did not find lun " + existingLun.getName() + " after refresh.");
                        luns = Collections.emptyList();
                    }
                }
            } else {
                ResponseContentType<ContainerType<LunInfoType>> result = marsController.getConnection().get("luns",
                                                                                                            new TypeReference<ResponseContentType<ContainerType<LunInfoType>>>() {
                                                                                                            });

                ContainerType<LunInfoType> lunContainer = result.getResult();
                lunMap = new SafeConcurrentHashMap<>();
                if (lunContainer != null && lunContainer.getInstancesCount() > 0) {
                    luns = lunContainer.getInstances();
                }
            }

            for (LunInfoType lunInfo : luns) {
                MarsLunImpl lunImpl = new MarsLunImpl(marsController);
                lunImpl.setPath(lunInfo.getName());
                lunImpl.setBlockSize(lunInfo.getBlockSize());
                lunImpl.setName(lunInfo.getName());
                lunImpl.setOnline(lunInfo.getLunState() == LunStateType.online);
                lunImpl.setReadOnly(lunInfo.getLunStatus() != LunStatusType.normal);
                lunImpl.setUuid(lunInfo.getUuid());
                if (lunInfo.getSpaceUsed() != null) {
                    lunImpl.setUsedSize(lunInfo.getSpaceUsed());
                }
                if (lunInfo.getSize() != null) {
                    lunImpl.setTotalSize(lunInfo.getSize());
                }
                lunImpl.setLastLoad(new Date());
                lunImpl.setType("");
                lunMap.put(lunImpl.getName(), lunImpl);
            }
            marsController.setLuns(lunMap);
        } catch (MarsApiFailedException e) {
            logger.error("An error occurred while loading the Mars luns: " + e.getMessage(), e);
            throw new DataLoaderException(e);
        }
        return true;
    }
}
