/*
 * NETAPP CONFIDENTIAL
 *
 * All information contained herein is, and remains 
 * the property of NetApp, Inc. and its suppliers if any.
 *
 * Copyright (c) 2014 NetApp, Inc.
 * All rights reserved.
 */
package com.netapp.offtap3.internal.mars.webmapi.util;

import java.util.Comparator;

import com.netapp.offtap3.internal.mars.webmapi.model.BaseDataType;

/**
 * Filter Condition defines condition to filter data records
 * 
 * @author Jing Shi
 * @version $Revision$, Mar 25, 2013
 */
public class FilterCondition extends BaseDataType implements Comparable<FilterCondition>, Comparator<FilterCondition> {
    public static final String QUERY_FILTER_KEY = "filter";
    public static final String QUERY_FILTER_FORMAT = "filter={field}({operator}){value}&more";
    public static String s_Query_LB = "(";
    public static String s_Query_RB = ")";
    public static String s_Query_EQ = "=";
    public static String s_LB = "[";
    public static String s_RB = "]";

    /**
     * FilterOp defines enumeration of filter operators
     */
    public enum FilterOp {
        eq, ne, gt, lt, ge, le;

        private static FilterOp[] s_values = FilterOp.values();

        public static FilterOp getType(String name) {
            if (name == null || name.length() == 0)
                throw new IllegalArgumentException("Invalid argument:" + name);
            String lowercase = name.toLowerCase();
            for (FilterOp op : s_values) {
                if (lowercase.equals(op.toString())) {
                    return op;
                }
            }

            throw new IllegalArgumentException("Cannot find filter type for:" + name);
        }
    }

    /** filter's priority, smaller number is higher priority */
    private Integer priority;

    /** filtered by field */
    private String field;

    /** filtered by operator */
    private FilterOp operator;

    /** filtered by field value */
    private String value;

    public FilterCondition() {
    }

    public FilterCondition(String field, FilterOp operator, String value) {
        if (field == null || operator == null || value == null)
            throw new IllegalArgumentException("Invalid filter condition arguments " + field + operator + value);

        this.field = field;
        this.operator = operator;
        this.value = value;
    }

    public Integer getPriority() {
        return priority;
    }

    public void setPriority(Integer priority) {
        this.priority = priority;
    }

    public String getField() {
        return field;
    }

    public void setField(String field) {
        this.field = field;
    }

    public FilterOp getOperator() {
        return this.operator;
    }

    public void setOperator(FilterOp operator) {
        this.operator = operator;
    }

    public String getValue() {
        return this.value;
    }

    public void setValue(String value) {
        this.value = value;
    }

    @Override
    public int hashCode() {
        return field.hashCode() ^ operator.hashCode() ^ value.hashCode();
    }

    @Override
    public boolean equals(Object obj) {
        if (obj == null)
            return false;
        if (obj instanceof FilterCondition) {
            FilterCondition that = (FilterCondition) obj;
            return this.field.equals(that.field) && this.operator.equals(that.operator) && this.value.equals(that.value);
        }

        return false;
    }

    @Override
    public int compareTo(FilterCondition o) {
        return compare(this, o);
    }

    @Override
    public int compare(FilterCondition o1, FilterCondition o2) {
        if (o1 == null && o2 == null)
            return 0;
        if (o1 == null)
            return -1;
        if (o2 == null)
            return 1;

        return (o1.priority - o2.priority);
    }

    public String toNiceString() {
        StringBuilder sb = new StringBuilder();
        sb.append(FilterCondition.QUERY_FILTER_KEY).append(getPriority()).append(s_Query_EQ).append(getField()).append(s_Query_LB).append(getOperator()).append(s_Query_RB).append(getValue());

        return sb.toString();
    }

    public static FilterCondition parse(String filterValue) throws Exception {
        assert filterValue != null;

        String field;
        FilterOp op;
        String value;

        int lbIdx = filterValue.indexOf(s_Query_LB);
        int rbIdx = filterValue.indexOf(s_Query_RB);
        if (lbIdx <= 0 || rbIdx <= 0 || lbIdx >= rbIdx || rbIdx >= filterValue.length())
            throw new Exception("Filter has invalid query string: " + filterValue);

        field = filterValue.substring(0, lbIdx);
        op = FilterOp.getType(filterValue.substring(lbIdx + 1, rbIdx));
        value = filterValue.substring(rbIdx + 1);

        return new FilterCondition(field, op, value);
    }
}
