/*
 * NETAPP CONFIDENTIAL
 *
 * All information contained herein is, and remains 
 * the property of NetApp, Inc. and its suppliers if any.
 *
 * Copyright (c) 2014 NetApp, Inc.
 * All rights reserved.
 */
package com.netapp.offtap3.internal.mars.webmapi.util;

import java.beans.BeanInfo;
import java.beans.Introspector;
import java.beans.PropertyDescriptor;
import java.lang.reflect.Method;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Iterator;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;
import java.util.WeakHashMap;

import com.netapp.offtap3.internal.mars.webmapi.model.common.Pair;

/**
 * ParseObjectType defines data type used as arguments.
 * <p>
 * Basically it is a map of name value pair in order. And it also support "marsh", "unmarsh" object to the map of object's field name and value using reflection
 * </p>
 * 
 * @author Jing Shi
 * @version $Revision$, May 2, 2013
 */
public class ParseObjectType extends LinkedHashMap<String, Object> {

    private static final long serialVersionUID = 1L;

    private static final WeakHashMap<Class<?>, LinkedHashMap<String, PropertyDescriptor>> s_fieldCheckers = new WeakHashMap<Class<?>, LinkedHashMap<String, PropertyDescriptor>>();

    /**
     * Constructor
     */
    public ParseObjectType() {
    }

    /**
     * Constructor
     * <p>
     * Put the fields of specified object as arguments
     * </p>
     * 
     * @param instance
     * @param ignoreNull
     */
    public ParseObjectType(Object instance, boolean ignoreNull) {
        setObject(instance, ignoreNull);
    }

    /**
     * Get object from an instance of specified class
     * 
     * @param clazz
     *            specified class
     * @param restrict
     *            true to RuntimeException for any invalid field; otherwise, to ignore invalid fields
     * @return instance of specified class
     */
    public <T> T getObject(Class<T> clazz, boolean restrict) throws RuntimeException {
        return unmarsh(clazz, restrict);
    }

    /**
     * Add the fields of specified object as arguments
     */
    public void setObject(Object instance, boolean ignoreNull) {
        marsh(instance, ignoreNull);
    }

    /**
     * Validate the arguments for specified class.
     * <p>
     * It is used for reflect from arguments to object instance. Check is based on the java reflection to class declared fields.
     * </p>
     * 
     * @param clazz
     * @return valid field map and invalid field list.
     */
    public Pair<Map<String, Object>, List<String>> validate(Class<?> clazz) {
        List<String> invalidFields = new ArrayList<String>();
        Map<String, Object> validFields = new LinkedHashMap<String, Object>();
        Map<String, PropertyDescriptor> fieldChecker = getClassChecker(clazz);

        Iterator<Map.Entry<String, Object>> itr = this.entrySet().iterator();
        while (itr.hasNext()) {
            Map.Entry<String, Object> ent = itr.next();
            try {
                if (!fieldChecker.containsKey(ent.getKey())) {
                    invalidFields.add(ent.getKey());
                } else {
                    validFields.put(ent.getKey(), ent.getValue());
                }
            } catch (Exception ex) {
                invalidFields.add(ent.getKey());
            }
        }

        return new Pair<Map<String, Object>, List<String>>(validFields, invalidFields);
    }

    /**
     * Unmarsh ParseObjectType instance to an instance of specified class
     * 
     * @param clazz
     *            specified class
     * @param restrict
     *            - true umarsh throws RuntimeException for any invalid field, otherwise, unmarsh ignores invalid fields
     * @return instance of specified class
     */
    protected <T> T unmarsh(Class<T> clazz, boolean restrict) {
        try {
            T instance = clazz.newInstance();
            Map<String, Object> fieldMap = toMap();
            Map<String, PropertyDescriptor> fieldChecker = getClassChecker(clazz);
            if (!restrict) {
                Pair<Map<String, Object>, List<String>> result = validate(clazz);
                fieldMap = result.getFirst();
                for (Map.Entry<String, Object> ent : fieldMap.entrySet()) {
                    try {
                        Method method = fieldChecker.get(ent.getKey()).getWriteMethod();
                        method.setAccessible(true);
                        method.invoke(instance, ent.getValue());
                    } catch (Exception ex) {
                        ex.printStackTrace();
                    }
                }
            } else {
                for (Map.Entry<String, Object> ent : fieldMap.entrySet()) {
                    Method method = fieldChecker.get(ent.getKey()).getWriteMethod();
                    method.setAccessible(true);
                    method.invoke(instance, ent.getValue());
                }
            }
            return instance;
        } catch (Exception ex) {
            ex.printStackTrace();
            throw new RuntimeException(ex);
        }
    }

    /**
     * Marsh the object's field value with getter methods as arguments
     * 
     * @param o
     *            the object
     * @param ignoreNull
     *            - true to ignore the field value is null, otherwise, there is null value as argument
     */
    protected void marsh(Object o, boolean ignoreNull) {
        try {
            Map<String, PropertyDescriptor> props = getClassChecker(o.getClass());

            for (PropertyDescriptor prop : props.values()) {
                String field = prop.getName();
                Method getter = prop.getReadMethod();
                if (getter != null) {
                    Object value = getter.invoke(o);

                    if (!ignoreNull || value != null)
                        put(field, value);
                }
            }
        } catch (Exception ex) {
            ex.printStackTrace();
            throw new RuntimeException(ex);
        }
    }

    private Map<String, PropertyDescriptor> getClassChecker(Class<?> clazz) {
        try {
            LinkedHashMap<String, PropertyDescriptor> checker = s_fieldCheckers.get(clazz);

            if (checker == null) {
                checker = new LinkedHashMap<String, PropertyDescriptor>();
                BeanInfo info = Introspector.getBeanInfo(clazz, Object.class);
                PropertyDescriptor[] props = info.getPropertyDescriptors();
                for (PropertyDescriptor prop : props) {
                    checker.put(prop.getName(), prop);
                }

                s_fieldCheckers.put(clazz, checker);
            }
            return checker;
        } catch (Exception ex) {
            ex.printStackTrace();
            throw new RuntimeException(ex);
        }
    }

    private Map<String, Object> toMap() {
        return this;

        // Map<String, Object> map = new LinkedHashMap<String, Object>();
        // Iterator<ArgumentType<?>> itr = this.iterator();
        // while(itr.hasNext()) {
        // ArgumentType<?> ent = itr.next();
        // map.put(ent.getName(), ent.getValue());
        // }
        //
        // return map;
    }

    public String toString() {
        StringBuilder sb = new StringBuilder();
        for (Map.Entry<String, Object> ent : this.entrySet()) {
            sb.append(ent.getKey()).append("=");
            Object value = ent.getValue();
            if (value != null && value.getClass().isArray()) {
                Object[] objAry = (Object[]) value;
                if (objAry != null && objAry.length > 10000)
                    sb.append("omit the large array of " + value.getClass());  //$NON-NLS-L$ 
                else
                    sb.append(Arrays.toString(objAry));
            } else {
                sb.append(String.valueOf(value));
            }
            sb.append(",");
        }

        if (sb.length() > 0)
            sb.deleteCharAt(sb.length() - 1);
        return sb.toString();
    }
}
