/*
 * NETAPP CONFIDENTIAL
 *
 * All information contained herein is, and remains
 * the property of NetApp, Inc. and its suppliers if any.
 *
 * Copyright (c) 2014 NetApp, Inc.
 * All rights reserved.
 */

package com.netapp.offtap3.internal.mars.webmapi.util;

import java.io.BufferedInputStream;
import java.io.BufferedOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.net.HttpURLConnection;
import java.net.ProtocolException;
import java.net.URI;
import java.net.URISyntaxException;
import java.security.KeyManagementException;
import java.security.NoSuchAlgorithmException;
import java.security.cert.X509Certificate;
import java.util.Scanner;

import javax.net.ssl.HostnameVerifier;
import javax.net.ssl.HttpsURLConnection;
import javax.net.ssl.SSLContext;
import javax.net.ssl.SSLSession;
import javax.net.ssl.SSLSocketFactory;
import javax.net.ssl.X509TrustManager;

import org.apache.log4j.Logger;

import com.netapp.offtap3.internal.utils.FilteredSslSocketFactory;
import com.netapp.offtap3.mars.MarsApiFailedException;
import com.netapp.offtap3.utils.StringUtils;


/**
 * WebMapiClient provides webmapi restful client operation functions
 */
public class WebMapiClient {
    private static final Logger logger = Logger.getLogger(WebMapiClient.class);

    private static final HostnameVerifier ACCEPT_ALL_HOSTNAME_VERIFIER = new HostnameVerifier() {
        @Override
        public boolean verify(String hostname, SSLSession session) {
            return true;
        }
    };

    public enum HttpConnectionType {GET, POST, PUT, DELETE}

    public static final String USER_AGENT = "webmapi";

    /* WebMAPI parameters */
    private String host;
    private int port;
    private String apiVersion;
    private String session;
    private boolean secureClient;

    private SSLSocketFactory sslSocketFactory;
    private static final String SSL_PROTOCOL = "TLSv1";

    /**
     * Constructor
     *
     * @param host
     *            cannot be null
     * @param apiVersion
     *            like "v1" cannot be null
     */
    public WebMapiClient(String host, String apiVersion, String session, int port, boolean secureClient) {
        /* GLOBALYZER_START_IGNORE */

        if (host == null)
            throw new IllegalArgumentException("Invalid webmapi client host: null");
        if (apiVersion == null)
            throw new IllegalArgumentException("Invalid webmapi client version: null");
        /* GLOBALYZER_END_IGNORE */

        this.host = host;
        this.port = port;
        this.apiVersion = apiVersion;
        this.session = session;
        this.secureClient = secureClient;
        // Definitely only TLS and no SSLv3
        sslSocketFactory = new FilteredSslSocketFactory(createSecureContext().getSocketFactory(), true, false);
    }

    private SSLContext createSecureContext() {
        SSLContext sslContext = null;
            try {
                sslContext = SSLContext.getInstance(SSL_PROTOCOL);
                X509TrustManager[] trustAllCerts = new X509TrustManager[] { new X509TrustManager() {
                    @Override
                    public X509Certificate[] getAcceptedIssuers() {
                        return null;
                    }

                    @Override
                    public void checkClientTrusted(X509Certificate[] certs, String authType) {
                    }

                    @Override
                    public void checkServerTrusted(X509Certificate[] certs, String authType) {
                    }
                } };
                sslContext.init(null, trustAllCerts, new java.security.SecureRandom());
            } catch (NoSuchAlgorithmException e) {
                logger.error("Unknown SSL protocol " + SSL_PROTOCOL, e);
            } catch (KeyManagementException e) {
                logger.error("Problem creating SSL context", e);
            }
        return sslContext;
    }


    public String connect(String requestId, String path, String query, HttpConnectionType connectionType, String body) throws MarsApiFailedException {
        URI targetUri = newURI(path, query);
        return connect(requestId, targetUri, connectionType, body);
    }

    public String connect(String requestId, URI targetUri, HttpConnectionType connectionType, String body) throws MarsApiFailedException {
        HttpURLConnection connection;
        try {
            connection = (HttpURLConnection)targetUri.toURL().openConnection();
        } catch (IOException e) {
            throw new MarsApiFailedException(MarsError.CONNECTION_FAILED, e);
        }
        if (connection instanceof HttpsURLConnection) {
            try {
                ((HttpsURLConnection) connection).setSSLSocketFactory(sslSocketFactory);
                ((HttpsURLConnection) connection).setHostnameVerifier(ACCEPT_ALL_HOSTNAME_VERIFIER);
            } catch (Exception e) {
                logger.error("Error ocurred while setting SSL factory on connection to url " + targetUri.toString());
            }
        }
        try {
            connection.setRequestMethod(connectionType.name());
        } catch (ProtocolException e) {
            throw new MarsApiFailedException(MarsError.INVALID_PROTOCOL, e);
        }

        connection.setDoOutput(true);
        connection.setRequestProperty("Connection", "keep-alive");
        connection.setRequestProperty("Content-Type", "application/json");
        connection.setRequestProperty("Accept", "application/json");
        connection.setRequestProperty("x-webmapi-version", apiVersion);
        connection.setRequestProperty("x-webmapi-requestid", requestId);
        connection.setRequestProperty("x-webmapi-useragent", USER_AGENT);
        if (!StringUtils.isEmpty(session)) {
            connection.setRequestProperty("x-webmapi-authorization", session);
        }

        try {
            if (!StringUtils.isEmpty(body)) {
                logger.debug("Sending request body: " + body);
                BufferedOutputStream requestStream = new BufferedOutputStream(connection.getOutputStream());
                requestStream.write(body.getBytes());
                requestStream.flush();
            }
            connection.connect();

            BufferedInputStream inputStream = new BufferedInputStream(connection.getInputStream());
            String response = convertStreamToString(inputStream);
            logger.debug("Received response for webmapi call " + response);
            return response;
        } catch (IOException e) {
            if (e.getMessage() != null && e.getMessage().contains("response code: 401")) {
                logger.debug("Received unauthorized response.");
                throw new MarsApiFailedException(MarsError.UNAUTHORIZED, e);
            } else {
                BufferedInputStream errorInputStream = new BufferedInputStream(connection.getErrorStream());
                String errorResponse = convertStreamToString(errorInputStream);
                logger.debug("Received error when executing WebMapi call: " + errorResponse);
                throw new MarsApiFailedException(MarsError.CONNECTION_FAILED, e);
            }
        }
    }

    private String convertStreamToString(InputStream is) {
        Scanner s = new java.util.Scanner(is, "UTF-8").useDelimiter("\\A");
        return s.hasNext() ? s.next() : "";
    }

    /**
     * New target URI from path and query
     *
     * @param path
     *            the resource path without the base path of web service, cannot be null
     * @param query
     *            query string can be null
     * @return target URI
     * @throws MarsApiFailedException
     */
    public URI newURI(String path, String query) throws MarsApiFailedException {
        String webmApiPath = "/webmapi/" + path;

        String schema = secureClient ? "https" : "http";

        URI targetURI;
        try {
            targetURI = new URI(schema, null, host, port, webmApiPath, query, null);
        } catch (URISyntaxException e) {
            throw new MarsApiFailedException(MarsError.INVALID_URI, e);
        }

        logger.trace("path:" + path + ", query:" + query + ", newURI:" + targetURI);
        return targetURI;
    }
}
