/*
 * $Id: //depot/prod/aeonflux/R3.0.1/src/main/java/com/netapp/offtap3/internal/ontap/ExportPolicyImpl.java#1 $
 *
 * Copyright (c) 2014 NetApp, Inc.
 * All rights reserved.
 */
package com.netapp.offtap3.internal.ontap;

import com.netapp.offtap3.ControllerConnection;
import com.netapp.offtap3.ExportOperationException;
import com.netapp.offtap3.Refreshable;
import com.netapp.offtap3.Resource;
import com.netapp.offtap3.StorageSystem;
import com.netapp.offtap3.internal.AbstractRefreshable;
import com.netapp.offtap3.internal.CachingOfftapProxy;
import com.netapp.offtap3.internal.DataCache;
import com.netapp.offtap3.internal.HasProxy;
import com.netapp.offtap3.internal.ontap.controller.DataSvmImpl;
import com.netapp.offtap3.internal.ontap.proxy.ExportPolicyProxy;
import com.netapp.offtap3.internal.ontap.support.ZapiCache;
import com.netapp.offtap3.ontap.ApiFailedException;
import com.netapp.offtap3.ontap.DataSvm;
import com.netapp.offtap3.ontap.ExportPolicy;
import com.netapp.offtap3.ontap.ExportPolicyRule;
import com.netapp.offtap3.ontap.ExportPolicyRuleSpec;
import com.netapp.offtap3.ontap.NaElement;
import org.apache.log4j.Logger;

import java.util.ArrayList;
import java.util.Collections;
import java.util.List;
import java.util.Set;
import java.util.concurrent.CopyOnWriteArrayList;

public class ExportPolicyImpl extends AbstractRefreshable implements ExportPolicy, HasProxy<ExportPolicy>, HasOntapController<DataSvm, DataSvmImpl> {
    private static final Logger LOGGER = Logger.getLogger(ExportPolicyImpl.class);
    private static final Set<StorageSystem> STORAGE_SYSTEMS = Collections.singleton(StorageSystem.datasvm);
    private List<ExportPolicyRule> rules = new CopyOnWriteArrayList<>();
    private DataSvmImpl dataSvm;
    private ExportPolicy proxy;
    private long id;
    private DataCache<NaElement, InternalZapiHandler> dataCache;

    public ExportPolicyImpl(DataSvmImpl dataSvm, DataCache<NaElement, InternalZapiHandler> dataCache) {
        super(dataSvm);
        this.dataSvm = dataSvm;
        this.dataCache = dataCache;
    }

    @Override
    public List<ExportPolicyRule> getRules() {
        return rules;
    }

    public void setRules(List<ExportPolicyRule> rules) {
        this.rules = rules;
    }

    @Override
    public long getId() {
        return id;
    }

    @Override
    public ExportPolicyRuleSpec.Builder createRuleSpecBuilder(String clientMatch, String roRule, String rwRule) {
        return new ExportPolicyRuleSpecImpl.BuilderImpl(clientMatch, roRule, rwRule);
    }

    public void setId(long id) {
        this.id = id;
    }


    @Override
    public void createRule(ExportPolicyRuleSpec ruleSpec) {
        LOGGER.debug("createRule: " + ruleSpec.getClientMatch() + " for policy: " + getName());

        NaElement xi = new NaElement("export-rule-create");
        xi.addNewChild("policy-name", getName());
        xi.addNewChild("client-match", ruleSpec.getClientMatch());

        NaElement roRuleElement = new NaElement("ro-rule");
        roRuleElement.addNewChild("security-flavor", ruleSpec.getRoRule());
        xi.addChildElem(roRuleElement);

        NaElement rwRuleElement = new NaElement("rw-rule");
        rwRuleElement.addNewChild("security-flavor", ruleSpec.getRwRule());
        xi.addChildElem(rwRuleElement);

        if (ruleSpec.isAllowSetUid() != null) {
            xi.addNewChild("is-allow-set-uid-enabled", String.valueOf(ruleSpec.isAllowSetUid()));
        }
        if (ruleSpec.isAllowDeviceCreation() != null) {
            xi.addNewChild("is-allow-dev-is-enabled", String.valueOf(ruleSpec.isAllowDeviceCreation()));
        }

        if (ruleSpec.getProtocol() != null) {
            NaElement protocolElement = new NaElement("protocol");
            protocolElement.addNewChild("access-protocol", ruleSpec.getProtocol());
            xi.addChildElem(protocolElement);
        }

        if (ruleSpec.getAnonymousUserId() != null) {
            xi.addNewChild("anonymous-user-id", ruleSpec.getAnonymousUserId());
        }

        if (ruleSpec.getSuperUserSecurity() != null) {
            NaElement superUserSecurityElement = new NaElement("super-user-security");
            superUserSecurityElement.addNewChild("security-flavor", ruleSpec.getSuperUserSecurity());
            xi.addChildElem(superUserSecurityElement);
        }

        if (ruleSpec.getRuleIndex() != null) {
            xi.addNewChild("rule-index", String.valueOf(ruleSpec.getRuleIndex()));
        }

        try {
            NaElement xo = dataSvm.invokeZapi(xi, ZapiCache.NO_CACHE);
            dataSvm.getZapiCache().blastCacheValues("export");
            LOGGER.info("Modified " + getName() + ":" + xo.getAttr("status") + ".");
        } catch (ApiFailedException e) {
            LOGGER.error("createRule: " + dataSvm.getName() + " Unable to modify export policy: " + getName());
            throw new ExportOperationException("Failed to modify export policy " + getName(), e);
        }
    }

    @Override
    public void destroy() {
        LOGGER.debug("destroy: " + getName());

        NaElement xi = new NaElement("export-policy-destroy");
        xi.addNewChild("policy-name", getName());
        try {
            NaElement xo = dataSvm.invokeZapi(xi, ZapiCache.NO_CACHE);
            dataSvm.getExportPolicies().remove(getName());
            dataSvm.getZapiCache().blastCacheValues("export");
            LOGGER.info("Destroyed export policy " + getName() + ":" + xo.getAttr("status") + ".");
        } catch(Exception ex) {
            LOGGER.error("Failed to destroy export policy " + getName(), ex);
            throw new ExportOperationException("Failed to destroy export policy " + getName() +".  Error: " +ex.getMessage(), ex);
        }
    }

    @Override
    public ExportPolicy getProxy() {
        if (proxy == null) {
            proxy = CachingOfftapProxy.create(ExportPolicy.class, ExportPolicyProxy.class, this, dataCache);
        }
        return proxy;
    }

    @Override
    public Set<StorageSystem> getStorageSystems() {
        return STORAGE_SYSTEMS;
    }

    @Override
    public ControllerConnection getConnection() {
        return getControllerImpl().getConnection();
    }

    @Override
    public DataSvm getController() {
        return dataSvm.getProxy();
    }

    @Override
    public DataSvmImpl getControllerImpl() {
        return dataSvm;
    }

    @Override
    public void refresh(Resource... resources) {
        // Handled by the proxy.
    }

    @Override
    public Refreshable refresh(Refreshable refreshable) {
        // Handled by the proxy
        return null;
    }

    @Override
    public ExportPolicy refresh() {
        return (ExportPolicy) super.refresh();
    }
}
