package com.netapp.offtap3.internal.ontap;

import java.net.HttpURLConnection;
import java.net.URLConnection;
import java.util.Date;
import java.util.concurrent.ConcurrentHashMap;

import javax.net.ssl.HttpsURLConnection;

import org.apache.log4j.Logger;
import org.apache.log4j.MDC;

import com.netapp.offtap3.OfftapPreferences;
import com.netapp.offtap3.factories.ControllerFactory;
import com.netapp.offtap3.internal.ontap.utils.StatsCollector;
import com.netapp.offtap3.utils.Pair;
import com.netapp.offtap3.utils.Utils;

public class MCCSlapperThread extends Thread {
	private static Logger logger = Logger.getLogger(MCCSlapperThread.class);
	
	private String controllerKey;
	private ConcurrentHashMap<Thread, Date> launchTimes = new ConcurrentHashMap<Thread, Date>();
	private boolean sawError = false;
	private OfftapPreferences prefs;
	private long nextWakeTime = -1;
	private String state;
	long lastMccSwitchCheck = System.currentTimeMillis(); // just initialize this to the current time

	
	public MCCSlapperThread(String controllerId, OfftapPreferences prefs) {
		super.setDaemon(true);
		this.controllerKey = controllerId;
		this.prefs = prefs;
		super.setName("MCC-Slapper-" + controllerId);
	}
	
	public synchronized void noteLaunch() {
		launchTimes.put(Thread.currentThread(), new Date());
	}
	
	public synchronized void noteLanding() {
		launchTimes.remove(Thread.currentThread());
	}
	
	public synchronized void noteError() {
		sawError = true;
	}
	
	public String getSupportabilityInformation() {
		String stateVal = "";
		if ("Sleeping".equals(state)) {
			long howMuchLonger = nextWakeTime - System.currentTimeMillis();
			long inSeconds = Math.round((double) howMuchLonger / 1000);
			stateVal = "Sleeping " + inSeconds + " more seconds";
		} else {
			stateVal = state;
		}		
		long ageOfLastCheck = System.currentTimeMillis() - lastMccSwitchCheck;
		long numberOfSeconds = Math.round(((double) ageOfLastCheck) / 1000);
		String lastCheckVal =  "Last checked " + numberOfSeconds + " seconds ago";
		return stateVal + ".  " + lastCheckVal;
	}
	
	public void run() {
		MDC.put("name", "MCCSlapper-" + controllerKey);
		// lets check once right at startup
		synchronouslyExecuteCheck(true);
		while (true) {
			try {
				int secondsToSleep = prefs.getMccSleepTimeInSeconds();
				if (secondsToSleep <= 0){
					state = "Disabled via properties";
					// allow MCC functionality to be disabled via preferences for someone who just enjoys failure.
					return;
				}
				nextWakeTime = System.currentTimeMillis() + (secondsToSleep*1000);
				state = "Sleeping";
				Utils.sleep(secondsToSleep*1000);
				state = "Thinking";
				boolean shouldCheck = false;
				synchronized(this) {
					for (Thread curr : launchTimes.keySet()) {
						Date d = launchTimes.get(curr);
						long ageOfZapi = System.currentTimeMillis() - d.getTime();
						if (ageOfZapi > prefs.getMccOldZapiTimeInSeconds()*1000) {
							shouldCheck = true;
						}
					}
					if (sawError) {
						sawError = false;
						shouldCheck = true;
					}
				}
				long ageOfLastMccSwitchCheck = System.currentTimeMillis() - lastMccSwitchCheck;
				if (shouldCheckForMCCEvents(ageOfLastMccSwitchCheck)) {
					shouldCheck = true;
				}
				//logger.debug("MCCSlapperThread - on " + controllerKey + " think shouldCheck=" + shouldCheck);					
				if (shouldCheck) {
					lastMccSwitchCheck = System.currentTimeMillis();
					state = "Checking";
					synchronouslyExecuteCheck(true);
				}
			} catch (Throwable t) {
				t.printStackTrace();
			} finally {
				synchronized(this) {
					this.notifyAll();
				}
			}
		}
	}
	
	private boolean shouldCheckForMCCEvents(long ageOfLastMccSwitchCheck) {
        return ageOfLastMccSwitchCheck > prefs.getMccCheckThresholdInSeconds() * 1000;
    }

	public boolean synchronouslyExecuteCheck(boolean interruptRunningThreads) {
		Pair newLocation = checkForMccFailover();
		logger.debug("MCCSlapperThread - on " + controllerKey + " see newLocation of " + newLocation);
		if (newLocation != null) {			
			boolean isDifferent = updateZapiInvokerStructures(newLocation);
			logger.debug("MCCSlapperThread - on " + controllerKey + " see newLocation of " + newLocation + " isDifferent is " + isDifferent);
			if (isDifferent && interruptRunningThreads) {
				logger.debug("MCCSlapperThread - Going to blast " + launchTimes.size() + " threads");
				synchronized(this) {
					for (Thread t : launchTimes.keySet()) {
						try {
							StatsCollector.getInstance().addCounter("mcc-thread-interrupt");
							logger.debug("Interrupting thread - dumping stack first");
							StackTraceElement[] ste = t.getStackTrace();
			                if (ste != null) {
			                    for (StackTraceElement aSte : ste) {
			                        logger.debug("\t" + aSte);
			                    }
			                }
							// note - thread.interrupt will not actually interrupt the blocking socket IO 
							// happening with the ZAPI runtime library.  So, lets see if we actually had a connection
							// in flight, and if so, lets interrupt that sucker.
							// Oh the joys of such things.
							boolean connectionInterrupted = false;
							URLConnection connection = MCCManager.getConnectionForThread(t);
							if (connection != null) {
								try {
									if (connection instanceof HttpURLConnection) {
										logger.debug("Disconnecting connection");
										connectionInterrupted = true;
							            ((HttpURLConnection) connection).disconnect();
							        }
							        if (connection instanceof HttpsURLConnection) {
							        	logger.debug("Disconnecting connection");
							        	connectionInterrupted = true;
							            ((HttpsURLConnection) connection).disconnect();
							        }
								} catch (Throwable t2) {
									t2.printStackTrace();
								}
							}
							if (!connectionInterrupted) {
								logger.debug("Actually firing interrupt");
								t.interrupt();
							}
						} catch (Throwable t2) {
							logger.debug("Unable to interrupt thread with " + t);
							t2.printStackTrace();
						}
					}
				}
			}
			return isDifferent;
		}		
		return false;
	}
	
	private Pair checkForMccFailover() {
		try {
			StatsCollector.getInstance().addCounter("mcc-check-for-failover");			
			return MCCManager.getInstance().checkForMccFailover(controllerKey);
		} catch (Exception e) {
			e.printStackTrace();
			return null;
		}
	}
	
	private boolean updateZapiInvokerStructures(Pair newValues) {
		return MCCManager.getInstance().addMCCFailover(controllerKey, (String) newValues.getA(), (String) newValues.getB());
	}
}
