package com.netapp.offtap3.internal.ontap;

import com.netapp.offtap3.ontap.OntapVersion;

import java.io.Serializable;

/**
 * Simple encapsulation of an 3 part ONTAP version number (assumed to be x.y.z).
 */
public class OntapVersionImpl implements Serializable, OntapVersion {
    private static final long serialVersionUID = 3545348702499387687L;

    public static final OntapVersionImpl SN_0 = new OntapVersionImpl("ONTAP dummy version for SN.0", 8, 2, 0);
    public static final OntapVersionImpl FS_0 = new OntapVersionImpl("ONTAP dummy version for FS.0", 8, 3, 0);

    private static final int GEN_FACTOR = 1000 * 1000;
    private static final int MAJOR_FACTOR = 1000;

    /**
     *  The encoded format is intentionally private and inaccessible, as users of this
     *  class should not be aware nor rely on this representation.
     *  Represents the version number as a single integer, by allocating three places
     *  for each component of the version, but assumes that the major and minor numbers
     *  will never exceed 999.
     */
    private final int encoded;

    private final int[] numbers;

    private final String rawVersion;

    public OntapVersionImpl(String rawVersion, String[] dotDelimited) {
        this(rawVersion, dotDelimited[0], dotDelimited[1], (dotDelimited.length > 2) ? dotDelimited[2] : "0");
    }

    public OntapVersionImpl(String rawVersion, String gen, String major, String minor) {
        this(rawVersion, Integer.parseInt(gen), Integer.parseInt(major), Integer.parseInt(minor));
    }

    public OntapVersionImpl(String rawVersion, int gen, int major, int minor) {
        this.rawVersion = rawVersion;
        encoded = gen*GEN_FACTOR + major*MAJOR_FACTOR + minor;
        numbers = new int[] {gen, major, minor};
    }

    public boolean equals(Object obj) {
        return obj instanceof OntapVersionImpl && ((OntapVersionImpl) obj).encoded == encoded;
    }

    @Override
    public int[] getVersionNumber() {
        return numbers;
    }

    @Override
    public String getFullVersion() {
        return rawVersion;
    }


    /** convenience method since it is a very popular version check */
    public boolean atLeastSN() {
        return this.compareTo(SN_0) >= 0;
    }

    public boolean atLeastFS() {
        return this.compareTo(FS_0) >= 0;
    }

    public boolean isSNDotAnything() {
        return atLeastSN() && !atLeastFS();
    }

    public int hashCode() {
        return encoded;
    }

    public String toString() {
        StringBuilder sb = new StringBuilder();
        for (int n : numbers) {
            if (sb.length() > 0) {
                sb.append(".");
            }
            sb.append(n);
        }
        return sb.toString();
    }

    @Override
    public int compareTo(OntapVersion version) {
        return encoded - ((OntapVersionImpl)version).encoded;
    }
}
