package com.netapp.offtap3.internal.ontap.aggregate;

import com.google.common.base.Objects;
import com.netapp.offtap3.AggregateOperationException;
import com.netapp.offtap3.ControllerResource;
import com.netapp.offtap3.Size;
import com.netapp.offtap3.internal.AbstractRefreshable;
import com.netapp.offtap3.internal.ontap.controller.AbstractOntapController;
import com.netapp.offtap3.internal.ontap.support.ZapiCache;
import com.netapp.offtap3.ontap.ApiFailedException;
import com.netapp.offtap3.ontap.Aggregate;
import com.netapp.offtap3.ontap.OntapController;
import com.netapp.offtap3.ontap.OntapDataController;

import com.netapp.offtap3.ontap.NaElement;

import org.apache.log4j.Logger;

import java.util.Collections;
import java.util.Date;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

public abstract class AggregateImpl extends AbstractRefreshable implements Aggregate {

    private static final Logger LOGGER = Logger.getLogger(AggregateImpl.class);

    private AbstractOntapController controller;

    private String uuid;
    private String state;
    private Size sizeUsed;
    private Size sizeTotal;
    private Size sizeAvailable;
    private int percentageUsed;
    private List<String> volumeNames;
    private Set<String> diskTypes = Collections.emptySet();
    private String blockType = null;
    private String stripingType;
    private boolean isStriped;
    private long diskCount;
    private boolean isHybrid;

    private static final Map<String, String> businessDiskTypes = new HashMap<String, String>();
    static {
        businessDiskTypes.put("ATA", "SATA");
        businessDiskTypes.put("BSAS", "SATA");
        businessDiskTypes.put("SAS", "SAS");
        businessDiskTypes.put("VMDISK", "SAS");
    }

    public AggregateImpl(AbstractOntapController<? extends OntapDataController> controller, String name, String uuid, String state, long sizeUsed,
                         int percentageUsed, long sizeTotal, long sizeAvailable, List<String> volumeNames,
                         long diskCount) {
        this(controller, name, sizeAvailable);
        this.uuid = uuid;
        this.state = state;
        this.sizeUsed = Size.createB(sizeUsed);
        this.percentageUsed = percentageUsed;
        this.sizeTotal = Size.createB(sizeTotal);
        this.volumeNames = Collections.unmodifiableList(volumeNames);
        this.diskCount = diskCount;
    }

    public AggregateImpl(AbstractOntapController controller, String name, long sizeAvailable) {
        super(name, new Date(), controller);
        this.controller = controller;
        this.sizeAvailable = Size.createB(sizeAvailable);
    }

    protected AbstractOntapController<? extends OntapController> getController() {
        return controller;
    }

    @Override
    public String getUuid() {
        return uuid;
    }

    public void setUuid(String uuid) {
        this.uuid = uuid;
    }

    @Override
    public String getState() {
        return state;
    }

    public void setState(String state) {
        this.state = state;
    }

    @Override
    public Size getSizeUsed() {
        return sizeUsed;
    }

    public void setSizeUsed(long sizeUsed) {
        this.sizeUsed = Size.createB(sizeUsed);
    }

    @Override
    public Size getSizeTotal() {
        return sizeTotal;
    }

    public void setSizeTotal(long sizeTotal) {
        this.sizeTotal = Size.createB(sizeTotal);
    }

    @Override
    public Size getSizeAvailable() {
        return sizeAvailable;
    }

    public void setSizeAvailable(long sizeAvailable) {
        this.sizeAvailable = Size.createB(sizeAvailable);
    }

    @Override
    public int getPercentageUsed() {
        return percentageUsed;
    }

    public void setPercentageUsed(int percentageUsed) {
        this.percentageUsed = percentageUsed;
    }

    @Override
    public List<String> getVolumeNames() {
        return volumeNames;
    }

    @Override
    public Set<String> getDiskTypes() {
        return diskTypes;
    }

    public void setDiskTypes(Set<String> diskTypes) {
        this.diskTypes = diskTypes;
    }

    @Override
    public Set<String> getBusinessDiskTypes() {
    	Set<String> retVal = new HashSet<String>();
    	for (String curr : getDiskTypes()) {
    	    String toUse = curr; 
    	    String override = businessDiskTypes.get("" + curr);
            if (override != null) {
                toUse = override;
            }
            retVal.add(toUse);
    	}
    	return retVal;
    }

    @Override
    public String getBlockType() {
        return blockType;
    }

    public void setBlockType(String blockType) {
        this.blockType = blockType;
    }

    @Override
    public String getStripingType() {
        return stripingType;
    }

    public void setStripingType(String stripingType) {
        this.stripingType = stripingType;
    }

    @Override
    public boolean isStriped() {
        return isStriped;
    }

    public void setStriped(boolean isStriped) {
        this.isStriped = isStriped;
    }

    @Override
    public long getDiskCount() {
        return diskCount;
    }

    public void setDiskCount(long diskCount) {
        this.diskCount = diskCount;
    }

    @Override
    public boolean isHybrid() {
        return isHybrid;
    }

    public void setHybrid(boolean isHybrid) {
        this.isHybrid = isHybrid;
    }

    @Override
    public void rename(String newName) {
        NaElement xi = new NaElement("aggr-rename");
        xi.addNewChild("aggregate", getName());
        xi.addNewChild("new-aggregate-name", newName);
        try {
            controller.invokeZapi(xi, ZapiCache.NO_CACHE);
            controller.getZapiCache().blastCacheValues("aggr");
            LOGGER.info(controller + " renamed aggr " + getName() + " to " + newName + ".");
            getController().getProxy().refresh(ControllerResource.aggregates, ControllerResource.volumes);
        } catch (ApiFailedException e) {
            LOGGER.error(controller + " failed to rename aggr: " + getName(), e);
            throw new AggregateOperationException(controller + " failed to rename aggr: " + getName(), e);
        }
    }

    @Override
    public Aggregate refresh() {
        return (Aggregate) super.refresh();
    }

    public boolean equals(Object obj) {
        return obj instanceof Aggregate && Objects.equal(getUuid(), ((Aggregate) obj).getUuid());
    }

    public int hashCode() {
        return getUuid().hashCode();
    }
}
