package com.netapp.offtap3.internal.ontap.controller;

import com.netapp.offtap3.ControllerResource;
import com.netapp.offtap3.FcpAdapter;
import com.netapp.offtap3.LogOperationException;
import com.netapp.offtap3.StorageSystem;
import com.netapp.offtap3.factories.DataLoaderException;
import com.netapp.offtap3.internal.CachingControllerProxy;
import com.netapp.offtap3.internal.ontap.EmsLoggerImpl;
import com.netapp.offtap3.internal.ontap.MCCManager;
import com.netapp.offtap3.internal.ontap.OntapVersionImpl;
import com.netapp.offtap3.internal.ontap.SnmpHostManager;
import com.netapp.offtap3.internal.ontap.dataloader.CModePerformanceObjectsLoaderDelegate;
import com.netapp.offtap3.internal.ontap.dataloader.OntapDataCache;
import com.netapp.offtap3.internal.ontap.proxy.AdminSvmProxy;
import com.netapp.offtap3.internal.ontap.support.ZapiCache;
import com.netapp.offtap3.ontap.AdminSvm;
import com.netapp.offtap3.ontap.Aggregate;
import com.netapp.offtap3.ontap.ApiFailedException;
import com.netapp.offtap3.ontap.CModeAggregate;
import com.netapp.offtap3.ontap.CModeIpInterface;
import com.netapp.offtap3.ontap.ClusterPeer;
import com.netapp.offtap3.ontap.ControllerWithSnmpSupport;
import com.netapp.offtap3.ontap.DataSvm;
import com.netapp.offtap3.ontap.EmsLogger;
import com.netapp.offtap3.ontap.EmsLoggerSpec;
import com.netapp.offtap3.ontap.NaElement;
import com.netapp.offtap3.ontap.NodeSvm;
import com.netapp.offtap3.ontap.SvmOperationException;
import com.netapp.offtap3.utils.SafeConcurrentHashMap;
import com.netapp.offtap3.utils.StringUtils;

import org.apache.log4j.Logger;

import java.util.Arrays;
import java.util.Collections;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.concurrent.CopyOnWriteArrayList;
import java.util.concurrent.CopyOnWriteArraySet;

public class AdminSvmImpl extends AbstractOntapController<AdminSvm> implements AdminSvm, ControllerWithSnmpSupport {
    private static final Logger logger = Logger.getLogger(AdminSvmImpl.class);
    private static final Set<StorageSystem> storageSystems = Collections.singleton(StorageSystem.adminsvm);
    private AdminSvm proxy;
    private String clusterState;
    private String clusterStateReason;

    private Map<String, DataSvm> dataSvms = new SafeConcurrentHashMap<String, DataSvm>();
    private Map<String, NodeSvm> nodeSvms = new SafeConcurrentHashMap<String, NodeSvm>();
    private String clusterId;
    private Set<String> snmpHosts = new CopyOnWriteArraySet<String>();
    private SnmpHostManager snmpHostManager;
    private CModePerformanceObjectsLoaderDelegate performanceObjectsLoader;
    private List<FcpAdapter> fcpAdapters = new CopyOnWriteArrayList<FcpAdapter>();
    private List<ClusterPeer> clusterPeers = new CopyOnWriteArrayList<ClusterPeer>();

    public AdminSvmImpl(OntapConnection connInfo, OntapVersionImpl version) {
        super(connInfo, version);
        snmpHostManager = new SnmpHostManager(this);
        performanceObjectsLoader = new CModePerformanceObjectsLoaderDelegate(this);
    }

    @Override
    public EmsLogger getLogger(EmsLoggerSpec spec) {
        // TODO - consider whether this is a good use of polymorphism, or just a bad idea.

        // it's kinda goofy to even try to send an EMS log to the admin, but AeonFlux
        // supports it today, and this is how they do it - pick a node, any node that is up.
        for (NodeSvm node : getProxy().getNodeSvms().values()) {
            if (node.isUp()) {
                return new EmsLoggerImpl(node.getConnection(), spec);
            }
        }
        throw new LogOperationException("None of the nodes are up!  EMS logger is unavailable");
    }

    @Override
    public String getModel() {
        // Technically speaking, you could have a heterogenous cluster,
        // in which case, there is not just a single model name for the cluster.  Models are
        // a physical concept.  They are a property of the node.
        Set<String> models = new HashSet<String>();
        for (NodeSvm svm : getProxy().getNodeSvms().values()) {
            String model = svm.getModel();
            if (model == null || model.isEmpty()) {
                logger.warn(svm + " has no model information");
            } else {
                models.add(model);
            }
        }
        if (models.isEmpty()) {
            throw new SvmOperationException(this + " does not have any nodes svms with model information!");
        } else if (models.size() > 1) {
            logger.warn(this + " is a mixed model cluster, consisting of: " + models);
        }
        return models.iterator().next();
    }

    @Override
    public Map<String, DataSvm> getDataSvms() {
        return dataSvms;
    }

    public void setDataSvms(List<DataSvm> dataSvmList) {
        Map<String, DataSvm> newSvms = new SafeConcurrentHashMap<String, DataSvm>();
        for (DataSvm svm : dataSvmList) {
            newSvms.put(svm.getName(), svm);
        }
        dataSvms = newSvms;
    }

    @Override
    public DataSvm getDataSvm(String name) {
        return dataSvms.get(name);
    }

    public void setNodeSvms(List<NodeSvm> nodeSvmList) {
        Map<String, NodeSvm> newSvms = new SafeConcurrentHashMap<String, NodeSvm>();
        for (NodeSvm svm : nodeSvmList) {
            newSvms.put(svm.getName(), svm);
        }
        nodeSvms = newSvms;
    }
    

    @Override
    public Map<String, NodeSvm> getNodeSvms() {
        return nodeSvms;
    }

    @Override
    public NodeSvm getNodeSvm(String name) {
        return nodeSvms.get(name);
    }

    @Override
    public List<ClusterPeer> getClusterPeers() {
        return clusterPeers;
    }

    public void setClusterPeers(List<ClusterPeer> clusterPeers) {
        this.clusterPeers = clusterPeers;
    }

    public String getClusterId() {
        return clusterId;
    }

    public void setClusterId(String clusterId) {
        this.clusterId = clusterId;
    }

    @Override
    public String getId() {
        String proposedId = super.getId();
        if ("00000000-0000-0000-0000-000000000000".equals(proposedId)) {
            // on some configs and/or versions of cDOT, the admin vserver uuid
            // was coming back as all 0's.  If that happens, we need a better
            // ID, so we use the connect IP address
            return getConnection().getIpAddress();
        }
        return proposedId;
    }
    
    @Override
    public String getMccState() {
    	return MCCManager.getInstance().getMccState(this.getId());
    }

    @Override
    public String getClusterState() {
        return clusterState;
    }

    public void setClusterState(String clusterState) {
        this.clusterState = clusterState;
    }

    @Override
    public String getClusterStateReason() {
        return clusterStateReason;
    }

    public void setClusterStateReason(String clusterStateReason) {
        this.clusterStateReason = clusterStateReason;
    }

    @Override
    public Set<String> getSnmpHosts() {
        return snmpHosts;
    }

    @Override
    public void setSnmpHosts(Set<String> snmpHosts) {
        this.snmpHosts = snmpHosts;
    }

    @Override
    public void addSnmpTrapListener(String hostIp) throws DataLoaderException {
        snmpHostManager.addSnmpTrapListener(hostIp);
    }

    @Override
    public void removeSnmpTrapListener(String hostIp) throws DataLoaderException {
        snmpHostManager.removeSnmpTrapListener(hostIp);
    }

    @Override
    public Map<String, DataSvm> getChildControllers() {
        return getDataSvms();
    }

    @Override
    public Map<String, CModeAggregate> getAggregates() {
        return (Map<String, CModeAggregate>) super.getAggregates();
    }

    @Override
    public Map<String, CModeIpInterface> getIpInterfaces() {
        return (Map<String, CModeIpInterface>) super.getIpInterfaces();
    }

    @Override
    public List<FcpAdapter> getFcpAdapters() {
        return fcpAdapters;
    }

    public void setFcpAdapters(List<FcpAdapter> fcpAdapters) {
        this.fcpAdapters = fcpAdapters;
    }

    @Override
    public List<String> getPerformanceObjectInstanceList(String objectName) {
        return performanceObjectsLoader.getPerformanceObjectInstanceList(objectName);
    }

    @Override
    public Map<String, Long> getPerformanceCounters(String counter, String objectName, List<String> objectInstances) {
        return performanceObjectsLoader.getPerformanceCounters(counter, objectName, objectInstances);
    }

    @Override
    public boolean enableNdmp() {
        NaElement xi = new NaElement("system-cli");
        xi.addNewChild("priv", "admin");
        NaElement xa = new NaElement("args");

        if (getVersion().atLeastSN()) {
            xa.addNewChild("arg", "system");
            xa.addNewChild("arg", "services");
        }
        xa.addNewChild("arg", "ndmp");
        xa.addNewChild("arg", "on");
        xa.addNewChild("arg", "*");
        xi.addChildElem(xa);

        boolean success = true;
        try {
            NaElement xo = invokeZapi(xi, ZapiCache.NO_CACHE);
            int rc = xo.getChildIntValue("cli-result-value", 0);
            String output = xo.getChildContent("cli-output");
            if (rc != 1) {
                logger.error(String.format("Unable to enable ndmp. rc = %s. output = %s", rc, output));
                success = false;
            }
        } catch (ApiFailedException e) {
            if (e.getErrno() == 13011) {
                logger.warn("DFM doesn't return anything...");
            } else {
                logger.error("Unable to enable ndmp. - " + e.getMessage(), e);
                success = false;
            }
        }
        return success;
    }

    @Override
    public DataSvm createSvm(String vServerName, Aggregate aggr) {
        NaElement xi = new NaElement("vserver-create");
        xi.addNewChild("vserver-name", vServerName);
        xi.addNewChild("root-volume", vServerName + "_rootvol");
        xi.addNewChild("root-volume-aggregate", aggr.getName());
        NaElement nsSwitch = new NaElement("name-server-switch");
        nsSwitch.addNewChild("nsswitch", "file");
        xi.addChildElem(nsSwitch);
        xi.addNewChild("root-volume-security-style", "unix");
        try {
            invokeZapi(xi, ZapiCache.NO_CACHE);
            getZapiCache().blastCacheValues("vserver-", "volume-");
            logger.info("Create vServer " + vServerName + ".");
        } catch (ApiFailedException e) {
            logger.error(this + " createVServer: " + vServerName + " Unable to create volume.", e);
            throw new SvmOperationException("Failed to createVServer " + vServerName, e);
        }
        // retrieve the new data svm via the proxy/loader mechanism
        AdminSvm proxy = getProxy();
        proxy.refresh(ControllerResource.svms);
        return proxy.getDataSvm(vServerName);
    }

    @Override
    public synchronized AdminSvm getProxy() {
        if (proxy == null) {
            proxy = (AdminSvm) CachingControllerProxy.create(AdminSvm.class, AdminSvmProxy.class, this,
                                                             new OntapDataCache(this));
        }
        return proxy;
    }

    @Override
    public void addLicenses(String... newLicenses) {
        if (getVersion().atLeastSN()) {
            try {
                NaElement xi = new NaElement("license-v2-add");
                NaElement codes = new NaElement("codes");
                for (String license : newLicenses) {
                    NaElement licenseCode = new NaElement("license-code-v2", license);
                    codes.addChildElem(licenseCode);
                }
                xi.addChildElem(codes);
                invokeZapi(xi, ZapiCache.NO_CACHE);
                getZapiCache().blastCacheValues("license-");
                getProxy().refresh(ControllerResource.licenses);
            } catch (ApiFailedException e) {
                logger.error(this + " failed to add licenses: " + Arrays.asList(newLicenses), e);
            }
        } else {
            throw new UnsupportedOperationException(this + " predates 8.2.0, and is unable to add licenses");
        }
    }

    @Override
    public void deleteLicense(String licensedPackage) {
        logger.debug(this + " deleteLicense package: " + licensedPackage);


        if (getVersion().atLeastSN()) {
            try {
                NaElement xi = new NaElement("license-v2-delete");
                NaElement license = new NaElement("package", licensedPackage);
                xi.addChildElem(license);
                invokeZapi(xi, ZapiCache.NO_CACHE);
                getZapiCache().blastCacheValues("license-");
                getProxy().refresh(ControllerResource.licenses);
            } catch (ApiFailedException e) {
                logger.error(this + " failed to deleteLicense package: " + licensedPackage, e);
            }
        } else {
            throw new UnsupportedOperationException(this + " predates 8.2.0 and does not support deleting licenses");
        }
    }

    @Override
    public String getNdmpPassword() {

        String password = null;

        if (getVersion().atLeastSN()) {
            NaElement xi = new NaElement("system-cli");
            xi.addNewChild("priv", "admin");
            NaElement xa = new NaElement("args");

            xa.addNewChild("arg", "vserver");
            xa.addNewChild("arg", "services");
            xa.addNewChild("arg", "ndmp");
            xa.addNewChild("arg", "generate-password");
            xa.addNewChild("arg", "-vserver");
            xa.addNewChild("arg", getName());
            xa.addNewChild("arg", "-user");
            xa.addNewChild("arg", getConnection().getUserName());
            xi.addChildElem(xa);

            try {
                NaElement xo = invokeZapi(xi, ZapiCache.NO_CACHE);
                int rc = xo.getChildIntValue("cli-result-value", 0);
                String output = xo.getChildContent("cli-output");
                if (rc != 1) {
                    logger.error(
                            String.format("Unable to get ndmp password on vserver. rc = %s. output = %s", rc, output));
                } else if (output != null) {
                    if (output.length() > 0) {
                        int ind = output.indexOf("Password");
                        if (ind != -1) {
                            password = output.substring(ind + 10).trim();
                        }
                    }
                }
            } catch (ApiFailedException e) {
                logger.error(this + " was unable to get ndmp password on vserver. - " + e.getMessage(), e);
                throw new SvmOperationException(
                        this + " was unable to get ndmp password on vserver. - " + e.getMessage(), e);
            }
        } else {
            //For RR mode, ndmp copy password is hardcoded to changeMe
            // Update the ndmp copy password on filer using command
            //ndmp password -node
            password = "changeMe";
        }
        return password;
    }
    

    @Override
    public Boolean isNdmpConfiguredInVserverAwareMode() {

        NaElement xi = new NaElement("system-cli");
        xi.addNewChild("priv", "admin");
        NaElement xa = new NaElement("args");
        xa.addNewChild("arg", "system service ndmp node-scope-mode status");
        xi.addChildElem(xa);

        try {
            NaElement xo = invokeZapi(xi, ZapiCache.LONG_CACHE);
            int rc = xo.getChildIntValue("cli-result-value", 0);
            String output = xo.getChildContent("cli-output");
            if (rc != 1) {
                logger.error(
                        String.format("Unable to get ndmp node-scope-mode status. rc = %s. output = %s", rc, output));
                throw new SvmOperationException(
                        "isNDMPConfiguredInVserverAwareMode: System-cli Zapi failed with rc:" + rc + output);
            } else if (StringUtils.isReal(output)) {
                return output.contains("is disabled.");
            } else {
                return null;
            }
        } catch (ApiFailedException e) {
            logger.error("Unable to get ndmp node-scope-mode status on vserver. - " + e.getMessage(), e);
            throw new SvmOperationException("Unable to get ndmp node-scope-mode status on vserver. - " + e.getMessage(),
                                            e);
        }
    }

    @Override
    public Set<StorageSystem> getStorageSystems() {
        return storageSystems;
    }
}
