package com.netapp.offtap3.internal.ontap.controller;

import com.google.common.collect.ArrayListMultimap;
import com.google.common.collect.ListMultimap;
import com.netapp.offtap3.CifsShare;
import com.netapp.offtap3.CifsShareOperationException;
import com.netapp.offtap3.Controller;
import com.netapp.offtap3.ControllerResource;
import com.netapp.offtap3.ExportOperationException;
import com.netapp.offtap3.FcpAdapter;
import com.netapp.offtap3.File;
import com.netapp.offtap3.IpInterface;
import com.netapp.offtap3.IscsiConnection;
import com.netapp.offtap3.IscsiSession;
import com.netapp.offtap3.Version;
import com.netapp.offtap3.VolumeOperationException;
import com.netapp.offtap3.internal.CifsShareImpl;
import com.netapp.offtap3.internal.IscsiSessionImpl;
import com.netapp.offtap3.internal.ontap.OntapVersionImpl;
import com.netapp.offtap3.internal.ontap.dataloader.OntapDataCache;
import com.netapp.offtap3.internal.ontap.lun.SevenModeLunImpl;
import com.netapp.offtap3.internal.ontap.support.ZapiCache;
import com.netapp.offtap3.ontap.ApiFailedException;
import com.netapp.offtap3.ontap.ClusterFailoverInfo;
import com.netapp.offtap3.ontap.Filer;
import com.netapp.offtap3.ontap.FilerAggregate;
import com.netapp.offtap3.ontap.FilerVolume;
import com.netapp.offtap3.ontap.LunOperationException;
import com.netapp.offtap3.ontap.NaElement;
import com.netapp.offtap3.ontap.OntapLun;
import com.netapp.offtap3.ontap.OntapLunSpec;
import com.netapp.offtap3.ontap.OntapVolumeSpec;
import com.netapp.offtap3.ontap.SnapMirrorOperationException;
import com.netapp.offtap3.ontap.SnapMirrorRelation;
import com.netapp.offtap3.ontap.SnapVault;
import com.netapp.offtap3.ontap.SpaceReserve;
import com.netapp.offtap3.utils.Utils;

import org.apache.log4j.Logger;

import java.util.ArrayList;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.concurrent.CopyOnWriteArrayList;
import java.util.concurrent.CopyOnWriteArraySet;


public abstract class FilerImpl<I extends Filer> extends AbstractOntapDataController<I> implements Filer {
    private static final Logger logger = Logger.getLogger(FilerImpl.class);

    private String etcVolName;
    private String vstorageStatus;
    private String uuid;

    private ClusterFailoverInfo cfInfo;
    private Set<String> iscsiPortalNames = new CopyOnWriteArraySet<String>();
    private List<SnapVault> primarySnapVaults = new ArrayList<SnapVault>();
    private List<SnapVault> secondarySnapVaults = new ArrayList<SnapVault>();
    private List<FcpAdapter> fcpAdapters = new CopyOnWriteArrayList<FcpAdapter>();

    private OntapDataCache dataCache;

    public FilerImpl(OntapConnection connInfo, Version version) {
        super(connInfo, version);
        dataCache = new OntapDataCache(connInfo);
    }

    // convenience method since 7-mode discovery of interfaces is so jumbled and piecemeal
    public void addInterface(IpInterface intf) {
        getIpInterfaces().put(intf.getIpAddress(), intf);
    }

    @Override
    public Map<String, IpInterface> getIpInterfaces() {
        return (Map<String, IpInterface>) super.getIpInterfaces();
    }

    @Override
    public Map<String, FilerVolume> getVolumes() {
        return (Map<String, FilerVolume>) super.getVolumes();
    }

    @Override
    public FilerVolume getVolume(String name) {
        return (FilerVolume) super.getVolume(name);
    }


    @Override
    public Map<String, FilerAggregate> getAggregates() {
        return (Map<String, FilerAggregate>) super.getAggregates();
    }

    /**
     * Checks if SnapMirror is available on the given controller.
     *
     * @return boolean contents of <is-available> tag from snapmirror-get-status
     */
    @Override
    public boolean isSnapMirrorAvailable() throws SnapMirrorOperationException {
        boolean available = false;

        try {
            logger.debug("Getting snapmirror-get-status for " + getName());
            NaElement xi = new NaElement("snapmirror-get-status");
            NaElement snapMirrorInfo = invokeZapi(xi, ZapiCache.LONG_CACHE);
            available = Boolean.parseBoolean(snapMirrorInfo.getChildContent("is-available"));
            logger.debug("Finished snapmirror-get-status for " + getName());
        } catch (Exception e) {
            logger.error("Exception while getting SnapMirrorInfo for " + getName(), e);
            throw new SnapMirrorOperationException(
                    "Exception while getting SnapMirrorInfo for " + getName() + ": " + e.getMessage(), e);
        }

        return available;
    }

    @Override
    public List<SnapMirrorRelation> getSnapMirrorRelations() {
        return (List<SnapMirrorRelation>) super.getSnapMirrorRelations();
    }

    @Override
    public void enableSnapMirror() {
        logger.debug("turnOnSnapMirror: " + getName());
          NaElement xi = new NaElement("snapmirror-on");
          try {
              invokeZapi(xi, ZapiCache.NO_CACHE);
              getZapiCache().blastCacheValues("snap");
              logger.debug("Turning on SnapMirror");
          } catch (ApiFailedException e) {
              logger.error("Unable to turn on SnapMirror.");
              throw new SnapMirrorOperationException("Unable to turn on SnapMirror" +e.getMessage(), e);
          }
    }

    @Override
    public void disableSnapMirror() {
        logger.debug("turnOffSnapMirror: " + getName());
        NaElement xi = new NaElement("snapmirror-off");
        try {
            invokeZapi(xi, ZapiCache.NO_CACHE);
            getZapiCache().blastCacheValues("snap");
            logger.debug(this + " turning off SnapMirror");
        } catch (ApiFailedException e) {
            logger.error("Unable to turn off SnapMirror.");
            throw new SnapMirrorOperationException("Unable to turn off SnapMirror" +e.getMessage(), e);
        }
    }

    /**
     * Helper method to handle all 3 possible cases:
     * 1) pfiler - return that connection
     * 2) direct vfiler - return that connection
     * 3) pass-through vfiler - return the pfiler's connection
     *
     * @return a connection that is not a passthrough vfiler connection
     */
    public abstract OntapConnection getNonPassThroughConnection();

    public Set<String> getIscsiPortalNames() {
        return iscsiPortalNames;
    }

    public void setIscsiPortalNames(Set<String> names) {
        iscsiPortalNames = names;
    }

    @Override
    public String getVstorageStatus() {
        return vstorageStatus;
    }

    public void setVstorageStatus(String status) {
        vstorageStatus = status;
    }

    public String getEtcVolName() {
        return etcVolName;
    }

    public void setEtcVolName(String etcVolName) {
        this.etcVolName = etcVolName;
    }

    public String getUuid() {
        return uuid;
    }

    public void setUuid(String uuid) {
        this.uuid = uuid;
    }

    public ClusterFailoverInfo getClusterFailoverInfo() {
        return cfInfo;
    }

    public void setClusterFailoverInfo(ClusterFailoverInfo info) {
        this.cfInfo = info;
    }

    // convenience method - not exposed on interface until proven necessary
    public boolean isDirectVFiler() {
        return false;
    }

    // convenience method - not exposed on interface until proven necessary
    public boolean isPassThroughVFiler() {
        return false;
    }

    /**
     * @return true only if this is SN (or better) and options cf.mode is 'ha'
     */
    public boolean isHaMode() {
        return "ha".equalsIgnoreCase(cfInfo.getCurrentMode());
    }

    @Override
    public List<FcpAdapter> getFcpAdapters() {
        return fcpAdapters;
    }

    public void setFcpAdapters(List<FcpAdapter> fcpAdapters) {
        this.fcpAdapters = fcpAdapters;
    }


    /**
     * Check on the cluster failover status for this filer.  This call
     * makes the network trip to the filer every time, to insure accurate reporting.
     *
     * @return true if the cluster failover status is either:
     * <ol>
     * <li>disabled</li>
     * <li>enabled and healthy/connected</li>
     * </ol>
     */
    public boolean isClusterUsable() {

        getProxy().refresh(ControllerResource.cf_info);

        // Two basic scenarios where the cluster is usable:
        // 1) The cluster is "turned off", which is evidenced by
        //   a) pre-SN will have a null CF state (could be disabled, could be unlicensed, etc.)
        //   b) on SN, we've seen a non-null CF state (like "ERROR"), so we ignore that if it is in non-HA mode
        // 2) The cluster is healthy and connected
        // "connected" means the CF is healthy and ready
        return cfInfo == null || cfInfo.getState() == null || cfInfo.getState().equalsIgnoreCase("connected") ||
                "non_ha".equalsIgnoreCase(cfInfo.getCurrentMode());
    }

    public void addSystemCliArgs(NaElement command) {
    }

    @Override
    public void modifyVStorageApiEnabled(boolean enabled) {
        super.modifyVStorageApiEnabled(enabled);
        try {
            //removed nfs.v4.enable per burt867572
            setNfsVStorageOption("nfs.v3.enable", enabled);
            setNfsVStorageOption("nfs.vstorage.enable", enabled);
        } catch (ApiFailedException e) {
            throw new ExportOperationException("An error occurred while enabling the vStorage APIs for controller " + getName() + ".", e);
        }
    }

    @Override
    public boolean isVStorageApiSupported() {
        return getVersion() != null && getVersion().compareTo(new OntapVersionImpl("8.1.1", 8, 1, 1)) >= 0;
    }

    private void setNfsVStorageOption(String option, boolean on) throws ApiFailedException {
        NaElement xi = new NaElement("options-set");
        xi.addNewChild("name", option);
        xi.addNewChild("value", (on ? "on" : "off"));
        NaElement xo = invokeZapi(xi, ZapiCache.NO_CACHE);
        getZapiCache().blastCacheValues("options-set");
        if (xo != null) {
            logger.debug("Set NFS option " + option + " to " + on + ": " + getName());
        }
    }

    @Override
    public boolean enableNdmp() {
        NaElement xi = new NaElement("system-cli");
        xi.addNewChild("priv", "admin");
        NaElement xa = new NaElement("args");

        addSystemCliArgs(xa);

        xa.addNewChild("arg", "ndmpd");
        xa.addNewChild("arg", "on");
        xi.addChildElem(xa);
        try {
            NaElement xo = getNonPassThroughConnection().invokeZapi(xi, ZapiCache.NO_CACHE);
            int rc = xo.getChildIntValue("cli-result-value", 0);
            String output = xo.getChildContent("cli-output");
            if (rc == 1) {
                return true;
            }
            logger.error(String.format(this + " was unable to enable ndmp. rc = %s. output = %s", rc, output));
        } catch (ApiFailedException e) {
            logger.error(this + " failed to enable ndmp", e);
        }
        return false;
    }

    @Override
    public CifsShare createCifsShare(String shareName, String sharePath) {
        NaElement xi = new NaElement("cifs-share-add");
        xi.addNewChild("share-name", shareName);
        xi.addNewChild("path", sharePath);
        try {
            invokeZapi(xi, ZapiCache.NO_CACHE);
            getZapiCache().blastCacheValues("cifs-");

            CifsShareImpl share = new CifsShareImpl(this);
            share.setName(shareName);
            share.setMountPoint(sharePath);
            return share;
        } catch (ApiFailedException e) {
            logger.error(this + " failed to create CIFS share named: " + shareName + " at path: " + sharePath, e);
            throw new CifsShareOperationException(
                    this + " failed to create CIFS share named: " + shareName + " at path: " + sharePath, e);
        }
    }

    @Override
    public List<SnapVault> getSnapVaultsPrimary() {
        return primarySnapVaults;
    }

    @Override
    public SnapVault getSnapVaultPrimary(File destinationPath) {
        for (SnapVault vault : primarySnapVaults) {
            if (vault.getDestinationPath().equals(destinationPath.getFullPath())) {
                return vault;
            }
        }
        return null;
    }

    public void setSnapVaultsPrimary(List<SnapVault> primarySnapVaults) {
        this.primarySnapVaults = primarySnapVaults;
    }

    @Override
    public List<SnapVault> getSnapVaultsSecondary() {
        return secondarySnapVaults;
    }

    @Override
    public SnapVault getSnapVaultSecondary(File destinationPath) {
        for (SnapVault vault : secondarySnapVaults) {
            if (vault.getDestinationPath().equals(destinationPath.getFullPath())) {
                return vault;
            }
        }
        return null;
    }

    public void setSnapVaultsSecondary(List<SnapVault> secondarySnapVaults) {
        this.secondarySnapVaults = secondarySnapVaults;
    }

    public OntapDataCache getDataCache() {
        return dataCache;
    }

    @Override
    public FilerVolume createVolume(OntapVolumeSpec volSpec) {
        logger.debug(this + " createVolume: " + volSpec);

        String volumeName = volSpec.getVolumeName();
        NaElement xi = new NaElement("volume-create");
        xi.addNewChild("volume", volumeName);
        xi.addNewChild("containing-aggr-name", volSpec.getAggregateName());
        xi.addNewChild("size", volSpec.getSize().toMB() + "m");
        xi.addNewChild("space-reserve", volSpec.getSpaceReserve().name());

        if (volSpec.getVmAlignSector() > 0 && volSpec.getVmAlignSuffix() != null) {
            xi.addNewChild("vm-align-sector", Integer.toString(volSpec.getVmAlignSector()));
            xi.addNewChild("vm-align-suffix", volSpec.getVmAlignSuffix());
        }

        try {
            NaElement xo = invokeZapi(xi, ZapiCache.NO_CACHE);
            getZapiCache().blastCacheValues("volume-");
            logger.info(this + " created volume " + volSpec.getVolumeName() + ":" + xo.getAttr("status") + ".");
        } catch (ApiFailedException e) {
            addZapiExecutionError("Failed to create volume named: " + volumeName, xi, e);
            logger.error(this + " failed to create volume named: " + volumeName, e);
            throw new VolumeOperationException(this + " failed to create volume named" + volumeName, e);
        }

        getProxy().refresh(ControllerResource.volumes);
        FilerVolume volume = getProxy().getVolume(volumeName);
        volume.setQtreeSecurity(volSpec.getSecurityStyle());
        return volume;
    }

    @Override
    public OntapLun createLun(OntapLunSpec lunSpec) {
        logger.debug("createLun: " + lunSpec);
        NaElement xi;
        int num = 0;

        if (lunSpec.getVolume() == null) {
            throw new IllegalArgumentException("No volume was specified in spec: " + lunSpec);
        }

        String newLunName = lunSpec.getLunName();
        String fullPath = "/vol/" + lunSpec.getVolume().getName() + "/" + newLunName;

        while ((getLun(fullPath) != null)) {
            num++;
            newLunName = String.format("%s%02d", lunSpec.getLunName(), num);
            fullPath = "/vol/" + lunSpec.getVolume().getName() + "/" + newLunName;
        }

        try {

            if (lunSpec.getPrefixSize() != null) {
                xi = new NaElement("system-cli");
                xi.addNewChild("priv", "diag");

                NaElement xa = new NaElement("args");

                OntapConnection conn = getNonPassThroughConnection();
                addSystemCliArgs(xa);

                xa.addNewChild("arg", "lun");
                xa.addNewChild("arg", "create");
                xa.addNewChild("arg", "-s");
                xa.addNewChild("arg", String.valueOf(lunSpec.getSize().toBytes()));
                xa.addNewChild("arg", "-t");
                xa.addNewChild("arg", "vmware");
                xa.addNewChild("arg", "-P");
                xa.addNewChild("arg", lunSpec.getPrefixSize().toString());

                if (lunSpec.getSpaceReserve() == SpaceReserve.none) {
                    xa.addNewChild("arg", "-o");
                    xa.addNewChild("arg", "noreserve");
                }

                xa.addNewChild("arg", fullPath);

                xi.addChildElem(xa);
                conn.invokeZapi(xi, ZapiCache.NO_CACHE);
            } else {
                xi = new NaElement("lun-create-by-size");
                xi.addNewChild("ostype", lunSpec.getType());
                xi.addNewChild("path", fullPath);
                xi.addNewChild("size", String.valueOf(lunSpec.getSize().toBytes()));
                if (lunSpec.getSpaceReserve() == SpaceReserve.none) {
                    xi.addNewChild("space-reservation-enabled", "false");
                } else {
                    xi.addNewChild("space-reservation-enabled", "true");
                }
                invokeZapi(xi, ZapiCache.NO_CACHE);

            }
            getZapiCache().blastCacheValues("lun-");
            logger.info("Create " + newLunName + ".");
        } catch (Exception e) {
            logger.error("Failed to create lun " + newLunName + ".", e);
            throw new LunOperationException("Failed to createLun" + fullPath + ".  Error: " + e.getMessage(), e);
        }

        // lun has been created - now load it up
        SevenModeLunImpl<I> lunImpl = new SevenModeLunImpl<I>(this);
        lunImpl.setPath(fullPath);
        OntapLun lun = lunImpl.refresh();
        if (lun == null) {
            Utils.sleep(1000);
        }
        return lunImpl.refresh();
    }

    public List<IscsiSession> getIscsiSessions() {

        logger.debug("getIscsiSessions: 7-mode (begin)");

        List<IscsiSession> sessions = new ArrayList<IscsiSession>();

        try {
            NaElement xo = invokeZapi(new NaElement("iscsi-session-list-info"), ZapiCache.NO_CACHE);
            List<NaElement> sessionInfoList = xo.getChildByName("iscsi-session-list-entries").getChildren();

            ListMultimap<Integer, IscsiConnection> connections = ArrayListMultimap.create();
            for (IscsiConnection conn : getIscsiConnections()) {
                connections.put((int)conn.getSessionId(), conn);
            }

            for (NaElement sessionInfo : sessionInfoList) {
                String nodename = sessionInfo.getChildContent("initiator-nodename");
                String alias = sessionInfo.getChildContent("initiator-alias");
                int sessionId = sessionInfo.getChildIntValue("target-session-id", 0);
                List<IscsiConnection> iscsiConnections = new ArrayList<IscsiConnection>();

                if (sessionInfo.getChildByName("iscsi-session-connection-list-entries").getChildContent("iscsi-session-connection-list-entry-info") != null) {
                    iscsiConnections = connections.get(sessionId);
                }

                IscsiSessionImpl iscsiSession = new IscsiSessionImpl();
                iscsiSession.setInitiatorNodeName(nodename);
                iscsiSession.setInitiatorAlias(alias);
                iscsiSession.setIscsiConnections(iscsiConnections);
                sessions.add(iscsiSession);
            }
        } catch (Exception e) {
            logger.error("Unable to getIscsiSessions.", e);
            throw new LunOperationException(this + " was unable to getIscsiSessions: " + e.getMessage(), e);
        }

        logger.debug("getIscsiSessions: 7-mode (end): " + sessions);

        return sessions;
    }

    public void addInterfacesToIscsiInitiators(List<String> initiatorNames, List<IpInterface> interfaces) {
        logger.debug("addInterfacesToIscsiInitiators: 7-mode (begin)");

        for (String init : initiatorNames) {
            StringBuilder sb = new StringBuilder();
            NaElement xi = new NaElement("system-cli");
            xi.addNewChild("priv", "admin");
            NaElement xa = new NaElement("args");

            OntapConnection conn = getNonPassThroughConnection();
            addSystemCliArgs(xa);

            xa.addNewChild("arg", "iscsi");
            xa.addNewChild("arg", "interface");
            xa.addNewChild("arg", "accesslist");
            xa.addNewChild("arg", "add");
            xa.addNewChild("arg", "-f");
            xa.addNewChild("arg", init);
            sb.append(init);
            for (IpInterface intf : interfaces){
                xa.addNewChild("arg", intf.getName());
                sb.append(" ").append(intf.getName());
            }
            logger.debug("addInterfacesToIscsiInitiators will submit:" +sb.toString());
            xi.addChildElem(xa);
            try {
                NaElement xo = conn.invokeZapi(xi, ZapiCache.NO_CACHE);
                getZapiCache().blastCacheValues("iscsi", "igroup");
                int rc = xo.getChildIntValue("cli-result-value", 0);
                String output = xo.getChildContent("cli-output");
                if (rc != 1) {
                    logger.error(String.format("Unable to add interfaces to iscsi initiators. rc = %s. output = %s", rc, output));
                }
            } catch (ApiFailedException ex) {
                logger.error("Failed to add interfaces to iscsi initiators.", ex);
                throw new LunOperationException(this + " failed to add interfaces to iscsi initiators.", ex);
            }
        }
        logger.debug("addInterfacesToIscsiInitiators: 7-mode (end)");
    }
}
