package com.netapp.offtap3.internal.ontap.controller;

import com.netapp.offtap3.ControllerConnection;
import com.netapp.offtap3.FcpAdapter;
import com.netapp.offtap3.Size;
import com.netapp.offtap3.StorageSystem;
import com.netapp.offtap3.internal.CachingControllerProxy;
import com.netapp.offtap3.internal.HasParentControllerImpl;
import com.netapp.offtap3.internal.HasProxy;
import com.netapp.offtap3.internal.ontap.dataloader.OntapDataCache;
import com.netapp.offtap3.internal.ontap.proxy.NodeSvmProxy;
import com.netapp.offtap3.internal.ontap.support.ZapiCache;
import com.netapp.offtap3.ontap.AdminSvm;
import com.netapp.offtap3.ontap.ApiFailedException;
import com.netapp.offtap3.ontap.CModeAggregate;
import com.netapp.offtap3.ontap.CModeIpInterface;
import com.netapp.offtap3.ontap.NaElement;
import com.netapp.offtap3.ontap.NodeSvm;
import org.apache.log4j.Logger;

import java.util.Collections;
import java.util.List;
import java.util.Map;
import java.util.Set;

public class NodeSvmImpl extends AbstractOntapController<NodeSvm> implements NodeSvm, HasProxy<NodeSvm>,
        HasParentControllerImpl {

    private static final Logger logger = Logger.getLogger(NodeSvmImpl.class);
    private static final Set<StorageSystem> storageSystems = Collections.singleton(StorageSystem.nodesvm);
    private Size flashCacheCapacity;
    private String highAvailabilityMode;
    private String clusterStatus;
    private String partner;
    private boolean asupEnabled;
    private AdminSvmImpl adminSvmImpl;
    private NodeSvm proxy;
    private OntapDataCache dataCache;

    public NodeSvmImpl(OntapConnection connInfo, AdminSvmImpl adminSvmImpl, OntapDataCache dataCache) {
        super(connInfo, adminSvmImpl.getVersion());
        this.adminSvmImpl = adminSvmImpl;
        this.dataCache = dataCache;
    }


    @Override
    public void setName(String name) {
        super.setName(name);
        getConnection().setContext(name);
    }

    @Override
    public void setConnection(ControllerConnection controllerConnection) {
        super.setConnection(controllerConnection);
        getConnection().setContext(getName());
    }

    @Override
    public boolean isAsupEnabled() {
        return asupEnabled;
    }

    public void setAsupEnabled(boolean on) {
        asupEnabled = on;
    }

    @Override
    public Size getFlashCacheCapacity() {
        return flashCacheCapacity;
    }

    //from ZAPI-DOC: capacity [0..2^31-1] integer Advertised capacity of the device, in gigabytes.
    public void setFlashCacheCapacity(int capacityInGB) {
        flashCacheCapacity = Size.createGB(capacityInGB);
    }

    @Override
    public String getHighAvailabilityMode() {
        return highAvailabilityMode;
    }

    public void setHighAvailabilityMode(String highAvailabilityMode) {
        this.highAvailabilityMode = highAvailabilityMode;
    }

    @Override
    public String getClusterStatus() {
        return clusterStatus;
    }

    public void setClusterStatus(String clusterStatus) {
        this.clusterStatus = clusterStatus;
    }

    @Override
    public String getPartner() {
        return partner;
    }

    public void setPartner(String partner) {
        this.partner = partner;
    }

    @Override
    public List<FcpAdapter> getFcpAdapters() {
        return getParentController().getFcpAdapters();
    }

    @Override
    public Map<String, CModeIpInterface> getIpInterfaces() {
        return (Map<String, CModeIpInterface>) super.getIpInterfaces();
    }

    @Override
    public boolean isUp() {
    	return getUptime() > 0;
    }
    
    @Override
    public long getUptime() {
        logger.debug("Enter getUptime for node: " + getName());
        NaElement xi = new NaElement("system-node-get-iter");
        NaElement query = new NaElement("query");
        NaElement nodeDetails = new NaElement("node-details-info");
        nodeDetails.addNewChild("node", getName());
        query.addChildElem(nodeDetails);
        xi.addChildElem(query);
        try {
        	// note - originally this had NO_CACHE, but really, caching is just used for
        	// unit testing speedup, lots of calls want information RIGHT NOW
            NaElement xo = getParentControllerImpl().invokeZapi(xi, ZapiCache.LONG_CACHE);
            if (xo != null && xo.getChildByName("attributes-list") != null) {
                NaElement details = xo.getChildByName("attributes-list").getChildByName("node-details-info");
                if (details != null && (details.getChildByName("node-uptime") != null)) {
                	return details.getChildLongValue("node-uptime", 0);
                } else {
                	return 0;
                }
            }
        } catch (ApiFailedException e) {
            logger.error("ZAPI failure trying to get uptime for node: " + getName(), e);
        } catch (Exception e) {
            logger.error(
                    "Unexpected error processing ZAPI result for node-uptime.  Assuming node is down: " + getName(), e);
        }
        logger.debug("No uptime information found.  Assuming node is down: " + getName());
        return 0;
    }

    @Override
    public AdminSvm getParentController() {
        return adminSvmImpl.getProxy();
    }

    @Override
    public AdminSvmImpl getParentControllerImpl() {
        return adminSvmImpl;
    }

    @Override
    public Map<String, CModeAggregate> getAggregates() {
        return (Map<String, CModeAggregate>) super.getAggregates();
    }

    @Override
    public synchronized NodeSvm getProxy() {
        if (proxy == null) {
            proxy = (NodeSvm) CachingControllerProxy.create(NodeSvm.class, NodeSvmProxy.class, this, dataCache);
        }
        return proxy;
    }

    @Override
    public Set<StorageSystem> getStorageSystems() {
        return storageSystems;
    }
}
