package com.netapp.offtap3.internal.ontap.controller;

import java.util.List;

import org.apache.log4j.Logger;
import org.w3c.dom.Document;

import com.netapp.offtap3.ControllerConnection;
import com.netapp.offtap3.DiscoveryError;
import com.netapp.offtap3.OfftapPreferences;
import com.netapp.offtap3.factories.ControllerType;
import com.netapp.offtap3.internal.ControllerConnectionImpl;
import com.netapp.offtap3.internal.ontap.InternalZapiHandler;
import com.netapp.offtap3.internal.ontap.utils.ExpiringArrayList;
import com.netapp.offtap3.internal.ontap.zapi.ZAPIInvoker;
import com.netapp.offtap3.ontap.ApiFailedException;
import com.netapp.offtap3.ontap.NaElement;
import com.netapp.offtap3.ontap.ZapiExecutionError;
import com.netapp.offtap3.ontap.ZapiHandler;
import com.netapp.offtap3.utils.StringUtils;
import com.netapp.offtap3.utils.Utils;

public class OntapConnection extends ControllerConnectionImpl implements InternalZapiHandler, ZapiHandler {

    private static final Logger LOGGER = Logger.getLogger(OntapConnection.class);

    private int connectionTimeout;
    private int timeout;
    private String context;
    private String uniqueKey = "";
    
    private List<DiscoveryError> discoveryErrors = new ExpiringArrayList<DiscoveryError>(getOfftapPreferences().getZapiErrorTimeoutInOntapConnection());
    private List<ZapiExecutionError> zapiErrors = new ExpiringArrayList<ZapiExecutionError>(getOfftapPreferences().getZapiErrorTimeoutInOntapConnection());
    private boolean useTls = true;

    public OntapConnection(OfftapPreferences preferences) {
        super(preferences);
        setControllerType(ControllerType.ONTAP);
        updateUniqueKey();
    }

    public OntapConnection(ControllerConnection connection) {
        super(connection);
        setControllerType(ControllerType.ONTAP);
        if (connection instanceof OntapConnection) {
            OntapConnection ontapConnection = (OntapConnection) connection;
            this.connectionTimeout = ontapConnection.connectionTimeout;
            this.timeout = ontapConnection.timeout;
            this.context = ontapConnection.context;
            this.uniqueKey = ontapConnection.uniqueKey;
            this.discoveryErrors.addAll(ontapConnection.discoveryErrors);
            this.zapiErrors.addAll(ontapConnection.zapiErrors);
            this.useTls = ontapConnection.useTls;            
        }
        updateUniqueKey();
    }

    /**
     * Determine the "caller method" that was invoked prior to hitting the invokeZapi layer.
     * NOTE: This method is hard coded to the method name invokeZAPI!
     *
     * @param t the exception that was thrown when the ZAPI failed.  the "caller method" will be
     *          located in this stack trace.
     * @return the method following the last invokeZAPI (in case of overload later).
     */
    protected static String getCaller(Throwable t) {
        String returnMe = null;
        try {
            if (t != null && t.getStackTrace() != null && t.getStackTrace().length > 2) {
                StackTraceElement caller = null;
                boolean grabNext = false;
                for (StackTraceElement ste : t.getStackTrace()) {
                    if ("invokeZAPI".equals(ste.getMethodName())) {
                        grabNext = true;
                    } else {
                        if (grabNext) {
                            caller = ste;
                            break;
                        }
                    }
                }
                if (caller != null) {
                    returnMe = "(called from " + caller.getClassName() + "." + caller.getMethodName() + "(...)" + " on line " + caller.getLineNumber() + ")";
                }
            }
        } catch (Exception e) {
            //ignore
        }
        return returnMe;
    }

    public int getConnectionTimeout() {
        return connectionTimeout;
    }

    public void setConnectionTimeout(int connectionTimeout) {
        this.connectionTimeout = connectionTimeout;
    }

    public int getTimeout() {
        return timeout;
    }

    public void setTimeout(int timeout) {
        this.timeout = timeout;
    }

    @Override
    public void setIpAddress(String ipAddress) {
        super.setIpAddress(ipAddress);
        updateUniqueKey();
    }

    public boolean useTls() {
        return useTls;
    }

    public void setTls(boolean useTls) {
        this.useTls = useTls;
    }

    public String getContext() {
        return context;
    }

    public void setContext(String context) {
        this.context = context;
        updateUniqueKey();
    }

    public String getUniqueKey() {
        return uniqueKey;
    }

    private void updateUniqueKey() {
        uniqueKey = getIpAddress();
        if (StringUtils.isReal(context)) {
            uniqueKey += ":" + context;
        }
    }

    public List<DiscoveryError> getDiscoveryErrors() {
        return discoveryErrors;
    }

    public List<ZapiExecutionError> getZapiExecutionErrors() {
        return zapiErrors;
    }

    public void addDiscoveryError(String message, Throwable ex) {       
        discoveryErrors.add(new DiscoveryError(message, ex));
    }

    public void addZapiExecutionError(String message, NaElement input, Throwable e) {     
        zapiErrors.add(new ZapiExecutionError(message, input, e));
    }

    public NaElement invokeZapi(NaElement xmlInput, long cacheDuration) throws ApiFailedException {
        ZAPIInvoker zapiInvoker = ZAPIInvoker.getLastInstance();
        if (zapiInvoker == null) {
            zapiInvoker = new ZAPIInvoker();
        }

        //handle null issues before we start
        if (xmlInput == null) {
            String message = "invokeZAPI: called with null!  controller=" + this + " " + getCaller(new Exception());
            LOGGER.error(message);
            throw new NullPointerException(message);
        }

        long startTime = System.currentTimeMillis();
        return zapiInvoker.invokeFilerBasedZapi(this, xmlInput, cacheDuration, startTime);
    }

    @Override
    public Document invokeZapi(Document input, long cacheDuration) throws ApiFailedException {
        // For now, we're going to convert from the W3C Element to NaElement. In the future, we hope to remove NaElement
        // entirely.
        NaElement naInput = Utils.convertToNaElement(input);
        NaElement naOutput = invokeZapi(naInput, cacheDuration);
        return Utils.convertToElement(naOutput);
    }
}
