package com.netapp.offtap3.internal.ontap.dataloader;

import com.netapp.offtap3.Refreshable;
import com.netapp.offtap3.Resource;
import com.netapp.offtap3.internal.DataCache;
import com.netapp.offtap3.internal.ontap.exports.ExportImpl;
import com.netapp.offtap3.internal.ontap.exports.ExportSecurityRuleImpl;
import com.netapp.offtap3.internal.ontap.exports.QtreeImpl;
import com.netapp.offtap3.internal.ontap.volume.OntapVolumeImpl;
import com.netapp.offtap3.ontap.Export;
import com.netapp.offtap3.ontap.ExportSecurityRule;
import com.netapp.offtap3.internal.ontap.InternalZapiHandler;
import com.netapp.offtap3.ontap.NaAuthenticationException;
import com.netapp.offtap3.ontap.Qtree;
import com.netapp.offtap3.ontap.SecurityStyle;
import com.netapp.offtap3.ontap.NaElement;
import com.netapp.offtap3.utils.SafeConcurrentHashMap;

import org.apache.log4j.Logger;

import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.concurrent.ConcurrentHashMap;


public abstract class AbstractExportDataLoader<V extends OntapVolumeImpl> extends OntapCachingDataLoader<V> {

    private static final Logger LOGGER = Logger.getLogger(AbstractExportDataLoader.class);

    protected AbstractExportDataLoader(DataCache<NaElement, InternalZapiHandler> dataCache) {
        super(dataCache);
    }

    @Override
    protected Refreshable doRefresh(V loadTarget, Resource resource, Refreshable refreshTarget) {
        doLoad(loadTarget, LoadType.refresh, Collections.singleton(resource));
        if (refreshTarget instanceof Qtree) {
            Qtree qtree = (Qtree) refreshTarget;
            return loadTarget.getQtree(qtree.getPath());
        } else {
            Export export = (Export) refreshTarget;
            return (Refreshable) loadTarget.getExports().get(export.getPath());
        }
    }

    protected abstract QtreeImpl createQtree(OntapVolumeImpl volume, String name);

    @Override
    protected boolean parse(OntapVolumeImpl volume, NaElement data, Resource resource) {
        LOGGER.debug("parse: begin");
        Map<String, Qtree> qtrees = new SafeConcurrentHashMap<String, Qtree>();

        Set<String> qtreesWithQuotas = new HashSet<String>();
        NaElement quotaList = data.getChildByName("quotas");
        for (NaElement quotaElem : quotaList.getChildren()) {
            String volumeName = quotaElem.getChildContent("volume");
            if (volumeName.equals(volume.getName())) {
                qtreesWithQuotas.add(quotaElem.getChildContent("quota"));
            }
        }

        List<NaElement> qtreesElem = data.getChildByName("qtrees").getChildren();
        for (NaElement qtreeElem : qtreesElem) {
            String owningVolume = qtreeElem.getChildContent("volume");
            if (owningVolume.equals(volume.getName())) {
                QtreeImpl qtree = createQtree(volume, qtreeElem.getChildContent("qtree"));
                qtree.setStatus(qtreeElem.getChildContent("status"));
                qtree.setId(qtreeElem.getChildIntValue("id", -1));
                qtree.setSecurityStyle(SecurityStyle.valueOf(qtreeElem.getChildContent("security-style")));
                qtree.setHasQuota(qtreesWithQuotas.contains(qtree.getName()));
                qtrees.put(qtree.getPath(), qtree);
            }
        }
        
        List<NaElement> exportsElem = data.getChildByName("rules").getChildren();
        for (NaElement exportElem : exportsElem) {
            String qtreePathname = exportElem.getChildContent("pathname");
            String exportPathname = exportElem.getChildContent("actual-pathname");
            String pathnameToUse = qtreePathname;
            if (exportPathname != null) {
                pathnameToUse = exportPathname;
            }
            QtreeImpl qtree = (QtreeImpl) qtrees.get(qtreePathname);
            if (qtree == null && (!(volume.getJunctionPath() != null && volume.getJunctionPath().equals(pathnameToUse)))) {
                // this just means we encountered an export for a qtree that is not on this volume
                continue;
            }
            ExportImpl export = new ExportImpl(volume);
            export.setPath(pathnameToUse);

            // 7-mode doesn't have a field for junction-path, but if the volume is
            // exported (aka qtree named "") then that export is basically the volume's junction path
            if ((qtree != null && (qtree.getName().equals("") && volume.getJunctionPath() == null))) {
                volume.setJunctionPath(export.getPath());
            }

            List<ExportSecurityRule> rules = new ArrayList<ExportSecurityRule>();
            for (NaElement securityRuleElem : exportElem.getChildByName("security-rules").getChildren()) {
                // First verify this rule applies to "sys" which means it applies to normal unix NFS clients.
                if (securityFlavorApplies(securityRuleElem, "sys")) {
                    rules.add(getExportSecurityRule(securityRuleElem));
                }
            }
            export.setSecurityRules(rules);
            if (qtree != null) {
                qtree.getExports().put(export.getPath(), export);
            } else {
                volume.getRawExports().put(export.getPath(), export);
            }
            
        }

        volume.setQtrees(qtrees);
        LOGGER.debug("parse: end");
        return true;
    }

    private boolean securityFlavorApplies(NaElement securityRuleInfo, String securityFlavor) {
        if (securityRuleInfo.getChildByName("sec-flavor") != null) {
            List<NaElement> flavors = securityRuleInfo.getChildByName("sec-flavor").getChildren();
            for (NaElement flavor : flavors) {
                if (securityFlavor.equals(flavor.getChildContent("flavor"))) {
                    return true;
                }
            }
        }
        return false;
    }

    private ExportSecurityRuleImpl getExportSecurityRule(NaElement exportElement) {
        ExportSecurityRuleImpl securityRule = new ExportSecurityRuleImpl();

        if (exportElement.getChildByName("anon") != null) {
            securityRule.setAnonymousUser(exportElement.getChildContent("anon"));
        }
        if (exportElement.getChildByName("read-only") != null) {
            addOnlyNew(securityRule.getReadOnlyHosts(), getHostnames(exportElement.getChildByName("read-only").getChildren()));
        }
        if (exportElement.getChildByName("read-write") != null) {
            addOnlyNew(securityRule.getReadWriteHosts(), getHostnames(exportElement.getChildByName("read-write").getChildren()));
        }
        if (exportElement.getChildByName("root") != null) {
            addOnlyNew(securityRule.getRootAccessHosts(), getHostnames(exportElement.getChildByName("root").getChildren()));
        }

        return securityRule;
    }

    private void addOnlyNew(List<String> existingHosts, List<String> newHosts) {
        if (!existingHosts.isEmpty()) {
            for (String host : newHosts) {
                if (!existingHosts.contains(host)) {
                    existingHosts.add(host);
                }
            }
        } else {
            existingHosts.addAll(newHosts);
        }
    }

    private List<String> getHostnames(List<NaElement> exportElements) {
        ArrayList<String> list = new ArrayList<String>();

        for (NaElement element : exportElements) {
            if (Boolean.parseBoolean(element.getChildContent("all-hosts"))) {
                list.add("all-hosts");
            } else {
                // If negate=true, it means to specifically exclude this host
                // Only add it to the list if negate = false
                if (!Boolean.parseBoolean(element.getChildContent("negate"))) {
                    list.add(element.getChildContent("name"));
                }
            }
        }
        return list;
    }

}
