/*
 * $Id: //depot/prod/aeonflux/R3.0.1/src/main/java/com/netapp/offtap3/internal/ontap/dataloader/AdminSvmFcpAdapterDataLoader.java#1 $
 *
 * Copyright (c) 2014 NetApp, Inc.
 * All rights reserved.
 */
package com.netapp.offtap3.internal.ontap.dataloader;

import com.netapp.offtap3.FcpAdapter;
import com.netapp.offtap3.Refreshable;
import com.netapp.offtap3.OfftapPreferences;
import com.netapp.offtap3.Resource;
import com.netapp.offtap3.factories.DataLoaderException;
import com.netapp.offtap3.internal.AbstractDataLoader;
import com.netapp.offtap3.internal.FcpAdapterImpl;
import com.netapp.offtap3.internal.ontap.controller.AdminSvmImpl;
import com.netapp.offtap3.internal.ontap.support.ZapiCache;
import com.netapp.offtap3.ontap.ApiFailedException;
import com.netapp.offtap3.ontap.NaElement;
import com.netapp.offtap3.utils.SafeConcurrentHashMap;

import org.apache.log4j.Logger;

import java.util.ArrayList;
import java.util.Collections;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

public class AdminSvmFcpAdapterDataLoader extends AbstractDataLoader<AdminSvmImpl> {

    private static final Logger logger = Logger.getLogger(AdminSvmFcpAdapterDataLoader.class);
    private OfftapPreferences prefs;

    public AdminSvmFcpAdapterDataLoader(OfftapPreferences prefs) {
        this.prefs = prefs;
    }

    @Override
    protected Refreshable doRefresh(AdminSvmImpl adminSvm, Resource resource, Refreshable refreshTarget) {
        doLoad(adminSvm, LoadType.refresh, Collections.singleton(resource), (FcpAdapter) refreshTarget);
        for (FcpAdapter adapter : adminSvm.getFcpAdapters()) {
            if (adapter.getName().equals(refreshTarget.getName())) {
                return adapter;
            }
        }
        return null;
    }

    @Override
    protected boolean doLoad(AdminSvmImpl adminSvm, LoadType loadType, Set<Resource> resources) {
        return doLoad(adminSvm, loadType, resources, null);
    }

    protected boolean doLoad(AdminSvmImpl adminSvm, LoadType loadType, Set<Resource> resources, FcpAdapter reloadTarget) {
        logger.debug("doLoad: start");
        NaElement initiatorInput = new NaElement("fcp-initiator-get-iter");
        NaElement adapterInput = new NaElement("fcp-adapter-get-iter");
        if (loadType == LoadType.refresh && reloadTarget != null) {
            adapterInput.addChildElem(getQuery(reloadTarget, "fcp-config-adapter-info"));
            initiatorInput.addChildElem(getQuery(reloadTarget, "fcp-adapter-initiators-info"));
        }

        Map<String, Set<String>> initiators = new SafeConcurrentHashMap<String, Set<String>>();
        List<FcpAdapter> fcpAdapters = new ArrayList<FcpAdapter>();

        CdotIterativeLoader<List<FcpAdapter>> adapterLoader = new CdotIterativeLoader<List<FcpAdapter>>(adminSvm, adapterInput,
                                                                                                 new FcpAdapterParser(
                                                                                                         initiators,
                                                                                                         adminSvm),
                                                                                                 prefs.getIteratorSize(),
                                                                                                 ZapiCache.LONG_CACHE
        );
        CdotIterativeLoader<Map<String, Set<String>>> initiatorLoader = new CdotIterativeLoader<Map<String, Set<String>>>(
                adminSvm, initiatorInput, new FcpInitiatorsParser(adminSvm.getName()), prefs.getIteratorSize(),
                ZapiCache.LONG_CACHE);


        try {
            initiatorLoader.load(initiators);
        } catch (ApiFailedException e) {
            throw new DataLoaderException(
                    "Error occurred while loading FCP initiators for Admin SVM " + adminSvm.getName(), e);
        }

        try {
            adapterLoader.load(fcpAdapters);
        } catch (ApiFailedException e) {
            throw new DataLoaderException(
                    "Error occurred while loading FCP adapters for Admin SVM " + adminSvm.getName(), e);
        }
        adminSvm.setFcpAdapters(fcpAdapters);
        logger.debug("doLoad: end");
        return true;
    }

    private NaElement getQuery(FcpAdapter reloadTarget, String configElementName) {
        NaElement query = new NaElement("query");
        NaElement configInfo = new NaElement(configElementName);
        configInfo.addNewChild("adapter", reloadTarget.getName());
        query.addChildElem(configInfo);
        return query;
    }

    private class FcpAdapterParser implements DataParser<List<FcpAdapter>> {
        private Map<String, Set<String>> initiators;
        private AdminSvmImpl adminSvm;

        public FcpAdapterParser(Map<String, Set<String>> initiators, AdminSvmImpl adminSvm) {
            this.initiators = initiators;
            this.adminSvm = adminSvm;
        }

        @Override
        public void parse(List<FcpAdapter> cummulativeResults, NaElement results) {
            logger.debug("getFcAdapters: start");
            NaElement attrList = results.getChildByName("attributes-list");
            if (attrList != null) {
                List<NaElement> adaptersInfo = attrList.getChildren();
                for (NaElement element : adaptersInfo) {
                    FcpAdapterImpl adapter = new FcpAdapterImpl(adminSvm);
                    adapter.setName(element.getChildContent("adapter"));
                    adapter.setPort(element.getChildContent("port-name"));
                    adapter.setState(element.getChildContent("state"));
                    if (initiators.containsKey(adapter.getName())) {
                        adapter.setInitiators(initiators.get(adapter.getName()));
                    }
                    cummulativeResults.add(adapter);
                    logger.debug("parse FCP adapters: " + adminSvm.getName() + " added " + adapter);
                }
            } else {
                logger.error("getFcAdapters: " + adminSvm.getName() + " fcAdapters returned null.");
            }
            logger.debug("parser: end");
        }
    }

    private class FcpInitiatorsParser implements DataParser<Map<String, Set<String>>> {
        private String controllerName;

        public FcpInitiatorsParser(String controllerName) {
            this.controllerName = controllerName;
        }

        @Override
        public void parse(Map<String, Set<String>> cummulativeResults, NaElement results) {
            logger.debug("parse FCP initiators: start " + controllerName);
            NaElement attrList = results.getChildByName("attributesList");
            if (attrList != null) {
                for (NaElement element : attrList.getChildren()) {
                    String adapter = element.getChildContent("adapter");
                    Set<String> portNames = new HashSet<String>();
                    if (element.getChildByName("fcp-connected-initiators").getChildContent(
                            "fcp-connected-initiator-info") != null) {
                        for (NaElement el1 : element.getChildByName("fcp-connected-initiators").getChildren()) {
                            if (el1.getChildContent("port-name") != null) {
                                portNames.add(el1.getChildContent("port-name"));
                            }
                        }
                    }
                    cummulativeResults.put(adapter, portNames);
                }
            }
            logger.debug("parse FCP initiators: end - " + cummulativeResults + " initators found");
        }
    }
}
