package com.netapp.offtap3.internal.ontap.dataloader;


import com.netapp.offtap3.Resource;
import com.netapp.offtap3.factories.DataLoaderException;
import com.netapp.offtap3.internal.AbstractDataLoader;
import com.netapp.offtap3.internal.ontap.controller.AbstractOntapController;
import com.netapp.offtap3.internal.ontap.controller.AdminSvmImpl;
import com.netapp.offtap3.internal.ontap.support.ZapiCache;
import com.netapp.offtap3.utils.Pair;
import com.netapp.offtap3.utils.SafeConcurrentHashMap;
import com.netapp.offtap3.utils.Utils;
import com.netapp.offtap3.internal.ontap.zapi.CModeZapiAccessPermissions;
import com.netapp.offtap3.internal.ontap.zapi.ZapiCommandMap;
import com.netapp.offtap3.ontap.ApiFailedException;
import com.netapp.offtap3.ontap.ZapiAccessPermissions;
import com.netapp.offtap3.ontap.NaElement;

import org.apache.log4j.Logger;

import java.util.ArrayList;
import java.util.List;
import java.util.Map;
import java.util.Set;


public class AdminSvmZapiAccessPermissionsDataLoader extends AbstractDataLoader<AdminSvmImpl> {

    private static final Logger logger = Logger.getLogger(AdminSvmZapiAccessPermissionsDataLoader.class);

    @Override
    protected boolean doLoad(AdminSvmImpl controller, LoadType loadType, Set<Resource> resources) throws DataLoaderException {

        ZapiAccessPermissions perms = getAccessPermissions(controller);
        controller.setZapiAccessPermissions(perms);
        return true;
    }


    ZapiAccessPermissions getAccessPermissions(AbstractOntapController controller) throws DataLoaderException {

        // first lets testGetVServerRolePairs
        // then foreach of those roles, lets AccessLevelCommandTreePairs
        // then for all of that crap, lets cross-reference to getZapiToCommandMappings
        long startTime = System.currentTimeMillis();
        List<Pair<String, String>> vserverRolePairs = getVServerRolePairs(controller, controller.getConnection().getUserName());
        CModeZapiAccessPermissions retVal = new CModeZapiAccessPermissions();


        for (Pair<String, String> p : vserverRolePairs) {
            String vServer = p.getA();
            String role = p.getB();
            Map<String, List<Map>> accessLevelsToCommandTrees = getAccessLevelCommandTreeMaps(controller, role);
            List<Map> accessLevelsToCommandTreesForThisVServer = new ArrayList<Map>();
            if (accessLevelsToCommandTrees.get(vServer) != null) {
                accessLevelsToCommandTreesForThisVServer.addAll(accessLevelsToCommandTrees.get(vServer));
            }

            retVal.setAccessLevelsForVServer(vServer, accessLevelsToCommandTreesForThisVServer);
            for (String zapi : ZapiCommandMap.getInstance().getZapis()) {
                String commandPattern = ZapiCommandMap.getInstance().getCommand(zapi, controller);
                String accessLevel = CModeZapiAccessPermissions.determineAccessLevel(commandPattern,
                                            accessLevelsToCommandTreesForThisVServer);
                // expect to see "none", "all", and "readonly"
                if ("all".equals(accessLevel)) {
                    // full permissions
                    retVal.addRWZapi(vServer, zapi);
                } else if ("readonly".equals(accessLevel)) {
                    // readonly
                    retVal.addROZapi(vServer, zapi);
                } else {
                    // none
                }
            }
        }

        long duration = System.currentTimeMillis() - startTime;
        logger.debug("Loading permissions - took " + duration + " ms");
        return retVal;
    }

    /**
     * Return a map of String vServer to list<pair> of access level, command trees
     */
    Map<String, List<Map>> getAccessLevelCommandTreeMaps(AbstractOntapController controller, String role) throws DataLoaderException {

        CdotIterativeLoader<Map<String, List<Map>>> loginRoleLoader = new CdotIterativeLoader<Map<String, List<Map>>>(controller,
                new NaElement("security-login-role-get-iter"),
                new AccessLevelDataParser(role),
                controller.getOfftapPreferences().getIteratorSize(),
                ZapiCache.SHORT_CACHE);

        Map<String, List<Map>> retVal = new SafeConcurrentHashMap<String, List<Map>>();

        try {
            loginRoleLoader.load(retVal);
        } catch (ApiFailedException e) {
            throw new DataLoaderException(controller + " failed to load security info for role: " + role);
        }
        return retVal;
    }

    List<Pair<String, String>> getVServerRolePairs(AbstractOntapController controller, String targetUsername) throws DataLoaderException {

        List<Pair<String, String>> securityList = new ArrayList<Pair<String, String>>();
        CdotIterativeLoader<List<Pair<String,String>>> vserverRoleLoader = new CdotIterativeLoader<List<Pair<String,String>>>(controller,
                 new NaElement("security-login-get-iter"),
                 new VServerRoleParser(targetUsername),
                 controller.getOfftapPreferences().getIteratorSize(),
                 ZapiCache.SHORT_CACHE);

         try {
             vserverRoleLoader.load(securityList);
         } catch (ApiFailedException e) {
             throw new DataLoaderException(controller + " failed to load security info for user: " + targetUsername, e);
         }
         return securityList;
    }


    static class AccessLevelDataParser implements DataParser<Map<String, List<Map>>> {

        private String role;

        AccessLevelDataParser(String role) {
            this.role = role;
        }

        @Override
        public void parse(Map<String, List<Map>> cummulativeResults, NaElement results) {
            List<NaElement> securityInfo = results.getChildByName("attributes-list").getChildren();

             for (NaElement element : securityInfo) {

                 String accessLevel = element.getChildContent("access-level");
                 String commandDirectory = element.getChildContent("command-directory-name");
                 String vserver = element.getChildContent("vserver");
                 String roleName = element.getChildContent("role-name");
                 if (accessLevel == null || commandDirectory == null || roleName == null ||
                         !role.equals(roleName)) {
                     continue;
                 }
                 List<Map> existingForVServer = cummulativeResults.get(vserver);

                 if (existingForVServer == null) {
                     existingForVServer = new ArrayList<Map>();
                     cummulativeResults.put(vserver, existingForVServer);
                 }
                 existingForVServer.add(Utils.toTypedMap(commandDirectory, accessLevel));
                 logger.debug("getAccessLevelCommandTreePairs:  Vserver(" + vserver +
                         "): added " + (new Pair<String, String>(commandDirectory, accessLevel)));
             }
        }
    }

    static class VServerRoleParser implements DataParser<List<Pair<String,String>>> {

        private final String targetUsername;

        VServerRoleParser(String targetUsername) {
            this.targetUsername = targetUsername;
        }

        @Override
        public void parse(List<Pair<String, String>> securityList, NaElement results) {
            List<NaElement> securityInfo = results.getChildByName("attributes-list").getChildren();

              for (NaElement element : securityInfo) {

                  String username = element.getChildContent("user-name");
                  String role = element.getChildContent("role-name");
                  String vserver = element.getChildContent("vserver");
                  String application = element.getChildContent("application");
                  if (username == null || role == null || vserver == null || !"ontapi".equals(application)) {
                      continue;
                  }
                  Pair<String, String> p = new Pair<String, String>(vserver, role);
                  if (targetUsername.equals(username)) {
                      securityList.add(p);
                  }
                  logger.debug("getVServerRolePairs: " + vserver + " added " + p);
              }
        }
    }
}
