package com.netapp.offtap3.internal.ontap.dataloader;

import com.netapp.offtap3.Refreshable;
import com.netapp.offtap3.OfftapPreferences;
import com.netapp.offtap3.Resource;
import com.netapp.offtap3.factories.DataLoaderException;
import com.netapp.offtap3.internal.AbstractDataLoader;
import com.netapp.offtap3.internal.ontap.controller.DataSvmImpl;
import com.netapp.offtap3.internal.ontap.intf.OntapIGroupImpl;
import com.netapp.offtap3.ontap.ApiFailedException;
import com.netapp.offtap3.ontap.DataSvm;
import com.netapp.offtap3.ontap.OntapIGroup;
import com.netapp.offtap3.ontap.NaElement;

import org.apache.log4j.Logger;

import java.util.ArrayList;
import java.util.HashSet;
import java.util.List;
import java.util.Set;
import java.util.concurrent.CopyOnWriteArrayList;

public class DataSvmIGroupDataLoader extends AbstractDataLoader<DataSvmImpl> {
    private static final Logger LOGGER = Logger.getLogger(DataSvmIGroupDataLoader.class);

    OfftapPreferences prefs;

    public DataSvmIGroupDataLoader(OfftapPreferences prefs) {
        this.prefs = prefs;
    }

    @Override
    protected boolean doLoad(DataSvmImpl dataSvm, LoadType loadType, Set<Resource> resources) throws DataLoaderException {
        try {
            populateIGroups(dataSvm, null);
        } catch (ApiFailedException e) {
            throw new DataLoaderException("Failed to load igroups: " + dataSvm, e);
        }
        return true;
    }

    @Override
    protected Refreshable doRefresh(DataSvmImpl dataSvm, Resource resource, Refreshable refreshTarget) {
        try {
            populateIGroups(dataSvm, refreshTarget.getName());
        } catch (ApiFailedException e) {
            throw new DataLoaderException("Failed to reload igroup " + refreshTarget.getName() + " on svm " + dataSvm.getName(), e);
        }
        return dataSvm.getIGroup(refreshTarget.getName());
    }

    void populateIGroups(final DataSvmImpl dataSvm, String iGroupName) throws ApiFailedException {
        LOGGER.debug("populateIGroup (C-mode): start. iGroupName (null OK): " + iGroupName + " vServer:" +
                             dataSvm.getName());
        //if the iGroupName is null we want to update all igroups, so we create a new map
        //if the iGroupName is NOT null, we just want to update the specifically named igroup,
        //so we use the current map.
        List<OntapIGroup> iGroups;
        if (iGroupName == null) {
            iGroups = new CopyOnWriteArrayList<OntapIGroup>();
        } else {
            iGroups = dataSvm.getIGroups();
            // Remove the old igroup before adding the new one.
            for (OntapIGroup iGroup : new ArrayList<OntapIGroup>(iGroups)) {
                if (iGroup.getName().equals(iGroupName)) {
                	iGroups.remove(iGroup);
                }
            }
        }

        NaElement xi = new NaElement("igroup-get-iter");
        if (iGroupName != null) {
            NaElement query = new NaElement("query");
            NaElement igroupInfo = new NaElement("initiator-group-info");
            igroupInfo.addNewChild("initiator-group-name", iGroupName);
            query.addChildElem(igroupInfo);
            xi.addChildElem(query);
        }
        SimpleCdotIterativeLoader iterativeLoader = new SimpleCdotIterativeLoader(dataSvm, xi);

        for (NaElement iGroupInfo : iterativeLoader.loadChildren()) {
            OntapIGroupImpl iGroup = new OntapIGroupImpl<DataSvmImpl, DataSvm>(dataSvm);
            iGroup.setName(iGroupInfo.getChildContent("initiator-group-name"));
            iGroup.setProtocol(iGroupInfo.getChildContent("initiator-group-type"));
            iGroup.setOsType(iGroupInfo.getChildContent("initiator-group-os-type"));
            iGroup.setUuid(iGroupInfo.getChildContent("initiator-group-uuid"));
            iGroup.setAluaEnabled(Boolean.parseBoolean(iGroupInfo.getChildContent("initiator-group-alua-enabled")));
            iGroup.setPortsetName(iGroupInfo.getChildContent("initiator-group-portset-name"));

            Set<String> initiators = new HashSet<String>();
            NaElement initiatorsList = iGroupInfo.getChildByName("initiators");
            if ((initiatorsList != null) && initiatorsList.getChildContent("initiator-info") != null) {
                for (NaElement el1 : initiatorsList.getChildren()) {
                    String iname = el1.getChildContent("initiator-name");
                    if (iname != null) {
                        initiators.add(iname);
                    }
                }
            }
            iGroup.setInitiatorIds(initiators);

            int lunId;
            if (iGroupInfo.getChildContent("lun-id") != null) {
                lunId = iGroupInfo.getChildIntValue("lun-id", 0);
            } else {
                lunId = -1;
            }
            iGroup.setLunId(lunId);
            iGroups.add(iGroup);
        }


        dataSvm.setIGroups(iGroups);
        LOGGER.debug("populateIGroup (C-mode): done. iGroupName (null OK): " + iGroupName + " vServer: " +
                             dataSvm.getName() + " dataVServer.iGroups: " + dataSvm.getIGroups().size());
    }
}
