package com.netapp.offtap3.internal.ontap.dataloader;

import java.util.Set;

import org.apache.log4j.Logger;

import com.netapp.offtap3.ControllerResource;
import com.netapp.offtap3.Resource;
import com.netapp.offtap3.StorageProtocol;
import com.netapp.offtap3.factories.DataLoaderException;
import com.netapp.offtap3.internal.AbstractDataLoader;
import com.netapp.offtap3.internal.ontap.NfsDetailsImpl;
import com.netapp.offtap3.internal.ontap.controller.DataSvmImpl;
import com.netapp.offtap3.internal.ontap.support.ZapiCache;
import com.netapp.offtap3.ontap.ApiFailedException;
import com.netapp.offtap3.ontap.NaElement;

public class DataSvmProtocolDataLoader extends AbstractDataLoader<DataSvmImpl> {

    private static final Logger logger = Logger.getLogger(DataSvmProtocolDataLoader.class);

    @Override
    protected boolean doLoad(DataSvmImpl dataVServer, LoadType loadType, Set<Resource> resources) throws DataLoaderException {
        logger.debug("Load Data Svm Protocols and vStorage: start, resources: " + resources);
        NaElement xo;

        if (resources.contains(ControllerResource.protocols)) {
            try {
                xo = dataVServer.invokeZapi(new NaElement("iscsi-service-get-iter"), ZapiCache.LONG_CACHE);
                NaElement iscsiAttributesList = xo.getChildByName("attributes-list");
                if (iscsiAttributesList != null) {
                    for (NaElement iscsiServiceInfo : iscsiAttributesList.getChildren()) {
                        if (Boolean.parseBoolean(iscsiServiceInfo.getChildContent("is-available"))) {
                            dataVServer.addSupportedProtocol(StorageProtocol.ISCSI);
                            dataVServer.setIscsiNodeName(iscsiServiceInfo.getChildContent("node-name"));
                            logger.info("Load Data Svm Protocols: " + dataVServer + " iscsi is licensed and enabled.");
                        } else {
                            dataVServer.removeSupportedProtocol(StorageProtocol.ISCSI);
                            logger.warn("Load Data Svm Protocols: " + dataVServer + " iscsi is licensed BUT NOT enabled.");
                        }
                    }
                } else {
                    dataVServer.removeSupportedProtocol(StorageProtocol.ISCSI);
                }
            } catch (ApiFailedException ex) {
                logger.error("Failed to get the iscsi service status", ex);
                dataVServer.removeSupportedProtocol(StorageProtocol.ISCSI);
            }

            try {
                xo = dataVServer.invokeZapi(new NaElement("fcp-service-get-iter"), ZapiCache.LONG_CACHE);
                NaElement fcpAttributesList = xo.getChildByName("attributes-list");
                if (fcpAttributesList != null) {
                    for (NaElement fcpServiceInfo : fcpAttributesList.getChildren()) {
                        if (Boolean.parseBoolean(fcpServiceInfo.getChildContent("is-available"))) {
                            dataVServer.addSupportedProtocol(StorageProtocol.FCP);
                            logger.info("Load Data Svm Protocols: " + dataVServer + " fcp is licensed and enabled.");
                        } else {
                            dataVServer.removeSupportedProtocol(StorageProtocol.FCP);
                            logger.warn("Load Data Svm Protocols: " + dataVServer + " fcp is licensed BUT NOT enabled.");
                        }
                    }
                } else {
                    dataVServer.removeSupportedProtocol(StorageProtocol.FCP);
                }
            } catch (ApiFailedException ex) {
                logger.error("Failed to get the fcp service status", ex);
                dataVServer.removeSupportedProtocol(StorageProtocol.FCP);
            }
        }

        NaElement input = new NaElement("nfs-service-get");
        NaElement desiredAttributes = new NaElement("desired-attributes");
        if (resources.contains(ControllerResource.protocols)) {
            desiredAttributes.addChildElem(new NaElement("is-nfs-access-enabled"));
            desiredAttributes.addChildElem(new NaElement("is-nfsv3-enabled"));
            desiredAttributes.addChildElem(new NaElement("is-nfsv41-enabled"));
        }
        if (resources.contains(ControllerResource.vstorage_status)) {
            desiredAttributes.addChildElem(new NaElement("is-vstorage-enabled"));
        }
        input.addChildElem(desiredAttributes);

        try {
            xo = dataVServer.invokeZapi(input, ZapiCache.LONG_CACHE);
            NaElement nfsAttributes = xo.getChildByName("attributes");

            if (nfsAttributes != null) {
                NaElement nfsServiceInfo = nfsAttributes.getChildByName("nfs-info");
                if (resources.contains(ControllerResource.protocols)) {
                    if (Boolean.parseBoolean(nfsServiceInfo.getChildContent("is-nfs-access-enabled"))) {
                        dataVServer.addSupportedProtocol(StorageProtocol.NFS);
                    } else {
                        dataVServer.removeSupportedProtocol(StorageProtocol.NFS);
                    }

                    logger.info("Load Data Svm Protocols: " + dataVServer + " nfs is licensed and enabled.");
                    NfsDetailsImpl nfsDetails = new NfsDetailsImpl();
                    if (Boolean.parseBoolean(nfsServiceInfo.getChildContent("is-nfsv3-enabled"))) {
                        nfsDetails.setNfsv3Enabled(true);
                        logger.info("Load Data Svm Protocols: " + dataVServer + " nfs v3 is enabled.");
                    }
                    if (Boolean.parseBoolean(nfsServiceInfo.getChildContent("is-nfsv41-enabled"))) {
                        nfsDetails.setNfsv41Enabled(true);
                        nfsDetails.setNfsv41ImplementationIdDomain(nfsServiceInfo.getChildContent("nfsv41-implementation-id-domain"));
                        nfsDetails.setNfsv41ImplementationIdName(nfsServiceInfo.getChildContent("nfsv41-implementation-id-name"));
                        if (Boolean.parseBoolean(nfsServiceInfo.getChildContent("is-nfsv41-pnfs-enabled"))) {
                            nfsDetails.setNfsv41pNFSEnabled(true);
                        }
                        logger.info("Load Data Svm Protocols: " + dataVServer + " nfs v41 is enabled. pNFS enable status=" + nfsDetails.isNfsv41pNFSEnabled());
                    }
                    dataVServer.setNfsDetails(nfsDetails);
                }

                if (resources.contains(ControllerResource.vstorage_status)) {
                    // Grab vStorage details as it's already part of the info returned on this call.
                    dataVServer.setVStorageApiEnabled(Boolean.parseBoolean(nfsServiceInfo.getChildContent("is-vstorage-enabled")));
                    logger.debug("vStorage enabled: " + dataVServer.isVStorageApiEnabled());
                }
            } else {
                if (resources.contains(ControllerResource.protocols)) {
                    dataVServer.removeSupportedProtocol(StorageProtocol.NFS);
                    logger.warn("Load Data Svm Protocols: " + dataVServer + " nfs is licensed BUT NOT enabled.");
                }
                if (resources.contains(ControllerResource.vstorage_status)) {
                    dataVServer.setVStorageApiEnabled(false);
                }
            }
        } catch (ApiFailedException ex) {
            if (ex.getMessage().contains("15661")) { // EOBJECTNOTFOUND
                logger.warn("Load Data Svm Protocols : " + dataVServer + " nfs is licensed BUT NOT enabled.");
            } else {
                logger.error("Failed to get the nfs status", ex);
            }
            if (resources.contains(ControllerResource.protocols)) {
                dataVServer.removeSupportedProtocol(StorageProtocol.NFS);
            }
            if (resources.contains(ControllerResource.vstorage_status)) {
                dataVServer.setVStorageApiEnabled(false);
            }
        }

        logger.debug("Load Data Svm Protocols and vStorage: end");
        return true;
    }
}
