package com.netapp.offtap3.internal.ontap.dataloader;


import com.netapp.offtap3.Refreshable;
import com.netapp.offtap3.Resource;
import com.netapp.offtap3.Size;
import com.netapp.offtap3.factories.DataLoaderException;
import com.netapp.offtap3.internal.AbstractDataLoader;
import com.netapp.offtap3.internal.ontap.Disk;
import com.netapp.offtap3.internal.ontap.aggregate.FilerAggregateImpl;
import com.netapp.offtap3.internal.ontap.controller.FilerImpl;
import com.netapp.offtap3.internal.ontap.support.ZapiCache;
import com.netapp.offtap3.ontap.ApiFailedException;
import com.netapp.offtap3.ontap.Filer;
import com.netapp.offtap3.ontap.FilerAggregate;
import com.netapp.offtap3.ontap.NaElement;
import com.netapp.offtap3.utils.SafeConcurrentHashMap;

import org.apache.log4j.Logger;

import java.util.ArrayList;
import java.util.Collections;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

public class FilerAggregateDataLoader extends AbstractDataLoader<FilerImpl<? extends Filer>> {
    private static final Logger LOGGER = Logger.getLogger(FilerAggregateDataLoader.class);

    @Override
    protected Refreshable doRefresh(FilerImpl<? extends Filer> loadTarget, Resource resource, Refreshable refreshTarget) {
        doLoad(loadTarget, LoadType.refresh, Collections.singleton(resource), refreshTarget);
        return loadTarget.getAggregates().get(refreshTarget.getName());
    }

    @Override
    protected boolean doLoad(FilerImpl<? extends Filer> filer, LoadType loadType, Set<Resource> resources) {
        return doLoad(filer, LoadType.refresh, resources, null);
    }

    protected boolean doLoad(FilerImpl<? extends Filer> filer, LoadType loadType, Set<Resource> resources, Refreshable reloadTarget) {
        LOGGER.debug("Enter populateAggregate");
        NaElement output;

        Map<String, Set<String>> plexes = new SafeConcurrentHashMap<String, Set<String>>();
        try {
            plexes = getPlexes(filer);
        } catch (ApiFailedException ex) {
            LOGGER.warn("Unable to build plexes: " + ex.getMessage());
        }

        long totalAllocated = 0;
        long totalUsed = 0;

        NaElement input = new NaElement("aggr-list-info");
        Map<String, FilerAggregate> aggrs = new SafeConcurrentHashMap<String, FilerAggregate>();
        if (loadType == LoadType.refresh && reloadTarget != null) {
            input.addNewChild("aggregate", reloadTarget.getName());
            aggrs.putAll(filer.getAggregates());
            aggrs.remove(reloadTarget.getName());
        }

        try {
            output = filer.invokeZapi(input, ZapiCache.SHORT_CACHE);
        } catch (ApiFailedException e) {
            throw new DataLoaderException("Failed to load aggrs for: " + filer, e);
        }

        for (NaElement aggrInfo : output.getChildByName("aggregates").getChildren()) {
            String name = aggrInfo.getChildContent("name");
            LOGGER.debug("populateAggregate: " + name);
            Set<String> diskTypes = new HashSet<String>();
            String uuid = aggrInfo.getChildContent("uuid");
            boolean traditional = "trad".equalsIgnoreCase(aggrInfo.getChildContent("type"));
            String state = aggrInfo.getChildContent("state");
            String blockType = aggrInfo.getChildContent("block-type");
            long sizeUsed = aggrInfo.getChildLongValue("size-used", 0L);
            int percentageUsed = aggrInfo.getChildIntValue("size-percentage-used", 0);
            long sizeTotal = aggrInfo.getChildLongValue("size-total", 0L);
            long sizeAvailable = aggrInfo.getChildLongValue("size-available", 0L);
            long diskCount = aggrInfo.getChildLongValue("disk-count", 0L);

            List<String> volumeNames = new ArrayList<String>();
            if (aggrInfo.getChildByName("volumes").getChildContent("contained-volume-info") != null) {
                for (NaElement el1 : aggrInfo.getChildByName("volumes").getChildren()) {
                    if (el1.getChildContent("name") != null) {
                        volumeNames.add(el1.getChildContent("name"));
                    }
                }
            }

            if (!plexes.isEmpty()) {
                if (aggrInfo.getChildByName("plexes").getChildContent("plex-info") != null) {
                    for (NaElement el1 : aggrInfo.getChildByName("plexes").getChildren()) {
                        if (el1.getChildContent("name") != null) {
                            diskTypes.addAll(plexes.get(el1.getChildContent("name")));
                        }
                    }
                }
            }

            FilerAggregateImpl aggr = new FilerAggregateImpl(filer, name, uuid, traditional, state, sizeUsed, percentageUsed,
                    sizeTotal, sizeAvailable, volumeNames, diskCount);
            aggr.setBlockType((blockType != null) ? blockType : "32_bit");
            aggr.setDiskTypes(diskTypes);

            totalAllocated += aggr.getSizeTotal().toBytes();
            totalUsed += aggr.getSizeUsed().toBytes();

            aggrs.put(aggr.getName(), aggr);
        }

        filer.getSpaceUsage().setAllocated(Size.createB(totalAllocated));
        filer.getSpaceUsage().setUsed(Size.createB(totalUsed));

        LOGGER.debug("populateAggregate: done");
        filer.setAggregates(aggrs);

        return true;
    }

    /**
     * A map of plex name (raid-name) to DiskTypes
     *
     * @throws com.netapp.offtap3.ontap.ApiFailedException
     */
    Map<String, Set<String>> getPlexes(final FilerImpl<? extends Filer> controller) throws ApiFailedException {
        LOGGER.debug("Enter getPlexes");
        Map<String, Set<String>> plexes = new SafeConcurrentHashMap<String, Set<String>>();
        Set<Disk> disks = getDisks(controller);

        NaElement xi;
        NaElement xo;

        xi = new NaElement("raid-info-listplex");

        xo = controller.invokeZapi(xi, ZapiCache.LONG_CACHE);
        for (NaElement aggrInfo : xo.getChildByName("plexes").getChildren()) {
            Set<String> diskTypesInPlex = new HashSet<String>();
            String plexName = aggrInfo.getChildContent("raid-name");
            long plexId = aggrInfo.getChildLongValue("id", 0L);
            for (Disk disk : disks) {
                if (disk.getPlexId() == plexId && disk.getType() != null) {
                    diskTypesInPlex.add(disk.getType());
                }
            }
            plexes.put(plexName, diskTypesInPlex);
        }
        LOGGER.debug("Leave getPlexes plexes.size=" + plexes.size());
        return plexes;
    }

    /**
     * A map of disk name (name || raid-name)plex's id (plex-id) to
     *
     * @throws com.netapp.offtap3.ontap.ApiFailedException
     */
    Set<Disk> getDisks(final FilerImpl<? extends Filer>controller) throws ApiFailedException {
        LOGGER.debug("Enter getDisks");
        Set<Disk> disks = new HashSet<Disk>();
        Map<String, String> diskTypes = getDisksType(controller);

        NaElement xi;
        NaElement xo;

        xi = new NaElement("raid-info-listdisk");

        xo = controller.invokeZapi(xi, ZapiCache.LONG_CACHE);
        for (NaElement raidInfo : xo.getChildByName("disks").getChildren()) {
            Disk disk = new Disk();
            disk.setName(raidInfo.getChildContent("name"));
            if (disk.getName() == null) {
                //name might not be there, so we'll use the raid scoped name
                disk.setName(raidInfo.getChildContent("raid-name"));
            }
            disk.setType(diskTypes.get(disk.getName()));
            disk.setPlexId(raidInfo.getChildLongValue("plex-id", 0L));
            disks.add(disk);
        }
        LOGGER.debug("Leave getDisks disks.size=" + disks.size());
        return disks;
    }

    /**
     * A map of disk name to disk type
     *
     * @throws com.netapp.offtap3.ontap.ApiFailedException
     */
    Map<String, String> getDisksType(final FilerImpl controller) throws ApiFailedException {
        LOGGER.debug("Enter getDisksType");
        Map<String, String> diskTypes = new SafeConcurrentHashMap<String, String>();

        NaElement xi;
        NaElement xo;

        xi = new NaElement("disk-list-info");

        xo = controller.invokeZapi(xi, ZapiCache.LONG_CACHE);
        for (NaElement aggrInfo : xo.getChildByName("disk-details").getChildren()) {
            diskTypes.put(aggrInfo.getChildContent("name"), aggrInfo.getChildContent("disk-type"));
        }
        LOGGER.debug("Leave getDisks diskTypes.size=" + diskTypes.size());
        return diskTypes;
    }
}
