package com.netapp.offtap3.internal.ontap.dataloader;

import com.netapp.offtap3.ControllerResource;
import com.netapp.offtap3.FcpAdapter;
import com.netapp.offtap3.Refreshable;
import com.netapp.offtap3.Resource;
import com.netapp.offtap3.StorageProtocol;
import com.netapp.offtap3.factories.DataLoaderException;
import com.netapp.offtap3.internal.AbstractDataLoader;
import com.netapp.offtap3.internal.FcpAdapterImpl;
import com.netapp.offtap3.internal.ontap.controller.FilerImpl;
import com.netapp.offtap3.internal.ontap.support.ZapiCache;
import com.netapp.offtap3.ontap.ApiFailedException;
import com.netapp.offtap3.ontap.NaElement;
import com.netapp.offtap3.utils.SafeConcurrentHashMap;

import org.apache.log4j.Logger;

import java.util.ArrayList;
import java.util.EnumSet;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

public class FilerFcpDataLoader extends AbstractDataLoader<FilerImpl> implements HasResourceDependencies {

    private static final Logger logger = Logger.getLogger(FilerFcpDataLoader.class);

    @Override
    protected boolean doLoad(FilerImpl filer, LoadType loadType, Set<Resource> resources) throws DataLoaderException {

        logger.debug(filer + " start");

        if (!filer.supportsProtocol(StorageProtocol.FCP))   {
            logger.warn(filer + " not populating FC adapters because FCP is disabled");
            return true;
        }

        try {
            Map<String,Set<String>> initiators = getFcpInitiators(filer);
            NaElement xo = filer.invokeZapi(new NaElement("fcp-adapter-list-info"), ZapiCache.LONG_CACHE);

            if (xo != null){
                List<NaElement> adaptersInfo = xo.getChildByName("fcp-config-adapters").getChildren();
                List<FcpAdapter> adapterList = new ArrayList<FcpAdapter>();

                for (NaElement element : adaptersInfo) {
                    FcpAdapterImpl fcpAdapter = new FcpAdapterImpl(filer);
                    fcpAdapter.setName(element.getChildContent("adapter"));
                    fcpAdapter.setPort(element.getChildContent("port-name"));
                    fcpAdapter.setState(element.getChildContent("state"));
                    if (initiators.containsKey(fcpAdapter.getName())) {
                        fcpAdapter.setInitiators(initiators.get(fcpAdapter.getName()));
                    }
                    adapterList.add(fcpAdapter);
                    logger.debug(filer + " added " + fcpAdapter);
                }
                filer.setFcpAdapters(adapterList);
            } else {
                logger.error(filer + " fcp-adapter-list-info returned null.");
            }

        }
        catch (ApiFailedException e) {
            logger.error(filer + " failed to load FCP information", e);
            throw new DataLoaderException(filer + " failed to load FCP information", e);
        }
        return true;
    }

    Map<String, Set<String>> getFcpInitiators(FilerImpl filer)
            throws ApiFailedException {
        logger.debug(filer + " getFcpInitiators: start ");
        Map<String, Set<String>> returnMe = new SafeConcurrentHashMap<String, Set<String>>();

        NaElement xo = filer.invokeZapi(new NaElement("fcp-adapter-initiators-list-info"), ZapiCache.LONG_CACHE);

        if (xo != null) {
            if (xo.getChildByName("fcp-adapters").getChildContent("fcp-adapter-initiators-info") != null) {
                for (NaElement element : xo.getChildByName("fcp-adapters").getChildren()) {
                    String adapter = element.getChildContent("adapter");
                    Set<String> portNames = new HashSet<String>();
                    if (element.getChildByName("fcp-connected-initiators").getChildContent(
                            "fcp-connected-initiator-info") != null) {
                        for (NaElement el1 : element.getChildByName("fcp-connected-initiators").getChildren()) {
                            if (el1.getChildContent("port-name") != null) {
                                portNames.add(el1.getChildContent("port-name"));
                            }
                        }
                    }
                    returnMe.put(adapter, portNames);
                }
            }
        } else {
            logger.error(filer + " fcp-adapter-initiators-list-info returned null.");
        }
        logger.debug(filer + " getFcpInitiators returning " + returnMe.size());
        return returnMe;
    }

    @Override
    public Set<ControllerResource> getResources(Refreshable refreshable) {
        return EnumSet.of(ControllerResource.protocols);
    }
}
