package com.netapp.offtap3.internal.ontap.dataloader;

import com.netapp.offtap3.ControllerResource;
import com.netapp.offtap3.IpInterface;
import com.netapp.offtap3.Refreshable;
import com.netapp.offtap3.Resource;
import com.netapp.offtap3.factories.DataLoaderException;
import com.netapp.offtap3.internal.AbstractDataLoader;
import com.netapp.offtap3.internal.ontap.intf.IpInterfaceImpl;
import com.netapp.offtap3.internal.ontap.controller.PFilerImpl;
import com.netapp.offtap3.internal.ontap.support.ZapiCache;
import com.netapp.offtap3.utils.IPAddressUtil;
import com.netapp.offtap3.ontap.ApiFailedException;
import com.netapp.offtap3.ontap.NaElement;
import org.apache.log4j.Logger;

import java.util.Collections;
import java.util.EnumSet;
import java.util.Set;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

public class FilerInterfaceDataLoader extends AbstractDataLoader<PFilerImpl>
        implements HasResourceDependencies {

    private static final Logger LOGGER = Logger.getLogger(FilerInterfaceDataLoader.class);

    private static final Pattern INTERFACE_PAT = Pattern.compile("^(.*): flags=");

    private static final Pattern INET_PAT = Pattern.compile(
            "inet\\s([0-9]+)\\.([0-9]+)\\.([0-9]+)\\.([0-9]+)\\snetmask\\s0x(\\S{2})(\\S{2})(\\S{2})(\\S{2})\\sbroadcast\\s(\\d\\d?\\d?\\.\\d\\d?\\d?\\.\\d\\d?\\d?\\.\\d\\d?\\d?)",
            Pattern.MULTILINE);
    private static final Pattern INET_PAT_2 = Pattern.compile(
            "inet\\s([0-9]+)\\.([0-9]+)\\.([0-9]+)\\.([0-9]+)\\snetmask-or-prefix\\s0x(\\S{2})(\\S{2})(\\S{2})(\\S{2})\\sbroadcast\\s(\\d\\d?\\d?\\.\\d\\d?\\d?\\.\\d\\d?\\d?\\.\\d\\d?\\d?)",
            Pattern.MULTILINE);
    private static final Pattern INET_6_PAT = Pattern.compile(
            "inet6\\s(.+)\\sprefixlen\\s(\\d+).?",
            Pattern.MULTILINE);

    @Override
    protected Refreshable doRefresh(PFilerImpl loadTarget, Resource resource, Refreshable refreshTarget) {
        doLoad(loadTarget, LoadType.refresh, Collections.singleton(resource));
        return loadTarget.getIpInterfaces().get(((IpInterface) refreshTarget).getIpAddress());
    }

    @Override
    protected boolean doLoad(PFilerImpl filer, LoadType loadType, Set<Resource> resources) {

        NaElement xi = new NaElement("system-cli");
        xi.addNewChild("priv", "admin");
        NaElement xa = new NaElement("args");
        xa.addNewChild("arg", "ifconfig");
        xa.addNewChild("arg", "-a");
        xi.addChildElem(xa);

        try {
            NaElement xo = filer.invokeZapi(xi, ZapiCache.NO_CACHE);
            int rc = xo.getChildIntValue("cli-result-value", 0);
            String output = xo.getChildContent("cli-output");
            if (rc != 1) {
                LOGGER.error(filer + String.format(" execution of ifconfig failed. rc = %s.  output = %s", rc, output));
            }
            int[] ip = new int[4];
            int[] mask = new int[4];
            Integer[] net = new Integer[4];

            Matcher devLine, m;
            String dev = null;

            for (String line : output.split("\\n")) {
                devLine = INTERFACE_PAT.matcher(line);

                if (devLine.find()) {
                    dev = devLine.group(1);
                    LOGGER.debug(filer + " found device - " + dev);
                } else {
                    m = INET_PAT.matcher(line);
                    boolean found = m.find();
                    if (!found) {
                        m = INET_PAT_2.matcher(line);
                        found = m.find();
                    }

                    // check IPv6
                    if (!found) {
                        m = INET_6_PAT.matcher(line);
                        found = m.find();
                    }

                    if (found) {
                        // skip all interfaces configured on device e0M, lo, losk, e0p, c0a, and c0b
                        if("e0m".equalsIgnoreCase(dev) || "lo".equalsIgnoreCase(dev) || "losk".equals(dev) ||
                                "e0p".equalsIgnoreCase(dev) || "c0a".equalsIgnoreCase(dev) || "c0b".equals(dev)) {
                            continue;
                        }

                        if ((m.groupCount() == 9)
                                && !(m.group(1).equals("127") && m.group(2).equals("0")
                                && m.group(3).equals("0") && m.group(4).equals("1"))) { // this is an IPv4 Address (not the loopback address).
                            for (int i = 0; i < 4; i++) {
                                ip[i] = Integer.parseInt(m.group(i + 1));
                                mask[i] = Integer.decode("0x" + m.group(i + 5));
                                net[i] = ip[i] & mask[i];
                            }

                            String ipv4 = m.group(1) + "." + m.group(2) + "." + m.group(3) + "." + m.group(4);
                            String netmask = String.format("%d.%d.%d.%d", mask[0], mask[1], mask[2], mask[3]);
                            String network = String.format("%d.%d.%d.%d", net[0], net[1], net[2], net[3]);

                            checkAndAddInterface(filer, ipv4, dev, network, netmask);

                        } else if(m.groupCount() == 2) { // this is an IPv6 Address.
                            String ipAddress = m.group(1);
                            String netmask = m.group(2);
                            String network = IPAddressUtil.getNetwork(ipAddress, netmask);

                            checkAndAddInterface(filer, ipAddress, dev, network, netmask);
                        }
                    }
                }
            }
        }
        catch (ApiFailedException e) {
            LOGGER.error(filer + " failed to execute ifconfig cli", e);
            throw new DataLoaderException(filer + " failed to execute ifconfig cli", e);
        }

        return true;
    }

    private void checkAndAddInterface(PFilerImpl filer, String ip, String dev, String network, String netmask) {
        if (!filer.hasVFilerInterface(ip)) {
            IpInterfaceImpl netInterface = new IpInterfaceImpl(ip, filer);
            netInterface.setNetmask(netmask);
            netInterface.setNetwork(network);
            netInterface.setName(dev);
            if (filer.getIscsiPortalNames().contains(dev)) {
                netInterface.getProtocols().add("iscsi");
            }
            filer.addInterface(netInterface);
            LOGGER.debug(filer + " added ip address on " + dev + " of ip " + ip);
        } else {
            LOGGER.debug(filer + " network device skipped (vfiler interface) - " + dev + ", ip " + ip);
        }
    }

    @Override
    public Set<ControllerResource> getResources(Refreshable refreshable) {
        // we need the vfilers in order to filter the vfiler interfaces out of the pfiler interfaces list
        // we need the protocols to populate the iSCSI portal names and add iSCSI protocol to the interface
        return EnumSet.of(ControllerResource.vfilers, ControllerResource.protocols);
    }
}
