package com.netapp.offtap3.internal.ontap.dataloader;


import com.netapp.offtap3.ControllerResource;
import com.netapp.offtap3.IscsiAdapter;
import com.netapp.offtap3.IscsiConnection;
import com.netapp.offtap3.Refreshable;
import com.netapp.offtap3.Resource;
import com.netapp.offtap3.StorageProtocol;
import com.netapp.offtap3.factories.DataLoaderException;
import com.netapp.offtap3.internal.AbstractDataLoader;
import com.netapp.offtap3.internal.IscsiAdapterImpl;
import com.netapp.offtap3.internal.ontap.controller.FilerImpl;
import com.netapp.offtap3.internal.ontap.intf.IscsiConnectionImpl;
import com.netapp.offtap3.internal.ontap.support.ZapiCache;
import com.netapp.offtap3.ontap.ApiFailedException;
import com.netapp.offtap3.ontap.Filer;
import com.netapp.offtap3.ontap.NaElement;
import com.netapp.offtap3.utils.SafeConcurrentHashMap;

import org.apache.log4j.Logger;

import java.util.ArrayList;
import java.util.Collections;
import java.util.EnumSet;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

public class FilerIscsiDataLoader extends AbstractDataLoader<FilerImpl<? extends Filer>> implements
        HasResourceDependencies {

    private final static Logger logger = Logger.getLogger(FilerIscsiDataLoader.class);

    @Override
    protected boolean doLoad(FilerImpl<? extends Filer> filer, LoadType loadType, Set<Resource> resources) {
        logger.debug(filer + " begin");
        if (!filer.supportsProtocol(StorageProtocol.ISCSI)) {
            logger.debug(filer + " not discovering iSCSI information because iSCSI is disabled");
            return true;
        }

        populateIscsiAdapters(filer);
        populateIscsiConnections(filer);

        return true;
    }

    @Override
    protected Refreshable doRefresh(FilerImpl<? extends Filer> loadTarget, Resource resource, Refreshable refreshTarget) {
        doLoad(loadTarget, LoadType.refresh, Collections.singleton(resource));
        if (refreshTarget instanceof IscsiAdapter) {
            return getFromList(loadTarget.getIscsiAdapters(), refreshTarget.getName());
        } else if (refreshTarget instanceof IscsiConnection) {
            IscsiConnection reloadConnection = (IscsiConnection) refreshTarget;
            for (IscsiConnection connection : loadTarget.getIscsiConnections()) {
                if (connection.getConnectionId() == reloadConnection.getConnectionId()) {
                    return connection;
                }
            }
        }
        return null;
    }

    protected void populateIscsiAdapters(FilerImpl filer) throws DataLoaderException {
        try {
            Map<String, Set<String>> initiators = getIscsiInitiators(filer);

            NaElement xo = filer.invokeZapi(new NaElement("iscsi-adapter-list-info"), ZapiCache.LONG_CACHE);
            if (xo != null) {
                List<NaElement> adaptersInfo = xo.getChildByName("iscsi-config-adapters").getChildren();
                List<IscsiAdapter> adapterList = new ArrayList<IscsiAdapter>();

                for (NaElement element : adaptersInfo) {
                    IscsiAdapterImpl iscsiAdapter = new IscsiAdapterImpl(filer);
                    iscsiAdapter.setName(element.getChildContent("name"));
                    iscsiAdapter.setState(element.getChildContent("state"));
                    if (initiators.containsKey(iscsiAdapter.getName())) {
                        iscsiAdapter.setInitiators(initiators.get(iscsiAdapter.getName()));
                    }
                    adapterList.add(iscsiAdapter);
                    logger.debug(filer + " added " + iscsiAdapter);
                }
                filer.setIscsiAdapters(adapterList);
            } else {
                logger.error(filer + " iscsi-adapter-list-info return null");
            }

        } catch (ApiFailedException e) {
            logger.error(filer + " failed to load iscsi adapters", e);
            throw new DataLoaderException(filer + " failed to load iscsi adapters", e);
        }
    }

    protected void populateIscsiConnections(final FilerImpl filer) {
        NaElement xo = null;
        logger.debug(filer + " populateISCSIConnections: start");

        List<IscsiConnection> connections = new ArrayList<IscsiConnection>();

        try {
            xo = filer.invokeZapi(new NaElement("iscsi-connection-list-info"), ZapiCache.LONG_CACHE);
        } catch (ApiFailedException e) {
            logger.info("populateISCSIConnections: Unable to retrieve iscsi connection information for " + filer);
            logger.debug("Exception while retrieving iscsi connection list info: " + e.getMessage());
        }
        if (xo != null && xo.getChildByName("iscsi-connection-list-entries") != null) {
            for (NaElement connectInfo : xo.getChildByName("iscsi-connection-list-entries").getChildren()) {
                IscsiConnectionImpl connection = new IscsiConnectionImpl(filer);
                connection.setConnectionId(connectInfo.getChildLongValue("connection-id", 0L));
                connection.setConnectionState(connectInfo.getChildContent("connection-state"));
                connection.setHasSession(Boolean.parseBoolean(connectInfo.getChildContent("has-session")));
                connection.setLocalInterface(connectInfo.getChildContent("interface-name"));
                connection.setLocalIp(connectInfo.getChildContent("local-ip-address"));
                connection.setLocalPort(connectInfo.getChildIntValue("local-ip-port", 0));
                connection.setRemoteIpAddress(connectInfo.getChildContent("remote-ip-address"));
                connection.setRemotePort(connectInfo.getChildIntValue("remote-ip-port", 0));
                connection.setSessionId(connectInfo.getChildLongValue("session-id", 0L));
                connections.add(connection);
            }
        }
        filer.setIscsiConnections(connections);
        logger.debug(filer + " populateIscsiConnections: done and found " + connections.size() + " connections");
    }

    protected Map<String, Set<String>> getIscsiInitiators(FilerImpl filer) throws ApiFailedException {
        logger.debug(filer + "  getIscsiInitiators: start ");
        Map<String, Set<String>> initiatorsMap = new SafeConcurrentHashMap<String, Set<String>>();

        NaElement xo = filer.invokeZapi(new NaElement("iscsi-adapter-initiators-list-info"), ZapiCache.LONG_CACHE);

        if (xo != null) {
            if (xo.getChildByName("iscsi-adapters").getChildContent("iscsi-adapter-initiators-info") != null) {
                for (NaElement element : xo.getChildByName("iscsi-adapters").getChildren()) {
                    String adapter = element.getChildContent("name");
                    Set<String> portNames = new HashSet<String>();
                    if (element.getChildByName("iscsi-connected-initiators")
                               .getChildContent("iscsi-connected-initiator-info") != null) {
                        for (NaElement el1 : element.getChildByName("iscsi-connected-initiators").getChildren()) {
                            if (el1.getChildContent("initiator-name") != null) {
                                portNames.add(el1.getChildContent("initiator-name"));
                            }
                        }
                    }
                    initiatorsMap.put(adapter, portNames);
                }
            }
        } else {
            logger.error(filer + " iscsi-adapter-initiators-list-info returned null.");
        }
        logger.debug(filer + " getIscsiInitiators: returning " + initiatorsMap.size() + " iscsi initiators");
        return initiatorsMap;
    }

    @Override
    public Set<ControllerResource> getResources(Refreshable refreshable) {
        // need this to avoid calling the iscsi-adapter-list-info ZAPI when there is no ISCSI support
        return EnumSet.of(ControllerResource.protocols);
    }
}
