package com.netapp.offtap3.internal.ontap.dataloader;

import com.netapp.offtap3.Resource;
import com.netapp.offtap3.StorageProtocol;
import com.netapp.offtap3.factories.DataLoaderException;
import com.netapp.offtap3.internal.AbstractDataLoader;
import com.netapp.offtap3.internal.ontap.controller.FilerImpl;
import com.netapp.offtap3.internal.ontap.support.ZapiCache;
import com.netapp.offtap3.ontap.ApiFailedException;
import com.netapp.offtap3.ontap.NaElement;
import org.apache.log4j.Logger;

import java.util.HashSet;
import java.util.List;
import java.util.Set;

public class FilerProtocolDataLoader extends AbstractDataLoader<FilerImpl> {
    private static Logger logger = Logger.getLogger(FilerProtocolDataLoader.class);


    @Override
    protected boolean doLoad(FilerImpl filer, LoadType loadType, Set<Resource> resources) throws DataLoaderException {
        NaElement xo;

        if (filer.isLicensed("iscsi")) {
            try {
                xo = filer.invokeZapi(new NaElement("iscsi-service-status"), ZapiCache.LONG_CACHE);
                if (Boolean.parseBoolean(xo.getChildContent("is-available"))){
                    filer.addSupportedProtocol(StorageProtocol.ISCSI);
                    logger.info(filer +" iscsi is licensed and enabled.");
                    // since we know iSCSI is working, this is a good time to load up the iscsi portal names
                    filer.setIscsiPortalNames(loadIsciPortalNames(filer));
                } else {
                    logger.warn(filer +" iscsi is licensed BUT NOT enabled.");
                }
            } catch(ApiFailedException ex) {
                if (ex.getErrno() == 13008) {
                    // don't fill up the log with stack traces just because it's disabled
                    // (some ONTAP versions won't even respond to the above ZAPI to allow us to determine status.)
                    logger.warn("iSCSI service status unknown for " + filer +".  It appears to be licensed BUT NOT enabled.");
                }
                else {
                    logger.error(filer +" failed to get the iscsi service status.", ex);
                }
            }

        }

        if (filer.isLicensed("fcp")) {
            try {
                xo = filer.invokeZapi(new NaElement("fcp-service-status"), ZapiCache.LONG_CACHE);
                if (Boolean.parseBoolean(xo.getChildContent("is-available"))) {
                    filer.addSupportedProtocol(StorageProtocol.FCP);
                    logger.info(filer +" fcp is licensed and enabled.");
                } else {
                    logger.warn(filer +" fcp is licensed BUT NOT enabled.");
                }
            } catch (ApiFailedException e) {
                if (e.getErrno() == 13008 || e.getErrno() == 13005) {
                    // don't fill up the log with stack traces just because it's disabled
                    // (some ONTAP versions won't even respond to the above ZAPI to allow us to determine status.)
                    logger.warn("FCP service status unknown for " + filer +" (received error=" + e.getErrno() + ").  It appears to be licensed BUT NOT enabled.");
                }
                else {
                    logger.error(filer + " failed to get fcp service status.", e);
                }
            }

        }

        if (filer.isLicensed("nfs")) {
            try {
                xo = filer.invokeZapi(new NaElement("nfs-status"), ZapiCache.LONG_CACHE);
                if (Boolean.parseBoolean(xo.getChildContent("is-enabled"))){
                    filer.addSupportedProtocol(StorageProtocol.NFS);
                    logger.info(filer +" nfs is licensed and enabled.");
                }else{
                    logger.warn(filer +" nfs is licensed BUT NOT enabled.");
                }
            } catch (ApiFailedException e) {
                logger.error(filer +" failed to get nfs service status.", e);
            }
        }

        return true;
    }

    protected Set<String> loadIsciPortalNames(FilerImpl filer) throws DataLoaderException {
        logger.debug(filer + " load iscsi portal names");

        Set<String> portalNames = new HashSet<String>();

        try {
            NaElement xo = filer.invokeZapi(new NaElement("iscsi-portal-list-info"), ZapiCache.SHORT_CACHE);
            List<NaElement> portals = xo.getChildByName("iscsi-portal-list-entries").getChildren();
            for (NaElement x : portals) {
                String intfName = x.getChildByName("interface-name").getContent();
                if (intfName != null) {
                    if (portalNames.add(intfName)) {
                        logger.debug(filer + " has an iscsi portal for " + intfName);
                    }
                }
            }
        } catch (ApiFailedException e) {
            logger.error(filer + " failed to get portal list.", e);
            throw new DataLoaderException(filer + " failed to get portal list. error: " +e.getMessage(), e);
        }
        return portalNames;
    }
}
