package com.netapp.offtap3.internal.ontap.dataloader;

import com.netapp.offtap3.Resource;
import com.netapp.offtap3.factories.DataLoaderException;
import com.netapp.offtap3.internal.AbstractDataLoader;
import com.netapp.offtap3.internal.ontap.controller.FilerImpl;
import com.netapp.offtap3.internal.ontap.support.ZapiCache;
import com.netapp.offtap3.ontap.ApiFailedException;
import com.netapp.offtap3.ontap.NaElement;
import com.netapp.offtap3.ontap.NaErrno;
import com.netapp.offtap3.ontap.ZapiExecutionError;
import com.netapp.offtap3.utils.Utils;

import org.apache.log4j.Logger;

import java.util.List;
import java.util.Set;

public class FilerSystemInfoDataLoader extends AbstractDataLoader<FilerImpl> {
    private static Logger logger = Logger.getLogger(FilerSystemInfoDataLoader.class);

    @Override
    protected boolean doLoad(FilerImpl filer, LoadType loadType, Set<Resource> resources) throws DataLoaderException {
        populateSystemInfo(filer);
        loadVStorageStatus(filer);

        return true;
    }

    private void populateSystemInfo(FilerImpl filer) throws DataLoaderException {
        if (filer.isPassThroughVFiler()) {
            logger.debug("Skip load for pass-through vfiler: " + filer + " since data found on parent");
            return;
        }

        NaElement xi = new NaElement("system-get-info");
        NaElement xo;

        try {
            xo = filer.invokeZapi(xi, ZapiCache.LONG_CACHE);
        } catch (ApiFailedException e) {
            throw new DataLoaderException("could not get system info for " + filer, e);
        }

        //get system info
        if (xo != null){
            String id = null;
            logger.trace("initializeFiler: " +xo);
            filer.setName(xo.getChildByName("system-info").getChildContent("system-name"));

            id = xo.getChildByName("system-info").getChildContent("system-id");
            if (id != null){
                filer.setSystemId(id);
                filer.setId(id,"0");
            }

            String model = xo.getChildByName("system-info").getChildContent("system-model");
            if (model != null){
                filer.setModel(model);
            }
        }
    }

    private void loadVStorageStatus(FilerImpl filer) {
        // get vaai support
        NaElement xi = new NaElement("copyoffload-show");
        try {
            NaElement xo = filer.invokeZapi(xi, ZapiCache.LONG_CACHE);
            filer.setVstorageStatus(xo.getChildContent("copyoffload-state"));
            logger.info("initializeFiler: " + filer + " Controller supports the VAAI primitatives");
        } catch (ApiFailedException e) {
            /* If the ONTAP version does not support the "copyoffload" feature,
             * a NaFailedException exception will be thrown.
             *
             * ONTAP is delivering the copyoffload features in phases. The
             * initial phase will only support same controller copies, which
             * means the Preferred IP address is not needed/implemented.
             * The next phase will add support for inter-controller copyoffload.
             * The preferred IP address will be added at that time.  The is
             * mentioned because it effects how the 'Not Found' exception is
             * handled.  If the "copyoffload-state" fails, then the controller
             * flat out doesn't support the copyoffload feature. However, if
             * it is not empty, then it means it supports Phase 1, but not
             * Phase 2.
             */
            if (e.getErrno() == NaErrno.EAPINOTFOUND
                    || e.getErrno() == NaErrno.EAPIERROR) {
                logger.info("initializeFiler: "+filer +" The VAAI primitatives are not supported by controller. (" +e.getReason()+")");
                filer.setVstorageStatus("No");

                // the original code read as:
                // filer.setErrorMessage(null);
                // so, I think the intent was to remove the last error as it is not important.                
                List<ZapiExecutionError> zapiErrors = filer.getZapiExecutionErrors();
                if (Utils.isReal(zapiErrors)) {
                	zapiErrors.remove(zapiErrors.size() - 1);
                } 
                
            }else{
                filer.setVstorageStatus("Unknown");
            }
        }
    }
}
