package com.netapp.offtap3.internal.ontap.dataloader;

import java.util.List;
import java.util.Map;

import org.apache.log4j.Logger;

import com.netapp.offtap3.ConnectionException;
import com.netapp.offtap3.Controller;
import com.netapp.offtap3.ControllerConnection;
import com.netapp.offtap3.Version;
import com.netapp.offtap3.factories.DataLoaderException;
import com.netapp.offtap3.factories.platform.PlatformDataLoader;
import com.netapp.offtap3.internal.ontap.OntapVersionImpl;
import com.netapp.offtap3.internal.ontap.aggregate.CModeAggregateImpl;
import com.netapp.offtap3.internal.ontap.controller.AdminSvmImpl;
import com.netapp.offtap3.internal.ontap.controller.DataSvmImpl;
import com.netapp.offtap3.internal.ontap.controller.OntapConnection;
import com.netapp.offtap3.internal.ontap.controller.PFilerImpl;
import com.netapp.offtap3.internal.ontap.controller.VFilerImpl;
import com.netapp.offtap3.internal.ontap.support.ZapiCache;
import com.netapp.offtap3.ontap.ApiFailedException;
import com.netapp.offtap3.ontap.CModeAggregate;
import com.netapp.offtap3.ontap.NaElement;
import com.netapp.offtap3.ontap.NaErrno;
import com.netapp.offtap3.utils.SafeConcurrentHashMap;

public class OntapDataLoader implements PlatformDataLoader {
    private static final Logger logger = Logger.getLogger(OntapDataLoader.class);

    private final SevenModeVersionBuilder sevenModeVersionBuilder = new SevenModeVersionBuilder();
    private final CModeVersionBuilder cModeVersionBuilder = new CModeVersionBuilder();

    private OntapConnection connection;
    private boolean isClustered;
    private OntapVersionImpl version;
    private boolean initialized;

    private boolean initialize(ControllerConnection controllerConnection, boolean throwOnFail) throws ConnectionException {
        if (initialized) {
            return true;
        }
        // Try to execute the 'system-get-version' ZAPI to check for 'isClustered'
        connection = new OntapConnection(controllerConnection);
        connection.setConnectionTimeout(controllerConnection.getOfftapPreferences().getZapiConnectionTimeout());
        connection.setTimeout(controllerConnection.getOfftapPreferences().getZapiTimeout());

        NaElement xi = new NaElement("system-get-version");
        NaElement xo;

        try {
            xo = connection.invokeZapi(xi, ZapiCache.LONG_CACHE);
        } catch (ApiFailedException e) {
            // TODO aeonflux-next - it would be great to differentiate
            // the case of "talking to ONTAP, but had a problem" (throw an exception)
            // from "not talking to ONTAP" (return false)
            // so far, we're just reporting the authorization failure separately.
            logger.trace("Failed to connect to : " + connection, e);
            if (throwOnFail || (e.getErrno() == NaErrno.EONTAPI_ENETUNREACH && !e.getMessage().contains("Unexpected HTTP response"))) {
                // Unexpected http response indicates we're probably not talking to ONTAP.
                throw new ConnectionException("Unable to communicate with ONTAP at " + connection, e);
            } else if (e.getErrno() == NaErrno.EAPIAUTHENTICATION) {
                throw new ConnectionException("Error: ONTAP authentication failed.", e);
            }
            return false;
        }

        isClustered = Boolean.parseBoolean(xo.getChildContent("is-clustered"));

        try {
            if (isClustered) {
                logger.info("Connected to clustered Data ONTAP at: " + connection);
                version = cModeVersionBuilder.build(xo);
            } else {
                logger.info("Connected to 7-mode Data ONTAP at: " + connection);
                version = sevenModeVersionBuilder.build(xo, connection);
            }
        } catch (DataLoaderException e) {
            throw new ConnectionException("Unable to retieve version info for ONTAP controller at + " + connection, e);
        }
        logger.info("Version is: " + version);
        initialized = true;
        return true;
    }

    @Override
    public boolean canHandle(ControllerConnection controllerConnection) throws ConnectionException {
        return initialize(controllerConnection, false);
    }

    @Override
    public Version getVersion(ControllerConnection controllerConnection) throws ConnectionException {
        initialize(controllerConnection, true);
        return version;
    }


    @Override
    public Controller load(ControllerConnection controllerConnection) throws DataLoaderException, ConnectionException {
        initialize(controllerConnection, true);
        Controller controller = null;
        if (!isClustered) {
            if (version.getFullVersion().endsWith("Multistore")) {
                // connecting directly to a vfiler
                VFilerImpl filer = new VFilerImpl(connection, version);
                controller = filer.getProxy();
            } else {
                // connecting to the filer
                PFilerImpl filer = new PFilerImpl(connection, version);
                controller = filer.getProxy();
            }
        } else {
            NaElement xi = new NaElement("vserver-get-iter");

            NaElement query = new NaElement("query");
            NaElement queryVServerInfo = new NaElement("vserver-info");
            NaElement queryVServerType = new NaElement("vserver-type");
            queryVServerType.setContent("admin,data,cluster");

            queryVServerInfo.addChildElem(queryVServerType);
            query.addChildElem(queryVServerInfo);
            xi.addChildElem(query);

            NaElement desiredAttributes = new NaElement("desired-attributes");
            NaElement desiredVServerInfo = new NaElement("vserver-info");
            desiredVServerInfo.addChildElem(new NaElement("vserver-type"));
            desiredVServerInfo.addChildElem(new NaElement("vserver-name"));
            desiredVServerInfo.addChildElem(new NaElement("uuid"));
            
            NaElement aggrInfoList = new NaElement("vserver-aggr-info-list");
            desiredVServerInfo.addChildElem(aggrInfoList);
            aggrInfoList.addChildElem(new NaElement("vserver-aggr-info"));
            desiredVServerInfo.addChildElem(new NaElement("vserver-subtype"));
            desiredAttributes.addChildElem(desiredVServerInfo);
            
            xi.addChildElem(desiredAttributes);

            CdotIterativeLoader<NaElement> vserverLoader = new CdotIterativeLoader<NaElement>(connection, xi,
                                                                                              new CumulativeDataParser(),
                                                                                              connection
                                                                                                      .getOfftapPreferences()
                                                                                                      .getIteratorSize(),
                                                                                              ZapiCache.LONG_CACHE);
            NaElement results = new NaElement("results");
            try {
                vserverLoader.load(results);

                if (results.getChildren().size() == 0) {
                    throw new RuntimeException("Error: Unable to load Storage Virtual Machine information.");
                }

                boolean adminIsSyncSource = false;
                AdminSvmImpl adminSvm = null;
                for (NaElement child : results.getChildren()) {
                    String type = child.getChildContent("vserver-type");

                    if (results.getChildren().size() == 1 && ("data".equals(type) || "cluster".equals(type))) {
                        DataSvmImpl dataSvm = new DataSvmImpl(connection, version);
                        dataSvm.setId(child.getChildContent("uuid"));
                        dataSvm.setName(child.getChildContent("vserver-name"));

                        // check to see if this is a MCC destination vserver.  If so, we should hide it and just fail over to it
                        // if we detect problems.
                        if ("sync_source".equals(child.getChildContent("vserver-subtype"))) {
                            dataSvm.setMccSource(true);
                        }
                        if ("sync_destination".equals(child.getChildContent("vserver-subtype"))) {
                            // normally my inclination would be to skip this.  But there is a single case where we actually want to keep it
                            // and use it.  That's the case of a direct connect vserver, which has been switched over.  In that case, we want to take
                            // the -mc off of the name, and we want to say its an mcc source.  That way when things get switched back, it will
                            // be in business and rolling.
                            logger.debug(
                                    "See direct connect vserver of sync_destination subtype.  This is a switched over MCC vserver and we are gonna keep it");
                            dataSvm.setMccSource(true);
                            if (dataSvm.getName().endsWith("-mc")) {
                                dataSvm.setName(dataSvm.getName().substring(0, dataSvm.getName().length() - 3));
                            }
                        }

                        Map<String, CModeAggregate> aggrs = new SafeConcurrentHashMap<String, CModeAggregate>();
                        NaElement aggrInfoListInner = child.getChildByName("vserver-aggr-info-list");
                        if (aggrInfoListInner != null) {
                            for (NaElement aggrInfo : aggrInfoListInner.getChildren()) {
                                String name = aggrInfo.getChildContent("aggr-name");
                                long sizeAvailable = aggrInfo.getChildLongValue("aggr-availsize", 0L);
                                CModeAggregateImpl aggr = new CModeAggregateImpl(dataSvm, name, sizeAvailable);
                                aggrs.put(name, aggr);
                            }
                        }
                        dataSvm.setAggregateList(aggrs);

                        controller = dataSvm.getProxy();
                        break;
                    } else if ("admin".equals(type)) {
                        adminSvm = new AdminSvmImpl(connection, version);
                        adminSvm.setClusterId(getClusterId(adminSvm));
                        adminSvm.setId(adminSvm.getClusterId());
                        adminSvm.setName(child.getChildContent("vserver-name"));
                        adminSvm.setMccSource(findMccSource(results.getChildren()));
                    	controller = adminSvm.getProxy();
                        break;
                    }
                }
            } catch (ApiFailedException e) {
                throw new DataLoaderException("Failed to obtain vserver info", e);
            }

            if (controller == null) {
                throw new RuntimeException("Unable to find admin or data SVM with this connection.");
            }

        }

        return controller;
    }

    private boolean findMccSource(List<NaElement> svms) {
        for (NaElement svm : svms) {
            if ("sync_source".equals(svm.getChildContent("vserver-subtype"))) {
                // At least one of the data svms is a sync source. Now we can mark the admin svm as a sync source.
                return true;
            }
        }
        return false;
    }

    private String getClusterId(AdminSvmImpl adminSvm) throws ApiFailedException {
        NaElement result = adminSvm.invokeZapi(new NaElement("cluster-identity-get"), ZapiCache.LONG_CACHE);
        NaElement attributes = result.getChildByName("attributes");
        NaElement info = attributes.getChildByName("cluster-identity-info");

        return info.getChildContent("cluster-uuid");
    }
}
