package com.netapp.offtap3.internal.ontap.dataloader;

import java.util.ArrayList;
import java.util.Collections;
import java.util.Date;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.apache.log4j.Logger;

import com.netapp.offtap3.FcInterface;
import com.netapp.offtap3.IpInterface;
import com.netapp.offtap3.OfftapPreferences;
import com.netapp.offtap3.Refreshable;
import com.netapp.offtap3.Resource;
import com.netapp.offtap3.factories.DataLoaderException;
import com.netapp.offtap3.internal.AbstractDataLoader;
import com.netapp.offtap3.internal.ontap.controller.AbstractOntapController;
import com.netapp.offtap3.internal.ontap.intf.CModeInterfaceImpl;
import com.netapp.offtap3.internal.ontap.intf.CModeIpInterfaceImpl;
import com.netapp.offtap3.internal.ontap.intf.FcInterfaceImpl;
import com.netapp.offtap3.internal.ontap.support.ZapiCache;
import com.netapp.offtap3.ontap.ApiFailedException;
import com.netapp.offtap3.ontap.CModeIpInterface;
import com.netapp.offtap3.ontap.NaElement;
import com.netapp.offtap3.ontap.OntapController;
import com.netapp.offtap3.utils.IPAddressUtil;
import com.netapp.offtap3.utils.SafeConcurrentHashMap;
import com.netapp.offtap3.utils.Utils;

public class SvmInterfaceDataLoader extends AbstractDataLoader<AbstractOntapController<? extends OntapController>> {
    private static final Logger LOGGER = Logger.getLogger(SvmInterfaceDataLoader.class);

    private OfftapPreferences prefs;

    public SvmInterfaceDataLoader(OfftapPreferences prefs) {
        this.prefs = prefs;
    }

    @Override
    protected Refreshable doRefresh(AbstractOntapController<? extends OntapController> loadTarget, Resource resource, Refreshable refreshTarget) {
        doLoad(loadTarget, LoadType.refresh, Collections.singleton(resource));
        if (refreshTarget instanceof FcInterface) {
            for (FcInterface fcInterface : loadTarget.getFcInterfaces()) {
                if (fcInterface.getWwpn().equals(((FcInterface) refreshTarget).getWwpn())) {
                    return fcInterface;
                }
            }
        } else if (refreshTarget instanceof IpInterface) {
            return loadTarget.getIpInterfaces().get(((IpInterface) refreshTarget).getIpAddress());
        }
        return null;
    }

    @Override
    protected boolean doLoad(AbstractOntapController<? extends OntapController> controller, LoadType loadType, Set<Resource> resources) throws DataLoaderException {
        LOGGER.debug("populateInterfaces: start");

        CdotIterativeLoader<NaElement> intfLoader = new CdotIterativeLoader<NaElement>(controller, new NaElement(
                "net-interface-get-iter"), new CumulativeDataParser(), prefs.getIteratorSize(), ZapiCache.SHORT_CACHE);

        Map<String, CModeIpInterface> ipInterfaces = new SafeConcurrentHashMap<String, CModeIpInterface>();
        List<FcInterface> fcInterfaces = new ArrayList<FcInterface>();

        NaElement results = new NaElement("results");

        try {
            intfLoader.load(results);
        } catch (ApiFailedException e) {
            throw new DataLoaderException(controller + " failed to load interfaces", e);
        }

        for (NaElement intf : results.getChildren()) {

            List<String> protocols = new ArrayList<String>();
            if (intf.getChildByName("data-protocols") != null) {
                List<NaElement> dataProtocols = intf.getChildByName("data-protocols").getChildren();
                for (NaElement naDataProtocol : dataProtocols) {
                    protocols.add(naDataProtocol.getContent());
                }
            }

            CModeInterfaceImpl cmodeInterface;

            if (protocols.contains("fcp")) {
                FcInterfaceImpl fcIntf = new FcInterfaceImpl(controller);
                fcIntf.setWwpn(intf.getChildContent("wwpn"));
                fcIntf.setPortAddress(intf.getChildContent("port-address"));
                fcInterfaces.add(fcIntf);

                cmodeInterface = fcIntf;
            } else {
                CModeIpInterfaceImpl ipIntf = new CModeIpInterfaceImpl(controller);
                ipIntf.setIpAddress(intf.getChildContent("address"));
                String netmask = intf.getChildContent("netmask");
                String network = IPAddressUtil.getNetwork(ipIntf.getIpAddress(), netmask);

                ipIntf.setNetmask(netmask);
                ipIntf.setNetwork(network);
                ipIntf.setFirewallPolicy(intf.getChildContent("firewall-policy"));
                ipIntf.setAddressFamily(intf.getChildContent("address-family"));
                ipInterfaces.put(ipIntf.getIpAddress(), ipIntf);

                cmodeInterface = ipIntf;
            }

            cmodeInterface.setProtocols(protocols);
            setCommonInterfaceFields(intf, cmodeInterface);

            LOGGER.debug("Found an interface: " + cmodeInterface);
        }

        controller.setIpInterfaces(ipInterfaces);
        controller.setFcInterfaces(fcInterfaces);
        if (fcInterfaces.size() > 0) {
        	// ok - we get some very basic information about fc interfaces here, but really, we want to be calling
        	// fcp-interface-get-iter to get everything.  So lets call that and use that information instead.
        	populateRealFcpInterfaces(controller, fcInterfaces);
        }

        LOGGER.debug("populateInterfaces: end");

        return true;
    }

	private void setCommonInterfaceFields(NaElement intf, CModeInterfaceImpl cmodeInterface) {
		cmodeInterface.setName(intf.getChildContent("interface-name"));

		cmodeInterface.setCurrentPort(intf.getChildContent("current-port"));		
		cmodeInterface.setRole(intf.getChildContent("role"));
		cmodeInterface.setOwningVserver(intf.getChildContent("vserver"));

		cmodeInterface.setIsUp("up".equals(intf.getChildContent("operational-status")));
		cmodeInterface.setIsHome(Boolean.parseBoolean(intf.getChildContent("is-home")));
		cmodeInterface.setCurrentNode(intf.getChildContent("current-node"));
		cmodeInterface.setHomeNode(intf.getChildContent("home-node"));
		cmodeInterface.setHomePort(intf.getChildContent("home-port"));
		cmodeInterface.setLastLoad(new Date());
		cmodeInterface.setUuid(intf.getChildContent("lif-uuid"));
	}
    
    private void populateRealFcpInterfaces(AbstractOntapController<? extends OntapController> controller, List<FcInterface> previousFcInterfaces) {
    	CdotIterativeLoader<NaElement> intfLoader = new CdotIterativeLoader<NaElement>(controller, new NaElement(
                "fcp-interface-get-iter"), new CumulativeDataParser(), prefs.getIteratorSize(), ZapiCache.LONG_CACHE);

        Map<String, IpInterface> ipInterfaces = new SafeConcurrentHashMap<String, IpInterface>();
        List<FcInterface> fcInterfaces = new ArrayList<FcInterface>();

        NaElement results = new NaElement("results");

        try {
            intfLoader.load(results);
        } catch (ApiFailedException e) {
            throw new DataLoaderException(controller + " failed to load fcp interfaces", e);
        }
        
        for (NaElement element : results.getChildren()) {

            FcInterfaceImpl fcIntf = new FcInterfaceImpl(controller);
            setCommonInterfaceFields(element, fcIntf);
            // lets see if we already had one of these based on loading things from previousFcInterfaces
            for (FcInterface curr : previousFcInterfaces) {
            	if (curr.getName() != null && curr.getName().equals(fcIntf.getName())) {
            		fcIntf = (FcInterfaceImpl) curr;
            	}
            }
            fcIntf.setProtocols(Utils.toList("fcp"));            
            fcIntf.setNodeName(element.getChildContent("node-name"));
            fcIntf.setPortAddress(element.getChildContent("port-address"));
            fcIntf.setWwpn(element.getChildContent("port-name"));
            // lets make sure we don't already have one for this vserver/inteface pair
            boolean add = true;
            for (FcInterface curr : fcInterfaces) {
            	if (curr.getName().equals(fcIntf.getName()) && curr.getOwningVserver().equals(fcIntf.getOwningVserver())) {
            		add = false;
            	}
            }
            if (add) {
            	fcInterfaces.add(fcIntf);
            }
        }
        controller.setFcInterfaces(fcInterfaces);
        
    }
}
