/*
 * $Id: //depot/prod/aeonflux/R3.0.1/src/main/java/com/netapp/offtap3/internal/ontap/dataloader/SvmsDataLoader.java#1 $
 *
 * Copyright (c) 2014 NetApp, Inc.
 * All rights reserved.
 */
package com.netapp.offtap3.internal.ontap.dataloader;

import java.lang.reflect.Proxy;
import java.util.ArrayList;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.concurrent.CopyOnWriteArrayList;

import org.apache.log4j.Logger;

import com.netapp.offtap3.ControllerResource;
import com.netapp.offtap3.OfftapPreferences;
import com.netapp.offtap3.Resource;
import com.netapp.offtap3.StorageProtocol;
import com.netapp.offtap3.factories.ControllerFactory;
import com.netapp.offtap3.factories.DataLoaderException;
import com.netapp.offtap3.internal.AbstractDataLoader;
import com.netapp.offtap3.internal.OfftapProxy;
import com.netapp.offtap3.internal.ontap.MCCManager;
import com.netapp.offtap3.internal.ontap.aggregate.CModeAggregateImpl;
import com.netapp.offtap3.internal.ontap.controller.AbstractOntapController;
import com.netapp.offtap3.internal.ontap.controller.AdminSvmImpl;
import com.netapp.offtap3.internal.ontap.controller.DataSvmImpl;
import com.netapp.offtap3.internal.ontap.controller.NodeSvmImpl;
import com.netapp.offtap3.internal.ontap.support.ZapiCache;
import com.netapp.offtap3.internal.ontap.zapi.ZAPIInvoker;
import com.netapp.offtap3.ontap.AdminSvm;
import com.netapp.offtap3.ontap.CModeAggregate;
import com.netapp.offtap3.ontap.DataSvm;
import com.netapp.offtap3.ontap.NaElement;
import com.netapp.offtap3.ontap.NodeSvm;
import com.netapp.offtap3.utils.SafeConcurrentHashMap;
import com.netapp.offtap3.utils.StringUtils;
import com.netapp.offtap3.utils.Utils;
import com.netapp.offtap3.utils.Utils.Callback;

public class SvmsDataLoader extends AbstractDataLoader<AbstractOntapController> {

    private static final Logger logger = Logger.getLogger(SvmsDataLoader.class);

    private OfftapPreferences prefs;

    public SvmsDataLoader(OfftapPreferences prefs) {
        this.prefs = prefs;
    }

    @Override
    protected boolean doLoad(AbstractOntapController controller, LoadType loadType,
                             Set<Resource> resources) throws DataLoaderException {
        // When performing a direct connect load, the controller itself will be the DataSvm.
        boolean directConnect = controller instanceof DataSvm;

        NaElement input = new NaElement("vserver-get-iter");
        NaElement desiredAttributes = new NaElement("desired-attributes");
        NaElement desiredVServerInfo = new NaElement("vserver-info");

        desiredVServerInfo.addChildElem(new NaElement("vserver-type"));
        desiredVServerInfo.addChildElem(new NaElement("root-volume"));
        desiredVServerInfo.addChildElem(new NaElement("root-volume-aggregate"));
        desiredVServerInfo.addChildElem(new NaElement("snapshot-policy"));
        desiredVServerInfo.addChildElem(new NaElement("uuid"));
        desiredVServerInfo.addChildElem(new NaElement("vserver-name"));
        desiredVServerInfo.addChildElem(new NaElement("state"));
        NaElement aggrInfoList = new NaElement("vserver-aggr-info-list");
        desiredVServerInfo.addChildElem(aggrInfoList);
        aggrInfoList.addChildElem(new NaElement("vserver-aggr-info"));
        desiredVServerInfo.addChildElem(new NaElement("vserver-subtype"));
        NaElement allowedProtocols = new NaElement("allowed-protocols");        
        allowedProtocols.addChildElem(new NaElement("protocol"));
        desiredVServerInfo.addChildElem(allowedProtocols);
        desiredVServerInfo.addChildElem(new NaElement("is-repository-vserver"));
        desiredVServerInfo.addChildElem(new NaElement("qos-policy-group"));
        desiredAttributes.addChildElem(desiredVServerInfo);
        input.addChildElem(desiredAttributes);

        NaElement query = new NaElement("query");
        NaElement vserverInfo = new NaElement("vserver-info");
        NaElement vserverName = new NaElement("vserver-type", "data,cluster,node,admin");  // note - if you add more vserver-types here, 
        																				   // you need logic in SvmDataParser to process them
        vserverInfo.addChildElem(vserverName);
        query.addChildElem(vserverInfo);
        input.addChildElem(query);

        CdotIterativeLoader<AbstractOntapController> loader = new CdotIterativeLoader<AbstractOntapController>(
                controller, input, new SvmDataParser(directConnect), prefs.getIteratorSize(), ZapiCache.LONG_CACHE);

        try {
            loader.load(controller);
        } catch (Exception e) {
            throw new DataLoaderException(
                    "Error occurred while loading storage virtual machine information for controller " + controller.getName(),
                    e);
        }

        return true;
    }

    private DataSvmImpl load(DataSvmImpl dataSvm, NaElement svmXml, boolean directConnect) {
        dataSvm.setRootVolume(svmXml.getChildContent("root-volume"));
        dataSvm.setRootVolumeAggregate(svmXml.getChildContent("root-volume-aggregate"));
        dataSvm.setSnapShotPolicy(svmXml.getChildContent("snapshot-policy"));
        String stateXml = svmXml.getChildContent("state");
        try {
            if (stateXml != null) {
                DataSvm.State state = DataSvm.State.valueOf(svmXml.getChildContent("state").toUpperCase());
                dataSvm.setState(state);
            } else {
                logger.warn(dataSvm + " reported a null state");
            }
        } catch (IllegalArgumentException e) {
            logger.error(dataSvm + " reported an unrecognized state: " + stateXml, e);
        }
        dataSvm.setId(svmXml.getChildContent("uuid"));
        dataSvm.setName(svmXml.getChildContent("vserver-name"));
        // check to see if this is a MCC destination vserver.  If so, we should hide it and just fail over to it
        // if we detect problems.        
        if ("sync_source".equals(svmXml.getChildContent("vserver-subtype"))) {
        	dataSvm.setMccSource(true);
        }
        if ("sync_destination".equals(svmXml.getChildContent("vserver-subtype"))) {
        	// normally my inclination would be to skip this.  But there is a single case where we actually want to keep it
        	// and use it.  That's the case of a direct connect vserver, which has been switched over.  In that case, we want to take
        	// the -mc off of the name, and we want to say its an mcc source.  That way when things get switched back, it will 
        	// be in business and rolling.
        	if (directConnect) {
        		logger.debug("See direct connect vserver of sync_destination subtype.  This is a switched over MCC vserver and we are gonna keep it");
        		dataSvm.setMccSource(true);	
        		if (dataSvm.getName().endsWith("-mc")) {
        			dataSvm.setName(dataSvm.getName().substring(0, dataSvm.getName().length()-3));
        		}
        	} else {
            	logger.debug("Skipping sync_destination vserver of " + svmXml.getChildContent("vserver-name"));
            	return null;
        	}
        }
        dataSvm.setQosPolicyGroupName(svmXml.getChildContent("qos-policy-group"));

        // find out if it is a repository vserver a.k.a Vserver with Infinite VolumeImpl
        String isRepository = svmXml.getChildContent("is-repository-vserver");
        if (isRepository != null) {
            dataSvm.setRepository(Boolean.parseBoolean(isRepository));
        } else {
            dataSvm.setRepository(false);
        }

        // load the allowed protocols - this is just establishing the protocols that
        // have a _chance_ of working.  When we hit the DataSvmProtocolDataLoader,
        // it will confirm that the protocol is actually enabled, and if not,
        // it will be removed.  This is helpful for oddball protocols like CIFS,
        // where we rely on the allowed-protocols.
        NaElement allowedProtocols = svmXml.getChildByName("allowed-protocols");
        if (allowedProtocols != null) {
            for (NaElement protocolXml : allowedProtocols.getChildren()) {
                String protocolName = protocolXml.getContent().toUpperCase();
                try {
                    dataSvm.addSupportedProtocol(StorageProtocol.valueOf(protocolName));
                } catch (IllegalArgumentException e) {
                    logger.warn("Ignoring unrecognized storage protocol: " + protocolName);
                }
            }
        }
        
        Map<String, CModeAggregate> aggrs = new SafeConcurrentHashMap<String, CModeAggregate>();
        NaElement aggrInfoList = svmXml.getChildByName("vserver-aggr-info-list");
        if (aggrInfoList != null) {
            for (NaElement aggrInfo : aggrInfoList.getChildren()) {
                String name = aggrInfo.getChildContent("aggr-name");
                long sizeAvailable = aggrInfo.getChildLongValue("aggr-availsize", 0L);
                CModeAggregateImpl aggr = new CModeAggregateImpl(dataSvm, name, sizeAvailable);
                aggrs.put(name, aggr);
            }
        }       
        dataSvm.setAggregateList(aggrs);
        if (dataSvm.isMccSource()) {
        	MCCManager.getInstance().getMCCSlapperThreadCanBeNull(dataSvm.getConnection());
        }
        
        return dataSvm;
    }

    private void load(NodeSvmImpl nodeSvm, NaElement svmXml) {
        nodeSvm.setId(svmXml.getChildContent("uuid"));
        nodeSvm.setName(svmXml.getChildContent("vserver-name"));
    }

    private class SvmDataParser implements DataParser<AbstractOntapController> {
        private boolean directConnect;

        public SvmDataParser(boolean directConnect) {
            this.directConnect = directConnect;
        }

        @Override
        public void parse(final AbstractOntapController controller, NaElement results) {
            NaElement attributesList = results.getChildByName("attributes-list");
            if (attributesList == null) {
                return;
            }

            List<DataSvm> oldDataSvms = new ArrayList<DataSvm>();
            List<NodeSvm> oldNodeSvms = new ArrayList<NodeSvm>();
            if (controller instanceof AdminSvm) {
            	oldDataSvms = new CopyOnWriteArrayList(((AdminSvm) controller).getDataSvms().values());
                oldNodeSvms = new CopyOnWriteArrayList(((AdminSvm) controller).getNodeSvms().values());
            }
            final List<DataSvm> oldDataSvmsToUse = oldDataSvms;
            final List<NodeSvm> oldNodeSvmsToUse = oldNodeSvms;
            final List<DataSvm> dataSvms = new CopyOnWriteArrayList<DataSvm>();
            final List<NodeSvm> nodeSvms = new CopyOnWriteArrayList<NodeSvm>();

            final List<NaElement> vServerInfo = attributesList.getChildren();
            
            // so - we're gonna pre-process this XML and get one very important piece of information.
            // that is how many nodes are in play here, and we're gonna setup a data structure in 
            // ZapiInvoker so that we know how many nodes we are allowed to talk to, so that we 
            // can discover the cluster at appropriate parallelism.
            int nodeCount = 0;
            Set<String> uuids = new HashSet<String>();
            String masterUuid = null;
            for (NaElement currVServerXML : vServerInfo) {
            	if ("node".equals(currVServerXML.getChildContent("vserver-type"))) {
            		nodeCount++;
            	}            	
            	String uuid = currVServerXML.getChildContent("uuid");
            	if ("admin".equals(currVServerXML.getChildContent("vserver-type"))) {
            		masterUuid = uuid;
            	}
            	if (StringUtils.isReal(uuid)) {
            		uuids.add(uuid);
            	}
            }
            if (nodeCount == 0) {
            	nodeCount = 1;
            }
            for (String uuid : uuids) {
            	if (uuid != null) {
	            	ZAPIInvoker.getLastInstance().addNodeCountByUuid(uuid, nodeCount);
	            	if (masterUuid != null) {
	            		ZAPIInvoker.getLastInstance().addUuidMapping(uuid, masterUuid);
	            	}
            	}            	
            }
            
            // Set up the data cache that will be shared between the data svms.
            final OntapDataCache dataCache = new OntapDataCache(controller);
            List callbacks = new ArrayList();
            for (NaElement currVServerXMLOuter : vServerInfo) {
            	final NaElement currVServerXML = currVServerXMLOuter;
            	callbacks.add(new Callback() {
            		public Object call() {
            			if ("cluster".equals(currVServerXML.getChildContent("vserver-type")) || "data".equals(
                                currVServerXML.getChildContent("vserver-type"))) {
                            //this is a data serving svm
                            if (directConnect) {
                                // We're loading a direct connect svm so use the instance provided.
                                load((DataSvmImpl) controller, currVServerXML, directConnect);
                            } else {
                                AdminSvmImpl adminSvm = (AdminSvmImpl) controller;                                
                                boolean shouldKeepVServer = true;
                                DataSvmImpl dataSvmImpl = new DataSvmImpl(adminSvm.getConnection(), adminSvm, dataCache);
                                dataSvmImpl = load(dataSvmImpl, currVServerXML, directConnect);
                                if (dataSvmImpl == null) {
                                	return null;
                                }
                                if (dataSvmImpl.isMccSource()) {
                                	adminSvm.setMccSource(true);
                                }
                                	
                                
                                // lets see if we already had this guy in memory, in which case, we'll basically just reload him
                                // instead of using a fresh one
                                DataSvm existingController = (DataSvm) ControllerFactory.getInstance().getController(dataSvmImpl.getId());
                                if (existingController != null) {
                                	DataSvmImpl priorImpl = (DataSvmImpl) ((OfftapProxy) Proxy.getInvocationHandler(existingController)).getProxiedObject();
                                	boolean load = true;
                                	if (MCCManager.getInstance().isFailedOverSomewhere(controller.getId()) && priorImpl != null && priorImpl.getParentController() != null && !priorImpl.getParentController().getId().equals(controller.getId())) {
                                		// this is case where we've had an MCC failver, we don't want to pick up any new vservers in that case.
                                        return null;
                                	}
                                	priorImpl = load(priorImpl, currVServerXML, directConnect);
                                	oldDataSvmsToUse.remove(priorImpl);
                                	priorImpl.getProxy().refresh(ControllerResource.protocols);
                                } 
                                
                                DataSvm dataSvm = dataSvmImpl.getProxy();
                                dataSvms.add(dataSvm); 
                                if (dataSvm.isMccSource()) {
                                	MCCManager.getInstance().getMCCSlapperThreadCanBeNull(dataSvm.getConnection());
                                }
                            }
                        } else if ("node".equals(currVServerXML.getChildContent("vserver-type"))) {
                            //this is a node svm
                            AdminSvmImpl adminSvm = (AdminSvmImpl) controller;
                            if (MCCManager.getInstance().isFailedOverSomewhere(adminSvm.getId())) {
                            	return null;
                            }
                            NodeSvmImpl nodeSvmImpl = new NodeSvmImpl(adminSvm.getConnection(), adminSvm, dataCache);
                            load(nodeSvmImpl, currVServerXML);
                            
                            // lets see if we already had this guy in memory, in which case, we'll basically just reload him
                            // instead of using a fresh one
                            NodeSvm existingController = (NodeSvm) ControllerFactory.getInstance().getController(nodeSvmImpl.getId());
                            if (existingController != null) {
                            	NodeSvmImpl priorImpl = (NodeSvmImpl) ((OfftapProxy) Proxy.getInvocationHandler(existingController)).getProxiedObject();
                            	load(priorImpl, currVServerXML);
                            	priorImpl.getProxy().refresh(ControllerResource.protocols);
                            }
                            
                            NodeSvm nodeSvm = nodeSvmImpl.getProxy();                    
                            nodeSvms.add(nodeSvm);
                        }
            			return null;
            		}
            		
            	});                
            }
            try {
            	Utils.executeCallbacksInParallel(callbacks);
            } catch (Exception e) {
            	// attempt to throw the original exception, but wrap it if we have to.
            	if (e instanceof RuntimeException) {
            		throw (RuntimeException) e;
            	}
            	throw new RuntimeException(e);            	
            }

            if (controller instanceof AdminSvmImpl) {
            	if (MCCManager.getInstance().isFailedOverSomewhere(controller.getId())) {
            		// pick up all the old ones that used to be in the memory data model.
            		dataSvms.addAll(oldDataSvms);
            		nodeSvms.addAll(oldNodeSvms);
            	}
                ((AdminSvmImpl) controller).setDataSvms(dataSvms);
                ((AdminSvmImpl) controller).setNodeSvms(nodeSvms);
            }
        }
    }
}
