package com.netapp.offtap3.internal.ontap.dataloader;

import com.netapp.offtap3.Resource;
import com.netapp.offtap3.internal.AbstractDataLoader;
import com.netapp.offtap3.internal.OfftapProxy;
import com.netapp.offtap3.internal.ontap.OntapVersionImpl;
import com.netapp.offtap3.internal.ontap.controller.FilerImpl;
import com.netapp.offtap3.internal.ontap.controller.OntapConnection;
import com.netapp.offtap3.internal.ontap.controller.PFilerImpl;
import com.netapp.offtap3.internal.ontap.controller.VFilerImpl;
import com.netapp.offtap3.internal.ontap.intf.IpInterfaceImpl;
import com.netapp.offtap3.internal.ontap.proxy.VFilerProxy;
import com.netapp.offtap3.internal.ontap.support.ZapiCache;
import com.netapp.offtap3.utils.IPAddressUtil;
import com.netapp.offtap3.ontap.ApiFailedException;
import com.netapp.offtap3.ontap.VFiler;
import com.netapp.offtap3.ontap.NaElement;
import com.netapp.offtap3.ontap.NaErrno;
import org.apache.log4j.Logger;

import java.util.List;
import java.util.Set;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

public class VFilerDataLoader extends AbstractDataLoader<FilerImpl> {

    final static Pattern volNamePat = Pattern.compile("^/vol/([^/]+)");
    private static Logger logger = Logger.getLogger(VFilerDataLoader.class);

    public VFilerDataLoader() {
    }

    @Override
    protected boolean doLoad(FilerImpl filer, LoadType loadType, Set<Resource> resources) {

        if (filer.isPassThroughVFiler()) {
            logger.debug("Skip load for pass-through vfiler: " + filer);
            return true;
        }

        OntapVersionImpl version = (OntapVersionImpl) filer.getVersion();
        if (filer.isLicensed("multistore") || version.atLeastSN()) {
            logger.debug("loading vfilers: " + filer + " multistore license present");
            NaElement xi = new NaElement("vfiler-list-info");
            NaElement xo = null;
            try {
                xo = filer.invokeZapi(xi, ZapiCache.LONG_CACHE);
            } catch (ApiFailedException e) {
                if ((e.getErrno() == NaErrno.EAPINOTFOUND) || (e.getErrno() == NaErrno.EAPIERROR)) {
                    logger.warn(
                            "initializeFiler: " + filer + " has multistore licensed, but is unable to get a list of vfilers");
                    filer.addDiscoveryError(
                            "Controller has multistore licensed, but is unable to get a list of vfilers and does not have Multistore in the version string.",
                            e);
                    throw new IllegalStateException(
                            "Unsupported configuration: " + filer + " has multistore licensed, but is unable to get a list of vfilers and does not have Multistore in the version string.");
                }
            }

            if (xo != null && xo.getChildByName("vfilers") != null) {
                List<NaElement> vFilerElems = xo.getChildByName("vfilers").getChildren();

                if (filer instanceof PFilerImpl) {
                    // don't count yourself (vfiler0)
                    logger.debug(filer + " has " + (vFilerElems.size() - 1) + " vFilers");
                } else if (filer instanceof VFilerImpl) {
                    logger.debug(filer + " is a direct vfiler.");
                } else {
                    throw new IllegalArgumentException("Unrecognized filer type: " + filer.getClass());
                }

                for (NaElement vfilerElem : vFilerElems) {
                    String vfilerName = vfilerElem.getChildContent("name");

                    if ("vfiler0".equals(vfilerName)) {
                        // found ourself (the pfiler) in the list of vfilers
                        filer.setUuid(vfilerElem.getChildContent("uuid"));
                    } else {
                        VFilerImpl vfilerImpl;
                        if (filer instanceof VFilerImpl) {
                            // found ourself (the direct vfiler) in the list of vfilers
                            vfilerImpl = (VFilerImpl) filer;
                            vfilerImpl.setUuid(vfilerElem.getChildContent("uuid"));
                            vfilerImpl.setId(filer.getSystemId(), vfilerImpl.getUuid());

                            // this seems strange to just slam in the connect IP address, but it's how the old code worked
                            String connectIpAddress = filer.getConnection().getIpAddress();
                            vfilerImpl.addInterface(new IpInterfaceImpl(connectIpAddress, filer));
                            populateInterfaces(vfilerImpl, vfilerElem.getChildByName("vfnets"));
                        } else {
                            // this is a vfiler managed by the pfiler
                            PFilerImpl pfiler = (PFilerImpl) filer;
                            OntapConnection vfilerConnection = new OntapConnection(pfiler.getConnection());
                            vfilerConnection.setContext(vfilerName);
                            vfilerImpl = new VFilerImpl(vfilerConnection, pfiler);

                            // assign the name, uuid, id information *before* creating the proxy, as that
                            // kicks off discovery, and we want that basic information available
                            vfilerImpl.setName(vfilerName);
                            vfilerImpl.setUuid(vfilerElem.getChildContent("uuid"));
                            vfilerImpl.setId(filer.getSystemId(), vfilerImpl.getUuid());
                            populateInterfaces(vfilerImpl, vfilerElem.getChildByName("vfnets"));

                            // TODO aeonflux-next OPTIMIZATION: this starts discovery of the pass-through vfiler RIGHT HERE.
                            // we probably want to discover the pfiler completely, first, because there is
                            // no need to load aggr, volume info, etc. for each vfiler, if we did it for the entire pfiler.
                            VFiler vfiler = OfftapProxy.create(VFiler.class, VFilerProxy.class, vfilerImpl);
                            pfiler.addVFiler(vfiler);
                        }
                        logger.debug("discovered vfiler: " + vfilerImpl);
                        assignEtcVolName(vfilerElem, vfilerImpl);
                    }
                }
            }
        }

        return true;
    }

    protected void populateInterfaces(VFilerImpl vfiler, NaElement vfnetsElem) {
        if (vfnetsElem != null) {
            for (NaElement vfnetInfo : vfnetsElem.getChildren()) {
                String ipAddress = vfnetInfo.getChildContent("ipaddress");

                String netmask = vfnetInfo.getChildContent("netmask");
                // There is a weird bug where the netmask field can come back as an empty string.
                // The safest approach is to skip those interfaces entirely
                if ((netmask == null) || netmask.equals("")) {
                    logger.warn(
                            "Netmask associated with IP " + ipAddress + " on " + vfiler + " was null or empty... skipping interface.");
                    continue;
                }

                String interfaceName = vfnetInfo.getChildContent("interface");
                String network = IPAddressUtil.getNetwork(ipAddress, netmask);


                IpInterfaceImpl vfInterface = new IpInterfaceImpl(ipAddress, vfiler);
                vfInterface.setNetmask(netmask);
                vfInterface.setNetwork(network);
                vfInterface.setName(interfaceName);

                logger.debug(
                        String.format(vfiler + " added network device: dev (%s), ip (%s), network (%s)", interfaceName,
                                      ipAddress, network));
                vfiler.addInterface(vfInterface);
            }
        }
    }

    private void assignEtcVolName(NaElement vfilerElem, VFilerImpl vfilerImpl) {
        NaElement vfStoresElem = vfilerElem.getChildByName("vfstores");
        if (vfStoresElem != null) {
            for (NaElement vfstoreinfo : vfStoresElem.getChildren()) {
                String path = vfstoreinfo.getChildContent("path");
                Matcher volMatcher = volNamePat.matcher(path);
                if (volMatcher.find()) {
                    String volName = volMatcher.group(1);
                    if (Boolean.parseBoolean(vfstoreinfo.getChildContent("is-etc"))) {
                        vfilerImpl.setEtcVolName(volName);
                        break;
                    }
                }
            }
        }
    }
}
