package com.netapp.offtap3.internal.ontap.dataloader.lun;

import com.netapp.offtap3.Refreshable;
import com.netapp.offtap3.Lun;
import com.netapp.offtap3.OfftapPreferences;
import com.netapp.offtap3.Resource;
import com.netapp.offtap3.factories.DataLoaderException;
import com.netapp.offtap3.internal.AbstractDataLoader;
import com.netapp.offtap3.internal.ontap.controller.DataSvmImpl;
import com.netapp.offtap3.internal.ontap.dataloader.CdotIterativeLoader;
import com.netapp.offtap3.internal.ontap.dataloader.DataParser;
import com.netapp.offtap3.internal.ontap.lun.CModeLunImpl;
import com.netapp.offtap3.internal.ontap.support.ZapiCache;
import com.netapp.offtap3.internal.ontap.utils.StatsCollector;
import com.netapp.offtap3.utils.SafeConcurrentHashMap;
import com.netapp.offtap3.utils.StringUtils;
import com.netapp.offtap3.ontap.ApiFailedException;
import com.netapp.offtap3.ontap.NaElement;
import com.netapp.offtap3.ontap.OntapLun;

import org.apache.log4j.Logger;

import java.util.Collections;
import java.util.Date;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

public class DataSvmLunDataLoader extends AbstractDataLoader<DataSvmImpl> {
    final static Pattern volOnly = Pattern.compile("^/vol/(.*)/(.*)");
    final static Pattern volAndQtreePat = Pattern.compile("^/vol/(.*)/(.*)/(.*)");
    private static Logger logger = Logger.getLogger(DataSvmLunDataLoader.class);
    private OfftapPreferences prefs;

    public DataSvmLunDataLoader(OfftapPreferences prefs) {
        this.prefs = prefs;
    }

    @Override
    protected Refreshable doRefresh(DataSvmImpl loadTarget, Resource resource, Refreshable refreshTarget) {
        Lun lun = (Lun) refreshTarget;
        doLoad(loadTarget, LoadType.refresh, Collections.singleton(resource), lun);
        return loadTarget.getLun(lun.getPath());
    }

    @Override
    protected boolean doLoad(DataSvmImpl dataSvm, LoadType loadType, Set<Resource> resources) throws DataLoaderException {
        return doLoad(dataSvm, loadType, resources, null);
    }

    protected boolean doLoad(DataSvmImpl dataSvm, LoadType loadType, Set<Resource> resources, Lun reloadTarget) {
        NaElement input = new NaElement("lun-get-iter");
        Map<String, OntapLun> luns;
        if (loadType == LoadType.refresh && reloadTarget != null) {
        	StatsCollector.getInstance().addCounter("refresh-targeted-lun");
            NaElement query = new NaElement("query");
            NaElement lunInfo = new NaElement("lun-info");
            lunInfo.addNewChild("path", reloadTarget.getPath());
            query.addChildElem(lunInfo);
            input.addChildElem(query);

            luns = dataSvm.getLuns();
            luns.remove(reloadTarget.getPath());
        } else {
        	StatsCollector.getInstance().addCounter("refresh-whole-lun");
            luns = new SafeConcurrentHashMap<String, OntapLun>();
        }
        NaElement desiredAttributes = new NaElement("desired-attributes");
        NaElement lunInfo = new NaElement("lun-info");
        lunInfo.addChildElem(new NaElement("path"));
        lunInfo.addChildElem(new NaElement("uuid"));
        lunInfo.addChildElem(new NaElement("multiprotocol-type"));
        lunInfo.addChildElem(new NaElement("share-state"));
        lunInfo.addChildElem(new NaElement("serial-number"));
        lunInfo.addChildElem(new NaElement("online"));
        lunInfo.addChildElem(new NaElement("read-only"));
        lunInfo.addChildElem(new NaElement("is-space-reservation-enabled"));
        lunInfo.addChildElem(new NaElement("is-space-alloc-enabled"));
        lunInfo.addChildElem(new NaElement("size"));
        lunInfo.addChildElem(new NaElement("qos-policy-group"));
        lunInfo.addChildElem(new NaElement("device-id"));
        lunInfo.addChildElem(new NaElement("prefix-size"));
        lunInfo.addChildElem(new NaElement("creation-timestamp"));
        if (dataSvm.getVersion().atLeastSN()) {
            lunInfo.addChildElem(new NaElement("class"));
        }
        desiredAttributes.addChildElem(lunInfo);
        input.addChildElem(desiredAttributes);
        CdotIterativeLoader<Map<String, OntapLun>> loader = new CdotIterativeLoader<Map<String, OntapLun>>(dataSvm, input,
                                                                                                 new CModeLunParser(
                                                                                                         dataSvm),
                                                                                                 prefs.getIteratorSize(),
                                                                                                 ZapiCache.SHORT_CACHE);

        try {
            loader.load(luns);
        } catch (ApiFailedException e) {
            throw new DataLoaderException("Error occurred while loading luns for data svm " + dataSvm.getName(), e);
        }

        dataSvm.setLuns(luns);

        return true;
    }

    private class CModeLunParser implements DataParser<Map<String, OntapLun>> {
        private DataSvmImpl dataSvm;

        public CModeLunParser(DataSvmImpl dataSvm) {
            this.dataSvm = dataSvm;
        }

        @Override
        public void parse(Map<String, OntapLun> luns, NaElement results) {
            NaElement attrList = results.getChildByName("attributes-list");
            if (attrList == null) {
                return;
            }

            List<NaElement> lunInfoList = attrList.getChildren();
            for (NaElement lunInfo : lunInfoList) {
                String path = lunInfo.getChildContent("path");
                if (path == null) {
                    continue;
                }

                logger.trace("getLun: start " + path);

                Matcher m = volAndQtreePat.matcher(path);
                String vol = "unknown";
                String name = "unknown";
                String qtree = "unknown";
                if (m.find()) {
                    logger.debug("returning (qtree) " + m.group(1));
                    vol = m.group(1);
                    qtree = m.group(2);
                    name = m.group(3);
                } else {
                    m = volOnly.matcher(path);
                    if (m.find()) {
                        logger.debug("returning (no qtree) " + m.group(1));
                        vol = m.group(1);
                        name = m.group(2);
                        qtree = null;
                    } else {
                        logger.error("Unable to find volume or qtree info from " + path);
                    }
                }
                logger.debug("buildLun: path elements; vol=" + vol + " qtree=" + qtree + " name=" + name);

                CModeLunImpl lun = new CModeLunImpl(dataSvm);
                lun.setName(name);
                lun.setQtree(qtree);
                lun.setVolumeName(vol);
                lun.setPath(path);
                lun.setUuid(lunInfo.getChildContent("uuid"));
                lun.setType(lunInfo.getChildContent("multiprotocol-type"));
                lun.setShareState(lunInfo.getChildContent("share-state"));
                lun.setSerial(lunInfo.getChildContent("serial-number"));
                lun.setOnline(Boolean.parseBoolean(lunInfo.getChildContent("online")));
                lun.setReadOnly(Boolean.parseBoolean(lunInfo.getChildContent("read-only")));
                lun.setSpaceReserve(Boolean.parseBoolean(lunInfo.getChildContent("is-space-reservation-enabled")));
                lun.setSpaceAlloc(Boolean.parseBoolean(lunInfo.getChildContent("is-space-alloc-enabled")));
                lun.setTotalSize(lunInfo.getChildLongValue("size", 0L));
                lun.setUsedSize(lunInfo.getChildLongValue("size-used", 0L));
                lun.setDeviceId(lunInfo.getChildIntValue("device-id", 0));
                lun.setPrefixBytes(Long.parseLong(lunInfo.getChildContent("prefix-size")));
                lun.setCreationTimestamp(lunInfo.getChildLongValue("creation-timestamp", 0));

                if (StringUtils.isReal(lunInfo.getChildContent("qos-policy-group"))) {
                    String policyGroupName = lunInfo.getChildContent("qos-policy-group");
                    lun.setQosPolicyGroup(policyGroupName);
                }

                String lunClass = lunInfo.getChildContent("class");
                if (lunClass == null) {
                    lunClass = "regular";
                }
                lun.setLunClass(lunClass);
                lun.setLastLoad(new Date());
                luns.put(lun.getPath(), lun.getProxy());
            }
        }
    }
}
