package com.netapp.offtap3.internal.ontap.dataloader.lun;

import com.netapp.offtap3.Lun;
import com.netapp.offtap3.OfftapPreferences;
import com.netapp.offtap3.Refreshable;
import com.netapp.offtap3.Resource;
import com.netapp.offtap3.factories.DataLoaderException;
import com.netapp.offtap3.internal.AbstractDataLoader;
import com.netapp.offtap3.internal.ontap.controller.FilerImpl;
import com.netapp.offtap3.internal.ontap.lun.SevenModeLunImpl;
import com.netapp.offtap3.internal.ontap.support.ZapiCache;
import com.netapp.offtap3.ontap.ApiFailedException;
import com.netapp.offtap3.ontap.Filer;
import com.netapp.offtap3.ontap.NaElement;
import com.netapp.offtap3.ontap.OntapLun;
import com.netapp.offtap3.ontap.utils.OntapCallable;
import com.netapp.offtap3.utils.SafeConcurrentHashMap;
import com.netapp.offtap3.utils.Utils.VscThreadFactory;

import org.apache.log4j.Logger;

import java.util.Collections;
import java.util.Date;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.concurrent.CompletionService;
import java.util.concurrent.ExecutorCompletionService;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.concurrent.Future;
import java.util.regex.Matcher;
import java.util.regex.Pattern;


public class FilerLunDataLoader<F extends Filer> extends AbstractDataLoader<FilerImpl<F>> {
    final static Pattern prefixPattern = Pattern.compile("Prefix Size:[^(]+\\(([0-9]+)\\)");
    final static Pattern volOnly = Pattern.compile("^/vol/(.*)/(.*)");
    final static Pattern volAndQtreePat = Pattern.compile("^/vol/(.*)/(.*)/(.*)");
    private final static Logger LOGGER = Logger.getLogger(FilerLunDataLoader.class);
    private final OfftapPreferences offtapPrefs;

    public FilerLunDataLoader(OfftapPreferences preferences) {
        offtapPrefs = preferences;
    }

    @Override
    protected Refreshable doRefresh(FilerImpl<F> loadTarget, Resource resource, Refreshable refreshTarget) {
        Lun lun = (Lun) refreshTarget;
        doLoad(loadTarget, LoadType.refresh, Collections.singleton(resource), lun);
        return loadTarget.getLun(lun.getPath());
    }

    @Override
    protected boolean doLoad(final FilerImpl<F> filer, LoadType loadType, Set<Resource> resources) {
        return doLoad(filer, loadType, resources, null);
    }

    protected boolean doLoad(final FilerImpl<F> filer, LoadType loadType, Set<Resource> resources, Lun reloadTarget) {
        LOGGER.debug(filer + " start");

        Map<String, OntapLun> luns = new SafeConcurrentHashMap<String, OntapLun>();
        NaElement input = new NaElement("lun-list-info");
        if (loadType == LoadType.refresh && reloadTarget != null) {
            input.addNewChild("path", reloadTarget.getPath());
        }

        final ExecutorService pool = Executors.newFixedThreadPool(offtapPrefs.getMaxRunningLunTaskThreads(), new VscThreadFactory("filerLunDataLoaderPool"));
        CompletionService<OntapLun> completionService = new ExecutorCompletionService<OntapLun>(pool);

        try {
            NaElement xo = filer.invokeZapi(input, ZapiCache.SHORT_CACHE);
            if (xo != null) {
                List<NaElement> lunElements = xo.getChildByName("luns").getChildren();

                for (final NaElement lunInfo : lunElements) {
                    completionService.submit(new OntapCallable<OntapLun>() {
                        public OntapLun callLocal() {
                            try {
                                return getLun(filer, lunInfo);
                            } catch (Exception e) {
                                LOGGER.error(filer + " Unable to getLun using:\n" + lunInfo + "\n", e);
                            }
                            return null;
                        }
                    });
                }

                for (NaElement lunElement : lunElements) {
                    Future<OntapLun> f = completionService.take();
                    OntapLun lun = f.get();
                    luns.put(lun.getPath(), lun);
                }
            }

            // don't set the reference to the new collection until we're done modifying it
            // this will ensure data integrity across threads
            if (reloadTarget != null) {
                // When handling a reload target, just update the existing list vs. setting the new list.
                OntapLun lun = luns.get(reloadTarget.getPath());
                if (lun == null) {
                    filer.getLuns().remove(reloadTarget.getPath());
                } else {
                    filer.getLuns().put(reloadTarget.getPath(), lun);
                }
            } else {
                filer.setLuns(luns);
            }

        } catch (Exception e) {
            LOGGER.error("populateLun (7-mode): " + filer + " Unable to load luns.", e);
            throw new DataLoaderException("populateLun (7-mode): " + filer + " Unable to load luns." + e.getMessage(),
                                          e);
        } finally {
            pool.shutdownNow();
        }

        LOGGER.debug("populateLun (7-mode): done " + filer);
        return true;
    }

    protected OntapLun getLun(FilerImpl<F> filer, NaElement lunInfo) throws ApiFailedException {
        String path = lunInfo.getChildContent("path");
        LOGGER.trace(filer + " getLun: start " + lunInfo);
        SevenModeLunImpl lun = new SevenModeLunImpl<F>(filer);

        Matcher m = volAndQtreePat.matcher(path);
        String vol = "unknown";
        String name = "unknown";
        String qtree = "unknown";
        if (m.find()) {
            LOGGER.debug("returning (qtree) " + m.group(1));
            vol = m.group(1);
            qtree = m.group(2);
            name = m.group(3);
        } else {
            m = volOnly.matcher(path);
            if (m.find()) {
                LOGGER.debug("returning (no qtree) " + m.group(1));
                vol = m.group(1);
                name = m.group(2);
                qtree = null;
            } else {
                LOGGER.error("Unable to find volume or qtree info from " + path);
            }
        }
        LOGGER.debug("buildLun: path elements; vol=" + vol + " qtree=" + qtree + " name=" + name);

        lun.setName(name);
        lun.setQtree(qtree);
        lun.setVolumeName(vol);
        lun.setPath(path);
        lun.setUuid(lunInfo.getChildContent("uuid"));
        lun.setType(lunInfo.getChildContent("multiprotocol-type"));
        lun.setShareState(lunInfo.getChildContent("share-state"));
        lun.setSerial(lunInfo.getChildContent("serial-number"));
        lun.setOnline(Boolean.parseBoolean(lunInfo.getChildContent("online")));
        lun.setReadOnly(Boolean.parseBoolean(lunInfo.getChildContent("read-only")));
        lun.setSpaceReserve(Boolean.parseBoolean(lunInfo.getChildContent("is-space-reservation-enabled")));
        lun.setTotalSize(lunInfo.getChildLongValue("size", 0L));
        lun.setUsedSize(lunInfo.getChildLongValue("size-used", 0L));
        lun.setDeviceId(lunInfo.getChildIntValue("device-id", 0));

        long maxResizeSize = -1;
        NaElement executeApi = new NaElement("lun-get-geometry");
        // get lun-get-geometry
        try {
            LOGGER.trace("Calling lun-get-geometry for " + path);
            executeApi.addNewChild("path", path);

            NaElement lunGeometryInfo = filer.invokeZapi(executeApi, ZapiCache.LONG_CACHE);
            maxResizeSize = lunGeometryInfo.getChildLongValue("max-resize-size", 0L);
        } catch (ApiFailedException e) {
            filer.addDiscoveryError("Failed to get lun geometry", e);
        } catch (Exception e) {
            LOGGER.error("getLun: " + filer + " " + lun.getPath() + " failed to get snapmirror status : " +
                                 e.getMessage());
            filer.addDiscoveryError("Exception while calling lun-get-geometry" + lun.getPath(), e);
        }

        // get lun prefix
        long prefix = 0;
        try {
            LOGGER.trace("Calling system-cli (lun show -v) for " + name);
            NaElement xi = new NaElement("system-cli");
            xi.addNewChild("priv", "diag");
            NaElement xa = new NaElement("args");

            filer.addSystemCliArgs(xa);

            xa.addNewChild("arg", "lun");
            xa.addNewChild("arg", "show");
            xa.addNewChild("arg", "-v");
            xa.addNewChild("arg", path);
            xi.addChildElem(xa);

            NaElement xo = filer.invokeZapi(xi, ZapiCache.SHORT_CACHE);
            if (xo == null) {
                throw new NullPointerException("invokeZAPI returned null");
            }

            NaElement cliOutput = xo.getChildByName("cli-output");
            if (cliOutput == null) {
                throw new Exception("cli-output node not found");
            }
            String output = cliOutput.getContent();
            if (output == null || output.length() == 0) {
                throw new Exception("cli-output had no content");
            }
            String[] outputLines = output.trim().split("\\n\\s*");
            for (String line : outputLines) {
                Matcher prefixMatcher = prefixPattern.matcher(line);
                if (prefixMatcher.find()) {
                    String sPrefix = prefixMatcher.group(1);
                    if (sPrefix.length() > 0) {
                        prefix = Integer.parseInt(sPrefix);
                    }
                }
            }
        } catch (Exception e) {
            LOGGER.error("Couldn't get LUN verbose info", e);
        }
        lun.setPrefixBytes(prefix);
        lun.setMaxResizeSize(maxResizeSize);
        lun.setLastLoad(new Date());

        return lun.getProxy();
    }

}
