package com.netapp.offtap3.internal.ontap.dataloader.volume;

import com.netapp.offtap3.Refreshable;
import com.netapp.offtap3.OfftapPreferences;
import com.netapp.offtap3.Resource;
import com.netapp.offtap3.Size;
import com.netapp.offtap3.internal.AbstractDataLoader;
import com.netapp.offtap3.internal.ontap.controller.FilerImpl;
import com.netapp.offtap3.internal.ontap.support.ZapiCache;
import com.netapp.offtap3.ontap.utils.OntapCallable;
import com.netapp.offtap3.internal.ontap.volume.DedupeImpl;
import com.netapp.offtap3.internal.ontap.volume.FilerVolumeImpl;
import com.netapp.offtap3.ontap.ApiFailedException;
import com.netapp.offtap3.ontap.Dedupe;
import com.netapp.offtap3.ontap.Filer;
import com.netapp.offtap3.ontap.FilerVolume;
import com.netapp.offtap3.ontap.NaElement;
import com.netapp.offtap3.ontap.OntapVolume;
import com.netapp.offtap3.utils.SafeConcurrentHashMap;
import com.netapp.offtap3.utils.Utils.VscThreadFactory;

import org.apache.log4j.Logger;

import java.util.Collections;
import java.util.Date;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.concurrent.CompletionService;
import java.util.concurrent.ExecutorCompletionService;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.concurrent.TimeUnit;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

public class FilerVolumeDataLoader extends AbstractDataLoader<FilerImpl<? extends Filer>> {
    private static Logger logger = Logger.getLogger(FilerVolumeDataLoader.class);

    OfftapPreferences prefs;

    public FilerVolumeDataLoader(OfftapPreferences prefs) {
        this.prefs = prefs;
    }


    @Override
    protected Refreshable doRefresh(FilerImpl<? extends Filer> loadTarget, Resource resource, Refreshable refreshTarget) {
        doLoad(loadTarget, LoadType.refresh, Collections.singleton(resource), refreshTarget);
        if (refreshTarget instanceof Dedupe) {
            OntapVolume volume = loadTarget.getVolume(((Dedupe) refreshTarget).getVolume().getName());
            if (volume != null) {
                return volume.getDedupe();
            }
        }
        return null;
    }

    @Override
    protected boolean doLoad(final FilerImpl<? extends Filer> filer, LoadType loadType, Set<Resource> resources) {
        return doLoad(filer, loadType, resources, null);
    }

    protected boolean doLoad(final FilerImpl<? extends Filer> filer, LoadType loadType, Set<Resource> resources, Refreshable reloadTarget) {
        String volumeName = null;
        if (loadType == LoadType.refresh && reloadTarget != null && reloadTarget instanceof Dedupe) {
            volumeName = ((Dedupe)reloadTarget).getVolume().getName();
        }
        Map<String, FilerVolume> volumes = buildVolumes(filer, volumeName);

        // don't set the reference to the new list until we're done modifying it
        // this will ensure data integrity across threads
        filer.setVolumes(volumes);

        logger.debug("populateVolume: done " + filer);

        return true;
    }

    protected Map<String, FilerVolume> buildVolumes(final FilerImpl<? extends Filer> filer, String volumeName) {
        logger.debug(filer + " start");
        NaElement xi;

        final Map<String, FilerVolume> volumes = new SafeConcurrentHashMap<String, FilerVolume>();

        xi = new NaElement("volume-list-info-iter-start");
        xi.addNewChild("verbose", "true");
        if (volumeName != null) {
            xi.addNewChild("volume", volumeName);
            volumes.putAll(filer.getVolumes());
            volumes.remove(volumeName);
        }

        final ExecutorService pool = Executors.newFixedThreadPool(prefs.getMaxRunningVolumeThreads(), new VscThreadFactory("filerVolumeDataLoaderPool"));
        final CompletionService<Void> completionService = new ExecutorCompletionService<Void>(pool);
        String tag = null;

        try {
            NaElement xo = filer.invokeZapi(xi, ZapiCache.NO_CACHE);
            int retrievedRecords = xo.getChildIntValue("records", 0);
            tag = xo.getChildContent("tag");
            logger.debug("populateVolume: volume-list-info-iter-start: records=" + retrievedRecords + " tag=" + tag);

            while (retrievedRecords > 0) {

                // ONTAP says it has retrievedRecords ready, but our preferences says to never fetch more
                // than some limit to avoid overruning the controller (in other words, don't trust ONTAP to throttle itself).
                int fetch = Math.min(retrievedRecords, prefs.getMaxVolumeListInfoRecords());

                xi = new NaElement("volume-list-info-iter-next");
                xi.addNewChild("tag", tag);
                xi.addNewChild("maximum", String.valueOf(fetch));
                logger.debug("populateVolume: volume-list-info-iter-next  IN: maximum=" + fetch + " tag=" + tag);

                // advance the iterator
                xo = filer.invokeZapi(xi, ZapiCache.NO_CACHE);

                retrievedRecords = xo.getChildIntValue("records", 0);
                logger.debug(
                        "populateVolume: volume-list-info-iter-next OUT: records=" + retrievedRecords + " tag=" + tag);

                if (retrievedRecords == 0) {
                    continue;
                }

                // submit this batch of XML results to the pool to create the NtapVolumes (requires additional ZAPIs)
                for (final NaElement volumeInfo : xo.getChildByName("volumes").getChildren()) {
                    completionService.submit(new OntapCallable<Void>() {
                        @Override
                        public Void callLocal() {
                            try {
                                FilerVolume vol = getVolume(filer, volumeInfo);
                                if (vol != null) {
                                    volumes.put(vol.getName(), vol);
                                }
                            } catch (Exception e) {
                                logger.error(
                                        "getVolumes: " + filer + " Unable to buildVolume using:\n" + volumeInfo + "\n",
                                        e);
                                filer.addDiscoveryError("Failed to build volume with: " + volumeInfo, e);
                            }
                            return null;
                        }
                    });
                }
            }

        } catch (ApiFailedException e) {
            filer.addDiscoveryError("Failed to load volumes", e);
            logger.error("populateVolume: " + filer + " Unable to load volumes.");
        } finally {
            if (tag != null) {
                // if tag is not null, that means that the iter was started -- and it's the client's job to clean up!
                xi = new NaElement("volume-list-info-iter-end");
                xi.addNewChild("tag", tag);
                try {
                    filer.invokeZapi(xi, ZapiCache.NO_CACHE);
                } catch (ApiFailedException e) {
                    logger.warn("Failed to end volume-iter for: " + filer, e);
                }
                logger.debug("populateVolume: volume-list-info-iter-end: tag=" + tag);
            }
            pool.shutdown();
            try {
                pool.awaitTermination(30, TimeUnit.MINUTES);
            } catch (InterruptedException ie) {
                logger.error("InterruptedException while waiting for pool to finish", ie);
            }
        }

        return volumes;

    }

    /**
     * Process NaElement and connect to filer when more info is needed.
     */
    public FilerVolume getVolume(FilerImpl<? extends Filer> filer, NaElement volInfo) {
        FilerVolumeImpl vol = new FilerVolumeImpl(filer);
        logger.trace("getVolume: " + filer.getName() + " " + volInfo);

        vol.setName(volInfo.getChildContent("name"));
        logger.debug("getVolume: " + vol.getName() + " on " + filer.getName());
        vol.setUuid(volInfo.getChildContent("uuid"));
        vol.setState(volInfo.getChildContent("state"));
        vol.setType(volInfo.getChildContent("type"));
        vol.setSnapshotPercentReserved(volInfo.getChildIntValue("snapshot-percent-reserved", 0));
        vol.setSnapshotReservePercentUsed(volInfo.getChildIntValue("percentage-snapshot-reserve-used", 0));
        vol.setSnapshotSpaceReserved(volInfo.getChildLongValue("snapshot-blocks-reserved", 0L) *
                                             1024); // these are reported as 1024 KB blocks

        String threshold = volInfo.getChildContent("spacemgmnt-high-threshold-trigger-percentage");
        if (threshold != null) {
            vol.setEarlyWarningThreshold(Integer.decode(threshold));
        }

        vol.setSizeTotal(volInfo.getChildLongValue("size-total", 0L));
        vol.setSizeUsed(volInfo.getChildLongValue("size-used", 0L));
        vol.setSizeAvailable(volInfo.getChildLongValue("size-available", 0L));
        vol.setPercentageUsed(volInfo.getChildIntValue("percentage-used", 0));

        vol.setInconsistent(Boolean.parseBoolean(volInfo.getChildContent("is-inconsistent")));

        if (volInfo.getChildByName("sis") != null) {
            // 7.3.1 style
            NaElement sisInfo = volInfo.getChildByName("sis").getChildByName("sis-info");
            if (sisInfo == null) {
                // 7.3.0 style
                sisInfo = volInfo.getChildByName("sis");
            }
            if (sisInfo != null) {
                DedupeImpl dedupe = new DedupeImpl(filer);
                dedupe.setEnabled("enabled".equals(sisInfo.getChildContent("state")));
                dedupe.setLastBeginDate(sisInfo.getChildContent("last-operation-begin"));
                dedupe.setLastEndDate(sisInfo.getChildContent("last-operation-end"));
                dedupe.setType(sisInfo.getChildContent("type"));
                dedupe.setStatus(sisInfo.getChildContent("status"));
                dedupe.setSchedule(sisInfo.getChildContent("schedule"));
                dedupe.setSizeShared(sisInfo.getChildLongValue("size-shared", 0L));
                dedupe.setSizeSaved(sisInfo.getChildLongValue("size-saved", 0L));
                dedupe.setPercentSaved(sisInfo.getChildIntValue("percentage-saved", 0));
                dedupe.setVolume(vol);
                vol.setDedupe(dedupe);
            }
        }

        if (volInfo.getChildByName("snap-autodelete") != null) {
            List<NaElement> autoDelInfoList = volInfo.getChildByName("snap-autodelete").getChildren();
            for (NaElement autoDelInfo : autoDelInfoList) {
                vol.setAutoDeleteStatus(autoDelInfo.getChildContent("state"));
                vol.setAutoDeleteTrigger(autoDelInfo.getChildContent("trigger"));
            }
        }

        if (volInfo.getChildByName("vm-align") != null) {
            NaElement vmAlignInfo = volInfo.getChildByName("vm-align").getChildByName("vm-align-info");
            vol.setVmAlignSector(vmAlignInfo.getChildIntValue("vm-align-sector", 0));
            vol.setVmAlignSuffix(vmAlignInfo.getChildContent("vm-align-suffix"));
        }

        if ("flex".equals(vol.getType())) {
            vol.setSpaceReserveEnabled(Boolean.parseBoolean(volInfo.getChildContent("space-reserve-enabled")));
            vol.setContainingAggregate(volInfo.getChildContent("containing-aggregate"));
            vol.setSpaceReserve(volInfo.getChildContent("space-reserve"));
            vol.setSnapLock(Boolean.parseBoolean(volInfo.getChildContent("is-snaplock")));
        }

        if (volInfo.getChildByName("clone-children") != null) {
            vol.setCloneChildCount(volInfo.getChildByName("clone-children").getChildren().size());
        }

        String parentSnapshotName = null;
        String parentVolumeName = null;
        if (volInfo.getChildByName("clone-parent") != null) {
            parentSnapshotName = volInfo.getChildByName("clone-parent").getChildByName("clone-parent-info")
                                        .getChildContent("parent-snapshot-name");
            parentVolumeName = volInfo.getChildByName("clone-parent").getChildByName("clone-parent-info")
                                      .getChildContent("parent-volume-name");
        }
        vol.setParentSnapshotName(parentSnapshotName);
        vol.setParentVolumeName(parentVolumeName);


        NaElement xi = new NaElement("volume-options-list-info");
        xi.addNewChild("volume", vol.getName());

        try {
            NaElement xo = filer.invokeZapi(xi, ZapiCache.SHORT_CACHE);
            List<NaElement> volOptInfoList = xo.getChildByName("options").getChildren();
            for (NaElement volOptInfo : volOptInfoList) {
                if (volOptInfo.getChildContent("name").equals("convert_ucode")) {
                    String value = volOptInfo.getChildContent("value");
                    vol.setConvertUcodeEnabled("on".equalsIgnoreCase(value));
                }
                if (volOptInfo.getChildContent("name").equals("create_ucode")) {
                    String value = volOptInfo.getChildContent("value");
                    vol.setCreateUcodeEnabled("on".equalsIgnoreCase(value));
                }
                if (volOptInfo.getChildContent("name").equals("no_atime_update")) {
                    String value = volOptInfo.getChildContent("value");
                    vol.setAtimeUpdateEnabled("off".equals(value));
                }
                if (volOptInfo.getChildContent("name").equals("fractional_reserve")) {
                    vol.setFractionalReserve(volOptInfo.getChildIntValue("value", 0));
                }
            }
        } catch (ApiFailedException e) {
            filer.addDiscoveryError("Failed to get volume options for " + vol.getName(), e);
        } catch (Exception e) {
            logger.error(
                    "getVolume: " + filer + " " + vol.getName() + " failed to get volume options : " + e.getMessage());
            filer.addDiscoveryError("Failed to get volume options for " + vol.getName(), e);
        }

        if ("online".equals(vol.getState()) && !"trad".equals(vol.getType()) && !vol.getSnapMirrorStatus().isSource() &&
                !vol.getSnapMirrorStatus().isDestination()) {
            try {
                logger.trace("Getting volume-autosize-get for " + vol.getName());
                xi = new NaElement("volume-autosize-get");
                xi.addNewChild("volume", vol.getName());
                NaElement autoSizeInfo = filer.invokeZapi(xi, ZapiCache.SHORT_CACHE);
                vol.setAutoSizeEnabled(Boolean.parseBoolean(autoSizeInfo.getChildContent("is-enabled")));

                // look out!  autosize increment and max is in KB!
                vol.setAutoSizeIncrement(Size.createKB(autoSizeInfo.getChildLongValue("increment-size", 0L)));
                vol.setAutoSizeMax(Size.createKB(autoSizeInfo.getChildLongValue("maximum-size", 0L)));

                logger.trace("Finished volume-autosize-get for " + vol.getName());
            } catch (Exception e) {
                filer.addDiscoveryError("Exception while getting autosize for " + vol.getName(), e);
                logger.error("Exception while getting autosize.", e);
            }
        }

        if ("online".equals(vol.getState())) {
            long totalSnapshotSpace = 0L;
            try {
                xi = new NaElement("system-cli");
                xi.addNewChild("priv", "admin");
                NaElement xa = new NaElement("args");

                filer.addSystemCliArgs(xa);

                xa.addNewChild("arg", "df");
                xa.addNewChild("arg", vol.getName());
                xi.addChildElem(xa);
                // use the connection from the physical filer to do this.
                NaElement xo = filer.getNonPassThroughConnection().invokeZapi(xi, ZapiCache.NO_CACHE);
                int rc = xo.getChildIntValue("cli-result-value", 0);
                String output = xo.getChildContent("cli-output");
                if (rc != 1) {
                    logger.error(
                            String.format("Get snapshot size (using df) failed. rc = %s.  output = %s", rc, output));
                }

                Pattern dfPattern = Pattern.compile("/vol/" + vol.getName() + "/.snapshot\\s+\\d+\\s+(\\d+)");
                for (String line : output.split("\\n")) {
                    Matcher matchedLine = dfPattern.matcher(line);
                    if (matchedLine.find()) {
                        String value = matchedLine.group(1);
                        totalSnapshotSpace = Long.parseLong(value);
                    }
                }
                logger.debug("Total snapshot space for " + vol.getName() + " : " + totalSnapshotSpace);
            } catch (Exception e) {
                filer.addDiscoveryError("Exception while getting snapshot space for " + vol.getName(), e);
                logger.error("Exception while getting snapshot space.", e);
            }
            vol.setTotalSnapshotSpace(totalSnapshotSpace);
        }

        vol.setLastLoad(new Date());

        return vol.getProxy();
    }
}
