/*
 * $Id: //depot/prod/aeonflux/R3.0.1/src/main/java/com/netapp/offtap3/internal/ontap/dataloader/volume/VolumeMaximumDataLoader.java#1 $
 *
 * Copyright (c) 2014 NetApp, Inc.
 * All rights reserved.
 */
package com.netapp.offtap3.internal.ontap.dataloader.volume;

import com.netapp.offtap3.Resource;
import com.netapp.offtap3.Size;
import com.netapp.offtap3.factories.DataLoaderException;
import com.netapp.offtap3.internal.AbstractDataLoader;
import com.netapp.offtap3.internal.HasParentControllerImpl;
import com.netapp.offtap3.internal.ontap.OntapVolumeMaximumsImpl;
import com.netapp.offtap3.internal.ontap.controller.AbstractOntapController;
import com.netapp.offtap3.internal.ontap.controller.FilerImpl;
import com.netapp.offtap3.internal.ontap.controller.OntapConnection;
import com.netapp.offtap3.internal.ontap.support.ZapiCache;
import com.netapp.offtap3.ontap.DataSvm;
import com.netapp.offtap3.ontap.NaElement;
import com.netapp.offtap3.ontap.OntapController;
import com.netapp.offtap3.ontap.OntapVolumeMaximums;
import com.netapp.offtap3.utils.StringUtils;
import org.apache.log4j.Logger;

import javax.xml.bind.JAXBContext;
import javax.xml.bind.JAXBException;
import javax.xml.bind.Unmarshaller;
import java.io.InputStream;
import java.util.List;
import java.util.Set;

public class VolumeMaximumDataLoader extends AbstractDataLoader<AbstractOntapController<? extends OntapController>> {

    private static final String VOLMAX_FILE_NAME = "/com/netapp/offtap3/internal/ontap/dataloader/volume/volumemaximums.xml";

    private static final Logger LOGGER = Logger.getLogger(VolumeMaximumDataLoader.class);
    private static final Class<?>[] DATA_CLASSES = {VolumeMaximum.class, OntapRelease.class, StorageController.class};

    // outside of this class, only referenced in volumeMaximums.xml
    private static final String DIRECT_VSERVER_MODEL = "direct-vserver";
    private static final OntapVolumeMaximumsImpl DEFAULT_MAXIMUMS = new OntapVolumeMaximumsImpl();

    public static final int DEFAULT_32BIT_DENSE_SIZE_TB = 14;
    public static final int DEFAULT_32BIT_SHARED_DENSE_SIZE_TB = 32;
    public static final int DEFAULT_64BIT_DENSE_SIZE_TB = 64;
    public static final int DEFAULT_64BIT_SHARED_DENSE_SIZE_TB = 64;

    static {
        DEFAULT_MAXIMUMS.setMax32BitDenseVolumeSize(Size.createTB(DEFAULT_32BIT_DENSE_SIZE_TB));
        DEFAULT_MAXIMUMS.setMax32BitDenseSharedVolumeSize(Size.createTB(DEFAULT_32BIT_SHARED_DENSE_SIZE_TB));
        DEFAULT_MAXIMUMS.setMax64BitDenseVolumeSize(Size.createTB(DEFAULT_64BIT_DENSE_SIZE_TB));
        DEFAULT_MAXIMUMS.setMax64BitDenseSharedVolumeSize(Size.createTB(DEFAULT_64BIT_SHARED_DENSE_SIZE_TB));
    }

    private volatile VolumeMaximum volumeMaximum;
    private static final Object volumeMaximumLoadLock = new Object();

    @Override
    protected boolean doLoad(AbstractOntapController<? extends OntapController> controller, LoadType loadType, Set<Resource> resources) throws DataLoaderException {
        OntapVolumeMaximums volMax = getFromParentController(controller);
        if (volMax == null) {
            if (hasVolumeMaximumsApi(controller)) {
                try {
                    if (controller instanceof FilerImpl) {
                        volMax = load7modeVolumeMaximumsApi((FilerImpl) controller);
                    } else {
                        volMax = loadCmodeVolumeMaximumsApi(controller);
                    }
                } catch (Exception e) {
                    LOGGER.warn("An error occurred while loading the volume maximums from the controller " +
                                        controller.getName() + ", attempting to fallback to XML.", e);
                }
            }
            if (volMax == null) {
                // Controller must not have supported getting the maximums through ZAPI, so just load them up from XML.
                volMax = loadVolumeMaximumsXml(controller);
            }
            // Final check. Still null? Just return defaults.
            if (volMax == null) {
                volMax = DEFAULT_MAXIMUMS;
            }
        }
        controller.setVolumeMaximums(volMax);
        return true;
    }

    private OntapVolumeMaximums getFromParentController(AbstractOntapController<? extends OntapController> controller) {
        OntapVolumeMaximums volumeMaximums = null;
        if (controller instanceof HasParentControllerImpl) {
            @SuppressWarnings("unchecked")
            AbstractOntapController<OntapController> parent = (AbstractOntapController)((HasParentControllerImpl)controller).getParentControllerImpl();
            volumeMaximums = parent.getVolumeMaximums();
        }
        return volumeMaximums;
    }

    /**
     * Returns the hardware model of the specified controller.
     */
    private String getModel(OntapController controller) {
        if (controller.getModel() != null) {
            return controller.getModel();
        } else if (controller instanceof DataSvm) {
            if (((DataSvm) controller).isDirectSvm()) {
                return DIRECT_VSERVER_MODEL;
            }
        }
        return null;
    }

    /**
     * Retruns true if the specified controller supports accessing the volume maximum information through a ZAPI.
     */
    private boolean hasVolumeMaximumsApi(OntapController controller) {
        boolean result = controller.getVersion().atLeastSN();
        if (controller instanceof DataSvm) {
            // Can't get at the node api if we're direct connect.
            result &= !((DataSvm) controller).isDirectSvm();
        }
        return result;
    }

    /**
     * Returns the release for the specified controller. This will be in the format: major#.minor#.
     */
    private String getRelease(OntapController controller) {
        return controller.getVersion().getVersionNumber()[0] + "." + controller.getVersion().getVersionNumber()[1];
    }

    /**
     * Loads the volume maximums for the specified controller from the static XML file volumemaximums.xml.
     *
     * @throws DataLoaderException If the volume maximums could not be loaded.
     */
    private OntapVolumeMaximums loadVolumeMaximumsXml(OntapController controller) throws DataLoaderException {
        OntapVolumeMaximums volMax = null;
        VolumeMaximum volumeMaximumsXml = loadVolumeMaximumsXml();
        String releaseString = getRelease(controller);
        String model = getModel(controller);
        if (releaseString != null && model != null) {
            for (OntapRelease release : volumeMaximumsXml.getOntapReleases()) {
                if (releaseString.equals(release.getRelease())) {
                    for (StorageController controllerMaximums : release.getStorageControllers()) {
                        if (controllerMaximums.getPlatform().equals(model)) {
                            volMax = controllerMaximums.toNtapVolumeMaximums();
                        }
                    }
                }
            }
        }
        return volMax;
    }

    /**
     * Loads the volume maximums from the static volumemaximums.xml file. This is used for older ONTAP versions in which
     * there is no ZAPI that provides this information.
     *
     * @return The VolumeMaximum object containing the volume maximums for all ONTAP releases and hardware models.
     * @throws DataLoaderException If the volume maximums could not be loaded.
     */
    private VolumeMaximum loadVolumeMaximumsXml() throws DataLoaderException {
        if (volumeMaximum == null) {
            synchronized (volumeMaximumLoadLock) {
                if (volumeMaximum == null) {
                    //load the xml file if needed.
                    try {
                        volumeMaximum = new VolumeMaximum();
                        InputStream volMaxStream = getClass().getResourceAsStream(VOLMAX_FILE_NAME);

                        if (volMaxStream != null) {
                            JAXBContext context = JAXBContext.newInstance(DATA_CLASSES);
                            Unmarshaller u = context.createUnmarshaller();
                            VolumeMaximum root = (VolumeMaximum) u.unmarshal(volMaxStream);
                            if (root.getOntapReleases() != null) {
                                volumeMaximum.setOntapReleases(root.getOntapReleases());
                            }
                        } else {
                            LOGGER.warn(
                                    "Warning: Unable to load volume maximums from the xml file " + VOLMAX_FILE_NAME +
                                            " which should be located in the manage offtap jar.");
                        }
                    } catch (JAXBException e) {
                        throw new DataLoaderException("Unable to unmarshall volume maximums xml file.", e);
                    }
                }
            }
        }
        return volumeMaximum;
    }

    /**
     * Loads the volume maximums for the specified 7-mode filer using a ZAPI.
     *
     * @param filer The 7-mode filer from which the volume maximums will be loaded.
     * @return The volume maximums for the filer
     */
    private OntapVolumeMaximums load7modeVolumeMaximumsApi(FilerImpl filer) throws Exception {
        NaElement input = new NaElement("system-get-info");
        NaElement output = filer.invokeZapi(input, ZapiCache.LONG_CACHE);
        NaElement systemInfo = output.getChildByName("system-info");
        if (systemInfo != null) {
            String maxVolumeSize = systemInfo.getChildContent("maximum-flexible-volume-size");
            OntapVolumeMaximumsImpl volumeMaximums = new OntapVolumeMaximumsImpl(DEFAULT_MAXIMUMS);
            volumeMaximums.setMax64BitDenseVolumeSize(Size.createB(Long.parseLong(maxVolumeSize)));
            return volumeMaximums;
        } else {
            throw new VolumeMaximumsException(
                    "Unable to load the maximum volume size for the controller " + filer.getName());
        }
    }

    /**
     * Loads the volume maximums for the specified svm using a ZAPI.
     *
     * @param svm The svm from which the volume maximums will be loaded.
     * @return The volume maximums for the svm
     */
    private OntapVolumeMaximums loadCmodeVolumeMaximumsApi(AbstractOntapController<? extends OntapController> svm) throws Exception {
        OntapConnection connection;
        if (svm instanceof HasParentControllerImpl) {
            @SuppressWarnings("unchecked")
            AbstractOntapController<? extends OntapController> parent = (AbstractOntapController)((HasParentControllerImpl)svm).getParentControllerImpl();
            connection = parent.getConnection();
        } else {
            connection = svm.getConnection();
        }
        NaElement input = new NaElement("system-node-get-iter");
        NaElement output = connection.invokeZapi(input, ZapiCache.LONG_CACHE);
        NaElement attributesList = output.getChildByName("attributes-list");
        if (attributesList != null) {
            @SuppressWarnings("unchecked")
            List<NaElement> attributes = attributesList.getChildren();
            for (NaElement nodeDetailsInfo : attributes) {
                String maxVolumeSize = nodeDetailsInfo.getChildContent("maximum-volume-size");
                if (StringUtils.isReal(maxVolumeSize)) {
                    OntapVolumeMaximumsImpl volumeMaximums = new OntapVolumeMaximumsImpl(DEFAULT_MAXIMUMS);
                    volumeMaximums.setMax64BitDenseVolumeSize(Size.createB(Long.parseLong(maxVolumeSize)));
                    return volumeMaximums;
                }
            }
        }
        throw new VolumeMaximumsException("Unable to load the maximum volume size for the controller " + svm.getName());
    }
}
