package com.netapp.offtap3.internal.ontap.exports;


import com.google.common.base.Objects;
import com.netapp.offtap3.ExportOperationException;
import com.netapp.offtap3.internal.AbstractRefreshable;
import com.netapp.offtap3.internal.ontap.controller.AbstractOntapDataController;
import com.netapp.offtap3.internal.ontap.support.ZapiCache;
import com.netapp.offtap3.internal.ontap.volume.OntapVolumeImpl;
import com.netapp.offtap3.ontap.ApiFailedException;
import com.netapp.offtap3.ontap.Export;
import com.netapp.offtap3.ontap.ExportSecurityRule;
import com.netapp.offtap3.ontap.OntapDataController;
import com.netapp.offtap3.ontap.OntapVolume;
import com.netapp.offtap3.ontap.NaElement;
import org.apache.log4j.Logger;

import java.util.ArrayList;
import java.util.HashSet;
import java.util.List;
import java.util.Set;

public class ExportImpl extends AbstractRefreshable implements Export {
    private static final Logger logger = Logger.getLogger(ExportImpl.class);

    public static final String ALL_HOSTS_7 = "all-hosts";
    public static final String ALL_HOSTS_C = "0.0.0.0/0";

    private final OntapVolumeImpl<? extends OntapVolume, ? extends OntapDataController, ? extends AbstractOntapDataController<? extends OntapDataController>> volume;
    private String path;
    private List<ExportSecurityRule> securityRules = new ArrayList<ExportSecurityRule>();

    public ExportImpl(OntapVolumeImpl<? extends OntapVolume, ? extends OntapDataController, ? extends AbstractOntapDataController<? extends OntapDataController>> volume) {
        super(volume);
        this.volume = volume;
    }

    @Override
    public String getPath() {
        return path;
    }

    public void setPath(String path) {
        this.path = path;
    }

    @Override
    public List<ExportSecurityRule> getSecurityRules() {
        return securityRules;
    }

    @Override
    public OntapVolume getVolume() {
        return volume.getProxy();
    }

    public void setSecurityRules(List<ExportSecurityRule> securityRules) {
        this.securityRules = securityRules;
    }

    public Set<String> getReadWriteHosts() {
        Set<String> allReadWriteHosts = new HashSet<>();
        for (ExportSecurityRule rule : securityRules) {
            allReadWriteHosts.addAll(rule.getReadWriteHosts());
        }
        return allReadWriteHosts;
    }

    public Set<String> getReadOnlyHosts() {
        Set<String> allReadOnlyHosts = new HashSet<>();
        for (ExportSecurityRule rule : securityRules) {
            allReadOnlyHosts.addAll(rule.getReadOnlyHosts());
        }
        return allReadOnlyHosts;
    }

    public Set<String> getRootAccessHosts() {
        Set<String> rootAccessHosts = new HashSet<>();
        for (ExportSecurityRule rule : securityRules) {
            rootAccessHosts.addAll(rule.getRootAccessHosts());
        }
        return rootAccessHosts;
    }

    @Override
    public boolean allowsClient(String clientIp, Permission permission) {
        AbstractOntapDataController controller = getController();
        logger.debug("isPathExported: checking " + path + " from " + controller.getName() + " to " + clientIp + " for " + permission.name());

        NaElement root = new NaElement("nfs-exportfs-check-permission");
        root.addNewChild("host", clientIp);
        root.addNewChild("pathname", path);
        root.addNewChild("permission", permission.zapiValue());

        try {
            NaElement result = controller.invokeZapi(root, ZapiCache.NO_CACHE);
            return Boolean.valueOf(result.getChildContent("is-permissible"));
        }
        catch (ApiFailedException e) {
            throw new ExportOperationException(controller + " could not check nfs export permission for " + this + " using client IP=" + clientIp, e);
        }
    }
    
    public boolean isReadWriteForAllHosts() {
        if (getReadWriteHosts() != null) {
            if (getReadWriteHosts().contains(ALL_HOSTS_7)) return true;
            if (getReadWriteHosts().contains(ALL_HOSTS_C)) return true;
        }
        return false;
    }

    @Override
    public void delete() {
        NaElement xi = new NaElement("nfs-exportfs-delete-rules");
        NaElement pathNames = new NaElement("pathnames");
        NaElement pathNameInfo = new NaElement("pathname-info");
        pathNameInfo.addNewChild("name", getPath());
        xi.addChildElem(pathNames);
        pathNames.addChildElem(pathNameInfo);
        xi.addNewChild("all-pathnames", "false");
        xi.addNewChild("persistent", "true");

        try {
            getController().invokeZapi(xi, ZapiCache.NO_CACHE);
            getController().getZapiCache().blastCacheValues("export");
        } catch (ApiFailedException e) {
            throw new ExportOperationException(getController() + " failed to delete rules for export: " + this, e);
        }

    }

    private AbstractOntapDataController getController() {
        return volume.getControllerImpl();
    }

    @Override
    public Export refresh() {
        return (Export) super.refresh();
    }

    public String toString() {
        return Objects.toStringHelper(this)
                .add("path", getPath())
                .add("rules", getSecurityRules())
                .toString();

    }
}
