package com.netapp.offtap3.internal.ontap.file;

import com.netapp.ndmp.NdmpException;
import com.netapp.offtap3.File;
import com.netapp.offtap3.FileCloneOperationException;
import com.netapp.offtap3.FileOperationException;
import com.netapp.offtap3.Lun;
import com.netapp.offtap3.StorageSystem;
import com.netapp.offtap3.internal.OfftapProxy;
import com.netapp.offtap3.internal.ontap.CModeCloneSpecImpl;
import com.netapp.offtap3.internal.ontap.MCCManager;
import com.netapp.offtap3.internal.ontap.controller.AbstractOntapDataController;
import com.netapp.offtap3.internal.ontap.controller.AdminSvmImpl;
import com.netapp.offtap3.internal.ontap.controller.DataSvmImpl;
import com.netapp.offtap3.internal.ontap.file.ndmp.NdmpCopyOperation;
import com.netapp.offtap3.internal.ontap.qos.QosPolicyGroupImpl;
import com.netapp.offtap3.internal.ontap.support.ZapiCache;
import com.netapp.offtap3.internal.ontap.utils.StatsCollector;
import com.netapp.offtap3.internal.ontap.volume.CModeVolumeImpl;
import com.netapp.offtap3.ontap.AdminSvm;
import com.netapp.offtap3.ontap.ApiFailedException;
import com.netapp.offtap3.ontap.CModeAggregate;
import com.netapp.offtap3.ontap.CModeCloneSpec;
import com.netapp.offtap3.ontap.CModeFile;
import com.netapp.offtap3.ontap.CModeLun;
import com.netapp.offtap3.ontap.CModeVolume;
import com.netapp.offtap3.ontap.CloneSpec;
import com.netapp.offtap3.ontap.DataSvm;
import com.netapp.offtap3.ontap.FileRange;
import com.netapp.offtap3.ontap.NaAuthenticationException;
import com.netapp.offtap3.ontap.NaElement;
import com.netapp.offtap3.ontap.NdmpCopyProcess;
import com.netapp.offtap3.ontap.OntapDataController;
import com.netapp.offtap3.ontap.OntapFile;
import com.netapp.offtap3.ontap.OntapVersion;
import com.netapp.offtap3.ontap.QosOperationException;
import com.netapp.offtap3.ontap.QosPolicyGroup;
import com.netapp.offtap3.ontap.utils.OntapCallable;
import com.netapp.offtap3.utils.StringUtils;
import com.netapp.offtap3.utils.Utils;

import org.apache.log4j.Logger;

import java.util.ArrayList;
import java.util.Collections;
import java.util.List;
import java.util.Set;
import java.util.concurrent.FutureTask;

public class CModeFileImpl extends
        OntapFileImpl<CModeFile, DataSvm, DataSvmImpl, CModeVolumeImpl, CModeVolume> implements CModeFile {

    private static final Logger LOGGER = Logger.getLogger(CModeFileImpl.class);
    private static final Set<StorageSystem> storageSystems = Collections.singleton(StorageSystem.datasvm);
    private static final long WAFL_BLOCK_SIZE = 4096L;
    private CModeFile proxy;

    private CModeFileImpl(DataSvmImpl controller, CModeVolumeImpl volume, String fullPath) {
        super(controller, volume, fullPath);
    }

    /**
     * Creates a <code>File</code> instance with the full path value.
     */
    public static CModeFile create(DataSvmImpl controller, CModeVolumeImpl volume, String path) {
        CModeFileImpl file = new CModeFileImpl(controller, volume, volume.toFullPath(path));
        CModeFile proxy = (CModeFile) OfftapProxy
                .create(new Class[]{CModeFile.class, InternalOntapFile.class}, FileProxy.class, file);
        file.setProxy(proxy);
        return proxy;
    }

    /**
     * Create a file relative to this directory
     *
     * @param path relative path
     */
    @Override
    public CModeFile createFile(String path) {
        return create(getControllerImpl(), getVolumeImpl(), concatPath(path));
    }

    @Override
    public CModeFile getParentDir() {
        return (CModeFile) super.getParentDir();
    }

    @Override
    public CModeLun createLun(String lunPath, String osType, boolean spaceReserve, String comment) {
        return (CModeLun) super.createLun(lunPath, osType, spaceReserve, comment);
    }

    @Override
    public List<CModeFile> listFiles() throws FileOperationException {
        String path = getFullPath();

        List<CModeFile> files = new ArrayList<CModeFile>();
        String nextTag = null;
        try {
            while (true) {
                NaElement xi = new NaElement("file-list-directory-iter");
                xi.addNewChild("path", path);
                if (getControllerImpl().getOfftapPreferences().getIteratorSize() > 0) {
                    xi.addNewChild("max-records", "" + getControllerImpl().getOfftapPreferences().getIteratorSize());
                }
                if (StringUtils.isReal(nextTag)) {
                    xi.addNewChild("tag", nextTag);
                }
                NaElement desiredAttributes = new NaElement("desired-attributes");
                NaElement fileInfo = new NaElement("file-info");
                fileInfo.addChildElem(new NaElement("name"));
                desiredAttributes.addChildElem(fileInfo);
                xi.addChildElem(desiredAttributes);
                NaElement xo = getControllerImpl().invokeZapi(xi, ZapiCache.NO_CACHE);

                NaElement attributesList = xo.getChildByName("attributes-list");
                if (attributesList == null) {
                    break;
                }
                List<NaElement> volumeInfo = attributesList.getChildren();
                for (NaElement currXML : volumeInfo) {
                    String filename = currXML.getChildContent("name");
                    if (!".".equals(filename) && !"..".equals(filename)) {
                        files.add(createFile(filename));
                    }
                }
                nextTag = xo.getChildContent("next-tag");
                if (!StringUtils.isReal(nextTag) || files.size() > 100000) {
                    // deadman switch to make sure we don't get stuck
                    break;
                }
            }
        } catch (ApiFailedException e) {
            LOGGER.error("unable to list directory contents", e);
            throw new FileOperationException("Unable to list file contents of " + path + ": " + e.getMessage(), e);
        }
        return files;
    }

    @Override
    public void setFileReservation(boolean on) throws FileOperationException {
        String path = getFullPath();
        LOGGER.debug("setFileReservation with path: " + path);

        NaElement xo;

        NaElement xi = new NaElement("file-set-space-reservation-info");
        xi.addNewChild("path", path);
        xi.addNewChild("is-overwrite-enabled", on ? "true" : "false");
        xi.addNewChild("is-fill-enabled", on ? "true" : "false");

        try {
            xo = getControllerImpl().invokeZapi(xi, ZapiCache.NO_CACHE);
            ZapiCache.getInstance().blastCacheValues("file-");
            LOGGER.info("Reserved space for file '" + path + "' :" + xo.getAttr("status") + ".");

        } catch (ApiFailedException ex) {
            LOGGER.error("Unable to reserve file. - " + ex.getMessage(), ex);
            throw new FileOperationException("Unable to reserve file. - " + ex.getMessage(), ex);
        }
    }
    
    public boolean getFileReservation()  throws FileOperationException {
        String path = getFullPath();
        NaElement xi = new NaElement("file-get-space-reservation-info");
        xi.addNewChild("path", path);
        try {
        	NaElement xo = getControllerImpl().invokeZapi(xi, ZapiCache.LONG_CACHE);
        	return "true".equals(xo.getChildContent("is-fill-enabled"));
        } catch (ApiFailedException ex) {
        	LOGGER.error("Unable to get file reservation. - " + ex.getMessage(), ex);
            throw new FileOperationException("Unable to reserve file. - " + ex.getMessage(), ex);
        }
    }

    /**
     * Invokes file-hole-punch-iter for c-mode as needed to process the complete Map of holes.
     * The zapidocs do not indicate the limit for how many holes can appear in a request.  Based on discussions with
     * WAFL team, 32 is safe.
     */
    @Override
    public void punchHoles(List<FileRange> holes) throws FileOperationException {
        String filePath = getFullPath();
        final int ITER_BATCH_LIMIT = 32;

        List<FileRange> zapiHoles = asMaxHoleList(holes, FileRange.DEFAULT_MAX_HOLE_SIZE);
        LOGGER.debug("Entering c-mode punchHoles for " + filePath + " with " + zapiHoles.size() + " holes to punch.");

        NaElement xi = new NaElement("file-punch-hole-iter");
        xi.addNewChild("path", filePath);

        NaElement xa = new NaElement("attributes-list");
        xi.addChildElem(xa);

        int holesPunched = 0;
        for (FileRange zapiHole : zapiHoles) {
            NaElement holeRange = new NaElement("hole-range-info");

            holeRange.addNewChild("hole-start", String.valueOf(zapiHole.getOffset()));
            holeRange.addNewChild("hole-size", String.valueOf(zapiHole.getLength()));

            xa.addChildElem(holeRange);
            holesPunched++;

            if ((holesPunched % ITER_BATCH_LIMIT) == 0) {
                filePunchHoleIter(xi);

                // we just punched 32 holes, so lets rebuild the zapi
                xi = new NaElement("file-punch-hole-iter");
                xi.addNewChild("path", filePath);

                xa = new NaElement("attributes-list");
                xi.addChildElem(xa);
            }
        }

        // punch any stragglers here
        if ((holesPunched % ITER_BATCH_LIMIT) != 0) {
            filePunchHoleIter(xi);
        }

        LOGGER.debug("Exiting c-mode punchHoles");
    }

    /**
     * Helper method to process a batch of holes.  In c-mode, the file-hole-punch-iter ZAPI <b>may</b> respond saying
     * that it only processed x out of n holes.  In that case, the ZAPI is meant to be re-invoked, passing in a tag
     * parameter that indicates what hole it is currently working on.
     * <p/>
     * It gets worse.  BURT 583171 means that the server may respond with the same hole over and over, forever.
     * In that case, it is up to this method to detect the situation and then remove that hole from the request.
     * The "detection algorithm" is:  did the server give back the same hole 4 times in a row, even while pausing
     * for one second between repetitions.
     *
     * @param zapiXml the XML for invoking file-hole-punch-iter
     */
    private void filePunchHoleIter(NaElement zapiXml) throws FileOperationException {
        String nextTag;
        String lastNextTag = null;

        // gross hack to deal with ONTAP BURT 583171 -- basically, give the ZAPI a few chances to process
        // the hole, but don't fall into an infinite loop!
        int repeatedNextTag = 0;
        final int MAX_RETRY = 4;
        try {
            do {
                NaElement xo = getControllerImpl().invokeZapi(zapiXml, ZapiCache.NO_CACHE);
                nextTag = xo.getChildContent("next-tag");
                // ZAPI response indicates that the server has not finished with this request
                if (nextTag != null) {
                    NaElement inputTag = zapiXml.getChildByName("tag");
                    // check if it says it is still processing the same hole
                    if (nextTag.equals(lastNextTag)) {
                        if (++repeatedNextTag >= MAX_RETRY) {
                            // at this point, the same ZAPI has been invoked MAX_RETRY times with no change in output.
                            // assume BURT 583171 and remove that hole from the request
                            int nextTagIndex = Integer.parseInt(nextTag);
                            LOGGER.warn("ZAPI returned same value of next-tag " + MAX_RETRY +
                                                " times in a row.  Removing hole at index " + nextTag +
                                                " from the list.");
                            List<NaElement> list = zapiXml.getChildByName("attributes-list").getChildren();
                            NaElement problem = list.remove(Integer.parseInt(nextTag));
                            // because a hole was removed from the list, nextTagIndex no longer points at the "problem" hole, but
                            // also check if it has moved past the end of the list
                            if (nextTagIndex >= list.size()) {
                                LOGGER.warn("Since last hole was removed, no more holes to punch in this batch");
                                break;
                            } else {
                                LOGGER.warn("ZAPI will be re-executed without " + problem);
                            }
                        } else {
                            // since the same tag was returned, it might just be taking a long time to free the blocks
                            // wait for a second to give the server a chance to process it
                            LOGGER.warn(
                                    "ZAPI response indicated server is still processing the same hole #" + lastNextTag +
                                            ".  Pausing for 1 second.");
                            Thread.sleep(1000);
                        }
                    } else {
                        LOGGER.debug("next-tag result changed from " + lastNextTag + " to " + nextTag);
                        repeatedNextTag = 0;
                        lastNextTag = nextTag;
                    }

                    // fix up the tag parameter for our next request (create or update as needed)
                    if (inputTag == null) {
                        zapiXml.addNewChild("tag", nextTag);
                    } else {
                        inputTag.setContent(nextTag);
                    }
                } else {
                    LOGGER.debug("Batch complete for " + zapiXml.getChildContent("path"));
                }
            } while (nextTag != null);
        } catch (InterruptedException e) {
            String filePath = zapiXml.getChildContent("path");
            LOGGER.debug("punch iter failed " + filePath, e);
            throw new FileOperationException("punch iter failed " + filePath + " with error: " + e.getMessage(), e);
        } catch (ApiFailedException e) {
            String filePath = zapiXml.getChildContent("path");
            LOGGER.debug("punch iter failed " + filePath, e);
            throw new FileOperationException("punch iter failed " + filePath + " with error: " + e.getMessage(), e);
        }

    }

    /**
     * Uses the c-mode vserver ZAPI file-rename-file to "move" this file within a volume.
     *
     * @param destination full path for new location
     */
    @Override
    public void move(File destination) throws FileOperationException {
        LOGGER.debug("moveFile: " + this + " " + destination);

        try {
            NaElement xi = new NaElement("file-rename-file");
            xi.addNewChild("from-path", getFullPath());
            xi.addNewChild("to-path", destination.getFullPath());

            NaElement xo = getControllerImpl().invokeZapi(xi, ZapiCache.NO_CACHE);
            getControllerImpl().getZapiCache().blastCacheValues("file-");

            String output = xo.getAttr("status");
            if ("passed".equals(output)) {
                LOGGER.debug("file moved.");
            } else {
                //something bad may have happened so we want to add
                //information to the logs here about which file is
                //still around.  If we don't find the source and we
                //find the destination we know we're good.

                LOGGER.debug("moveFile is checking for source file...");

                //we wrap this so its exception doesn't isn't confused with
                //the exception file-rename-file might throw
                //this makes the method retryable (from the application)
                if (exists()) {
                    throw new FileOperationException(
                            "moveFile - Move of " + this + " to " + destination + " failed. (Source still exists.) - " +
                                    output);
                }

                LOGGER.debug("moveFile is checking for destination file...");

                if (!destination.exists()) {
                    throw new FileOperationException("moveFile - Move of " + this + " to " + destination +
                                                             " failed. (Destination does not exist.) - " + output);
                }
            }

        } catch (ApiFailedException e) {
            LOGGER.error("Move of " + this + " to " + destination + " failed.", e);
            throw new FileOperationException("Move of " + this + " to " + destination + " failed: " + e.getMessage(),
                                             e);
        }
    }

    @Override
    public NdmpCopyProcess ndmpCopy(OntapFile destFile) {
        @SuppressWarnings(
                "unchecked") InternalOntapFile<OntapDataController, AbstractOntapDataController<OntapDataController>> internalDestFile = (InternalOntapFile) destFile;
        return doNdmpCopy(internalDestFile, false, false);
    }

    @Override
    public NdmpCopyProcess asyncNdmpCopy(OntapFile destFile) {
        @SuppressWarnings(
                "unchecked") InternalOntapFile<OntapDataController, AbstractOntapDataController<OntapDataController>> internalDestFile = (InternalOntapFile) destFile;
        return doNdmpCopy(internalDestFile, true, false);
    }
    
    @Override
    public void asyncNdmpCopy(OntapFile destFile, boolean skipPollingOnNdmpProgress) {
        @SuppressWarnings(
                "unchecked") InternalOntapFile<OntapDataController, AbstractOntapDataController<OntapDataController>> internalDestFile = (InternalOntapFile) destFile;
        doNdmpCopy(internalDestFile, true, skipPollingOnNdmpProgress);
    }

    private NdmpCopyProcess doNdmpCopy(final InternalOntapFile<OntapDataController, AbstractOntapDataController<OntapDataController>> destFile, final boolean async, boolean skipPollingOnNdmpProgress) {
        final DataSvmImpl srcSvm = getControllerImpl();
        InternalOntapFile finalDestFile = destFile;
        final DataSvm destSvm = (DataSvm) finalDestFile.getController();
        LOGGER.debug(srcSvm + " to ndmpcopy " + this + " to " + destSvm + " and path " + finalDestFile + " starting.");

        srcSvm.enableNdmp();
        destSvm.enableNdmp();

        AdminSvm srcAdmin = srcSvm.getParentControllerImpl();
        AdminSvm destAdmin = destSvm.getParentController();

        if (srcAdmin == null || destAdmin == null) {
            throw new UnsupportedOperationException("NDMP copy is not supported for direct SVM");
        }

        srcAdmin.enableNdmp();
        destAdmin.enableNdmp();

        final String srcAdminUserPassword = srcAdmin.getNdmpPassword();

        if (srcAdminUserPassword == null) {
            LOGGER.error("Could not get source Admin password");
            throw new FileOperationException("Could not get source Admin password");
        }

        final String dstAdminUserPassword = destAdmin.getNdmpPassword();
        if (dstAdminUserPassword == null) {
            LOGGER.error("Could not get destination Admin password");
            throw new FileOperationException("Could not get destination Admin password");
        }   
        
        
        finalDestFile = (InternalOntapFile) finalDestFile.getParentDir().createFile("tempDir/");
        final String destPath = finalDestFile.getNdmpPath();

        LOGGER.info("ndmpCopy has determined NDMP destination path is: " + destPath);
        final NdmpCopyOperation ndmpCopyOperation = new NdmpCopyOperation.Builder()
                .withSourceIpAddress(srcAdmin.getConnection().getIpAddress())
                .withSourceLogin(srcAdmin.getConnection().getUserName()).withSourcePassword(srcAdminUserPassword)
                .withSourcePathInfo(getNdmpPath()).withDestinationIpAddress(destAdmin.getConnection().getIpAddress())
                .withDestinationLogin(destAdmin.getConnection().getUserName())
                .withDestinationPassword(dstAdminUserPassword).withDestinationPathInfo(destPath).build();

        FutureTask<Integer> ndmpCopyTask = new FutureTask<Integer>(new OntapCallable<Integer>() {
            @Override
            public Integer callLocal() throws Exception {
                long startTime = System.currentTimeMillis();

                try {
                	LOGGER.info("doNdmpCopy: Going to start ndmpcopy");
                    ndmpCopyOperation.startCopyOperation();
                } catch (NdmpException e) {
                    LOGGER.error(srcSvm + " failed to ndmpcopy " + this + " to " + destSvm + " with path " + destPath +
                                         " failed: " + e.getMessage(), e);
                    throw new FileOperationException(
                            srcSvm + " failed to ndmpcopy " + this + " to " + destSvm + " with path " + destPath +
                                    " failed: " + e.getMessage(), e);
                }

                long duration = System.currentTimeMillis() - startTime;
                StatsCollector.getInstance().addCounter("ndmp-copy", duration);
                return 0;
            }
        });

		NdmpCopyProcessImpl copyProcess = null;
		String fileName = getProxy().getName();
		InternalOntapFile tmpDestFile = (InternalOntapFile) finalDestFile.getParentDir().createFile("" + fileName);
		copyProcess = new NdmpCopyProcessImpl(ndmpCopyTask, srcSvm, getProxy(), destFile, tmpDestFile);
		copyProcess.start(async,skipPollingOnNdmpProgress);
		LOGGER.info(srcSvm + " ndmpcopied to " + destSvm + " at path " + finalDestFile + " finished.");
		return copyProcess;
    }

    @Override
    public void sisClone(OntapFile newLocation) throws FileOperationException {
        sisClone(getCloneSpecBuilder(newLocation).build());
    }

    @Override
    public CModeCloneSpecImpl.Builder getCloneSpecBuilder(File destinationFile) {
        return new CModeCloneSpecImpl.Builder(destinationFile);
    }


    @Override
    public void sisClone(CloneSpec spec) {
    	CModeCloneSpec cloneSpec = (CModeCloneSpec) spec;
    	InternalOntapFile destFile = (InternalOntapFile) cloneSpec.getDestination();
    	boolean fileThoughtToExist = cloneSpec.fileCertainToExist();
    	try {
	    	try {
	    		sisCloneInternal(spec, fileThoughtToExist);
	    	} catch (ApiFailedException e) {
	    		// 3449 * @errdef      ESIS_CLONE_CLONE_EXIST 14952
	    		// 3450 * @desc        Source file exists
	    		if (14952 == e.getErrno()) {
	    			fileThoughtToExist = true;
	    			sisCloneInternal(spec, fileThoughtToExist);
	    		} else {
	    			throw e;
	    		}	    		
	    	}
        } catch (ApiFailedException e) {
            LOGGER.error(this + " failed to be sis-cloned to " + destFile, e);
            throw new FileCloneOperationException(this.getFullPath() + " failed to be sis-cloned to " + destFile.getFullPath(), e);
        }
    }
    
    public void sisCloneInternal(CloneSpec spec, boolean fileExists) throws ApiFailedException {
        CModeCloneSpec cloneSpec = (CModeCloneSpec) spec;
        InternalOntapFile destFile = (InternalOntapFile) cloneSpec.getDestination();
        OntapVersion version = getController().getVersion();
        LOGGER.debug("Entered clone with srcPath=" + getFullPath() + " destPath=" + destFile.getFullPath());

        if (!getController().getId().equals(destFile.getController().getId())) {
            throw new IllegalArgumentException(
                    "sis-clone only works within the same SVM." + " Instead saw source controller " + getController() +
                            " and destination controller " + destFile.getController());
        }

        String destVolumeName = destFile.getVolume().getName();
        String sourceVolumeName = getVolumeImpl().getName();
        if (!sourceVolumeName.equals(destVolumeName)) {
            throw new IllegalArgumentException(
                    "Sis-clone only works within a volume - instead saw source volume " + sourceVolumeName +
                            " and destination volume " + destVolumeName);
        }

        String service = "clone-create";
        NaElement xi = new NaElement(service);
        xi.addNewChild("volume", destVolumeName);
        xi.addNewChild("source-path", getVolumeRelativePath());
        xi.addNewChild("destination-path", destFile.getVolumeRelativePath());

        // we used to check here with file.exists().  But Ontap appears to cache this information for some period
        // of time, so that a call to file-get-file BEFORE you do the clone-start means that you got usually 30 seconds
        // of polling on a failed file-get-file in the clone-validation phase.  So, instead, we're gonna assume that the file
        // didn't exist unless someone said it did, and we'll just catch the failure in the immediately calling method and retry.
        // And while it is tempting to set it to true always, that in fact doesn't work.  If you set it to true and the file isn't 
        // there, you get an error.  If you omit it and the file is there, you get an error.  So, you have to set it precisely right
        // or you get unhappiness
        if (fileExists && version.atLeastSN()) {
            xi.addNewChild("destination-exists", "true");
        }

        if (version.atLeastFS()) {
            addTokenIfAppropriate(xi, cloneSpec);
        }

        for (int i = 0; i < getControllerImpl().getOfftapPreferences().getSiscloneRetryTimesOnEbusy(); i++) {
            try {
                getControllerImpl().invokeZapi(xi, ZapiCache.NO_CACHE);
                getControllerImpl().getZapiCache().blastCacheValues("file-", "lun-");
                break;
            } catch (ApiFailedException e) {
                if (e.getMessage().trim().contains("Device busy") || e.getMessage().trim().contains("(errno=16)")) {
                    /*
                     * tdeepak - Ontap expects offtap tools to retry in some cases.
                     * 1. First clone operation on the Volume(initializing the clone metafiles).
					 * 2. Hierarchy.
					 * 3. Ondemand operation is in progress on the source file or fence is set..
		             * 4. Clone + Iron..
					 */
                    StatsCollector.getInstance().addCounter("sisclone-retry-EBUSY");
                    Utils.sleep(getControllerImpl().getOfftapPreferences().getSiscloneIntervalBetweenRetryInSecs() *
                                        1000);
                } else if (e.getMessage().trim().contains("Requested space exceeds") ||
                        e.getMessage().trim().contains("(errno=14974)")) {
                    // Retry on BackPressure
                    StatsCollector.getInstance().addCounter("sisclone-retry-BackPressure");
                    Utils.sleep(getControllerImpl().getOfftapPreferences().getSiscloneIntervalBetweenRetryInSecs() *
                                        1000);
                } else {
                    throw e;
                }
            }
        }
        validateCloneReallyAppearedIfNecessary(destFile);

          /*tdeepak: SN does support clone autodelete, however another Zapi(unlike clone-create in FS) clone-autodelete needs to be used to set/unset it.
           *  SN : Autodelete is on by default for lun clones. Autodelete is not supported for file clones
           *  FS : Autodelete is off by default for lun/file clones
           */

        Lun sourceLun = getController().getLun(getFullPath());
        if (getController().getVersion().isSNDotAnything() && (sourceLun != null)) {
            //need to turn off autodelete for SN lun clones
            disableCloneAutodeleteForSN(destFile);
        }
    }

    private void addTokenIfAppropriate(NaElement input, CModeCloneSpec cloneSpec) throws ApiFailedException {
        boolean tokenExpired = false;
        if (cloneSpec.autoDelete()) {
            input.addNewChild("autodelete", "true");
        }
        if (cloneSpec.getToken() != null) {
            //check if passed in token is valid
            CModeVolume volume = getController().getVolume(getVolumeImpl().getName());
            String nodeName = volume.getContainingAggregate().getCurrentNodeName();
            long remainingTime = getSisCloneTokenRemainingTimeInSecs(cloneSpec.getToken(), nodeName);
            if (remainingTime > 0) {
                input.addNewChild("token-uuid", cloneSpec.getToken());
            } else {
                StatsCollector.getInstance().addCounter("sisclone-expired-split-load-token");
                LOGGER.error("clone:  sisClone split load token with id:" + cloneSpec.getToken() +
                                     " has expired!, Skipping using the token");
                tokenExpired = true;
            }
        }
        if (cloneSpec.createToken() || tokenExpired) {
            // acquire a split load token before attempting the clone
            // find file/lun size
            String tokenId = createSisCloneToken(cloneSpec.getTokenRetryInterval(), cloneSpec.getTokenRetryCount());
            if (tokenId != null) {
                input.addNewChild("token-uuid", tokenId);
            } else {
                LOGGER.error("clone:  Unable to create sisClone split load token, Skipping using the token");
            }
        }
    }

    public boolean isLun() {
        return getController().getLun(getFullPath()) != null;
    }

    @Override
    public long getSisCloneTokenRemainingTimeInSecs(String tokenId, String nodeName) throws FileOperationException {
        long remainingTime = 0L;
        AdminSvmImpl adminSvm = getControllerImpl().getParentControllerImpl();
        if (adminSvm == null) {
            throw new FileOperationException(
                    "The sis clone token remaining time cannot be queried through a directly connected SVM.");
        }

        try {
            String service = "clone-token-get";
            NaElement xi = new NaElement(service);
            xi.addNewChild("node", nodeName);
            xi.addNewChild("token-uuid", tokenId);

            NaElement xo = adminSvm.invokeZapi(xi, ZapiCache.NO_CACHE);
            NaElement cloneTokenInfo = xo.getChildByName("clone-token-info");
            if (cloneTokenInfo != null) {
                remainingTime = cloneTokenInfo.getChildLongValue("reservation-duration-remaining", 0L);
            }
        } catch (ApiFailedException e) {
            if (e.getMessage().trim().equals("FlexClone not licensed (errno=14616)")) {
                throw new FileCloneOperationException("FlexClone not licensed");
            } else if (e.getMessage().trim().contains("Reason: entry doesn't exist. (errno=13001)")) {
                //tdeepak: tokens expire in 60 secs, don't throw exception here.
                //Old Ontap Behavior pre 7/10/2014
                remainingTime = 0L;
            } else if(e.getMessage().trim().contains("Token expired or not found")) {
                //tdeepak: tokens expire in 60 secs, don't throw exception here
                //New Ontap Behavior post 7/10/2014
                remainingTime = 0L;
            } else {
                String message = "An error occurred while getting the time remaining for a clone token on Admin SVM " +
                        adminSvm.getName() + " for node " + nodeName;
                LOGGER.error(message, e);
                throw new FileCloneOperationException(message, e);
            }
        }
        return remainingTime;
    }

    public String createSisCloneToken(int tokenRetryIntervalInSecs, int tokenRetryCount) throws FileOperationException {
        String tokenId = null;

        if (tokenRetryIntervalInSecs <= 0) {
            tokenRetryIntervalInSecs = 60;
        }
        if (tokenRetryCount < 0) {
            tokenRetryCount = 0;
        }

        
        CModeAggregate aggr = getVolume().getContainingAggregate();
        // figure out the node the volume lives on.
        
        if(getControllerImpl().isDirectSvm()){
            throw new UnsupportedOperationException(
                    getControllerImpl() + " can not create a sisclone-token.  Requires admin svm access");
        }
        
        AdminSvmImpl adminSvm = getControllerImpl().getParentControllerImpl();  
        CModeAggregate aggrFromAdminVserver = adminSvm.getAggregates().get(aggr.getName()); 
        if(aggrFromAdminVserver!=null){
            aggr = aggrFromAdminVserver;  
        }             
       
        String nodeName = aggr.getCurrentNodeName();
        LOGGER.debug("createSisCloneToken: aggr=" + aggr.getName() +" nodeName="+nodeName);

        if (Utils.isEmpty(nodeName)) {
            throw new FileOperationException(
                    "Unable to determine current node value for volume " + getVolume().getName() + " on aggregate " +
                            aggr);
        }

        NaElement xi = new NaElement("clone-token-create");
        xi.addNewChild("node", nodeName);
        xi.addNewChild("reserve-size", String.valueOf(getReserveSizeForCloneToken()));
        

        // clone-token-create ZAPI must run against the cluster admin
        try {
            /*
             * tdeepak: We expect a very very large % of sis-clones to be
			 * non-throttled, so we try aquiring the token without checking the
			 * split load(saving us a Zapi call). However if we hit max split load , we enter retry
			 * logic that checks for split load available
			 */
            NaElement xo = adminSvm.invokeZapi(xi, ZapiCache.NO_CACHE);
            tokenId = xo.getChildContent("token-uuid");
            StatsCollector.getInstance().addCounter("sisclone-created-split-load-token");
        } catch (ApiFailedException e) {
            if (e.getMessage().trim().contains("Reached maximum split load allowed on the node")) {
                StatsCollector.getInstance().addCounter("sisclone-split-load-hit-maximum");
                Utils.sleep(tokenRetryCount * 1000);
                // enter retry logic
                try {
                    tokenId = getSisCloneTokenRetry(nodeName, tokenRetryIntervalInSecs, tokenRetryCount);
                } catch (Exception e2) {
                    handleSisCloneException(e2, adminSvm);
                }
			} else if (e.getMessage().trim().contains("Invalid value specified for \"node\" element")) {
				try {
					// refresh aggr info, this might be a MCC switchover burt:866207
					nodeName = aggr.refresh().getCurrentNodeName();
					tokenId = getSisCloneTokenRetry(nodeName, tokenRetryIntervalInSecs, tokenRetryCount);
				} catch (Exception e2) {
					handleSisCloneException(e2, adminSvm);
				}
			} else {
                handleSisCloneException(e, adminSvm);
            }
        } catch (Exception e) {
            handleSisCloneException(e, adminSvm);
        }

        return tokenId;
    }

    /**
     * tdeepak 3/19/2014 see burt:809278 Ontap now expects us to request for 2X as the reserve size where X = size of file to be cloned
     * tdeepak 4/18/2014 see burt:817535 We need to use the bytes used by first finding the blocks used and then converting that to bytes
     */
    private long getReserveSizeForCloneToken() {
        long blocksUsed = (long) Math.ceil(getProxy().getFileSize().toBytes() / (double) WAFL_BLOCK_SIZE);
        return 2L * blocksUsed * WAFL_BLOCK_SIZE;
    }

    private void handleSisCloneException(Exception e, AdminSvm adminSvm) {
        if (e instanceof ApiFailedException) {
            if (e.getMessage().trim().equals("FlexClone not licensed (errno=14616)")) {
                throw new FileCloneOperationException("FlexClone not licensed");
            } else {
                throw new FileCloneOperationException(
                        "An error occurred while creating the cloning token for file " + getFullPath(), e);
            }
        } else {
            LOGGER.error("createSisCloneToken:  Exception occured when trying to create split load token " +
                                 adminSvm.getName(), e);
            throw new FileCloneOperationException(
                    "createSisCloneToken:  Exception occured when trying to create split load token " +
                            adminSvm.getName(), e);
        }
    }

    private String getSisCloneTokenRetry(String nodeName, int tokenRetryIntervalInSecs, int tokenRetryCount) throws ApiFailedException, NaAuthenticationException {
        String tokenId = null;
        long reserveSize = getReserveSizeForCloneToken();
        NaElement xi = new NaElement("clone-token-create");
        xi.addNewChild("node", nodeName);
        xi.addNewChild("reserve-size", String.valueOf(reserveSize));

        for (int i = 0; i < tokenRetryCount; i++) {
            try {
                StatsCollector.getInstance().addCounter("sisclone-split-load-retry-count");
                if (isCloneTokenAvailable(nodeName, reserveSize)) {
                    NaElement xo = getControllerImpl().getParentControllerImpl().invokeZapi(xi, ZapiCache.NO_CACHE);
                    tokenId = xo.getChildContent("token-uuid");
                    break;
                } else {
                    //sleep for tokenRetryIntervalInSecs
                    StatsCollector.getInstance().addCounter("sisclone-split-load-retry-wait");
                    Utils.sleep(tokenRetryIntervalInSecs * 1000);
                }
            } catch (ApiFailedException e) {
                if (e.getMessage().trim().contains("Reason: Reached maximum split load allowed on the node.")) {
                    /* tdeepak: should'nt be here since we just checked the split load before requesting a token,
                     however if this is a freak case try again if retry count is less than tokenRetryCount*/
                    StatsCollector.getInstance().addCounter("sisclone-split-load-hit-maximum");
                    tokenId = null;
                    Utils.sleep(tokenRetryIntervalInSecs * 1000);
                } else {
                    throw e;
                }
            }
        }
        return tokenId;
    }

    @Override
    public boolean isCloneTokenAvailable(String nodeName, long requestedSize) throws FileOperationException {
        long availableSplitLoad = 0L;

        NaElement xi = new NaElement("clone-split-load-get-iter");
        NaElement query = new NaElement("query");
        xi.addChildElem(query);
        NaElement cloneSplitLoadInfo = new NaElement("clone-split-load-info");
        query.addChildElem(cloneSplitLoadInfo);
        cloneSplitLoadInfo.addNewChild("node", nodeName);

        try {
            NaElement xo = getControllerImpl().getParentControllerImpl().invokeZapi(xi, ZapiCache.NO_CACHE);
            NaElement attributes = xo.getChildByName("attributes-list");
            if (attributes != null) {
                NaElement cloneSplitLoadInfoResponse = attributes.getChildByName("clone-split-load-info");
                if (cloneSplitLoadInfoResponse != null) {
                    availableSplitLoad = cloneSplitLoadInfoResponse.getChildLongValue("available-split-load", 0L);
                }
            }
            return requestedSize <= availableSplitLoad;
        } catch (ApiFailedException e) {
            throw new FileOperationException("Error while checking if a cloning token is available: " + e.getMessage(),
                                             e);
        }
    }

    private void validateCloneReallyAppearedIfNecessary(File destFile) {
        if (getControllerImpl().getOfftapPreferences().getWantsCloneValidation()) {
            // dsarisky 7/12/12.  Unfortunately this will tend to return before the clone actually results in a file that
            // you can use.  So we need to poll on the silly thing until it reports that it exists.
            long startTime = System.currentTimeMillis();
            while (!destFile.exists() && (System.currentTimeMillis() < startTime + 120000)) {
                ZapiCache.getInstance().blastCacheValues("file-");
                ZapiCache.getInstance().blastCacheValues("lun-");

                StatsCollector.getInstance()
                              .addCounter("clone-validation-time", System.currentTimeMillis() - startTime);
                Utils.sleep(2000);
            }
        }
    }

    private void disableCloneAutodeleteForSN(InternalOntapFile destFile) throws ApiFailedException {
           /*
            *  this Zapi will fail if used for file clones in SN.
   		 *  "Clone autodelete enable failed. Reason: File type is not vdisk."
   		 */
        String service = "clone-autodelete";
        NaElement xi = new NaElement(service);
        xi.addNewChild("enable", "false");
        xi.addNewChild("clone-path", destFile.getVolumeRelativePath());
        xi.addNewChild("volume", destFile.getVolume().getName());
        getControllerImpl().invokeZapi(xi, ZapiCache.NO_CACHE);
    }


    @Override
    public CModeFile getProxy() {
        return proxy;
    }

    private void setProxy(CModeFile proxy) {
        this.proxy = proxy;
    }

    @Override
    public Set<StorageSystem> getStorageSystems() {
        return storageSystems;
    }

    @Override
    public void assignQos(QosPolicyGroup policyGroup) {
        DataSvmImpl svm = getControllerImpl();
        if (svm.isDirectSvm()) {
            throw new UnsupportedOperationException(
                    svm + " can not assign QoS policy groups.  Requires admin svm access");
        }

        if (!svm.getVersion().atLeastSN()) {
            throw new UnsupportedOperationException(svm + " does not support QoS policy groups.  Version is " +
                                                            svm.getVersion() + " and must be 8.2 or later.");
        }

        LOGGER.debug("Enter assignQosPolicyToFile with policy: " + policyGroup + " and volume: " + getVolumeImpl() +
                             " and path: " + getFullPath());

        NaElement xi = new NaElement("file-assign-qos");
        xi.addNewChild("qos-policy-group-name", policyGroup.getName());

        xi.addNewChild("vserver", MCCManager.getInstance().getVServerNameByUuid(svm));
        xi.addNewChild("volume", getVolumeImpl().getName());

        // when the file parameter is used for QoS, it is *RELATIVE* to the volume.
        xi.addNewChild("file", getVolumeRelativePath());

        try {
            NaElement xo = svm.getParentControllerImpl().invokeZapi(xi, ZapiCache.NO_CACHE);
            LOGGER.info(svm + " assigned QoS policy '" + policyGroup + "' to file '" + getFullPath() +
                                "' completed with status: " + xo.getAttr("status"));
        } catch (ApiFailedException e) {
            LOGGER.error(svm + " failed to assign qos policy: '" +
                                 policyGroup + "' to file '" + getFullPath() + "' due to: " + e, e);
            throw new QosOperationException(svm + " failed to assign qos policy: '" +
                                                    policyGroup + "' to file '" + getFullPath() + "' due to: " + e, e);
        }
    }

    @Override
    public void removeQos() {
        assignQos(QosPolicyGroupImpl.NONE);
    }

    @Override
    public String getOptimalInterface() throws FileOperationException {
        if (getController().getVersion().atLeastSN()) {
            String service = "file-get-optimal-access-address";
            NaElement xi = new NaElement(service);
            xi.addNewChild("path", getFullPath());
            try {
                NaElement xo = getControllerImpl().invokeZapi(xi, ZapiCache.SHORT_CACHE);
                return xo.getChildContent("ip-address");
            } catch (ApiFailedException e) {
                throw new FileOperationException(
                        "Get optimal access address of " + getFullPath() + " failed:" + e.getMessage(), e);
            }
        } else {
            return "";
        }
    }

    @Override
    public CModeFile refresh() {
        return (CModeFile) super.refresh();
    }
}
