package com.netapp.offtap3.internal.ontap.file;

import com.netapp.offtap3.FileResource;
import com.netapp.offtap3.Refreshable;
import com.netapp.offtap3.Resource;
import com.netapp.offtap3.Size;
import com.netapp.offtap3.internal.AbstractRefreshable;
import com.netapp.offtap3.internal.HasProxy;
import com.netapp.offtap3.utils.StringUtils;
import com.netapp.offtap3.File;
import com.netapp.offtap3.FileOperationException;

import java.util.Date;
import java.util.List;

public abstract class FileImpl<I extends File> extends AbstractRefreshable implements File, HasProxy<I> {

    private final String fullPath;

    private int accessedTimestamp;
    private String aclType;
    private int changedTimestamp;
    private int creationTimestamp;
    private Size fileSize;
    private String fileType;
    private int groupID;
    private int hardLinksCount;
    private int inodeNumber;
    private int modifiedTimestamp;
    private int ownerID;
    private int perm;
    private Size used;

    private boolean exists;

    protected FileImpl(String path, HasProxy loader) {
        super(loader);
        this.fullPath = path;
    }

    /**
     * Returns the full path of the file
     *
     * @return the full path of the file
     */
    @Override
    public String getFullPath() {
        return fullPath;
    }

    @Override
    public File getParentDir() {
        int lastIndexOfSlash = fullPath.lastIndexOf("/");
        String parentDirectory = fullPath.substring(0, lastIndexOfSlash);
        return createFile(parentDirectory);
    }

    @Override
    public String read(long offset, long length) throws FileOperationException {
        return read(offset, length, null);
    }

    @Override
    public void write(String data) throws FileOperationException {
        write(data, 0L, false, true, null);
    }

    @Override
    public void write(String data, long offset, boolean overwrite, String streamName) throws FileOperationException {
        write(data, offset, overwrite, false, streamName);
    }

    @Override
    public void write(String data, long offset, boolean overwrite) throws FileOperationException {
        write(data, offset, overwrite, false, null);
    }

    @Override
    public void write(String fileData, long offset, boolean overwrite, boolean update) throws FileOperationException {
        write(fileData, offset, overwrite, update, null);
    }

    public void write(String fileData, long offset, boolean overwrite, boolean update, String streamName)
            throws FileOperationException {
        if (fileData == null) {
            throw new NullPointerException();
        }

        StringBuffer data = convertStringToBase64Encoding(fileData);

        if (update && !StringUtils.isReal(streamName)) {
            try {
                delete();
            } catch (FileOperationException e) {
                // ignore
            }
        }

        write(data.toString(), offset, overwrite, streamName);
    }

     private StringBuffer convertStringToBase64Encoding(String fileData) {
        byte[] fileDataBytes = fileData.getBytes();
        StringBuffer data = new StringBuffer();
        String hexValue;
        int len;

        for (byte dataByte : fileDataBytes) {
            hexValue = Integer.toHexString(dataByte);
            // pad with 0 (if needed)
            len = hexValue.length();
            while (len++ < 2) {
                data.append("0");
            }
            // add hex value
            data.append(hexValue);
        }
        return data;
    }

    @Override
    public void mkdir() {
        mkdir("755");
    }

    @Override
    public void rmdir(boolean removeContents) {
        if (removeContents) {
            List<? extends File> files = listFiles();
            for (File file : files) {
            	if(file.isDirectory()){
            		file.rmdir(true);
            	} else {
            		file.delete();
            	}                
            }
        }
        rmdir();
    }

    /**
     * @return the ACL type of the file
     */
    @Override
    public String getAclType() {
        return aclType;
    }

    /**
     * @return the file type of the file
     */
    @Override
    public String getFileType() {
        return fileType;
    }

    @Override
    public boolean isDirectory() {
        return "directory".equals(getFileType());
    }

    /**
     * @return the accessed timestamp of the file
     */
    @Override
    public int getAccessedTimestamp() {
        return accessedTimestamp;
    }

    /**
     * @return the changed timestamp of the file
     */
    @Override
    public int getChangedTimestamp() {
        return changedTimestamp;
    }

    /**
     * @return the creation timestamp of the file
     */
    @Override
    public int getCreationTimestamp() {
        return creationTimestamp;
    }

    /**
     * @return the file size of the file
     */
    @Override
    public Size getFileSize() {
        return fileSize;
    }

    /**
     * @return the group ID of the file
     */
    @Override
    public int getGroupID() {
        return groupID;
    }

    /**
     * @return the number of hard links of the file
     */
    @Override
    public int getHardLinksCount() {
        return hardLinksCount;
    }

    /**
     * @return the inode number of the file
     */
    @Override
    public int getInodeNumber() {
        return inodeNumber;
    }

    /**
     * @return the modified timestamp of the file
     */
    public int getModifiedTimestamp() {
        return modifiedTimestamp;
    }

    /**
     * @return the owner ID of the file
     */
    @Override
    public int getOwnerId() {
        return ownerID;
    }

    /**
     * @return the file permission bits of the file
     */
    @Override
    public int getPerm() {
        return perm;
    }

    @Override
    public Size getUsed() {
        return used;
    }

    @Override
    public boolean exists() { return exists; }

    public void setAccessedTimestamp(int accessedTimestamp) {
        this.accessedTimestamp = accessedTimestamp;
    }

    public void setAclType(String aclType) {
        this.aclType = aclType;
    }

    public void setChangedTimestamp(int changedTimestamp) {
        this.changedTimestamp = changedTimestamp;
    }

    public void setCreationTimestamp(int creationTimestamp) {
        this.creationTimestamp = creationTimestamp;
    }

    public void setFileSize(Size fileSize) {
        this.fileSize = fileSize;
    }

    public void setFileType(String fileType) {
        this.fileType = fileType;
    }

    public void setGroupID(int groupID) {
        this.groupID = groupID;
    }

    public void setHardLinksCount(int hardLinksCount) {
        this.hardLinksCount = hardLinksCount;
    }

    public void setInodeNumber(int inodeNumber) {
        this.inodeNumber = inodeNumber;
    }

    public void setModifiedTimestamp(int modifiedTimestamp) {
        this.modifiedTimestamp = modifiedTimestamp;
    }

    public void setOwnerID(int ownerID) {
        this.ownerID = ownerID;
    }

    public void setPerm(int perm) {
        this.perm = perm;
    }

    public void setBytesUsed(long bytesUsed) {
        this.used = Size.createB(bytesUsed);
    }

    public void setExists(boolean exists) {
        this.exists = exists;
    }

    @Override
    public void refresh(Resource... resources) {
        // Handled by the proxy.
    }

    @Override
    public Refreshable refresh(Refreshable refreshable) {
        // Handled by the proxy
        return null;
    }

    @Override
    public File refresh() {
        getProxy().refresh(FileResource.values());
        setLastLoad(new Date());
        return this;
    }

    @Override
    public boolean equals(Object obj) {
        return obj instanceof File && getFullPath().equals(((File) obj).getFullPath());
    }

    @Override
    public int hashCode() {
        return getFullPath().hashCode();
    }

    @Override
    public String toString() {
        if (fileSize == null) {
            return "File [fullPath = " + fullPath + "]";
        } else {
            return "File [accessedTimestamp=" + accessedTimestamp +
                    ", aclType=" + aclType +
                    ", changedTimestamp=" + changedTimestamp +
                    ", creationTimestamp=" + creationTimestamp +
                    ", fileSize=" + fileSize +
                    ", fileType=" + fileType +
                    ", fullPath=" + fullPath +
                    ", groupID=" + groupID +
                    ", hardLinksCount=" + hardLinksCount +
                    ", inodeNumber=" + inodeNumber +
                    ", modifiedTimestamp=" + modifiedTimestamp +
                    ", name=" + getName() +
                    ", ownerID=" + ownerID +
                    ", perm=" + perm + "]";
        }
    }
}
