/*
 * $Id: //depot/prod/aeonflux/R3.0.1/src/main/java/com/netapp/offtap3/internal/ontap/file/NdmpCopyProcessImpl.java#1 $
 *
 * Copyright (c) 2014 NetApp, Inc.
 * All rights reserved.
 */
package com.netapp.offtap3.internal.ontap.file;

import com.netapp.offtap3.ControllerResource;
import com.netapp.offtap3.File;
import com.netapp.offtap3.FileCloneOperationException;
import com.netapp.offtap3.FileOperationException;
import com.netapp.offtap3.internal.ontap.controller.AbstractOntapDataController;
import com.netapp.offtap3.internal.ontap.controller.FilerImpl;
import com.netapp.offtap3.utils.Utils;
import com.netapp.offtap3.ontap.NdmpCopyProcess;
import com.netapp.offtap3.ontap.OntapLun;

import org.apache.log4j.Logger;

import java.util.concurrent.ExecutionException;
import java.util.concurrent.FutureTask;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.TimeoutException;

/**
 * Monitors the NDMP copy process.
 */
public class NdmpCopyProcessImpl implements NdmpCopyProcess, Runnable {
    private static final Logger LOGGER = Logger.getLogger(NdmpCopyProcessImpl.class);

    private static final int MAX_MAIN_COPY_PERCENT = 95;

    private AbstractOntapDataController dataController;
    private FutureTask<Integer> ndmpCopyTask;
    private File srcFile, destFile, tmpDestFile;
    private int returnCode = NdmpCopyProcess.COPY_IN_PROGRESS_RC;
    private Throwable copyException;
    private long srcFileSize, destFileSize;
    private int percentComplete;
    private boolean complete;
    private final int copyBeginChecksMax, copyProgressChecksMax;
    private final int destinationCheckInterval;

    public NdmpCopyProcessImpl(FutureTask<Integer> ndmpCopyTask, AbstractOntapDataController dataController,
                               File srcFile, File destFile, File tmpDestFile) {
        this.dataController = dataController;
        this.ndmpCopyTask = ndmpCopyTask;
        srcFileSize = srcFile.getFileSize().toBytes();
        this.srcFile = srcFile;
        this.destFile = destFile;
        this.tmpDestFile = tmpDestFile;
        destinationCheckInterval = dataController.getOfftapPreferences().getNdmpDestinationCheckInterval();
        copyBeginChecksMax = calculateCopyBeginChecksMax();
        copyProgressChecksMax = calculateCopyProgressChecksMax();
    }

    public NdmpCopyProcessImpl(FutureTask<Integer> ndmpCopyTask, AbstractOntapDataController dataController,
                               File srcFile, File destFile) {
        this(ndmpCopyTask, dataController, srcFile, destFile, destFile);
    }

    public void start(boolean async) {
    	start(async, false);
    }    
    
    public void start(boolean async, boolean skipPollingOnNdmpProgress) {
        new Thread(ndmpCopyTask).start();
        if(!skipPollingOnNdmpProgress){
        	if (async) {
                new Thread(this).start();
            } else {
                run();
            }
        }        
    }

    public void run() {
        long fileSize = 0L;
        try {
            boolean copyComplete = false;
            int copyBeginChecks = 0;
            int copyProgressChecks = 0;
            Long lastDestFileSize = 0L;
            while (!copyComplete && !hasError()) {
                Utils.sleep(destinationCheckInterval * 1000);

                checkClusterStatus();
                // we allow the file to not appear 30 times (1 minute 30 seconds) (most of which will occur at the start of the process)
                if (!tmpDestFile.exists()) {
                    LOGGER.debug("Number of get file tries is now " + copyBeginChecks);
                    if (copyBeginChecks > copyBeginChecksMax) {
                        String message = "The NDMP copy progress has failed. It did not begin after the timeout of " + dataController
                                .getOfftapPreferences().getNdmpBeginTimeout() + " seconds.";
                        LOGGER.error(message);
                        copyException = new FileCloneOperationException(message);
                    } else {
                        copyBeginChecks++;
                    }
                } else {
                    destFileSize = tmpDestFile.getFileSize().toBytes();
                    if (srcFileSize > 0) {
                        System.out.println("srcFileSize = " + srcFileSize);
                        System.out.println("destFileSize = " + destFileSize);
                        if (destFileSize >= srcFileSize) {
                            copyComplete = true;
                            percentComplete = MAX_MAIN_COPY_PERCENT;
                        } else {
                            percentComplete = (int) (((double) destFileSize / (double) srcFileSize) * MAX_MAIN_COPY_PERCENT);
                            LOGGER.debug(percentComplete + "% complete.  File size is now " + destFileSize +
                                                 " out of " + srcFileSize + ".");
                        }
                    } else {
                        percentComplete = MAX_MAIN_COPY_PERCENT;
                        copyComplete = true;
                        LOGGER.debug("Dest file info:" + tmpDestFile);                        
                    }

                    LOGGER.trace("ndmpCopy : " + tmpDestFile.getFullPath() + " current size: " + fileSize +
                                         " last reported size: " + lastDestFileSize);
                    //we allow the file dest size to not increase 60 times (5 minutes)
                    if (destFileSize > lastDestFileSize) {
                        lastDestFileSize = destFileSize;
                        copyProgressChecks = 0;
                    } else if (copyProgressChecks < copyProgressChecksMax) {
                        copyProgressChecks++;
                    } else {
                        String message = "The NDMP copy process has failed. It did not progress after the timeout of " + dataController
                                .getOfftapPreferences().getNdmpProgressTimeout() + " seconds.";
                        LOGGER.error(message);
                        copyException = new FileCloneOperationException(message);
                    }
                }
            }

            if (!hasError()) {
                LOGGER.debug("File copy complete. Waiting up to 5 minutes for NDMP process to wrap up.");
                // no errors were found, so we wait up to 5 minutes for the ndmp to shut down
                ndmpCopyTask.get(5, TimeUnit.MINUTES);
            } else {
                LOGGER.debug("File copy failed. Waiting up to 30 seconds for NDMP process to wrap up.");
                // the copy is on error, so we wait 30 seconds for the thread to end, then throw an exception.
                ndmpCopyTask.get(30, TimeUnit.SECONDS);
            }
        } catch (TimeoutException ex) {
            LOGGER.error("Copy of " + srcFile.getFullPath() + " to " + destFile.getFullPath() + " timed out. - ", ex);
            copyException = new FileCloneOperationException(
                    "Copy of " + srcFile.getFullPath() + " to " + destFile.getFullPath() + " timed out. - " +
                            ex.getMessage(), ex);
        } catch (Exception e) {
            LOGGER.error("Copy of " + srcFile.getFullPath() + " to " + destFile.getFullPath() + " failed. - ", e);
            copyException = new FileCloneOperationException(
                    "Copy of " + srcFile.getFullPath() + " to " + destFile.getFullPath() + " failed. - " +
                            e.getMessage(), e);
        }
       
        // Move the temporary file to the final destination if needed.
        moveTempToFinal();

        complete = true;
        percentComplete = 100;
        LOGGER.info("ndmpcopy " + srcFile.getFullPath() + " " + destFile.getFullPath() + " finished.");
    }

    private void checkClusterStatus() {
        if (dataController instanceof FilerImpl) {
            // make sure the cluster is still OK
            if (!((FilerImpl) dataController).isClusterUsable()) {
                String message = "Cluster even during ndmpCopyTask file may be inconsistent.";
                LOGGER.fatal(message);
                copyException = new FileCloneOperationException(message);
            }
        }
    }

    private void moveTempToFinal() {
        try {
            if (!tmpDestFile.equals(destFile)) {
            	File tmpDestDir = tmpDestFile.getParentDir();
            	if(tmpDestFile.getFileType().equals("lun")){            		            	
            		OntapLun tmpDestLun = dataController.getLun(tmpDestFile.getFullPath());
            		if(tmpDestLun == null){
            			dataController.refresh();
            			tmpDestLun = dataController.getLun(tmpDestFile.getFullPath());
            		}
            		tmpDestLun.move(destFile);            		
            	} else {
            		tmpDestFile.move(destFile);
            	}                                
                tmpDestDir.rmdir(true);
            }
        } catch (FileOperationException e) {
            if (copyException == null) {
                copyException = e;
            }
            LOGGER.error("Error occurred while moving temporary NDMP file to final destination. ", e);
        }
    }

    private int calculateCopyProgressChecksMax() {
        int progressTimeout = dataController.getOfftapPreferences().getNdmpProgressTimeout();
        return progressTimeout / destinationCheckInterval;
    }

    private int calculateCopyBeginChecksMax() {
        int beginTimeout = dataController.getOfftapPreferences().getNdmpBeginTimeout();
        return beginTimeout / destinationCheckInterval;
    }

    public int getReturnCode() {
        if (returnCode == NdmpCopyProcess.COPY_IN_PROGRESS_RC) {
            try {
                returnCode = ndmpCopyTask.get(5, TimeUnit.SECONDS);
            } catch (TimeoutException t) {
                // expected and ignored
            } catch (InterruptedException e) {
                // ignored
            } catch (ExecutionException e) {
                // not good. save for later and set return code to exception occurred.
                copyException = e.getCause();
                complete = true;
                returnCode = NdmpCopyProcess.COPY_EXCEPTION_RC;
                if (copyException instanceof Error) {
                    throw (Error) copyException;
                }
            }
        }
        return returnCode;
    }

    @Override
    public long getDestinationFileSize() {
        return destFileSize;
    }

    @Override
    public long getSourceFileSize() {
        return srcFileSize;
    }

    @Override
    public int getPercentComplete() {
        return percentComplete;
    }

    @Override
    public boolean isComplete() {
        return complete;
    }

    @Override
    public boolean hasError() {
        return copyException != null;
    }

    @Override
    public Throwable getException() {
        return copyException;
    }
}
