package com.netapp.offtap3.internal.ontap.file;

import com.netapp.offtap3.File;
import com.netapp.offtap3.FileOperationException;
import com.netapp.offtap3.utils.Utils;
import org.apache.log4j.Logger;

import java.io.UnsupportedEncodingException;
import java.nio.ByteBuffer;
import java.nio.ByteOrder;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;

public class NtfsReader {

    // MBR specific constants
    private static final int MBR_PARTITION_SIZE = 0x10;
    private static final int MBR_PARTITION_TYPE_OFFSET = 0x04;
    private static final int MBR_PARTITION_TYPE_NTFS = 0x07;
    private static final int MBR_PARTITION_TYPE_EXTENDED_1 = 0x05;
    private static final int MBR_PARTITION_TYPE_EXTENDED_2 = 0x0F;
    private static final int MBR_PARTITION_TYPE_EXTENDED_3 = 0x85;

    // GPT specific constants
    private static final int GPT_PARTITION_ENTRY_START = 0x400;
    private static final int GPT_PARTITION_ENTRY_SIZE = 0x80;

    // NTFS constants
    private static final int NTFS_ID_OFFSET = 0x03;
    private static final int NTFS_BYTES_PER_SECTOR_OFFSET = 0x0B;
    private static final int NTFS_SECTORS_PER_BLOCK_OFFSET = 0x0D;
    private static final int NTFS_SECTORS_IN_VOLUME = 0x28;
    private static final int NTFS_LCN_OF_MFT_OFFSET = 0x30;
    private static final int NTFS_BLOCKS_PER_MFT_RECORD_OFFSET = 0x40;
    private static final int MFT_RECORD_OFFSET_TO_FIRST_ATTRIBUTE = 0x14;

    private static final int MFT_RECORD_ATTRIBUTE_HEADER_LENGTH = 0x18;
    private static final String BITMAP_FILE_NAME = "$Bitmap";
    private static final String LOG_FILE_NAME = "$LogFile";
    private static final String VOL_FILE_NAME = "$VolumeImpl";

    private static final long NTFS_FILE_ATTR_SYSTEM = 0x00000004L;
    private static final long NTFS_FILE_ATTR_SPARSE_FILE = 0x00000200L;
    private static final long NTFS_FILE_ATTR_COMPRESSED = 0x00000800L;


    public static final String NTFS_UNCLEAN = "NTFS volume was not unmounted cleanly.";
    public static final String NTFS_INVALID = "invalid NTFS volume.";
    public static final String NTFS_UNSUPPORTED_VERSION = "unsupported NTFS version.";

    private static final long SECTOR_SIZE = 0x200;


    private static final Logger logger = Logger.getLogger(NtfsReader.class);
    private final File file;

    public NtfsReader(File file) {
        this.file = file;
    }

    /**
     * Check the offset of each partition for the NTFS magic string
     */
    public boolean hasNtfsPartition() throws FileOperationException {
        List<Long> offsets = findPartitionOffsets();

        logger.debug("Checking NTFS partition for: " + file.getName());

        for (Long startingOffset : offsets) {
            String id = file.read(startingOffset + NTFS_ID_OFFSET, 8);
            logger.debug(String.format("Partition id (%s) for offset (%s)", id, startingOffset));

            if ("NTFS".equals(id.trim())) {
                return true;
            }
        }

        return false;
    }

    /**
     * Returns a list of partition offsets
     */
    public List<Long> findPartitionOffsets() throws FileOperationException {
        logger.debug(String.format("Entering findPartitionOffsets (file: %s)", file));

        List<Long> offsets = new ArrayList<Long>();

        if (isMbr()) {
            logger.debug(String.format("Checking an MBR partition (file: %s)", file.getFullPath()));

            // The offsets of the 4 primary partition tables
            List<Integer> partitionTable = new ArrayList<Integer>();
            partitionTable.add(0x1BE);
            partitionTable.add(0x1CE);
            partitionTable.add(0x1DE);
            partitionTable.add(0x1EE);

            // loops through each of the primary partition tables
            for (Integer offset : partitionTable) {
                logger.debug(String.format("Checking offset (file: %s): %d", file.getFullPath(), offset));

                int[] partitionRecord = Utils.getBytes(file.readBin(offset, MBR_PARTITION_SIZE));
                int type = partitionRecord[MBR_PARTITION_TYPE_OFFSET];
                long lba = getLong(Arrays.copyOfRange(partitionRecord, 0x08, 0x0C));

                // determine if it is an extended type
                if ((type == MBR_PARTITION_TYPE_EXTENDED_1) || (type == MBR_PARTITION_TYPE_EXTENDED_2) || (type == MBR_PARTITION_TYPE_EXTENDED_3)) {
                    boolean hasEbr = true;
                    long startingEbrOffset = lba * SECTOR_SIZE;
                    long nextEbr = 0;
                    List<Long> ebrPartitionTable = new ArrayList<Long>();
                    ebrPartitionTable.add(0x1BEL);
                    ebrPartitionTable.add(0x1CEL);

                    while (hasEbr) {
                        long ebrOffset = startingEbrOffset + (nextEbr * SECTOR_SIZE);
                        for (long ebrPartitionOffset : ebrPartitionTable) {
                            ebrPartitionOffset = ebrOffset + ebrPartitionOffset;
                            partitionRecord = Utils.getBytes(file.readBin(ebrPartitionOffset, MBR_PARTITION_SIZE));

                            type = partitionRecord[MBR_PARTITION_TYPE_OFFSET];
                            lba = getLong(Arrays.copyOfRange(partitionRecord, 0x08, 0x0C));

                            // There are no more EBRs to read.
                            if (lba == 0) {
                                hasEbr = false;
                            }

                            if ((type == MBR_PARTITION_TYPE_EXTENDED_1) || (type == MBR_PARTITION_TYPE_EXTENDED_2) || (type == MBR_PARTITION_TYPE_EXTENDED_3)) {
                                nextEbr = lba;
                            } else if (type == MBR_PARTITION_TYPE_NTFS) {
                                offsets.add(ebrOffset + (lba * SECTOR_SIZE));
                            }
                        }
                    }
                } else if (type == MBR_PARTITION_TYPE_NTFS) {
                    // if it isn't, just add the offset to the start of the primary partition
                    offsets.add(lba * SECTOR_SIZE);
                }
            }
        } else if (isGpt()) {
            int offset = GPT_PARTITION_ENTRY_START;
            boolean hasEntry = true;

            while (hasEntry) {
                int[] partitionRecord = Utils.getBytes(file.readBin(offset, GPT_PARTITION_ENTRY_SIZE));
                long lba = getLong(Arrays.copyOfRange(partitionRecord, 0x20, 0x28));
                offset += GPT_PARTITION_ENTRY_SIZE;

                if (lba != 0) {
                    offsets.add(lba * SECTOR_SIZE);
                } else {
                    hasEntry = false;
                }
            }
        }

        logger.debug("Exiting findPartitionOffsets (file: " + file.getFullPath() + "):" + offsets);
        return offsets;
    }

    /**
     * Return a BitmapFile for an NTFS volume.
     * Verifies the NTFS volume is clean.
     */
    public NtfsBitmapFile buildNtfsBitmap(long startingOffset) throws FileOperationException {
        String id = file.read(startingOffset + NTFS_ID_OFFSET, 8);
        if (!"NTFS".equals(id.trim())) {
            logger.warn(file.getFullPath() + "Did not find NTFS at " + startingOffset + " (" + id + ")");
            return new NtfsBitmapFile(NTFS_INVALID);
        }

        long bytesPerSector = getLong(Utils.getBytes(file.readBin(startingOffset + NTFS_BYTES_PER_SECTOR_OFFSET, 2)));
        long sectorsPerCluster = getLong(Utils.getBytes(file.readBin(startingOffset + NTFS_SECTORS_PER_BLOCK_OFFSET, 1)));
        long clusterSize = bytesPerSector * sectorsPerCluster;
        logger.debug(String.format("Allocation unit size for NTFS partition (file: %s): %d", file.getFullPath(), clusterSize));

        long numberOfSectorsInVolume = getLong(Utils.getBytes(file.readBin(startingOffset + NTFS_SECTORS_IN_VOLUME, 8)));
        long partitionSize = numberOfSectorsInVolume * SECTOR_SIZE;

        long lcnOfMftStart = getLong(Utils.getBytes(file.readBin(startingOffset + NTFS_LCN_OF_MFT_OFFSET, 8)));
        logger.debug(String.format("LCN of the start of the MFT (file: %s): %d", file.getFullPath(), lcnOfMftStart));

        byte blocksPerMftRecord = (byte) getLong(Utils.getBytes(file.readBin(startingOffset + NTFS_BLOCKS_PER_MFT_RECORD_OFFSET, 1)));

        long bytesPerMftRecord;
        if (blocksPerMftRecord < 0) {
            bytesPerMftRecord = (long) Math.pow(2, (-1 * blocksPerMftRecord));
        } else {
            bytesPerMftRecord = blocksPerMftRecord * clusterSize;
        }

        long mftStart = startingOffset + (lcnOfMftStart * clusterSize);
        logger.debug(String.format("MFT start (file: %s): %d", file.getFullPath(), mftStart));

        boolean foundBitmapRecord = false;
        long recordOffset = mftStart;

        long majorVersion = 0;
        long minorVersion = 0;

        long volFlags;
        String volName = null;
        String fileName = null;
        boolean compressed = false;
        boolean sparse = false;
        while (!foundBitmapRecord) {
            //get the offset of the first attribute for this record
            long attrOffset = recordOffset + getLong(Utils.getBytes(file.readBin(recordOffset + MFT_RECORD_OFFSET_TO_FIRST_ATTRIBUTE, 2)));

            boolean hasAttrs = true;
            while (hasAttrs) {
                //read the attribute type
                long attrType = getLong(Utils.getBytes(file.readBin(attrOffset, 4)));
                //read the length
                long attrLength = getLong(Utils.getBytes(file.readBin(attrOffset + 4, 4)));
                logger.trace("MFT attribute " + NTFS_ATTR_TYPE.getType(attrType) + " has a length of " + attrLength);

                //if this is the file name attribute set fileName and see if its one we care about.
                if (attrType == NTFS_ATTR_TYPE.FILE_NAME.getValue()) {
                    long fileNameLengthOffset = attrOffset + MFT_RECORD_ATTRIBUTE_HEADER_LENGTH + 0x40;
                    long fileNameLength = getLong(Utils.getBytes(file.readBin(fileNameLengthOffset, 1)));
                    String data = file.readBin(fileNameLengthOffset + 2, (int) (fileNameLength * 2));

                    // the file name is in UTF-16LE format.  We need to decode it here.
                    String mychar;
                    if (data != null) {
                        if (data.length() > 0) {
                            int len = (data.length() / 2);
                            byte[] dataBytes = new byte[len];
                            int x = 0;
                            for (int i = 0; i < len; i++) {
                                mychar = data.substring(x, x + 2);
                                x = (x + 2);
                                dataBytes[i] = (byte) (Integer.parseInt(mychar, 16));
                            }
                            try {
                                fileName = new String(dataBytes, 0, dataBytes.length, "UTF-16LE");
                                logger.trace(String.format("Filename we are currently reading (file: %s): %s", file.getFullPath(), fileName));
                            } catch (UnsupportedEncodingException ex) {
                                logger.error(String.format("Unable to decode the string (file: %s): %s", file.getFullPath(), ex.getMessage()));
                            }
                        }
                    }

                    //if this is a filename we care about, keep going.  Otherwise, break out of the while loop
                    if (!BITMAP_FILE_NAME.equals(fileName) && !LOG_FILE_NAME.equals(fileName) && !VOL_FILE_NAME.equals(fileName)) {
                        //break out of while loop (stop processing attributes for this file)
                        break;
                    }

                } else if (attrType == NTFS_ATTR_TYPE.STD_INFO.getValue()) {
                    //pull the whole attribute in (cheaper than multiple reads)
                    int[] data = Utils.getBytes(file.readBin(attrOffset, attrLength));
                    long flags = getLong(subArray(data, 56, 60));
                    if ((flags & NTFS_FILE_ATTR_SYSTEM) == NTFS_FILE_ATTR_SYSTEM) {
                        if ((flags & NTFS_FILE_ATTR_SPARSE_FILE) == NTFS_FILE_ATTR_SPARSE_FILE) {
                            logger.warn("The sparse flag is set on a system file.");
                            sparse = true;
                        } else {
                            sparse = false;
                        }
                        if ((flags & NTFS_FILE_ATTR_COMPRESSED) == NTFS_FILE_ATTR_COMPRESSED) {
                            logger.warn("The compressed flag is set on a system file.");
                            compressed = true;
                        } else {
                            compressed = false;
                        }
                    } else {
                        logger.error("This is not a system file.");
                        return new NtfsBitmapFile(NTFS_INVALID);
                    }

                } else if (attrType == NTFS_ATTR_TYPE.VOL_NAME.getValue()) {
                    //pull the whole attribute in (cheaper than multiple reads)
                    int[] data = Utils.getBytes(file.readBin(attrOffset, attrLength));
                    Long dataOffset = getLong(subArray(data, 20, 22)); //24
                    volName = getStringValue(subArray(data, dataOffset.intValue(), (int) attrLength));

                } else if (attrType == NTFS_ATTR_TYPE.VOL_INFO.getValue()) {
                    //pull the whole attribute in (cheaper than multiple reads)
                    int[] data = Utils.getBytes(file.readBin(attrOffset, attrLength));
                    Long dataOffset = getLong(subArray(data, 20, 22)); //24
                    majorVersion = getLong(subArray(data, dataOffset.intValue() + 8, dataOffset.intValue() + 9));
                    minorVersion = getLong(subArray(data, dataOffset.intValue() + 9, dataOffset.intValue() + 10));
                    volFlags = getLong(subArray(data, dataOffset.intValue() + 10, dataOffset.intValue() + 12));
                    if (volFlags != 0) {
                        logger.warn("NTFS volume flags were set. volFlags=" + volFlags);
                    }

                     /* Version to OS mapping
                      * 1.1 indicates NT4.0
                      * 1.2 indicates NT4.0
                      * 2.x indicates win2k beta
                      * 3.0 indicates win2k
                      * 3.1 indicates xp, win2k3, vista, win2k8
                      */
                    //we support major version 1-3
                    if (majorVersion > 0 && majorVersion < 4) {
                        //we support only minor versions 0 and 1 of major version 3
                        if (majorVersion == 3 && minorVersion < 2) {
                            logger.info("Valid NTFS version found.  Version " + majorVersion + "." + minorVersion);
                        } else {
                            logger.error("Unsupported NTFS version found.  Version " + majorVersion + "." + minorVersion);
                            return new NtfsBitmapFile(NTFS_UNSUPPORTED_VERSION + " " + majorVersion + "." + minorVersion);
                        }
                    } else {
                        logger.error("Unsupported NTFS version found.  Version " + majorVersion + "." + minorVersion);
                        return new NtfsBitmapFile(NTFS_UNSUPPORTED_VERSION + " " + majorVersion + "." + minorVersion);
                    }

                } else if (attrType == NTFS_ATTR_TYPE.DATA.getValue()) {
                    //pull the whole attribute in (cheaper than multiple reads)
                    String data = file.readBin(attrOffset, attrLength);
                    String nonResident = data.substring(16, 18);

                    if ("01".equals(nonResident)) {
                        if (attrLength > 70) {
                            //verify that our data is not compressed, else break out of the loop
                            String compressionUnit = data.substring(68, 70);
                            if ("00".equals(compressionUnit)) {
                                logger.trace(fileName + "/$Data is non compressed (expected).");
                            } else {
                                logger.error(fileName + "/$Data is compressed, this is not normal.  Bail out!");
                                return new NtfsBitmapFile(NTFS_INVALID);
                            }
                        }
                    }

                    if (BITMAP_FILE_NAME.equals(fileName)) {

                        //verify that our data is not resident, else bust out of the loop
                        if ("01".equals(nonResident)) {
                            logger.trace(fileName + "/$Data is non resident (expected).");
                        } else {
                            logger.error(fileName + "/$Data is resident, this is not normal.  Bail out!");
                            return new NtfsBitmapFile(NTFS_INVALID);
                        }

                        //verify that we are not compressed
                        if (!compressed) {
                            logger.trace(fileName + "/$Data is not compressed.");
                        } else {
                            logger.error(fileName + "/$Data is compressed.  Bail out!");
                            return new NtfsBitmapFile(NTFS_INVALID);
                        }

                        //verify that we are not sparse
                        if (!sparse) {
                            logger.trace(fileName + "/$Data is not sparse.");
                        } else {
                            logger.error(fileName + "/$Data is sparse.  Bail out!");
                            return new NtfsBitmapFile(NTFS_INVALID);
                        }

                        long bitmapLength = ((partitionSize / clusterSize) / 8) + 1;  // 1 bit per cluster in the bitmap
                        long runListOffset = attrOffset + 0x40;

                        Map<Long, Long> runList = getRunList(startingOffset, runListOffset, clusterSize, bitmapLength);
                        logger.info("NTFS volume (" + volName + ") found.  Version " + majorVersion + "." + minorVersion);
                        return new NtfsBitmapFile(startingOffset, runList, bitmapLength, clusterSize);

                    } else if (LOG_FILE_NAME.equals(fileName)) {

                        //verify that our data is not resident, else bust out of the loop
                        if ("01".equals(nonResident)) {
                            logger.trace(fileName + "/$Data is non resident (expected).");
                        } else {
                            logger.error(fileName + "/$Data is resident, this is not normal.  Bail out!");
                            return new NtfsBitmapFile(NTFS_INVALID);
                        }

                        //verify that we are not compressed
                        if (!compressed) {
                            logger.trace(fileName + "/$Data is not compressed.");
                        } else {
                            logger.error(fileName + "/$Data is compressed.  Bail out!");
                            return new NtfsBitmapFile(NTFS_INVALID);
                        }

                        //verify that we are not sparse
                        if (!sparse) {
                            logger.trace(fileName + "/$Data is not sparse.");
                        } else {
                            logger.error(fileName + "/$Data is sparse.  Bail out!");
                            return new NtfsBitmapFile(NTFS_INVALID);
                        }

                        long runListOffset = attrOffset + 64;
                        Map<Long, Long> runList = getRunList(startingOffset, runListOffset, clusterSize);

                        //grab 2 pages worth of data from the run
                        int[] binData = readRun(runList, 8192);

                        //verify the first restart area
                        NtfsRestartArea restart1;
                        try {
                            restart1 = new NtfsRestartArea(subArray(binData, 0, 4096));
                        } catch (IllegalStateException e) {
                            logger.error("Restart 1 - " + e.getMessage());
                            return new NtfsBitmapFile(NTFS_UNCLEAN);
                        }

                        //verify the second restart area
                        NtfsRestartArea restart2;
                        try {
                            restart2 = new NtfsRestartArea(subArray(binData, 4096, 8192));
                        } catch (IllegalStateException e) {
                            logger.error("Restart 2 - " + e.getMessage());
                            return new NtfsBitmapFile(NTFS_UNCLEAN);
                        }

                        //verify that the 2 restart areas agree
                        if (restart1 == null || !restart1.equals(restart2)) {
                            logger.error("This NTFS volume is in an unknown state. " + restart1 + " " + restart2);
                            return new NtfsBitmapFile(NTFS_UNCLEAN);
                        } else {
                            logger.trace("$Logfile/$Data has been verified.  This volume is clean. " + restart1 + " " + restart2);
                        }

                    }
                    //the is nothing (we care about) after the data attribute
                    //saves us a read
                    hasAttrs = false;

                } else if (attrType == NTFS_ATTR_TYPE.END.getValue()) {
                    // no more attributes
                    hasAttrs = false;
                }
                attrOffset += attrLength;
            }

            recordOffset += bytesPerMftRecord;
        }

        return new NtfsBitmapFile(NTFS_INVALID);
    }

    /**
     * NTFS file data is stored in runs.  This method decodes runs into a Map that we can use in other methods.
     *
     * @param startingOffset - The offset of the start of the NTFS volume
     * @param runListOffset  - The offset of the runList we wish to process
     * @param clusterSize    - The NTFS cluster (block) size (4096 in most cases)
     */
    public Map<Long, Long> getRunList(long startingOffset, long runListOffset, long clusterSize) throws FileOperationException {
        return getRunList(startingOffset, runListOffset, clusterSize, null);
    }

    /**
     * NTFS file data is stored in runs.  This method decodes runs into a Map that we can use in other methods.
     * NOTE: This method does not support compressed files!
     *
     * @param startingOffset - The offset of the start of the NTFS volume
     * @param runListOffset  - The offset of the runList we wish to process
     * @param clusterSize    - The NTFS cluster (block) size (4096 in most cases)
     * @param bitmapLength   - The size of the bitmap (if you're interested in reading a bitmap, otherwise use the other signature)
     */
    public Map<Long, Long> getRunList(long startingOffset, long runListOffset, long clusterSize, Long bitmapLength) throws FileOperationException {
        long lcnOffset = 0;
        Map<Long, Long> runList = new LinkedHashMap<Long, Long>();
        long totalRunLength = 0;
        long finalRunLength = (bitmapLength != null) ? bitmapLength : 0L;

        // read the run list
        int[] header = Utils.getBytes(file.readBin(runListOffset, 1));
        while (header[0] != 0) {
            int lengthNibble = (header[0] & 0x0f); // read the length nibble
            int offsetNibble = (header[0] >> 4); // read the offset nibble

            //if offsetNibble is zero, this is a sparse run.
            //therefore we need to skip it
            if (offsetNibble != 0) {
                int[] runOffsetBytes = Utils.getBytes(file.readBin(runListOffset + 1 + lengthNibble, offsetNibble));

                // check for a negative number in the most significant byte
                byte mostSignificantByte = (byte) runOffsetBytes[runOffsetBytes.length - 1];
                long runOffsetLcn = getLong(runOffsetBytes);

                if (mostSignificantByte < 0) {
                    // If this is a scrambled file (see burt 562570), the current run is
                    // behind the previous.
                    long mask = getMask(runOffsetBytes.length);
                    runOffsetLcn = (~runOffsetLcn + 1) & mask;
                    lcnOffset -= runOffsetLcn;
                } else {
                    lcnOffset += runOffsetLcn;
                }
                long runOffset = startingOffset + (lcnOffset * clusterSize);

                long runLengthInClusters = getLong(Utils.getBytes(file.readBin(runListOffset + 1, lengthNibble)));
                long runLength = runLengthInClusters * clusterSize;

                // The bitmap run length is the number of clusters in a run.  This may total more than the actual length of the bitmap.
                // If it does, we set the length of the final run to the remaining size of the bitmap.
                totalRunLength += runLength;
                if (bitmapLength != null && totalRunLength > bitmapLength) {
                    runList.put(runOffset, finalRunLength);
                } else {
                    runList.put(runOffset, runLength);
                    if (bitmapLength != null) {
                        finalRunLength -= runLength;
                    }
                }
            } else {
                logger.trace("getRunList: Skipping sparse run.  offset=" + offsetNibble + " length=" + lengthNibble);
            }

            runListOffset += (lengthNibble + offsetNibble + 1);
            header = Utils.getBytes(file.readBin(runListOffset, 1));
        }
        return runList;
    }

    /**
     * Read data in an NTFS run
     *
     * @return array of int , 1 int per byte in the sector (to handle unsigned values)
     */
    public int[] readRun(Map<Long, Long> runList) throws FileOperationException {
        return readRun(runList, 0);
    }

    /**
     * Read data in an NTFS run
     *
     * @param bytes   - limit the bytes returned to this value
     * @return array of int , 1 int per byte in the sector (to handle unsigned values)
     */
    public int[] readRun(Map<Long, Long> runList, long bytes) throws FileOperationException {
        logger.debug("readRun: " + file);
        int allData[] = new int[0];

        long maxReadSize = 1048576L; // maximum file-read-file size is 1MB
        long totalBytesRead = 0;
        if (bytes > 0 && maxReadSize > bytes) {
            maxReadSize = bytes;
        }

        for (long runOffset : runList.keySet()) {
            long runLength = runList.get(runOffset);

            int bytesRead = 0;
            while (bytesRead < runLength) {
                long offsetToReadFrom = runOffset + bytesRead;
                long bytesToReadInRun = runLength - bytesRead;

                long bytesToRead = Math.min(bytesToReadInRun, maxReadSize);
                logger.debug(String.format("Bytes to read (file: %s) : %d", file.getFullPath(), bytesToRead));

                int[] data = Utils.getBytes(file.readBin(offsetToReadFrom, bytesToRead));
                allData = addAll(allData, data);
                bytesRead += bytesToRead;
                totalBytesRead += bytesRead;
                if (bytes > 0 && totalBytesRead >= bytes) {
                    //We have what was requested, break now.
                    break;
                }
            }
        }
        return allData;
    }


    /**
     * Returns an array of int , 1 int per byte in the sector (to handle unsigned values)
     */
    public int[] getSectors(long start, long count) {
        try {
            return Utils.getBytes(file.readBin(start * SECTOR_SIZE, count * SECTOR_SIZE));
        } catch (Exception ex) {
            logger.warn("Unable to read " + file, ex);
        }
        return null;
    }

    /**
      * Use System.arraycopy to create a new array based on 2 int arrays.
      */
     public static int[]  addAll(int[] array1, int[] array2) {
         if (array1 == null) throw new NullPointerException("First array was null.");
         if (array2 == null) throw new NullPointerException("Second array was null.");

         int[] returnMe = new int[array1.length + array2.length];

         System.arraycopy(array1, 0, returnMe, 0, array1.length);
         System.arraycopy(array2, 0, returnMe, array1.length, array2.length);

         return returnMe;
     }

    /**
     * @return true if the virtual disk has a MBR partition table
     */
    public boolean isMbr() {
        int[] sectors = getSectors(0, 2);
        if (sectors != null) {
            if ((sectors[510] == 0x55) && (sectors[511] == 0xAA)) {
                String gptSig = getStringValue(Arrays.copyOfRange(sectors, 512, 520));
                //Just checking for the GPT signature is not enough, we now check for
                //partition type (@offset 450) of 0xee and a starting lba (@offset 454) of 01
                return !((gptSig.equals("EFI PART")) && (sectors[450] == 0xEE) && (sectors[454] == 0x01));
            } else {
                return false;
            }
        } else {
            return false;
        }
    }

    /**
     * @return true if the virtual disk has a GPT partition table
     */
    public boolean isGpt() {
        int[] sectors = getSectors(0, 2);
        if (sectors != null) {
            if ((sectors[510] == 0x55) && (sectors[511] == 0xAA)) {
                String gptSig = getStringValue(Arrays.copyOfRange(sectors, 512, 520));
                //Just checking for the GPT signature is not enough, we now check for
                //partition type (@offset 450) of 0xee and a starting lba (@offset 454) of 01
                return (gptSig.equals("EFI PART")) && (sectors[450] == 0xEE) && (sectors[454] == 0x01);
            } else {
                return false;
            }
        } else {
            return false;
        }
    }

    /**
     * Returns the string value of the given array
     *
     * @param data the array to convert to a string
     * @return the string value
     */
    String getStringValue(int[] data) {

        byte[] bytes = new byte[data.length];
        for (int i = 0; i < data.length; i++) {
            bytes[i] = (byte) data[i];
        }
        return new String(bytes);
    }

    /**
      * Returns a long value across the given array
      * @param input the array to convert to a long
      * @return a long value across the given array
      */
     public long getLong(int[] input) {
         int size = (input.length > 7) ? input.length : 8;

         ByteBuffer bb = ByteBuffer.allocate(size);
         bb.order(ByteOrder.LITTLE_ENDIAN);
         for (int anInput : input) {
             bb.put((byte) anInput);
         }

         return bb.getLong(0);
     }

    /**
     * Returns a mask used to get variable length bytes from the bitmap data run offset
     * @param length the length of the data run offset
     *
     * @return a mask used to get variable length bytes from the bitmap data run offset
     */
    public long getMask(int length) {
        switch(length) {
            case 1: return 0xFFL;
            case 2: return 0xFFFFL;
            case 3: return 0xFFFFFFL;
            case 4: return 0xFFFFFFFFL;
            case 5: return 0xFFFFFFFFFFL;
            case 6: return 0xFFFFFFFFFFFFL;
            case 7: return 0xFFFFFFFFFFFFFFL;
            case 8: return 0xFFFFFFFFFFFFFFFFL;
            default: throw new IllegalArgumentException("Don't do this!");
        }
    }

    /**
     * Get a new array of int based on an existing array.
     * @param array - starting offset
     * @param start - ending offset
     * @param end - data before end index is returned, but not end index
     */
    public int[] subArray(int[] array, int start, int end) {
        if (array == null) return null;
        if (start < 0) start = 0;
        if (end > array.length) end = array.length;

        int subSize = end - start;

        if (subSize <= 0) return new int[0];

        int[] subArray = new int[subSize];
        System.arraycopy(array, start, subArray, 0, subSize);
        return subArray;
    }

    /**
     * This describes NTFS attribute types in MFT records
     *
     * @author forgette
     * @version $Revision: #1 $ $Date: 2015/07/22 $
     */
    enum NTFS_ATTR_TYPE {
        UNUSED(0), STD_INFO(0x10), ATTR_LIST(0x20), FILE_NAME(0x30), OBJECT_ID(0x40), SECURITY(0x50),
        VOL_NAME(0x60), VOL_INFO(0x70), DATA(0x80), INDEX_ROOT(0x90), INDEX_ALLOCATION(0xa0),
        BITMAP(0xb0), REPARSE_POINT(0xc0), EA_INFO(0xd0), EA(0xe0), PROP_SET(0xf0), LOGGED_UTIL_STEAM(0x100),
        FIRST_USER_ATTR(0x1000), END(4294967295L);

        private long value;

        private NTFS_ATTR_TYPE(long code) {
            this.value = code;
        }

        static String getType(long attrType) {
            for (NTFS_ATTR_TYPE type : NTFS_ATTR_TYPE.values()) {
                if (type.getValue() == attrType) {
                    return type.toString();
                }
            }
            return null;
        }

        long getValue() {
            return value;
        }
    }

    /**
     * This class describes the restart area of the $Logfile
     *
     * @author forgette
     * @version $Revision: #1 $ $Date: 2015/07/22 $
     */
    class NtfsRestartArea {
        final private String magic;
        final private Long restartOffset;
        final private Long major;
        final private Long minor;
        final private Long clientInUseList;
        final private long flags;

        final static long RESTART_VOLUME_IS_CLEAN = 2L;
        final static long LOGFILE_NO_CLIENT = 65535L;
        final static String CHKD = "CHKD";
        final static String RSTR = "RSTR";

        /**
         * @throws IllegalStateException
         */
        NtfsRestartArea(int[] data) throws IllegalStateException {
            magic = getStringValue(subArray(data, 0, 4));
            restartOffset = getLong(subArray(data, 24, 26));
            major = getLong(subArray(data, 26, 28));
            minor = getLong(subArray(data, 28, 30));
            clientInUseList = getLong(subArray(data, restartOffset.intValue() + 12, restartOffset.intValue() + 14));
            flags = getLong(subArray(data, restartOffset.intValue() + 14, restartOffset.intValue() + 16));

            //Check to see if chkdsk has messed with the volume
            if (CHKD.equals(magic)) {
                throw new IllegalStateException("The restart header page in $LogFile has been modified by chkdsk.  This NTFS volume is in an unknown state.");
            }
            //Check to make sure we have the right header
            if (!RSTR.equals(magic)) {
                throw new IllegalStateException("The restart header page in $LogFile has an unexpected value (" + magic + ").  This NTFS volume is in an unknown state.");
            }
            //Check to see that we have a valid offset
            if (restartOffset < 31) {
                throw new IllegalStateException("The restart header page in $LogFile has an illegal offset.  This NTFS volume is in an unknown state.");
            }
            //Check to see that we have a valid version number
            if (major == null || minor == null) {
                throw new IllegalStateException("The restart header page in $LogFile has a version of " + major + "." + minor + ".  This NTFS volume is in an unknown state.");
            }
            //Check to see if the volume has active clients (log left open) and was not shut down clean.
            if ((!clientInUseList.equals(LOGFILE_NO_CLIENT)) && (flags & RESTART_VOLUME_IS_CLEAN) != RESTART_VOLUME_IS_CLEAN) {
                throw new IllegalStateException("This NTFS volume is in an unknown state.  clientInUseList: " + clientInUseList + " flag: " + flags);
            }

        }

        @Override
        public String toString() {
            return "NTFSRestartArea [magic=" + magic + ", restartOffset=" + restartOffset + ", major=" + major + ", minor=" + minor + ", clientInUseList=" + clientInUseList + ", flags=" + flags + "]";
        }

        @Override
        public int hashCode() {
            final int prime = 31;
            int result = 1;
            result = prime * result + (magic.hashCode());
            result = prime * result + (major.hashCode());
            result = prime * result + (minor.hashCode());
            result = prime * result + (restartOffset.hashCode());
            return result;
        }

        @Override
        public boolean equals(Object obj) {
            if (this == obj)
                return true;
            if (obj == null)
                return false;
            if (getClass() != obj.getClass())
                return false;
            NtfsRestartArea other = (NtfsRestartArea) obj;
            if (!magic.equals(other.magic))
                return false;
            if (!major.equals(other.major))
                return false;
            if (!minor.equals(other.minor))
                return false;
            return restartOffset.equals(other.restartOffset);
        }
    }

}
