package com.netapp.offtap3.internal.ontap.file;

import java.util.ArrayList;
import java.util.Collections;
import java.util.List;

import org.apache.log4j.Logger;

import com.netapp.offtap3.ControllerConnection;
import com.netapp.offtap3.ControllerResource;
import com.netapp.offtap3.File;
import com.netapp.offtap3.FileOperationException;
import com.netapp.offtap3.FileResource;
import com.netapp.offtap3.Size;
import com.netapp.offtap3.internal.ontap.MCCManager;
import com.netapp.offtap3.internal.ontap.controller.AbstractOntapDataController;
import com.netapp.offtap3.internal.ontap.support.ZapiCache;
import com.netapp.offtap3.internal.ontap.volume.OntapVolumeImpl;
import com.netapp.offtap3.ontap.ApiFailedException;
import com.netapp.offtap3.ontap.DataSvm;
import com.netapp.offtap3.ontap.FileRange;
import com.netapp.offtap3.ontap.LunOperationException;
import com.netapp.offtap3.ontap.NaElement;
import com.netapp.offtap3.ontap.OntapDataController;
import com.netapp.offtap3.ontap.OntapFile;
import com.netapp.offtap3.ontap.OntapLun;
import com.netapp.offtap3.ontap.OntapVolume;
import com.netapp.offtap3.utils.StringUtils;
import com.netapp.offtap3.utils.Utils;

public abstract class OntapFileImpl<F extends OntapFile, C extends OntapDataController, M extends AbstractOntapDataController<C>, V extends OntapVolumeImpl<I, C, M>, I extends OntapVolume> extends
        FileImpl<F> implements OntapFile, InternalOntapFile<C, M> {

    private static final long WAFL_BLOCK_SIZE = 4096L;

    private static final Logger LOGGER = Logger.getLogger(OntapFileImpl.class);
    private final M controller;
    private final V volume;
    private final NtfsReader ntfsReader;

    protected OntapFileImpl(M controller, V volume, String fullPath) {
        super(fullPath, controller);
        this.volume = volume;
        this.controller = controller;
        this.ntfsReader = new NtfsReader(this);
    }

    @Override
    public C getController() {
        return controller.getProxy();
    }

    @Override
    public M getControllerImpl() {
        return controller;
    }

    public V getVolumeImpl() {
        return volume;
    }

    @Override
    public I getVolume() {
        return volume.getProxy();
    }

    public void copy(File destination) {
        // TODO aeonflux-next
        // try to figure out the necessary generics voodoo
        // to get this to work
        if (destination instanceof OntapFile) {
            OntapFile destFile = (OntapFile) destination;
            if (getVolume().equals((destFile).getVolume())) {
                sisClone(destFile);
            } else {
                ndmpCopy(destFile);
            }
        } else {
            throw new UnsupportedOperationException(
                    "Don't know how to copy to a file of type: " + destination.getClass() + " named: " + destination);
        }
    }

    @Override
    public long getReserveSize() {
        // tdeepak 4/18/2014 see burt:817535 We need to use the bytes used by first finding
        // the blocks used and then converting that to bytes
        long blocksUsed = (long) Math.ceil(getProxy().getFileSize().toBytes() / (double) WAFL_BLOCK_SIZE);
        // but, add 2 blocks to this count to account for VMaligned volumes. See burt 823944
        blocksUsed = blocksUsed + 2;
        return 2L * blocksUsed * WAFL_BLOCK_SIZE;
    }

    public boolean exists() {
        try {
            getProxy().refresh(FileResource.file_metadata);
        } catch (FileOperationException e) {
            setExists(false);
        }
        return super.exists();
    }

    @Override
    public OntapLun createLun(String lunPath, String osType, boolean spaceReserve, String comment) {
        String fullLunPath = volume.toFullPath(lunPath);
        NaElement xi = new NaElement("lun-create-from-file");
        xi.addNewChild("file-name", getFullPath());
        xi.addNewChild("ostype", osType);
        xi.addNewChild("space-reservation-enabled", String.valueOf(spaceReserve));
        xi.addNewChild("path", fullLunPath);

        try {
            controller.invokeZapi(xi, ZapiCache.NO_CACHE);
            controller.getZapiCache().blastCacheValues("lun-");
            LOGGER.info("Create lun:" + getFullPath() + ".");
        } catch (Exception e) {
            LOGGER.error("Failed to create lun from file " + getFullPath() + ".", e);
            throw new LunOperationException(controller + " failed to createLun " + getFullPath() +
                                                    ".  Error: " + e.getMessage(), e);
        }

        getController().refresh(ControllerResource.luns);
        OntapLun lun = getController().getLun(fullLunPath);
        if (comment != null) {
            lun.setComment(comment);
        }
        return lun;
    }

    public String concatPath(String path2) {
        StringBuilder newPath = new StringBuilder(getFullPath());
        if (!getFullPath().endsWith("/")) {
            newPath.append("/");
        }
        newPath.append(path2);
        return newPath.toString();
    }

    @Override
    public void write(String data, long offset, boolean overwrite, String streamName) throws FileOperationException {
        String path = getFullPath();

        NaElement xi = new NaElement("file-write-file");
        xi.addNewChild("path", path);
        xi.addNewChild("offset", Long.toString(offset));
        xi.addNewChild("overwrite", Boolean.toString(overwrite));
        xi.addNewChild("data", data);
        if (StringUtils.isReal(streamName) && getController().getVersion().atLeastFS()) {
            xi.addNewChild("stream-name", streamName);
        }

        try {
            try {
                getControllerImpl().invokeZapi(xi, ZapiCache.NO_CACHE);
            } catch (ApiFailedException t) {
                if (t.getMessage() != null && t.getMessage().contains("13001")) {
                    Utils.sleep(2000);
                    getControllerImpl().invokeZapi(xi, ZapiCache.NO_CACHE);
                } else {
                    throw t;
                }
            }
            getControllerImpl().getZapiCache().blastCacheValues("file-", "volume-", "aggr-");
        } catch (ApiFailedException e) {
            throw new FileOperationException(
                    "file-write-file failed to write to path " + path + " with error: " + e.getMessage(), e);
        }
    }

    @Override
    public void delete() throws FileOperationException {
        delete(false);
    }

    @Override
    public void delete(boolean isCloneFile) throws FileOperationException {
        String path = getFullPath();

        NaElement xi = new NaElement("file-delete-file");
        xi.addNewChild("path", path);

        if (getController().getVersion().atLeastFS()) {
            // in FS, sis-clones should ideally be deleted via the clone engine. This requirement only exists for files and not luns
            xi.addNewChild("is-clone-file", "" + isCloneFile);
        }

        try {
            getControllerImpl().invokeZapi(xi, ZapiCache.NO_CACHE);
            getControllerImpl().getZapiCache().blastCacheValues("file-", "volume-", "aggr-");
        } catch (ApiFailedException e) {
            throw new FileOperationException("deleteFile failed " + path + " with " + e, e);
        }
    }

    @Override
    public String readBin(long offset, long length) throws FileOperationException {
        NaElement xi = new NaElement("file-read-file");
        String data = null;

        if (length == 0) {
            length = getProxy().getFileSize().toBytes();
        }

        xi.addNewChild("path", getFullPath());
        xi.addNewChild("offset", Long.toString(offset));
        xi.addNewChild("length", Long.toString(length));

        try {
            NaElement xo = getControllerImpl().invokeZapi(xi, ZapiCache.SHORT_CACHE);
            if (xo.getAttr("status").equals("passed")) {
                if (xo.getChildIntValue("length", 0) > 0) {
                    data = xo.getChildContent("data");
                }
            }
        } catch (ApiFailedException e) {
            LOGGER.error("readFileBin failed " + getFullPath(), e);
            throw new FileOperationException("readFileBin failed " + getFullPath() + " error: " + e.getMessage(), e);
        }

        return data;
    }

    @Override
    public String read(long offset, long length, String streamName) throws FileOperationException {
        LOGGER.debug("fileReadFile :" + getFullPath());
        NaElement xi = new NaElement("file-read-file");
        String asciiHex = null;

        if (length == 0) {
            length = getProxy().getFileSize().toBytes();
        }

        xi.addNewChild("path", getFullPath());
        xi.addNewChild("offset", Long.toString(offset));
        xi.addNewChild("length", Long.toString(length));

        if (StringUtils.isReal(streamName) && getController().getVersion().atLeastFS()) {
            xi.addNewChild("stream-name", streamName);
        }

        try {
            NaElement xo = getControllerImpl().invokeZapi(xi, ZapiCache.SHORT_CACHE);
            if (xo.getAttr("status").equals("passed")) {
                if (xo.getChildIntValue("length", 0) > 0) {
                    asciiHex = xo.getChildContent("data");
                }
            }
        } catch (ApiFailedException e) {
            LOGGER.error("readFile failed " + getFullPath(), e);
            throw new FileOperationException("readFile failed " + getFullPath() + " error: " + e.getMessage(), e);
        }

        return convertBytes(asciiHex);
    }

    protected String convertBytes(String data) {
        try {
            String mychar;
            if (data != null) {
                if (data.length() > 0) {
                    int len = (data.length() / 2);
                    byte[] dataBytes = new byte[len];
                    int x = 0;
                    for (int i = 0; i < len; i++) {
                        mychar = data.substring(x, x + 2);
                        x = (x + 2);
                        dataBytes[i] = (byte) (Integer.parseInt(mychar, 16));
                    }
                    return new String(dataBytes);
                }
            }
        } catch (RuntimeException e) {
            return data;
        }
        return null;
    }

    @Override
    public void truncate(Size size) throws FileOperationException {
        String path = getFullPath();

        NaElement xi = new NaElement("file-truncate-file");
        xi.addNewChild("path", path);
        xi.addNewChild("size", Long.toString(size.toBytes()));

        try {
            getControllerImpl().invokeZapi(xi, ZapiCache.NO_CACHE);
            getControllerImpl().getZapiCache().blastCacheValues("file-");
        } catch (ApiFailedException e) {
            throw new FileOperationException(
                    "file-truncate-file failed to truncate " + path + " with error: " + e.getMessage(), e);
        }
    }

    @Override
    public String getVolumeRelativePath() {
        String fullPath = getFullPath();
        String volumePrefix = getVolumeImpl().getFile("/").getFullPath();
        if ((fullPath + "/").equals(volumePrefix)) {
            return "";
        }

        if (fullPath.startsWith(volumePrefix)) {
            return fullPath.substring(volumePrefix.length());
        } else {
            throw new IllegalStateException(
                    "internal error with file path format.  File: " + this + " does not appear within " + volumePrefix);
        }
    }

    @Override
    public String getNdmpPath() {
        OntapVolumeImpl vol = getVolumeImpl();
        String volRelative = getVolumeRelativePath();
        StringBuilder builder = new StringBuilder();
        if (vol.getController() instanceof DataSvm) {
        	builder.append("/").append(MCCManager.getInstance().getVServerNameByUuid((DataSvm) vol.getController())).append("/").append(vol.getName());        	
        } else {
        	builder.append("/").append(vol.getController().getName()).append("/").append(vol.getName());
        }

        if (!volRelative.isEmpty()) {
            builder.append("/").append(volRelative);
        }

        return builder.toString();
    }

    @Override
    public OntapFile getParentDir() {
        int lastIndexOfSlash = getFullPath().lastIndexOf("/");
        String parentDirectory = getFullPath().substring(0, lastIndexOfSlash);
        return getVolumeImpl().getFile(parentDirectory);
    }

    @Override
    public void mkdir(String permissions) throws FileOperationException {
        String path = getFullPath();

        LOGGER.debug("mkdir: " + path);
        NaElement xi;
        if (path.endsWith("/")) {
            path = path.substring(0, path.length() - 1);
        }
        xi = new NaElement("file-create-directory");
        xi.addNewChild("path", path);
        xi.addNewChild("perm", permissions);
        try {
            getControllerImpl().invokeZapi(xi, ZapiCache.NO_CACHE);
            getControllerImpl().getZapiCache().blastCacheValues("file-");
            LOGGER.info("Make Directory " + path + ".");
        } catch (ApiFailedException e) {
            if (e.getErrno() == 17) {
                LOGGER.warn("Make Directory " + path + " failed. " + e.getMessage());
            } else {
                LOGGER.error("Make Directory " + path + " failed. " + e.getMessage(), e);
                throw new FileOperationException("Make directory " + path + " failed: " + e.getMessage(), e);
            }
        }
    }

    @Override
    public void rmdir() throws FileOperationException {
        String path = getFullPath();

        LOGGER.debug("rmdir: " + path);

        NaElement xi = new NaElement("file-delete-directory");
        xi.addNewChild("path", path);
        try {
            NaElement xo = getControllerImpl().invokeZapi(xi, ZapiCache.NO_CACHE);
            getControllerImpl().getZapiCache().blastCacheValues("file-");
            LOGGER.info("Remove Directory " + path + ":" + xo.getAttr("status") + ".");
        } catch (ApiFailedException e) {
            LOGGER.error("Remove Directory " + path + "failed.", e);
            throw new FileOperationException("Remove Directory " + path + " failed: " + e.getMessage(), e);
        }
    }

    @Override
    public long getUniqueBytes() throws FileOperationException {
        String path = getFullPath();

        try {
            NaElement xi = new NaElement("file-usage-get");
            xi.addNewChild("path", path);
            NaElement xo = getControllerImpl().invokeZapi(xi, ZapiCache.NO_CACHE);
            try {
                return Long.parseLong(xo.getChildContent("unique-bytes"));
            } catch (NumberFormatException e) {
                return 0;
            }
        } catch (ApiFailedException e) {
            throw new FileOperationException("Unable to find unqiue bytes for " + path + " saw " + e.getMessage(), e);
        }
    }

    /**
     * Reclaim space for this file, assuming it is NTFS formatted.
     */
    @Override
    public String reclaimSpace() throws FileOperationException {
        List<Long> offsets = ntfsReader.findPartitionOffsets();
        String errorInfo = null;

        for (Long startingOffset : offsets) {
            NtfsBitmapFile partition = ntfsReader.buildNtfsBitmap(startingOffset);

            if (partition != null) {
                if (partition.isValid()) {
                    LOGGER.debug("Reclaiming space from NTFS partition: " + partition);
                    punchNtfsVolume(partition);
                } else {
                    errorInfo = (errorInfo == null) ?
                            "At offset " + startingOffset + " " + partition.getErrorReason() :
                            errorInfo + " " + "At offset " + startingOffset + " " + partition.getErrorReason();
                }
            }
        }
        return errorInfo;
    }

    public abstract void punchHoles(List<FileRange> range);

    /**
     * Punch holes in an NTFS volume
     */
    private void punchNtfsVolume(NtfsBitmapFile bitmapFile) throws FileOperationException {
        LOGGER.debug(String.format("Entering punchPartition (file: %s)", getFullPath()));

        long maxReadSize = 1048576L; // maximum file-read-file size is 1MB

        List<FileRange> holes = new ArrayList<FileRange>();

        long bitCount = 0;
        long startingBit = 0;

        for (long runOffset : bitmapFile.getBitmapRunList().keySet()) {
            long runLength = bitmapFile.getBitmapRunList().get(runOffset);

            int bytesRead = 0;
            while (bytesRead < runLength) {
                long offsetToReadFrom = runOffset + bytesRead;
                long bytesToReadInRun = runLength - bytesRead;

                long bytesToRead = Math.min(bytesToReadInRun, maxReadSize);
                LOGGER.debug(String.format("Bytes to read (file: %s) : %d", getFullPath(), bytesToRead));

                int[] bitmap = Utils.getBytes(readBin(offsetToReadFrom, bytesToRead));

                for (int mapbit : bitmap) {
                    short mask = 1;
                    while (mask < 256) {
                        boolean bit = ((mapbit & mask) != 0);

                        if (!bit) {
                            // we have a block to free
                            if (startingBit == 0) {
                                startingBit = bitCount;
                            }
                        } else {
                            // check if we finished a free block range
                            if (startingBit != 0) {
                                if (bitCount - startingBit > 10) {
                                    long numberOfBytes = (bitCount - startingBit) * bitmapFile.getClusterSize();
                                    long byteOffset = bitmapFile.getStartingOffset() +
                                            (startingBit * bitmapFile.getClusterSize());
                                    holes.add(new FileRange(byteOffset, numberOfBytes));
                                }
                                startingBit = 0;
                            }
                        }
                        mask <<= 1;
                        // increment the bit count and grab the right bit
                        bitCount++;
                    }
                }
                bytesRead += bytesToRead;
            }
        }

        if (holes.isEmpty()) {
            LOGGER.debug(String.format("No holes to punch (file: %s)", getFullPath()));
            return;
        }

        LOGGER.debug(String.format("Starting offset (file: %s): %d", getFullPath(), bitmapFile.getStartingOffset()));

        LOGGER.debug(String.format("*** Begin - Ranges to punch (file: %s) ***", getFullPath()));
        for (FileRange hole : holes) {
            LOGGER.debug(String.format("  Hole start (file: %s) - %s", getFullPath(), hole));
            LOGGER.debug(String.format("  End offset (file: %s) - %s", getFullPath(),
                                       Long.toHexString(hole.getOffset() + hole.getLength())));

        }
        LOGGER.debug(String.format("*** End - Ranges to punch (file: %s) ***", getFullPath()));

        punchHoles(holes);
    }

    /**
     * Transforms the underlying list of NtapHole into a list where no NtapHole has a length that is
     * greater than the maxHoleSize requested.
     *
     * @param maxHoleSize limit on length of a NtapHole
     * @return new list covering the same holes as underlying list, but with length values limited to maxHoleSize
     */
    protected List<FileRange> asMaxHoleList(List<FileRange> holes, long maxHoleSize) {
        List<FileRange> biteSizeHoles = new ArrayList<FileRange>();

        // insure they are ordered by offset
        Collections.sort(holes);

        for (FileRange hole : holes) {
            long byteOffset = hole.getOffset();
            long numberOfBytes = hole.getLength();
            while (numberOfBytes > 0) {
                biteSizeHoles.add(new FileRange(byteOffset, Math.min(numberOfBytes, maxHoleSize)));
                numberOfBytes -= maxHoleSize;
                byteOffset += maxHoleSize;
            }
        }
        return biteSizeHoles;
    }

    @Override
    public ControllerConnection getConnection() {
        return getControllerImpl().getConnection();
    }

    @Override
    public OntapFile refresh() {
        return (OntapFile) super.refresh();
    }
}
