package com.netapp.offtap3.internal.ontap.file;

import com.netapp.offtap3.File;
import com.netapp.offtap3.FileCloneOperationException;
import com.netapp.offtap3.internal.ontap.controller.FilerImpl;
import com.netapp.offtap3.internal.ontap.controller.OntapConnection;
import com.netapp.offtap3.internal.ontap.support.ZapiCache;
import com.netapp.offtap3.ontap.utils.OntapCallable;
import com.netapp.offtap3.utils.Utils;
import com.netapp.offtap3.ontap.ApiFailedException;
import com.netapp.offtap3.ontap.NaAuthenticationException;
import com.netapp.offtap3.ontap.NaElement;
import com.netapp.offtap3.ontap.NaErrno;
import com.netapp.offtap3.ontap.NaProtocolException;
import com.netapp.offtap3.ontap.NdmpCopyProcess;
import com.netapp.offtap3.ontap.OntapFile;
import com.netapp.offtap3.ontap.SevenModeCloneSpec;
import org.apache.log4j.Logger;

import java.io.IOException;
import java.util.concurrent.FutureTask;
import java.util.concurrent.LinkedBlockingQueue;

public class SevenModeFileCopier {

    private static final Logger LOGGER = Logger.getLogger(SevenModeFileCopier.class);

    private FilerImpl filer;


    public SevenModeFileCopier(FilerImpl filer) {
        this.filer = filer;
    }

    public NdmpCopyProcess ndmpCopy(OntapFile srcFile, File destFile,
                  final String options, boolean async) throws FileCloneOperationException {

        final String ndmpSourcePath = srcFile.getFullPath();
        final String destPath = destFile.getFullPath();

        if (!srcFile.exists()) {
            throw new FileCloneOperationException("Error: The source file " + srcFile.getFullPath() + " does not exist.)");
        }
        LOGGER.debug(ndmpSourcePath + " to " + destPath + " starting.");

        final File tmpDestDir = destFile.getParentDir().createFile(".tmpNdmpDir");
        tmpDestDir.mkdir();
        File tmpDestFile = tmpDestDir.createFile(srcFile.getName());

        filer.enableNdmp();

        FutureTask<Integer> ndmpCopyTask = new FutureTask<Integer>(new OntapCallable<Integer>() {
            @Override
            public Integer callLocal() throws Exception {

                NaElement xi = new NaElement("system-cli");
                xi.addNewChild("priv", "admin");
                NaElement xa = new NaElement("args");

                OntapConnection conn = filer.getNonPassThroughConnection();
                filer.addSystemCliArgs(xa);

                xa.addNewChild("arg", "ndmpcopy");

                if (options != null) {
                    xa.addNewChild("arg", "-e");
                    xa.addNewChild("arg", options);
                }

                xa.addNewChild("arg", ndmpSourcePath);

                xa.addNewChild("arg", tmpDestDir.getFullPath());
                xi.addChildElem(xa);
                NaElement xo;
                int rc = -9;
                try {
                    xo = conn.invokeZapi(xi, ZapiCache.NO_CACHE);
                    rc = xo.getChildIntValue("cli-result-value", 0);
                    if (rc < 0) {
                        String output = xo.getChildContent("cli-output");
                        LOGGER.error("ndmpcopy of " + ndmpSourcePath + " to " + tmpDestDir + " failed. - " + output);
                        throw new FileCloneOperationException("ndmpcopy of " + ndmpSourcePath + " to " + tmpDestDir + " failed. - " + output);
                    }
                } catch (ApiFailedException e) {
                    //dfm proxy doesn't give us anything back... so we guess that everything is ok
                    //if it isn't we'll die later.  Not pretty but it works.
                    if (e.getErrno() == 13011) {
                        rc = 2;
                    } else {
                        LOGGER.error("Error starting copy of " + ndmpSourcePath + ".", e);
                    }
                }

                return rc;
            }
        });

        NdmpCopyProcessImpl copyProcess = new NdmpCopyProcessImpl(ndmpCopyTask, filer, srcFile, destFile, tmpDestFile);
        copyProcess.start(async);
        return copyProcess;
    }

    /**
     * Invokes a sis clone on the specified filer/naServer
     *
     * @throws FileCloneOperationException
     */
    void sisClone(OntapFile srcFile, SevenModeCloneSpec cloneSpec) throws FileCloneOperationException {
        File destFile = cloneSpec.getDestination();
        boolean createSnapshotFirst = cloneSpec.createSnapshot();

        String sourcePath = srcFile.getFullPath();
        String destPath = destFile.getFullPath();

        if (!srcFile.exists()) {
            throw new FileCloneOperationException("Error: The source file " + srcFile.getFullPath() + " does not exist.");
        }
        LOGGER.debug("clone: " + sourcePath + " " + destPath + " starting.");


        String status = "badNews";
        int failures;
        NaElement xo = cloneStart(sourcePath, destPath, createSnapshotFirst);
        if (xo != null) {
            status = xo.getAttr("status");
        }

        if (xo != null && status.equals("passed")) {
            NaElement cloneId = xo.getChildByName("clone-id").getChildByName("clone-id-info");
            failures = monitorCloneStatus(cloneId, srcFile, cloneSpec);
        } else {
            LOGGER.error("clone of " + sourcePath + " to " + destPath + " failed.");
            throw new FileCloneOperationException("clone of " + sourcePath + " to " + destPath + " failed.");
        }
        LOGGER.debug("clone " + sourcePath + " " + destPath + " finished, (errors=" + failures + ")");

        if (failures >= 10) {
            throw new FileCloneOperationException("clone " + sourcePath + " " + destPath + " failed.");
        }
    }

    private NaElement cloneStart(String sourcePath, String destPath, boolean createSnapshotFirst)
            throws FileCloneOperationException {
        NaElement xi;
        NaElement xo = null;
        xi = new NaElement("clone-start");
        xi.addNewChild("source-path", sourcePath);
        xi.addNewChild("destination-path", destPath);
        if (createSnapshotFirst) {
            xi.addNewChild("no-snap", "false");
        } else {
            xi.addNewChild("no-snap", "true");
        }

        String badNews = null;
        boolean notStarted = true;
        int attempts = 0;
        while (notStarted && attempts < 3) {
            attempts++;
            try {
                xo = filer.invokeZapi(xi, ZapiCache.NO_CACHE);
                notStarted = false;
            } catch (ApiFailedException e) {
                if (e.getErrno() == NaErrno.EDENSE_SNAPSHOT_CREATE) {
                    LOGGER.warn("The snapshot request failed.  We will try without the snapshot.");
                    xi = new NaElement("clone-start");
                    xi.addNewChild("source-path", sourcePath);
                    xi.addNewChild("destination-path", destPath);
                    xi.addNewChild("no-snap", "true");
                } else if (e.getErrno() == NaErrno.EDENSE_SHUTDOWN) {
                    LOGGER.warn("Node shutdown in progress.");
                    try {
                        Thread.sleep(30000);
                    } catch (InterruptedException ie) {
                        LOGGER.info("Woke from sleep while waiting on EDENSE_SHUTDOWN.");
                    }
                } else if (e.getErrno() == NaErrno.EDENSE_FAILOVER) {
                    LOGGER.warn("Cluster failover in progress.");
                    try {
                        Thread.sleep(30000);
                    } catch (InterruptedException ie) {
                        LOGGER.info("Woke from sleep while waiting on EDENSE_FAILOVER.");
                    }
                } else {
                    LOGGER.error("clone of " + sourcePath + " to " + destPath + " failed to start. - " + e.getMessage(),
                                 e);
                    badNews = e.getMessage();
                }
            }
        }

        if (notStarted) {
            throw new FileCloneOperationException(badNews);
        }
        return xo;
    }

    private int monitorCloneStatus(NaElement cloneId, OntapFile srcFile, SevenModeCloneSpec cloneSpec) {
        int failures = 0;
        boolean running = true;
        NaElement cloneOpsInfo, xi;
        int percentDone;
        while (running && failures < 10) {
            xi = new NaElement("clone-list-status");
            NaElement ci = new NaElement("clone-id");
            ci.addChildElem(cloneId);
            xi.addChildElem(ci);
            try {
                NaElement xo = filer.invokeZapi(xi, ZapiCache.NO_CACHE);
                filer.getZapiCache().blastCacheValues("file-");
                NaElement cloneOpsStatus = xo.getChildByName("status");
                if (cloneOpsStatus == null) {
                    LOGGER.debug("clone-list-status returned as null.  This is running RR.");
                    break;
                }

                cloneOpsInfo = cloneOpsStatus.getChildByName("ops-info");
                if (!cloneOpsInfo.getChildContent("clone-state").equals("running")) {
                    if (cloneOpsInfo.getChildContent("clone-state").equals("failed")) {
                        int error = cloneOpsInfo.getChildIntValue("error", 0);
                        LOGGER.debug("clone-state is failed.  error=" + error + ". " + cloneOpsInfo);
                        if (error == NaErrno.EDENSE_FAILOVER) {
                            LOGGER.warn("Waiting for clone of " + cloneOpsInfo.getChildContent("destination-file") +
                                                " to start again.  Cluster failover in progress.");
                            failures++;
                            try {
                                Thread.sleep(10000);
                            } catch (InterruptedException e) {
                                LOGGER.info("Woke from sleep while waiting on EDENSE_FAILOVER.");
                            }
                        } else if (error == NaErrno.EDENSE_SHUTDOWN) {
                            LOGGER.warn("Waiting for clone of " + cloneOpsInfo.getChildContent("destination-file") +
                                                " to start again.  Node shutdown in progress.");
                            failures++;
                            try {
                                Thread.sleep(10000);
                            } catch (InterruptedException e) {
                                LOGGER.info("Woke from sleep while waiting on EDENSE_SHUTDOWN.");
                            }
                        } else {
                            LOGGER.error(
                                    "Unable to create clone " + cloneOpsInfo.getChildContent("destination-file") + "." +
                                            " ");
                            running = false;
                            cloneClear(ci);

                            if (cloneOpsInfo.getChildContent("reason") != null) {
                                throw new FileCloneOperationException(cloneOpsInfo.getChildContent("reason"));
                            } else {
                                throw new FileCloneOperationException("Unable to create clone");
                            }
                        }
                    } else {
                        running = false;
                    }
                } else {
                    percentDone = cloneOpsInfo.getChildIntValue("percent-done", 0);
                    long blocksCopied = cloneOpsInfo.getChildLongValue("blocks-copied", 0L);
                    double totalBlocks = cloneOpsInfo.getChildLongValue("total-blocks", 0L);
                    double percentCopied = (((double) blocksCopied / totalBlocks) * 100);

                    LOGGER.debug(cloneOpsInfo.getChildContent("destination-file") + " clone is " +
                                         cloneOpsInfo.getChildContent("clone-state") + " and is " + percentDone +
                                         "% done. " + blocksCopied + " blocks have required duplication.");

                    if (cloneSpec.getStatusQueue() != null) {
                        try {
                            cloneSpec.getStatusQueue().add(percentDone);
                        } catch (IllegalStateException e) {
                            LOGGER.error(
                                    "IllegalStateException caught while adding percentDone to statusQueue.  percentDone=" +
                                            percentDone);
                        }
                    }

                    if ((percentDone < 99) && (blocksCopied > 0)) {
                        LOGGER.info(cloneOpsInfo.getChildContent("destination-file") + " hit max shared blocks.");

                        if (cloneSpec.fallBackToCopy() && (percentCopied > cloneSpec.getFallBackThreshold())) {
                            LOGGER.debug("Requesting copy in place of clone " + cloneId);

                            try {
                                LOGGER.debug("requesting clone-stop on " + cloneId);
                                xi = new NaElement("clone-stop");
                                xi.addChildElem(ci);
                                filer.invokeZapi(xi, ZapiCache.NO_CACHE);
                            } catch (Exception e) {
                                LOGGER.warn("clone-stop request returned " + e.getMessage() + " for " + cloneId);
                            }
                            if (cloneSpec.getStatusQueue() != null) {
                                NdmpCopyProcess copyProcess = ndmpCopy(srcFile, cloneSpec.getDestination(), null, true);
                                updateStatusQueue(cloneSpec.getStatusQueue(), copyProcess);
                            } else {
                                ndmpCopy(srcFile, cloneSpec.getDestination(), null, false);
                            }
                            running = false;
                        } else {
                            LOGGER.debug(
                                    "We either haven't hit the threshold or the user does not want to fall back to ndmp copy");
                        }
                    }

                    try {
                        Thread.sleep(1000);
                    } catch (InterruptedException e) {
                        LOGGER.info("Woke from sleep.");
                    }
                }
            } catch (NaAuthenticationException e) {
                LOGGER.error("NaAuthenticationException - get clone status of " + srcFile.getFullPath() + " to " +
                                     cloneSpec.getDestination().getFullPath() + " failed. " + e.getMessage());
                failures = 99;
            } catch (ApiFailedException e) {
                if (e.getErrno() == NaErrno.EDENSE_SHUTDOWN) {
                    LOGGER.warn("Waiting for clone operation to start again.  Node shutdown in progress.");
                    try {
                        Thread.sleep(10000);
                    } catch (InterruptedException ie) {
                        LOGGER.info("Woke from sleep while waiting on EDENSE_SHUTDOWN.");
                    }
                } else if (e.getErrno() == NaErrno.EDENSE_FAILOVER) {
                    LOGGER.warn("Waiting for clone operation to start again.  Cluster failover in progress.");
                    try {
                        Thread.sleep(10000);
                    } catch (InterruptedException ie) {
                        LOGGER.info("Woke from sleep while waiting on EDENSE_FAILOVER.");
                    }
                } else {
                    LOGGER.error("NaAPIFailedException - get clone status of " + srcFile.getFullPath() + " to " +
                                         cloneSpec.getDestination().getFullPath() + " failed. " + e.getMessage());
                    failures++;
                }
            } catch (NaProtocolException e) {
                LOGGER.error("NaProtocolException - get clone status of " + srcFile.getFullPath() + " to " +
                                     cloneSpec.getDestination().getFullPath() + " failed. " + e.getMessage());
                failures++;
            } catch (IOException e) {
                LOGGER.error("IOException - get clone status of " + srcFile.getFullPath() + " to " +
                                     cloneSpec.getDestination().getFullPath() + " failed. " + e.getMessage());
                failures++;
            }
        }
        return failures;
    }

    private void updateStatusQueue(LinkedBlockingQueue<Integer> statusQueue, NdmpCopyProcess copyProcess) {
        while (!copyProcess.isComplete()) {
            statusQueue.add(copyProcess.getPercentComplete());
            Utils.sleep(2000);
        }
    }

    private void cloneClear(NaElement cloneID) throws NaAuthenticationException, ApiFailedException, NaProtocolException, IOException {
        LOGGER.debug("cloneClear: " + cloneID);

        NaElement xi = new NaElement("clone-clear");
        xi.addChildElem(cloneID);

        NaElement xo = filer.invokeZapi(xi, ZapiCache.NO_CACHE);

        if (xo.getAttr("status").equals("passed")) {
            LOGGER.debug("cloneClear passed.");
        }
    }
}
