package com.netapp.offtap3.internal.ontap.file;

import com.netapp.offtap3.File;
import com.netapp.offtap3.FileOperationException;
import com.netapp.offtap3.StorageSystem;
import com.netapp.offtap3.internal.OfftapProxy;
import com.netapp.offtap3.internal.ontap.SevenModeCloneSpecImpl;
import com.netapp.offtap3.internal.ontap.controller.FilerImpl;
import com.netapp.offtap3.internal.ontap.controller.OntapConnection;
import com.netapp.offtap3.internal.ontap.support.ZapiCache;
import com.netapp.offtap3.internal.ontap.volume.FilerVolumeImpl;
import com.netapp.offtap3.ontap.ApiFailedException;
import com.netapp.offtap3.ontap.CloneSpec;
import com.netapp.offtap3.ontap.FileRange;
import com.netapp.offtap3.ontap.Filer;
import com.netapp.offtap3.ontap.FilerVolume;
import com.netapp.offtap3.ontap.NaElement;
import com.netapp.offtap3.ontap.NdmpCopyProcess;
import com.netapp.offtap3.ontap.OntapFile;
import com.netapp.offtap3.ontap.SevenModeCloneSpec;
import org.apache.log4j.Logger;

import java.util.ArrayList;
import java.util.EnumSet;
import java.util.List;
import java.util.Set;

public class SevenModeFileImpl<F extends Filer> extends OntapFileImpl<OntapFile, F, FilerImpl<F>, FilerVolumeImpl<F>, FilerVolume> {

    private static final Logger LOGGER = Logger.getLogger(SevenModeFileImpl.class);

    private SevenModeFileCopier copier;

    private OntapFile proxy;
    private static final Set<StorageSystem> storageSystems = EnumSet.of(StorageSystem.pfiler, StorageSystem.vfiler);

    private SevenModeFileImpl(FilerImpl<F> controller, FilerVolumeImpl<F> volume, String fullPath) {
        super(controller, volume, fullPath);
        copier = new SevenModeFileCopier(controller);
    }

    /**
     * Creates a <code>File</code> instance with the full path value.
     */
    public static <F extends Filer> OntapFile create(FilerImpl<F> controller, FilerVolumeImpl<F> volume, String path) {
        SevenModeFileImpl file = new SevenModeFileImpl<F>(controller, volume, volume.toFullPath(path));
        OntapFile proxy = (OntapFile) OfftapProxy.create(new Class[] {OntapFile.class, InternalOntapFile.class}, FileProxy.class, file);
        file.proxy = proxy;
        return proxy;
    }

    public OntapFile createFile(String path) {
        return create(getControllerImpl(), getVolumeImpl(), concatPath(path));
    }

    @Override
    public void setFileReservation(boolean on) throws FileOperationException {
        LOGGER.debug("setFileReservation with path: " + getFullPath());
        NaElement xi;
        NaElement xo;
        NaElement xa;

        xi = new NaElement("system-cli");
        xi.addNewChild("priv", "admin");
        xa = new NaElement("args");

        FilerImpl filer = getControllerImpl();
        filer.addSystemCliArgs(xa);

        xa.addNewChild("arg", "file");
        xa.addNewChild("arg", "reservation");
        xa.addNewChild("arg", getFullPath());
        xa.addNewChild("arg", on ? "enable" : "disable");
        xi.addChildElem(xa);
        try {
            xo = filer.getNonPassThroughConnection().invokeZapi(xi, ZapiCache.NO_CACHE);
            int rc = xo.getChildIntValue("cli-result-value", 0);
            String output = xo.getChildContent("cli-output");
            if (rc != 1) {
                LOGGER.error(String.format("Unable to reserve file. rc = %s. output = %s", rc, output));
            }

            if ((output != null) && (output.contains("file: Can't find path"))) {
                throw new FileOperationException("Unable to reserve file. - " +output);
            }
        } catch (ApiFailedException ex) {
            LOGGER.error("Unable to reserve file. - " + ex.getMessage(), ex);
            throw new FileOperationException("Unable to reserve file. - "+ ex.getMessage(), ex);
        }
    }

    @Override
    public List<OntapFile> listFiles() throws FileOperationException {
        List<OntapFile> files = new ArrayList<OntapFile>();
        long maxGrab = getControllerImpl().getOfftapPreferences().getIteratorSize();
        NaElement api = new NaElement("file-list-directory-iter-start");

        api.addChildElem(new NaElement("path", getFullPath()));
        try {
            NaElement xo = getControllerImpl().invokeZapi(api, ZapiCache.NO_CACHE);

            Long records = xo.getChildLongValue("records", 0L);
            String tag = xo.getChildContent("tag");

            boolean done = false;
            // grab is the maximum records to get per iteration.
            Long grab = records;

            while (!done) {
                if (grab > maxGrab) {
                    grab = maxGrab;
                }
                if (records < 1) {
                    break;
                }
                api = new NaElement("file-list-directory-iter-next");
                api.addNewChild("tag", tag);
                api.addNewChild("maximum", grab + "");
                xo = getControllerImpl().invokeZapi(api, ZapiCache.NO_CACHE);
                if (xo != null) {
                    if (xo.getChildByName("files") != null) {
                        List<NaElement> fileInfoList = xo.getChildByName("files").getChildren();
                        for (final NaElement fileInfo : fileInfoList) {
                            String filename = fileInfo.getChildContent("name");
                            if (!".".equals(filename) && !"..".equals(filename)) {
                                files.add(getVolumeImpl().getFile(getFullPath() + "/" + filename));
                            }
                        }
                    }
                }
                records = records - grab;
                if (records < 1) {
                    done = true;
                } else {
                    grab = records;
                }
            }

            api = new NaElement("file-list-directory-iter-end");
            api.addNewChild("tag", tag);
            getControllerImpl().invokeZapi(api, ZapiCache.NO_CACHE);
        }
        catch (ApiFailedException e) {
            throw new FileOperationException("Failure while listing directory for " + getFullPath(), e);
        }

        return files;
    }

    @Override
    public void punchHoles(List<FileRange> holes) throws FileOperationException {
        String filePath = getFullPath();
        LOGGER.debug("Entering 7-mode punchHoles for " + filePath);

        NaElement xi = new NaElement("file-punch-hole");
        xi.addNewChild("path", filePath);

        NaElement xa = new NaElement("hole-info");
        xi.addChildElem(xa);

        int holesPunched = 0;
        for (FileRange zapiHole : asMaxHoleList(holes, FileRange.DEFAULT_MAX_HOLE_SIZE)) {
            NaElement holeRange = new NaElement("hole-range-info");

            holeRange.addNewChild("hole-start", String.valueOf(zapiHole.getOffset()));
            holeRange.addNewChild("hole-size", String.valueOf(zapiHole.getLength()));

            xa.addChildElem(holeRange);
            holesPunched++;

            if((holesPunched % 16) == 0) {
                try {
                    getControllerImpl().invokeZapi(xi, ZapiCache.NO_CACHE);
                } catch (ApiFailedException e) {
                    LOGGER.debug("punchHoles failed " + filePath, e);
                    throw new FileOperationException("punchHoles failed " + filePath +" with error: " +e.getMessage(), e);
                }

                // we just punched 16 holes, so lets rebuild the zapi
                xi = new NaElement("file-punch-hole");
                xi.addNewChild("path", filePath);

                xa = new NaElement("hole-info");
                xi.addChildElem(xa);
            }
        }

        // punch any stragglers here
        if((holesPunched % 16) != 0) {
            try {
                getControllerImpl().invokeZapi(xi, ZapiCache.NO_CACHE);
            } catch (ApiFailedException e) {
                LOGGER.debug("punchHoles failed " + filePath, e);
                throw new FileOperationException("punchHoles failed " + filePath +" with error: " +e.getMessage(), e);
            }
        }

        LOGGER.debug("Exiting 7-mode punchHoles");
    }

    @Override
    public void move(File destination) throws FileOperationException {
        if (getControllerImpl().isDirectVFiler()) {
            // no support for move, why not try copy and delete?
            copy(destination);
            delete();
            return;
        }

        LOGGER.debug("moveFile: " + this + " " + destination);
        NaElement xi;
        NaElement xo;
        NaElement xa;

        xi = new NaElement("system-cli");
        xi.addNewChild("priv", "admin");
        xa = new NaElement("args");

        FilerImpl filer = getControllerImpl();

        OntapConnection localConnection = filer.getNonPassThroughConnection();

        xa.addNewChild("arg", "mv");
        xa.addNewChild("arg", getFullPath());
        xa.addNewChild("arg", destination.getFullPath());
        xi.addChildElem(xa);

        try {
            //Make sure we have something to move.
            if (exists()) {
                xo = localConnection.invokeZapi(xi, ZapiCache.NO_CACHE);
                filer.getZapiCache().blastCacheValues("file-");
                int rc = xo.getChildIntValue("cli-result-value", 0);
                String output = xo.getChildContent("cli-output");

                LOGGER.debug("moveFile (rc): " + rc);
                LOGGER.debug("moveFile (output): " + output);

                if (rc != 1) {
                    LOGGER.error(String.format("Move of %s to %s failed. rc = %s. output = %s", this, destination, rc,
                                               output));
                }

                if (output != null) {
                    //We expect output to NOT be null, but also not be longer than zero
                    if (output.length() > 0) {
                        LOGGER.warn("moveFile result was... " + output.replace("\n", ""));
                        LOGGER.debug("moveFile is checking for source file...");
                        if (exists()) {
                            throw new FileOperationException("moveFile - Move of " + this + " to " + destination + " failed. (Source still exists.) - " + output);
                        }
                        LOGGER.debug("moveFile is checking for destination file...");
                        if (!destination.exists()) {
                            throw new FileOperationException("moveFile - Move of " + this + " to " + destination + " failed. (Destination does not exist.) - " + output);
                        }
                    }
                } else {
                    LOGGER.debug("mv request did not return anything.");
                }
            } else {
                throw new FileOperationException("moveFile - Move of " + this + " to " + destination + " failed. (Source not found.)");
            }
        } catch (ApiFailedException e) {
            LOGGER.error("Move of " + this + " to " + destination + " failed.", e);
            throw new FileOperationException("Move of " + this + " to " + destination + " failed: " +e.getMessage(), e);
        }
    }

    @Override
    public NdmpCopyProcess ndmpCopy(OntapFile newLocation) throws FileOperationException {
        return copier.ndmpCopy(getProxy(), newLocation, null, false);
    }

    @Override
    public NdmpCopyProcess asyncNdmpCopy(final OntapFile newLocation) throws FileOperationException {
        return copier.ndmpCopy(getProxy(), newLocation, null, true);
    }

    @Override
    public void sisClone(OntapFile newLocation) throws FileOperationException {
        copier.sisClone(getProxy(), getCloneSpecBuilder(newLocation).build());
    }

    @Override
    public void sisClone(CloneSpec cloneSpec) {
        copier.sisClone(getProxy(), (SevenModeCloneSpec)cloneSpec);
    }

    @Override
    public SevenModeCloneSpec.Builder getCloneSpecBuilder(File destinationFile) {
        return new SevenModeCloneSpecImpl.Builder(destinationFile);
    }

    @Override
    public OntapFile getProxy() {
        return proxy;
    }

    @Override
    public Set<StorageSystem> getStorageSystems() {
        return storageSystems;
    }
}
