/*
 * 
 * Copyright (c) 2012 NetApp, Inc.
 * All rights reserved.
 */
package com.netapp.offtap3.internal.ontap.file.ndmp;

import com.google.common.annotations.VisibleForTesting;
import com.google.common.collect.ImmutableSet;
import com.netapp.ndmp.NdmpSession;
import com.netapp.ndmp.rpcgen.ndmp_log_message_post;
import org.acplt.oncrpc.XdrAble;

import java.util.concurrent.Callable;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;


/**
 * Asynchronous proxy to a NdmpSession.
 *
 * Each NdmpAsyncSession has its own thread. The thread only exists while the session is connected, to save
 * system resources. There is only one thread; method calls and their callbacks will be processed in the
 * order of execution.
 */
public class NdmpAsyncSession {
    private NdmpSession ndmpSession;

    /**
     * The executor is lazy-initialized in the execute method.
     */
    @VisibleForTesting
    ExecutorService executor;

    /**
     * @see com.netapp.ndmp.NdmpSession#NdmpSession(String, int, int, int)
     */
    public NdmpAsyncSession(String hostname, int port, int connectTimeout, int readTimeout) {
        this.ndmpSession = new NdmpSession(hostname, port, connectTimeout, readTimeout);
    }

    /**
     * @see com.netapp.ndmp.NdmpSession#NdmpSession(String, int, int)
     */
    public NdmpAsyncSession(String hostname, int connectTimeout, int readTimeout) {
        this.ndmpSession = new NdmpSession(hostname, connectTimeout, readTimeout);
    }

    /**
     * Wrap an existing NdmpSession. Be careful not to use the NdmpSession's synchronized methods while
     * the wrapper's async methods are executing in the background!
     */
    public NdmpAsyncSession(NdmpSession session) {
        this.ndmpSession = session;
    }

    /**
     * Be careful not to use the NdmpSession's synchronized methods while the wrapper's async methods
     * are executing in the background!
     */
    @VisibleForTesting
    public NdmpSession getSession() {
        return this.ndmpSession;
    }

    /**
     * @see com.netapp.ndmp.NdmpSession#getExtensionVersion(short)
     */
    public short getExtensionVersion(short extension) {
        return ndmpSession.getExtensionVersion(extension);
    }

    @VisibleForTesting
    <T> void executeAsync(final Callable<T> c, final NdmpAsyncCallback<T> callback) {
        if (executor == null || executor.isShutdown()) {
            executor = Executors.newSingleThreadExecutor();
        }

        executor.execute(new Runnable() {
            public void run() {
                T result;
                try {
                    result = c.call();

                    // Do not call onSuccess inside here; we want to call either onSuccess or onFailure, but
                    // not both.

                } catch (Exception e) {
                    callback.onFailure(e);
                    return;
                }

                callback.onSuccess(result);
            }
        });
    }

    /**
     * @see com.netapp.ndmp.NdmpSession#connect(String, String)
     */
    public void connect(final String login, final String password, final NdmpAsyncCallback<Void> callback) {
        executeAsync(
            new Callable<Void>() {
                public Void call() throws Exception {
                    ndmpSession.connect(login, password);
                    return null;
                }
            },
            callback);
    }

    /**
     * @see com.netapp.ndmp.NdmpSession#connect(String, String, com.google.common.collect.ImmutableSet)
     */
    public void connect(final String login, final String password, final ImmutableSet<Short> extensionList, final NdmpAsyncCallback<Void> callback) {
        executeAsync(
            new Callable<Void>() {
                public Void call() throws Exception {
                    ndmpSession.connect(login, password, extensionList);
                    return null;
                }
            },
            callback);
    }

    /**
     * @see com.netapp.ndmp.NdmpSession#close()
     */
    public void close(final NdmpAsyncCallback<Void> callback) {
        executeAsync(
            new Callable<Void>() {
                public Void call() throws Exception {
                    ndmpSession.close();
                    return null;
                }
            },
            callback);

        executor.shutdown();
    }

    /**
     * @see com.netapp.ndmp.NdmpSession#sendMessage(int, org.acplt.oncrpc.XdrAble, org.acplt.oncrpc.XdrAble)
     *
     * @param reply Modification of this object before the callback is called
     *              will result in undefined behavior. Should be null iff the
     *              message is expected not to reply.
     */
    public <T extends XdrAble> void sendMessage(final int messageCode, final XdrAble request, final T reply, final NdmpAsyncCallback<T> callback) {
        executeAsync(
            new Callable<T>() {
                public T call() throws Exception {
                    return ndmpSession.sendMessage(messageCode, request, reply);
                }
            },
            callback);
    }

    /**
     * @see com.netapp.ndmp.NdmpSession#waitForAndProcessNotification(int, Class)
     */
    public <T extends XdrAble> void waitForAndProcessNotification(final int messageCode, final Class<T> clazz, final NdmpAsyncCallback<T> callback) {
        executeAsync(
            new Callable<T>() {
                public T call() throws Exception {
                    return ndmpSession.waitForAndProcessNotification(messageCode, clazz);
                }
            },
            callback);
    }

    /**
     * @see com.netapp.ndmp.NdmpSession#waitForAndProcessNotification(int, Class, com.netapp.ndmp.rpcgen.ndmp_log_message_post)
     *
     * @param logMessage Modification of this object before the callback is called
     *                   will result in undefined behavior.
     */
    public <T extends XdrAble> void waitForAndProcessNotification(final int messageCode, final Class<T> clazz, final ndmp_log_message_post logMessage, final NdmpAsyncCallback<T> callback) {
        executeAsync(
            new Callable<T>() {
                public T call() throws Exception {
                    return ndmpSession.waitForAndProcessNotification(messageCode, clazz, logMessage);
                }
            },
            callback);
    }
}
