/*
 * 
 *
 * Copyright (c) 2012 NetApp, Inc.
 * All rights reserved.
 */

package com.netapp.offtap3.internal.ontap.file.ndmp;

import com.netapp.ndmp.NdmpException;
import com.netapp.ndmp.NdmpSession;

/*
 Referenced the below sources
 - //depot/prod/ontap/R7.3.7x/prod/common/java/netapp/cmds/ndmpc/NdmpCopyOperation.java
 - //depot/prod/champagne/champagne-aquarius-ms3/src/libjob/restore_common.c
*/

/**
 * NdmpCopyOperation transfers files between storage systems using the Network Data Management
 * Protocol (NDMP). The method, startCopyOperation() is a synchronous operation and it returns
 * after completing the file copy operation.
 */

public class NdmpCopyOperation {


    private NdmpSession srcSession;
    private NdmpSession destSession;

    private String dumpName; // unique data copy relationship name

    private String srcIpAddress;
    private String srcLogin;
    private String srcPassword;

    /**
     * The format of srcPathInfo is /[vserverName]/[volumeName]/.snapshot/[snapshotName]/[srcPath].
     */
    private String srcPathInfo;

    private String destIpAddress;
    private String destLogin;
    private String destPassword;

    /**
     * The format of destPathInfo is /[vserverName]/[volumeName]/[destPath].
     * If [destPath] is not specfied, files are copied to the root directory of the volume.
     */
    private String destPathInfo;

    public NdmpCopyOperation(Builder builder) {
        srcIpAddress = builder.srcIpAddress;
        srcLogin = builder.srcLogin;
        srcPassword = builder.srcPassword;
        srcPathInfo = builder.srcPathInfo;
        destIpAddress = builder.destIpAddress;
        destLogin = builder.destLogin;
        destPassword = builder.destPassword;
        destPathInfo = builder.destPathInfo;
    }

    public void startCopyOperation() throws NdmpException {

        try {
            srcSession = NdmpCopyUtil.getInstance().ndmpConnect(srcIpAddress, srcLogin, srcPassword);
            destSession = NdmpCopyUtil.getInstance().ndmpConnect(destIpAddress, destLogin, destPassword);

            dumpName = NdmpCopyUtil.getInstance().getDumpName(srcIpAddress, srcPathInfo, destIpAddress, destPathInfo);

            NdmpCopyUtil.getInstance().ndmpCopyPrepare(srcSession, srcPathInfo, destSession, destPathInfo);

            performFileCopy();
        } finally {
            cleanup();
        }
    }

    private void processMessages() throws NdmpException {
        /*
         * Wait until copying is done. We look for NOTIFY_DATA_HALTED
         * from both source and destination.
         */
        NdmpCopyUtil.getInstance().waitForDataHaltedNotification(srcSession, srcIpAddress, srcPathInfo, destPathInfo);
        NdmpCopyUtil.getInstance().waitForDataHaltedNotification(destSession, destIpAddress, srcPathInfo, destPathInfo);
    }

    private void performFileCopy() throws NdmpException {
        /*
         * Tell destination to get ready for restore and then tell source to send data to the destination.
         */
        NdmpCopyUtil.getInstance().startDestinationRestore(destSession, destPathInfo, srcPathInfo, dumpName);
        NdmpCopyUtil.getInstance().startSourceDump(srcSession, srcPathInfo, dumpName);
        processMessages();
    }

    private void cleanup() throws NdmpException {
        if (srcSession != null) {
            NdmpCopyUtil.getInstance().cleanup(srcSession);
            srcSession = null;
        }

        if (destSession != null) {
            NdmpCopyUtil.getInstance().cleanup(destSession);
            destSession = null;
        }
    }

    public static class Builder {

        private String srcIpAddress;
        private String srcLogin;
        private String srcPassword;
        private String srcPathInfo;
        private String destIpAddress;
        private String destLogin;
        private String destPassword;
        private String destPathInfo;

        public Builder withSourceIpAddress(String value) {
            srcIpAddress = value;
            return this;
        }

        public Builder withSourceLogin(String value) {
            srcLogin = value;
            return this;
        }

        public Builder withSourcePassword(String value) {
            srcPassword = value;
            return this;
        }

        public Builder withSourcePathInfo(String value) {
            srcPathInfo = value;
            return this;
        }

        public Builder withDestinationIpAddress(String value) {
            destIpAddress = value;
            return this;
        }

        public Builder withDestinationLogin(String value) {
            destLogin = value;
            return this;
        }

        public Builder withDestinationPassword(String value) {
            destPassword = value;
            return this;
        }

        public Builder withDestinationPathInfo(String value) {
            destPathInfo = value;
            return this;
        }

        public NdmpCopyOperation build() {
            return new NdmpCopyOperation(this);
        }
    }

}
