/*
 * 
 * Copyright (c) 2012 NetApp, Inc.
 * All rights reserved.
 */

package com.netapp.offtap3.internal.ontap.file.ndmp;

import com.google.common.collect.ImmutableSet;
import com.netapp.ndmp.NdmpDataHaltedException;
import com.netapp.ndmp.NdmpException;
import com.netapp.ndmp.NdmpSession;
import com.netapp.ndmp.rpcgen.ndmp_addr;
import com.netapp.ndmp.rpcgen.ndmp_addr_type;
import com.netapp.ndmp.rpcgen.ndmp_config_get_host_info_reply;
import com.netapp.ndmp.rpcgen.ndmp_data_abort_reply;
import com.netapp.ndmp.rpcgen.ndmp_data_conn_prepare_reply;
import com.netapp.ndmp.rpcgen.ndmp_data_conn_prepare_request;
import com.netapp.ndmp.rpcgen.ndmp_data_connect_reply;
import com.netapp.ndmp.rpcgen.ndmp_data_connect_request;
import com.netapp.ndmp.rpcgen.ndmp_data_halt_reason;
import com.netapp.ndmp.rpcgen.ndmp_data_listen_reply;
import com.netapp.ndmp.rpcgen.ndmp_data_listen_request;
import com.netapp.ndmp.rpcgen.ndmp_data_start_backup_reply;
import com.netapp.ndmp.rpcgen.ndmp_data_start_backup_request;
import com.netapp.ndmp.rpcgen.ndmp_data_start_recover_reply;
import com.netapp.ndmp.rpcgen.ndmp_data_start_recover_request;
import com.netapp.ndmp.rpcgen.ndmp_data_stop_reply;
import com.netapp.ndmp.rpcgen.ndmp_log_message_post;
import com.netapp.ndmp.rpcgen.ndmp_message;
import com.netapp.ndmp.rpcgen.ndmp_name;
import com.netapp.ndmp.rpcgen.ndmp_notify_data_halted_post;
import com.netapp.ndmp.rpcgen.ndmp_notify_data_read_post;
import com.netapp.ndmp.rpcgen.ndmp_pval;
import com.netapp.ndmp.rpcgen.ndmp_tcp_addr;
import com.netapp.ndmp.rpcgen.ndmp_u_quad;


public class NdmpCopyUtil {

    private static final int CONNECTION_TIMEOUT = 30000; // milliseconds
    private static final int READ_TIMEOUT = 0;

    /*
     * The ONTAP error messages on which we depend for identifying detailed NDMP errors
     */
    public static final String ONTAP_ERROR_MESSAGE_VOL_NOT_FOUND = "ERROR: VOL NOT FOUND";
    public static final String ONTAP_ERROR_MESSAGE_DESTINATION_VOL_OUT_OF_SPACE = "The destination volume is either out of inode/disk space";
    
    private static NdmpCopyUtil instanceOfNdmpCopyUtil;

    public NdmpCopyUtil() { }
    
    public static synchronized NdmpCopyUtil getInstance() {
    	if (instanceOfNdmpCopyUtil == null) {
    		instanceOfNdmpCopyUtil = new NdmpCopyUtil();
    	}
    	return instanceOfNdmpCopyUtil;
    }
    

    /**
     * Tests the reachability of the given NDMP Server and validates the login and password.
     * If the NDMP server is not reachable, it throws an {@link com.netapp.ndmp.NdmpException}.
     */
    public void pingNdmpServer(String ipAddress, String login, String password) throws NdmpException {
        NdmpSession session = ndmpConnect(ipAddress, login, password);
        session.close();
    }

    /**
     * Connect to NDMP server and return NDMP session.
     *
     * @param ipAddress IP address to connect to. For Data ONTAP 8.1, this is the node management IP address.
     * @param login Login name for NDMP.
     * @param password Password for NDMP.
     * @return NDMP session.
     * @throws com.netapp.ndmp.NdmpException when failure to connect to NDMP server occurs.
     */
    public NdmpSession ndmpConnect(String ipAddress, String login, String password) throws NdmpException {
        NdmpSession session = new NdmpSession(ipAddress, CONNECTION_TIMEOUT, READ_TIMEOUT);
        session.connect(login, password, ImmutableSet.of(NdmpSession.CAB_EXT_CLASS));

        validateStorageSystemType(session);

        return session;
    }

    /**
     * Get the address that the destination is listening on (using NDMP_DATA_LISTEN),
     * and tell source to connect to that address (using NDMP_DATA_CONNECT).
     *
     * @param srcSession NDMP session to the source system.
     * @param sourcePath Path to copy from.
     * @param destSession NDMP session to the destination system.
     * @param destinationPath Path to copy to.
     * @throws com.netapp.ndmp.NdmpException on failure.
     */
    public void ndmpCopyPrepare(NdmpSession srcSession, String sourcePath, NdmpSession destSession, String destinationPath) throws NdmpException {
        if (srcSession.getExtensionVersion(NdmpSession.CAB_EXT_CLASS) != 0) {
            sendDataConnPrepare(srcSession, sourcePath);
        }

        if (destSession.getExtensionVersion(NdmpSession.CAB_EXT_CLASS) != 0) {
            sendDataConnPrepare(destSession, destinationPath);
        }

        ndmp_data_listen_request dataListenRequest = new ndmp_data_listen_request();
        dataListenRequest.addr_type = ndmp_addr_type.NDMP_ADDR_TCP;
        ndmp_data_listen_reply dataListenReply = new ndmp_data_listen_reply();
        destSession.sendMessage(ndmp_message.NDMP_DATA_LISTEN, dataListenRequest, dataListenReply);
        if (dataListenReply.connect_addr.tcp_addr.length < 1) {
            throw new NdmpException(Messages.get("ndmp.no.addr.in.data.listen.reply"));
        }

        ndmp_data_connect_reply dataConnectReply = new ndmp_data_connect_reply();
        ndmp_data_connect_request dataConnectRequest = new ndmp_data_connect_request();
        dataConnectRequest.addr = new ndmp_addr();
        dataConnectRequest.addr.addr_type = dataListenReply.connect_addr.addr_type;
        dataConnectRequest.addr.tcp_addr = new ndmp_tcp_addr[1];
        dataConnectRequest.addr.tcp_addr[0] = new ndmp_tcp_addr();
        dataConnectRequest.addr.tcp_addr[0].ip_addr = dataListenReply.connect_addr.tcp_addr[0].ip_addr;
        dataConnectRequest.addr.tcp_addr[0].port = dataListenReply.connect_addr.tcp_addr[0].port;
        dataConnectRequest.addr.tcp_addr[0].addr_env = new ndmp_pval[0];
        srcSession.sendMessage(ndmp_message.NDMP_DATA_CONNECT, dataConnectRequest, dataConnectReply);
    }

    /**
     * Get the "dump name" that can be passed to NDMP_DATA_START_RECOVER_REQUEST.
     *
     * @param sourceIpAddress IP address of the source system.
     * @param sourcePath Path to copy from.
     * @param destinationIpAddress IP address of the destination system.
     * @param destinationPath Path to copy to.
     * @return string representing the dump name.
     */
    public String getDumpName(String sourceIpAddress, String sourcePath, String destinationIpAddress, String destinationPath) {
        return sourceIpAddress + ":" + sourcePath + " " + destinationIpAddress + ":" + destinationPath;
    }

    /**
     * Tell destination storage system to start restoring data.
     *
     * @param destinationSession {@link com.netapp.ndmp.NdmpSession} to the destination system
     * @param destinationPath Path to copy the data to.
     * @param sourcePath Path to copy the data from.
     * @param dumpName Dump name obtained from {@link #getDumpName}
     * @param destinationVolumeName Destination volume name
     * @throws com.netapp.ndmp.NdmpException
     */
    public void startDestinationRestore(NdmpSession destinationSession, String destinationPath, String sourcePath, String dumpName, String destinationVolumeName) throws NdmpException {

        ndmp_data_start_recover_reply dataRecoverReply = new ndmp_data_start_recover_reply();
        ndmp_data_start_recover_request dataRecoverRequest = new ndmp_data_start_recover_request();

        dataRecoverRequest.butype_name = "dump";

        dataRecoverRequest.env = new ndmp_pval[5];

        dataRecoverRequest.env[0] = new ndmp_pval();
        dataRecoverRequest.env[0].name = "DMP_NAME";
        dataRecoverRequest.env[0].value = dumpName;

        dataRecoverRequest.env[1] = new ndmp_pval();
        dataRecoverRequest.env[1].name = "UPDATE";
        dataRecoverRequest.env[1].value = "Y";

        dataRecoverRequest.env[2] = new ndmp_pval();
        dataRecoverRequest.env[2].name = "BASE_DATE";
        dataRecoverRequest.env[2].value = "0";

        dataRecoverRequest.env[3] = new ndmp_pval();
        dataRecoverRequest.env[3].name = "LEVEL";
        dataRecoverRequest.env[3].value = "0";

        dataRecoverRequest.env[4] = new ndmp_pval();
        dataRecoverRequest.env[4].name = "FILESYSTEM";
        dataRecoverRequest.env[4].value = sourcePath;

        dataRecoverRequest.nlist = new ndmp_name[1];
        dataRecoverRequest.nlist[0] = new ndmp_name();
        dataRecoverRequest.nlist[0].original_path = "/";
        dataRecoverRequest.nlist[0].destination_dir = destinationPath + "/";
        dataRecoverRequest.nlist[0].name = "";
        dataRecoverRequest.nlist[0].other_name = "";
        dataRecoverRequest.nlist[0].node = new ndmp_u_quad();
        dataRecoverRequest.nlist[0].node.high = 0;
        dataRecoverRequest.nlist[0].node.low = 0;
        dataRecoverRequest.nlist[0].fh_info = new ndmp_u_quad();
        dataRecoverRequest.nlist[0].fh_info.high = 0;
        dataRecoverRequest.nlist[0].fh_info.low = 0;

        try {
            destinationSession.sendMessage(ndmp_message.NDMP_DATA_START_RECOVER, dataRecoverRequest, dataRecoverReply);
        } catch (NdmpException e) {
            /*
             * NDMP_DATA_START_RECOVER message
             *
             * The reason for the 'NDMP_ILLEGAL_ARGS_ERR' reply error on the NDMP v4 protocol documenation:
             *
             *     "The start recover request specified an invalid recovery method (butype_name), an invalid backup
             *     environment (env) or an invalid file recovery list (nlist). Recovery types and backup environment
             *     variables are Data Server implementation dependent."
             *
             * This means that the 'destination_dir' nlist is not valid, and the NDMP service cannot restore data to the volume.
             * From the below reasons:
             *  - The 'NDMP_CAB_DATA_CONN_PREPARE' message already verified that the destination volume existed.
             *  - When the volume is offline, the NDMP server just sends a log message instead of generating an error message.
             *  - The volume space is not full either.
             *
             * we can assume that the destination volume is not writable.
             *
             */
            if (e.getMessage() != null && e.getMessage().contains("NDMP_ILLEGAL_ARGS_ERR")) {
                throw new RuntimeException(
                        Messages.get("ndmp.destination.volume.not.writable.error", destinationVolumeName), e);
            } else {
                throw e;
            }
        }

        ndmp_notify_data_read_post dataReadNotification;
        boolean destVolumeOffline = false;

        /*
         * When the destination volume is offline, the NDMP service sends a log message first, and then sends a data-halted
         * message with the 'NDMP_DATA_HALT_INTERNAL_ERROR'.
         *
         * The NDMP log message when the destination volume is offline:
         *     "DATA: Operation terminated: Nlist processing failure on RPC error: ERROR: VOL NOT FOUND (for dest_vol_name)"
         */
        do {
            ndmp_log_message_post logMsg = new ndmp_log_message_post();
            try {
                dataReadNotification = destinationSession.waitForAndProcessNotification(ndmp_message.NDMP_NOTIFY_DATA_READ, ndmp_notify_data_read_post.class, logMsg);
            } catch (NdmpDataHaltedException e) {
                if (destVolumeOffline && e.getReasonCode() == ndmp_data_halt_reason.NDMP_DATA_HALT_INTERNAL_ERROR) {
                    throw new RuntimeException(
                            Messages.get("ndmp.data.halt.destination.volume.offline.error", destinationVolumeName), e);
                } else {
                    throw e;
                }
            }

            if (dataReadNotification == null) {
               // storeNdmpLogMessage(messageLogger, logMsg);
                /*
                 * When the destination volume is offline, the below log message is sent first.
                 */
                if (logMsg.entry != null && logMsg.entry.contains(ONTAP_ERROR_MESSAGE_VOL_NOT_FOUND)) {
                    destVolumeOffline = true;
                }
            }
        } while (dataReadNotification == null);
    }

    public void startDestinationRestore(NdmpSession destinationSession, String destinationPath, String sourcePath, String dumpName) throws NdmpException {
        startDestinationRestore(destinationSession, destinationPath, sourcePath, dumpName, null);
    }

    /**
     * Tell source system to start dumping data to the destination.
     *
     * @param sourceSession {@link com.netapp.ndmp.NdmpSession} to the source system
     * @param sourcePath Path to copy data from.
     * @param dumpName Path to copy data to. Use {@link #getDumpName}.
     * @throws com.netapp.ndmp.NdmpException
     */
    public void startSourceDump(NdmpSession sourceSession, String sourcePath, String dumpName) throws NdmpException {

        ndmp_data_start_backup_request dataBackupRequest = new ndmp_data_start_backup_request();
        ndmp_data_start_backup_reply dataBackupReply = new ndmp_data_start_backup_reply();

        dataBackupRequest.butype_name = "dump";

        dataBackupRequest.env = new ndmp_pval[5];
        dataBackupRequest.env[0] = new ndmp_pval();
        dataBackupRequest.env[0].name = "DMP_NAME";
        dataBackupRequest.env[0].value = dumpName;

        dataBackupRequest.env[1] = new ndmp_pval();
        dataBackupRequest.env[1].name = "UPDATE";
        dataBackupRequest.env[1].value = "Y";

        dataBackupRequest.env[2] = new ndmp_pval();
        dataBackupRequest.env[2].name = "BASE_DATE";
        dataBackupRequest.env[2].value = "0";

        dataBackupRequest.env[3] = new ndmp_pval();
        dataBackupRequest.env[3].name = "LEVEL";
        dataBackupRequest.env[3].value = "0";

        dataBackupRequest.env[4] = new ndmp_pval();
        dataBackupRequest.env[4].name = "FILESYSTEM";
        dataBackupRequest.env[4].value = sourcePath;

        sourceSession.sendMessage(ndmp_message.NDMP_DATA_START_BACKUP, dataBackupRequest, dataBackupReply);
    }

    /**
     * Wait for NDMP_NOTIFY_DATA_HALTED from the storage system. This indicates completion of NDMP copy.
     * @param session {@link com.netapp.ndmp.NdmpSession} to source or destination storage system.
     * @param ipAddress IP address of NDMP server for logging.
     * @param sourceVolumeName Source volume name
     * @param destinationVolumeName Destination volume name
     * @throws com.netapp.ndmp.NdmpException
     */
    public void waitForDataHaltedNotification(NdmpSession session, String ipAddress, String sourceVolumeName, String destinationVolumeName) throws NdmpException {
        ndmp_notify_data_halted_post dataHaltedNotification;
        boolean noSpaceLeftError = false;
        do {
            ndmp_log_message_post logMsg = new ndmp_log_message_post();
            dataHaltedNotification = session.waitForAndProcessNotification(ndmp_message.NDMP_NOTIFY_DATA_HALTED, ndmp_notify_data_halted_post.class, logMsg);
            if (dataHaltedNotification == null) {
                //storeNdmpLogMessage(messageLogger, logMsg);
                /*
                 * When the destination volume has no space for restore, the NDMP service sends a log message first, and then sends a data-halted
                 * message with the 'NDMP_DATA_HALT_INTERNAL_ERROR'.
                 *
                 * The NDMP log message when the destination volume has no space for restore:
                 *     "RESTORE: The destination volume is either out of inode/disk space, or the directory . has reached its size limit."
                 *
                 */
                if (logMsg.entry != null && logMsg.entry.contains(ONTAP_ERROR_MESSAGE_DESTINATION_VOL_OUT_OF_SPACE)) {
                    noSpaceLeftError = true;
                }
            } else if (ndmp_data_halt_reason.NDMP_DATA_HALT_SUCCESSFUL != dataHaltedNotification.reason) {
                if (dataHaltedNotification.reason == ndmp_data_halt_reason.NDMP_DATA_HALT_INTERNAL_ERROR) {
                    if (noSpaceLeftError) {
                        throw new RuntimeException(
                                Messages.get("ndmp.data.halt.no.space.left.error", destinationVolumeName));
                    } else {
                        /*
                         * At this point, both the source volume and the destination volume are valid and ready for data copy. The only reason
                         * for the 'NDMP_DATA_HALT_INTERNAL_ERROR' failure could be out of space error.
                         * When the destination volume has no space for restore, the source NDMP server could send a data-halted message first.
                         */
                        //throw new NdmpException(MsgKey.NDMP_DATA_HALT_DUMP_RESTORE_ERROR.localize(sourceVolumeName, destinationVolumeName));
                    	throw new NdmpException(Messages.get("ndmp.data.halt.dump.restore.error", sourceVolumeName,
                                                             destinationVolumeName));
                    }
                } else {
                    throw new NdmpException(getNdmpDataHaltReasonMessage(ipAddress, dataHaltedNotification.reason));
                }
            }
        } while (dataHaltedNotification == null);
    }


    private static class DataHaltedNotificationCallback implements NdmpAsyncCallback<ndmp_notify_data_halted_post> {
        private NdmpAsyncSession session;
        private String ipAddress;
        private String sourceVolumeName;
        private String destinationVolumeName;
        private NdmpAsyncCallback<Void> callback;
        private boolean noSpaceLeftError;
        private Holder<ndmp_log_message_post> logMsgHolder;

        public DataHaltedNotificationCallback(final NdmpAsyncSession session, final String ipAddress, final String sourceVolumeName, final String destinationVolumeName, final NdmpAsyncCallback<Void> callback, Holder<ndmp_log_message_post> logMsgHolder) {
            this.session = session;
            this.ipAddress = ipAddress;
            this.sourceVolumeName = sourceVolumeName;
            this.destinationVolumeName = destinationVolumeName;
            this.callback = callback;
            this.noSpaceLeftError = false;
            this.logMsgHolder = logMsgHolder;
        }

        public void onSuccess(ndmp_notify_data_halted_post haltedPost) {

            if (haltedPost == null) {
               // storeNdmpLogMessage(messageLogger, logMsgHolder.value);
                /*
                 * When the destination volume has no space for restore, the NDMP service sends a log message first, and then sends a data-halted
                 * message with the 'NDMP_DATA_HALT_INTERNAL_ERROR'.
                 *
                 * The NDMP log message when the destination volume has no space for restore:
                 *     "RESTORE: The destination volume is either out of inode/disk space, or the directory . has reached its size limit."
                 *
                 */
                if (logMsgHolder.value.entry != null && logMsgHolder.value.entry.contains(ONTAP_ERROR_MESSAGE_DESTINATION_VOL_OUT_OF_SPACE)) {
                    noSpaceLeftError = true;
                }

                logMsgHolder.value = new ndmp_log_message_post();
                session.waitForAndProcessNotification(ndmp_message.NDMP_NOTIFY_DATA_HALTED, ndmp_notify_data_halted_post.class, logMsgHolder.value, this);
            } else if (haltedPost.reason != ndmp_data_halt_reason.NDMP_DATA_HALT_SUCCESSFUL) {
                if (haltedPost.reason == ndmp_data_halt_reason.NDMP_DATA_HALT_INTERNAL_ERROR) {
                    if (noSpaceLeftError) {
                        callback.onFailure(new NdmpException(
                                Messages.get("ndmp.data.halt.no.space.left.error", destinationVolumeName)));
                    } else {
                        /*
                         * At this point, both the source volume and the destination volume are valid and ready for data copy. The only reason
                         * for the 'NDMP_DATA_HALT_INTERNAL_ERROR' failure could be out of space error.
                         * When the destination volume has no space for restore, the source NDMP server could send a data-halted message first.
                         */
                        callback.onFailure(new NdmpException(
                                Messages.get("ndmp.data.halt.dump.restore.error", sourceVolumeName,
                                             destinationVolumeName)));
                    }
                } else {
                    callback.onFailure(new NdmpException(getNdmpDataHaltReasonMessage(ipAddress, haltedPost.reason)));
                }
            } else {
                callback.onSuccess(null);
            }
        }

        public void onFailure(Exception e) {
            callback.onFailure(e);
        }
    }

    public void asyncWaitForDataHaltedNotification(final NdmpAsyncSession session, final String ipAddress, final String sourceVolumeName, final String destinationVolumeName, final NdmpAsyncCallback<Void> callback) {
        final Holder<ndmp_log_message_post> logMessageHolder = Holder.create(new ndmp_log_message_post());

        final NdmpAsyncCallback<ndmp_notify_data_halted_post> haltedCallback = new DataHaltedNotificationCallback(session, ipAddress, sourceVolumeName, destinationVolumeName, callback, logMessageHolder);

        session.waitForAndProcessNotification(ndmp_message.NDMP_NOTIFY_DATA_HALTED, ndmp_notify_data_halted_post.class, logMessageHolder.value, haltedCallback);
    }



    /**
     * Clean up NDMP session after NDMP copy operation.
     * @param session {@link com.netapp.ndmp.NdmpSession} to source or destination storage system.
     */
    public void cleanup(NdmpSession session) throws NdmpException {
        if (session != null) {
            ndmp_data_stop_reply dataStopReply = new ndmp_data_stop_reply();
            session.sendMessage(ndmp_message.NDMP_DATA_STOP, null, dataStopReply);
            session.close();
        }
    }

    /**
     * Abort an NDMP session.
     * @param session {@link com.netapp.ndmp.NdmpSession} to source or destination storage system.
     * @throws com.netapp.ndmp.NdmpException
     */
    public void abortSession(NdmpSession session) throws NdmpException {
        if (session != null) {
            session.sendMessage(ndmp_message.NDMP_DATA_ABORT, null, new ndmp_data_abort_reply());
        }
    }


    /**
     * Send an NDMP_CAB_DATA_CONN_PREPARE message to the NDMP server which supports CAB extension.
     */
    public void sendDataConnPrepare(NdmpSession session, String fileSystem) throws NdmpException {
        ndmp_data_conn_prepare_request dataConnPrepareRequest = new ndmp_data_conn_prepare_request();
        ndmp_data_conn_prepare_reply dataConnPrepareReply = new ndmp_data_conn_prepare_reply();
        dataConnPrepareRequest.addr_type = ndmp_addr_type.NDMP_ADDR_TCP;
        dataConnPrepareRequest.butype_name = "dump";

        dataConnPrepareRequest.env = new ndmp_pval[1];
        dataConnPrepareRequest.env[0] = new ndmp_pval();
        dataConnPrepareRequest.env[0].name = "FILESYSTEM";
        dataConnPrepareRequest.env[0].value = fileSystem;

        /*
         * NDMP_CAB_DATA_CONN_PREPARE message
         *
         * The reason for the 'NDMP_ILLEGAL_ARGS_ERR' reply error on the NDMP v4 protocol documenation:
         *
         *     "The prepare request specified an invalid or backup type or an invalid backup environment variable
         *     or the FILESYSTEM environment variable was not present."
         *
         * This means that the 'FILESYSTEM' argument is not valid, and the volume for the path does not exist.
         *
         */
        try {
            session.sendMessage(ndmp_message.NDMP_CAB_DATA_CONN_PREPARE, dataConnPrepareRequest, dataConnPrepareReply);
        } catch (NdmpException e) {
            if (e.getMessage() != null && e.getMessage().contains("NDMP_ILLEGAL_ARGS_ERR")) {
                throw new RuntimeException((Messages.get("ndmp.could.not.find.volume.for.the.path", fileSystem)), e);
            } else {
                throw e;
            }
        }
    }



    private static void validateStorageSystemType(NdmpSession s) throws NdmpException {

        ndmp_config_get_host_info_reply hostInfoReply = new ndmp_config_get_host_info_reply();
        s.sendMessage(ndmp_message.NDMP_CONFIG_GET_HOST_INFO, null, hostInfoReply);

        /**
         * TODO: need to add IBM storage systems too
         */
        if (!hostInfoReply.os_type.equals("NetApp")) {
            //throw new NdmpException(MsgKey.NDMP_NON_NETAPP_STORAGE_SYSTEM.localize(hostInfoReply.hostname));
        	throw new NdmpException(Messages.get("ndmp.non.netapp.storage.system", hostInfoReply.hostname));
        }
    }


    static String getNdmpDataHaltReasonMessage(String ipAddress, int reasonCode) {
        if (reasonCode == ndmp_data_halt_reason.NDMP_DATA_HALT_SUCCESSFUL) {
        	return Messages.get("ndmp.data.halt.successful", ipAddress, reasonCode);
        } else if (reasonCode == ndmp_data_halt_reason.NDMP_DATA_HALT_CONNECT_ERROR) {
        	return Messages.get("ndmp.data.halt.connect.error", ipAddress, reasonCode);
        } else if (reasonCode == ndmp_data_halt_reason.NDMP_DATA_HALT_ABORTED) {
        	return Messages.get("ndmp.data.halt.aborted", ipAddress, reasonCode);
        } else if (reasonCode == ndmp_data_halt_reason.NDMP_DATA_HALT_INTERNAL_ERROR) {
        	return Messages.get("ndmp.data.halt.internal.error", ipAddress, reasonCode);
        } else {        	
        	return Messages.get("ndmp.data.halt.error", ipAddress, reasonCode);
        }
    }
}
