package com.netapp.offtap3.internal.ontap.lun;

import com.netapp.offtap3.ControllerResource;
import com.netapp.offtap3.StorageSystem;
import com.netapp.offtap3.factories.DataLoaderException;
import com.netapp.offtap3.internal.HasProxy;
import com.netapp.offtap3.internal.OfftapProxy;
import com.netapp.offtap3.internal.ontap.MCCManager;
import com.netapp.offtap3.internal.ontap.controller.AbstractOntapDataController;
import com.netapp.offtap3.internal.ontap.controller.DataSvmImpl;
import com.netapp.offtap3.internal.ontap.controller.OntapConnection;
import com.netapp.offtap3.internal.ontap.dataloader.SimpleCdotIterativeLoader;
import com.netapp.offtap3.internal.ontap.proxy.LunProxy;
import com.netapp.offtap3.internal.ontap.support.ZapiCache;
import com.netapp.offtap3.utils.StringUtils;
import com.netapp.offtap3.ontap.ApiFailedException;
import com.netapp.offtap3.ontap.CModeLun;
import com.netapp.offtap3.ontap.DataSvm;
import com.netapp.offtap3.ontap.LunBinding;
import com.netapp.offtap3.ontap.LunOperationException;
import com.netapp.offtap3.ontap.NaElement;
import com.netapp.offtap3.ontap.NaErrno;
import com.netapp.offtap3.ontap.OntapLun;
import com.netapp.offtap3.ontap.OntapVolume;
import com.netapp.offtap3.ontap.QosPolicyGroup;
import com.netapp.offtap3.ontap.Snapshot;

import org.apache.log4j.Logger;

import java.util.Collections;
import java.util.Set;

public class CModeLunImpl extends OntapLunImpl<CModeLun, DataSvm, DataSvmImpl> implements CModeLun, HasProxy<CModeLun> {

    private static final int LUN_ATTRIBUTE_SIZE_FRAGMENTATION_VALUE = 3900;

    private static final Logger logger = Logger.getLogger(CModeLunImpl.class);
    private static final Set<StorageSystem> storageSystems = Collections.singleton(StorageSystem.datasvm);
    private CModeLun proxy;
    private String qosPolicyGroupName;
    private long creationTimestamp;
    private boolean spaceAlloc;

    public CModeLunImpl(DataSvmImpl controller) {
        super(controller);
    }

    @Override
    public long getCreationTimestamp() {
        return creationTimestamp;
    }

    public void setCreationTimestamp(long creationTimestamp) {
        this.creationTimestamp = creationTimestamp;
    }
    
    

    public void loadStats() throws DataLoaderException {
        AbstractOntapDataController dataSvm = getControllerImpl();

        NaElement xi = new NaElement("lun-stats-get-iter");
        NaElement query = new NaElement("query");
        xi.addChildElem(query);
        NaElement lunStatsInfo = new NaElement("lun-stats-info");
        query.addChildElem(lunStatsInfo);
        lunStatsInfo.addNewChild("path", getPath());
        lunStatsInfo.addNewChild("vserver",  MCCManager.getInstance().getVServerNameByUuid((DataSvm) dataSvm));
        SimpleCdotIterativeLoader iterativeLoader = new SimpleCdotIterativeLoader(dataSvm, xi);

        try {
            for (NaElement statsXmlElement : iterativeLoader.loadChildren()) {
                LunStatsImpl lunStats = new LunStatsImpl(statsXmlElement, this);
                if (lunStats.getPath().equals(getPath())) {
                    setLunStats(lunStats);
                } else {
                    logger.error(dataSvm + " did not find lun stats for " + this + " but for " + lunStats.getPath());
                }
            }
        } catch (ApiFailedException e) {
            throw new DataLoaderException(dataSvm + " failed to load stats for " + this, e);
        }
    }

    @Override
    public void setAttribute(String key, String value) {
        int counter = 0;
        for (String newValue : StringUtils.splitIntoChunks(value, getEffectiveFragmentationSizeValue())) {
            NaElement lunApi = new NaElement("lun-set-attribute");
            if (!StringUtils.isReal(newValue)) {
                lunApi = new NaElement("lun-unset-attribute");
            }
            lunApi.addNewChild("path", getPath());
            String localKey = key;
            if (counter > 0) {
                localKey = key + "_" + counter;
            }
            lunApi.addNewChild("name", localKey);
            if (StringUtils.isReal(newValue)) {
                lunApi.addNewChild("value", newValue);
            }
            try {
                getControllerImpl().invokeZapi(lunApi, ZapiCache.NO_CACHE);
                getControllerImpl().getZapiCache().blastCacheValues("lun-");
            } catch (Exception e) {
                if (StringUtils.isReal(newValue)) {
                    throw new LunOperationException(
                            getController() + " was unable to set lun attribute of " + key + "=" + value + " on " +
                                    getPath() + ".  Error: " + e.getMessage(), e);
                }
            }
            counter++;
        }
    }

    private int getEffectiveFragmentationSizeValue() {
        if (getController().getVersion().atLeastSN()) {
            return LUN_ATTRIBUTE_SIZE_FRAGMENTATION_VALUE;
        } else {
            return 450;
        }
    }

    @Override
    public void setSpaceAllocation(boolean enable) {
        NaElement input = new NaElement("lun-set-space-alloc");
        input.addNewChild("enable", Boolean.toString(enable));
        input.addNewChild("path", getPath());
        try {
            getControllerImpl().invokeZapi(input, ZapiCache.NO_CACHE);
            getControllerImpl().getZapiCache().blastCacheValues("lun-");
            logger.info("Enabled lun space allocation to " + getPath() + ".");
        } catch (ApiFailedException e) {
            if (e.getErrno() == NaErrno.EAPIPRIVILEGE) { // ignore this error
                logger.warn("Controller reported that this user cannot invoke lun-set-space-alloc: " + e.getMessage());
            } else {
                logger.error("Failed to enable lun space allocation on lun " + getPath() + ".", e);
                throw new LunOperationException(
                        "Failed to enable lun space allocation on lun " + getPath() + ".  Error: " + e.getMessage());
            }
        }
    }

    @Override
    public LunBinding bind(String vvolPePath) throws LunOperationException {
        NaElement moveLunApi = new NaElement("lun-bind");
        moveLunApi.addNewChild("protocol-endpoint-path", vvolPePath);
        moveLunApi.addNewChild("vvol-path", getPath());
        try {
        	NaElement xo = getControllerImpl().invokeZapi(moveLunApi, ZapiCache.NO_CACHE);
            // refreshing these from controller is horribly expensive, we're just going to make a fake one so that everything is wired up correctly.
            LunBinding transientReturnValue = new LunBindingImpl(true, vvolPePath, 1, Long.parseLong(xo.getChildContent("secondary-lun")), MCCManager.getInstance().getVServerNameByUuid(this.getController()), this.getPath(), null);
            return transientReturnValue;

        } catch (Exception e) {
            throw new LunOperationException(
                    "Failed to bind lun from :" + getPath() + "to: " + vvolPePath + ".  Error: " + e.getMessage(), e);
        }
    }

    @Override
    public void unBind(String vvolPePath) throws LunOperationException {
        NaElement moveLunApi = new NaElement("lun-unbind");
        moveLunApi.addNewChild("protocol-endpoint-path", vvolPePath);
        moveLunApi.addNewChild("vvol-path", getPath());
        try {
            getControllerImpl().invokeZapi(moveLunApi, ZapiCache.NO_CACHE);
            getControllerImpl().getZapiCache().blastCacheValues("lun-");
        } catch (Exception e) {
            throw new LunOperationException(
                    "Failed to unbind lun from vvol PE path :" + vvolPePath + ".  Error: " + e.getMessage(), e);
        }
    }

    @Override
    public QosPolicyGroup getQosPolicyGroup() {
        QosPolicyGroup group = getController().getQosPolicyGroup(qosPolicyGroupName);
        if (group == null && StringUtils.isReal(qosPolicyGroupName)) {
            // Maybe the policy groups on the controller were out of date. Let's refresh just in case.
            getController().refresh(ControllerResource.qos);
            group = getController().getQosPolicyGroup(qosPolicyGroupName);
        }
        return group;
    }

    public void setQosPolicyGroup(String qosPolicyGroupName) {
        this.qosPolicyGroupName = qosPolicyGroupName;
    }

    @Override
    public OntapLun createSnapshotClone(String targetLunPath, Snapshot snapshot, boolean spaceReservationEnabled) {
        NaElement input;
        OntapVolume volume = getController().getVolume(getVolumeName());
        String fullTargetLunPath = volume.toFullPath(targetLunPath);
        input = new NaElement("clone-create");
        input.addNewChild("volume", getVolumeName());
        input.addNewChild("source-path", getRelativePath());
        input.addNewChild("snapshot-name", snapshot.getName());
        input.addNewChild("destination-path", targetLunPath);

        try {
            getControllerImpl().invokeZapi(input, ZapiCache.NO_CACHE);
            getControllerImpl().getZapiCache().blastCacheValues("lun-");
            getController().refresh(ControllerResource.luns);
            return getController().getLun(fullTargetLunPath);
        } catch (ApiFailedException e) {
            throw new LunOperationException(
                    "Error while creating clone for lun " + getPath() + " from snapshot " + snapshot.getName());
        }

    }

    @Override
    public synchronized CModeLun getProxy() {
        if (proxy == null) {
            proxy = OfftapProxy.create(CModeLun.class, LunProxy.class, this);
        }
        return proxy;
    }

	@Override
    public Set<StorageSystem> getStorageSystems() {
        return storageSystems;
    }

    @Override
    public OntapConnection getConnection() {
        return getControllerImpl().getConnection();
    }
    
    
    @Override
    public boolean isSpaceAlloc() {
        return spaceAlloc;
    }    
    
    public void setSpaceAlloc(boolean spaceAlloc) {
    	this.spaceAlloc = spaceAlloc;
    }

    @Override
    public CModeLun refresh() {
        return (CModeLun) super.refresh();
    }
}
