package com.netapp.offtap3.internal.ontap.lun;

import java.util.Set;

import org.apache.log4j.Logger;

import com.google.common.base.Objects;
import com.netapp.offtap3.File;
import com.netapp.offtap3.LunResource;
import com.netapp.offtap3.Refreshable;
import com.netapp.offtap3.Resource;
import com.netapp.offtap3.Size;
import com.netapp.offtap3.factories.DataLoaderException;
import com.netapp.offtap3.internal.HasProxy;
import com.netapp.offtap3.internal.LunImpl;
import com.netapp.offtap3.internal.ontap.HasOntapController;
import com.netapp.offtap3.internal.ontap.controller.AbstractOntapDataController;
import com.netapp.offtap3.internal.ontap.support.ZapiCache;
import com.netapp.offtap3.ontap.ApiFailedException;
import com.netapp.offtap3.ontap.OntapIGroup;
import com.netapp.offtap3.ontap.LunOperationException;
import com.netapp.offtap3.ontap.LunStats;
import com.netapp.offtap3.ontap.NaElement;
import com.netapp.offtap3.ontap.OntapDataController;
import com.netapp.offtap3.ontap.OntapFile;
import com.netapp.offtap3.ontap.OntapLun;
import com.netapp.offtap3.ontap.OntapVolume;

public abstract class OntapLunImpl<I extends OntapLun, C extends OntapDataController, M extends AbstractOntapDataController<C>> extends
        LunImpl implements OntapLun, HasProxy<I>, HasOntapController<C, M> {

    private static final Logger logger = Logger.getLogger(OntapLunImpl.class);
    private final M controller;
    private String volumeName;
    private String qtree;
    private boolean spaceReserve = false;
    private boolean multiStateAlua;
    private long prefixBytes;
    private String shareState;
    private boolean mapped;
    private int deviceId;
    // Added for MHC compatibility.  This flag should eventually get removed when MHC gets fixed!
    // TODO Remove this when MHC data model gets fixed.
    private OntapIGroup lastIGroup;
    private long maxResizeSize;
    private String lunClass;
    private LunStats lunStats = null;

    protected OntapLunImpl(M controller) {
        super(controller);
        this.controller = controller;
    }

    @Override
    public C getController() {
        return controller.getProxy();
    }

    @Override
    public M getControllerImpl() {
        return controller;
    }

    public abstract void loadStats() throws DataLoaderException;

    @Override
    public String getQtree() {
        return qtree;
    }

    public void setQtree(String qtree) {
        this.qtree = qtree;
    }

    @Override
    public boolean isSpaceReserve() {
        return spaceReserve;
    }

    public void setSpaceReserve(boolean spaceReserve) {
        this.spaceReserve = spaceReserve;
    }

    @Override
    public String getVolumeName() {
        return volumeName;
    }

    public void setVolumeName(String volumeName) {
        if (volumeName != null) {
            volumeName = volumeName.intern();
        }
        this.volumeName = volumeName;
    }

    @Override
    public boolean isMultiStateAlua() {
        return multiStateAlua;
    }

    public void setMultiStateAlua(boolean b) {
        this.multiStateAlua = b;
    }

    @Override
    @SuppressWarnings("unchecked")
    public Set<OntapIGroup> getIgroups() {
        return (Set<OntapIGroup>)super.getIgroups();
    }

    @Override
    public OntapIGroup getLastIGroup() {
        return lastIGroup;
    }

    public void setLastIGroup(OntapIGroup lastIGroup) {
        this.lastIGroup = lastIGroup;
    }

    public String getRelativePath() {
        String relativePath = getPath();
        String fullPathPrefix = "/vol/" + volumeName;
        if (relativePath.startsWith(fullPathPrefix)) {
            relativePath = relativePath.substring(fullPathPrefix.length() + 1);
        }
        return relativePath;
    }

    @Override
    public String getShareState() {
        return shareState;
    }

    public void setShareState(String shareState) {
        if (shareState != null) {
            shareState = shareState.intern();
        }
        this.shareState = shareState;
    }

    @Override
    public boolean isMapped() {
        return mapped;
    }

    public void setMapped(boolean mapped) {
        this.mapped = mapped;
    }

    @Override
    public int getDeviceId() {
        return deviceId;
    }

    public void setDeviceId(int deviceId) {
        this.deviceId = deviceId;
    }

    @Override
    public long getMaxResizeSize() {
        return maxResizeSize;
    }

    public void setMaxResizeSize(long maxResizeSize) {
        this.maxResizeSize = maxResizeSize;
    }

    @Override
    public long getPrefixBytes() {
        return prefixBytes;
    }

    public void setPrefixBytes(long prefixBytes) {
        this.prefixBytes = prefixBytes;
    }

    @Override
    public LunStats getLunStats() {
        return lunStats;
    }

    public void setLunStats(LunStats lunStats) {
        this.lunStats = lunStats;
    }

    @Override
    public void setAttribute(String name, boolean on) {
        setAttribute(name, String.valueOf(on));
    }

    @Override
    public String getAttribute(String name) {
        NaElement lunApi = new NaElement("lun-get-attribute");
        lunApi.addNewChild("path", getPath());
        lunApi.addNewChild("name", name);

        try {
            NaElement xo = getControllerImpl().invokeZapi(lunApi, ZapiCache.SHORT_CACHE);
            return xo.getChildContent("value");
        } catch (ApiFailedException nfe) {
            if (nfe.getMessage() != null && nfe.getMessage().contains("No such attribute")) {
                return null;
            } else {
                throw new LunOperationException("Error occurred while retrieving value for attribute " + name, nfe);
            }
        }
    }

    @Override
    public void offline() {
        NaElement offlineLunApi = new NaElement("lun-offline");
        offlineLunApi.addNewChild("path", getPath());
        try {
            controller.invokeZapi(offlineLunApi, ZapiCache.NO_CACHE);
            controller.getZapiCache().blastCacheValues("lun-");
            logger.info(controller + " offlined lun:" + getPath() + ".");
        } catch (ApiFailedException e) {
            //9014 - already offline error
            if (e.getErrno() != 9014) {
                logger.error("Failed to offline " + getPath() + ".", e);
                throw new LunOperationException(
                        controller + " failed to offline lun " + getPath() + ".  Error: " + e.getMessage(), e);
            }
        }
    }

    @Override
    public void online() {
        NaElement offlineLunApi = new NaElement("lun-online");
        offlineLunApi.addNewChild("path", getPath());
        try {
            controller.invokeZapi(offlineLunApi, ZapiCache.NO_CACHE);
            controller.getZapiCache().blastCacheValues("lun-");
            logger.info(controller + " onlined lun:" + getPath() + ".");
        } catch (ApiFailedException e) {
            //9015 - already online error
            if (e.getErrno() != 9015) {
                logger.error("Failed to online " + getPath() + ".", e);
                throw new LunOperationException(
                        controller + " failed to online lun " + getPath() + ".  Error: " + e.getMessage(), e);
            }
            else {
                logger.debug("The lun with path " + getPath() + " is already online.");
            }
        }
    }

    @Override
    public Size resize(Size newSize) {
        logger.debug("resizeLun: " + getPath() + " to " + newSize.toMB() + " MB.");

        NaElement resizeLunApi = new NaElement("lun-resize");
        resizeLunApi.addNewChild("path", getPath());
        resizeLunApi.addNewChild("size", String.valueOf(newSize.toBytes()));
        resizeLunApi.addNewChild("force", Boolean.toString(true));
        NaElement resizeApiOutput;

        try {
            resizeApiOutput = getControllerImpl().invokeZapi(resizeLunApi, ZapiCache.NO_CACHE);
            getControllerImpl().getZapiCache().blastCacheValues("lun-");
            logger.info("Resize lun:" + getPath() + ".");
            Size retVal = Size.createB(resizeApiOutput.getChildLongValue("actual-size", -1));
            setTotalSize(retVal.toBytes());
            return retVal;
        } catch (Exception e) {
            logger.error("Failed to resize lun " + getPath() + ".", e);
            throw new LunOperationException(
                    controller + " failed to resize lun " + getPath() + ".  Error: " + e.getMessage(), e);
        }
    }

    @Override
    public File getFile() {
        OntapVolume vol = getController().getVolume(getVolumeName());
        return vol.getFile(getPath());
    }

    @Override
    public void move(File newLocation) {
        // assert that this is within the same volume!
        if (newLocation instanceof OntapFile) {
            OntapVolume destVol = ((OntapFile) newLocation).getVolume();
            OntapVolume thisVol = controller.getVolume(getVolumeName());
            if (!destVol.equals(thisVol)) {
                throw new IllegalArgumentException(
                        "cross-volume move is not supported! src=" + thisVol + " dest=" + destVol);
            }
        }
        NaElement xi = new NaElement("lun-move");
        xi.addNewChild("path", getPath());
        xi.addNewChild("new-path", newLocation.getFullPath());
        try {
            controller.invokeZapi(xi, ZapiCache.NO_CACHE);
            controller.getZapiCache().blastCacheValues("lun-");
            logger.info("moved  " + getPath() + " " + newLocation.getFullPath());
            controller.getLuns().remove(getPath());
            setPath(newLocation.getFullPath());
            controller.getLuns().put(newLocation.getFullPath(), this);
        } catch (Exception e) {
            logger.error(controller + " failed to move lun " + getPath() + "to " + newLocation.getFullPath(), e);
            throw new LunOperationException(
                    controller + " failed to move lun " + getPath() + "to " + newLocation.getFileType() + ". Error: " + e.getMessage(),
                    e);
        }

    }

    @Override
    public void setComment(String comment) {
        NaElement xi = new NaElement("lun-set-comment");
        xi.addNewChild("comment", comment);
        xi.addNewChild("path", getPath());
        try {
            getControllerImpl().invokeZapi(xi, ZapiCache.NO_CACHE);
            getControllerImpl().getZapiCache().blastCacheValues("lun-");
            logger.info("Added comment to " + getPath() + ".");
        } catch (Exception e) {
            logger.error("Failed to set comment on lun " + getPath() + ".", e);
            throw new LunOperationException(
                    controller + " failed to set comment on lun " + getPath() + ".  Error: " + e.getMessage(), e);
        }
    }

    @Override
    public void setSpaceReservation(boolean enabled) {
        NaElement xi = new NaElement("lun-set-space-reservation-info");
        xi.addNewChild("enable", Boolean.toString(enabled));
        xi.addNewChild("path", getPath());
        try {
            getControllerImpl().invokeZapi(xi, ZapiCache.NO_CACHE);
            getControllerImpl().getZapiCache().blastCacheValues("lun-");
        } catch (ApiFailedException e) {
            throw new LunOperationException(getController() + " failed to set space reservation for " + this, e);
        }
    }

    @Override
    public long getOccupiedSize() {
        NaElement xi = new NaElement("lun-get-occupied-size");
        xi.addNewChild("path", getPath());
        try {
            NaElement xo = getControllerImpl().invokeZapi(xi, ZapiCache.SHORT_CACHE);
            return xo.getChildLongValue("occupied-size", 0);
        } catch (ApiFailedException e) {
            throw new LunOperationException(controller + " failed to get occupied size for " + this, e);
        }
    }

    @Override
    public void writeRaw(long offset, String data) {
        if (data == null) {
            throw new NullPointerException("Raw lun data must not be null");
        }

        NaElement xi = new NaElement("lun-write-raw");
        xi.addNewChild("path", getPath());
        xi.addNewChild("offset", Long.toString(offset));
        xi.addNewChild("length", "" + data.length());
        xi.addNewChild("encoded-data", data);
        try {
            getControllerImpl().invokeZapi(xi, ZapiCache.NO_CACHE);
            getControllerImpl().getZapiCache().blastCacheValues("lun-read");

        } catch (ApiFailedException e) {
            throw new LunOperationException(
                    "lun-write-raw failed to write to path " + getPath() + " with error: " + e.getMessage(), e);
        }
    }

    @Override
    public String readRaw(long offset, long length) {
        logger.debug("readRaw :" + getPath());
        NaElement xi = new NaElement("lun-read-raw");
        String base64data = null;

        if (length == 0) {
            length = getTotalSize().toBytes();
        }

        xi.addNewChild("path", getPath());
        xi.addNewChild("offset", Long.toString(offset));
        xi.addNewChild("length", Long.toString(length));

        try {
            NaElement xo = getControllerImpl().invokeZapi(xi, ZapiCache.SHORT_CACHE);
            if (xo.getAttr("status").equals("passed")) {
                base64data = xo.getChildContent("encoded-data");
            }
        } catch (ApiFailedException e) {
            logger.error(this + " readRaw failed " + getPath(), e);
            throw new LunOperationException(this + " readRaw failed " + getPath() + " error: " + e.getMessage(), e);
        }

        return base64data;
    }

    @Override
    public void map(OntapIGroup igroup) {
        map(igroup, false);
    }

    @Override
    public void map(OntapIGroup igroup, boolean force) {
        logger.debug("mapLun: " + getPath() + " to igroup:" + igroup + "using flag force:" + force);
        int newLunId = -1;
        NaElement mapLunApi = new NaElement("lun-map");
        mapLunApi.addNewChild("path", getPath());
        mapLunApi.addNewChild("initiator-group", igroup.getName());

        if (igroup.getLunId() > -1) {
            mapLunApi.addNewChild("lun-id", String.valueOf(igroup.getLunId()));
        }

        if (force) {
            mapLunApi.addNewChild("force", "true");
        }

        try {
            NaElement mapApiOutput = controller.invokeZapi(mapLunApi, ZapiCache.NO_CACHE);
            controller.getZapiCache().blastCacheValues("lun-");
            newLunId = mapApiOutput.getChildIntValue("lun-id-assigned", 0);
            logger.info("Mapped " + getPath() + " to " + newLunId + ".");
        } catch (ApiFailedException e) {
            if (e.getErrno() == 9023) {
                logger.warn(
                        "Controller reported that this lun (" + getPath() + ") is already mapped to this group (" + igroup + ").");
            } else {
                logger.error("Failed to map lun " + getPath() + " to " + igroup + ".", e);
                throw new LunOperationException(
                        controller + " failed to map lun " + getPath() + ".  Error: " + e.getMessage(), e);
            }
        }

        getProxy().refresh(LunResource.map_list);

        logger.debug("MapLun Completed for lun:" + getPath() + "with id:" + newLunId);
    }

    @Override
    public void unmap(OntapIGroup igroup) {
        logger.debug("unmapLun: " + getPath() + " for igroup:" + igroup);
        NaElement unmapLunApi = new NaElement("lun-unmap");
        unmapLunApi.addNewChild("path", getPath());
        unmapLunApi.addNewChild("initiator-group", igroup.getName());

        try {
            controller.invokeZapi(unmapLunApi, ZapiCache.NO_CACHE);
            controller.getZapiCache().blastCacheValues("lun-");
            logger.info("UnMap lun: " + getPath() + ".");
        } catch (ApiFailedException e) {
            if (e.getErrno() == 9016) {
                logger.warn(
                        "Controller reported that this lun (" + getPath() + ") was not mapped to this group (" + igroup + ").");
            } else {
                logger.error("Failed to unmap lun " + getPath() + ".", e);
                throw new LunOperationException(
                        controller + " failed to unmap lun " + getPath() + ".  Error: " + e.getMessage(), e);
            }
        }
        getProxy().refresh(LunResource.map_list);

        logger.debug("UnMapLun Completed for lun:" + getPath());
    }

    @Override
    public void destroy() {
        String lunPath = getPath();
        logger.debug(getController() + " deletelun: " + lunPath);

        // dsarisky - 10/27/2011 - in C-Mode, offlining the lun and deleting it causes a failure of
        // netapp.manage.NaAPIFailedException: Failed to delete
        // lun/vol/nfs_vs_vol1/lun-1319719362484. Error: Read timed out (errno=51) (errno=911)
        // so we will only delete for 7-mode - and presume the the "force=true" will handle the
        // internals of offlining and whacking it
        NaElement deleteLunApi = new NaElement("lun-destroy");
        deleteLunApi.addNewChild("path", lunPath);
        deleteLunApi.addNewChild("force", "true");

        try {
            controller.invokeZapi(deleteLunApi, ZapiCache.NO_CACHE);
            controller.getLuns().remove(lunPath);
            controller.getZapiCache().blastCacheValues("lun-");
            logger.info(controller + " deleted lun:" + lunPath + ".");
        } catch (ApiFailedException e) {
            logger.error(controller + " failed to delete lun " + lunPath + ".", e);
            throw new LunOperationException(
                    controller + " failed to delete lun" + lunPath + ".  Error: " + e.getMessage(), e);
        }
    }

    @Override
    public void refresh(Resource... resources) {
        // Handled by the proxy.
    }

    @Override
    public Refreshable refresh(Refreshable refreshable) {
        // Handled by the proxy
        return null;
    }

    @Override
    public OntapLun refresh() {
        return (OntapLun) super.refresh();
    }

    @Override
    public String toString() {
        return "Lun[name=" + getName() + ", " + (getSerial() != null ? "serial=" + getSerial() + ", " : "") + (getType() != null ? "type=" + getType() + ", " : "") + "totalSize=" + getTotalSize() + "]";
    }

    public boolean equals(Object obj) {
        if (obj instanceof OntapLun) {
            OntapLun other = (OntapLun) obj;
            return Objects.equal(getType(), other.getType()) && Objects.equal(getSerial(), other.getSerial());
        }
        return false;
    }

    public int hashCode() {
        return getSerial().hashCode();
    }

    public String getLunClass() {
        return lunClass;
    }

    public void setLunClass(String lunClass) {
        if (lunClass != null) {
            lunClass = lunClass.intern();
        }
        this.lunClass = lunClass;
    }
}
