package com.netapp.offtap3.internal.ontap.lun;

import com.netapp.offtap3.ControllerResource;
import com.netapp.offtap3.StorageSystem;
import com.netapp.offtap3.factories.DataLoaderException;
import com.netapp.offtap3.internal.HasProxy;
import com.netapp.offtap3.internal.OfftapProxy;
import com.netapp.offtap3.internal.ontap.controller.AbstractOntapDataController;
import com.netapp.offtap3.internal.ontap.controller.FilerImpl;
import com.netapp.offtap3.internal.ontap.controller.OntapConnection;
import com.netapp.offtap3.internal.ontap.proxy.LunProxy;
import com.netapp.offtap3.internal.ontap.support.ZapiCache;
import com.netapp.offtap3.ontap.ApiFailedException;
import com.netapp.offtap3.ontap.Filer;
import com.netapp.offtap3.ontap.OntapLun;
import com.netapp.offtap3.ontap.LunOperationException;
import com.netapp.offtap3.ontap.OntapVolume;
import com.netapp.offtap3.ontap.Snapshot;
import com.netapp.offtap3.ontap.NaElement;
import org.apache.log4j.Logger;

import java.util.EnumSet;
import java.util.List;
import java.util.Set;

public class SevenModeLunImpl<F extends Filer> extends OntapLunImpl<OntapLun, F, FilerImpl<F>> implements HasProxy<OntapLun> {
    private static final Logger LOGGER = Logger.getLogger(SevenModeLunImpl.class);
    private static final Set<StorageSystem> storageSystems = EnumSet.of(StorageSystem.pfiler, StorageSystem.vfiler);
    private OntapLun proxy;

    public SevenModeLunImpl(FilerImpl<F> controller) {
        super(controller);
    }

    @Override
    public void loadStats() throws DataLoaderException {
        AbstractOntapDataController filer = getControllerImpl();

        NaElement xi = new NaElement("lun-stats-list-info");
        xi.addNewChild("path", getPath());

        try {
            NaElement xo = filer.invokeZapi(xi, ZapiCache.LONG_CACHE);

            if (xo.getChildByName("lun-stats") != null) {
                List<NaElement> pps = xo.getChildByName("lun-stats").getChildren();
                for (final NaElement lunStatsXml : pps) {
                    LunStatsImpl lunStats = new LunStatsImpl(lunStatsXml, this);
                    if (lunStats.getPath().equals(getPath())) {
                        setLunStats(lunStats);
                    } else {
                        LOGGER.error(filer + " did not find lun stats for " + this + " but for " + lunStats.getPath());
                    }
                }
            }
        } catch (ApiFailedException e) {
            throw new DataLoaderException(filer + " failed to get lun stats", e);
        }
    }

    @Override
    public void setAttribute(String attributeName, String enabled) {
        NaElement xi = new NaElement("system-cli");
        xi.addNewChild("priv", "admin");

        NaElement xa = new NaElement("args");

        getControllerImpl().addSystemCliArgs(xa);

        OntapConnection conn = getControllerImpl().getNonPassThroughConnection();
        xa.addNewChild("arg", "lun");
        xa.addNewChild("arg", "set");
        xa.addNewChild("arg", attributeName);
        xa.addNewChild("arg", getPath());
        xa.addNewChild("arg", Boolean.valueOf(enabled) ? "enable" : "disable");
        xi.addChildElem(xa);

        try {
            NaElement xo = conn.invokeZapi(xi, ZapiCache.NO_CACHE);
            getControllerImpl().getZapiCache().blastCacheValues("lun-");
            int rc = xo.getChildIntValue("cli-result-value", 0);
            String output = xo.getChildContent("cli-output");
            if (rc != 1) {
                LOGGER.error(getController() + " was unable to set LUN Attribute (" + attributeName +
                                     ") on " + getPath() + ". Response from server =" + output);
            } else if ((output != null) && output.length() > 1) {
                // this will fail prior to 7.3.2 and on vfilers.  keep rolling if it does.
                LOGGER.warn("Unable to set LUN Attribute." + output);
            } else {
                LOGGER.info(String.format("Set attribute (%s) on LUN (%s) to (%s)", attributeName, getPath(), enabled));
            }
        } catch (ApiFailedException ex) {
            // this will fail prior to 7.3.2.  keep rolling if it does.
            LOGGER.error(String.format("Failed to set attribute (%s) on LUN (%s) to (%s)", attributeName, getPath(),
                                       enabled), ex);
        }
    }

    @Override
    public OntapLun createSnapshotClone(String targetLunPath, Snapshot snapshot, boolean spaceReservationEnabled) {
        NaElement input;
        OntapVolume volume = getController().getVolume(getVolumeName());
        String fullTargetLunPath = volume.toFullPath(targetLunPath);
        input = new NaElement("lun-create-clone");
        input.addNewChild("path", fullTargetLunPath);
        input.addNewChild("parent-lun-path", getPath());
        input.addNewChild("parent-snap", snapshot.getName());
        input.addNewChild("space-reservation-enabled", String.valueOf(spaceReservationEnabled));

        try {
            getControllerImpl().invokeZapi(input, ZapiCache.NO_CACHE);
            getControllerImpl().getZapiCache().blastCacheValues("lun-");
            getController().refresh(ControllerResource.luns);
            return getController().getLun(fullTargetLunPath);
        } catch (ApiFailedException e) {
            throw new LunOperationException(
                    "Error while creating clone for lun " + getPath() + " from snapshot " + snapshot.getName());
        }

    }

    @Override
    public void destroy() {
        // dsarisky - 10/27/2011 - in C-Mode, offlining the lun and deleting it causes a failure of
        // netapp.manage.NaAPIFailedException: Failed to delete
        // lun/vol/nfs_vs_vol1/lun-1319719362484. Error: Read timed out (errno=51) (errno=911)
        // so we will only delete for 7-mode - and presume the the "force=true" will handle the
        // internals of offlining and whacking it
        offline();
        super.destroy();
    }

    @Override
    public synchronized OntapLun getProxy() {
        if (proxy == null) {
            proxy = OfftapProxy.create(OntapLun.class, LunProxy.class, this);
        }
        return proxy;
    }

    @Override
    public Set<StorageSystem> getStorageSystems() {
        return storageSystems;
    }

    @Override
    public OntapConnection getConnection() {
        return getControllerImpl().getConnection();
    }
}
