package com.netapp.offtap3.internal.ontap.snap;

import com.netapp.offtap3.VolumeOperationException;
import com.netapp.offtap3.internal.ontap.controller.AbstractOntapDataController;
import com.netapp.offtap3.internal.ontap.support.ZapiCache;
import com.netapp.offtap3.ontap.ApiFailedException;
import com.netapp.offtap3.ontap.ConsistencyGroup;
import com.netapp.offtap3.ontap.OntapVolume;
import com.netapp.offtap3.ontap.SnapshotOperationException;
import com.netapp.offtap3.ontap.NaElement;

import java.util.ArrayList;
import java.util.List;

public class ConsistencyGroupImpl implements ConsistencyGroup {

    // updated every time "start" is called.  will be used in subsequent "commit"
    private int id;

    private final AbstractOntapDataController controller;
    private final String snapshotName;
    private final List<OntapVolume> volumes;

    public ConsistencyGroupImpl(AbstractOntapDataController controller,
                                List<OntapVolume> volumes, String snapshotName) {
        this.controller = controller;
        this.snapshotName = snapshotName;
        this.volumes = (volumes == null) ? new ArrayList<OntapVolume>() : volumes;
    }

    @Override
    public void start() {
        start(null);
    }

    @Override
    public void start(String optionalTimeout) {
        NaElement xi = new NaElement("cg-start");
        xi.addNewChild("snapshot", snapshotName);
        NaElement volumesElem = new NaElement("volumes");
        if (optionalTimeout != null) {
            xi.addNewChild("timeout", optionalTimeout);
        }
        xi.addChildElem(volumesElem);
        for (OntapVolume currVolume : volumes) {
            volumesElem.addNewChild("volume-name", currVolume.getName());
        }

        try {
            NaElement xo = controller.invokeZapi(xi, ZapiCache.NO_CACHE);
            id = xo.getChildIntValue("cg-id", -1);
        } catch (ApiFailedException e) {
            throw new SnapshotOperationException(this + " failed to start cg named: " + snapshotName + "" +
                    "on volumes: " + volumes, e);
        }
    }

    @Override
    public void orchestrate() {
        orchestrate(null);
    }

    @Override
    public void waflSync() {
        NaElement xi = new NaElement("wafl-sync");

        NaElement volumesElem = new NaElement("volumes");
        xi.addChildElem(volumesElem);
        for (OntapVolume currVolume : volumes) {
            volumesElem.addNewChild("volume-name", currVolume.getName());
        }
        try {
            controller.invokeZapi(xi, ZapiCache.NO_CACHE);
        }
        catch (ApiFailedException e) {
            throw new VolumeOperationException(this + " failed to perform WAFL sync across: " + volumes, e);
        }
    }

    @Override
    public void orchestrate(String optionalTimeout) {
        if (volumes.isEmpty()) {
            throw new UnsupportedOperationException("No volumes specified, orchestrateConsistencyGroup requires explicit volumes to be specified");
        }
        waflSync();
       	start(optionalTimeout);
       	commit();
    }

    @Override
    public void commit() {
        NaElement xi = new NaElement("cg-commit");
        xi.addNewChild("cg-id", String.valueOf(id));
        try {
            controller.invokeZapi(xi, ZapiCache.NO_CACHE);
            controller.getZapiCache().blastCacheValues("snapshot-");
        } catch (ApiFailedException e) {
            throw new SnapshotOperationException(controller + " failed to commit CG " + id +
            " across: " + volumes, e);
        }
    }

    @Override
    public void delete() {
        NaElement xi = new NaElement("cg-delete");
        xi.addNewChild("snapshot", snapshotName);
        NaElement volumesElem = new NaElement("volumes");
        xi.addChildElem(volumesElem);
        for (OntapVolume currVolume : volumes) {
            volumesElem.addNewChild("volume-name", currVolume.getName());
        }
        try {
            controller.invokeZapi(xi, ZapiCache.NO_CACHE);
            controller.getZapiCache().blastCacheValues("snapshot-");
        } catch (ApiFailedException e) {
            throw new SnapshotOperationException(controller + " failed to delete CG " + id +
                    " across: " + volumes, e);
        }
    }

    @Override
    public String getSnapshotName() {
        return snapshotName;
    }

    @Override
    public int getId() {
        return id;
    }

    @Override
    public List<OntapVolume> getVolumes() {
        return volumes;
    }
}
