package com.netapp.offtap3.internal.ontap.snap;

import com.google.common.base.Objects;
import com.netapp.offtap3.ControllerResource;
import com.netapp.offtap3.File;
import com.netapp.offtap3.internal.AbstractRefreshable;
import com.netapp.offtap3.internal.Messages;
import com.netapp.offtap3.internal.ontap.VolumeCloneSpecImpl;
import com.netapp.offtap3.internal.ontap.controller.AbstractOntapDataController;
import com.netapp.offtap3.internal.ontap.controller.DataSvmImpl;
import com.netapp.offtap3.internal.ontap.support.ZapiCache;
import com.netapp.offtap3.internal.ontap.volume.OntapVolumeImpl;
import com.netapp.offtap3.ontap.ApiFailedException;
import com.netapp.offtap3.ontap.OntapDataController;
import com.netapp.offtap3.ontap.OntapVolume;
import com.netapp.offtap3.ontap.Snapshot;
import com.netapp.offtap3.ontap.SnapshotOperationException;
import com.netapp.offtap3.ontap.SpaceReserve;
import com.netapp.offtap3.ontap.NaElement;
import com.netapp.offtap3.ontap.VolumeCloneSpec;
import com.netapp.offtap3.ontap.VolumeCloneSpec.VolumeCloneSpecBuilder;

import org.apache.log4j.Logger;

import java.util.ArrayList;
import java.util.List;

public class SnapshotImpl<V extends OntapVolumeImpl<? extends OntapVolume, C, M>, C extends OntapDataController, M extends AbstractOntapDataController<C>> extends
        AbstractRefreshable implements Snapshot {
    private static final Logger logger = Logger.getLogger(SnapshotImpl.class);

    static final String CLONE_NAME_FORMAT = "%s_%03d";

    private long date;
    private final V volume;
    private final M controller;

    public SnapshotImpl(V volume) {
        super(volume);
        this.volume = volume;
        this.controller = volume.getControllerImpl();
    }

    public long getDate() {
        return date;
    }

    protected M getController() {
        return controller;
    }

    @Override
    public void rename(String newName) {
        NaElement xi = new NaElement("snapshot-rename");
        xi.addNewChild("volume", volume.getName());
        xi.addNewChild("current-name", getName());
        xi.addNewChild("new-name", newName);
        try {
            controller.invokeZapi(xi, ZapiCache.NO_CACHE);
            controller.getZapiCache().blastCacheValues("snapshot-");
            setName(newName);
        }
        catch (ApiFailedException e) {
            String msg = Messages.getInstance().getMessage("snapshot.rename.fail", getName(), newName, controller.toString(), e.getMessage());
            throw new SnapshotOperationException(msg, e);
        }
    }

    @Override
    public void restoreFile(File fileToRestore, File restoreFileLocation) {
        NaElement xi = new NaElement("snapshot-restore-file");
        xi.addNewChild("path", fileToRestore.getFullPath());
        xi.addNewChild("restore-path", restoreFileLocation.getFullPath());
        xi.addNewChild("snapshot", getName());

        try {
            controller.invokeZapi(xi, ZapiCache.NO_CACHE);
            controller.getZapiCache().blastCacheValues("snapshot-", "file-");
        } catch (ApiFailedException e) {
            String msg = Messages.getInstance().getMessage("snapshot.restore.file.fail", fileToRestore, getName(), controller.toString());
            throw new SnapshotOperationException(msg, e);
        }
    }

    @Override
    public void delete() {
        try {
            NaElement xi = new NaElement("snapshot-delete");
            xi.addNewChild("volume", volume.getName());
            xi.addNewChild("snapshot", getName());
            controller.invokeZapi(xi, ZapiCache.NO_CACHE);
            controller.getZapiCache().blastCacheValues("snapshot-");
            volume.getSnapshots().remove(this);
        } catch (ApiFailedException e) {
            String msg = Messages.getInstance().getMessage("snapshot.delete.fail", getName(), volume.getName(), e.getMessage());
            throw new SnapshotOperationException(msg, e);
        }
    }

    @Override
    public void restore() {
        String snapshotName = getName();
        String volumeName = volume.getName();
        logger.debug("restoreFromSnapshot on volume " + volumeName + " with name " + snapshotName);


        NaElement xi = new NaElement("snapshot-restore-volume");
        xi.addNewChild("volume", volumeName);
        xi.addNewChild("snapshot", snapshotName);
        try {
            controller.invokeZapi(xi, ZapiCache.NO_CACHE);
            controller.getZapiCache().blastCacheValues("snapshot-");
        } catch (ApiFailedException e) {
            String msg = Messages.getInstance().getMessage("snapshot.restore.fail", snapshotName, volumeName, e.getMessage());
            throw new SnapshotOperationException(msg, e);
        }
    }

    public void setDate(long date) {
        this.date = date;
    }

    @Override
    public OntapVolume getVolume() {
        return volume.getProxy();
    }
    
    @Override
    public VolumeCloneSpecBuilder createCloneBuilder(int numClones, SpaceReserve spaceReserve) {
        return new VolumeCloneSpecImpl.VolumeCloneSpecBuilderImpl(numClones, spaceReserve);
    }
    
    @Override
    public VolumeCloneSpecBuilder createCloneBuilder(List<String> cloneNames, SpaceReserve spaceReserve) {
        return new VolumeCloneSpecImpl.VolumeCloneSpecBuilderImpl(cloneNames, spaceReserve);
    }

    @Override
    public List<OntapVolume> createVolumeClones(int numClones, SpaceReserve spaceReserve) {
        VolumeCloneSpec spec = new VolumeCloneSpecImpl.VolumeCloneSpecBuilderImpl(numClones, spaceReserve)
            .withJunctionPath(true).build();
        return createVolumeClones(spec);
    }
    
    @Override
    public List<OntapVolume> createVolumeClones(List<String> cloneNames, SpaceReserve spaceReserve) {
        VolumeCloneSpec spec = new VolumeCloneSpecImpl.VolumeCloneSpecBuilderImpl(cloneNames, spaceReserve)
            .withJunctionPath(true).build();
        return createVolumeClones(spec);
        
    }

    @Override
    public List<OntapVolume> createVolumeClones(VolumeCloneSpec spec) {
        List<String> cloneNames = new ArrayList<String>();
        if (spec.getNumClones() != -1) {
            for (int i=0; i < spec.getNumClones(); i++) {
                String clonename = String.format(CLONE_NAME_FORMAT, volume.getName(), i+1);
                cloneNames.add(clonename);
            }
        } else {
            cloneNames = spec.getCloneNames();
        }
        return createVolumeClonesHelper(cloneNames, spec.getSpaceReserve(), spec.isWithJunctionPath());
    }
    
    
    private List<OntapVolume> createVolumeClonesHelper(List<String> cloneNames, SpaceReserve spaceReserve, boolean withJunctionPath) {
        logger.debug("createVolumeClones: " + getVolume().getName() + " " + getName() + " " + spaceReserve);
        int clone = 0;
        List<String> volumeNames = new ArrayList<String>();
        ArrayList<OntapVolume> createdVolumes = new ArrayList<OntapVolume>();

        for (String cloneName : cloneNames) {
            String shortName = volume.getName();

            while (controller.getVolumes().get(cloneName) != null) {
                clone++;
                cloneName = String.format(CLONE_NAME_FORMAT, shortName, clone);
            }

            NaElement xi = new NaElement("volume-clone-create");
            xi.addNewChild("parent-snapshot", getName());
            xi.addNewChild("parent-volume", volume.getName());
            xi.addNewChild("space-reserve", spaceReserve.name());
            xi.addNewChild("volume", cloneName);

            if (withJunctionPath && controller instanceof DataSvmImpl) {
                xi.addNewChild("junction-path", "/" + cloneName);
            }
            
            try {
                getController().invokeZapi(xi, ZapiCache.NO_CACHE);
                getController().getZapiCache().blastCacheValues("volume-");
                getController().getZapiCache().blastCacheValues("lun-");
                getController().getZapiCache().blastCacheValues("file-");
            } catch (ApiFailedException ex) {
                String errMsg = Messages.getInstance().getMessage("snapshot.volume.clone.fail", getName(), getVolume(), ex.getMessage());
                logger.error(errMsg, ex);
                throw new SnapshotOperationException(errMsg, ex);
            }
            volumeNames.add(cloneName);
        }

        // refresh once, and we should find all the volumes
        OntapDataController proxy = getController().getProxy();
        proxy.refresh(ControllerResource.volumes);
        for (String name : volumeNames) {
            createdVolumes.add(proxy.getVolume(name));
        }
        return createdVolumes;
    }

    @Override
    public Snapshot refresh() {
        return (Snapshot) super.refresh();
    }

    @Override
    public boolean equals(Object o) {
        if (o instanceof Snapshot) {
            Snapshot other = (Snapshot) o;
            return Objects.equal(getName(), other.getName()) &&
                    Objects.equal(getVolume(), other.getVolume());
        }
        return false;
    }

    @Override
    public int hashCode() {
        int result = super.hashCode();
        result = 31 * result + (volume.getName() != null ? volume.getName().hashCode() : 0);
        result = 31 * result + (getName() != null ? getName().hashCode() : 0);
        return result;
    }

    @Override
    public String toString() {
        return Objects.toStringHelper(this)
                .add("name", getName())
                .add("volumeName", getVolume().getName())
                .add("date", getDate())
                .toString();
    }




}
