package com.netapp.offtap3.internal.ontap.support;


import com.netapp.offtap3.utils.Pair;
import com.netapp.offtap3.internal.ontap.utils.StatsCollector;
import com.netapp.offtap3.utils.SafeConcurrentHashMap;
import com.netapp.offtap3.utils.StringUtils;
import com.netapp.offtap3.utils.Utils;
import com.netapp.offtap3.ontap.NaElement;

import org.apache.log4j.Logger;

import java.util.ArrayList;
import java.util.Collections;
import java.util.Comparator;
import java.util.Date;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.ConcurrentSkipListSet;
import java.util.concurrent.CopyOnWriteArrayList;
import java.util.concurrent.atomic.AtomicLong;

/**
 * Helper object for supportability and performance - to understand what ZAPI calls we made, how
 * long they took, etc. Used for in-the-field performance improvement, as well as local version
 * without having to delve into running a profiler.
 * 
 * Marking everything as synchronized in external accesses to avoid any carnage on multi-threaed for
 * this mostly debug level class.
 */
public class ZAPIStatsCounter {
    private long loadTime = System.currentTimeMillis();
    private long firstCallTime = -1;
    private final static String TOTAL_CALLS_KEY = "Total";

    private Map<String, List<Long>> zapiCallsToDurations = new SafeConcurrentHashMap<String, List<Long>>();
    private Map<String, List<Date>> zapiCallsToDates = new SafeConcurrentHashMap<String, List<Date>>();
    private Map<String, AtomicLong> zapiCallToDurationSum = new SafeConcurrentHashMap<String, AtomicLong>();
    private Map<String, AtomicLong> zapiCallToDurationCount = new SafeConcurrentHashMap<String, AtomicLong>();

    private Map<String, Long> extraCounters = new SafeConcurrentHashMap<String, Long>();

    private static volatile ZAPIStatsCounter instance = null;
    
    private ThreadLocal<List<String>> recordedKey = new ThreadLocal<List<String>>();
    private ThreadLocal<Long> operationStartTime = new ThreadLocal<Long>();
    
    private Map<String,Set<String>> recordedCalls = new SafeConcurrentHashMap<String,Set<String>>();
    private boolean isSynchronized = true;
    
    private boolean demandRecordedKey = false;
    
    // lets keep track of how many things come into here so that we don't grow absolutely forever.
    private int rolloverCounter = 0;
    private final static int MAX_SIZE_UNTIL_ROLLOVER = 50000;
    private Map<String, Map<String, Pair<AtomicLong, AtomicLong>>> costBreakdowns = new SafeConcurrentHashMap<String, Map<String, Pair<AtomicLong, AtomicLong>>>();
    private boolean costBreakdownMode = false;

    public static ZAPIStatsCounter getInstance() {
        // Double checked locking for speed. This will work since instance variable is volatile. Note that this sort of
        // thing is NOT guaranteed in Java unless you are working with a volatile variable.
        if (instance == null) {
            synchronized (ZAPIStatsCounter.class) {
                if (instance == null) {
                    instance = new ZAPIStatsCounter();
                }
            }
        }
        return instance;
    }
    
    public Map<String, Long> getDurationsAndAverages() {
    	Map<String, Long> retVal = new HashMap<String, Long>();
    	for (String key : zapiCallToDurationSum.keySet()) {
    		try {
	    		long duration = zapiCallToDurationSum.get(key).longValue();
	    		long count = zapiCallToDurationCount.get(key).longValue();
	    		String zapiKey = key;
	    		if(!key.startsWith("zapi") && !key.startsWith("lock") && !key.startsWith("persistence") && !key.startsWith("vsc")) {
	        		zapiKey = "zapi-"+key;
	        	}
	    		retVal.put(zapiKey+"-total", duration);
	    		retVal.put(zapiKey+"-average", duration / count);
    		} catch (Exception e) {
    			// just won't have this one this time, not a big deal
    		}
    	}
    	return retVal;
    }
    
    private void rolloverIfNecessary() {
    	rolloverCounter++;
    	if (rolloverCounter > MAX_SIZE_UNTIL_ROLLOVER) {
    		rolloverCounter = 0;
    		StatsCollector.getInstance().addCounter("rollover-zapis");
    		// we have enough things that we keep forever now, we don't need to write this report.
    		// in addition to it making some noise, it also holds a synchronized lock on actually adding the 
    		// information.
    		//this.writeReportToFinalOutput("Intermediate ZAPIStatsCounter report before rolling over to save memory as of " + new Date(), false);
    		reset(false);
    	}
    }
    
    public synchronized void reset() {
    	reset(true);
    }
    
    public synchronized void reset(boolean full) {
    	rolloverCounter=0;
        zapiCallsToDurations.clear();
        zapiCallsToDates.clear();
        extraCounters.clear();        
        if (full) {
        	recordedCalls.clear();
	        zapiCallToDurationSum.clear();
	        zapiCallToDurationCount.clear();
	        costBreakdowns.clear();
        }
    }

    public synchronized void addZAPICall(NaElement zapiCall, long duration) {
        if (zapiCall != null) {
        	String callName = getZapiPerformanceKey(zapiCall.toString());
        	if(!callName.startsWith("zapi")) {
        		callName = "zapi-"+callName;
        	}
            addZAPICall(callName, duration);
        }
    }
    
    private <L> List<L> getNewList() {
        if (isSynchronized) {
            return new CopyOnWriteArrayList<L>();
        } else {
            return new ArrayList<L>();
        }
    }

    private <S> Set<S> getNewSet() {
        if (isSynchronized) {
            return new ConcurrentSkipListSet<S>();
        } else {
            return new HashSet<S>();
        }
    }

    public void addZAPICall(String callName, long duration) {
    	long startTime = System.currentTimeMillis();
    	synchronized (this) {
    		rolloverIfNecessary();
    	}
    	
    	// not marking the recorded key work synchronized - we certainly might lose a call or two at startup
    	// time, but I can live with that, cost of having this synchronized does start to add up over time
    	List<String> rkList = getRecordedKey();
    	synchronized (rkList) {
	    	for (String myRecordedKey : rkList) {
	    		if (myRecordedKey != null) {
                    Set<String> recordedForThisValue = recordedCalls.get(myRecordedKey);
	    			if (recordedForThisValue == null) {
	    				recordedForThisValue = getNewSet();
	    				recordedCalls.put(myRecordedKey, recordedForThisValue);
	    			}
	    			recordedForThisValue.add(callName);
	    			if (costBreakdownMode) {
	    				Map<String, Pair<AtomicLong, AtomicLong>> costBreakdownsForThisZapi = costBreakdowns.get(myRecordedKey);
	    				if (costBreakdownsForThisZapi == null) {
	    					costBreakdownsForThisZapi = new SafeConcurrentHashMap<String, Pair<AtomicLong, AtomicLong>>();
	    					costBreakdowns.put(myRecordedKey, costBreakdownsForThisZapi);
	    				}
	    				Pair<AtomicLong, AtomicLong> durationsForThisCallOnThisBreakdown = costBreakdownsForThisZapi.get(callName);
	    				if (durationsForThisCallOnThisBreakdown == null) {
	    					durationsForThisCallOnThisBreakdown = new Pair<AtomicLong, AtomicLong>(new AtomicLong(1), new AtomicLong(1));
	    					costBreakdownsForThisZapi.put(callName, durationsForThisCallOnThisBreakdown);
	    				} else {
	    					durationsForThisCallOnThisBreakdown.getA().incrementAndGet();	
	    				}            	
	    				durationsForThisCallOnThisBreakdown.getB().addAndGet(duration);
	    			}
	    		} else {
	    			if (demandRecordedKey) {
	    				throw new IllegalArgumentException("Demand Recorded Key set when we received zapi of " + 
	    						callName + " but no recorded key present - are we running in a thread that didn't transfer over the ThreadLocal information?");
	    			}
	    		}
	    	}
    	}
        if (firstCallTime == -1) {
            firstCallTime = System.currentTimeMillis();
        }
        
        List<Long> durationsForCall = zapiCallsToDurations.get(callName);
        if (durationsForCall == null) {
            durationsForCall = getNewList();
            zapiCallsToDurations.put(callName, durationsForCall);
        }
        List<Date> datesForCall = zapiCallsToDates.get(callName);
        if (datesForCall == null) {
            datesForCall = getNewList();
            zapiCallsToDates.put(callName, datesForCall);
        }        
        if (durationsForCall.size() < 3000 && datesForCall.size() < 3000) {
            // lets not keep > 1000 values for any one thing - otherwise the CopyOnWriteArrayList
            // tends to get too large and shows up in performance profiling.
            durationsForCall.add(duration);
            datesForCall.add(new Date());
        }
        AtomicLong currentSum = zapiCallToDurationSum.get(callName);
        if (currentSum == null) {
        	currentSum = new AtomicLong();
        	zapiCallToDurationSum.put(callName, currentSum);
        }
        currentSum.getAndAdd(duration);
        AtomicLong currentCount = zapiCallToDurationCount.get(callName);
        if (currentCount == null) {
        	currentCount = new AtomicLong();
        	zapiCallToDurationCount.put(callName, currentCount);
        }
        currentCount.incrementAndGet();
        long accountingDuration = System.currentTimeMillis() - startTime;
        StatsCollector.getInstance().addCounter("zsc-accounting-duration", accountingDuration);
    }

    public void addCounter(String key) {
        addCounter(key, 1);
    }

    public void addCounter(String key, long counter) {
    	synchronized(this) {
    		rolloverIfNecessary();
    	}
        Long existingValue = extraCounters.get(key);
        if (existingValue == null) {
            existingValue = 0L;
        }
        existingValue += counter;
        extraCounters.put(key, existingValue);
    }
    
    public void writePeriodicReport() {
	    StatsCollector.getInstance().captureLongTermTrendingData();
		ZAPIStatsCounter.getInstance().writeReportToFinalOutput("Intermediate report at " + new Date() + StringUtils.getNewLine(), true);
    }
    
    public void writeReportToFinalOutput(String header, boolean includeStatsCollector) {
    	if (ZapiCache.isWriteReport()) {
            System.out.println(writeReport(header, includeStatsCollector));
        }
        if (ZapiCache.isWriteReportToLogger()) {
            Logger logger = Logger.getLogger("PERFORMANCE");
            logger.info(writeReport(header, includeStatsCollector));
        }
    }
    
    public String writeReport() {
    	return writeReport("Aeonflux Performance information as of " + new Date() + ".\n", true);
    }
    
    public Map<String, Pair<AtomicLong, AtomicLong>> getCostBreakdownForKey(String key) {
    	return costBreakdowns.get(key);
    }

    // this whole thing used to be synchronized, but that was bad since it could take a bit of time and under truly impressive load
    // and without enough CPU resource from the underlying ESX host it took so long that it got us some watchdog timeouts because it
    // was holding the lock for this object and everyone else was waiting to add their information to it.  So, we will instead copy off
    // the pieces of information we need and then do the slower operations on them without the lock.
    public String writeReport(String header, boolean includeStatsCollector) {
        if (firstCallTime == -1) {
            firstCallTime = System.currentTimeMillis();
        }
        StringBuilder report = new StringBuilder();
        report.append(header);
        long now = System.currentTimeMillis();
        report.append("System has been up for ").append(convertMillisToMinutes(now - loadTime)).append(
                " minutes of uptime and ").append(convertMillisToMinutes(now - firstCallTime)).append(
                " minutes since first call came in.\nSpecific calls:\n");
        Map<String, Long> callsToTotalDurations = null;
        Map<String, List<Long>> masterMap = null;
        Map<String, List<Long>> myZapiCallsToDuration = new HashMap<String, List<Long>>();
        synchronized(this) {
        	callsToTotalDurations = calculateTotalDurations();
        	masterMap = calculateMasterMap();
        	for (Entry<String, List<Long>> keyEntry : zapiCallsToDurations.entrySet()) {
        		String key = keyEntry.getKey();
        		List<Long> theRealList = keyEntry.getValue();
        		ArrayList<Long> myCopyOfRealList = new ArrayList(theRealList);
        		myZapiCallsToDuration.put(key, myCopyOfRealList);
        	}
        }
        report.append(writeLine(TOTAL_CALLS_KEY, masterMap));
        List<String> keysByDuration = Utils.getKeysSortedByValues(callsToTotalDurations);
        Collections.reverse(keysByDuration);
        for (String key : keysByDuration) {
            report.append(writeLine(key, myZapiCallsToDuration));
        }
        Map<String, List<String>> recordedValues = this.getRecordedZapisSorted();
        if (recordedValues.size() > 0) {
            report.append("Recorded Values:\n");
            for (String key : Utils.getSortedList(recordedValues.keySet())) {
                String printableKey = key;
                if (printableKey.length() > 80) {
                    printableKey = printableKey.substring(0, 80);
                }
                report.append(printableKey).append(": ").append(recordedValues.get(key)).append("\n");
                if (costBreakdownMode) {
                	Map<String, Pair<AtomicLong, AtomicLong>> currentCostBreakdown = costBreakdowns.get(key);
                	Set<Long> totals = new HashSet<Long>();
                	if (currentCostBreakdown != null) {
	                	for (Pair<AtomicLong, AtomicLong> longPair : currentCostBreakdown.values()) {
	                		totals.add(longPair.getB().get());
	                	}
	                	List<Long> values = Utils.getSortedList(totals);
	                	Collections.reverse(values);
	                	for (Long curr : values) {
	                		for (String innerKey : currentCostBreakdown.keySet()) {
	                    		if (currentCostBreakdown.get(innerKey).getB().get() == curr) {
	                    			report.append("    Caused-by ").append(innerKey).append(" invoked ").append(
                                            currentCostBreakdown.get(innerKey).getA()).append(
                                            " times at average cost ").append(
                                            currentCostBreakdown.get(innerKey).getA().get() / currentCostBreakdown.get(
                                                    innerKey).getB().get()).append(" for total of ").append(
                                            currentCostBreakdown.get(innerKey).getB().get()).append("\n");
	                    		}
	                    	}
	                	}
                	}
                }
            }
        }
        
        if (extraCounters.size() > 0) {
            report.append("Extra Counters:\n");
            for (String key : Utils.getSortedList(extraCounters.keySet())) {
                String printableKey = key;
                if (printableKey.length() > 80) {
                    printableKey = printableKey.substring(0, 80);
                }
                report.append(printableKey).append(": ").append(extraCounters.get(key)).append("\n");
            }
        }
        if (includeStatsCollector) {
        	report.append(StatsCollector.getInstance().writeReport());
        	StatsCollector.getInstance().writePerformanceCsvIfApplicable();
        }
        return report.toString();
    }
    
    public Map getAllValues() {
        Map<String, List<Long>> total = new HashMap<String, List<Long>>(calculateMasterMap());
        for (String key : zapiCallsToDurations.keySet()) {
            total.put(key, zapiCallsToDurations.get(key));
        }
        return total;
    }
    
    public Map getAllDates() {
        Map<String, List<Date>> total = new HashMap<String, List<Date>>();
        List<Date> wholeThing = getNewList();
        for (List<Date> date : zapiCallsToDates.values()) {
            wholeThing.addAll(date);
        }
        total.put("TOTAL_CALLS_KEY", wholeThing);
        for (String key : zapiCallsToDates.keySet()) {
            total.put(key, zapiCallsToDates.get(key));
        }
        return total;
    }

    private Map<String, List<Long>> calculateMasterMap() {
        List<Long> wholeThing = getNewList();
        for (List<Long> values : zapiCallsToDurations.values()) {
            wholeThing.addAll(values);
        }
        return Utils.toTypedMap(TOTAL_CALLS_KEY, wholeThing);
    }

    private String writeLine(String zapiCallName, Map<String, List<Long>> zapiCallsToDurations) {
        StringBuilder retVal = new StringBuilder();
        List<Long> durations = new ArrayList<Long>(zapiCallsToDurations.get(zapiCallName));

        if (!Utils.isReal(durations)) {
            return "no values for " + zapiCallName;
        }

        Collections.sort(durations);

        long total = 0;
        for (long curr : durations) {
            total += curr;
        }
        long average = total / durations.size();
        retVal.append(zapiCallName).append(" ").append(durations.size()).append(" calls @ ").append(average).append(
                "ms for total of ").append(total).append(" ms.  Low=");
        retVal.append(durations.get(0));
        retVal.append(" ms, High=");
        retVal.append(durations.get(durations.size() - 1));
        retVal.append(" ms.  Specific values were ");
        retVal.append(writeDurationsCompressed(durations));
        retVal.append(" ms.\n");
        return retVal.toString();
    }
    
    public static String writeDurationsCompressed(List<? extends Number> durations) {
    	return writeDurationsCompressed(durations, true);
    }

    public static String writeDurationsCompressed(List<? extends Number> durations, boolean allowSuperCompress) {
        Collections.sort(durations, new Comparator<Number>() {
            @Override
            public int compare(Number o1, Number o2) {
                return o1.intValue() - o2.intValue();
            }
        });
        StringBuilder retVal = new StringBuilder();
        long prevLong = -1;
        int prevValCount = 0;
        int distinctCounts = 0;
        retVal.append("[");
        for (Number currLongNum : durations) {
            long currLong = currLongNum.longValue();
            if (currLong == prevLong) {
                prevValCount++;
            } else {
            	distinctCounts++;
                if (prevLong != -1) {
                    if (prevValCount > 1) {
                        retVal.append(prevLong).append("(").append(prevValCount).append(")");
                    } else {
                        retVal.append(prevLong);
                    }
                    retVal.append(",");
                }
                prevLong = currLong;
                prevValCount = 1;
            }
        }
        if (prevValCount > 0) {
            if (prevValCount > 1) {
                retVal.append(prevLong).append("(").append(prevValCount).append(")");
            } else {
                retVal.append(prevLong);
            }
        }
        retVal.append("]");
        if (distinctCounts > 15 && allowSuperCompress) {
        	return writeDurationsSuperCompressed(durations);
        }
        return retVal.toString();
    }
    
    public static String writeDurationsSuperCompressed(List<? extends Number> durations) {
    	// ok - lets see what level we need to work at.  We can take the last number off, the last 
    	// two numbers, or the last 3 numbers.  Lets keep doing that until we get a 10:1 compression.
    	Set<Long> oneNumberRemoved = new HashSet<Long>();
    	Set<Long> twoNumberRemoved = new HashSet<Long>();
    	for (Number currLongNum : durations) {
    		long currLong = currLongNum.longValue();
    		oneNumberRemoved.add(currLong / 10);
    		twoNumberRemoved.add(currLong / 100);
    	}
    	// the goal here is to super-compress to at least 10:1
    	int divisor = 10;
    	if (oneNumberRemoved.size() > 20) {
    		divisor = 100;
    	}
    	if (twoNumberRemoved.size() > 20) {
    		divisor=1000;
    	}
    	List<Long> numbers = new ArrayList<Long>();
    	for (Number currLongNum : durations) {
            long currLong = currLongNum.longValue();
            currLong = currLong / divisor;
            currLong = currLong * divisor;
            numbers.add(currLong);
    	}
    	return writeDurationsCompressed(numbers, false);
    }

    private Map<String, Long> calculateTotalDurations() {
        Map<String, Long> retVal = new HashMap<String, Long>();
        for (String key : new CopyOnWriteArrayList<String>(zapiCallsToDurations.keySet())) {
            List<Long> calls = zapiCallsToDurations.get(key);
            long total = 0;
            if (calls != null) {
	            for (Long currDuration : calls) {
	                total += currDuration;
	            }
            }
            retVal.put(key, total);
        }
        return retVal;
    }

    private long convertMillisToMinutes(long duration) {
        return duration / 1000 / 60;
    }

    public static String getZapiPerformanceKey(String filerBasedZapi) {
        if (filerBasedZapi == null || !filerBasedZapi.contains("<")) {
            return filerBasedZapi == null ? "" : filerBasedZapi;
        }
        int beginIndex = filerBasedZapi.indexOf("<");
        int candidateEndOne = filerBasedZapi.indexOf(">");
        int candidateEndTwo = filerBasedZapi.indexOf("/>");
        int endIndex = -1;
        if (candidateEndOne == -1) {
            endIndex = candidateEndTwo;
        }
        if (candidateEndTwo == -1) {
            endIndex = candidateEndOne;
        }
        if (candidateEndOne != -1 && candidateEndTwo != -1) {
            endIndex = Math.min(candidateEndOne, candidateEndTwo);
        }
        if (endIndex == -1) {
            return "" + filerBasedZapi;
        }
        return filerBasedZapi.substring(beginIndex + 1, endIndex);
    }
    
    public int getCountMatchingKey(String key) {
    	int retVal = 0;
    	Map<String, AtomicLong> recordedZapis = new HashMap<String, AtomicLong>(zapiCallToDurationCount);
    	for (Map.Entry<String, AtomicLong> entry : recordedZapis.entrySet()) {
    		if (entry.getKey().contains(key)) {
    			retVal += entry.getValue().intValue();
    		}
    	}
    	return retVal;
    }

    public void setRecordedKey(List<String> recordedKeyParam) {
    	if (recordedKeyParam != null && this.isSynchronized && !(recordedKeyParam instanceof CopyOnWriteArrayList)) {
    		List<String> temp = getNewList();
    		temp.addAll(recordedKeyParam);
    		recordedKeyParam = temp;
    	}
        recordedKey.set(recordedKeyParam);
    }
    
    public List<String> getRecordedKey() {
    	if(recordedKey.get() == null) {
    		recordedKey.set(this.<String>getNewList());
    	}
        return (recordedKey.get());
    }
    
    public void setOperationStartTime(Long operationStartTime) {
    	this.operationStartTime.set(operationStartTime);
    }
    
    public Long getOperationStartTime() {
    	return this.operationStartTime.get();
    }

    public Map<String, Set<String>> getRecordedCalls() {
        return new HashMap<String, Set<String>>(recordedCalls);
    }

    public void setRecordedCalls(Map<String, Set<String>> recordedCalls) {
        this.recordedCalls = recordedCalls;
    }

    public synchronized Map<String, List<String>> getRecordedZapisSorted() {
        Map<String, Set<String>> recordedCallsCopy = getRecordedCalls();
        Map<String,List<String>> recordedZapis = new HashMap<String,List<String>>();
        for (Map.Entry<String, Set<String>> entry : recordedCallsCopy.entrySet()) {
            recordedZapis.put(entry.getKey(), Utils.getSortedList(entry.getValue()));
        }
        return recordedZapis;
    }

    public boolean isDemandRecordedKey() {
        return demandRecordedKey;
    }

    public void setDemandRecordedKey(boolean demandRecordedKey) {
        this.demandRecordedKey = demandRecordedKey;
    }

    public void setSynchronized(boolean isSynchronized) {
        this.isSynchronized = isSynchronized;
    }
    
    public void setCostBreakdownMode(boolean val) {
    	costBreakdownMode = val;
    }
    
}
