package com.netapp.offtap3.internal.ontap.support;


import java.io.IOException;
import java.io.Serializable;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Date;
import java.util.HashMap;
import java.util.Map;
import java.util.Set;
import java.util.concurrent.CopyOnWriteArraySet;

import org.apache.log4j.Logger;
import org.xml.sax.SAXException;

import com.netapp.offtap3.internal.ontap.controller.DataSvmImpl;
import com.netapp.offtap3.internal.ontap.utils.StatsCollector;
import com.netapp.offtap3.internal.ontap.zapi.NaElementFactory;
import com.netapp.offtap3.internal.ontap.zapi.ZAPIInvoker;
import com.netapp.offtap3.ontap.NaElement;
import com.netapp.offtap3.utils.Pair;
import com.netapp.offtap3.utils.Utils;

/**
 * Class that manages caching ZAPI information that we retrieved from a filer. Cache is persisted
 * onto disk to stay between processes (common use case is for unit tests)
 * 
 * For risk reduction - must be enabled by an external caller. Came initially from the vvol
 * prototype code.
 * 
 * This class is normally accessed as a singleton, but it capable of being replaced for unit test
 * purposes in ControllerManager.
 */
public class ZapiCache {
    private static final Logger LOGGER = Logger.getLogger(ZapiCache.class);
    public final static long NO_CACHE = 0;
    public final static long SHORT_CACHE = 600000; // 10 minutes
    public final static long LONG_CACHE = 1260000; // 2 hours

    private Map zapiMessagesCacheValueMap = Collections.synchronizedMap(new HashMap());
    private Map zapiMessagesCacheTimeMap = Collections.synchronizedMap(new HashMap());
    private static boolean enableCaching = false;
    private static boolean writeReport = false;
    private static boolean writeReportToLogger = false;
    private static ZapiCache instance = null;
    private int cacheHit = 0;
    private int cacheMiss = 0;
    private boolean dumb = false;
    private static String performanceCsvFile = null;
    private static boolean haveAddedShutdownHook = false;
    private static Set<String> pendingBlasts = new CopyOnWriteArraySet<String>(); 

    public ZapiCache() {

    }

    public static synchronized ZapiCache getInstance() {
        if (instance == null) {
            instance = new ZapiCache();
        }
        return instance;
    }
    
    public void goDumb() {
        dumb = true;
    }
    
    public void goSmart() {
        dumb = false;
    }

    public void blastCacheValues(String... keys) {
        for (String key: keys) {
            blastCacheValues(key);
        }
    }

    public void blastCacheValues(String keyString) {
    	if (dumb) {
    		return;
    	}
    	int sleepCount = 0;
    	pendingBlasts.add(keyString);
    	while (ZAPIInvoker.getInFlightCount() > 0) {
    		// spin while someone is in flight - want to avoid case where we cache a response that is just slightly out of date.  We are trying to avoid
    		// the case of:
    		// ask for lun-get-iter
    		// Create lun B
    		// blastCacheValues() since we know we just created B
    		// lun-get-iter returns, thinking it has everything and caches it response which does not have B.
    		StatsCollector.sleep(100, "zapi-cache-blast");
    		sleepCount++;
    		if (sleepCount % 100 == 99) {
    			LOGGER.debug("Waiting to clear zapi cache for " + sleepCount);
    		}
    	}
    	try {
	        synchronized (zapiMessagesCacheValueMap) {
	            for (String key : new ArrayList<String>(zapiMessagesCacheValueMap.keySet())) {
	                if (key.contains(keyString)) {
	                    zapiMessagesCacheValueMap.remove(key);
	                    zapiMessagesCacheTimeMap.remove(key);
	                }
	            }
	        }
    	} finally {
    		pendingBlasts.remove(keyString);
    	}
    }

    public String getZapiCacheKey(Object filer, Object server, NaElement zapi) {
        try {
            return filer + ":" + server + ":" + zapi;
        } catch (RuntimeException e) {
            return null;
        }
    }

    public Object getValueFromCache(Object filer, Object server, NaElement zapi,
            long howLongToCacheFor) {
        if (dumb) {
            return null;
        }
        if (pendingBlasts.size() > 0) {
        	// someone has a pending blast, lets not allow anyone to return a value here that might have been 
        	// requested from JUST before the blast was issued.  I'd rather return nothing than risk returning
        	// something that is wrong.
        	// See note in blastCacheValues()
        	return null;
        }
        Object cachedValue = null;
        Object cachedTime = null;
        boolean useZapiCache = false;
        String zapiCacheKey = getZapiCacheKey(filer, server, zapi);
        if (zapiCacheKey == null) {
            return null;
        }

        synchronized (zapiMessagesCacheValueMap) {
            cachedValue = zapiMessagesCacheValueMap.get(zapiCacheKey);
            cachedTime = zapiMessagesCacheTimeMap.get(zapiCacheKey);
        }
        
        long howFarBack = 0;

        if (cachedTime != null) {
            // ok - it was there. What to do about it
            howFarBack = System.currentTimeMillis() - ((Long) cachedTime);
            if (howFarBack < howLongToCacheFor) {
                useZapiCache = true;
            }
        }

        if (useZapiCache && enableCaching) {
            if (cachedValue instanceof String) {
                try {
                    NaElement cachedElement = NaElementFactory.fromString((String) cachedValue);
                    StatsCollector.getInstance().addCounter("zsc-cache-hit");
                    cacheHit++;
                    return cachedElement;
                } catch (IOException e) {
                    LOGGER.warn("Could not build element from cached value " + cachedValue, e);
                } catch (SAXException e) {
                    LOGGER.warn("Could not build element from cached value " + cachedValue, e);
                }
            }
            if (cachedValue instanceof Pair) {
            	StatsCollector.getInstance().addCounter("zsc-cache-hit");
                cacheHit++;
                return cachedValue;
            }
        }
        StatsCollector.getInstance().addCounter("zsc-cache-miss");
        cacheMiss++;
        return null;
    }

    public void addZapiToCache(Object filer, Object server, NaElement zapi, Object results) {
        String zapiCacheKey = getZapiCacheKey(filer, server, zapi);
        if (zapiCacheKey == null || dumb) {
            return;
        }
        synchronized (zapiMessagesCacheValueMap) {
            // make sure it doesn't get arbitrarily large and suck the memory out of the world
            if (zapiMessagesCacheValueMap.size() > 5000) {
                zapiMessagesCacheValueMap.clear();
                zapiMessagesCacheTimeMap.clear();
            }
        }
        if (results instanceof NaElement) {
            zapiMessagesCacheValueMap.put(zapiCacheKey, "" + results);
        } else {
            zapiMessagesCacheValueMap.put(zapiCacheKey, results);
        }
        zapiMessagesCacheTimeMap.put(zapiCacheKey, System.currentTimeMillis());
    }
    
    public static void enableCaching() {
    	getInstance().enableCachingInternal();
    }

    public void enableCachingInternal() {
        enableCaching = true;
        final String messageCacheFilename = "zapi_message_cache.ser";
        final String timeCacheFilename = "zapi_time_cache.ser";
        try {
            Object existingMessagesCache = Utils.readObjectFromFile(messageCacheFilename);
            Object existingTimingsCache = Utils.readObjectFromFile(timeCacheFilename);
            synchronized (getInstance().zapiMessagesCacheValueMap) {
                if (existingMessagesCache instanceof Map) {
                	boolean load = true;
                	if (getInstance().zapiMessagesCacheTimeMap != null && getInstance().zapiMessagesCacheTimeMap.size() > 0) {
                		load = false;
                	}
                	if (load) {
	                    Map existingMessageCacheMap = (Map) existingMessagesCache;
	                    for (Object message : existingMessageCacheMap.keySet()) {
	                        getInstance().zapiMessagesCacheValueMap.put(message,
	                                ((Map) existingMessagesCache).get(message));                        
	                        getInstance().zapiMessagesCacheTimeMap.put(message,
	                                ((Map) existingTimingsCache).get(message));
	                    }
                	}
                }

            }
        } catch (Exception e) {
        }

        if (!haveAddedShutdownHook) {
        	haveAddedShutdownHook = true;
	        Runtime.getRuntime().addShutdownHook(new Thread(new Runnable() {
	            public void run() {
	                try {
	                    if (enableCaching) {
	                        Utils.writeObjectToFile(messageCacheFilename,
	                                (Serializable) getInstance().zapiMessagesCacheValueMap);
	                        Utils.writeObjectToFile(timeCacheFilename,
	                                (Serializable) getInstance().zapiMessagesCacheTimeMap);
	                    }                    
	                    if (writeReport) {                        
	                        System.out.println(getInstance().cacheHit + " cache hits and "
	                                + getInstance().cacheMiss + " cache misses");                        
	                    }
	                    if (writeReportToLogger) {
	                        Logger logger = Logger.getLogger("PERFORMANCE");
	                        logger.info(getInstance().cacheHit + " cache hits and "
	                                + getInstance().cacheMiss + " cache misses");
	                    }
	                    ZAPIStatsCounter.getInstance().writeReportToFinalOutput("Aeonflux Performance information as of " + new Date() + ".\n", true);
	                } catch (Exception e) {
	                }
	            }
	        }));
        }
    }

    public static void disableCaching() {
        enableCaching = false;
    }

    public static boolean getEnableCaching() {
        return enableCaching;
    }

    public static synchronized void setWriteReport(boolean writeReport) {
        ZapiCache.writeReport = writeReport;
    }
    
    public static synchronized void setWriteReportToLogger(boolean writeReport) {
        ZapiCache.writeReportToLogger = writeReport;
    }

	public static boolean isWriteReport() {
		return writeReport;
	}

	public static boolean isWriteReportToLogger() {
		return writeReportToLogger;
	}

	public static String getPerformanceCsvFile() {
		return performanceCsvFile;
	}

	public static void setPerformanceCsvFile(String arg) {
		performanceCsvFile = arg;
	}
	
	public long getZapiCacheSize() {
		return zapiMessagesCacheValueMap.size();
	}
}
