package com.netapp.offtap3.internal.ontap.utils;

import com.netapp.offtap3.internal.ontap.support.ZAPIStatsCounter;
import com.netapp.offtap3.internal.ontap.support.ZapiCache;
import com.netapp.offtap3.utils.SafeConcurrentHashMap;
import com.netapp.offtap3.utils.StringUtils;
import com.netapp.offtap3.utils.Utils;

import org.apache.log4j.Logger;

import java.io.File;
import java.io.Serializable;
import java.util.Date;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;
import java.util.concurrent.ConcurrentHashMap;

/**
 * Helper class that manages long term trending data for StatsCollector.
 * Originally this was simply in memory, but over time it began to 
 * become a larger and larger memory footprint and impact.  So we'll 
 * rig something up to spool things out to disk so that we can in fact
 * live forever.
 */
public class LongTermTrender {
    private static final Logger LOGGER = Logger.getLogger(LongTermTrender.class);

	private Map<Date, String> longTermTrendingData = new SafeConcurrentHashMap<Date, String>();
	private Map<String, Object> valuesToIncludeInLongTermTrend = new SafeConcurrentHashMap<String, Object>();
	private static EngagementPercentageCalculator engagementCalculator = new EngagementPercentageCalculator();
	private final static String SPOOL_DIR = Utils.addFS(".") + "spool";
	private static String overrideSpoolDir = null;
	static boolean haveInitialized = false;	

	public LongTermTrender() {
	}
	
	public static String getSpoolDir() {
	    if (overrideSpoolDir != null) {
	        return overrideSpoolDir;
	    } else {
	        return SPOOL_DIR;
	    }
	}
	
	public static void setOverrideSpoolDir(String param) {
		if (!Utils.areEqual(param, overrideSpoolDir)) {
			haveInitialized = false;
		}
	    overrideSpoolDir = param;
	}

    public void captureLongTermTrendingData() {
		if (!haveInitialized) {
			haveInitialized = true;
			if (ZapiCache.getPerformanceCsvFile() != null) {
				if (new File(getSpoolDir()).exists()) {
					Utils.deleteFilesRecursive(getSpoolDir());
				}
				try {
					new File(getSpoolDir()).mkdirs();
				} catch (Exception e) {
                    LOGGER.debug("Error while creating directory for capturing long term trending data: " + e.getMessage());
					// not much to be able to do about it now.
				}
			}
		}
		Date now = new Date();
    	Map zapiInformation = ZAPIStatsCounter.getInstance().getDurationsAndAverages();
    	zapiInformation.putAll(StatsCollector.getInstance().getDurationsAndAverages());
    	System.gc();  // lets try to make this memory information as accurate as we can.
    	long freeMem = Runtime.getRuntime().freeMemory();
    	long totalMem = Runtime.getRuntime().totalMemory();
    	zapiInformation.put("memory-free", freeMem);
    	zapiInformation.put("memory-max", Runtime.getRuntime().maxMemory());
    	zapiInformation.put("memory-total", totalMem);
    	zapiInformation.put("memory-used", totalMem-freeMem);
    	zapiInformation.putAll(valuesToIncludeInLongTermTrend);
    	zapiInformation.putAll(engagementCalculator.calculateEngagementPercentages(zapiInformation));
    	try {
	    	byte[] compressedData = Utils.compressObject((Serializable) zapiInformation);
	    	String filename = Utils.addFS(getSpoolDir()) + System.currentTimeMillis() + ".ser";
	    	Utils.writeFileBinary(filename, compressedData, false);
	    	longTermTrendingData.put(now, filename);
    	} catch (Exception e) {
    		LOGGER.debug("Error occurred while writing long term trending data: " + e.getMessage());
    	}

    	writePerformanceCsvIfApplicable();
	}

	public void addValueForNextLongTermTrend(String key, Object value) {
    	valuesToIncludeInLongTermTrend.put(key, value);
    }

	public void removeValueForNextLongTermTrend(String key) {
		valuesToIncludeInLongTermTrend.remove(key);
	}

	public Set<String> getValuesForNextLongTermTrend() {
		return valuesToIncludeInLongTermTrend.keySet();
	}

    public Object getValueForNextLongTermTrend(String key) {
        return valuesToIncludeInLongTermTrend.get(key);
    }

	private Map<String, Long> getValuesForDate(Date d) throws Exception {
		String filename = longTermTrendingData.get(d);
		if (filename != null) {
		    try {
		        byte[] bytes = Utils.readFileBinary(filename);
		        return (Map<String, Long>) Utils.decompressObject(bytes);
		    } catch (Exception e) {
		        // having some trouble with these in NFS mount unit test environment where we've asked to have the file
		        // written but its not really there yet - so we'll sleep a bit and see if its there the second time.
		        Utils.sleep(2000);
		        byte[] bytes = Utils.readFileBinary(filename);
                return (Map<String, Long>) Utils.decompressObject(bytes);
		    }
		}
		return new HashMap<String, Long>();
	}

	public void writePerformanceCsvIfApplicable() {
        if (ZapiCache.getPerformanceCsvFile() != null) {
            Set<String> keys = new HashSet<String>();
            for (Date d : longTermTrendingData.keySet()) {
                try {
                    Map<String, Long> values = getValuesForDate(d);
                    keys.addAll(values.keySet());
                } catch (Exception e) {
                    LOGGER.debug("Error occurred while retrieving performance values for date " + d + ": " + e.getMessage());
                }
            }
            StringBuilder report = new StringBuilder();
            report.append("Date,");
            for (String key : Utils.getSortedList(keys)) {
                key = key.replaceAll(",", "");
                if (key.length() > 200) {
                    key = key.substring(0, 200);
                }
                report.append(key);
                report.append(",");
            }
            report.setLength(report.length()-1);
            report.append(StringUtils.getNewLine());
            for (Date d : Utils.getSortedList(longTermTrendingData.keySet())) {
                StringBuffer temp = new StringBuffer();
                temp.append(d);
                temp.append(",");
                
                try {
                    Map<String, Long> currMap = getValuesForDate(d);
                    for (String key : Utils.getSortedList(keys)) {
                        Object value = currMap.get(key);
                        if (value == null) {
                            value = "";
                        }
                        temp.append(value);
                        temp.append(",");
                    }
                    temp.setLength(temp.length()-1);
                    temp.append(StringUtils.getNewLine());
                    report.append(temp);
                } catch (Exception e) {
                    e.printStackTrace();
                    LOGGER.debug("Error occurred while retrieving performance values for date " + d + ": " + e.getMessage());
                }               
            }
            // previously we were just writing it directly, but in an OOM condition, we got to 
            // where the delete succeeded and the subsequent write failed.  Which bites, because this
            // is a file that would have been very helpful in understanding the growth to our OOM.
            // So - we'll do a song and dance using a .tmp file to reduce risk that things disappear
            String renamedFile = ZapiCache.getPerformanceCsvFile() + ".tmp";
            if (new File(ZapiCache.getPerformanceCsvFile()).exists()) {
                // only whack .tmp if the .csv file exists
                new File(renamedFile).delete();
            }
            // move .csv (if it exists) 
            try {
                new File(ZapiCache.getPerformanceCsvFile()).renameTo(new File(renamedFile));
            } catch (RuntimeException e) {
                // whatever
            }
            Utils.writeFileSafe(ZapiCache.getPerformanceCsvFile(), report.toString());
            if (new File(ZapiCache.getPerformanceCsvFile()).exists()) {
                // only whack .tmp if the .csv file exists
                new File(renamedFile).delete();
            }
        }
    }
}
