package com.netapp.offtap3.internal.ontap.utils;


import org.apache.log4j.Logger;

import java.util.Arrays;
import java.util.HashSet;
import java.util.Set;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

/**
 * Handles the parsing of the version number, and failing that a few random string matches to determine the ONTAP version
 * from an ONTAP version string.  Also, has the ability to map ONTAPI versions to ONTAP versions.
 * <p/>
 * Currently, this class is only used for 7-mode, since c-mode system-get-version returns an XML-formatted version tuple,
 * which eliminates the parsing problems.
 */
public class OntapVersionGuesser {

    // version pattern
    // not clear why we are using {0,3} and {0,7}
    // but the trailing group is to handle "x.y.z 7-Mode:" and "x.y.z Cluster-Mode:" formats, in addition to the "x.y.z:" format
    final static Pattern versionPat = Pattern.compile("Release ?R?((\\d\\.\\d\\.\\d).{0,3}|(\\d\\.\\d\\.\\d).{0,7}|(\\d\\.\\d).{0,3}|(\\d\\.\\d).{0,7}|(\\d\\.\\dRC\\d).{0,7})((:)|( [0-9A-Za-z\\-]+:))");

    private static Logger logger = Logger.getLogger(OntapVersionGuesser.class);

    private static final Set<String> SUFFIXES = new HashSet<String>(Arrays.asList("X", "RC", "P", "D", "R", "L"));

    /**
     * Take a stab at parsing the ONTAP version
     * @param xoVersion version string from ONTAP (the long form:  NetApp Release blah-blah  some-date )
     * @return gen.major or gen.major.minor format, if parse succeeded.  null if could not be determined.
     */
    public String fromVersionString(String xoVersion) {
        String version = null;
        Matcher m = versionPat.matcher(xoVersion);

        if (m.find()) {
            version = m.group(1);

            for (String suffix : SUFFIXES) {
                int index = version.indexOf(suffix);
                if (index > 0) {
                    version = version.substring(0, index);
                }
            }
        } else { // let's try some nightly build codenames (note: this is 7-mode format)
            if (xoVersion.contains("RironcityN") || xoVersion.contains("IronCity")) {
                version = "7.3.3";
            }
            else if (xoVersion.contains("RrollingrockN") || xoVersion.contains("RollingRock")) {
                version = "8.1";
            }
            else if (xoVersion.contains("RsierranevadaN") || xoVersion.contains("SierraNevada")) {
                version = "8.2";
            }
            else if (xoVersion.contains("RfullsteamN") || xoVersion.contains("Fullsteam")) {
                version = "8.3";
            }
        }

        if (version != null) {
            // before blindly returning the string, confirm it has been sanitized
            String[] digits = version.split("\\.");
            for (String d : digits) {
                try {
                    //noinspection ResultOfMethodCallIgnored
                    Integer.parseInt(d);
                }
                catch (NumberFormatException nfe) {
                    // failed to parse
                    return null;
                }
            }
        }
        return version;

    }

    /**
     * Maps an ONTAPI version to an ONTAP version
     * @param minor the ONTAPI minor version (since major is always 1)
     * @return gen.major or gen.major.minor format for version or null if unrecognized
     */
    public String fromOntapiVersion(int minor) {
        // see BURT 498586 and 589473
        // for a particular DOT release, see ng_shared/src/util/zapi_version.h

        switch (minor) {

            case 21 :
                return "8.2.1";
            case 20 :
                return "8.2";

            case 19 :
                return "8.1.2";
            case 17 :
                return "8.1.1";
            case 15 :
                return "8.1";

            case 18 :
                return "8.0.4";
            case 13 : // 8.0.3, 8.0.2, 8.0.1
            case 11 : // 8.0
                return "8.0";

            case 16 : // 7.3.6, 7.3.7
            case 14 : // 7.3.5
                return "7.3.5"; // treat all like 7.3.5, since that's closer to the mark than 7.3
            case 12 : // 7.3.3, 7.3.4
                return "7.3";

            default: // 10 is not used; 9 and earlier are not supported
                logger.warn("Unrecognized ONTAPI version: 1." + minor);
                return null;
        }

    }
}

