package com.netapp.offtap3.internal.ontap.utils;

import com.netapp.offtap3.internal.ontap.support.ZAPIStatsCounter;
import com.netapp.offtap3.internal.ontap.support.ZapiCache;
import com.netapp.offtap3.utils.SafeConcurrentHashMap;
import com.netapp.offtap3.utils.Utils;

import org.apache.log4j.Logger;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.Date;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.CopyOnWriteArrayList;
import java.util.concurrent.atomic.AtomicLong;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

public class StatsCollector {

    protected final static Map<String, List<Long>> stats = new SafeConcurrentHashMap<String, List<Long>>();
    protected final static Map<String, List<Date>> statsDates = new SafeConcurrentHashMap<String, List<Date>>();
    private final static int MAX_SIZE_UNTIL_ROLLOVER = 50000;
    private static Pattern timeoutMessage = Pattern.compile("running (\\d+) milliseconds");
    private static StatsCollector instance;
    protected final Map<String, AtomicLong> statsToDurationSum = new SafeConcurrentHashMap<String, AtomicLong>();
    protected final Map<String, AtomicLong> statsToDurationCount = new SafeConcurrentHashMap<String, AtomicLong>();
    private LongTermTrender trender = new LongTermTrender();
    // flag to say its only being accessed by a single thread - otherwise the CopyOnWriteArrayList performance
    // hit causes the OutputCollatorTest to start taking minutes instead of seconds.
    private boolean goSingleThreaded = false;
    // lets keep track of how many things come into here so that we don't grow absolutely forever.
    private int rolloverCounter = 0;
    private final Object rolloverCounterLock = new Object();


    // Note - I am allowing construction here for unit test purposes - even though normally
    // things will be Spring singleton
    public StatsCollector() {
        instance = this;
    }

    public static synchronized StatsCollector getInstance() {
        if (instance == null) {
            instance = new StatsCollector();
        }
        return instance;
    }
    
    public void reset() {
    	stats.clear();
    	statsDates.clear();
    	statsToDurationSum.clear();
    	statsToDurationCount.clear();
    	rolloverCounter = 0;
    }

    public void addCounter(String counter) {
        addCounter(counter, 1);
    }

    public void addValueForNextLongTermTrend(String key, Object value) {
        trender.addValueForNextLongTermTrend(key, value);
    }

    public void removeValueForNextLongTermTrend(String key) {
        trender.removeValueForNextLongTermTrend(key);
    }

    public Set<String> getValuesForNextLongTermTrend() {
        return trender.getValuesForNextLongTermTrend();
    }

    public Object getValueForNextLongTermTrend(String key) {
        return trender.getValueForNextLongTermTrend(key);
    }

    public void addCounter(String counter, long amount) {
        if (counter == null) {
            return;
        }
        if (counter.length() > 300) {
            counter = counter.substring(0, 270) + "..." + counter.substring(counter.length() - 30, counter.length());
        }
        if (counter.contains("Operation timed out")) {
            Matcher makeMatch = timeoutMessage.matcher(counter);
            if (makeMatch.find()) {
                String beenGoingDuration = makeMatch.group(1);

                int numberOfZeroes = beenGoingDuration.length() - 1;
                long numberToRoundOff = Long.parseLong(beenGoingDuration);
                numberToRoundOff = (long) (numberToRoundOff - numberToRoundOff % (Math.pow(10, numberOfZeroes)));
                counter = counter.replaceFirst(beenGoingDuration, Long.toString(numberToRoundOff));
            }
        }
        if (counter.contains("Task of ")) {
            counter = counter.replaceAll("Task of \\d+", "Task");
        }
        if (counter.contains("Vvol uuid of ")) {
            counter = counter.replaceAll("Vvol uuid of \\S+ ", "Vvol uuid ");
        }
        if (counter.contains("rfc4122")) {
        	counter = counter.replaceAll("rfc4122.{37}", "uuid");
        }
        if (counter.contains("naa.")) {
        	counter = counter.replaceAll("naa.{33}", "uuid");
        }
        if (!counter.contains("depth") && (counter.startsWith("persistence") && !counter.equals(
                "persistence-manager-any-operation")) || counter.startsWith("lock") || counter.startsWith("vsc-")) {
            ZAPIStatsCounter.getInstance().addZAPICall(counter, amount);
        }
        synchronized (rolloverCounterLock) {
            rolloverCounter++;
            if (rolloverCounter > MAX_SIZE_UNTIL_ROLLOVER) {
                rolloverCounter = 0;
                addCounter("rollover-stats");
                // lets get all of the rollover-* counters.  Those are the only history we have
                // of how many rollovers we've done, and we want to keep those.
                Map<String, List<Long>> statsToKeep = new SafeConcurrentHashMap<String, List<Long>>();
                Map<String, List<Date>> statsDatesToKeep = new SafeConcurrentHashMap<String, List<Date>>();
                for (String key : stats.keySet()) {
                    if (key.startsWith("rollover")) {
                        statsToKeep.put(key, stats.get(key));
                        statsDatesToKeep.put(key, statsDates.get(key));
                    }
                }

                writeReportToFinalOutput(
                        "Intermediate StatsCollector report before rolling over to save memory as of " + new Date());
                stats.clear();
                statsDates.clear();
                stats.putAll(statsToKeep);
                statsDates.putAll(statsDatesToKeep);
            }
        }
        List<Long> values = stats.get(counter);
        if (values == null) {
            values = makeList();
            stats.put(counter, values);
        }
        List<Date> dates = statsDates.get(counter);
        if (dates == null) {
            dates = makeList();
            statsDates.put(counter, dates);
        }
        if (values.size() < 300 && dates.size() < 300) {
            // we get into degenerate cases where these lists are very very large, and since they have to be
            // synchronized, we have to copy them before the update.  Lets only keep the specific value if there's
            // a small number of values, since the AtomicLongs below is what we use for most of the processing            
            values.add(amount);
            dates.add(new Date());
        }

        AtomicLong currentSum = statsToDurationSum.get(counter);
        if (currentSum == null) {
            currentSum = new AtomicLong();
            statsToDurationSum.put(counter, currentSum);
        }
        currentSum.getAndAdd(amount);
        AtomicLong currentCount = statsToDurationCount.get(counter);
        if (currentCount == null) {
            currentCount = new AtomicLong();
            statsToDurationCount.put(counter, currentCount);
        }
        currentCount.incrementAndGet();
    }

    public Map<String, Long> getDurationsAndAverages() {
    	return getDurationsAndAverages("");
    }
    
    public Map<String, Long> getDurationsAndAverages(String keyBase) {
    	Map<String, Long> retVal = new HashMap();
    	for (String key : statsToDurationSum.keySet()) {
    		if (key.contains(keyBase)) {
	    		Long duration = statsToDurationSum.get(key).longValue();
	    		Long count = statsToDurationCount.get(key).longValue();
	    		retVal.put(key+"-total", duration);
	    		retVal.put(key+"-average", new Long (duration / count));
    		}
    	}
    	return retVal;
    }

    public long getTotalValueForCounterSubstring(String counter) {
        long accumulator = 0;
        for (String key : this.getKeys()) {
            if (key.contains(counter)) {
                accumulator += getTotalValueForCounter(key);
            }
        }
        return accumulator;
    }

    public long getTotalValueForCounter(String counter) {
        AtomicLong val = statsToDurationSum.get(counter);
        if (val != null) {
            return val.longValue();
        }
        return 0;
    }

    public long getAverageValueForCounter(String counter) {
        return Utils.average(getValuesForCounter(counter));
    }

    public List<Long> getValuesForCounter(String counter) {
        List<Long> retVal = stats.get(counter);
        if (retVal == null) {
            retVal = makeList();
        }
        return retVal;
    }

    public List<Date> getDatesForCounter(String counter) {
        List<Date> retVal = statsDates.get(counter);
        if (retVal == null) {
            retVal = makeList();
        }
        return retVal;
    }

    private <T> List<T> makeList() {
        if (goSingleThreaded) {
            return Collections.synchronizedList(new ArrayList<T>());
        } else {
            return new CopyOnWriteArrayList<T>();
        }
    }

    public Collection<String> getKeys() {
        return Utils.getSortedList(statsToDurationSum.keySet());
    }

    public void captureLongTermTrendingData() {
        trender.captureLongTermTrendingData();
    }

    public void writePerformanceCsvIfApplicable() {
        trender.writePerformanceCsvIfApplicable();
    }

    public void writeReportToFinalOutput(String header) {
        if (ZapiCache.isWriteReport()) {
            System.out.println(header);
            System.out.println(writeReport());
        }
        if (ZapiCache.isWriteReportToLogger()) {
            Logger logger = Logger.getLogger("PERFORMANCE");
            logger.info(header);
            logger.info(writeReport());
        }
    }

    public String writeReport() {
        StringBuilder retVal = new StringBuilder();
        retVal.append("Aeonflux statistics:\n");
        Map<String, List<Long>> statsCopy;
        synchronized (stats) {
            statsCopy = new HashMap<String, List<Long>>(stats);
        }
        for (String counter : Utils.getSortedList(statsCopy.keySet())) {
            long total = 0;
            for (Long value : statsCopy.get(counter)) {
                total += value;
            }
            retVal.append(counter).append("=").append(total).append("(").append(statsCopy.get(counter).size()).append(")\n");
        }

        return retVal.toString();
    }
    
    public static void sleep(long amount, String lockName) {
    	Utils.sleep(amount);
    	StatsCollector.getInstance().addCounter("lock-" + lockName, amount);
    }
    
    /**
     * Should be used only in cases where we are sure that someone is going to put in a stat saying why we were sleeping
     */
    public static void sleepSilently(long amount) {
    	Utils.sleep(amount);
    }

    public boolean isGoSingleThreaded() {
        return goSingleThreaded;
    }

    public void setGoSingleThreaded(boolean goSingleThreaded) {
        this.goSingleThreaded = goSingleThreaded;
    }
}
