package com.netapp.offtap3.internal.ontap.volume;

import com.netapp.offtap3.ControllerConnection;
import com.netapp.offtap3.ControllerResource;
import com.netapp.offtap3.StorageSystem;
import com.netapp.offtap3.VolumeOperationException;
import com.netapp.offtap3.internal.CachingOfftapProxy;
import com.netapp.offtap3.internal.DataCache;
import com.netapp.offtap3.internal.UnixPermissionImpl;
import com.netapp.offtap3.internal.ontap.InternalZapiHandler;
import com.netapp.offtap3.internal.ontap.MCCManager;
import com.netapp.offtap3.internal.ontap.controller.AdminSvmImpl;
import com.netapp.offtap3.internal.ontap.controller.DataSvmImpl;
import com.netapp.offtap3.internal.ontap.controller.OntapConnection;
import com.netapp.offtap3.internal.ontap.file.CModeFileImpl;
import com.netapp.offtap3.internal.ontap.proxy.CModeVolumeProxy;
import com.netapp.offtap3.internal.ontap.snap.VolumeSnapMirrorStatusImpl;
import com.netapp.offtap3.internal.ontap.support.ZapiCache;
import com.netapp.offtap3.ontap.Aggregate;
import com.netapp.offtap3.ontap.ApiFailedException;
import com.netapp.offtap3.ontap.CModeAggregate;
import com.netapp.offtap3.ontap.CModeDedupe;
import com.netapp.offtap3.ontap.CModeFile;
import com.netapp.offtap3.ontap.CModeLun;
import com.netapp.offtap3.ontap.CModeSnapMirrorRelation;
import com.netapp.offtap3.ontap.CModeSnapshot;
import com.netapp.offtap3.ontap.CModeVolume;
import com.netapp.offtap3.ontap.DataSvm;
import com.netapp.offtap3.ontap.ExportPolicy;
import com.netapp.offtap3.ontap.NaElement;
import com.netapp.offtap3.ontap.QosPolicyGroup;
import com.netapp.offtap3.ontap.SnapMirrorOperationException;
import com.netapp.offtap3.ontap.SnapMirrorRelation;
import com.netapp.offtap3.utils.StringUtils;
import com.netapp.offtap3.utils.Utils;

import org.apache.log4j.Logger;

import java.util.EnumSet;
import java.util.List;
import java.util.Set;

public class CModeVolumeImpl extends OntapVolumeImpl<CModeVolume, DataSvm, DataSvmImpl> implements CModeVolume {

    private static final Logger LOGGER = Logger.getLogger(CModeVolumeImpl.class);
    private static final Set<StorageSystem> STORAGE_SYSTEMS = EnumSet.of(StorageSystem.datasvm);
    private final DataCache<NaElement, InternalZapiHandler> dataCache;
    private CModeVolume proxy;
    //Inode info
    private long filesTotal;
    private long filesUsed;

    private int stripeWidth;
    private String flexvoltype;

    private boolean isNodeRoot;
    private boolean inNvfailedState;
    private boolean isClusterVolume;
    private boolean vserverRoot;
    private String aggregateUUID;

    private String exportPolicyName;
    private boolean isSnapDirAccessEnabled;

    // snapshot auto delete
    private boolean isAutoDeleteEnabled;
    private String autoDeleteCommitment;

    // Unix permissions
    private UnixPermissionImpl unixPermission;

    //qos related
    private String qosPolicyGroupName;

    public CModeVolumeImpl(DataSvmImpl controller, DataCache<NaElement, InternalZapiHandler> dataCache) {
        super(controller);
        this.dataCache = dataCache;
        setDedupe(new CModeDedupeImpl(this, controller.getOfftapPreferences()));
    }

    @Override
    public UnixPermissionImpl getUnixPermission() {
        return unixPermission;
    }

    public void setUnixPermissions(UnixPermissionImpl permission) {
        unixPermission = permission;
    }

    /**
     * Returns the total inodes
     *
     * @return the total inodes
     */
    @Override
    public long getFilesTotal() {
        return filesTotal;
    }

    public void setFilesTotal(long filesTotal) {
        this.filesTotal = filesTotal;
    }

    /**
     * Returns the number of inodes used
     *
     * @return the number of inodes used
     */
    @Override
    public long getFilesUsed() {
        return filesUsed;
    }

    public void setFilesUsed(long filesUsed) {
        this.filesUsed = filesUsed;
    }

    /**
     * Get RAID stripe width (0 if not applicable)
     *
     * @return RAID stripe width.
     */
    @Override
    public int getStripeWidth() {
        return stripeWidth;
    }

    /**
     * Set RAID stripe width.  0 if not applicable.
     *
     * @param stripeWidth RAID stripe width.
     */
    public void setStripeWidth(int stripeWidth) {
        this.stripeWidth = stripeWidth;
    }

    /**
     * Root node flag. (in ZAPI XML as volume-state-attributes -> is-node-root)
     *
     * @return Root node flag.
     */
    @Override
    public boolean isNodeRoot() {
        return isNodeRoot;
    }

    /**
     * Set root node flag.
     *
     * @param nodeRoot Root node flag value.
     */
    public void setNodeRoot(boolean nodeRoot) {
        isNodeRoot = nodeRoot;
    }

    /**
     * Return cluster volume flag.
     *
     * @return Cluster volume flag.
     */
    @Override
    public boolean isClusterVolume() {
        return isClusterVolume;
    }

    /**
     * Set cluster volume flag.
     *
     * @param clusterVolume Cluster volume flag.
     */
    public void setClusterVolume(boolean clusterVolume) {
        isClusterVolume = clusterVolume;
    }

    @Override
    public boolean isVserverRoot() {
        return vserverRoot;
    }

    public void setVserverRoot(boolean vserverRoot) {
        this.vserverRoot = vserverRoot;
    }

    @Override
    public boolean isSnapDirAccessEnabled() {
        return isSnapDirAccessEnabled;
    }

    public void setSnapDirAccessEnabled(boolean isSnapDirAccessEnabled) {
        this.isSnapDirAccessEnabled = isSnapDirAccessEnabled;
    }

    @Override
    public String getExportPolicyName() {
        return exportPolicyName;
    }

    public void setExportPolicyName(String exportPolicyName) {
        this.exportPolicyName = exportPolicyName;
    }

    @Override
    public void setExportPolicy(ExportPolicy exportPolicy) {
        NaElement modify = new NaElement("volume-modify-iter");

        // Add volume name for query purposes.
        NaElement query = new NaElement("query");
        NaElement queryAttributes = new NaElement("volume-attributes");
        NaElement queryVolumeId = new NaElement("volume-id-attributes");
        queryVolumeId.addNewChild("name", getName());
        queryAttributes.addChildElem(queryVolumeId);
        query.addChildElem(queryAttributes);
        modify.addChildElem(query);

        // Set up the unix security attributes
        NaElement attributes = new NaElement("attributes");
        NaElement volumeAttributes = new NaElement("volume-attributes");
        NaElement exportAttributes = new NaElement("volume-export-attributes");
        exportAttributes.addNewChild("policy", exportPolicy.getName());

        volumeAttributes.addChildElem(exportAttributes);
        attributes.addChildElem(volumeAttributes);
        modify.addChildElem(attributes);

        try {
            getControllerImpl().invokeZapi(modify, ZapiCache.NO_CACHE);
        } catch (Exception e) {
            LOGGER.error("Failed to modify unix security.", e);
            throw new VolumeOperationException("Failed to modify export policy. error: " + e.getMessage(), e);
        }
        getControllerImpl().getZapiCache().blastCacheValues("volume-");
    }

    /**
     * @return the volume qos policy group name
     */
    @Override
    public QosPolicyGroup getQosPolicyGroup() {
        QosPolicyGroup group = getController().getQosPolicyGroup(qosPolicyGroupName);
        if (group == null && StringUtils.isReal(qosPolicyGroupName)) {
            // Maybe the policy groups on the controller were out of date. Let's refresh just in case.
            getController().refresh(ControllerResource.qos);
            group = getController().getQosPolicyGroup(qosPolicyGroupName);
        }
        return group;
    }

    public void setQosPolicyGroup(String qosPolicyGroupName) {
        this.qosPolicyGroupName = qosPolicyGroupName;
    }

    @Override
    public String getAggregateUUID() {
        return aggregateUUID;
    }

    public void setAggregateUUID(String aggregateUUID) {
        this.aggregateUUID = aggregateUUID;
    }

    @Override
    public CModeAggregate getContainingAggregate() {
        return (CModeAggregate) super.getContainingAggregate();
    }

    /**
     * @return the type of this flex volume.  I.e 'rw' , 'ls' , 'dp' , 'dc' , 'tmp'
     */
    @Override
    public String getFlexVolType() {
        return flexvoltype;
    }

    public void setFlexVolType(String flexvoltype) {
        this.flexvoltype = flexvoltype;
    }

    /**
     * @return <code>true</code> if this volume has auto delete enabled;
     * <code>false</code> otherwise
     */
    @Override
    public boolean isAutoDeleteEnabled() {
        return isAutoDeleteEnabled;
    }

    public void setAutoDeleteEnabled(boolean isSnapshotAutoDeleteEnabled) {
        this.isAutoDeleteEnabled = isSnapshotAutoDeleteEnabled;
    }

    /**
     * Returns the commitment as to how snapshot auto delete kicks in
     */
    @Override
    public String getAutoDeleteCommitment() {
        return autoDeleteCommitment;
    }

    public void setAutoDeleteCommitment(String autoDeleteCommitment) {
        this.autoDeleteCommitment = autoDeleteCommitment;
    }

    public CModeDedupe getDedupe() {
        return (CModeDedupe) super.getDedupe();
    }

    @Override
    public String getSnapMirrorFormat() {
        //  [cluster:]//vserver:volume
        StringBuilder format = new StringBuilder();
        DataSvm vserver = getController();
        // TODO aeonflux-next - think about when we want the cluster specified.
        //        if (!vserver.isDirectSvm()) {
        //            format.append(vserver.getParentController().getName()).append(":");
        //            format.append("//")
        //        }
        format.append(MCCManager.getInstance().getVServerNameByUuid(vserver));
        format.append(":").append(getName());
        return format.toString();
    }

    @Override
    public CModeSnapshot getSnapshotByName(String snapName) {
        return (CModeSnapshot) super.getSnapshotByName(snapName);
    }
    
    @Override
    public void createAndInitializeSnapMirror(String sourceVserver, String sourceVolume, CModeSnapMirrorRelation.Type type) {
        // If this is not a standard mirror, use snapmirror-create first
        if (type != null) {
            String destinationLocation = getSnapMirrorFormat();
            LOGGER.debug("Enter createSnapMirror: " + destinationLocation + ", source: " + sourceVserver + ":" + sourceVolume);
            NaElement xi;

            xi = new NaElement("snapmirror-create");
            xi.addNewChild("destination-location", destinationLocation);

            if (StringUtils.isReal(sourceVolume) && StringUtils.isReal(sourceVserver)) {
                //TODO: may want to consider situations where we would also need to specify the cluster name as well 
                //      (8.1 or 8.2 with relationship control plane = v1) See zapi docs.
                xi.addNewChild("source-vserver", sourceVserver);
                xi.addNewChild("source-volume", sourceVolume);
            }

            xi.addNewChild("relationship-type", type.name());

            try {
                getControllerImpl().invokeZapi(xi, ZapiCache.NO_CACHE);
                getControllerImpl().getZapiCache().blastCacheValues("snap");
            } catch (ApiFailedException e) {
                LOGGER.error("Unable to initialize SnapMirror.", e);
                throw new SnapMirrorOperationException("Unable to initialize SnapMirror.", e);
            }
        }
        super.inititalizeSnapMirror(sourceVserver, sourceVolume);
    }

    @Override
    public void initializeSnapMirror(CModeVolume sourceVol, CModeSnapMirrorRelation.Type type) {

        // If this is not a standard mirror, use snapmirror-create first
        if (type != null && type != CModeSnapMirrorRelation.Type.data_protection) {
            String destinationLocation = getSnapMirrorFormat();
            String sourceLocation = sourceVol.getSnapMirrorFormat();
            LOGGER.debug("Enter createSnapMirror: " + destinationLocation + ", source: " + sourceLocation);
            NaElement xi;
            
            xi = new NaElement("snapmirror-create");
            xi.addNewChild("destination-location", destinationLocation);

            if (sourceLocation != null) {
                xi.addNewChild("source-location", sourceLocation);
            }

            xi.addNewChild("relationship-type", type.name());

            try {
                getControllerImpl().invokeZapi(xi, ZapiCache.NO_CACHE);
                getControllerImpl().getZapiCache().blastCacheValues("snap");
            } catch (ApiFailedException e) {
                LOGGER.error("Unable to initialize SnapMirror.", e);
                throw new SnapMirrorOperationException("Unable to initialize SnapMirror.", e);
            }
        }

        // now, either way, we are ready to initialize the mirror
        super.initializeSnapMirror(sourceVol);
    }


    @Override
    public void updateLSMirror(int timeout) throws SnapMirrorOperationException {
        DataSvmImpl vserver = getControllerImpl();
        LOGGER.debug(vserver + " updating LS mirror set for volume " + getName());
        if (!vserver.supportsDirectLsMirrorUpdate()) {
            throw new IllegalArgumentException("SnapMirror operation requires admin vserver access and " +
                                                       vserver.getName() + " is a direct-connect vserver.");
        }

        NaElement xi = new NaElement("snapmirror-update-ls-set");
        OntapConnection conn = vserver.getConnection();
        if (!vserver.isDirectSvm()) {
            conn = vserver.getParentControllerImpl().getConnection();
            xi.addNewChild("source-cluster", vserver.getParentControllerImpl().getName());
        }

        try {
            xi.addNewChild("source-vserver",  MCCManager.getInstance().getVServerNameByUuid(vserver));
            xi.addNewChild("source-volume", getName());

            NaElement xo = conn.invokeZapi(xi, ZapiCache.NO_CACHE);
            String status = xo.getChildContent("result-status");
            String jobIdString = xo.getChildContent("result-jobid");
            LOGGER.debug("LS mirror update requested. Result status: " + status + " and job ID: " + jobIdString);
            if (jobIdString != null) {
                long jobId = Long.parseLong(jobIdString);
                // wait 30 seconds, polling every 2 -- configurable?
                String completionState = vserver.blockUntilJobCompletion(jobId, timeout, 2);
                if (completionState == null) {
                    LOGGER.error("LS mirror update job (" + jobId + ") did not complete!");
                } else if (!completionState.equals("success")) {
                    LOGGER.warn("LS mirror update job (" + jobId + ") did not complete successfully!  The volume " +
                                        getName() + " may fail to mount.");
                }
            }
            LOGGER.debug("Done updating LS mirror.");
        } catch (ApiFailedException e) {
            LOGGER.error(vserver + " exception while updating LS mirror for volume " + getName(), e);
            throw new SnapMirrorOperationException(
                    vserver + " exception while updating LS mirror for volume " + getName(), e);
        }
    }

    public boolean matchesDestination(SnapMirrorRelation rel) {
        CModeSnapMirrorRelation relation = (CModeSnapMirrorRelation) rel;
        if (rel == null) {
            LOGGER.error("Why were we passed a null relation for vol + " + this);
            return false;
        }
        return Utils.areEqual(relation.getDestinationVolume(), this) &&
                Utils.areEqual(relation.getDestinationSvm(), getController());
    }

    public boolean matchesSource(SnapMirrorRelation rel) {
        CModeSnapMirrorRelation relation = (CModeSnapMirrorRelation) rel;
        if (rel == null) {
            return false;
        }
        return Utils.areEqual(relation.getSourceVolume(), getName()) &&
                Utils.areEqual(relation.getSourceSvm(),  MCCManager.getInstance().getVServerNameByUuid(getController()));
    }

    @Override
    public void offline() {
        try {
            unmount();
        } catch (VolumeOperationException e) {
            // expected and was ignored before
            LOGGER.info(this + " unmount prior to offlining failed", e);
        }

        // The unmount operation takes some time to complete.  If we don't wait here,
        // the subsequent offline operation will fail.
        try {
            Thread.sleep(1000);
        } catch (InterruptedException ex) {
            LOGGER.warn("Interrupted exception while sleeping after unmount");
        }
        super.offline();
    }

    @Override
    public void mount(String exportPath) throws VolumeOperationException {
        NaElement xi;
        xi = new NaElement("volume-mount");
        xi.addNewChild("volume-name", getName());
        xi.addNewChild("junction-path", exportPath);
        String message = "Mounted " + getName() + " to " + exportPath;
        try {
            getControllerImpl().invokeZapi(xi, ZapiCache.NO_CACHE);
            getControllerImpl().getZapiCache().blastCacheValues("volume-");
            LOGGER.info(message);
        } catch (Exception ex) {
            LOGGER.error("Failed to " + message, ex);
            throw new VolumeOperationException(message + " Error: " + ex.getMessage(), ex);
        }
    }

    @Override
    public void unmount() throws VolumeOperationException {
        try {
            LOGGER.info("Unmount " + getName() + ".");
            NaElement xi = new NaElement("volume-unmount");
            xi.addNewChild("volume-name", getName());
            xi.addNewChild("force", "true");

            getControllerImpl().invokeZapi(xi, ZapiCache.NO_CACHE);
            LOGGER.debug(getName() + " is now unmounted.");
        } catch (ApiFailedException ex) {
            if (ex.getErrno() == 14716) { //TODO: replace with NaErrno.EVOL_NOT_MOUNTED after updating manage ontap SDK
                //the volume is not mounted
            } else {
                LOGGER.error("Failed to unmount " + getName() + ".", ex);
                throw new VolumeOperationException("Failed to unmount " + getName() + ".  Error: " + ex.getMessage(),
                                                   ex);
            }
        }
        setJunctionPath(null);
    }

    @Override
    public void setUnixPermission(Integer userId, Integer groupId, String permissions) {
        NaElement modify = new NaElement("volume-modify-iter");

        // Add volume name for query purposes.
        NaElement query = new NaElement("query");
        NaElement queryAttributes = new NaElement("volume-attributes");
        NaElement queryVolumeId = new NaElement("volume-id-attributes");
        queryVolumeId.addNewChild("name", getName());
        queryAttributes.addChildElem(queryVolumeId);
        query.addChildElem(queryAttributes);
        modify.addChildElem(query);

        // Set up the unix security attributes
        NaElement attributes = new NaElement("attributes");
        NaElement volumeAttributes = new NaElement("volume-attributes");
        NaElement secAttributes = new NaElement("volume-security-attributes");
        NaElement unixSecAttributes = new NaElement("volume-security-unix-attributes");
        if (permissions != null) {
            unixSecAttributes.addNewChild("permissions", permissions);
        }
        if (userId != null) {
            unixSecAttributes.addNewChild("user-id", String.valueOf(userId));
        }
        if (groupId != null) {
            unixSecAttributes.addNewChild("group-id", String.valueOf(groupId));
        }
        secAttributes.addChildElem(unixSecAttributes);
        volumeAttributes.addChildElem(secAttributes);
        attributes.addChildElem(volumeAttributes);
        modify.addChildElem(attributes);

        try {
            getControllerImpl().invokeZapi(modify, ZapiCache.NO_CACHE);
            getControllerImpl().getZapiCache().blastCacheValues("volume-");
        } catch (ApiFailedException e) {
            LOGGER.error(this + " failed to modify unix security.", e);
            throw new VolumeOperationException(this + " failed to modify unix security. error: " + e.getMessage(), e);
        }

    }

    @Override
    public void move(Aggregate newAggr, boolean sync) {
        if (getController().isDirectSvm()) {
            throw new UnsupportedOperationException(
                    "volume move is not supported by direct-connect SVM - " + getController());
        }

        LOGGER.debug(this + " is being moved to " + newAggr);

        NaElement xi = new NaElement("volume-move-start");
        xi.addNewChild("source-volume", getName());
        xi.addNewChild("dest-aggr", newAggr.getName());
        xi.addNewChild("vserver",  MCCManager.getInstance().getVServerNameByUuid(getController()));
        try {
            AdminSvmImpl admin = getControllerImpl().getParentControllerImpl();
            admin.invokeZapi(xi, ZapiCache.NO_CACHE);
            admin.getZapiCache().blastCacheValues("volume");
        } catch (ApiFailedException e) {
            LOGGER.error(this + " failed to move to " + newAggr, e);
            throw new VolumeOperationException(this + " failed to move to " + newAggr + " error: " + e.getMessage(), e);
        }

        if (sync) {
            LOGGER.debug("move called synchronously.  waiting to complete.");
            boolean stillRunning = true;
            int count = 0;
            while (stillRunning) {
                Utils.sleep(1000);
                stillRunning = isMoving();
                if (count > 600) {
                    throw new IllegalArgumentException("vol move not completed in 10 minutes, breaking");
                }
            }
        }
        LOGGER.debug(this + " has been moved to " + newAggr.getName());
    }

    @Override
    public VolumeSnapMirrorStatusImpl getSnapMirrorStatus() {
        LOGGER.debug("begin getSnapMirrorStatus, volume = " + getName() + ", svm = " + getControllerImpl());
        VolumeSnapMirrorStatusImpl status = new VolumeSnapMirrorStatusImpl();
        for (CModeSnapMirrorRelation relation : getController().getSnapMirrorRelations()) {
            if (matchesSource(relation)) {
                status.setSource(true);
            }
            if (matchesDestination(relation)) {
                status.setDest(true);
                if ("broken-off".equalsIgnoreCase(relation.getMirrorState())) {
                    status.setBroken(true);
                }
                if ("transferring".equalsIgnoreCase(relation.getStatus()) ||
                        "quiescing".equalsIgnoreCase(relation.getStatus())) {
                    status.setInProgress(true);
                }
            }
        }
        LOGGER.debug("end getsnapMirrorStatus");
        return status;
    }

    @Override
    public CModeLun getLun(String path) {
        return (CModeLun) super.getLun(path);
    }

    @Override
    public boolean isMoving() {
        NaElement xi;
        boolean stillRunning = false;
        xi = new NaElement("volume-move-get-iter");
        NaElement child = new NaElement("query");
        child.addNewChild("volume", getName());
        xi.addChildElem(child);
        try {
            NaElement xo = getControllerImpl().getParentControllerImpl().invokeZapi(xi, ZapiCache.NO_CACHE);
            List<NaElement> theVals = xo.getChildByName("attributes-list").getChildren();
            if (theVals != null) {
                for (NaElement currChild : theVals) {
                    String state = currChild.getChildContent("state");
                    if (!"done".equals(state)) {
                        stillRunning = true;
                    }
                }
            }
        } catch (ApiFailedException e) {
            throw new VolumeOperationException(getController() + " could not determine if " + this + " was moving", e);
        }

        return stillRunning;
    }

    @Override
    public CModeFile getFile(String path) {
        return CModeFileImpl.create(getControllerImpl(), this, path);
    }


    @Override
    public void export(String exportPath, boolean writeEnable, boolean readWriteAccess, boolean rootAccess, boolean exportPathIsActual, Set<String> hosts) {
        throw new UnsupportedOperationException(
                "not implemented - and don't use append-rules-2... need to use Kamino logic to find and update rules");
    }

    @Override
    public void export(Set<String> hosts) {
        export("/" + getName(), true, hosts);
    }

    @Override
    public CModeVolume getProxy() {
        if (proxy == null) {
            proxy = CachingOfftapProxy.create(CModeVolume.class, CModeVolumeProxy.class, this, dataCache);
        }
        return proxy;
    }

    @Override
    public void setSnapDirAccess(boolean enabled) {
        LOGGER.debug("Enter changeVolumeSnapDirAccess for controller: " + getController() + " and volume: " + this);
        NaElement modify = new NaElement("volume-modify-iter");

        // Add volume name for query purposes.
        NaElement query = new NaElement("query");
        NaElement queryAttributes = new NaElement("volume-attributes");
        NaElement queryVolumeId = new NaElement("volume-id-attributes");
        queryVolumeId.addNewChild("name", getName());
        queryAttributes.addChildElem(queryVolumeId);
        query.addChildElem(queryAttributes);
        modify.addChildElem(query);

        // change the snapdir-access option
        NaElement attributes = new NaElement("attributes");
        NaElement volumeAttributes = new NaElement("volume-attributes");
        NaElement snapshotAttributes = new NaElement("volume-snapshot-attributes");
        snapshotAttributes.addNewChild("snapdir-access-enabled", Boolean.toString(enabled));

        volumeAttributes.addChildElem(snapshotAttributes);
        attributes.addChildElem(volumeAttributes);
        modify.addChildElem(attributes);

        try {
            getControllerImpl().invokeZapi(modify, ZapiCache.NO_CACHE);
            getControllerImpl().getZapiCache().blastCacheValues("volume-");
        } catch (Exception e) {
            LOGGER.error("Failed to modify snapdir-access.", e);
            throw new VolumeOperationException("Failed to modify snapdir-access. error: " + e.getMessage(), e);
        }
    }

    @Override
    public Set<StorageSystem> getStorageSystems() {
        return STORAGE_SYSTEMS;
    }

    @Override
    public ControllerConnection getConnection() {
        return getControllerImpl().getConnection();
    }

    @Override
    public CModeVolume refresh() {
        return (CModeVolume) super.refresh();
    }

    @Override
	public boolean isInNvfailedState() {
		return inNvfailedState;
	}

	public void setInNvfailedState(boolean inNvfailedState) {
		this.inNvfailedState = inNvfailedState;
	}
}
