package com.netapp.offtap3.internal.ontap.volume;

import com.netapp.offtap3.ControllerConnection;
import com.netapp.offtap3.ExportOperationException;
import com.netapp.offtap3.StorageSystem;
import com.netapp.offtap3.VolumeOperationException;
import com.netapp.offtap3.internal.CachingOfftapProxy;
import com.netapp.offtap3.internal.ontap.controller.FilerImpl;
import com.netapp.offtap3.internal.ontap.file.SevenModeFileImpl;
import com.netapp.offtap3.internal.ontap.proxy.FilerVolumeProxy;
import com.netapp.offtap3.internal.ontap.snap.SnapVaultScheduleImpl;
import com.netapp.offtap3.internal.ontap.support.ZapiCache;
import com.netapp.offtap3.ontap.Aggregate;
import com.netapp.offtap3.ontap.Filer;
import com.netapp.offtap3.ontap.FilerVolume;
import com.netapp.offtap3.ontap.OntapFile;
import com.netapp.offtap3.ontap.SecurityStyle;
import com.netapp.offtap3.ontap.SnapMirrorRelation;
import com.netapp.offtap3.ontap.SnapVaultSchedule;
import com.netapp.offtap3.ontap.VFiler;

import com.netapp.offtap3.ontap.NaElement;

import org.apache.log4j.Logger;

import java.util.ArrayList;
import java.util.EnumSet;
import java.util.List;
import java.util.Set;

public class FilerVolumeImpl<F extends Filer> extends OntapVolumeImpl<FilerVolume, F, FilerImpl<F>> implements FilerVolume {

    private static final Logger logger = Logger.getLogger(FilerVolumeImpl.class);

    private FilerVolume proxy;
    private static final Set<StorageSystem> storageSystems = EnumSet.of(StorageSystem.vfiler, StorageSystem.pfiler);

    private int earlyWarningThreshold;
    private boolean isSnapLock;
    private List<SnapVaultSchedule> snapVaultSchedules = new ArrayList<SnapVaultSchedule>();

    public FilerVolumeImpl(FilerImpl<F> controller) {
        super(controller);
    }

    @Override
    public OntapFile getFile(String path) {
        return SevenModeFileImpl.create(getControllerImpl(), this, path);
    }

    /**
     * Getter method used to get the early warning threshold percentage value.
     *
     * @return The early warning threshold value
     */
    @Override
    public Integer getEarlyWarningThreshold() {
        return earlyWarningThreshold;
    }

    /**
     * Setter method used to set the early warning threshold percentage value.
     *
     * @param warningThreshold The early warning threshold value
     */
    public void setEarlyWarningThreshold(int warningThreshold) {
        earlyWarningThreshold = warningThreshold;
    }

    /**
     * Returns the snaplock setting of this volume
     *
     * @return the snaplock setting of this volume
     */
    @Override
    public boolean isSnapLock() {
        return isSnapLock;
    }

    public void setSnapLock(boolean isSnapLock) {
        this.isSnapLock = isSnapLock;
    }

    @Override
    public List<SnapVaultSchedule> getSnapVaultSchedules() {
        return snapVaultSchedules;
    }

    public void setSnapVaultSchedules(List<SnapVaultSchedule> snapVaultSchedules) {
        this.snapVaultSchedules = snapVaultSchedules;
    }

    @Override
    public SnapVaultSchedule createSnapVaultSchedule(String scheduleName) {
        return new SnapVaultScheduleImpl(scheduleName, this, SnapVaultSchedule.Relation.primary);
    }

    @Override
    public String getSnapMirrorFormat() {
        // <filer>:/vol/<volume-name>
        return getController().getName() + ":" + "/vol/" + getName();
    }

    @Override
    public boolean matchesDestination(SnapMirrorRelation relation) {
        String snapMirrorFormat = getSnapMirrorFormat();
        return relation.getDestinationLocation().startsWith(snapMirrorFormat);
    }

    @Override
    public boolean matchesSource(SnapMirrorRelation relation) {
        String snapMirrorFormat = getSnapMirrorFormat();
        return relation.getSourceLocation().startsWith(snapMirrorFormat);
    }

    @Override
    public void move(Aggregate newAggr, boolean sync) {
        // TODO aeonflux-next - we *can* implement this, but AeonFlux 2.x did not, so
        // not going to do it now.
        throw new UnsupportedOperationException("not implemented for 7-mode");
    }

    @Override
    public void setQtreeSecurity(SecurityStyle securityStyle) {
        logger.debug("setQtreeSecurity:  " + getName() + ": " + securityStyle);
        NaElement xi;
        NaElement xo;
        NaElement xa;

        xi = new NaElement("system-cli");
        xi.addNewChild("priv", "admin");
        xa = new NaElement("args");

        FilerImpl filer = getControllerImpl();
        filer.addSystemCliArgs(xa);

        xa.addNewChild("arg", "qtree");
        xa.addNewChild("arg", "security");
        xa.addNewChild("arg", "/vol/" + getName());
        xa.addNewChild("arg", securityStyle.name());
        xi.addChildElem(xa);
        try {
            xo = filer.getNonPassThroughConnection().invokeZapi(xi, ZapiCache.NO_CACHE);
            filer.getZapiCache().blastCacheValues("qtree");
            int rc = xo.getChildIntValue("cli-result-value", 0);
            String output = xo.getChildContent("cli-output");
            if (rc != 1) {
                logger.error(String.format("Qtree (volume) security set to %s on %s failed. rc = %s. output = %s", securityStyle, getName(), rc, output));
            }
        } catch (Exception e) {
            logger.error(filer + " failed to set Qtree (volume) security to " + securityStyle + " on " + getName(), e);
            throw new VolumeOperationException(filer + " failed to set Qtree (volume) security  to " + securityStyle + " on " + getName(), e);
        }
    }

    @Override
    public void addToVFiler(VFiler vfiler) {
        logger.debug("addToVFiler: " + getName() + " vfiler: " + vfiler.getName());

        NaElement xi = new NaElement("vfiler-add-storage");
        xi.addNewChild("vfiler", vfiler.getName());
        xi.addNewChild("storage-path", "/vol/" + getName());

        try {
            NaElement xo = getControllerImpl().invokeZapi(xi, ZapiCache.NO_CACHE);
            logger.info("Added storage path: /vol/" + getName() + " to vfiler: " + vfiler + ". status: " + xo.getAttr("status"));
        } catch (Exception ex) {
            logger.error(getController() + " failed to add volume: " + this + " to vfiler: " + vfiler, ex);
            throw new VolumeOperationException(getController() + " failed to add storage path /vol/" + getName()
                    + " to vfiler: " + vfiler.getName() + ".  Error: " + ex.getMessage(), ex);
        }
    }

    @Override
    public void export(String exportPath, boolean writeEnable, boolean readWriteAccess, boolean rootAccess, boolean exportPathIsActual, Set<String> hosts) {
        logger.debug("Enter exportVolume: " + this + ", write = " + String.valueOf(writeEnable));
         logger.debug(String.format("Export path (%s)", exportPath));

         NaElement listRules = new NaElement("nfs-exportfs-list-rules-2");
         listRules.addNewChild("pathname", exportPath); //Only actually query about the volume we care about...
         NaElement xo = null;
         NaElement xiRules = null;
         String actualPath = "/vol/" + getName();

         String anonValue = null;

         /* It's not entirely clear to me the purpose of this block.
          * We are getting the list of existing exports and trying to find the path we're looking for in them.
          * 1) In 7-mode, a volume is given a default export string as soon as it's created so this will never fall through
          * 2) The ZAPIDoc for nfs-exportfs-modify-rule-2 explicitly says it's functionally similer to append-2 except it only works
          *    if the rule exists. Which will always be the case. So the different ZAPI would have no different behavior.
          * 3) The syntax on modify-rule-2 and append-rules-2 are not identical, so were we somehow to not find the export rule,
          *    the ZAPI would probably fail because the XML is formatted wrong.
          *
          *    What is going on here...?
          */
         NaElement xi = new NaElement("nfs-exportfs-append-rules-2");
         xiRules = new NaElement("rules");
         try {
             xo = getControllerImpl().invokeZapi(listRules, ZapiCache.NO_CACHE);
             getControllerImpl().getZapiCache().blastCacheValues("nfs");
             if (xo.getAttr("status").equals("passed")) {
                 if (xo.getChildByName("rules").getChildContent("exports-rule-info-2") != null) {
                     for (NaElement rule : xo.getChildByName("rules").getChildren()) {
                         if (rule.getChildContent("pathname").equals(exportPath)) {
                             anonValue = rule.getChildByName("security-rules").getChildByName("security-rule-info").getChildContent("anon");

                             xi = new NaElement("nfs-exportfs-modify-rule-2");
                             xiRules = new NaElement("rule");
                             break;
                         }
                     }
                 }
             }

         } catch (Exception e) {
             logger.error("Failed to list exports " + this + ".", e);
             throw new VolumeOperationException("Failed to list exports " + this + ". error: " +e.getMessage(), e);
         }

         xi.addChildElem(xiRules);
         xi.addNewChild("persistent", "true");

         NaElement eri2 = new NaElement("exports-rule-info-2");
         xiRules.addChildElem(eri2);

         eri2.addNewChild("pathname", exportPath);

         if (!exportPath.equals(actualPath) && !exportPathIsActual) {
             eri2.addNewChild("actual-pathname", actualPath);
         }

         NaElement xiSecRules = new NaElement("security-rules");
         eri2.addChildElem(xiSecRules);

         NaElement xiSecRuleInfo = new NaElement("security-rule-info");
         xiSecRules.addChildElem(xiSecRuleInfo);

         if(readWriteAccess) {
             NaElement xiPermissions;
             if (writeEnable){
                 xiPermissions = new NaElement("read-write");
             } else {
                 xiPermissions = new NaElement("read-only");
             }
             xiSecRuleInfo.addChildElem(xiPermissions);

             for(String ip : hosts) {
                 // if the IP address contains a ":", we're dealing with IPv6
                 if (ip.contains(":")) {
                     ip = String.format("[%s]", ip);
                 }

                 NaElement xiRwExpHostInfo = new NaElement("exports-hostname-info");
                 xiPermissions.addChildElem(xiRwExpHostInfo);
                 xiRwExpHostInfo.addNewChild("name", ip);
                 logger.debug("adding rw access to " + getName() + " for " + ip);
             }
         }

         if(rootAccess) {
             NaElement xiRoot = new NaElement("root");
             xiSecRuleInfo.addChildElem(xiRoot);

             for(String ip : hosts) {
                 // if the IP address contains a ":", we're dealing with IPv6
                 if (ip.contains(":")) {
                     ip = String.format("[%s]", ip);
                 }

                 NaElement xiRootExpHostInfo = new NaElement("exports-hostname-info");
                 xiRoot.addChildElem(xiRootExpHostInfo);
                 xiRootExpHostInfo.addNewChild("name", ip);
                 logger.debug("adding root access to " + getName() + " for " + ip);
             }
         }

         xiSecRuleInfo.addNewChild("nosuid", "true");
         xiSecRuleInfo.addNewChild("sec-flavor", "sys");
         if(anonValue != null){
             xiSecRuleInfo.addNewChild("anon", anonValue);
         }

         try {
             xo = getControllerImpl().invokeZapi(xi, ZapiCache.NO_CACHE);
             logger.info("Export " + getName() + ":" + xo.getAttr("status") + ".");

             xi = new NaElement("nfs-exportfs-load-exports");
             xo = getControllerImpl().invokeZapi(xi, ZapiCache.NO_CACHE);
             logger.debug("Reload exports :" + xo.getAttr("status") + ".");
         } catch (Exception ex) {
             logger.error("Failed to export " + this + ".", ex);
             throw new ExportOperationException("Failed to export " + this +".  Error: " +ex.getMessage(), ex);
         }
    }

    @Override
    public void export(Set<String> hosts) {
        export("/vol/"+getName(), true, hosts);
    }

    @Override
    public synchronized FilerVolume getProxy() {
        if (proxy == null) {
            proxy = CachingOfftapProxy.create(FilerVolume.class, FilerVolumeProxy.class, this, getControllerImpl().getDataCache());
        }
        return proxy;
    }

    @Override
    public Set<StorageSystem> getStorageSystems() {
        return storageSystems;
    }

    @Override
    public ControllerConnection getConnection() {
        return getControllerImpl().getConnection();
    }

    @Override
    public FilerVolume refresh() {
        return (FilerVolume) super.refresh();
    }
}
