package com.netapp.offtap3.internal.ontap.volume;

import com.google.common.base.Objects;
import com.netapp.offtap3.*;
import com.netapp.offtap3.internal.AbstractRefreshable;
import com.netapp.offtap3.internal.HasProxy;
import com.netapp.offtap3.internal.ontap.HasOntapController;
import com.netapp.offtap3.internal.ontap.controller.AbstractOntapDataController;
import com.netapp.offtap3.internal.ontap.controller.FilerImpl;
import com.netapp.offtap3.internal.ontap.controller.OntapConnection;
import com.netapp.offtap3.internal.ontap.lun.OntapLunSpecImpl;
import com.netapp.offtap3.internal.ontap.snap.VolumeSnapMirrorStatusImpl;
import com.netapp.offtap3.internal.ontap.support.ZapiCache;
import com.netapp.offtap3.ontap.*;
import com.netapp.offtap3.utils.SafeConcurrentHashMap;
import com.netapp.offtap3.utils.StringUtils;
import org.apache.log4j.Logger;

import java.util.*;

public abstract class OntapVolumeImpl<I extends OntapVolume, C extends OntapDataController, M extends AbstractOntapDataController<C>> extends
        AbstractRefreshable implements OntapVolume, HasProxy<I>, HasOntapController<C, M> {
    private static final Logger LOGGER = Logger.getLogger(OntapVolumeImpl.class);

    private M controller;

    private String uuid;
    private String type;
    private String state;
    private boolean spaceReserveEnabled;
    private String spaceReserve;

    private Size sizeUsed;
    private Size sizeTotal;
    private Size sizeAvailable;
    private int percentageUsed;
    private boolean isInconsistent;
    private String containingAggregate;

    // snapshot reserves
    private int snapshotPercentReserved;
    private Size snapshotSpaceReserved;
    private int snapshotReservePercentUsed;
    private Size snapshotReserveUsed;

    // auto delete
    private String autoDeleteTrigger = null;
    private String autoDeleteStatus = null;

    private String parentSnapshotName;
    private String parentVolumeName;


    private Size totalSnapshotSpace;
    private int cloneChildCount;

    private Dedupe dedupe;

    // auto-size info
    private Size autoSizeIncrement;
    private Size autoSizeMax;
    private boolean autoSizeEnabled;

    private VolumeSnapMirrorStatusImpl snapMirror = new VolumeSnapMirrorStatusImpl();

    private boolean isConvertUcodeEnabled;
    private boolean isCreateUcodeEnabled;
    private boolean isAtimeUpdateEnabled;
    private int fractionalReserve;

    // qtrees
    private Map<String, Qtree> qtrees = new SafeConcurrentHashMap<String, Qtree>();
    private Map<String, Export> rawExports = new SafeConcurrentHashMap<String, Export>();

    private String junctionPath;
    private List<Snapshot> snapshots = new ArrayList<Snapshot>();

    //alignment
    private int vmAlignSector = 0;
    private String vmAlignSuffix = null;

    /**
     * Creates a <code>VolumeImpl</code> instance with no values set.
     */
    public OntapVolumeImpl(M controller) {
        super(controller);
        this.controller = controller;
    }

    @Override
    public C getController() {
        return controller.getProxy();
    }

    @Override
    public M getControllerImpl() {
        return controller;
    }

    /**
     * @return the free space of this volume
     */
    @Override
    public Size getFreeSpace() {
        return sizeAvailable;
    }

    /**
     * @return the uuid of this volume
     */
    @Override
    public String getUuid() {
        return uuid;
    }

    public void setUuid(String uuid) {
        this.uuid = uuid;
    }

    /**
     * @return the type of this volume
     */
    @Override
    public String getType() {
        return type;
    }

    public void setType(String type) {
        this.type = type;
    }

    /**
     * @return the state of this volume
     */
    @Override
    public String getState() {
        return state;
    }

    /**
     * Sets the state of this volume
     *
     * @param state the state of this volume
     */
    public void setState(String state) {
        this.state = state;
    }

    /**
     * @return the space reserve enabled setting
     */
    @Override
    public boolean isSpaceReserveEnabled() {
        return spaceReserveEnabled;
    }

    public void setSpaceReserveEnabled(boolean spaceReserveEnabled) {
        this.spaceReserveEnabled = spaceReserveEnabled;
    }

    /**
     * @return the space reserve value
     */
    @Override
    public String getSpaceReserve() {
        return spaceReserve;
    }

    public void setSpaceReserve(String spaceReserve) {
        this.spaceReserve = spaceReserve;
    }

    /**
     * Returns <code>true</code> if this volume is thin provisioned;
     * <code>false</code> otherwise
     *
     * @return <code>true</code> if this volume is thin provisioned;
     * <code>false</code> otherwise
     */
    @Override
    public boolean isThinProvisioned() {
        return "none".equals(spaceReserve);
    }

    /**
     * @return the used size of this volume
     */
    @Override
    public Size getSizeUsed() {
        return sizeUsed;
    }

    public void setSizeUsed(long sizeUsed) {
        this.sizeUsed = Size.createB(sizeUsed);
    }

    /**
     * @return the total size of this volume
     */
    @Override
    public Size getSizeTotal() {
        return sizeTotal;
    }

    public void setSizeTotal(long sizeTotal) {
        this.sizeTotal = Size.createB(sizeTotal);
    }

    /**
     * Returns <code>true</code> if this volume is inconsistent;
     * <code>false</code> otherwise
     *
     * @return <code>true</code> if this volume is inconsistent;
     * <code>false</code> otherwise
     */
    @Override
    public boolean isInconsistent() {
        return isInconsistent;
    }

    public void setInconsistent(boolean isInconsistent) {
        this.isInconsistent = isInconsistent;
    }
    
    
    /**
     * Returns the containing aggregate name of this volume
     *
     * @return the containing aggregate name of this volume
     */
    @Override
    public String getContainingAggregateName() {
        return containingAggregate;
    }    

    /**
     * Returns the containing aggregate of this volume
     *
     * @return the containing aggregate of this volume
     */
    @Override
    public Aggregate getContainingAggregate() {
        return getController().getAggregates().get(containingAggregate);
    }

    /**
     * Set the name of the AggregateImpl
     *
     * @param containingAggregate aggr name
     */
    public void setContainingAggregate(String containingAggregate) {
        this.containingAggregate = containingAggregate;
    }

    /**
     * Returns the used percentage of this volume
     *
     * @return the used percentage of this volume
     */
    @Override
    public int getPercentageUsed() {
        return percentageUsed;
    }

    public void setPercentageUsed(int percentageUsed) {
        this.percentageUsed = percentageUsed;
    }

    /**
     * @return the available size of this volume
     */
    @Override
    public Size getSizeAvailable() {
        return sizeAvailable;
    }

    public void setSizeAvailable(long sizeAvailable) {
        this.sizeAvailable = Size.createB(sizeAvailable);
    }

    /**
     * Returns the parent snapshot name
     *
     * @return the parent snapshot name
     */
    @Override
    public String getParentSnapshotName() {
        return parentSnapshotName;
    }

    public void setParentSnapshotName(String parentSnapshotName) {
        this.parentSnapshotName = parentSnapshotName;
    }

    /**
     * Returns the parent volume name
     *
     * @return the parent volume name
     */
    @Override
    public String getParentVolumeName() {
        return parentVolumeName;
    }

    public void setParentVolumeName(String parentVolumeName) {
        this.parentVolumeName = parentVolumeName;
    }

    /**
     * Returns the total snapshot space
     *
     * @return the total snapshot space
     */
    @Override
    public Size getTotalSnapshotSpace() {
        return totalSnapshotSpace;
    }

    /**
     * Number of 1024 byte blocks in the snapshot. In other words, this value is
     * in KB.
     */
    public void setTotalSnapshotSpace(long totalSnapshotSpace) {
        this.totalSnapshotSpace = new Size(totalSnapshotSpace, Size.Unit.KB);
    }

    /**
     * Returns the snapshot percent reserved
     *
     * @return the snapshot percent reserved
     */
    @Override
    public int getSnapshotPercentReserved() {
        return snapshotPercentReserved;
    }

    public void setSnapshotPercentReserved(int snapshotPercentReserved) {
        this.snapshotPercentReserved = snapshotPercentReserved;
    }

    /**
     * Returns the snapshot space reserved
     *
     * @return the snapshot space reserved
     */
    @Override
    public Size getSnapshotSpaceReserved() {
        return snapshotSpaceReserved;
    }

    public void setSnapshotSpaceReserved(long snapshotSpaceReserved) {
        this.snapshotSpaceReserved = Size.createB(snapshotSpaceReserved);
    }

    /**
     * Get snapshot reserve percentage used.
     *
     * @return Percentage of snapshot reserve used.
     */
    @Override
    public int getSnapshotReservePercentUsed() {
        return snapshotReservePercentUsed;
    }

    /**
     * Set snapshot reserve percentage used.
     *
     * @param snapshotReservePercentUsed Percentage of usage.
     */
    public void setSnapshotReservePercentUsed(int snapshotReservePercentUsed) {
        this.snapshotReservePercentUsed = snapshotReservePercentUsed;
    }

    @Override
    public Size getSnapshotReserveUsed() {
        return snapshotReserveUsed;
    }

    public void setSnapshotReserveUsed(long snapshotReserveUsed) {
        this.snapshotReserveUsed = Size.createB(snapshotReserveUsed);
    }

    /**
     * Returns the clone child count
     *
     * @return the clone child count
     */
    @Override
    public int getCloneChildCount() {
        return cloneChildCount;
    }

    public void setCloneChildCount(int cloneChildCount) {
        this.cloneChildCount = cloneChildCount;
    }

    @Override
    public Dedupe getDedupe() {
        return dedupe;
    }

    public void setDedupe(Dedupe dedupe) {
        this.dedupe = dedupe;
    }

    /**
     * @return the trigger as to when snapshot auto delete kicks in
     */
    @Override
    public String getAutoDeleteTrigger() {
        return autoDeleteTrigger;
    }

    public void setAutoDeleteTrigger(String autoDeleteTrigger) {
        this.autoDeleteTrigger = autoDeleteTrigger;
    }

    /**
     * @return The string status of snapshot autodelete.
     */
    @Override
    public String getAutoDeleteStatus() {
        return autoDeleteStatus;
    }

    /**
     * Set the snapshot autodelete status string.
     *
     * @param autoDeleteStatus The status string for snapshot. autodelete
     */
    public void setAutoDeleteStatus(String autoDeleteStatus) {
        this.autoDeleteStatus = autoDeleteStatus;
    }

    /**
     * @return the shared SIS size
     */
    @Override
    public Size getSharedSize() {
        return getDedupe().getSizeSaved().plus(getSizeUsed());
    }

    /**
     * @return the auto size increment value
     */
    @Override
    public Size getAutoSizeIncrement() {
        return autoSizeIncrement;
    }

    public void setAutoSizeIncrement(Size increment) {
        this.autoSizeIncrement = increment;
    }

    public void setAutoSizeIncrement(long autoSizeIncrement) {
        this.autoSizeIncrement = Size.createB(autoSizeIncrement);
    }

    /**
     * @return the auto size maximum value
     */
    @Override
    public Size getAutoSizeMax() {
        return autoSizeMax;
    }

    public void setAutoSizeMax(Size max) {
        this.autoSizeMax = max;
    }

    public void setAutoSizeMax(long autoSizeMax) {
        this.autoSizeMax = Size.createB(autoSizeMax);
    }

    /**
     * Return <code>true</code> if auto-size is eanbled on this VolumeImpl;
     * <code>false</code> otherwise
     *
     * @return <code>true</code> if auto-size is eanbled on this VolumeImpl;
     * <code>false</code> otherwise
     */
    @Override
    public boolean isAutoSizeEnabled() {
        return autoSizeEnabled;
    }

    public void setAutoSizeEnabled(boolean autoSizeEnabled) {
        this.autoSizeEnabled = autoSizeEnabled;
    }

    @Override
    public VolumeSnapMirrorStatusImpl getSnapMirrorStatus() {
        return snapMirror;
    }

    public void setSnapMirrorStatus(VolumeSnapMirrorStatusImpl mirror) {
        this.snapMirror = mirror;
    }

    @Override
    public boolean isConvertUcodeEnabled() {
        return isConvertUcodeEnabled;
    }

    public void setConvertUcodeEnabled(boolean isConvertUcodeEnabled) {
        this.isConvertUcodeEnabled = isConvertUcodeEnabled;
    }

    @Override
    public boolean isCreateUcodeEnabled() {
        return isCreateUcodeEnabled;
    }

    public void setCreateUcodeEnabled(boolean isCreateUcodeEnabled) {
        this.isCreateUcodeEnabled = isCreateUcodeEnabled;
    }

    @Override
    public boolean isAtimeUpdateEnabled() {
        return isAtimeUpdateEnabled;
    }

    public void setAtimeUpdateEnabled(boolean isAtimeUpdateEnabled) {
        this.isAtimeUpdateEnabled = isAtimeUpdateEnabled;
    }

    @Override
    public int getFractionalReserve() {
        return fractionalReserve;
    }

    public void setFractionalReserve(int fractionalReserve) {
        this.fractionalReserve = fractionalReserve;
    }

    @Override
    public Map<String, Export> getExports() {
        Map<String, Export> exports = new SafeConcurrentHashMap<String, Export>();
        for (Qtree qtree : getQtrees().values()) {
            exports.putAll(qtree.getExports());
        }
        exports.putAll(rawExports);
        return exports;
    }



    public Map<String, Export> getRawExports() {
        return rawExports;
    }

    @Override
    public OntapLunSpec.Builder getLunSpecBuilder(String path, String type, Size size) {
        return new OntapLunSpecImpl.BuilderImpl(path, this, type, size);
    }

    @Override
    public Map<String, Qtree> getQtrees() {
        return qtrees;
    }

    public void setQtrees(Map<String, Qtree> qtrees) {
        this.qtrees = qtrees;
    }

    @Override
    public Qtree getQtree(String relativePath) {
        String qtreePath = "/vol/" + getName();
        if (relativePath != null) {
            if (relativePath.startsWith("/")) {
                qtreePath = relativePath;
            } else {
                qtreePath += "/" + relativePath;
            }
        }
        return qtrees.get(qtreePath);
    }

    @Override
    public String getJunctionPath() {
        return junctionPath;
    }

    public void setJunctionPath(String junctionPath) {
        this.junctionPath = junctionPath;
    }

    @Override
    public OntapLun getLun(String path) {
        return getController().getLun(toFullPath(path));
    }

    @Override
    public Snapshot getSnapshotByName(String name) {
        for (Snapshot curr : getSnapshots()) {
            if (curr.getName() != null && curr.getName().equals(name)) {
                return curr;
            }
        }
        for (Snapshot curr : getSnapshots()) {
            // so - we append a .0 when creating snapshots - not really sure why - but don't want to
            // change that without knowledge of who cares.  So we'll make getSnapshotByName
            // smarter so that it looks for the proper name, as well as the .0 name and
            // encapsulate that hack at least a little bit.

            if (curr.getName() != null && curr.getName().equals(name + ".0")) {
                return curr;
            }
        }
        return null;
    }

    @Override
    public List<Snapshot> getSnapshots() {
        return snapshots;
    }

    public void setSnapshots(List<Snapshot> snapshots) {
        this.snapshots = snapshots;
    }

    @Override
    public void inititalizeSnapMirror(String sourceVserver, String sourceVolume) {
        NaElement xi = new NaElement("snapmirror-initialize");

        xi.addNewChild("destination-location", getSnapMirrorFormat());

        if (StringUtils.isReal(sourceVolume) && StringUtils.isReal(sourceVserver)) {
            xi.addNewChild("source-vserver", sourceVserver);
            xi.addNewChild("source-volume", sourceVolume);
        }

        try {
            getControllerImpl().invokeZapi(xi, ZapiCache.NO_CACHE);
            getControllerImpl().getZapiCache().blastCacheValues("snap");
        } catch (ApiFailedException e) {
            LOGGER.error("Unable to initialize SnapMirror.", e);
            throw new SnapMirrorOperationException("Unable to initialize SnapMirror.", e);
        }
    }


    @Override
    public void initializeSnapMirror(OntapVolume sourceVol) {
        NaElement xi = new NaElement("snapmirror-initialize");

        xi.addNewChild("destination-location", getSnapMirrorFormat());

        if (sourceVol != null) {
            xi.addNewChild("source-location", sourceVol.getSnapMirrorFormat());
        }

        try {
            getControllerImpl().invokeZapi(xi, ZapiCache.NO_CACHE);
            getControllerImpl().getZapiCache().blastCacheValues("snap");
        } catch (ApiFailedException e) {
            LOGGER.error("Unable to initialize SnapMirror.", e);
            throw new SnapMirrorOperationException("Unable to initialize SnapMirror.", e);
        }
    }

    @Override
    public List<SnapMirrorRelation> getSnapMirrorRelations() {
        List<SnapMirrorRelation> onVolume = new ArrayList<SnapMirrorRelation>();
        for (SnapMirrorRelation relation : getController().getSnapMirrorRelations()) {
            if (matchesDestination(relation) || matchesSource(relation)) {
                onVolume.add(relation);
            }
        }
        return onVolume;
    }

    public List<SnapMirrorRelation> getSnapMirrorDestinationRelations() {
        List<SnapMirrorRelation> onVolume = new ArrayList<SnapMirrorRelation>();
        List<? extends SnapMirrorRelation> snapMirrorRelations = getController().getSnapMirrorRelations();
        LOGGER.debug("getSnapMirrorDestinationRelations, svm = " + getControllerImpl() + ", relations size = " + snapMirrorRelations.size());
        for (SnapMirrorRelation relation : snapMirrorRelations) {
            if (matchesDestination(relation)) {
                onVolume.add(relation);
            }
        }
        LOGGER.debug("found " + onVolume.size() + " relations on volume.");
        return onVolume;
    }

    public List<SnapMirrorRelation> getSnapMirrorDestinationRelations(OntapVolume source) {
        if (source == null) {
            throw new NullPointerException("the source volume was null");
        }

        List<SnapMirrorRelation> onVolume = new ArrayList<SnapMirrorRelation>();
        for (SnapMirrorRelation relation : getController().getSnapMirrorRelations()) {
            if (matchesDestination(relation) && source.matchesSource(relation)) {
                onVolume.add(relation);
            }
        }
        return onVolume;
    }

    /**
     * Get the shim size in sectors (0-7) used for alignment correction
     */
    @Override
    public int getVmAlignSector() {
        return vmAlignSector;
    }

    public void setVmAlignSector(int vmAlignSector) {
        this.vmAlignSector = vmAlignSector;
    }

    /**
     * @return filename suffix that vmAlign shim is applied to.
     */
    @Override
    public String getVmAlignSuffix() {
        return vmAlignSuffix;
    }

    public void setVmAlignSuffix(String vmAlignSuffix) {
        this.vmAlignSuffix = vmAlignSuffix;
    }

    public boolean equals(Object obj) {
        if (obj instanceof OntapVolume) {
            return Objects.equal(getUuid(), ((OntapVolume) obj).getUuid());
        }
        return false;
    }

    public int hashCode() {
        return getUuid().hashCode();
    }

    @Override
    public String toString() {
        StringBuilder builder = new StringBuilder("OntapVolume[name=");
        builder.append(getName());
        builder.append(", containingAggregate=");
        builder.append(containingAggregate);
        builder.append(", sizeAvailable=");
        builder.append(sizeAvailable);
        builder.append(", sizeTotal=");
        builder.append(sizeTotal);
        builder.append(", sizeUsed=");
        builder.append(sizeUsed);
        builder.append(", state=");
        builder.append(state);
        builder.append(", type=");
        builder.append(type);
        builder.append(", uuid=");
        builder.append(uuid);
        builder.append("]");
        return builder.toString();
    }

    @Override
    public String getPath(boolean nfsStyle) {
        NaElement xi = new NaElement("volume-get-volume-path");
        xi.addNewChild("volume", getName());
        xi.addNewChild("is-style-cifs", "" + !nfsStyle);
        try {
            NaElement xo = controller.invokeZapi(xi, ZapiCache.NO_CACHE);
            return xo.getChildContent("junction");
        } catch (ApiFailedException nfe) {
            return null;
        }
    }

    @Override
    public void rename(String newVolumeName) throws VolumeOperationException {
        NaElement xi = new NaElement("volume-rename");
        xi.addNewChild("volume", getName());
        xi.addNewChild("new-volume-name", newVolumeName);
        try {
            controller.invokeZapi(xi, ZapiCache.NO_CACHE);
            controller.getZapiCache().blastCacheValues("volume");
            LOGGER.info("Renamed volume " + getName() + " to " + newVolumeName + ".");
            setName(newVolumeName);
        } catch (ApiFailedException e) {
            LOGGER.error("volumeRename: " + controller + " Unable to renameVolume.", e);
            throw new VolumeOperationException("volumeRename: " + controller + " Unable to renameVolume.", e);
        }
    }

    @Override
    public void restrict() throws VolumeOperationException {
        LOGGER.trace("restrict(" + controller.getName() + ":" + getName() + "): enter");

        NaElement xi;

        try {
            LOGGER.info("Restricting " + getName() + ".");
            xi = new NaElement("volume-restrict");
            xi.addNewChild("name", getName());
            controller.invokeZapi(xi, ZapiCache.NO_CACHE);
            LOGGER.debug(getName() + " is now restricted.");

        } catch (ApiFailedException e) {
            if (e.getErrno() == 13061) {
                LOGGER.debug(getName() + " was already restricted.");
            } else {
                LOGGER.error("Failed to restrict " + getName() + ".", e);
                throw new VolumeOperationException("Failed to restrict volume " + getName(), e);
            }
        }

        setState("restricted");

        LOGGER.debug("restrict(" + controller.getName() + ":" + getName() + "): exit");
    }

    @Override
    public void destroy(boolean destroyParentIfLastClone) {
        destroy();

        if (destroyParentIfLastClone) {
            OntapDataController controller = getController();
            OntapVolume parent = controller.getVolume(getParentVolumeName());
            if (parent.getCloneChildCount() == 0) {
                LOGGER.debug("destroyVolume removing parent " + parent.getName());
                parent.destroy();
            } else {
                LOGGER.debug("not destroying parent because it still has child clones");
            }
        }
    }

    @Override
    public void destroy() throws VolumeOperationException {

        offline();

        String volumeName = getName();
        AbstractOntapDataController controller = getControllerImpl();
        LOGGER.info(controller + " is going to destroy " + volumeName + ".");
        NaElement xi = new NaElement("volume-destroy");
        xi.addNewChild("name", volumeName);

        try {
            controller.invokeZapi(xi, ZapiCache.NO_CACHE);
            controller.getVolumes().remove(getName());
            controller.getZapiCache().blastCacheValues("volume-", "aggr-");
            LOGGER.debug(controller + " has destroyed " + volumeName);
        } catch (ApiFailedException ex) {
            controller.addZapiExecutionError("Failed to destroy " + volumeName, xi, ex);
            LOGGER.error(controller + " failed to destroy  " + volumeName + ".", ex);
            throw new VolumeOperationException(controller + " failed to destroy " + volumeName, ex);
        }

    }

    @Override
    public void offline() throws VolumeOperationException {
        LOGGER.debug(controller + " will offline volume: " + getName());

        try {
            NaElement xi = new NaElement("volume-offline");
            xi.addNewChild("name", getName());

            controller.invokeZapi(xi, ZapiCache.NO_CACHE);
            controller.getZapiCache().blastCacheValues("volume-");
            LOGGER.debug(controller + " has offlined volume : " + getName());
        } catch (ApiFailedException ex) {
            if (ex.getErrno() == NaErrno.EVOLUMEOFFLINE) {
                LOGGER.debug(controller + " volume " + getName() + " was already offline");
            } else {
                LOGGER.error(controller + " failed to offline " + getName() + ".", ex);
                throw new VolumeOperationException(
                        controller + " failed to offline " + getName() + ".  Error: " + ex.getMessage(), ex);
            }
        }
        setState("offline");
    }

    @Override
    public void online() throws VolumeOperationException {
        try {
            NaElement xi = new NaElement("volume-online");
            xi.addNewChild("name", getName());
            controller.invokeZapi(xi, ZapiCache.NO_CACHE);
            controller.getZapiCache().blastCacheValues("volume-");
            LOGGER.debug(getName() + " is now online.");
        } catch (ApiFailedException ex) {
            // Fail silently if the volume is already online (errno=13060)
            if (ex.getErrno() != 13060) {
                LOGGER.error("Failed to online " + getName() + ".", ex);
                throw new VolumeOperationException("Failed to online " + getName() + ".  Error: " + ex.getMessage(), ex);
            }
            else {
                LOGGER.debug("The volume with name " + getName() + " is already online.");
            }
        }
        setState("online");
    }

    @Override
    public void resize(Size size) throws VolumeOperationException {
        LOGGER.debug(controller + " resizeVolume: " + getName() + " to " + size.toMB() + " MB.");

        // if the fs_size_fixed option is on, turn it off prior to resizing
        NaElement xi = new NaElement("volume-set-option");
        xi.addNewChild("volume", getName());
        xi.addNewChild("option-name", "fs_size_fixed");
        xi.addNewChild("option-value", "off");
        try {
            controller.invokeZapi(xi, ZapiCache.NO_CACHE);
            controller.getZapiCache().blastCacheValues("volume-");
        } catch (Exception ex) {
            // log the error and go
            LOGGER.warn(String.format("WARNING ONLY: Failed to set the fs_size_fixed option on volume (%s) to off",
                    getName()));
        }

        // check for passthrough as it no longer works in BR.2 or RR RC2
        OntapConnection localConnection = controller.getConnection();
        if (controller instanceof FilerImpl) {
            localConnection = ((FilerImpl) controller).getNonPassThroughConnection();
        }

        xi = new NaElement("volume-size");
        xi.addNewChild("volume", getName());
        xi.addNewChild("new-size", size.toMB() + "m");
        try {
            NaElement xo = localConnection.invokeZapi(xi, ZapiCache.NO_CACHE);
            controller.getZapiCache().blastCacheValues("volume-");
            String sizeResponse = xo.getChildContent("volume-size");
            try {
                setSizeTotal(Size.parse(xo.getChildContent("volume-size")).toBytes());
            } catch (RuntimeException e) {
                LOGGER.warn("Unexpected response in resize ZAPI (" + sizeResponse +
                        ") that could not be parsed.  Will be safely ignored: " + e.getMessage());
                setSizeTotal(size.toBytes());
            }

        } catch (ApiFailedException ex) {
            LOGGER.error("Failed to resize volume " + getName() + " to " + size.toMB() + " MB.", ex);
            throw new VolumeOperationException(
                    "Failed to resize volume " + getName() + " to " + size.toMB() + " MB.  Error: " + ex.getMessage(),
                    ex);
        }
    }

    @Override
    public void setAutosize(Size increment, Size max, boolean enabled) throws VolumeOperationException {
        LOGGER.debug("setVolAutoSize: " + getName() + " inc " + increment + " max " + max +
                " enable " + enabled);
        NaElement xi;
        NaElement xo;
        xi = new NaElement("volume-autosize-set");
        xi.addNewChild("volume", getName());
        xi.addNewChild("is-enabled", String.valueOf(enabled));
        xi.addNewChild("increment-size", increment.toMB() + "m");
        xi.addNewChild("maximum-size", max.toMB() + "m");

        try {
            xo = controller.invokeZapi(xi, ZapiCache.NO_CACHE);
            setAutoSizeIncrement(increment);
            setAutoSizeMax(max);
            setAutoSizeEnabled(enabled);
            controller.getZapiCache().blastCacheValues("volume-");
            LOGGER.info("AutoSize for " + getName() + ":" + xo.getAttr("status") + ".");
        } catch (ApiFailedException ex) {
            LOGGER.error("Failed to set AutoSize for " + getName() + ".", ex);
            throw new VolumeOperationException(
                    "Failed to set AutoSize for " + getName() + ".  Error: " + ex.getMessage(), ex);
        }
    }

    @Override
    public void setSis(boolean dedupeEnabled, boolean dedupeOn, boolean rescan) throws VolumeOperationException {
        LOGGER.debug(controller + " setVolumeSis: start for " + getName());

        if (dedupeEnabled) {
            if (!getDedupe().isEnabled()) {
                NaElement xi = new NaElement("sis-enable");
                xi.addNewChild("path", "/vol/" + getName());
                try {
                    controller.invokeZapi(xi, ZapiCache.NO_CACHE);
                    controller.getZapiCache().blastCacheValues("volume-", "sis-");
                } catch (ApiFailedException ex) {
                    LOGGER.error(controller + " failed to enable sis on " + getName(), ex);
                    throw new VolumeOperationException(controller + " failed to enable sis on " +
                            getName() + ". Error: " + ex.getMessage(), ex);
                }
            }
            if (dedupeOn) {
                NaElement xi = new NaElement("sis-start");
                xi.addNewChild("path", "/vol/" + getName());
                xi.addNewChild("scan", rescan ? "true" : "false");
                try {
                    controller.invokeZapi(xi, ZapiCache.NO_CACHE);
                    controller.getZapiCache().blastCacheValues("volume-", "sis-");
                } catch (ApiFailedException ex) {
                    LOGGER.error(controller + " failed to start sis for " + getName(), ex);
                    throw new VolumeOperationException(
                            controller + " failed to start sis for " + getName() + ". Error: " + ex.getMessage(), ex);
                }
            }
            if (!dedupeOn && getDedupe().isActive()) {
                LOGGER.debug(controller + " issuing the sis-stop command for " + getName());
                NaElement xi = new NaElement("sis-stop");
                xi.addNewChild("path", "/vol/" + getName());
                try {
                    controller.invokeZapi(xi, ZapiCache.NO_CACHE);
                    controller.getZapiCache().blastCacheValues("volume-", "sis-");
                } catch (ApiFailedException ex) {
                    LOGGER.error(controller + " failed to stop sis for " + getName(), ex);
                    throw new VolumeOperationException(
                            controller + " failed to stop sis for " + getName() + ". Error: " + ex.getMessage(), ex);
                }
            }

        } else {
            boolean dedupeStopped = false;
            if (!dedupeOn && getDedupe().isActive()) {
                LOGGER.debug(controller + " issuing the sis-stop command for " + getName());
                NaElement xi = new NaElement("sis-stop");
                xi.addNewChild("path", "/vol/" + getName());
                try {
                    controller.invokeZapi(xi, ZapiCache.NO_CACHE);
                    controller.getZapiCache().blastCacheValues("volume-", "sis-");
                    dedupeStopped = true;
                } catch (ApiFailedException ex) {
                    LOGGER.error(controller + " failed to stop sis for " + getName(), ex);
                    throw new VolumeOperationException(
                            controller + " failed to stop sis fpr " + getName() + ". Error: " + ex.getMessage(), ex);
                }
            }

            // if we issued a sis-stop, pause for 3 seconds before issuing a sis-disable.  Otherwise the ZAPI goes too early and we receive
            // an error stating that the sis operation is currently active
            if (dedupeStopped) {
                try {
                    Thread.sleep(3000);
                } catch (InterruptedException ex) {
                    // no need to bubble this up
                }
            }

            if (getDedupe().isEnabled()) {
                LOGGER.debug(controller + " issuing the sis-disable command for " + getName());
                NaElement xi = new NaElement("sis-disable");
                xi.addNewChild("path", "/vol/" + getName());
                try {
                    controller.invokeZapi(xi, ZapiCache.NO_CACHE);
                    controller.getZapiCache().blastCacheValues("volume-", "sis-");
                } catch (ApiFailedException ex) {
                    LOGGER.error(controller + " failed to disable sis for " + getName(), ex);
                    throw new VolumeOperationException(
                            controller + " failed to disable sis for " + getName() + ". Error: " + ex.getMessage(), ex);
                }
            }
        }

        LOGGER.debug(controller + " setSis for " + getName() + " end.");
    }

    @Override
    public void setOptions(String[][] options) throws VolumeOperationException {
        LOGGER.debug("setVolumeOptions (start): " + getName());

        for (String[] option : options) {
            NaElement xi = new NaElement("volume-set-option");
            xi.addNewChild("volume", getName());
            xi.addNewChild("option-name", option[0]);
            xi.addNewChild("option-value", option[1]);
            try {
                controller.invokeZapi(xi, ZapiCache.NO_CACHE);
                controller.getZapiCache().blastCacheValues("volume-");
            } catch (ApiFailedException ex) {
                if (ex.getErrno() == 13114) {
                    // we can ignore this.
                    // VolumeImpl option fractional_reserve can be modified only if the volume has its guarantee option set to 'volume' (errno=13114)
                } else if (ex.getErrno() == 160) {
                    // Modification of field(s) auto-snapshots-enabled not allowed for a(n) Read-Write volume. (errno=160)
                } else {
                    LOGGER.error("Failed to set volume options.", ex);
                    throw new VolumeOperationException("Failed to set volume options. Error: " + ex.getMessage(), ex);
                }
            }
            LOGGER.info("Applied option " + option[0] + " " + option[1] + " to " + getName() + ".");
        }
        LOGGER.debug("setVolumeOptions (end): " + getName());
    }

    @Override
    public void setSnapReserve(int percent) throws VolumeOperationException {
        LOGGER.debug("setVolSnapReserve (start): " + getName() + " " + percent);

        NaElement xi;
        NaElement xo;
        xi = new NaElement("snapshot-set-reserve");
        xi.addNewChild("volume", getName());
        xi.addNewChild("percentage", String.valueOf(percent));
        try {
            xo = controller.invokeZapi(xi, ZapiCache.NO_CACHE);
            controller.getZapiCache().blastCacheValues("volume-", "snapshot");
            LOGGER.info("Set SnapReserve to " + String.valueOf(percent) + "% for " + getName() + ":" +
                    xo.getAttr("status") + ".");
        } catch (ApiFailedException ex) {
            LOGGER.error("Failed to set volume snap reserve for " + getName() + ".", ex);
            throw new VolumeOperationException(
                    "Failed to set volume snap reserve for " + getName() + ". Error: " + ex.getMessage(), ex);
        }

        LOGGER.debug("setVolSnapReserve (end)");
    }

    public String toFullPath(String path) {
        if (path.equals(".") || (path.equals("/"))) {
            path = "";
        }
        if (!path.startsWith("/vol/" + getName())) {
            path = "/vol/" + getName() + "/" + path;
        }
        return path;
    }

    @Override
    public Snapshot createSnapshot(String snapshotName) {
        String volumeName = getName();
        LOGGER.debug("createSnapshot on volume " + volumeName + " with name " + snapshotName);

        String name = snapshotName + ".0";
        List<Snapshot> snapshots = getProxy().getSnapshots();
        int version = 0;
        for (int i = 0; i < snapshots.size(); i++) {
            if (snapshots.get(i).getName().equals(name)) {
                version++;
                i = 0;
                name = snapshotName + "." + version;
            }
        }

        NaElement xi = new NaElement("snapshot-create");
        xi.addNewChild("volume", volumeName);
        xi.addNewChild("snapshot", name);
        try {
            controller.invokeZapi(xi, ZapiCache.NO_CACHE);
            controller.getZapiCache().blastCacheValues("snapshot-");

            getProxy().refresh(VolumeResource.snapshots);

            Snapshot snapshot = getSnapshotByName(name);
            LOGGER.info("Created Snapshot " + name + ".");
            return snapshot;
        } catch (ApiFailedException e) {
            LOGGER.error(controller + " failed to create Snapshot " + name + " on " + this, e);
            throw new VolumeOperationException(controller + " failed to create Snapshot " + name + " on " + this, e);
        }
    }

    @Override
    public void export(String exportPath, boolean writeEnable, boolean readWriteAccess, boolean rootAccess, Set<String> hosts) {
        export(exportPath, writeEnable, readWriteAccess, rootAccess, true, hosts);
    }

    @Override
    public void export(String exportPath, boolean writeEnable, Set<String> hosts) {
        export(exportPath, writeEnable, true, true, hosts);
    }

    @Override
    public void refresh(Resource... resources) {
        // Handled by the proxy.
    }

    @Override
    public Refreshable refresh(Refreshable refreshable) {
        // Handled by the proxy
        return null;
    }

    @Override
    public Volume refresh() {
        getProxy().refresh(VolumeResource.values());
        setLastLoad(new Date());
        return this;
    }
}
