package com.netapp.offtap3.internal.ontap.volume;


import com.netapp.offtap3.Size;
import com.netapp.offtap3.ontap.Aggregate;
import com.netapp.offtap3.ontap.OntapVolumeSpec;
import com.netapp.offtap3.ontap.SecurityStyle;
import com.netapp.offtap3.ontap.SpaceReserve;
import com.netapp.offtap3.ontap.VolumeType;

/**
 * Specifies the parameters for creating a volume.  Provides a builder with a fluent interface to
 * accommodate the many options:<ul>
 * <li> volumeName - Name for the new volume.
 * <li> aggrName - Name of the aggragate to create the volume on
 * <li> size - The size of the volume to be created
 * <li> spaceReserve - none, file or volume
 * <li> securityStyle - the security stype to apply to qtree0 (the volume)
 * <li> shimSize - The size in sectors of the "shim" to be placed in files that match shimSuffix. (vm-align-sector)
 * <li> shimSuffix - The suffix of file names that WAFL should apply the shim to. (vm-align-suffix)
 * <li> snapshotPolicy - The name of the snapshot policy to use (snapshot-policy).
 *                       If null, the parameter will not be set and 'default' will be used.
 *                       </ul>
 */
public class OntapVolumeSpecImpl implements OntapVolumeSpec {

    private final String volumeName;
    private final String aggregateName;
    private final Size size;
    private final SpaceReserve spaceReserve;
    private final SecurityStyle securityStyle;
    private final VolumeType volumeType;
    private final String securityPermissions;
    private final int userId;
    private final int groupId;
    private final String junctionPath;
    private final int vmAlignSector;
    private final String vmAlignSuffix;
    private final String snapshotPolicy;
    private final String qosPolicyGroupName;
    private final String exportPolicy;

    private OntapVolumeSpecImpl(Builder builder) {
        volumeName = builder.volumeName;
        securityPermissions = builder.securityPermissions;
        aggregateName = builder.aggregateName;
        size = builder.size;
        spaceReserve = builder.spaceReserve;
        securityStyle = builder.securityStyle;
        volumeType = builder.volumeType;
        userId = builder.userId;
        groupId = builder.groupId;
        junctionPath = builder.junctionPath;
        vmAlignSector = builder.vmAlignSector;
        vmAlignSuffix = builder.vmAlignSuffix;
        snapshotPolicy = builder.snapshotPolicy;
        qosPolicyGroupName = builder.qosPolicyGroupName;
        exportPolicy = builder.exportPolicy;
    }

    public String getVolumeName() {
        return volumeName;
    }

    public String getAggregateName() {
        return aggregateName;
    }

    public Size getSize() {
        return size;
    }

    public SpaceReserve getSpaceReserve() {
        return spaceReserve;
    }

    public SecurityStyle getSecurityStyle() {
        return securityStyle;
    }

    public String getSecurityPermissions() {
      return securityPermissions;
    }

    public int getUserId() {
      return userId;
    }

    public int getGroupId() {
      return groupId;
    }

    public VolumeType getVolumeType() {
      return volumeType;
    }

    public String getJunctionPath() {
        return junctionPath;
    }

    public int getVmAlignSector() {
        return vmAlignSector;
    }

    public String getVmAlignSuffix() {
        return vmAlignSuffix;
    }

    public String getSnapshotPolicy() {
        return snapshotPolicy;
    }

    public String getQosPolicyGroupName() {
        return qosPolicyGroupName;
    }

    public String getExportPolicy() {
      return exportPolicy;
    }

    @Override
  public String toString() {
        StringBuilder output = new StringBuilder();
        output.append("VolumeImpl: ").append(volumeName);
        output.append(" in aggr ").append(aggregateName);
        output.append(" with size ").append(size.toMB()).append(" MB ");
        output.append(" and space reserve ").append(spaceReserve);
        output.append(" and volume type ").append(volumeType);
        output.append(" and snapshot policy ").append(snapshotPolicy);
        output.append(" and qos group policy ").append(qosPolicyGroupName);
        output.append(" and security style ").append(securityStyle);
        output.append(" and security permissions ").append(securityPermissions);
        output.append(" and user ID ").append(userId);
        output.append(" and group ID ").append(groupId);
        output.append(" and export policy ").append(exportPolicy);
        output.append("(shimSize ").append(vmAlignSector).append(" sectors for files types: ").append(vmAlignSuffix).append(")");
        return output.toString();
    }

    public static class Builder implements OntapVolumeSpec.Builder {

        // required parameters
        private final String volumeName;
        private final String aggregateName;
        private final Size size;

        // optional parameters with reasonable defaults where needed
        private SpaceReserve spaceReserve = SpaceReserve.volume;
        private SecurityStyle securityStyle = SecurityStyle.unix;
        private VolumeType volumeType = VolumeType.rw;
        private String securityPermissions;
        // Default the user and group IDs to -1 so that the volume defaults will be used if they are not set.
        private int userId = -1;
        private int groupId = -1;
        private String junctionPath;
        private int vmAlignSector;
        private String vmAlignSuffix;
        private String snapshotPolicy;

        // Only for cDOT
        private String exportPolicy;

        // only valid for SN.0 or better running cDOT
        private String qosPolicyGroupName;

        /**
         * A volume cannot be specified without the following required bits of information:
         * @param volumeName name for the volume
         * @param aggregate the aggregate that will contain the volume
         * @param size how large it should be
         */
        public Builder(String volumeName, Aggregate aggregate, Size size) {
            this.volumeName = volumeName;
            this.aggregateName = aggregate.getName();
            this.size = size;
        }

        public Builder securityStyle(SecurityStyle style) {
            this.securityStyle = style;
            return this;
        }

        public Builder securityPermissions(String permissions) {
          this.securityPermissions = permissions;
          return this;
        }

        public Builder userId(int userId) {
          this.userId = userId;
          return this;
        }

        public Builder groupId(int groupId) {
          this.groupId = groupId;
          return this;
        }

        public Builder spaceReserve(SpaceReserve val) {
            this.spaceReserve = val;
            return this;
        }

        public Builder volumeType(VolumeType type) {
          this.volumeType = type;
          return this;
        }

        public Builder junctionPath(String val) {
            junctionPath = val;
            return this;
        }

        public Builder vmAlign(int sector, String suffix) {
            vmAlignSector = sector;
            vmAlignSuffix = suffix;
            return this;
        }

        public Builder snapshotPolicy(String val) {
            snapshotPolicy = val;
            return this;
        }

        public Builder qosPolicy(String policyName) {
            qosPolicyGroupName = policyName;
            return this;
        }

        public Builder exportPolicy(String exportPolicy) {
          this.exportPolicy = exportPolicy;
          return this;
        }

        public OntapVolumeSpecImpl build() {
            return new OntapVolumeSpecImpl(this);
        }
    }
}
