package com.netapp.offtap3.internal.ontap.zapi;

import java.io.BufferedOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.OutputStream;
import java.io.OutputStreamWriter;
import java.io.Reader;
import java.io.StringReader;
import java.io.StringWriter;
import java.io.Writer;
import java.net.HttpURLConnection;
import java.net.URL;
import java.net.URLConnection;
import java.security.KeyManagementException;
import java.security.NoSuchAlgorithmException;
import java.security.SecureRandom;
import java.security.cert.CertificateException;
import java.security.cert.X509Certificate;

import javax.net.ssl.HostnameVerifier;
import javax.net.ssl.HttpsURLConnection;
import javax.net.ssl.SSLContext;
import javax.net.ssl.SSLSession;
import javax.net.ssl.SSLSocketFactory;
import javax.net.ssl.X509TrustManager;
import javax.xml.bind.DatatypeConverter;

import org.xml.sax.SAXException;

import com.netapp.offtap3.internal.ontap.MCCManager;
import com.netapp.offtap3.internal.ontap.utils.StatsCollector;
import com.netapp.offtap3.internal.utils.FilteredSslSocketFactory;
import com.netapp.offtap3.ontap.ApiFailedException;
import com.netapp.offtap3.ontap.NaAuthenticationException;
import com.netapp.offtap3.ontap.NaElement;
import com.netapp.offtap3.ontap.NaProtocolException;

/**
 * Class that does raw http invocations of ZAPIs to increase performance.  This is based on some converations with
 * Joe Sunday and some micro-benchmarking dsarisky performed with AutoZAPI and manage-ontap.  I saw differences
 * of 300 milliseconds for a aggr-get-iter with a single aggr with manageontap vs. 30 milliseconds with this
 * implementation and with AutoZAPI.
 * <p/>
 * Per conversation with Brian Mcgiverin, lets try to do some raw HTTP and see how well that performs instead of
 * trying to take the runtime stack out of AutoZAPI and use it.  There's just too many build dependencies and
 * tentacles in AutoZAPI to deal with - I just want a tentacle - not an entire octopus.
 * <p/>
 * Some code borrowed from AutoZAPI project.
 * Some code borrowed from ManageOntap project.
 */
public class RawHttpZAPIInvoker {
    static final HostnameVerifier ACCEPT_ALL_HOSTNAME_VERIFIER = new HostnameVerifier() {
        @Override
        public boolean verify(String hostname, SSLSession session) {
            return true;
        }
    };
    public static Long nextHttpCodeOverride = null;
    public static String nextHttpResponseOverride = null;
    public static boolean nextHttpConnectionResetOnRead = false;
    public static boolean nextHttpConnectionResetOnWrite = false;
    private static ThreadLocal<NaElement> lastTraceOutput = new ThreadLocal<NaElement>();
    private static SSLSocketFactory socketFactoryWithTls;
    private static SSLSocketFactory socketFactoryWithoutTls;
    private static boolean wasSslv3Enabled = false;
    
    /**
     * For testability.
     * 
     * @return SSLSocketFactory
     */
    public SSLSocketFactory getSocketFactoryWithTls() {
        return socketFactoryWithTls;
    }
    
    /**
     * For testability.
     * 
     * @return SSLSocketFactory
     */
    public SSLSocketFactory getSocketFactoryWithoutTls() {
        return socketFactoryWithoutTls;
    }

    private static NaElement processIntoNaElement(String body) throws NaProtocolException, ApiFailedException {
        // this method was totally taken from manage-ontap library

        if (nextHttpResponseOverride != null) {
            body = nextHttpResponseOverride;
            nextHttpResponseOverride = null;
        }

        if (0 == body.length()) {
            throw new NaProtocolException("body of server response empty");
        }

		/*
         * The parser doesn't like the <!DOCTYPE> entry which
		 * contains a relative URI.  I wasn't able to figure out how
		 * to tell the parser what our parent URL/hostname/whatever was,
		 * so that it could handle the relative URI,
		 * so to finesse this, we just remove the <!DOCTYPE> tag entirely.
		 *
		 * It's also possible that the <!DOCTYPE> doesn't appear at all,
		 * in which case we've got a fairly degenerate situation (like
		 * a 404 Not Found response)
		 */
        int pos1 = body.indexOf("<!DOCTYPE");
        int pos2 = body.indexOf("<netapp");

		/*
         * Maybe we failed to find a <!DOCTYPE>
		 *
		 * This could be an authentication error.  It could also be that
		 * the server doesn't have a clue (404 Not Found), which might be
		 * the case if the server is listening but is not a ZAPI server.
		 *
		 * We err on the side of hinting to the client to try again.
		 */
        if (pos1 < 0) {
            throw new NaProtocolException("server response did not " + "contain <!DOCTYPE>");
        }

		/*
         * Maybe we just failed to find <netapp>
		 */
        if (pos2 < 0) {
            /*
             * We found a <!DOCTYPE>, but there's no netapp directive.
			 * We'll try to return whatever follows the <!DOCTYPE>
			 */
            pos2 = body.indexOf("<", pos1);

            if (pos2 < 0) {
                throw new NaProtocolException("Output from server doesn't " + "contain <netapp> element.");
            }

            throw new NaProtocolException("Unexpected tag found: " +
                                                  body.substring(0, pos1) + body.substring(pos2));
        }

		/*
         * strip <!DOCTYPE> from body
		 */
        body = body.substring(0, pos1) + body.substring(pos2);

        NaElement xo;

        try {
            xo = NaElementFactory.fromString(body);
        } catch (IOException e) {
            throw new RuntimeException("I/O error during XML parsing", e);
        } catch (SAXException e) {
            throw new NaProtocolException("Malformed XML", e);
        }


		/*
		 * We expect a <netapp> element
		 */
        if (null == xo || !xo.getName().equals("netapp")) {
            throw new NaProtocolException("Output from server doesn't " + "contain <netapp> element.");
        }

        NaElement results = xo.getChildByName("results");
        NaElement localNaElement3 = xo.getChildByName("trace-output");
        lastTraceOutput.set(localNaElement3);

		/*
		 *  If the API succeeded, return the results now
		 */
        if ("passed".equals(results.getAttr("status"))) {
            return results;
        }

		/*
		 *  Otherwise, something went wrong, and we need to build and
		 *  throw an NaAPIFailedException.
		 */
        int errno;
        String s = results.getAttr("errno");
        try {
            errno = Integer.parseInt(s);
        } catch (NumberFormatException e) {
            throw new NaProtocolException("Malformed errno code: " + s, e);
        }
        throw new ApiFailedException(errno, results.getAttr("reason"));
    }

    public NaElement invokeZapi(String username, String password, String hostname, String port,
                                       String vServerName, NaElement xi, int zapiConnectionTimeoutInSeconds, int zapiTimeoutInSeconds, boolean ssl,
                                       boolean wantsTraceOutput,
                                       boolean useTls,
                                       boolean useSslv3) throws IOException, NaAuthenticationException, NaProtocolException, ApiFailedException {
        long startTime = System.currentTimeMillis();
        String rawOutput = invokeZapiViaRawHttp(username, password, hostname, port, vServerName, xi,
                                                zapiConnectionTimeoutInSeconds, zapiTimeoutInSeconds, ssl, wantsTraceOutput, useTls, useSslv3);

        long now = System.currentTimeMillis();
        StatsCollector.getInstance().addCounter("raw-invocation-network", now - startTime);

        NaElement retVal = processIntoNaElement(rawOutput);
        long now2 = System.currentTimeMillis();
        StatsCollector.getInstance().addCounter("raw-invocation-xml", now2 - now);

        return retVal;
    }

    public static NaElement getLastTraceOutput() {
        return lastTraceOutput.get();
    }

    private static SSLSocketFactory makeSSLFactory(boolean useTls, boolean useSslv3) {
        /*
         * If we changed the useSslv3 setting, we cannot use a cached SSLSocketFactory
         * unless it had the same setting.
         */
        if (useSslv3 != wasSslv3Enabled) {
            socketFactoryWithTls = null;
            socketFactoryWithoutTls = null;
        }
        wasSslv3Enabled = useSslv3;
        
        SSLSocketFactory factory = useTls ? socketFactoryWithTls : socketFactoryWithoutTls;
        // method from AutoZAPI
        if (factory == null) {
            X509TrustManager[] tm = {new X509TrustManager() {
                public void checkClientTrusted(X509Certificate[] chain, String authType) throws CertificateException {
                }

                public void checkServerTrusted(X509Certificate[] chain, String authType) throws CertificateException {
                }

                public X509Certificate[] getAcceptedIssuers() {
                    return new X509Certificate[0];
                }
            }};
            try {
                SSLContext ctx = SSLContext.getInstance("SSL");
                ctx.init(null, tm, new SecureRandom());
                factory = new FilteredSslSocketFactory(ctx.getSocketFactory(), useTls, useSslv3);
                if (useTls) {
                    socketFactoryWithTls = factory;
                } else {
                    socketFactoryWithoutTls = factory;
                }
            } catch (NoSuchAlgorithmException e) {
                return null;
            } catch (KeyManagementException e) {
                return null;
            }
        }
        return factory;
    }

    private static String invokeZapiViaRawHttp(String username, String password, String hostname, String port,
                                               String vServerName, NaElement xi, int zapiConnectionTimeout, int zapiTimeout, boolean ssl,
                                               boolean wantsTraceOutput,
                                               boolean useTls,
                                               boolean useSslv3) throws IOException, NaAuthenticationException, NaProtocolException {

        lastTraceOutput.set(null);
        if (password == null) {
            password = "";
        }
        String protocol = "http://";
        if (ssl) {
            protocol = "https://";
        }

        URLConnection connection = new URL(
                protocol + hostname + ":" + port + "/servlets/netapp.servlets.admin.XMLrequest_filer").openConnection();
        if (connection instanceof HttpURLConnection) {
            ((HttpURLConnection) connection).setRequestMethod("POST");
        }
        if (connection instanceof HttpsURLConnection) {
            ((HttpsURLConnection) connection).setRequestMethod("POST");
            ((HttpsURLConnection) connection).setSSLSocketFactory(makeSSLFactory(useTls, useSslv3));
            ((HttpsURLConnection) connection).setHostnameVerifier(ACCEPT_ALL_HOSTNAME_VERIFIER);
        }
        connection.setDoOutput(true);
        connection.setRequestProperty("Connection", "keep-alive");
        connection.setRequestProperty("Content-Type", "text/xml");
        connection.setConnectTimeout(zapiConnectionTimeout * 1000);
        connection.setReadTimeout(zapiTimeout * 1000);
        if (wantsTraceOutput) {
            connection.setRequestProperty("X-Trace-Threshold", "0");
        }

        final String authString = username + ":" + password;

        String header = "Basic " + DatatypeConverter.printBase64Binary((authString.getBytes("UTF-8")));
        connection.setRequestProperty("Authorization", header);
        int responseCode = 0;
        try {
	        MCCManager.addConnectionForThread(connection);
	        OutputStream requestStream = new BufferedOutputStream(connection.getOutputStream());
	        StringBuilder outgoing = new StringBuilder(
	                "<?xml version='1.0' encoding='UTF-8'?><!DOCTYPE netapp SYSTEM 'file:/etc/netapp_filer.dtd'><netapp xmlns=\"http://www.netapp.com/filer/admin\" ");
	        if (vServerName != null) {
	            outgoing.append(" vfiler=\"").append(vServerName).append("\" ");
	        }
	
	        outgoing.append("version=\"1.7\">");
	        outgoing.append(xi.toString());
	        outgoing.append("</netapp>");
	
	        OutputStreamWriter osw = new OutputStreamWriter(requestStream);
	        if (nextHttpConnectionResetOnWrite) {
	            nextHttpConnectionResetOnWrite = false;
	            throw new IOException("Connection reset by peer");
	        }
	        copy(new StringReader(outgoing.toString()), osw);
	        osw.flush();
	        connection.connect();	
	        
	        if (connection instanceof HttpURLConnection) {
	            responseCode = ((HttpURLConnection) connection).getResponseCode();
	        }
	        if (connection instanceof HttpsURLConnection) {
	            responseCode = ((HttpsURLConnection) connection).getResponseCode();
	        }
        } finally {
        	MCCManager.removeConnectionForThread();
        }
        if (nextHttpCodeOverride != null) {
            responseCode = nextHttpCodeOverride.intValue();
            nextHttpCodeOverride = null;
        }

        if (401 == responseCode) {
            throw new NaAuthenticationException("Authorization failed");
        } else if (200 != responseCode) {
            String goodMessage = "";
            try {
                InputStream output = connection.getInputStream();
                goodMessage = toString(output);
            } catch (Exception e) {
                // no problem - just won't have a goodMessage
            }
            String message = "Unexpected HTTP response: " + responseCode;
            if (goodMessage != null) {
                message += " specifically " + goodMessage;
            }
            throw new NaProtocolException(message);
        }
        if (nextHttpConnectionResetOnRead) {
            nextHttpConnectionResetOnRead = false;
            throw new IOException("Connection reset by peer");
        }

        InputStream output = connection.getInputStream();
        return toString(output);

        // I realize we're not obviously closing HTTP connections or input streams here
        // after some reading on stackoverflow - the general consensus is that so long as you read the response -
        // that is all that's necessary to make Http keepalives work and avoid any resource leaking.
        // Extensive vvol unit testing hasn't shown any obvious resource leak problems so far.
    }

    private static String toString(InputStream is) throws IOException {
        StringWriter output = new StringWriter();
        InputStreamReader isr = new InputStreamReader(is);
        copy(isr, output);
        return output.getBuffer().toString();
    }

    private static void copy(Reader inputStream, Writer output) throws IOException {
        char[] buffer = new char[4096];
        int n;
        while (-1 != (n = inputStream.read(buffer))) {
            output.write(buffer, 0, n);
        }
    }
}
