package com.netapp.offtap3.internal.ontap.zapi;

import com.netapp.offtap3.ontap.ApiFailedException;
import com.netapp.offtap3.Controller;
import com.netapp.offtap3.OfftapPreferences;
import com.netapp.offtap3.factories.ControllerFactory;
import com.netapp.offtap3.internal.ontap.MCCManager;
import com.netapp.offtap3.internal.ontap.MCCSlapperThread;
import com.netapp.offtap3.internal.ontap.controller.OntapConnection;
import com.netapp.offtap3.internal.ontap.support.TlsRetryException;
import com.netapp.offtap3.internal.ontap.support.ZAPIStatsCounter;
import com.netapp.offtap3.internal.ontap.support.ZapiCache;
import com.netapp.offtap3.internal.ontap.utils.StatsCollector;
import com.netapp.offtap3.ontap.*;
import com.netapp.offtap3.utils.*;
import org.apache.log4j.Logger;

import javax.net.ssl.SSLException;
import java.io.File;
import java.io.IOException;
import java.net.ConnectException;
import java.util.*;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.atomic.AtomicInteger;

public class ZAPIInvoker {
    private static final Logger ZAPI_LOGGER = Logger.getLogger("ZAPI_LOGGER");
    private static final Logger ZAPI_ERROR_LOGGER = Logger.getLogger("ZAPI_ERROR_LOGGER");
    private static final Logger LOGGER = Logger.getLogger(ZAPIInvoker.class);
    private static final ConcurrentHashMap<String, FlightDetailsForController> inFlightCounts = new SafeConcurrentHashMap<String, FlightDetailsForController>();
    private static final WhatsHappeningMonitor IN_FLIGHT_DETAILS = WhatsHappeningMonitor.getMonitor("zapi-in-flight");

    // I'm not sure if we really need to keep the string instances to lock on, or whether JVM will ensure that we only
    // have a single instance in memory for a given string.  This is easy enough to do, so we'll just do it to be sure.
    private static final ConcurrentHashMap<String, String> TRACE_OUTPUT_LOCK_KEYS = new SafeConcurrentHashMap<String, String>();

    /**
     * Set of strings that are used to see if a ZAPI failure is recoverable and we should wait and retry after a second
     */
    private static final List<String> RECOVERABLE_ZAPI_FAILURE_MESSAGES = Arrays
            .asList("Authorization failed", "errno=51", "memory allocation failed", "Specified Vserver not found",
                    "errno=15698", "internal network error", "errno=14450", "errno=13158", "errno=13001");

    // 15661 is a snapmirror get when there's no snap mirror relationship
    private static final List<String> SILENT_DISCARD_ERROR_MESSAGES = Arrays.asList("15661","17130");

    private static final Set<String> SILENT_DISCARD_FAKE_PASSED_ZAPI_FRAGMENTS = Utils.toSet("snapmirror");    
    private static final Set<String> SILENT_DISCARD_EXPECTED_ERROR_DO_NOT_LOG = Utils.toSet("nfs-service-get", "sis-get","snapmirror-abort");

    private static final AtomicInteger IN_FLIGHT_ALL_CONTROLLERS_COUNT = new AtomicInteger();
    private static final AtomicInteger MY_ZAPI_COUNTER_STATIC = new AtomicInteger();
    private static boolean checkedTraceDirectory = false;
    private static ZAPIInvoker lastInstance = null;
    private ZapiCache zapiCache = ZapiCache.getInstance();
    private Priority defaultPriority;
    private RawHttpZAPIInvoker rawZapiInvoker = new RawHttpZAPIInvoker();
    private Map<String, Integer> nodeCountsByUuid = new SafeConcurrentHashMap<String, Integer>();
    private Map<String, String> uuidMapping = new SafeConcurrentHashMap<String, String>();

    public ZAPIInvoker() {
        // disable the max allowable size for this monitor, let it go as high as anyone wants.
        IN_FLIGHT_DETAILS.setMaxAllowableSize(10000);
        defaultPriority = Priority.MEDIUM;
        lastInstance = this;
    }

    public static ZAPIInvoker getLastInstance() {
        return lastInstance;
    }
    
    public void addNodeCountByUuid(String uuid, Integer nodeCount) {
    	nodeCountsByUuid.put(uuid, nodeCount);
    }
    
    public void addUuidMapping(String childUuid, String adminVsUuid) {
    	uuidMapping.put(childUuid, adminVsUuid);
    }
    
    public static int getInFlightCount() {
    	return IN_FLIGHT_ALL_CONTROLLERS_COUNT.get();
    }

    private Priority getPriority(OntapConnection connection, String caller, String zapi) {
        Priority retVal = defaultPriority;
        if (caller == null) {
            caller = "";
        }
        if (zapi == null) {
            zapi = "";
        }
        if (caller.contains(",")) {
            // ok - got something that's being called on behalf of a couple different people.  Lets run through and see if we found a high, medium, or
            // low priority and then take the highest we found
            Priority foundPriority = defaultPriority;
            for (String delimitedCaller : StringUtils.splitIntoList(caller, ",")) {
                Priority curr = getPriorityForCallerAndZapi(connection, delimitedCaller, zapi, retVal);
                if (curr.compareTo(foundPriority) > 0) {
                    foundPriority = curr;
                }
            }
            return foundPriority;
        } else {
            return getPriorityForCallerAndZapi(connection, caller, zapi, retVal);
        }
    }

    Priority getPriorityForCallerAndZapi(OntapConnection connection, String caller, String zapi, Priority retVal) {
        caller = caller.replaceAll("\\[", "");
        caller = caller.replaceAll("\\]", "");
        if ((StringUtils.isReal(caller) && connection.getOfftapPreferences().getHighPriorityZapiRequestors().contains(caller)) ||
                (StringUtils.isReal(zapi) && connection.getOfftapPreferences().getHighPriorityZapiRequestors().contains(zapi))) {
            return Priority.HIGH;
        } else if ((StringUtils.isReal(caller) && connection.getOfftapPreferences().getLowPriorityZapiRequestors().contains(caller)) ||
                (StringUtils.isReal(zapi) && connection.getOfftapPreferences().getLowPriorityZapiRequestors().contains(zapi))) {
            return Priority.LOW;
        } else if ((StringUtils.isReal(caller) && connection.getOfftapPreferences().getMediumPriorityZapiRequestors().contains(caller)) ||
                (StringUtils.isReal(zapi) && connection.getOfftapPreferences().getMediumPriorityZapiRequestors().contains(zapi))) {
            return Priority.MEDIUM;
        } else {
            return retVal;
        }
    }

    public NaElement invokeFilerBasedZapi(OntapConnection connection, NaElement filerBasedZapi, long cacheDuration, long startTime) throws ApiFailedException {
        long actualInFlightDuration = 0;
        int myZapiCounter = MY_ZAPI_COUNTER_STATIC.incrementAndGet();
        int myDepth = -1;
        int allControllersTotalDepth = 0;
        long throttleDuration = 0;
        Priority myPriority = null;
        String myInFlightKey = null;
        boolean succeeded = false;
        NaElement results = null;

        if (connection.getOfftapPreferences().getWantsZapiValidation()) {
        	connection.getOfftapPreferences().getZapiValidator().isValid(filerBasedZapi);
        }


        try {

            // straight filer based ZAPI - lets try to get from cache
            ZAPI_LOGGER.debug("Sending-" + myZapiCounter + ": " + filerBasedZapi);
            if (zapiCache != null) {
                Object fromCache = zapiCache
                        .getValueFromCache(connection.getContext(), connection.getIpAddress(), filerBasedZapi,
                                           cacheDuration);
                if (fromCache != null) {
                    checkForTimeout(connection);
                    if (fromCache instanceof NaElement) {
                        ZAPI_LOGGER.debug("Received-" + myZapiCounter + " (Cache): " + fromCache);
                        if (connection.getOfftapPreferences().getZapiSleepTime() > 0) {
                            StatsCollector.sleep(connection.getOfftapPreferences().getZapiSleepTime(), "zapi-prefs-sleep-time");
                        }
                        return (NaElement) fromCache;
                    } else if (fromCache instanceof Pair) {
                        // we cache error messages under certain circumstances - ones associated with normal discovery - so that integration
                        // test environments which run with caching can quickly discover filers and not get gobs of errors.
                        ZAPI_LOGGER.debug("Received-" + myZapiCounter + " (Cached-Error): " + fromCache);
                        Pair<Integer, String> asPair = (Pair) fromCache;
                        throw new ApiFailedException(asPair.getA(), asPair.getB());
                    }
                }
            }
            myInFlightKey = ZAPIStatsCounter.getZapiPerformanceKey("" + filerBasedZapi);
            String myThreadLocalVal = null;
            if (!ZAPIStatsCounter.getInstance().getRecordedKey().isEmpty()) {
                myThreadLocalVal = "" + ZAPIStatsCounter.getInstance().getRecordedKey();
            }
            myPriority = getPriority(connection, myThreadLocalVal, myInFlightKey);
            String inFlightKey = connection.getUniqueKey();
            if (StringUtils.isReal(myThreadLocalVal)) {
                myInFlightKey += " (" + myThreadLocalVal + ":" + myPriority + ")";
            }
            IN_FLIGHT_DETAILS.registerStartCall(inFlightKey, myInFlightKey);
            String inFlightCountsKey = connection.getIpAddress();
            if (inFlightCountsKey == null) {
                inFlightCountsKey = "";
            }
            FlightDetailsForController flightDetailsForThisController = new FlightDetailsForController();
		    
		    boolean found = false;		    
		    Integer nodeCountForUuid = 0;
		    String controllerId = "";
		    if (StringUtils.isReal(connection.getFilerIdOfTarget())) {
		    	controllerId = connection.getFilerIdOfTarget();
		    	nodeCountForUuid = this.nodeCountsByUuid.get(controllerId);
		    	if (nodeCountForUuid == null) {
		    		nodeCountForUuid = new Integer(0);
		    	}
		    }
			if (connection.getOfftapPreferences().getNewCmodeZapiThrottle()) {
				//C-mode Zapi throttling needs accounting at the cluster level
				String adminVserverId = uuidMapping.get(controllerId);
				if (adminVserverId == null) {
					adminVserverId = "unknown";
				}
			    found = true;
			    flightDetailsForThisController = inFlightCounts.get(adminVserverId);
			    if (flightDetailsForThisController == null) {
			    	flightDetailsForThisController = new FlightDetailsForController();
			    	inFlightCounts.put(adminVserverId, flightDetailsForThisController);		 		    	
			    }
			} 
			
			if (!found) {
				// direct connect vserver or 7 mode filer
				flightDetailsForThisController = inFlightCounts.get(inFlightCountsKey);
				if (flightDetailsForThisController == null) {
					flightDetailsForThisController = new FlightDetailsForController();
					inFlightCounts.put(inFlightCountsKey, flightDetailsForThisController);
				}
			}

            boolean decrementAllControllersCount = false;
            ApiFailedException toCacheError = null;
            try {
                long throttleEnterTime = System.currentTimeMillis();
                while (true) {
                    String originalInFlightKey = myInFlightKey;
                    // ok - now we have the flight information for this guy.  We need to see whether we can launch, or whether we need to wait.
                    boolean launch = true;
                    synchronized (flightDetailsForThisController) {
                    	myDepth = flightDetailsForThisController.inFlightCount.get();				  
				    	if(connection.getOfftapPreferences().getNewCmodeZapiThrottle()){
				    		/*tdeepak: new c-mode throttling based on burt: 830923
				    		  Details : https://wikid.netapp.com/w/User:Tdeepak/Cmode_Zapi_Throttling#Retry_mechanism_in_Aeonflux*/
				    		launch = !shouldNewCmodeThrottle(connection, nodeCountForUuid, myDepth);
				    	} else if (myDepth >= connection.getOfftapPreferences().getZapiThrottleSize()) {
				    		launch = false;
				    	}

                        // lets see if we've timed out or died
                        checkForTimeout(connection);

                        if (launch) {
                            if (originalInFlightKey != null) {
                                IN_FLIGHT_DETAILS.updateName(originalInFlightKey);
                            }
                            flightDetailsForThisController.launchAtPriority(myPriority);
                            break;
                        }
                    }
                    if (!launch) {
                        if (originalInFlightKey != null) {
                            IN_FLIGHT_DETAILS.updateName(originalInFlightKey + " (WAITING TO LAUNCH)");
                        }
                        flightDetailsForThisController.waitAtPriority(myPriority);
                    }
                }
                allControllersTotalDepth = IN_FLIGHT_ALL_CONTROLLERS_COUNT.incrementAndGet();
                decrementAllControllersCount = true;
                throttleDuration = System.currentTimeMillis() - throttleEnterTime;
                try {
                    long startTime2 = System.currentTimeMillis();
                    results = invokeZapi(connection, filerBasedZapi);
                    succeeded = true;
                    actualInFlightDuration = System.currentTimeMillis() - startTime2;
                    String toPrint = results.toString();
                    if (toPrint.length() > 10000) {
                        toPrint = toPrint.substring(0, 10000);
                    }
                    ZAPI_LOGGER.debug("Received-" + myZapiCounter + ": " + toPrint);
                    if (connection.getOfftapPreferences().getZapiSleepTime() > 0) {
                        StatsCollector.sleep(connection.getOfftapPreferences().getZapiSleepTime(), "zapi-prefs-sleep-time");
                    }
                } catch (NaException nae) {
                    // I suppose this is double logging, and that might hack someone off - but I want to clearly and without
                    // a doubt know which ZAPIs generated an error and be able to act on them.
                    String identifyingString = connection.getUniqueKey();
                    if (zapiCache != null && nae instanceof ApiFailedException && cacheDuration > 0) {
                        toCacheError = (ApiFailedException) nae;
                    }
                    boolean logError = true;
                    for (String curr : SILENT_DISCARD_ERROR_MESSAGES) {
                        if (nae.toString().contains(curr)) {
                            for (String currZapi : SILENT_DISCARD_FAKE_PASSED_ZAPI_FRAGMENTS) {
                                if (filerBasedZapi.toString().contains(currZapi)) {
                                    ZAPI_LOGGER.error("Returning empty XML for silent discard of " + nae);
                                    results = NaElementFactory.fromString("<results status=\"passed\"/>");
                                }
                            }
                            for (String currZapi : SILENT_DISCARD_EXPECTED_ERROR_DO_NOT_LOG) {
                            	if (filerBasedZapi.toString().contains(currZapi)) {
                            		logError = false;
                            	}
                            }

                        }
                    }
                    if (results == null) {
                    	if (logError) {
	                        ZAPI_LOGGER.error("Sent " + myZapiCounter + " (and received error) to " + identifyingString +
	                                                  filerBasedZapi);
	                        ZAPI_LOGGER.error("NaException " + nae);
	                        ZAPI_ERROR_LOGGER
	                                .error("Sent " + myZapiCounter + " (and received error) to " + identifyingString +
	                                               filerBasedZapi);
	                        ZAPI_ERROR_LOGGER.error("NaException " + nae);
                    	}
                        results = retryIfApplicable(connection, filerBasedZapi, nae);
                    } else {
                        throw nae;
                    }
                }
            } finally {
                synchronized (flightDetailsForThisController) {
                    flightDetailsForThisController.finishAtPriority(myPriority);
                }
                if (decrementAllControllersCount) {
                    IN_FLIGHT_ALL_CONTROLLERS_COUNT.decrementAndGet();
                }
                IN_FLIGHT_DETAILS.registerEndCall();
                if (toCacheError != null) {
                    zapiCache.addZapiToCache(connection.getContext(), connection.getIpAddress(), filerBasedZapi,
                                             new Pair(toCacheError.getErrno(), toCacheError.getMessage()));
                }
            }


            if (zapiCache != null) {
                zapiCache.addZapiToCache(connection.getContext(), connection.getIpAddress(), filerBasedZapi, results);
            }
        } catch (NaAuthenticationException e) {
            LOGGER.warn("invokeZAPI(" + myDepth + "): failed " + connection + " FAILED authentication. " +
                                Utils.getCaller(e));
            recordErrorDetails(connection, filerBasedZapi, e);
            String message = e.getMessage();
            throw new ApiFailedException(NaErrno.EAPIAUTHENTICATION, message, e);

        } catch (NaProtocolException e) {
            try {
                if (e.getMessage() != null && (e.getMessage().contains("500") || e.getMessage().contains("502"))) {
                    results = retryIfApplicable(connection, filerBasedZapi, new ApiFailedException(51, e.getMessage(), e));
                    if (results != null) {
                        return results;
                    }
                }
            } catch (Exception e2) {
                // no problem - just keep going.
            }
            LOGGER.warn(
                    "invokeZAPI(" + myDepth + "): failed " + connection + " protocol exception : " + e.getMessage() +
                            " " + Utils.getCaller(e));
            recordErrorDetails(connection, filerBasedZapi, e);
            throw new ApiFailedException(NaErrno.EAPIERROR, e.getMessage(), e);

        } catch (ConnectException e) {
        	try {
	        	results = retryIfApplicable(connection, filerBasedZapi, new ApiFailedException(777, e.getMessage()));
	            if (results != null) {
	                return results;
	            }
        	} catch (Exception e2) {
        		// no problem - keep going
        	}
            String errorMessage = e.getMessage();
            LOGGER.warn("invokeZAPI(" + myDepth + "): failed " + connection + " using " +
                                (connection.isSSL() ? "HTTPS" : "HTTP") + ": " + errorMessage + " " +
                                Utils.getCaller(e));
            recordErrorDetails(connection, filerBasedZapi, e);
            throw new ApiFailedException(NaErrno.EAPI_CONNECTION, e.getMessage(), e);

        } catch (IOException e) {
            String msg = e.getMessage();
            if (connection.useTls() && (e instanceof SSLException) && (msg != null) &&
                    (msg.contains("bad_record_mac"))) {
                connection.setTls(false);
                return invokeFilerBasedZapi(connection, filerBasedZapi, cacheDuration, startTime);
            } else {
                try {
                    results = retryIfApplicable(connection, filerBasedZapi, new ApiFailedException(51, e.getMessage()));
                    if (results != null) {
                        return results;
                    }
                } catch (Exception e2) {
                    // no problem - just keep going.
                }
            }
            LOGGER.warn(
                    "invokeZAPI(" + myDepth + "): failed " + connection + " (IOException): " + e.getMessage() + " " +
                            Utils.getCaller(e));
            recordErrorDetails(connection, filerBasedZapi, e);
            throw new ApiFailedException(NaErrno.EONTAPI_EIO, e.getMessage(), e);

        } catch (ApiFailedException e) {
            //handle some common issues, log error and then pass along the bad news.
            LOGGER.warn("invokeZAPI(" + myDepth + "): failed " + connection + " API failed. " + e.getMessage() + " " +
                                Utils.getCaller(e));
            recordErrorDetails(connection, filerBasedZapi, e);
            throw e;

        } catch (Exception e) {
        	if (String.valueOf(e).contains("Interrupted")) {
				try {
					results = retryIfApplicable(connection, filerBasedZapi,
							new ApiFailedException(51, "Interrupted"));
					if (results != null) {
						return results;
					}
				} catch (Exception e2) {
					// no problem - just keep going.
				}
			}
            LOGGER.warn(
                    "invokeZAPI(" + myDepth + "): failed " + connection + " General Exception " + e.getMessage() + " " +
                            Utils.getCaller(e));
            recordErrorDetails(connection, filerBasedZapi, e);
            throw new ApiFailedException(NaErrno.EAPIERROR, e.getMessage(), e);
        } finally {
            if (zapiCache != null) {
                long duration = System.currentTimeMillis() - startTime;
                ZAPIStatsCounter.getInstance().addZAPICall(filerBasedZapi, (duration));
                if (actualInFlightDuration > 0) {
                    StatsCollector.getInstance().addCounter("zapi-overhead", duration - actualInFlightDuration);
                }
                if (duration > 10000) {
                	String slowZapiKey = connection.getIpAddress();
                	if (StringUtils.isReal(connection.getContext())) {
                		slowZapiKey += "(" + connection.getContext() + ")";
                	}
                	
                    ZAPI_LOGGER.debug("slow-zapi-" + myZapiCounter + " " + slowZapiKey + ": " + duration + " ms");
                }
                if (myDepth > -1) {

                    StatsCollector.getInstance().addCounter("depth-" + myDepth);
                    String throttleKey = "zapi-throttle-wait-" + myPriority;
                    StatsCollector.getInstance().addCounter(throttleKey, throttleDuration);
                    if (myDepth != allControllersTotalDepth) {
                        StatsCollector.getInstance().addCounter("all-controllers-depth-" + allControllersTotalDepth);
                    }
                }
            }
            if (succeeded) {
                writeTraceOutputIfDesired(connection, filerBasedZapi, myInFlightKey);
            }
        }
        return results;
    }

    private NaElement retryIfApplicable(OntapConnection connection, NaElement filerBasedZapi, NaException nae) throws InterruptedException, NaException {
        NaElement results = null;
        // Always retry if the call failed because TSL was enabled for SSL. TLS should be now disabled and hopefully the call will work.
        if (connection.getOfftapPreferences().getWantsRetryOnRecoverableZapiFailure() || nae instanceof TlsRetryException) {        	
			if (!connection.isAllowRetry()) {
				throw nae;
			}			
			// one special case we want to fool with - mcc may change the vserver names.
			if (String.valueOf(nae).contains("15698") && connection.getFilerIdOfTarget() != null) {
				Controller controller = ControllerFactory.getInstance().getController(connection.getFilerIdOfTarget());
				MCCManager.getInstance().synchronouslyCheckForChangedVServerName(connection);	
			}
			
            String message = nae.toString();
            // having trouble getting a match here
            if (nae.getCause() != null && nae.getCause().getMessage() != null) {
                message += nae.getCause() + "\n" + nae.getMessage();
            }            
            boolean isRecoverable = false;
            for (String currRecoverableFragment : RECOVERABLE_ZAPI_FAILURE_MESSAGES) {
                if (message.contains(currRecoverableFragment)) {
                    isRecoverable = true;
                }
            }
            boolean wasVolumeBusy = false;
            boolean mayHaveBeenTimeout = false;
            if (nae instanceof ApiFailedException) {
                if (((ApiFailedException) nae).getErrno() == 51) {
                    isRecoverable = true;
                    mayHaveBeenTimeout = true;
                }
                if (((ApiFailedException) nae).getErrno() == 13158) {
                	wasVolumeBusy = true;
                }                
            }
            
            if (mayHaveBeenTimeout) {
            	hangOutWhileMccRuns(connection);
            }
            
            // lets see if this connection just did an mcc failover.
            OntapConnection connection2 = MCCManager.getInstance().possiblyReviseHelpfulServerForMCC(connection);
            if (connection2 != null && !(Utils.areEqual(connection2.getContext(), connection.getContext()) && (Utils.areEqual(connection2.getIpAddress(), connection.getIpAddress())))) {
            	isRecoverable = true;
            }
            LOGGER.debug("retryIfApplicable: message=" + message + " nae=" + nae.toString() + " and isRecoverable=" + isRecoverable);
            if (isRecoverable) {
		    	// normally we sleep a second.  But I think that volume busy is liable to need longer than that, so lets give it some extra room to breathe if that's what we got.		    	
		    	if (wasVolumeBusy) {
		    		StatsCollector.sleep(7000, "zapi-retry-volume-busy");
		    	} else {
		    		StatsCollector.sleep(1000, "zapi-retry-normal");
		    	}
		    	
                try {
                    LOGGER.debug("retryIfApplicable: Going to retry Zapi: " + filerBasedZapi.toString());                    
                    StatsCollector.getInstance().addCounter("recoverable-zapi-error-attempt");
                    results = invokeZapi(connection, filerBasedZapi);
                    StatsCollector.getInstance().addCounter("recoverable-zapi-error-success");
                } catch (Exception e) {
                    StatsCollector.getInstance().addCounter("recoverable-zapi-error-failure");
                    throw nae;
                }
            } else {
                throw nae;
            }
        }
        return results;    
    }

	private void hangOutWhileMccRuns(OntapConnection connection) {
		MCCSlapperThread slapper = MCCManager.getInstance().getMCCSlapperThreadCanBeNull(connection);
		if (slapper != null) {
			long startTime = System.currentTimeMillis();
			try {
				// It should never be 30 seconds, but just in case, lets put this timer on it so that it doesn't time out
				synchronized(slapper) {
					slapper.wait(30000);
				}
			} catch (Throwable t) {
				// no problem
			}
			StatsCollector.getInstance().addCounter("lock-sleep-mcc-slap", System.currentTimeMillis() - startTime);
		}
	}
    
	private boolean shouldNewCmodeThrottle(OntapConnection connection, Integer nodeCountForUuid, int myDepth) {
		boolean toReturn = false;
		int totalDepthForCluster = myDepth;
		Integer totalNodes = nodeCountForUuid;
		
		if (totalNodes == null || totalNodes == 0) {
			//partial discovery case
			totalNodes = 1;
		}
		
		double acceptableDepth = ((double) totalNodes * (double) connection.getOfftapPreferences().getZapiThrottleSize() * connection.getOfftapPreferences().getZapiThrottleFactor());

		if (totalDepthForCluster >= acceptableDepth) {
			
			LOGGER.debug("shouldUseNewCmodeThrottle: Throttling Zapi. Details: totalDepthForCluster="
					+ totalDepthForCluster + " Total Nodes=" + totalNodes + " Zapi throttle Size="
					+ connection.getOfftapPreferences().getZapiThrottleSize() + " Zapi throttle factor=" + connection.getOfftapPreferences().getZapiThrottleFactor());
			
			toReturn = true;
			StatsCollector.getInstance().addCounter("cluster-level-zapi-throttle-invoked");
		}
		return toReturn;
	}


    void recordErrorDetails(OntapConnection connection, NaElement lastZapi, Exception thrownException) {
        connection.addZapiExecutionError(thrownException.getMessage(), lastZapi, thrownException);
    }

    public NaElement invokeZapi(OntapConnection connection, NaElement filerBasedZapi) throws IOException, NaAuthenticationException, ApiFailedException, NaProtocolException {

    	boolean shouldLand = false;
		boolean succeeded = false;
		MCCManager mccManager = MCCManager.getInstance();
		MCCSlapperThread thread = null;
		try {										
				
			if (connection.isAllowRetry()) {
				// lets see if per chance this is a node level zapi and we're in mcc failover.
				boolean nodeLevelPunt = false;
				try {
					Controller controller = ControllerFactory.getInstance().getController(connection.getFilerIdOfTarget());
					if (controller instanceof NodeSvm) {
						NodeSvm asNode = (NodeSvm) controller;
						if (asNode.getParentController() != null) {
							if (MCCManager.getInstance().isFailedOverSomewhere(asNode.getParentController().getId())) {
								nodeLevelPunt = true;
							}
						}
					}
				} catch (Exception e) {
					e.printStackTrace();
				}
				if (nodeLevelPunt) {
					throw new ApiFailedException(987656, "Unable to invoke node level ZAPIs against " + connection + " as it is in mcc failover state");
				}
				
				connection = mccManager.possiblyReviseHelpfulServerForMCC(connection);
				thread = mccManager.getMCCSlapperThreadCanBeNull(connection);
			}
			if (thread != null) {
				thread.noteLaunch();
			}
			shouldLand = true;
			NaElement results = this.rawZapiInvoker.invokeZapi(connection.getUserName(), connection.getPassword(), connection.getIpAddress(), ""+connection.getPort(), 
					connection.getContext(), filerBasedZapi, getConnectionTimeout(connection.getOfftapPreferences(), connection),
                    getTimeout(connection.getOfftapPreferences(), connection), connection.isSSL(), connection.getOfftapPreferences().getWantsTraceOutput(),
                    connection.useTls(), connection.getOfftapPreferences().getSslv3Enabled());
			succeeded = true;
			return results;
		} finally {
			if (shouldLand) {					
				if (thread != null) {
					thread.noteLanding();
				}
			}
			if (!succeeded) {
				if (thread != null) {
					thread.noteError();
				}
			}
		}
    }
    
    public RawHttpZAPIInvoker getRawZapiInvoker() {
		return rawZapiInvoker;
	}

	public void setRawZapiInvoker(RawHttpZAPIInvoker rawZapiInvoker) {
		this.rawZapiInvoker = rawZapiInvoker;
	}

    private int getConnectionTimeout(OfftapPreferences prefs, OntapConnection connection) {
        if (connection.getZapiConnectionTimeout() > 0) {
            return connection.getZapiConnectionTimeout();
        }
        return prefs.getZapiConnectionTimeout();
    }
    
	private int getTimeout(OfftapPreferences prefs, OntapConnection connection) {
		if (connection.getZapiTimeout() > 0) {
			return connection.getZapiTimeout();
		}
		return prefs.getZapiTimeout();
	}

    private void writeTraceOutputIfDesired(OntapConnection connection, NaElement sent, String myInFlightKey) throws ApiFailedException {
        if (connection.getOfftapPreferences().getWantsTraceOutput()) {
            NaElement lastTraceOutput = null;
            lastTraceOutput = RawHttpZAPIInvoker.getLastTraceOutput();

            if (lastTraceOutput != null) {
                String traceOutput = lastTraceOutput.getContent();

                // now we have to put it into the right file.  We're liable to be highly parallel here, so
                // we need to find a way to lock the file.
                String lockKey = myInFlightKey;
                lockKey = lockKey.replaceAll(":", "");
                lockKey = lockKey.replaceAll("\\(", "");
                lockKey = lockKey.replaceAll("\\)", "");
                lockKey = lockKey.replaceAll(" ", "");
                String traceDirectory = Utils.addFS(".") + "traceoutput";
                synchronized (ZAPIInvoker.class) {
                    if (!checkedTraceDirectory) {
                        if (!new File(traceDirectory).exists()) {
                            new File(traceDirectory).mkdirs();
                        } else {
                            Utils.deleteFilesRecursive(traceDirectory);
                            new File(traceDirectory).mkdirs();
                        }
                        if (new File(traceDirectory).exists()) {
                            checkedTraceDirectory = true;
                        }
                    }
                    TRACE_OUTPUT_LOCK_KEYS.putIfAbsent(lockKey, lockKey);
                    lockKey = TRACE_OUTPUT_LOCK_KEYS.get(lockKey);
                }
                synchronized (lockKey) {
                    // ok - now we have a lock for the right file.
                    String filename = Utils.addFS(traceDirectory) + lockKey + ".txt";
                    try {
                        Utils.writeFile(filename, "Sent " + sent + "\nReceived\n " + traceOutput, true);
                    } catch (Exception e) {
                        // TODO - understand - saw this once - maybe locking is goofy
                    }
                }
            }
        }
    }

    public void checkForTimeout(OntapConnection connection) throws ApiFailedException {
        Long operationStartTime = ZAPIStatsCounter.getInstance().getOperationStartTime();
        if (operationStartTime != null) {
            long beenGoingDuration = System.currentTimeMillis() - operationStartTime;
            long operationalTimeoutPeriod = connection.getOfftapPreferences().getOperationTimeoutPeriod();
            if (operationalTimeoutPeriod > 0 && beenGoingDuration > operationalTimeoutPeriod) {
                // ok - this guy is over his limit.  Lets just check to see if this is an immortal requestor,
                // ie, someone that should never be timed out.  Vvols uses this for periodic discoveries to be
                // able to complete forever.
                String immortalRequestors = connection.getOfftapPreferences().getOperationTimeoutPeriodExemptions();
                if (StringUtils.isReal(immortalRequestors)) {
                    List<String> exemptions = StringUtils.splitIntoList(immortalRequestors, ",");
                    for (String exemption : exemptions) {
                        if (ZAPIStatsCounter.getInstance().getRecordedKey() != null &&
                                !ZAPIStatsCounter.getInstance().getRecordedKey().isEmpty()) {
                            for (String myRecordedKey : ZAPIStatsCounter.getInstance().getRecordedKey()) {
                                if (myRecordedKey.contains(exemption)) {
                                    return;
                                }
                            }
                        }
                    }
                }

                // in unit test environment, this is indicative of everything getting gummed up.  So if we know
                // that we're in hudson, we're gonna dump everything we know.
                if (System.getenv("HUDSON_URL") != null) {
                    String allStacks = Utils.dumpStackTraceForAllThreads();
                    System.err.println(allStacks);
                    LOGGER.error(allStacks);
                }
                throw new ApiFailedException(0, "Operation timed out - been running " + beenGoingDuration +
                        " milliseconds, with allowable maximum of " + connection.getOfftapPreferences().getOperationTimeoutPeriod());
            }
        }
    }

    public enum Priority {
        LOW, MEDIUM, HIGH
    }

    private class FlightDetailsForController {
        private AtomicInteger inFlightCount = new AtomicInteger(0);
        private AtomicInteger callCount = new AtomicInteger(0);
        private Map<Priority, AtomicInteger> inFlightCountByPriority = new EnumMap<Priority, AtomicInteger>(
                Priority.class);
        private Map<Priority, AtomicInteger> waitingByPriority = new EnumMap<Priority, AtomicInteger>(Priority.class);

        public FlightDetailsForController() {
            for (Priority priority : Priority.values()) {
                inFlightCountByPriority.put(priority, new AtomicInteger(0));
                waitingByPriority.put(priority, new AtomicInteger(0));
            }
        }

        void waitAtPriority(Priority p) {
            AtomicInteger toUse = waitingByPriority.get(p);
            toUse.incrementAndGet();
            synchronized (toUse) {
                try {
                    toUse.wait();
                } catch (InterruptedException e) {
                    // whatever
                }
            }
            toUse.decrementAndGet();
        }


        void launchAtPriority(Priority p) {
            inFlightCount.incrementAndGet();
            AtomicInteger toUse = inFlightCountByPriority.get(p);
            toUse.incrementAndGet();
        }

        void finishAtPriority(Priority p) {
            inFlightCount.decrementAndGet();
            AtomicInteger toUse = inFlightCountByPriority.get(p);
            toUse.decrementAndGet();

            // lets give our slot to someone.
            int ourCounter = callCount.incrementAndGet() % 20;
            AtomicInteger waitingLowPriority = waitingByPriority.get(Priority.LOW);
            AtomicInteger waitingMediumPriority = waitingByPriority.get(Priority.MEDIUM);
            AtomicInteger waitingHighPriority = waitingByPriority.get(Priority.HIGH);
            if (ourCounter == 19 && waitingLowPriority.get() > 0) {
                synchronized (waitingLowPriority) {
                    waitingLowPriority.notifyAll();
                }
            } else if ((ourCounter == 16 || ourCounter == 17 || ourCounter == 18) && waitingMediumPriority.get() > 0) {
                synchronized (waitingMediumPriority) {
                    waitingMediumPriority.notifyAll();
                }
            } else if (waitingHighPriority.get() > 0) {
                synchronized (waitingHighPriority) {
                    waitingHighPriority.notifyAll();
                }
            } else if (waitingMediumPriority.get() > 0) {
                synchronized (waitingMediumPriority) {
                    waitingMediumPriority.notifyAll();
                }
            } else if (waitingLowPriority.get() > 0) {
                synchronized (waitingLowPriority) {
                    waitingLowPriority.notifyAll();
                }
            }
        }
    }
}
