/*
 * $Id: //depot/prod/aeonflux/R3.0.1/src/main/java/com/netapp/offtap3/ontap/NaElement.java#1 $
 *
 * Copyright (c) 2014 NetApp, Inc.
 * All rights reserved.
 */

package com.netapp.offtap3.ontap;

import com.netapp.offtap3.internal.ontap.zapi.ARCFour;
import org.xml.sax.Attributes;

import java.util.HashMap;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Set;

/**
 *  This class is one element of a XML-like tree of elements.
 */
public class NaElement {

    private String name;
    private String content;
    private Map<String, String> attributes;
    private List<NaElement> children;

    /**
     * Construct a <code>NaElement</code> with the specified name.
     *
     * @param    name        the element name.
     */
    public NaElement(String name) {
        this(name, null);
    }

    /**
     * Construct a <code>NaElement</code> with the specified name
     * and content.
     *
     * @param    name        the element name.
     * @param    content        the element content.
     */
    public NaElement(String name, String content) {
        this(name, content, null);
    }

    /**
     * Construct a <code>NaElement</code> with a name, content, and
     * attributes (which are passed in by the XML sax parser).
     *
     * @param    name        the element name.
     * @param    content        the element content.
     * @param    attrs        the XML attributes.
     */
    public NaElement(String name, String content, Attributes attrs) {
        attributes = new HashMap<String, String>();
        children = new LinkedList<NaElement>();

        if (null != attrs) {
			/*
			 * The Attributes value gets cleared as each element
			 * is parsed, so we need to copy things out of it.
			 * This was important in the server-side version of
			 * this code - not so sure for the client-side.
			 */
            for (int i = 0; i < attrs.getLength(); i++) {
                setAttr(attrs.getQName(i), attrs.getValue(i));
            }
        }

        setName(name);
        setContent(content);
    }

    /**
     * Replace certain characters with their SGML escape sequences.
     *
     * @param    str    the string to be escaped.
     * @return the escaped string.
     */
    private static String sgmlEscape(String str) {
        StringBuilder buf = new StringBuilder();

        for (char c : str.toCharArray()) {
			/*
			 *  Pray that Java compiles this switch statement into a jump
			 *  table instead of a if-else-if chain.
			 */
            switch (c) {
                case '&':
                    buf.append("&amp;");
                    break;
                case '<':
                    buf.append("&lt;");
                    break;
                case '>':
                    buf.append("&gt;");
                    break;
                case '\'':
                    buf.append("&apos;");
                    break;
                case '"':
                    buf.append("&quot;");
                    break;
                default:
                    buf.append(c);
                    break;
            }
        }

        return buf.toString();
    }
    
    /**
     * Removes the child with the specified element name.
     * @param name the name of the child to remove.
     */
    public void removeChildByName(String name) {
        for (NaElement o : children) {
            if (o.name.equals(name)) {
                children.remove(o);
            }
        }
    }

    /**
     * Append the XML-string-encoded form of this element and all its
     * children to a buffer. The <b>pretty</b> flag specifies that the
     * encoding should include newlines and indentation. XML tags deeper
     * than the maximum nesting level are represented with
     * <code>"..."</code>.
     *
     * @param    s            the <code>StringBuilder</code> to append to.
     * @param    pretty        flag to add newlines/indenting to the XML.
     * @param    depth        nesting level of this element.
     * @param    maxDepth    maximum nesting level to stringify.
     */
    private void appendEncoded(StringBuilder s, boolean pretty, int depth, int maxDepth) {
        if (pretty) {
			/*
			 *  Add newline/indentation
			 */
            s.append('\n');
            for (int i = 0; i < depth; i++) {
                s.append('\t');
            }
        }

		/*
		 *  Append the start tag
		 */
        s.append('<');
        assert (null != name);
        s.append(name);
        for (String attrName : attributes.keySet()) {
            s.append(' ');
            s.append(attrName);
            s.append("='");

			/*
			 * We need to make sure there are no single-quotes
			 * in the value, because that would screw up
			 * the XML parsing on the other end.
			 */
            String t = getAttr(attrName);
            if (0 <= t.indexOf('\'')) {
                t = t.replace('\'', '"');
            }
            s.append(t);
            s.append('\'');
        }
        if (!hasChildren() && null == content) {
            s.append("/>");
            return;
        }
        s.append('>');

		/*
		 *  Important note: if an element has both content and children
		 *  we only append the content (the children are ignored entirely).
		 *  This matches the C implementation.
		 */

        if (null != content) {
			/*
			 *  Append content
			 */
            s.append(sgmlEscape(content));
        } else {
            if (depth < maxDepth) {
				/*
				 *  Append child tags
				 */
                for (NaElement o : children) {
                    assert (null != o);
                    o.appendEncoded(s, pretty, depth + 1, maxDepth);
                }
                if (pretty) {
					/*
					 *  Add newline/indentation
					 */
                    s.append('\n');
                    for (int i = 0; i < depth; i++) {
                        s.append('\t');
                    }
                }
            } else {
                s.append("...");
            }
        }

		/*
		 *  Append end tag
		 */
        s.append("</");
        s.append(name);
        s.append('>');
    }

    /**
     * Convert an element and all its children to a pretty XML string.
     * Readable/indented output is produced, with a prefix at the
     * beginning of each line. XML tags deeper than the maximum nesting
     * level are represented with <code>"..."</code>.
     *
     * @param    prefix        string to add to the beginning of each line.
     * @param    maxDepth    maximum nesting level to stringify.
     * @return string value containing XML.
     */
    public String toPrettyString(String prefix, int maxDepth) {
        if (null == prefix) {
            throw new NullPointerException("prefix must not be null");
        }
        StringBuilder s = new StringBuilder();
        appendEncoded(s, true, 0, maxDepth);
        String raw = s.toString();
        String pretty = raw.replaceAll("\n", "\n" + prefix);
        return pretty.substring(1) + "\n";
    }

    /**
     * Convert an element and all its children to a pretty XML string.
     * Readable/indented output is produced, with a prefix at the
     * beginning of each line.
     *
     * @param    prefix        string to add to the beginning of each line.
     * @return string value containing XML.
     */
    public String toPrettyString(String prefix) {
        return toPrettyString(prefix, Integer.MAX_VALUE);
    }

    /**
     * Convert an element and all its children to a single
     * string, formatted as XML.
     *
     * @return String value containing XML.
     */
    public String toString() {
        StringBuilder s = new StringBuilder();
        appendEncoded(s, false, 0, Integer.MAX_VALUE);
        return s.toString();
    }

    /**
     * Set the name of this element.
     *
     * @param    name    the new name.
     */
    public void setName(String name) {
        if (null == name) {
            throw new NullPointerException("name must not be null");
        }
        this.name = name;
    }

    /**
     * Get the name of this element.
     *
     * @return name of the element.
     */
    public String getName() {
        assert (null != name);
        return name;
    }

    /**
     * Find a child with the specified name and return its string content.
     * Returns null if it's unable to find child with the specified name.
     *
     * @param    name    name of the desired child.
     * @return content of the child.
     */
    public String getChildContent(String name) {
        NaElement child = getChildByName(name);
        return child == null ? null : child.getContent();
    }

    /**
     * Find a child with the specified name and return its content as
     * as an int value.
     *
     * @param    name    name of the desired child.
     * @param    defVal    if the named child does not exist,
     * this value is returned.
     * @return int value of that child.
     */
    public int getChildIntValue(String name, int defVal) {
        try {
            return Integer.parseInt(getChildContent(name));
        } catch (NumberFormatException e) {
            return defVal;
        }
    }

    /**
     * Find a child with the specified name and return its content as
     * as an int value.
     *
     * @param    name    name of the desired child.
     * @param    defVal    if the named child does not exist,
     * this value is returned.
     * @return long value of that child.
     */
    public long getChildLongValue(String name, long defVal) {
        try {
            return Long.parseLong(getChildContent(name));
        } catch (NumberFormatException e) {
            return defVal;
        }
    }

    /**
     * Get the list of children.
     *
     * @return    <code>List</code> containing <code>NaElement</code> values.
     */
    public List<NaElement> getChildren() {
        return children;
    }

    /**
     * Find and return the child element with a specified name.
     * The value returned is the actual child, not a cloned copy,
     * so modifications to it will affect the actual child.
     * Returns null if it's unable to find child element with
     * specified name.
     *
     * @param    name    rhe name of the child to find.
     * @return a <code>NaElement</code> which is the child.
     */
    public NaElement getChildByName(String name) {
        for (NaElement o : children) {
            if (o.name.equals(name)) {
                return o;
            }
        }
        return null;
    }

    /**
     * Removes the children with the specified element name.
     * @param name the name of the children to remove.
     */
    public void removeChildrenByName(String name) {
        for (Iterator<NaElement> iter = children.iterator(); iter.hasNext();) {
            if (iter.next().name.equals(name)) {
                iter.remove();
            }
        }
    }

    /**
     * Determine if there are children.
     *
     * @return false if there are no children, true otherwise.
     */
    public boolean hasChildren() {
        return (0 != children.size());
    }

    /**
     * Set the value of the content of this element.
     *
     * @param    content    the content.
     */
    public void setContent(String content) {
        this.content = content;
    }

    /**
     * Encrypts the given value and sets the encrypted value as the
     * content of this element.
     *
     * @param    content    The value to be encrypted.
     */
    public void setEncryptedContent(String content) {
        this.content = ARCFour.encryptAndEncode(content);
    }

    /**
     * Get the value of the content of this element.
     *
     * @return the content.
     */
    public String getContent() {
        return (content == null) ? "" : content;
    }

    /**
     * Add a new child element with a given name and content.
     *
     * @param    name        The name of the new child element.
     * @param    content        The content value of the new child element.
     */
    public void addNewChild(String name, String content) {
        addChildElem(new NaElement(name, content));
    }

    /**
     * Add a child element.
     *
     * @param    elem    the child elemet to add.
     */
    public void addChildElem(NaElement elem) {
        children.add(elem);
    }

    /**
     * Set the value of an attribute, or add a new attribute.
     *
     * @param    name    attribute name.
     * @param    value    attribute value.
     */
    public void setAttr(String name, String value) {
        attributes.put(name, value);
    }

    /**
     * Get the value of an attribute.
     *
     * @param    name    attribute name.
     * @return value of the attribute (or null).
     * Returns null if it's unable to get value for
     * a given attribute.
     */
    public String getAttr(String name) {
        return attributes.get(name);
    }

    /**
     * Gets the names of the attributes.
     *
     * @return the attribute names (String objects)
     */
    public Set<String> getAttrNames() {
        return attributes.keySet();
    }

    /**
     * Encrypts data contained in content. Adds a new child element with a
     * given name and encrypted content.
     *
     * @param name The name of new child element.
     * @param    content        The content value of new child element.
     */

    public void addNewEncryptedChild(String name, String content) {
        String output;
        output = ARCFour.encryptAndEncode(content);
        addNewChild(name, output);
    }

    /**
     * Finds a child with the specified name , decrypts it's string content
     * and returns decrypted content. Returns null if it's unable to find
     * child element with specified name.
     *
     * @param name The name of child element.
     * @return The content of child element.
     */

    public String getChildEncryptContent(String name) {
        String output;
        String content;
        content = getChildContent(name);
        if (null == content) {
            return null;
        }

        output = ARCFour.decodeAndDecrypt(content);
        return output;
    }

    /**
     * Decrypts and returns the value of content of this element.
     *
     * @return The content of this element.
     */

    public String getEncryptContent() {
        String output;
        String content;
        content = getContent();
        output = ARCFour.decodeAndDecrypt(content);
        return output;
    }
}
