package com.netapp.offtap3.ontap;

import java.util.regex.Matcher;
import java.util.regex.Pattern;

public class QosThroughput {

    /**
     * Identifies the valid Qos Max-throughput Units
     */
    public enum Unit {
        IOPS("iops"), BPS("B/s"), KBPS("KB/s"), MBPS("MB/s"), GBPS("GB/s"), TBPS("TB/s"), PBPS("PB/s"), NONE("none"), INF("INF");

        Unit(String format) {
            defaultFormat = format;
        }

        // typical string representation of unit
        private String defaultFormat;

        public String getDefaultFormat() {
            return defaultFormat;
        }

    }

   // this is the magic limit from ONTAP
    static final Double UPPER_BOUND_MBPS = 8192.0 * 1024 * 1024 * 1024;

    private static final Pattern decimalPattern = Pattern.compile("\\d+(\\.\\d+)?");

    private double value;
    private Unit unit;

    public QosThroughput(double value, Unit unit) {
        this.value = value;
        this.unit = unit;
        Double mbps = convertToMbps();
        if (mbps != null) {
            if (mbps.equals(UPPER_BOUND_MBPS)) {
                setUnit(Unit.INF);
            }
            else if (mbps > UPPER_BOUND_MBPS) {
                throw new IllegalArgumentException("throughput value exceeds maximum");
            }
        }
    }

    /**
     * Parses the string representation of a max-throughput value found
     * in ZAPI XML and assigns the value and unit to the returned instance.
     * Examples of valid specifications are: 100B/S, 10KB/S, 1GB/S,
     * 500MB/S, 1TB/S, and 100IOPS
     *
     * @param throughput string representation of max-throughput
     * @throws IllegalArgumentException if input cannot be parsed
     */
    public QosThroughput(String throughput) {
        Matcher matcher = decimalPattern.matcher(throughput);
        if (matcher.find()) {
            value = Double.valueOf(matcher.group());
            throughput = throughput.substring(matcher.group().length());
        }

        String unitString = throughput.toLowerCase().trim();
        QosThroughput.Unit parsedUnit = null;
        for (QosThroughput.Unit unit : QosThroughput.Unit.values()) {
            if (unitString.equals(unit.getDefaultFormat().toLowerCase())) {
                parsedUnit = unit;
                break;
            }
        }

        if (parsedUnit == null) {
            throw new IllegalArgumentException("Invalid QoS expression '" + throughput + "'");
        }
        setUnit(parsedUnit);
    }

    public double getValue() {
        return value;
    }

    public Unit getUnit() {
        return unit;
    }

    private void setUnit(Unit unit) {
        this.unit = unit;

        // reset value if it no longer makes sense
        if (unit == Unit.INF || unit == Unit.NONE) {
            value = 0.0;
        }
    }

    /**
     * Converts the throughput value for this instance to Mbps, unless
     * it is in IOPS, NONE, or INF, in which case null is returned.
     * @return throughput value in Mbps or null
     */
    public Double convertToMbps() {
        switch (unit) {
            case MBPS:
                return value;
            case BPS:
                return ((value / 1024) / 1024);
            case KBPS:
                return value / 1024;
            case GBPS:
                return value * 1024;
            case TBPS:
                return value * 1024 * 1024;
            case PBPS:
                return value * 1024 * 1024 * 1024;

            case IOPS:
            case NONE:
            case INF:
            default:
                return null;
        }
    }

    public QosThroughput toONTAPResolution() {
        // ONTAP resolution is already precise for these units
        if (unit == Unit.IOPS || unit == Unit.NONE || unit == Unit.INF) {
            return this;
        }

        double ontapValue = value;
        Unit ontapUnit = unit;

        // any value >= 1024 should be represented with the next highest unit
        // but, ONTAP does not support units beyond PBPS
        while (ontapValue >= 1024 && ontapUnit != Unit.PBPS) {
            ontapValue /= 1024;
            ontapUnit = Unit.values()[ontapUnit.ordinal()+1];
        }

        // also, the value with that unit must be rounded to 2 decimal places
        ontapValue = Math.round(ontapValue * 100.0) / 100.0;

        return new QosThroughput(ontapValue, ontapUnit);
    }

    public String toString() {
        switch (unit) {
            case NONE:
            case INF:
                return unit.getDefaultFormat();

            case IOPS:
                //Cast to long to avoid trailing .0 decimal for IOPS which ONTAP dislikes
                return ((long) value) + unit.getDefaultFormat();

            case BPS:
            case KBPS:
            case MBPS:
            case GBPS:
            case TBPS:
            case PBPS:
                return value + unit.getDefaultFormat();
            default:
                throw new IllegalArgumentException("Unknown unit: " + unit);
        }
    }

    public boolean equals(Object obj) {
        if (obj instanceof QosThroughput) {
            QosThroughput tp = (QosThroughput) obj;
            if (unit == Unit.INF || unit == Unit.NONE) {
                return unit == tp.unit;
            }
            else if (unit == Unit.IOPS) {
                return value == tp.value;
            }
            else {
                return convertToMbps().equals(tp.convertToMbps()) ;
            }
        }
        return false;
    }

    public int hashCode() {
        return ((int) value) ^ ((unit == null) ? 0 : unit.hashCode());
    }
}
