package com.netapp.offtap3.utils;

import java.util.HashMap;
import java.util.Map;
import java.util.concurrent.Executors;
import java.util.concurrent.ScheduledExecutorService;
import java.util.concurrent.TimeUnit;


public class ExpiringHashMap<K, V> extends HashMap<K, V> {
	private Map<K, Long> insertionTimes = new HashMap<K, Long>();
	private long keepDuration = 0;
	private final static ScheduledExecutorService executor = Executors.newScheduledThreadPool(2);

	public ExpiringHashMap(long durationToKeep) {
		keepDuration = durationToKeep;
		executor.scheduleWithFixedDelay(new RemoveExpiredObjects(this), this.keepDuration / 2, this.keepDuration, TimeUnit.MILLISECONDS);
	}

	private class RemoveExpiredObjects extends Thread {
		ExpiringHashMap<K, V> cacheHashMap;
		public RemoveExpiredObjects(ExpiringHashMap<K, V> cacheHashMap) {
			this.cacheHashMap = cacheHashMap;
		}

		public void run() {
			synchronized(cacheHashMap) {
				long whackEverythingBeforeDate = System.currentTimeMillis() - keepDuration;
                for (Map.Entry<K, Long> entry : insertionTimes.entrySet()) {
                    if (entry.getValue() < whackEverythingBeforeDate) {
                        cacheHashMap.remove(entry.getKey());
                    }
                }
			}
		}

	}

	@Override
	public synchronized V put(K key, V value) {
		if (key != null) {
			insertionTimes.put(key, System.currentTimeMillis());
		}
		return super.put(key, value);
	}

	@Override
	public synchronized V remove(Object key) {
		insertionTimes.remove(key);
		return super.remove(key);
	}

	@Override
	public synchronized void clear() {
		insertionTimes.clear();
		super.clear();
	}

	@Override
	public synchronized V get(Object key) {
		Long insertionTime = this.insertionTimes.get(key);
		if (insertionTime == null) return null;
		long whackEverythingBeforeDate = System.currentTimeMillis() - this.keepDuration;
		if (insertionTime < whackEverythingBeforeDate) {
			this.remove(key);
			return null;
		}
		return super.get(key);
	}
}
