/*
 * $Id: //depot/prod/aeonflux/R3.0.1/src/main/java/com/netapp/offtap3/utils/IPAddressUtil.java#3 $
 */
package com.netapp.offtap3.utils;

import com.google.common.net.InetAddresses;
import org.apache.log4j.Logger;

import com.netapp.offtap3.IpInterface;
import com.netapp.offtap3.internal.ontap.utils.StatsCollector;
import com.netapp.offtap3.ontap.OntapController;

import java.net.Inet4Address;
import java.net.Inet6Address;
import java.net.InetAddress;
import java.net.UnknownHostException;
import java.util.Map;
import java.util.regex.Pattern;

/**
 * Converts String hosts or IP addresses to InetAddress objects.
 */
public class IPAddressUtil {
    private static Logger logger = Logger.getLogger(IPAddressUtil.class);

    private static final String _255 = "(?:25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)";
    private static final String _4HEX = "[0-9a-fA-F]{1,4}";
    private static final Pattern IPV4_PATTERN = Pattern.compile( "^(?:" + _255 + "\\.){3}" + _255 + "$");
    private static final Pattern IPV6_STD_PATTERN = Pattern.compile("^(?:" + _4HEX + ":){7}" + _4HEX + "$");
    private static final Pattern IPV6_HEX_COMPRESSED_PATTERN = Pattern.compile("^((?:" + _4HEX + "(?::" + _4HEX + ")*)?)::((?:" + _4HEX + "(?::" + _4HEX + ")*)?)$");

    /**
     * True if the given address is an IPV4 address.
     *
     * @param address the address to test
     * @return true if the address is an IPV4 address.
     */
    public static boolean isIPV4(String address) {
        return IPV4_PATTERN.matcher(address).matches();
    }

    /**
     * True if the given address is an IPV6 address.
     *
     * @param address the address to test
     * @return true if the address is a standard or hex compressed IPV6 address.
     */
    public static boolean isIPV6(String address) {
        return IPV6_STD_PATTERN.matcher(address).matches() || IPV6_HEX_COMPRESSED_PATTERN.matcher(address).matches();
    }

    /**
     * Returns the IP address for the given host.  If the hostname is not resolvable, then the host is returned.
     *
     * @param host hostname
     * @return String IP address
     */
    public static String toIPV4(String host) {
        String ipAddress = host;

        if (!IPV4_PATTERN.matcher(host).matches()) {
            try {
                ipAddress = InetAddress.getByName(host).getHostAddress();
                logger.debug(String.format("Resolving %s to %s", host, ipAddress));

            } catch (UnknownHostException uhe) {
                logger.warn(uhe.getMessage(), uhe);
            }
        }

        return ipAddress;
    }

    /**
     * Returns the byte[] array of a String representing an IP address.  If the argument
     * is not an IP address, null is returned.
     *
     * @param ip
     * @return byte[] - IP address as a byte array, or null if ip is not a valid IP address.
     */
    public static byte[] textToNumericFormatV4(String ip) {
        logger.trace("textToNumericFormatV4(" + ip + "): enter");
        byte[] result = null;
        try {
            InetAddress addr = InetAddresses.forString(ip);
            if (addr instanceof Inet6Address)
            {
                logger.trace(ip + " is IPv6, but asked to parse as IPv4, so returning null");
                return null;
            }
            result = addr.getAddress();
        }
        catch (IllegalArgumentException iae) {
            logger.trace("WARNING: " + ip + " was not valid", iae);
        }
        logger.trace("textToNumericFormatV4(" + ip + "): exit");
        return result;
    }

    /**
     * Returns the int[] array of a String representing an IPv6 address.  If the argument
     * is not an IPv6 address, null is returned.
     * <p/>
     * @param ipString
     * @return int[] - IPv6 address as an int array, or null if ipString is not a valid IPv6 address.
     */
    public static int[] textToNumericFormatV6(String ipString) {
        int[] ipArray = null;
        try {
            InetAddress addr = InetAddresses.forString(ipString);
            byte[] addrArray = addr.getAddress();

            // just in case we were fed an IPv4 address using IPv6 formatting
            if (addr instanceof Inet4Address) {
                logger.trace("Asked to parse " + ipString + " as IPv6, but is really IPv4");
                return null;
            }

            // now pack them up as int[]
            ipArray = new int[8];
            for (int i=0; i < ipArray.length; i++) {
                ipArray[i] = ((int)addrArray[i*2]) << 8 & 0x0000ff00
                        | ((int)addrArray[(i*2)+1]) & 0x000000ff;
            }
    }
        catch (IllegalArgumentException iae) {
            logger.trace("WARNING: " + ipString + " was not valid", iae);
        }
        return ipArray;
    }

    /**
     * Returns the given IP address or hostname as an InetAddress
     *
     * @param host
     * @return InetAddress
     */
    public static InetAddress getInetAddress(String host) throws UnknownHostException {
        InetAddress inetAddr = null;

        byte[] address = textToNumericFormatV4(host);

        if (address == null) {
            inetAddr = InetAddress.getByName(host);
            logger.trace("getInetAddress(" + host + "): DNS lookup");

        } else {
            inetAddr = InetAddress.getByAddress(host, address);
            logger.trace("getInetAddress(" + host + "): *no* DNS lookup, IP address used");
        }

        return inetAddr;
    }

    /**
     * Attempts to convert a hostname or ip address to a valid ip address.
     * Unknown host exception will cause null to be returned.
     *
     * @param host
     * @return Returns the given IP address or hostname as a String
     */
    public static String getIPAddress(String host) {
        InetAddress inetAddr = null;
        String ip = null;

        byte[] address = textToNumericFormatV4(host);

        try {
            if (address == null) {
                inetAddr = InetAddress.getByName(host);
                ip = inetAddr.getHostAddress();
                logger.trace("getIPAddress(" + host + "): DNS lookup resulted in " + ip);
            } else {
                inetAddr = InetAddress.getByAddress(host, address);
                ip = inetAddr.getHostAddress();
                logger.trace("getIPAddress(" + host + "): *no* DNS lookup, ip is " + ip);
            }
        } catch (UnknownHostException e) {
            StatsCollector.getInstance().addCounter("error-failed to find ip address for " + host);
            logger.warn("getIPAddress: failed to find ip address for " + host + ". " + e.getMessage());
        }
        return ip;
    }

    public static int[] buildIpV6Netmask(int prefixLength) {
        int mask[] = new int[8];

        if (prefixLength <= 16) {
            mask[0] = (0xffff << (16 - prefixLength)) & 0xffff;
            mask[1] = 0;
            mask[2] = 0;
            mask[3] = 0;
            mask[4] = 0;
            mask[5] = 0;
            mask[6] = 0;
            mask[7] = 0;
        } else if ((prefixLength > 16) && (prefixLength <= 32)) {
            mask[0] = 0xffff;
            mask[1] = (0xffff << (32 - prefixLength)) & 0xffff;
            mask[2] = 0;
            mask[3] = 0;
            mask[4] = 0;
            mask[5] = 0;
            mask[6] = 0;
            mask[7] = 0;
        } else if ((prefixLength > 32) && (prefixLength <= 48)) {
            mask[0] = 0xffff;
            mask[1] = 0xffff;
            mask[2] = (0xffff << (48 - prefixLength)) & 0xffff;
            mask[3] = 0;
            mask[4] = 0;
            mask[5] = 0;
            mask[6] = 0;
            mask[7] = 0;
        } else if ((prefixLength > 48) && (prefixLength <= 64)) {
            mask[0] = 0xffff;
            mask[1] = 0xffff;
            mask[2] = 0xffff;
            mask[3] = (0xffff << (64 - prefixLength)) & 0xffff;
            mask[4] = 0;
            mask[5] = 0;
            mask[6] = 0;
            mask[7] = 0;
        } else if ((prefixLength > 64) && (prefixLength <= 80)) {
            mask[0] = 0xffff;
            mask[1] = 0xffff;
            mask[2] = 0xffff;
            mask[3] = 0xffff;
            mask[4] = (0xffff << (80 - prefixLength)) & 0xffff;
            mask[5] = 0;
            mask[6] = 0;
            mask[7] = 0;
        } else if ((prefixLength > 80) && (prefixLength <= 96)) {
            mask[0] = 0xffff;
            mask[1] = 0xffff;
            mask[2] = 0xffff;
            mask[3] = 0xffff;
            mask[4] = 0xffff;
            mask[5] = (0xffff << (96 - prefixLength)) & 0xffff;
            mask[6] = 0;
            mask[7] = 0;
        } else if ((prefixLength > 96) && (prefixLength <= 112)) {
            mask[0] = 0xffff;
            mask[1] = 0xffff;
            mask[2] = 0xffff;
            mask[3] = 0xffff;
            mask[4] = 0xffff;
            mask[5] = 0xffff;
            mask[6] = (0xffff << (112 - prefixLength)) & 0xffff;
            mask[7] = 0;
        } else if ((prefixLength > 112) && (prefixLength <= 128)) {
            mask[0] = 0xffff;
            mask[1] = 0xffff;
            mask[2] = 0xffff;
            mask[3] = 0xffff;
            mask[4] = 0xffff;
            mask[5] = 0xffff;
            mask[6] = 0xffff;
            mask[7] = (0xffff << (128 - prefixLength)) & 0xffff;
        }

        return mask;
    }

    public static String getIpV6Network(int[] ip, int[] netmask) {
        if ((ip.length != 8) || (netmask.length != 8)) {
            return null;
        }

        int[] netArray = new int[8];
        for (int i = 0; i < 8; i++) {
            netArray[i] = ip[i] & netmask[i];
        }

        String network = String.format("%s:%s:%s:%s:%s:%s:%s:%s", Integer.toHexString(netArray[0]), Integer.toHexString(netArray[1]),
                Integer.toHexString(netArray[2]), Integer.toHexString(netArray[3]), Integer.toHexString(netArray[4]),
                Integer.toHexString(netArray[5]), Integer.toHexString(netArray[6]), Integer.toHexString(netArray[7]));

        return network;
    }

    public static String getNetwork(String ipAddress, String netmask) {
        if (isIPV4(ipAddress)) { // we have an IPv4 address
            byte[] ipArray = textToNumericFormatV4(ipAddress);
            byte[] maskArray = textToNumericFormatV4(netmask);
            Integer[] netArray = new Integer[4];

            for (int i = 0; i < 4; i++) {
                netArray[i] = byteToInt(ipArray[i]) & byteToInt(maskArray[i]);
            }
            return String.format("%d.%d.%d.%d", netArray[0], netArray[1], netArray[2], netArray[3]);
        } else { //we have an IPv6 address
            if (ipAddress.contains("%")) {
                ipAddress = ipAddress.substring(0, ipAddress.indexOf("%"));
            }

            // build the IP array
            int[] ipArray = textToNumericFormatV6(ipAddress);

            // build the netmask array
            int prefixLength = 0;
            try {
                prefixLength = Integer.parseInt(netmask);
            } catch (NumberFormatException nfe) {
                // we're seeing ffff:ffff:ffff:ffff:: coming back from filers, in that case
                // lets set prefix length somewhat reasonably.  I am far from an ipv6 expert
                // so lets just do what we can.
                if (netmask != null) {
                    netmask = netmask.replaceAll(":", "");
                    // I want to use Long.parseLong, but that's failing for values like ffffffffffffffff for reasons that I really don't 
                    // understand.  So we'll ghetto rig it.
                    for (char c : netmask.toCharArray()) {
                        if ('f' == c) {
                            prefixLength += 4;
                        }
                        if ('8' == c) {
                            prefixLength += 3;
                        }
                        if ('4' == c) {
                            prefixLength += 2;
                        }
                        if ('2' == c) {
                            prefixLength += 1;
                        }
                    }
                }
            }
            int[] maskArray = buildIpV6Netmask(prefixLength);

            return getIpV6Network(ipArray, maskArray);
        }
    }

    public static String getNetworkDev(OntapController controller, String ipAddress) {
        if (ipAddress != null) {
            ipAddress = toIPV4(ipAddress);
            Map<String, ? extends IpInterface> ipInterfaces = controller.getIpInterfaces();

            if (ipInterfaces.containsKey(ipAddress)) {
                return ipInterfaces.get(ipAddress).getName();
            }
        }

        return null;
    }

    public static IpInterface getInterface(OntapController controller, String ipAddress) {
        if (ipAddress != null){
            ipAddress = toIPV4(ipAddress);
            //TODO: check usables
            for (IpInterface inter : controller.getIpInterfaces().values()){
                if (inter != null && areIPOrHostnameEquivalent(ipAddress, inter.getIpAddress())) {
                    return inter;
                }
            }
        }

        return null;
    }

    public static String getiSCSINetworkDev(OntapController controller, String ipAddress) {
        if (ipAddress != null) {
            ipAddress = toIPV4(ipAddress);
            Map<String, ? extends IpInterface> interfaces = controller.getIpInterfaces();
            if (interfaces.containsKey(ipAddress)) {
                IpInterface inter = interfaces.get(ipAddress);
                if (inter.isIscsiEnabled()){
                    return inter.getName();
                }
            }
        }

        return null;
    }

    public static boolean areIPOrHostnameEquivalent(String host1, String host2) {
        if (Utils.areEqual(host1, host2)) {
            return true;
        }
        // ok - if they are not precisely equal - lets go and pass them thru a hostname
        // resolver and see if what they resolve to is equal
        long startTime = System.currentTimeMillis();
        try {
            String effectiveHost1 = null;
            if (StringUtils.isReal(host1)) {
                effectiveHost1 = IPAddressUtil.getIPAddress(host1);
            } else {
                effectiveHost1 = "";
            }
            String effectiveHost2 = null;
            if (StringUtils.isReal(host2)) {
                effectiveHost2 = IPAddressUtil.getIPAddress(host2);
            } else {
                effectiveHost2 = "";
            }
            // special case - two unresolvables - shouldn't be equal by default.
            if (effectiveHost1 == null && effectiveHost2 == null) {
                return false;
            }
            return Utils.areEqual(effectiveHost1, effectiveHost2);
        } finally {
            StatsCollector.getInstance().addCounter("hostname-resolution-time", System.currentTimeMillis() - startTime);
        }
    }

    private static int byteToInt(byte b) {
        return (int) b & 0xFF;
    }
}


