package com.netapp.offtap3.utils;


import com.netapp.offtap3.internal.ontap.support.ZAPIStatsCounter;
import com.netapp.offtap3.ontap.NaElement;
import com.netapp.offtap3.ontap.utils.OntapCallable;
import com.netapp.offtap3.ontap.utils.OntapThreadSpecificInfo;

import org.apache.log4j.Logger;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.NamedNodeMap;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;

import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;

import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.Closeable;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.Flushable;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.io.Serializable;
import java.net.URL;
import java.nio.ByteBuffer;
import java.nio.ByteOrder;
import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.Collections;
import java.util.Comparator;
import java.util.Date;
import java.util.HashSet;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.concurrent.CompletionService;
import java.util.concurrent.ExecutorCompletionService;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.concurrent.Future;
import java.util.concurrent.ThreadFactory;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.atomic.AtomicInteger;
import java.util.zip.GZIPInputStream;
import java.util.zip.GZIPOutputStream;
import java.util.zip.ZipEntry;
import java.util.zip.ZipOutputStream;

/**
 * Helper class for utility functions - can be refactored as needed
 *
 * @author dsarisky
 */
public class Utils {
    private static final Logger LOGGER = Logger.getLogger(Utils.class);
    private static DocumentBuilder DOCUMENT_BUILDER;

    static {
        try {
            DOCUMENT_BUILDER = DocumentBuilderFactory.newInstance().newDocumentBuilder();
        } catch (ParserConfigurationException e) {
            LOGGER.error("Error: Could not build the XML parser. ZAPI parsing may not work correctly.", e);
        }
    }

    public static <T> Set<T> toSet(T... objects) {
        HashSet<T> retVal = new HashSet<T>();
        if (objects != null) {
            Collections.addAll(retVal, objects);
        }
        return retVal;
    }

    public static <K, V> Map<K, V> toTypedMap(K key, V value) {
        Map<K, V> retVal = new LinkedHashMap<K, V>();
        retVal.put(key, value);
        return retVal;
    }

    public static Map toMap(Object... objects) {
        Object nullObjectNeverUsed = "null_object_never_used";
        Map retVal = new LinkedHashMap();
        Object currKey = nullObjectNeverUsed;
        for (Object current : objects) {
            if (nullObjectNeverUsed == currKey) {
                currKey = current;
            } else {
                retVal.put(currKey, current);
                currKey = nullObjectNeverUsed;
            }
        }
        return retVal;
    }

    public static <T> List<T> toList(T... values) {
        List<T> list = new ArrayList<T>(values.length);
        Collections.addAll(list, values);
        return list;
    }

    /**
     * Helper method that gets a URL and returns the contents.
     *
     * @return The contents as a string
     */
    public static String getURLContentAsString(String url) throws IOException {
        StringBuilder retVal = new StringBuilder();
        InputStream inputStream = new URL(url).openStream();
        try {
            BufferedReader br = new BufferedReader(new InputStreamReader(inputStream));
            String line;
            while ((line = br.readLine()) != null) {
                retVal.append(line);
                retVal.append("\n");
            }
            return retVal.toString().trim();
        } finally {
            inputStream.close();
        }
    }

    public static String formatDate(Date d) {
        if (d == null) {
            return "<Empty>";
        }
        SimpleDateFormat sdf = new SimpleDateFormat("MM/dd/yy HH:mm:ss");
        return sdf.format(d);
    }

    /**
     * Helper method that gets a URL and returns the contents.
     *
     * @return The contents as a String if it can read them, a null if there is a problem
     */
    public static String getURLContentAsStringSafe(String url) {
        try {
            return getURLContentAsString(url);
        } catch (Exception e) {
            return "";
        }
    }

    public static <T> List<T> removeDuplicates(Collection<T> incoming) {
        if (incoming == null) {
            return null;
        }
        return new ArrayList<T>(new HashSet<T>(incoming));
    }

    public static <X> List<X> getUICentricSortedList(Collection<X> c) {
        ArrayList<X> retVal = new ArrayList<X>(c);
        try {
            Collections.sort((List) retVal, new UICentricComparator());
        } catch (Exception e) {
            // whatever, won't be sorted
        }
        return retVal;
    }

    public static boolean isReal(Collection c) {
        return (c != null && c.size() > 0);
    }

    public static boolean isReal(Map c) {
        return (c != null && c.size() > 0);
    }

    public static boolean isWindows() {
        return System.getProperty("os.name").toUpperCase().contains("WIN");
    }

    public static long median(List<? extends Number> values) {
        if (!Utils.isReal(values)) {
            return 0;
        }
        List<? extends Number> sorted = Utils.getSortedList(values);
        Number n = sorted.get(sorted.size() / 2);
        if (n != null) {
            return n.longValue();
        } else {
            return 0;
        }
    }

    public static int parseInt(String s) {
        try {
            return Integer.parseInt(s);
        } catch (Exception e) {
            return 0;
        }
    }

    public static long parseLong(String s) {
        try {
            return Long.parseLong(s);
        } catch (Exception e) {
            return 0;
        }
    }

    /**
     * Returns the string value of the given array
     *
     * @param data the array to convert to a string
     * @return the string value
     */
    public static String getStringValue(int[] data) {
        byte[] bytes = new byte[data.length];
        for (int i = 0; i < data.length; i++) {
            bytes[i] = (byte) data[i];
        }

        return new String(bytes);
    }

    /**
     * Returns a long value across the given array
     *
     * @param input the array to convert to a long
     * @return a long value across the given array
     */
    public static long getLong(int[] input) {
        int size = (input.length > 7) ? input.length : 8;

        ByteBuffer bb = ByteBuffer.allocate(size);
        bb.order(ByteOrder.LITTLE_ENDIAN);
        for (int anInput : input) {
            bb.put((byte) anInput);
        }

        return bb.getLong(0);
    }

    /**
     * Get a new array of int based on an existing array.
     *
     * @param array - starting offset
     * @param start - ending offset
     * @param end   - data before end index is returned, but not end index
     */
    public static int[] subArray(int[] array, int start, int end) {
        if (array == null) {
            return new int[0];
        }
        if (start < 0) {
            start = 0;
        }
        if (end > array.length) {
            end = array.length;
        }

        int subSize = end - start;

        if (subSize <= 0) {
            return new int[0];
        }

        int[] subArray = new int[subSize];
        System.arraycopy(array, start, subArray, 0, subSize);
        return subArray;
    }

    public static List<String> getFilesRecursive(String directory, boolean includeDirectories) {
        List<String> fullPaths = new ArrayList<String>();
        getAllFilesRecursive(directory, fullPaths, includeDirectories);
        return fullPaths;
    }

    public static List<String> getLinesSafe(String filename) throws Exception {
        try {
            return getLines(filename);
        } catch (Exception e) {
            return new ArrayList<String>();
        }
    }

    /**
     * Returns a mask used to get variable length bytes from the bitmap data run offset
     * @param length the length of the data run offset
     *
     * @return a mask used to get variable length bytes from the bitmap data run offset
     */
    public static long getMask(int length) {
        switch(length) {
            case 1: return 0xFFL;
            case 2: return 0xFFFFL;
            case 3: return 0xFFFFFFL;
            case 4: return 0xFFFFFFFFL;
            case 5: return 0xFFFFFFFFFFL;
            case 6: return 0xFFFFFFFFFFFFL;
            case 7: return 0xFFFFFFFFFFFFFFL;
            case 8: return 0xFFFFFFFFFFFFFFFFL;
            default: throw new IllegalArgumentException("Don't do this!");
        }
    }

    public static long sumCollection(Collection<? extends Number> collection) {
        if (collection == null) {
            return 0;
        }
        long accumulator = 0;
        for (Number number : collection) {
            accumulator += number.longValue();
        }
        return accumulator;
    }

    private static void getAllFilesRecursive(String root, List<String> soFar, boolean includeDirectories) {
        String[] children = new File(root).list();
        if (children != null) {
            for (String kid : children) {
                File kidFile = new File(root + File.separator + kid);
                if (kidFile.isDirectory()) {
                    if (includeDirectories) {
                        soFar.add(kidFile.getAbsolutePath());
                    }
                    getAllFilesRecursive(kidFile.getAbsolutePath(), soFar, includeDirectories);
                } else {
                    soFar.add(kidFile.getAbsolutePath());
                }
            }
        }
    }

    public static void deleteFilesRecursive(String rootPath) {
        // not try and order this, lets just loop that bad boy until we run out of files
        // or cannot make any more progress.
        while (true) {
            boolean anySucceedeed = false;
            List<String> files = getFilesRecursive(rootPath, true);
            for (String file : new ArrayList<String>(files)) {
                if (file.contains(".nfs000")) {
                    // lets not freak out about nfs locking files - we'll just act like they aren't there
                    files.remove(file);
                }
            }
            if (files.size() == 0) {
                break;
            }
            for (String file : files) {
                if (new File(file).delete()) {
                    anySucceedeed = true;
                }
            }
            if (!anySucceedeed && files.size() > 0) {
                throw new RuntimeException("Unable to delete all files in " + rootPath + " still see " + files);
            }
        }
        new File(rootPath).delete();
    }

    public static void writeFile(String filename, String contents) throws IOException {
        writeFile(filename, contents, false);
    }

    public static void writeFile(String filename, String contents, boolean append) throws IOException {
        BufferedWriter bw = null;
        try {
            bw = new BufferedWriter(new FileWriter(filename, append));
            bw.write(contents);
            bw.flush();
        } finally {
            safeClose(bw);
        }
    }

    public static void writeFileBinary(String filename, byte[] contents, boolean append) throws IOException {
        FileOutputStream bw = null;
        try {
            bw = new FileOutputStream(filename, append);
            bw.write(contents);
            bw.flush();
            bw.close();
        } finally {
            safeClose(bw);
        }
    }

    public static byte[] readFileBinary(String filename) throws IOException {
        FileInputStream fis = new FileInputStream(filename);
        ByteArrayOutputStream baos = new ByteArrayOutputStream();
        try {
            byte[] bytes = new byte[16384];
            int length;
            while ((length = fis.read(bytes)) != -1) {
                baos.write(bytes, 0, length);
            }
            return baos.toByteArray();
        } finally {
            safeClose(fis);
        }
    }

    public static void writeFileSafe(String filename, String contents) {
        try {
            writeFile(filename, contents);
        } catch (IOException e) {
            // whatever
        }
    }

    public static List<String> getKeysSortedByValues(Map<String, Long> theMap) {
        List<Long> values = new ArrayList<Long>(theMap.values());
        List<String> keysToReturn = new ArrayList<String>();
        try {
            Collections.sort(values);
        } catch (RuntimeException e) {
            // no problem
        }
        for (Long curr : values) {
            for (Map.Entry<String, Long> entry : theMap.entrySet()) {
                if (curr.equals(entry.getValue())) {
                    keysToReturn.add(entry.getKey());
                }
            }
        }
        return keysToReturn;
    }

    public static <X> List<X> getSortedList(Collection<X> c) {
        ArrayList<X> retVal = new ArrayList<X>(c);
        try {
            Collections.sort((List) retVal);
        } catch (RuntimeException e) {
            // whatever, won't be sorted
        }
        return retVal;
    }

    public static long convertSizeInMBToBytes(Long incoming) {
        if (incoming == null) {
            return 0;
        }
        return incoming.longValue() * 1024 * 1024;
    }

    public static long convertSizeInBytesToMB(Long incoming) {
        return Math.round((double) incoming / 1024 / 1024);
    }

    public static byte[] compressObject(Serializable object) throws Exception {
        ObjectOutputStream oos = null;
        ByteArrayOutputStream baos = new ByteArrayOutputStream();
        try {
            oos = new ObjectOutputStream(new GZIPOutputStream(baos));
            oos.writeObject(object);
        } finally {
            safeFlush(oos);
            safeClose(oos);
        }
        return baos.toByteArray();
    }

    public static Object decompressObject(byte[] bytes) throws Exception {
        ObjectInputStream oos = null;
        ByteArrayInputStream bais = new ByteArrayInputStream(bytes);
        try {
            oos = new ObjectInputStream(new GZIPInputStream(bais));
            return oos.readObject();
        } finally {
            safeClose(oos);
        }
    }

    public static void safeClose(Closeable stream) {
        if (stream != null) {
            try {
                stream.close();
            } catch (IOException e) {
                // ignore
            }
        }
    }

    public static void safeFlush(Flushable stream) {
        if (stream != null) {
            try {
                stream.flush();
            } catch (IOException e) {
                // ignore
            }
        }
    }

    public static void writeObjectToFile(String filename, Serializable object) throws Exception {
        ObjectOutputStream oos = null;
        try {
            oos = new ObjectOutputStream(new FileOutputStream(filename));
            oos.writeObject(object);
        } finally {
            safeFlush(oos);
            safeClose(oos);
        }
    }

    public static Object readObjectFromFile(String filename) throws Exception {
        ObjectInputStream oos = null;
        try {
            oos = new ObjectInputStream(new FileInputStream(filename));
            return oos.readObject();
        } finally {
            safeClose(oos);
        }
    }

    public static void sleep(long l) {
        try {
            Thread.sleep(l);
        } catch (InterruptedException e) {
            // whatever
        }
    }

    public static String join(Collection<String> value, String delimiter) {
        if (value == null || delimiter == null) {
            return "";
        }

        StringBuilder retVal = new StringBuilder();
        for (String s : value) {
            retVal.append(s);
            retVal.append(delimiter);
        }
        if (retVal.length() > delimiter.length()) {
            retVal.setLength(retVal.length() - delimiter.length());
        }
        return retVal.toString();
    }

    public static String join(String[] value, String delimiter) {
        if (value != null) {
            return join(Arrays.asList(value), delimiter);
        }
        return "";
    }

    public static ProcessLaunchResults executeProcess(String[] command, boolean waitForFinish, String textToWaitFor) throws Exception {
        Process p = Runtime.getRuntime().exec(command);
        ProcessLaunchResults plr = new ProcessLaunchResults();
        InputSucker outputStreamSucker = new InputSucker(p.getInputStream(), plr.getOutputLines());
        outputStreamSucker.start();
        InputSucker errorSucker = new InputSucker(p.getErrorStream(), plr.getErrorLines());
        errorSucker.start();
        if (waitForFinish) {
            p.waitFor();
            plr.setExitCode(p.exitValue());
        }
        if (StringUtils.isReal(textToWaitFor)) {
            while (true) {
                Thread.sleep(250);
                if (("" + plr.getOutputLines()).contains(textToWaitFor)) {
                    break;
                }
                if (("" + plr.getErrorLines()).contains(textToWaitFor)) {
                    break;
                }
                try {
                    if (p.exitValue() != 0) {
                        break;
                    }
                } catch (Exception e) {
                    // no problem
                }
            }
        }
        if (waitForFinish) {
            outputStreamSucker.join();
            errorSucker.join();
        }

        plr.setProcess(p);
        return plr;
    }

    public static List<String> getLines(String filename) throws Exception {
        List<String> retVal = new ArrayList<String>();
        FileReader fr = new FileReader(filename);
        BufferedReader br = new BufferedReader(fr);
        String line;
        while ((line = br.readLine()) != null) {
            retVal.add(line);
        }
        br.close();
        fr.close();
        return retVal;
    }

    public static long average(List<? extends Number> values) {
        long total = 0;
        if (values != null) {
            for (Number curr : values) {
                if (curr != null) {
                    total += curr.longValue();
                }
            }
            double d = (double) total / (double) values.size();
            return Math.round(d);
        }
        return 0;
    }

    /**
     * Add a file separator
     */
    public static String addFS(String pathPossiblyWithFileSeparator) {
        if (pathPossiblyWithFileSeparator != null) {
            if (!pathPossiblyWithFileSeparator.endsWith(java.io.File.separator)) {
                pathPossiblyWithFileSeparator += java.io.File.separator;
            }
        }
        return pathPossiblyWithFileSeparator;
    }

    /**
     * Add a file separator
     * @return
     */
    public static String addFS2(String... pathsPossiblyWithFileSeparator) {
        String temp ="";
        for(String pathPossiblyWithFileSeparator : pathsPossiblyWithFileSeparator){
            if (pathPossiblyWithFileSeparator != null) {
                if (!pathPossiblyWithFileSeparator.endsWith(java.io.File.separator)) {
                    pathPossiblyWithFileSeparator += java.io.File.separator;
                    temp = temp + pathPossiblyWithFileSeparator;
                }
            }
        }
        return temp;
    }

    /**
     * Returns the bytes for the given String read from file-read-file
     *
     * @param data the String read from file-read-file
     * @return the bytes
     */
    public static int[] getBytes(String data) {
        int[] bytes = new int[data.length() / 2];

        for (int i = 0, j = 0; i < data.length() / 2; i++, j += 2) {
            bytes[i] = Integer.parseInt(data.substring(j, j + 2), 16);
        }
        return bytes;
    }

    public static String dumpStackTraceForAllThreads() {
        StringBuilder retVal = new StringBuilder();
        retVal.append("Full thread dump:");
        Map<Thread, StackTraceElement[]> stacks = Thread.getAllStackTraces();
        if (stacks != null) {
            for (Map.Entry<Thread, StackTraceElement[]> entry : stacks.entrySet()) {
                retVal.append("\n\n\nStack for ").append(entry.getKey()).append("\n");
                StackTraceElement[] ste = entry.getValue();
                if (ste != null) {
                    for (StackTraceElement aSte : ste) {
                        retVal.append(aSte).append("\n");
                    }
                }
            }
        }
        return retVal.toString();
    }

    public static void dumpStackTrace(String message) {
        try {
            throw new Exception(message + " via Utils.dumpStack()");
        } catch (Exception e) {
            e.printStackTrace();
        }
    }

    /**
     * Converts kb,mb,gb,tb,pb to the multiplier. Used to convert to bytes
     *
     * @param space String Ex "kb","gb"
     * @return multiplier
     * multiplier
     */
    public static Long spaceStringToBytes(String space) {
        if (space.equalsIgnoreCase("kb")) {
            return 1024L;
        } else if (space.equalsIgnoreCase("mb")) {
            return 1024L * 1024L;
        } else if (space.equalsIgnoreCase("gb")) {
            return 1024L * 1024L * 1024L;
        } else if (space.equalsIgnoreCase("tb")) {
            return 1024L * 1024L * 1024L * 1024L;
        } else if (space.equalsIgnoreCase("pb")) {
            return 1024L * 1024L * 1024L * 1024L * 1024L;
        } else {
            //should never be in this case
            return 0L;
        }
    }

    /**
     * This method is hard coded to the method name invokeZAPI.
     * It will return the method following the last invokeZAPI (in case of overload later).
     */
    public static String getCaller(Exception e) {
        String returnMe = null;
        try {
            if (e != null && e.getStackTrace() != null && e.getStackTrace().length > 2) {
                StackTraceElement caller = null;
                boolean grabNext = false;
                for (StackTraceElement ste : e.getStackTrace()) {
                    if ("invokeZAPI".equals(ste.getMethodName())) {
                        grabNext = true;
                    } else {
                        if (grabNext) {
                            caller = ste;
                            break;
                        }
                    }
                }
                if (caller != null) {
                    returnMe = "(called from " + caller.getClassName() + "." + caller.getMethodName() + "(...)" +
                            " on line " + caller.getLineNumber() + ")";
                }
            }
        } catch (Exception e2) {
            //ignore
        }
        return returnMe;
    }

    public static boolean isEmpty(String value) {
        return value == null || value.length() == 0;
    }

    public static NaElement convertToNaElement(Node element) {
        NaElement naElement = new NaElement(element.getNodeName());
        if (!isEmpty(element.getTextContent())) {
            naElement.setContent(element.getTextContent());
        }
        if (element.hasAttributes()) {
            NamedNodeMap attrMap = element.getAttributes();
            for (int i = 0; i < element.getAttributes().getLength(); i++) {
                Node attr = attrMap.item(i);
                naElement.setAttr(attr.getNodeName(), attr.getNodeValue());
            }
        }
        if (element.hasChildNodes()) {
            NodeList childNodes = element.getChildNodes();
            for (int i = 0; i < childNodes.getLength(); i++) {
                Node childNode = childNodes.item(i);
                naElement.addChildElem(convertToNaElement(childNode));
            }
        }
        return naElement;
    }

    public static Document convertToElement(NaElement naElement) {
        Document doc = DOCUMENT_BUILDER.newDocument();
        doc.appendChild(convertToElement(naElement, doc));
        return doc;
    }

    public static Element convertToElement(NaElement naElement, Document doc) {
        Element element = doc.createElement(naElement.getName());
        if (!isEmpty(naElement.getContent())) {
            element.setTextContent(naElement.getContent());
        }
        if (naElement.getAttrNames().size() > 0) {
            for (String name : naElement.getAttrNames()) {
                element.setAttribute(name, naElement.getAttr(name));
            }
        }
        if (naElement.hasChildren()) {
            for (NaElement childNaElement : naElement.getChildren()) {
                element.appendChild(convertToElement(childNaElement, doc));
            }
        }
        return element;
    }

    /**
     * Null safe equals method.
     */
    public static boolean areEqual(Object o1, Object o2) {
        return o1 == null && o2 == null || (o1 != null && o2 != null && o1.equals(o2));
    }

    /**
     * Use System.arraycopy to create a new array based on 2 int arrays.
     */
    public static int[] addAll(int[] array1, int[] array2) {
        if (array1 == null) {
            throw new NullPointerException("First array was null.");
        }
        if (array2 == null) {
            throw new NullPointerException("Second array was null.");
        }

        int[] returnMe = new int[array1.length + array2.length];

        System.arraycopy(array1, 0, returnMe, 0, array1.length);
        System.arraycopy(array2, 0, returnMe, array1.length, array2.length);

        return returnMe;
    }

    public static void createZipFile(List<String> fileNames, String outputFileName, String baseDir) throws IOException {
        ZipOutputStream outStream = null;
        try {

            // Create input and output streams
            outStream = new ZipOutputStream(new FileOutputStream(outputFileName));
            outStream.setLevel(9);

            for (String inputFileName : new HashSet<String>(fileNames)) {
                FileInputStream inStream = null;
                try {
                    inStream = new FileInputStream(inputFileName);

                    // Add a zip entry to the output stream
                    if (baseDir != null && inputFileName.startsWith(baseDir)) {
                        inputFileName = inputFileName.substring(baseDir.length());
                    }
                    ZipEntry ze = new ZipEntry(inputFileName);
                    outStream.putNextEntry(ze);

                    byte[] buffer = new byte[1024];
                    int bytesRead;

                    // Each chunk of data read from the input stream
                    // is written to the output stream
                    while ((bytesRead = inStream.read(buffer)) > 0) {
                        outStream.write(buffer, 0, bytesRead);
                    }
                } catch (Exception e) {
                    //System.err.println("Unable to zip " + inputFileName + " into " + outputFileName + " because of " + e);
                    // lets not let a single file ruin all the rest of our zip.
                } finally {
                    // Close zip entry and file streams
                    try {
                        outStream.closeEntry();
                    } catch (Exception e) {
                        // no problem
                    }
                    try {
                        if (inStream != null) {
                            inStream.close();
                        }
                    } catch (Exception e) {
                        // no problem
                    }
                }
            }

        } catch (IOException ex) {
            ex.printStackTrace();
        } finally {
            if (outStream != null) {
                outStream.close();
            }
        }
    }

    public static <T> List<CallbackResult<T>> executeCallbacksInParallel(List<Callback<T>> callbacks) throws Exception {
        return executeCallbacksInParallel(callbacks, false);
    }

    public static <T> List<CallbackResult<T>> executeCallbacksInParallel(List<Callback<T>> callbacks, boolean throwAllExceptions) throws Exception {
        return executeCallbacksInParallel(callbacks, throwAllExceptions, -1);
    }

    public static <T> List<CallbackResult<T>> executeCallbacksInParallelWithTimeoutEatExceptions(List<Callback<T>> callbacks, long timeoutInMilliseconds) {
        try {
            return executeCallbacksInParallel(callbacks, false, timeoutInMilliseconds);
        } catch (Exception e) {
            e.printStackTrace();
            return Collections.emptyList();
        }
    }

    public static <T> List<CallbackResult<T>> executeCallbacksInParallelWithTimeoutThrowExceptions(List<Callback<T>> callbacks, long timeoutInMilliseconds) throws Exception {
        return executeCallbacksInParallel(callbacks, false, timeoutInMilliseconds);
    }

    private static <T> List<CallbackResult<T>> executeCallbacksInParallelThreadStyle(List<Callback<T>> callbacks, boolean throwAllExceptions, long timeout) throws Exception {
        return executeCallbacksInParallelThreadStyle(callbacks, throwAllExceptions, timeout, false);
    }

    private static <T> List<CallbackResult<T>> executeCallbacksInParallelThreadStyleIncludeExceptionsInReturnedArray(List<Callback<T>> callbacks, long timeout) {
        try {
            return executeCallbacksInParallelThreadStyle(callbacks, false, timeout, true);
        } catch (Exception e) {
            // should never happen - exceptions are inlined
            throw new RuntimeException(e);
        }
    }
    
    public static <T> List<CallbackResult<T>> executeCallbacksInParallelThreadStyle(List<Callback<T>> callbacks, boolean throwAllExceptions, long timeout, boolean inlineExceptions) throws Exception {
    	return executeCallbacksInParallelThreadStyle(callbacks, throwAllExceptions, timeout, inlineExceptions, false);
    }
    
    public static String getStackTrace(StackTraceElement[] ste) {
        StringBuffer retVal = new StringBuffer();
        if (ste != null) {
	        for (int i=0; i<ste.length; i++) {
	            StackTraceElement curr = ste[i];
	            retVal.append(curr.getClassName() + "." + curr.getMethodName() + "() : " + curr.getLineNumber() + "\n");
	        }
        }
        return retVal.toString();
    }

    public static <T> List<CallbackResult<T>> executeCallbacksInParallelThreadStyle(List<Callback<T>> callbacks, boolean throwAllExceptions, long timeout, 
    		boolean inlineExceptions, boolean dumpStackOnTimeout) throws Exception {
        long timeUntilWeAreDone = System.currentTimeMillis() + timeout;
        List<ParallelWorkerThread<T>> launchedThreads = new ArrayList<ParallelWorkerThread<T>>();
        final List<CallbackResult<T>> results = Collections.synchronizedList(new ArrayList<CallbackResult<T>>());
        int desiredPriority = Thread.currentThread().getPriority();
        for (Callback<T> curr : callbacks) {
            ParallelWorkerThread<T> parallelWorkerThread = new ParallelWorkerThread<T>(curr, inlineExceptions);
            parallelWorkerThread.setPriority(desiredPriority);
            parallelWorkerThread.start();
            launchedThreads.add(parallelWorkerThread);
        }
        for (ParallelWorkerThread<T> curr : launchedThreads) {
            long howLongToWait = timeUntilWeAreDone - System.currentTimeMillis();
            if (howLongToWait <= 0) {
                howLongToWait = 1;
            }
            curr.join(howLongToWait);
            CallbackResult<T> o = curr.getReturnValue();
            if (dumpStackOnTimeout && (o == null || (o.getResult() == null && o.getException() == null))) {
            	LOGGER.error("Dumping stack trace for callback that ran with timeout of " + timeout + " but had not completed:\n" + getStackTrace(curr.getStackTrace()));
            }
            results.add(o);
        }
        if (!inlineExceptions) {
            processExceptions(throwAllExceptions, results);
        }
        return results;
    }

    private static <T> CallbackResult<T> executeInParallel(final Callback<T> callback, boolean throwException) throws Exception {
        try {
            String requestor = "";
            if (!ZAPIStatsCounter.getInstance().getRecordedKey().isEmpty()) {
                requestor = ZAPIStatsCounter.getInstance().getRecordedKey() + "";
            }
            // disable max allowable size for this guy
            WhatsHappeningMonitor.getMonitor("parallel-callbacks").setMaxAllowableSize(10000);
            WhatsHappeningMonitor.getMonitor("parallel-callbacks")
                                 .registerStartCall(requestor, callback.getClass().getName());
            return new CallbackResult<T>(callback.call());
        } catch (Exception e) {
            if (throwException) {
                throw e;
            } else {
                return new CallbackResult<T>(e);
            }
        } finally {
            WhatsHappeningMonitor.getMonitor("parallel-callbacks").registerEndCall();
        }
    }

    public static <T> List<CallbackResult<T>> executeCallbacksInParallel(List<Callback<T>> callbacks, boolean throwAllExceptions, long timeout) throws Exception {
        return executeCallbacksInParallel(callbacks, throwAllExceptions, timeout, 100);
    }

    public static <T> List<CallbackResult<T>> executeCallbacksInParallelGuarenteeOrderingOfResponseAndIncludeThrowablesInTheList(List<Callback<T>> callbacks, long timeout) {
        return executeCallbacksInParallelThreadStyleIncludeExceptionsInReturnedArray(callbacks, timeout);
    }

    /**
     * Execute a series of callbacks in parallel - used to process multi-threaded tasks more easily.
     *
     * @throws Exception
     */
    public static <T> List<CallbackResult<T>> executeCallbacksInParallel(List<Callback<T>> callbacks, boolean throwAllExceptions, long timeout, int numberOfThreads) throws Exception {
        if (timeout > 0) {
            // unfortunately the ExecutorCompletionService when executed with a timeout decides to
            // interrupt things that have timed out.  When you are in the middle of (for example)
            // writing to a derby database, this InterruptedException can cause very bad things
            // to happen.  So, if someone wants to run with a callback, we're going to have to
            // manage our own basic threads to cause this to work without the interruption issues.
            return executeCallbacksInParallelThreadStyle(callbacks, throwAllExceptions, timeout);
        }
        long howLongUntilWeGottaBeDone = System.currentTimeMillis() + timeout;
        List<CallbackResult<T>> results = new ArrayList<CallbackResult<T>>();
        final ExecutorService utilsPool;
        utilsPool = Executors.newFixedThreadPool(numberOfThreads, new VscThreadFactory("utilsPool"));
        CompletionService<CallbackResult<T>> completionService = new ExecutorCompletionService<CallbackResult<T>>(utilsPool);

        for (final Callback<T> callback : callbacks) {
            completionService.submit(new OntapCallable<CallbackResult<T>>() {
                public CallbackResult<T> callLocal() throws Exception {
                    return executeInParallel(callback, false);
                }
            });
        }

        for (Callback callback : callbacks) {
            CallbackResult<T> result = null;
            try {
                long howLongUntilWeDemandResult = howLongUntilWeGottaBeDone - System.currentTimeMillis();
                if (howLongUntilWeDemandResult < 0) {
                    howLongUntilWeDemandResult = 1;
                }
                if (timeout < 0) {
                    result = completionService.take().get();
                } else {
                    // the documentation says that you can simply call get with a timeout value,
                    // but that absolutely positively did not work for me in unit tests.  I did find that
                    // poll works reliably, so we'll go with that instead.
                    Future<CallbackResult<T>> future = completionService.poll(howLongUntilWeDemandResult, TimeUnit.MILLISECONDS);
                    if (future != null) {
                        result = future.get();
                    }
                }
            } catch (Exception e) {
                result = new CallbackResult<T>(e);
            }
            if (result != null) {
                results.add(result);
            }
        }

        // We're done, stop any threads that are still going and kill the pool.
        utilsPool.shutdownNow();
        processExceptions(throwAllExceptions, results);
        return results;
    }

    public static class CallbackResult<T> {
        private T result;
        private boolean hasResult;
        private Exception exception;
        private boolean hasException;

        public CallbackResult() {
        }

        public CallbackResult(T result) {
            this.result = result;
            hasResult = true;
        }

        public CallbackResult(Exception exception) {
            this.exception = exception;
            hasException = true;
        }

        public boolean hasResult() {
            return hasResult;
        }

        public T getResult() {
            return result;
        }

        public boolean hasException() {
            return hasException;
        }

        public Exception getException() {
            return exception;
        }

        @Override
        public boolean equals(Object o) {
            if (this == o) {
                return true;
            }
            if (o == null || getClass() != o.getClass()) {
                return false;
            }
            CallbackResult that = (CallbackResult) o;

            return hasResult == that.hasResult &&
                    !(exception != null ? !exception.equals(that.exception) : that.exception != null) &&
                    !(result != null ? !result.equals(that.result) : that.result != null);

        }

        @Override
        public int hashCode() {
            int result1 = result != null ? result.hashCode() : 0;
            result1 = 31 * result1 + (hasResult ? 1 : 0);
            result1 = 31 * result1 + (exception != null ? exception.hashCode() : 0);
            return result1;
        }
    }

    private static <T> void processExceptions(boolean throwAllExceptions, final List<CallbackResult<T>> results) throws Exception {
        List<Exception> exceptions = new ArrayList<Exception>();
        for (CallbackResult<T> result : results) {
            if (result.hasException()) {
                exceptions.add(result.getException());
            }
        }
        if (exceptions.size() > 0) {
            if (exceptions.size() > 1) {
                StringBuilder allExceptions = new StringBuilder(
                        "Parallel callbacks - saw " + exceptions.size() + " exceptions:  ");
                for (Exception e : exceptions) {
                    allExceptions.append(e).append("\n");
                    e.printStackTrace();
                }
                throw new Exception(allExceptions.toString());
            } else {
                throw (exceptions.get(0));
            }
        }
    }

    public interface Callback<T> {
        public T call() throws Exception;
    }

    private static class UICentricComparator implements Comparator {
        private final List<String> UPPER_CASE_UI_IDENTIFIERS = Arrays.asList("UUID", "NAME", "PATH");

        public int compare(Object o1, Object o2) {
            int tentativeResult = 0;
            try {
                tentativeResult = ((Comparable) o1).compareTo(o2);
            } catch (Exception e) {
                // guess it isn't comparable
            }
            if ("NAME".equals(String.valueOf(o1).toUpperCase())) {
                return -2;
            }
            if ("NAME".equals(String.valueOf(o2).toUpperCase())) {
                return 2;
            }
            if (UPPER_CASE_UI_IDENTIFIERS.contains(String.valueOf(o1).toUpperCase())) {
                return -1;
            }
            if (UPPER_CASE_UI_IDENTIFIERS.contains(String.valueOf(o2).toUpperCase())) {
                return 1;
            }
            return tentativeResult;
        }

    }

    private static class InputSucker extends Thread {
        InputStream stream;
        List<String> destination;

        InputSucker(InputStream stream, List<String> destination) {
            this.stream = stream;
            this.destination = destination;
        }

        public void run() {
            BufferedReader bis = new BufferedReader(new InputStreamReader(stream));
            while (true) {
                try {
                    String s = bis.readLine();
                    if (s == null) {
                        break;
                    }
                    destination.add(s);
                } catch (Exception e) {
                    // whatever
                }
            }
        }

    }

    private static class ParallelWorkerThread<T> extends Thread {
        Callback<T> callback;
        CallbackResult<T> returnValue = new CallbackResult<T>();
        private OntapThreadSpecificInfo threadInfo;
        private boolean wantsExceptionThrown;
        // any changes here need to go into NtapCallable too

        ParallelWorkerThread(Callback<T> callback, boolean wantsExceptionThrown) {
            this.callback = callback;
            threadInfo = new OntapThreadSpecificInfo();
            this.wantsExceptionThrown = wantsExceptionThrown;
        }

        public void run() {
            threadInfo.applyToCurrentThread();
            try {
                returnValue = executeInParallel(callback, wantsExceptionThrown);
            } catch (Exception e) {
                returnValue = new CallbackResult<T>(e);
                if ((e instanceof RuntimeException)) {
                    throw (RuntimeException) e;
                } else {
                    throw new RuntimeException(e);
                }
            }
        }

        public CallbackResult<T> getReturnValue() {
            return returnValue;
        }
    }

    public static class VscThreadFactory implements ThreadFactory {
        private static final AtomicInteger poolIndex = new AtomicInteger(1);
        private int myPoolIndex;
        private final ThreadGroup group;
        private final AtomicInteger threadIndex = new AtomicInteger(1);
        private String threadName;
        private String myPoolName;

        public VscThreadFactory(String poolName) {
            SecurityManager s = System.getSecurityManager();
            group = (s != null) ? s.getThreadGroup() : Thread.currentThread().getThreadGroup();
            myPoolIndex = poolIndex.getAndIncrement();
            myPoolName = poolName;
        }

        public Thread newThread(Runnable r) {
        	threadName = myPoolName + "-" + myPoolIndex + "-thread-" + threadIndex.getAndIncrement();
            Thread t = new Thread(group, r, threadName, 0);
            if (t.isDaemon()) {
                t.setDaemon(false);
            }
            t.setPriority(Thread.currentThread().getPriority());

            return t;
        }
    }
}
