package com.netapp.offtap3.utils;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.Date;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import com.netapp.offtap3.internal.ontap.utils.StatsCollector;

/**
 * Class that manages some information about what is happening RIGHT NOW
 * Intended for a page that shows a window into what threads are running, what they
 * are doing, etc, for debuggability and supportability
 */
public class WhatsHappeningMonitor {
    private Map<Thread, String> currentlyInFlightThreads = new SafeConcurrentHashMap<Thread, String>();
    private Map<Thread, Date> launchTime = new SafeConcurrentHashMap<Thread, Date>();
    private Map<Thread, String> requestors = new SafeConcurrentHashMap<Thread, String>();
    private String name;
    private int maxSize = 0;
    private Date lastLaunchDate;
    protected long somethingStartedRunningTime = 0; 
    private int maxAllowableSize = 100;
    private static Map<String, WhatsHappeningMonitor> whoAreWeWatching = new SafeConcurrentHashMap<String, WhatsHappeningMonitor>();
    
    protected WhatsHappeningMonitor(String name) {
        this.name = name;
    }
    
    public void registerStartCall(String requestor, String name) {
        if (name == null) {
            name = "unknown";
        }
        if (requestor == null) {
            requestor = "unknown";
        }
        if (currentlyInFlightThreads.size() == 0) {
        	somethingStartedRunningTime = System.currentTimeMillis();
        }
        synchronized (this) {
	        int currSize = currentlyInFlightThreads.size();
	        if (currSize >= maxAllowableSize) {
	        	// we're gonna toss an error.  Lets say what exactly was running so it can be obvious who took up all the run slots.
	        	Map<String, Integer> whoIsRunning = new HashMap<String, Integer>();
	        	for (String currRequestor : this.currentlyInFlightThreads.values()) {
	        		Integer currentCount = whoIsRunning.get(currRequestor);
	        		if (currentCount == null) {
	        			currentCount = new Integer(0);
	        		}
	        		whoIsRunning.put(currRequestor, currentCount+1);
	        	}
	        	List<String> whoWasRunningFragments = new ArrayList<String>();
	        	List<Integer> vals = Utils.getSortedList(whoIsRunning.values());
	        	Collections.reverse(vals);
	        	for (Integer val : vals) {
	        		for (String key : Utils.getSortedList(whoIsRunning.keySet())) {
	        			if (whoIsRunning.get(key).equals(val)) {
	        				whoWasRunningFragments.add(key + ": " + val + " instances");
	        			}
	        		}
	        	}
	        	
	        	throw new RuntimeException("Request of " + name + " from " + requestor + " exceeds maximum allowable concurrent run limit of " + maxAllowableSize + 
	        			".  Currently running saw " + Utils.join(whoWasRunningFragments, ", "));
	        }
	        currentlyInFlightThreads.put(Thread.currentThread(), name);
	        Date now = new Date();
	        lastLaunchDate = now;
	        launchTime.put(Thread.currentThread(), now);
	        requestors.put(Thread.currentThread(), requestor);
	        currSize = currentlyInFlightThreads.size();
	        if (currSize > maxSize) {
	            maxSize = currSize;
	        }        
        }
    }
    
    public void updateName(String newName) {
    	currentlyInFlightThreads.put(Thread.currentThread(), newName);
    }
    
    public void registerEndCall() {
        currentlyInFlightThreads.remove(Thread.currentThread());
        launchTime.remove(Thread.currentThread());
        requestors.remove(Thread.currentThread());
        synchronized (this) {
	        long howLongWasSomethingRunning = System.currentTimeMillis() - somethingStartedRunningTime;
	        if (currentlyInFlightThreads.size() == 0) {        	
	        	if (howLongWasSomethingRunning > 0 && somethingStartedRunningTime > 0) {
	        		StatsCollector.getInstance().addCounter("wall-clock-" + name, howLongWasSomethingRunning);
	        	}
	        } else {
	        	// something was still running.  in performance and longevity testing, we sometimes have something running
	        	// basically forever.  So lets see if this is above a certain value, and if so, lets carve it off.
	        	if (howLongWasSomethingRunning > 300000) {
	        		StatsCollector.getInstance().addCounter("wall-clock-" + name, howLongWasSomethingRunning);
	        		somethingStartedRunningTime = System.currentTimeMillis();
	        	}
	        }
        }
    }
    
    public Collection<Thread> getCurrentlyInFlightThreads() {
        return currentlyInFlightThreads.keySet();
    }
    
    public Date getLaunchTime(Thread client) {
        return launchTime.get(client);
    }
    
    public String getRequestor(Thread client) {
        return requestors.get(client);
    }
    
    public String getValue(Thread client) {
        return currentlyInFlightThreads.get(client);
    }

    public Date getLastLaunchDate() {
        return lastLaunchDate;
    }

    public String getName() {
        return name;
    }

    public int getMaxSize() {
        return maxSize;
    }

    public static synchronized WhatsHappeningMonitor getMonitor(String identifyingName) {
        WhatsHappeningMonitor retVal = whoAreWeWatching.get(identifyingName);
        if (retVal == null) {
            retVal = new WhatsHappeningMonitor(identifyingName);
            whoAreWeWatching.put(identifyingName, retVal);
        }
        return retVal;
    }
    
    public static Collection<WhatsHappeningMonitor> getAllMonitors() {
        return whoAreWeWatching.values();
    }

	public int getMaxAllowableSize() {
		return maxAllowableSize;
	}

	public void setMaxAllowableSize(int maxAllowableSize) {
		this.maxAllowableSize = maxAllowableSize;
	}
}
