/*
 * $Id:  $
 *
 * Copyright (c) 2015 NetApp, Inc.
 * All rights reserved.
 */
package com.netapp.security;

import java.io.IOException;
import java.net.InetAddress;
import java.net.Socket;
import java.security.AlgorithmConstraints;
import java.security.KeyManagementException;
import java.security.NoSuchAlgorithmException;
import java.security.Provider;
import java.security.SecureRandom;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;
import javax.net.ssl.HostnameVerifier;
import javax.net.ssl.KeyManager;
import javax.net.ssl.SSLContext;
import javax.net.ssl.SSLParameters;
import javax.net.ssl.SSLSocket;
import javax.net.ssl.SSLSocketFactory;
import javax.net.ssl.TrustManager;

/**
 * TODO Move to a common security library
 */
public class ConfigurableSecurityContext implements SecurityContext {

    private HostnameVerifier hostnameVerifier = new TrustAllHostnameVerifier(); // by default, trust hostnames
    private String protocolVersion = "TLS";
    private Provider provider;

    private AlgorithmConstraints algorithmConstraints;
    private List<String> cipherSuites;
    private String endpointIdentificationAlgorithm;
    private List<String> supportedProtocols;
    private List<String> defaultSupportedProtocols = Arrays.asList("TLSv1", "TLSv1.1", "TLSv1.2");

    private List<KeyManager> keyManagers;
    private List<TrustManager> trustManagers;
    private SecureRandom secureRandom;

    private SSLSocketFactory sslSocketFactory;


    @Override
    public HostnameVerifier getHostnameVerifier() {
        return hostnameVerifier;
    }

    @Override
    public SSLSocketFactory getSslSocketFactory() {
        synchronized (this) {
            if (sslSocketFactory == null) {
                try {
                    SSLContext sc = provider == null ? SSLContext.getInstance(protocolVersion) : SSLContext.getInstance(protocolVersion, provider);
                    sc.init(buildKeyManagerArray(), buildTrustManagerArray(), secureRandom);

                    final SSLSocketFactory delegateSslSocketFactory = sc.getSocketFactory();
                    sslSocketFactory = new SSLSocketFactory() {
                        @Override
                        public String[] getDefaultCipherSuites() {
                            return delegateSslSocketFactory.getDefaultCipherSuites();
                        }

                        @Override
                        public String[] getSupportedCipherSuites() {
                            return delegateSslSocketFactory.getSupportedCipherSuites();
                        }

                        @Override
                        public Socket createSocket(Socket socket, String host, int port, boolean autoClose) throws IOException {
                            return decorateSslSocket((SSLSocket) delegateSslSocketFactory.createSocket(socket, host, port, autoClose));
                        }

                        @Override
                        public Socket createSocket(String host, int port) throws IOException {
                            return decorateSslSocket((SSLSocket) delegateSslSocketFactory.createSocket(host, port));
                        }

                        @Override
                        public Socket createSocket(String host, int port, InetAddress clientAddress, int clientPort) throws IOException {
                            return decorateSslSocket((SSLSocket) delegateSslSocketFactory.createSocket(host, port, clientAddress, clientPort));
                        }

                        @Override
                        public Socket createSocket(InetAddress address, int port) throws IOException {
                            return decorateSslSocket((SSLSocket) delegateSslSocketFactory.createSocket(address, port));
                        }

                        @Override
                        public Socket createSocket(InetAddress address, int port, InetAddress clientAddress, int clientPort) throws IOException {
                            return decorateSslSocket((SSLSocket) delegateSslSocketFactory.createSocket(address, port, clientAddress, clientPort));
                        }

                        private SSLSocket decorateSslSocket(SSLSocket sslSocket) {
                            String[] cipherSuites = buildStringArray(getCipherSuites());
                            String[] protocols = buildStringArray(getSupportedProtocols() == null ? defaultSupportedProtocols : getSupportedProtocols());

                            SSLParameters sslParameters = new SSLParameters();
                            sslParameters.setAlgorithmConstraints(getAlgorithmConstraints());

                            if (cipherSuites != null) {
                                sslParameters.setCipherSuites(cipherSuites);
                            }

                            sslParameters.setEndpointIdentificationAlgorithm(getEndpointIdentificationAlgorithm());

                            if (protocols != null) {
                                sslParameters.setProtocols(protocols);
                            }

                            sslSocket.setSSLParameters(sslParameters);

                            return sslSocket;
                        }
                    };
                } catch (NoSuchAlgorithmException e) {
                    throw new RuntimeException(e);
                } catch (KeyManagementException e) {
                    throw new RuntimeException(e);
                }
            }
        }

        return sslSocketFactory;
    }

    private String[] buildStringArray(List<String> strings) {
        return strings == null ? null : strings.toArray(new String[strings.size()]);
    }

    private KeyManager[] buildKeyManagerArray() {
        return keyManagers == null ? null : keyManagers.toArray(new KeyManager[keyManagers.size()]);
    }


    private TrustManager[] buildTrustManagerArray() {
        return trustManagers == null ? null : trustManagers.toArray(new TrustManager[trustManagers.size()]);
    }


    // fluent api
    public ConfigurableSecurityContext algorithmConstraints(AlgorithmConstraints algorithmConstraints) {
        setAlgorithmConstraints(algorithmConstraints);
        return this;
    }

    public ConfigurableSecurityContext cipherSuite(String cipherSuite) {
        addCipherSuite(cipherSuite);
        return this;
    }

    public ConfigurableSecurityContext cipherSuites(List<String> cipherSuites) {
        setCipherSuites(cipherSuites);
        return this;
    }

    public ConfigurableSecurityContext endpointIdentificationAlgorithm(String endpointIdentificationAlgorithm) {
        setEndpointIdentificationAlgorithm(endpointIdentificationAlgorithm);
        return this;
    }

    public ConfigurableSecurityContext supportedProtocol(String protocol) {
        if (supportedProtocols == null) {
            supportedProtocols = new ArrayList<String>();
        }

        supportedProtocols.add(protocol);
        return this;
    }

    public ConfigurableSecurityContext supportedProtocols(List<String> protocols) {
        setSupportedProtocols(protocols);
        return this;
    }

    public ConfigurableSecurityContext hostnameVerifier(HostnameVerifier hostnameVerifier) {
        setHostnameVerifier(hostnameVerifier);
        return this;
    }

    public ConfigurableSecurityContext keyManager(KeyManager keyManager) {
        addKeyManager(keyManager);
        return this;
    }

    public ConfigurableSecurityContext keyManagers(List<KeyManager> keyManagers) {
        setKeyManagers(keyManagers);
        return this;
    }

    public ConfigurableSecurityContext protocol(String protocol) {
        setProtocolVersion(protocol);
        return this;
    }

    public ConfigurableSecurityContext provider(Provider provider) {
        setProvider(provider);
        return this;
    }

    public ConfigurableSecurityContext secureRandom(SecureRandom secureRandom) {
        setSecureRandom(secureRandom);
        return this;
    }

    public ConfigurableSecurityContext trustManager(TrustManager trustManager) {
        addTrustManager(trustManager);
        return this;
    }

    public ConfigurableSecurityContext trustManagers(List<TrustManager>trustManagers) {
        setTrustManagers(trustManagers);
        return this;
    }


    // field accessors
    public AlgorithmConstraints getAlgorithmConstraints() {
        return algorithmConstraints;
    }

    public void setAlgorithmConstraints(AlgorithmConstraints algorithmConstraints) {
        this.algorithmConstraints = algorithmConstraints;
    }

    public List<String> getCipherSuites() {
        return cipherSuites;
    }

    public void setCipherSuites(List<String> cipherSuites) {
        this.cipherSuites = cipherSuites;
    }

    private void addCipherSuite(String cipherSuite) {
        if (cipherSuites == null) {
            cipherSuites = new ArrayList<String>();
        }

        cipherSuites.add(cipherSuite);
    }

    public String getEndpointIdentificationAlgorithm() {
        return endpointIdentificationAlgorithm;
    }

    public void setEndpointIdentificationAlgorithm(String endpointIdentificationAlgorithm) {
        this.endpointIdentificationAlgorithm = endpointIdentificationAlgorithm;
    }

    public List<String> getSupportedProtocols() {
        return supportedProtocols;
    }

    public void setSupportedProtocols(List<String> supportedProtocols) {
        this.supportedProtocols = supportedProtocols;
    }

    public void setHostnameVerifier(HostnameVerifier hostnameVerifier) {
        this.hostnameVerifier = hostnameVerifier;
    }

    public List<KeyManager> getKeyManagers() {
        return keyManagers;
    }

    public void setKeyManagers(List<KeyManager> keyManagers) {
        this.keyManagers = keyManagers;
    }

    private void addKeyManager(KeyManager keyManager) {
        if (keyManagers == null) {
            keyManagers = new ArrayList<KeyManager>();
        }

        keyManagers.add(keyManager);
    }

    public String getProtocolVersion() {
        return protocolVersion;
    }

    public void setProtocolVersion(String protocolVersion) {
        this.protocolVersion = protocolVersion;
    }

    public Provider getProvider() {
        return provider;
    }

    public void setProvider(Provider provider) {
        this.provider = provider;
    }

    public SecureRandom getSecureRandom() {
        return secureRandom;
    }

    public void setSecureRandom(SecureRandom secureRandom) {
        this.secureRandom = secureRandom;
    }

    public List<TrustManager> getTrustManagers() {
        return trustManagers;
    }

    public void setTrustManagers(List<TrustManager> trustManagers) {
        this.trustManagers = trustManagers;
    }

    private void addTrustManager(TrustManager trustManager) {
        if (trustManagers == null) {
            trustManagers = new ArrayList<TrustManager>();
        }

        trustManagers.add(trustManager);
    }
}
