package com.netapp.vee.security;

import java.io.File;
import java.io.IOException;
import java.security.KeyPair;
import java.security.KeyStore;
import java.security.KeyStoreException;
import java.security.PrivateKey;
import java.security.Security;
import java.security.cert.CertPathBuilder;
import java.security.cert.CertPathBuilderResult;
import java.security.cert.CertStore;
import java.security.cert.CertStoreParameters;
import java.security.cert.CollectionCertStoreParameters;
import java.security.cert.PKIXBuilderParameters;
import java.security.cert.TrustAnchor;
import java.security.cert.X509CRL;
import java.security.cert.X509CertSelector;
import java.security.cert.X509Certificate;
import java.util.Collection;
import java.util.Enumeration;
import java.util.HashSet;
import java.util.Set;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.bouncycastle.asn1.x500.X500Name;
import org.bouncycastle.pkcs.PKCS10CertificationRequest;


/**
 * The class is a singleton that provides a front-end
 * to general certificate and security functions.
 * 
 * @author cliftonj
 *
 */
public class CertSecurityMgr implements SecurityConstants {
    private static Log log = LogFactory.getLog(CertSecurityMgr.class);

	/**
	 * Debug logging - wrap Dev-ony code in this variable
	 */
	private static final boolean DEBUG = false;
	
	/**
	 * No duplicate trust certificates allowed.
	 */
	private static final boolean NO_DUPLICATES = true;
    public static String CRL_SUFFIX = "_CRL"; // File suffix for CRLs
	
	private KeyStore keyStore = null;
	private String keyStorePath = null;
	private String keyStorePassword = null;
	private Collection<X509CRL> crls = null; // Currently non-persistent
	private String ownCertAlias = DEFAULT_OWN_CERT_ALIAS;
	

    /**
     * Thread-safe singleton
     */
    private static class Singleton {
        static final CertSecurityMgr INSTANCE = new CertSecurityMgr();
    }
    
    /**
     * Get SecurityMgr instance.
     * 
     * @return The SecurityMgr
     */
    public static CertSecurityMgr getInstance() {
        //      We may need to manage multiple key store sessions.
        //      But for now we are happily a singleton.
        return Singleton.INSTANCE;
    }
    
    private CertSecurityMgr() {
        Security.addProvider(new org.bouncycastle.jce.provider.BouncyCastleProvider());
        
        if (DEBUG) {
            // This is to make sure that we don't leave DEBUG enabled
            log.warn("WARNING: Developer (NON-PRODUCTION) debug enabled!", new RuntimeException());
        }
    }
    
	/**
	 * Open existing or optionally create key store file.
	 * 
	 * @param ksPath File path to key store.
	 * @param ksPassword Key store password.
	 * @param create <code>true</code> to create if keystore doesn't exist.
	 * @return KeyStore instance on success.
	 * @throws KeyStoreMgrException on failure
	 * @throws CertUtilsException on failure
	 */
	public KeyStore openKeyStoreSession(String ksPath, String ksPassword, boolean create)
	        throws KeyStoreMgrException, CertUtilsException {
	    
	    keyStorePath = ksPath;
	    keyStorePassword = ksPassword;
	    File ksFile = new File(ksPath);
	    
	    if (ksFile.exists()) {
	        keyStore = KeyStoreMgr.loadKeyStoreFile(ksPath, ksPassword);
	    } else if (create) {
	        keyStore = KeyStoreMgr.getNewKeyStore();
	        // Write to the file
	        flushKeyStoreSession();
	    } else {
            KeyStoreMgr.logAndThrow("Key store path doesn't exist: " + ksPath, null);
	    }
	    
	    // Load CRL file.
	    // TODO - Re-enenable CRL encryption code - Clif
	    //        Need encrypt/decrypt code
	    String crlPath = ksPath + CRL_SUFFIX;
	    File crlFile = new File(crlPath);
	    if (crlFile.exists()) {
	        // We stored some CRLs
	        String crlPEMs;
            try {
                crlPEMs = CryptUtils.readEncryptedFile(crlPath);
            }
            catch (IOException e) {
                throw new CertUtilsException("Error loading CRL file", e);
            }
            setCRLs(CertUtils.importPEMCRL(crlPEMs));
	    } else {
	        // No CRLs
	        setCRLs(null);;
	    }
	    
	    return keyStore;
	}
	
	/**
	 * Write current key store session to disk.
	 * This call is valid between an openKeyStore and closeKeyStoreSession calls.
	 * 
	 * @throws KeyStoreMgrException On error.
	 * @throws CertUtilsException On error.
	 */
	public void flushKeyStoreSession()
	        throws KeyStoreMgrException, CertUtilsException {
	    if (keyStore != null) {
	        if (keyStorePath != null || keyStorePath.length() == 0) {
	            KeyStoreMgr.saveKeyStoreFile(keyStore, keyStorePath, keyStorePassword);
	        } else {
                KeyStoreMgr.logAndThrow("Key store path empty", null);
	        }
	        
	        // Save CRLs
            String crlPath = keyStorePath + CRL_SUFFIX;
            
            // TODO Re-enable CRLs - Clif
            // Need encrypt/decrypt file.
    	    if (SecurityUtils.isReal(crls)) {
    	        // Persist our CRL list
    	        String crlPems = CertUtils.exportCRLToPEM(crls);
    	        try {
                    CryptUtils.writeEncryptedFile(crlPems, crlPath);
                }
                catch (IOException e) {
                    throw new CertUtilsException("Error saving CRL file", e);
                }
    	    } else {
    	        // Delete existing CRLs
    	        File crlF = new File(crlPath);
    	        if (crlF.exists()) {
    	            if(!crlF.delete()) {
    	                throw new CertUtilsException("Failed to delete CRL file");
    	            }
    	        }
    	    }
	    }
	}
	
	/**
	 * End key store session and optionally write remaining changes to disk.
	 * Key store information is invalid after this call.
	 * 
	 * @param flush <code>true</code> to write changes to disk.
	 * 
	 * @throws KeyStoreMgrException On error.
	 * @throws CertUtilsException On error.
	 */
	public void closeKeyStoreSession(boolean flush)
	        throws KeyStoreMgrException, CertUtilsException {
	    if (flush) {
	        flushKeyStoreSession();
	    }
	    
	    keyStore = null;
	    keyStorePath = null;
	    keyStorePassword = null;
	    crls = null;
	}
    
    /**
     * @return the keyStore
     */
    public KeyStore getKeyStore() {
        return keyStore;
    }

    /**
     * @param keyStore the keyStore to set
     */
    protected void setKeyStore(KeyStore keyStore) {
        this.keyStore = keyStore;
    }

    /**
     * @return the keyStorePath
     */
    public String getKeyStorePath() {
        return keyStorePath;
    }

    /**
     * @param keyStorePath the keyStorePath to set
     */
    public void setKeyStorePath(String keyStorePath) {
        this.keyStorePath = keyStorePath;
    }

    /**
     * @return the keyStorePassword
     */
    public String getKeyStorePassword() {
        return keyStorePassword;
    }

    /**
     * @param keyStorePassword the keyStorePassword to set
     */
    public void setKeyStorePassword(String keyStorePassword) {
        this.keyStorePassword = keyStorePassword;
    }
    
    /**
     * Check the certificate security manager session status.
     * 
     * @return <code>true</code> if session is open.
     */
    public boolean isSessionOpen() {
        return (keyStore != null);
    }
    
    private final void checkKeyStoreSession() throws KeyStoreMgrException {
        if (!isSessionOpen()) {
            KeyStoreMgr.logAndThrow("Key store session not open", null);
        }
    }
    
    /**
     * Initialize our own private key and certificate.
     * The old private key and certificate is replaced.
     * The self-signed certificate is NOT a CA.
     * 
     * @param subjectDN Subject name for generated certificate
     * @return Newly generated own certificate.
     * @throws KeyUtilsException On key generation error.
     * @throws CertUtilsException On certificate generation error.
     * @throws KeyStoreMgrException On key/certificate storage error.
     */
    public X509Certificate initOwnCertificate(String subjectDN)
            throws KeyUtilsException, CertUtilsException, KeyStoreMgrException {
        return initOwnCertificate(subjectDN, null);
    }
    
    /**
     * Initialize our own private key and certificate.
     * The old private key and certificate is replaced.
     * The self-signed certificate is NOT a CA.
     * 
     * @param subjectDN Subject name for generated certificate
     * @param newOwnAlias Use the non-default alias for the own certificate.
     * @return Newly generated own certificate.
     * @throws KeyUtilsException On key generation error.
     * @throws CertUtilsException On certificate generation error.
     * @throws KeyStoreMgrException On key/certificate storage error.
     */
    public X509Certificate initOwnCertificate(String subjectDN, String newOwnAlias)
            throws KeyUtilsException, CertUtilsException, KeyStoreMgrException {
        return initOwnCertificate(subjectDN, newOwnAlias, false);
    }
    
    /**
     * Initialize our own private key and certificate.
     * The old private key and certificate is replaced.
     * 
     * @param subjectDN Subject name for generated certificate
     * @param newOwnAlias Use the non-default alias for the own certificate.
	 * @param isCA <code>true</code> if the signed certificate will be a certificate authority.
     * @return Newly generated own certificate.
     * @throws KeyUtilsException On key generation error.
     * @throws CertUtilsException On certificate generation error.
     * @throws KeyStoreMgrException On key/certificate storage error.
     */
    public X509Certificate initOwnCertificate(String subjectDN, String newOwnAlias, boolean isCA)
            throws KeyUtilsException, CertUtilsException, KeyStoreMgrException {
        X509Certificate ret = null;
         
        // Sanity check
        checkKeyStoreSession();
        
        KeyPair kp = KeyUtils.genKeyPair();

        ret = CertUtils.genSelfSignedCert(kp, subjectDN, isCA);
        X509Certificate chain[] = new X509Certificate[1];
        
        try {
            if (keyStore.containsAlias(ownCertAlias)) {
                log.debug("Found existing own cert entry...deleting alias: " + ownCertAlias);
                if (DEBUG) {
                    log.debug("Old cert entry: " + keyStore.getCertificate(ownCertAlias));
                }
                keyStore.deleteEntry(ownCertAlias);
            }
            
            chain[0] = ret;
            if (DEBUG) {
                log.debug("Saving new self-signed cert for private key=" + kp.getPrivate());
            }
            
            if (newOwnAlias != null && newOwnAlias.length() > 0) {
                ownCertAlias = newOwnAlias;
            }
            
            keyStore.setKeyEntry(ownCertAlias, kp.getPrivate(), keyStorePassword.toCharArray(), chain);
            log.debug("Saved new self-signed own cert under alias: " + ownCertAlias + ", cert=\n" + ret);
        }
        catch (Exception e) {
            KeyStoreMgr.logAndThrow("Error storing own certificate in key store.", e);
        }
        
        return ret;
    }
    
    /**
     * Get our own certificate.
     * 
     * @return Get our own certificate or <code>null</code> if not set.
      * @throws KeyStoreMgrException On error.
     */
    public X509Certificate getOwnCertificate()
            throws KeyStoreMgrException {
        X509Certificate ret = null;
        
        // Sanity check
        checkKeyStoreSession();
        
        try {
            if (keyStore.containsAlias(ownCertAlias)) {
                log.debug("Found existing own cert entry");
                
                ret = (X509Certificate) keyStore.getCertificate(ownCertAlias);
                
                if (DEBUG) {
                    log.debug("Own cert entry: " + ret);
                }
            }
        }
        catch (Exception e) {
            KeyStoreMgr.logAndThrow("Error retrieving own certificate from key store.", e);
        }
        
        return ret;
    }
    
    /**
     * Test own certificate info for initialization.
     * 
     * @return <code>true</code> if own certificate is initialized.
     */
    public boolean isOwnCertificateInitialized() {
        boolean ret = false;
        
        try {
            X509Certificate ownCert = getOwnCertificate();
            PrivateKey privKey = (PrivateKey)keyStore.getKey(ownCertAlias, keyStorePassword.toCharArray());
            if (ownCert != null && privKey != null) {
                ret = true;
            }
        }
        catch (Exception e) {
            // ignore
            if (DEBUG) {
                log.debug("Exception on isOwnCertificateIntialized", e);
            }
        }
        
        return ret;
    }
    
    /**
     * Set our own certificate.
     * You can only set your own certificate if an own key pair exists
     * and the public key verifies the given certificate.  The normal
     * steps to get here are to generate a CSR for the own certificate,
     * sign it with a CA and then replace the existing own certificate
     * with the newly signed one.
     *
     * @param ownCert New own certificate.
     * @throws KeyStoreMgrException On error.
     * @throws CertUtilsException On error.
     */
    public void setOwnCertificate(X509Certificate ownCert)
            throws KeyStoreMgrException, CertUtilsException {
        // Sanity checks
        if (ownCert == null) {
            KeyStoreMgr.logAndThrow("Null input certificate", null);
        }
        checkKeyStoreSession();
        if (!isOwnCertificateInitialized()) {
            KeyStoreMgr.logAndThrow("Own certificate info not initialized", null);
        }
        
        X509Certificate origOwnCert = getOwnCertificate();
        if (origOwnCert.equals(ownCert)) {
            // Same certificate, No-Op
            return;
        }
        
        // Check for our public key
        if (!origOwnCert.getPublicKey().equals(ownCert.getPublicKey())) {
            if (DEBUG) {
                log.error("Set own certificate failed: public key mismatch");
            }
            KeyStoreMgr.logAndThrow("Set own certificate verification failed", null);
        }
        
        // Make sure that we are not self-signed
        if (CertUtils.isSelfSigned(ownCert)) {
            if (DEBUG) {
                log.error("Set own certificate failed: self-signed");
            }
            KeyStoreMgr.logAndThrow("Set own certificate verification failed", null);
        }
        
        // Verify trust of our certificate
        verifyCertificate(ownCert);
        
        // Looks good. We'll take it.
        PrivateKey privKey = null;
        try {
            privKey = (PrivateKey)keyStore.getKey(ownCertAlias, keyStorePassword.toCharArray());
            keyStore.deleteEntry(ownCertAlias);
            X509Certificate chain[] = new X509Certificate[1];
            chain[0] = ownCert;
            keyStore.setKeyEntry(ownCertAlias, privKey, keyStorePassword.toCharArray(), chain);
        }
        catch (Exception e) {
            KeyStoreMgr.logAndThrow("Error setting own certificate", e);
        }
    }
    
    /**
     * Generate a certificate signing request (CSR) to sign our own certificate.
     * 
     * @return Generated CSR for our own certificate.
     * @throws KeyStoreMgrException On error.
     * @throws CertUtilsException On error.
     */
    public PKCS10CertificationRequest getOwnCSR()
            throws KeyStoreMgrException {
        PKCS10CertificationRequest ret = null;
        // Sanity check
        checkKeyStoreSession();
        X509Certificate ownCert = getOwnCertificate();
        PrivateKey privKey = null;
        try {
            privKey = (PrivateKey)keyStore.getKey(ownCertAlias, keyStorePassword.toCharArray());
            KeyPair kp = new KeyPair(ownCert.getPublicKey(), privKey);
            ret = CertUtils.genCSR(kp, new X500Name(ownCert.getSubjectDN().getName()), CertUtils.isCA(ownCert));
        }
        catch (Exception e) {
            KeyStoreMgr.logAndThrow("Error building CSR", e);
        }
        
        return ret;
    }
    
    /**
     * Add the CA certificate to the trust store under the desired alias or pattern.
     * 
     * @param trustCert Trusted CA cert.
     * @param alias Alias or alias pattern.
     * @param findUniqueAlias <code>true</code> to use alias as a pattern for uniqueness.
     * @return Alias used to store the trusted certificate.
     * @throws KeyStoreMgrException
     */
    public String addTrustCertificate(X509Certificate trustCert, String alias, boolean findUniqueAlias)
            throws KeyStoreMgrException {
        // Sanity check
        checkKeyStoreSession();
        
        if (NO_DUPLICATES) {
            if (trustCertificateExists(trustCert)) {
                KeyStoreMgr.logAndThrow("Certificate already exists in trust store", null);
            }
        }
        
        String uniqueAlias = KeyStoreMgr.findUnusedAlias(keyStore, alias);
        if (!findUniqueAlias && !uniqueAlias.equals(alias)) {
            // We wanted an exact alias and it is in use.
            KeyStoreMgrException ex = new KeyStoreMgrException("Cannot add trust certificate under in-use alias: " + alias);
            log.error(ex.getMessage(), ex);
            throw ex;
        }
        
        try {
            keyStore.setCertificateEntry(uniqueAlias, trustCert);
        }
        catch (KeyStoreException e) {
            KeyStoreMgr.logAndThrow("Error adding trust certificate to key store.", e);
        }
        
        return uniqueAlias;
    }
    
    /**
     * Remove a trust certificate from the trust store.
     * 
     * @param trustCert Certificate to remove.
     * @return <code>true</code> if the trust certificate was found.
     * @throws KeyStoreMgrException On error.
     */
    public boolean removeTrustCertificate(X509Certificate trustCert)
            throws KeyStoreMgrException {
        // Sanity check
        checkKeyStoreSession();
        boolean found = false;
        log.debug("remoteTrustCertificate: " + trustCert);
        
        try {
            String alias = keyStore.getCertificateAlias(trustCert);
            // Our own certificate show
            if (alias != null && alias.equals(ownCertAlias)) {
                // This certificate is also our own cert so a comprehensive search
                // must occur.
                // Note: This appears to be unreachable code on the JKS but reading the spec,
                //       I'm leaving the logic in place just in case.
                log.debug("Trust cert being deleted matches own cert");
                Enumeration<String> aliases = keyStore.aliases();
                while (aliases.hasMoreElements()) {
                    alias = aliases.nextElement();
                    if (alias.equals(ownCertAlias) || !keyStore.getCertificate(alias).equals(trustCert)) {
                        if (DEBUG) {
                            log.debug("Skipping non-matching trust cert alias: " + alias);
                        }
                        continue;
                    }
                    found = true;
                    log.debug("Deleting trust certificate alias entry: " + alias);
                    keyStore.deleteEntry(alias);                   
                }
            } else {
                // Fast delete
                log.debug("Fast delete trust cert ");
                while (alias != null) {
                    found = true;
                    log.debug("Deleting trust certificate alias entry: " + alias);
                    keyStore.deleteEntry(alias);
                    alias = keyStore.getCertificateAlias(trustCert);
                }
            }
        }
        catch (KeyStoreException e) {
                KeyStoreMgr.logAndThrow("Error removing trust certificate", e);
        }
        
        
        return found;
    }
    
    /**
     * Remove all trust certificates.
     * 
     * @throws KeyStoreMgrException On error
     */
    public void clearTrustCertificates()
            throws KeyStoreMgrException {
        log.debug("clearTrustCertificates");
        // Sanity check
        checkKeyStoreSession();
        String alias = null;

        try {
            Enumeration<String> aliases = keyStore.aliases();
            while (aliases.hasMoreElements()) {
                alias = aliases.nextElement();
                if (alias.equals(ownCertAlias)) {
                    if (DEBUG) {
                        log.debug("Skipping non-matching trust cert alias: " + alias);
                    }
                    continue;
                }
                log.debug("Deleting trust certificate alias entry: " + alias);
                keyStore.deleteEntry(alias);                   
            }
            crls = null;
        }
        catch (Exception e) {
            KeyStoreMgr.logAndThrow("Error clearing trust certificates", e);
        }
    }
    
    /**
     * Check for the existence of a trust certificate.
     * 
     * @param trustCert Certificate to test.
     * @return <code>true</code> if trust certificate found in key store.
     * @throws KeyStoreMgrException
     */
    public boolean trustCertificateExists(X509Certificate trustCert)
            throws KeyStoreMgrException {
        boolean exists = false;    
        // Sanity check
        checkKeyStoreSession();
        
        if (trustCert != null) {
            try {
                String alias = keyStore.getCertificateAlias(trustCert);
                exists = (alias != null && !alias.equals(ownCertAlias));
            }
            catch (KeyStoreException e) {
                KeyStoreMgr.logAndThrow("Error checking for existence of trust certificate.", e);
            }
        }
        
        return exists;
    }
    
    /**
     * Get the current trust certificates.
     * 
     * @param rootOnly <code>true</code> for root CAs only.
     * @return List of matching certificates.
     * @throws KeyStoreMgrException On error.
     * @throws CertUtilsException On error.
     */
    public Set<X509Certificate> getTrustCertificates(boolean rootOnly)
            throws KeyStoreMgrException, CertUtilsException {
        Set<X509Certificate> certs = new HashSet<X509Certificate>();
       
        try {
            Enumeration<String> aliases = keyStore.aliases();
            String alias = null;
            X509Certificate trustCert = null;
            while (aliases.hasMoreElements()) {
                alias = aliases.nextElement();
                if (keyStore.isCertificateEntry(alias)) {
                    trustCert = (X509Certificate)keyStore.getCertificate(alias);
                    if (CertUtils.isSelfSigned(trustCert)) {
                        if (DEBUG) {
                            log.debug("getTrustCertificates: Adding trust anchor cert: " + trustCert);
                        }
                        certs.add(trustCert);
                    } else if (!rootOnly) {
                        if (DEBUG) {
                            log.debug("getTrustCertificates: Adding intermediate trust cert: " + trustCert);
                        }
                        certs.add(trustCert);
                    }
                }
            }
        }
        catch (KeyStoreException e) {
            KeyStoreMgr.logAndThrow("Key store error during certificate verification.", e);
        }
        catch (Exception e) {
            CertUtils.logAndThrow("Certificate verification failed", e);
        }
        
        return certs;
    }
    
    /**
     * Verify the given certificate against the current trust store.
     * 
     * @param unverifiedCert Certificate under test.
     * @throws KeyStoreMgrException On verification failure.
     * @throws CertUtilsException On verification failure.
     */
    public void verifyCertificate(X509Certificate unverifiedCert)
            throws KeyStoreMgrException, CertUtilsException {
        Set<TrustAnchor> rootCAs = new HashSet<TrustAnchor>();
        Set<X509Certificate> intermediateCAs = new HashSet<X509Certificate>();
        if (DEBUG) {
            log.debug("verifyCertificate: " + unverifiedCert);
        }

        
        try {
            // Sort our root and intermediate CA certs.
            // * Optimization possibility
            // * This can be cached if we are careful
            // * about keeping the cache in sync with the
            // * current trust store state.
            Enumeration<String> aliases = keyStore.aliases();
            String alias = null;
            X509Certificate trustCert = null;
            while (aliases.hasMoreElements()) {
                alias = aliases.nextElement();
                if (keyStore.isCertificateEntry(alias)) {
                    trustCert = (X509Certificate)keyStore.getCertificate(alias);
                    if (CertUtils.isSelfSigned(trustCert)) {
                        if (DEBUG) {
                            log.debug("verifyCertificate: Adding trust anchor cert: " + trustCert);
                        }
                        rootCAs.add(new TrustAnchor(trustCert, null));
                    } else {
                        if (DEBUG) {
                            log.debug("verifyCertificate: Adding intermediate trust cert: " + trustCert);
                        }
                        intermediateCAs.add(trustCert);
                    }
                }
            }
    
            X509CertSelector target = new X509CertSelector();
            target.setCertificate(unverifiedCert);
        
            PKIXBuilderParameters params = new PKIXBuilderParameters(rootCAs, target);
            CertStoreParameters intermediates = new CollectionCertStoreParameters(intermediateCAs);
            params.addCertStore(CertStore.getInstance("Collection", intermediates));

            if (SecurityUtils.isReal(crls)) {
                if (DEBUG) {
                    log.debug("Enabling CRL checks");
                }
                CertStoreParameters revoked = new CollectionCertStoreParameters(crls);
                params.addCertStore(CertStore.getInstance("Collection", revoked));
            } else {
                if (DEBUG) {
                    log.debug("Disabling CRL checks");
                }
                params.setRevocationEnabled(false);
            }
            
            CertPathBuilder builder = CertPathBuilder.getInstance("PKIX");
            /* 
             * If build() returns successfully, the certificate is valid. More details 
             * about the valid path can be obtained through the PKIXBuilderResult.
             * If no valid path can be found, a CertPathBuilderException is thrown.
             */
            CertPathBuilderResult result = builder.build(params);
            if (DEBUG) {
                log.debug("verifyCertificate result=" + result);
            }
        }
        catch (KeyStoreException e) {
            KeyStoreMgr.logAndThrow("Key store error during certificate verification.", e);
        }
        catch (Exception e) {
            CertUtils.logAndThrow("Certificate verification failed", e);
        }

    }
    
    /**
     * Set our CRL list.
     * Note: The new value overwrites the existing list.
     * 
     * @param crls List of CRL objects or <code>null</code> to clear.
     */
    public void setCRLs(Collection<X509CRL> crls) {
        this.crls = crls;
    }
    
    /**
     * Get our CRL list.
     * 
     * @return List of CRLs or <code>null</code> if empty.
     */
    public Collection<X509CRL> getCRLs() {
        return crls;
    }

    /**
     * Get the current own certificate alias.
     * 
     * @return the ownCertAlias
     */
    public String getOwnCertAlias() {
        return ownCertAlias;
    }

    /**
     * Set the own certificate alias.
     * NOTE: This value is NOT persistent.
     *       It must be set after each openKeyStoreSession call.
     *       
     * @param ownCertAlias the ownCertAlias to set
     */
    public void setOwnCertAlias(String ownCertAlias) {
        this.ownCertAlias = ownCertAlias;
    }
}
