package com.netapp.vee.security;

import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStreamReader;
import java.lang.reflect.Method;
import java.util.HashMap;
import java.util.Set;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

public final class CertSecurityMgrCli implements SecurityConstants {
    private static Log log = LogFactory.getLog(CertSecurityMgrCli.class);
    
    private static final String CMD_HELP = "help";
    private static final String CMD_INIT_CERT = "initcert";
    private static final String CMD_DUMP = "dump";
    
	/**
	 * Debug logging - wrap Dev-ony code in this variable
	 */
	private static final boolean DEBUG = false;
	
	private static HashMap<String,String> COMMANDS;
	
	static {
        COMMANDS = new HashMap<String,String>();
        
        COMMANDS.put(CMD_HELP, "- Print this page");
        COMMANDS.put(CMD_INIT_CERT, "<key store filename> <Subject DN> <alias> - Initialize own key and certificate");
        COMMANDS.put(CMD_DUMP, "<key store filename> - Dump key store");

        if (DEBUG) {
            // This is to make sure that we don't leave DEBUG enabled
            LogFactory.getLog(CertSecurityMgrCli.class).warn("WARNING: Developer (NON-PRODUCTION) debug enabled!");
        }
	}
	
	private CertSecurityMgrCli() {
	    // empty
	}
	
    private static void printLogError(String msg, Throwable t) {
        System.err.print("ERROR: " + msg);
        String reason = "";
        if (t != null) {
            reason = ": " + t;
        }
        System.err.println(reason);
        log.error(msg, t);
    }
    
    private static String getPassword() throws IOException {
        System.out.print("Enter key store password: ");
        BufferedReader reader = new BufferedReader(new InputStreamReader(System.in));
        String password = reader.readLine();
        reader.close();
        return password;
    }
    
    private static boolean isNonBlank(String input) {
        return (input != null && input.length() > 0);
    }
    
    /**
     * Initialize the given key store (if needed).
     * Create or replace own certificate.
     * Save changes to key store file.
     * 
     * @return <code>true</code> for success.
     */
    private static boolean doInitcert(String[] args) {
        boolean ok = true;
        String keyStorePath = null;
        String subjectDN = null;
        String password = null;
        String alias = null;
        
        if (args.length != 4) {
            printLogError("Syntax error", null);
            doHelp(args);
            ok = false;
        } else if (isNonBlank(args[1]) && isNonBlank(args[2]) && isNonBlank(args[3])) {
            keyStorePath = args[1];
            subjectDN = args[2];
            alias = args[3];
            try {
                password = getPassword();
            }
            catch (IOException e) {
                printLogError("Error reading password input", e);
                ok = false;
            }
        } else {
            printLogError("Arguments cannot be blank", null);
            ok = false;
        }        
        
        if (DEBUG) {
            log.debug("Enter initOwnCert: keyStorePath=" + keyStorePath +
                      ", subjectDN=" + subjectDN + ", alias=" + alias +
                      ", password=" + password);
        }
        
        if (ok) {
            try {
                CertSecurityMgr secMgr = CertSecurityMgr.getInstance();
                secMgr.openKeyStoreSession(keyStorePath, password, true);
                secMgr.initOwnCertificate(subjectDN, alias);
                secMgr.closeKeyStoreSession(true);
                System.out.println("Own key/certificate initialized");
            }
            catch (Exception e) {
                printLogError("Own key/certificate generation failed", e);
                ok = false;
            }
        }
        
        return ok;
    }
    
    /**
     * Dump key store.
     * 
     * @return <code>true</code> for success.
     */
    private static boolean doDump(String[] args) {
        boolean ok = true;
        String keyStorePath = null;
        String password = null;
        
        if (args.length != 2) {
            printLogError("Syntax error", null);
            doHelp(args);
            ok = false;
        } else if (isNonBlank(args[1])) {
            keyStorePath = args[1];
            try {
                password = getPassword();
            }
            catch (IOException e) {
                printLogError("Error reading password input", e);
                ok = false;
            }
        } else {
            printLogError("Arguments cannot be blank", null);
            ok = false;
        }        
        
        if (DEBUG) {
            log.debug("Enter dump: keyStorePath=" + keyStorePath +
                      ", password=" + password);
        }
        
        if (ok) {
            try {
                CertSecurityMgr secMgr = CertSecurityMgr.getInstance();
                secMgr.openKeyStoreSession(keyStorePath, password, true);
                System.out.println(KeyStoreMgr.dumpKeyStore(secMgr.getKeyStore(), password, true, true));
                secMgr.closeKeyStoreSession(true);
            }
            catch (Exception e) {
                printLogError("Own key/certificate generation failed", e);
                ok = false;
            }
        }
        
        return ok;
    }
    
    private static boolean doHelp(String[] args) {
        System.out.print("usage: " + CertSecurityMgrCli.class.getCanonicalName() + " [");
        boolean first = true;
        Set<String> commands = COMMANDS.keySet();
        for (String command : commands) {
            if (first) {
                first = false;
            } else {
                System.out.print(" | ");
            }
            System.out.print(command);
        }
        System.out.println("]");
        
        for (String command : commands) {
            System.out.println("  " + command + " " + COMMANDS.get(command));
        }
        
        return false;
    }
    
    /**
     * CLI interface to CertSecurityMgr
     * 
     * @param args CLI args
     * @see CertSecurityMgr
     */
    public static void main(String[] args) {
        int retCode = runCommand(args);        
        System.exit(retCode);
    }

    public static int runCommand(String[] args) {
        int retCode = 0;
        
        if (args.length < 1) {
            // Just do help.
            doHelp(args);
            retCode = 1;
        } else if (COMMANDS.containsKey(args[0].toLowerCase())) {
            StringBuffer buf = new StringBuffer(args[0].toLowerCase());
            buf.setCharAt(0, Character.toUpperCase(buf.charAt(0)));
            buf.insert(0, "do");
            
            if (DEBUG) {
                log.debug("Invoking command method: " + buf.toString());
            }
            
            try {
                Method meth = CertSecurityMgrCli.class.getDeclaredMethod(buf.toString(), args.getClass());
                Object[] o = new Object[1];
                o[0] = args;
                Object ret =  meth.invoke(null, o);
                if (ret.equals(Boolean.FALSE)) {
                    retCode = 4;
                }
            }
            catch (NoSuchMethodException e) {
                retCode = 3;
                System.err.println("Unknown command: " + args[0]);
                doHelp(args);
            }
            catch (Exception e) {
                retCode = 2;
                printLogError("Unexpected error", e);
            }
        } else {
                retCode = 3;
                System.err.println("Unknown command: " + args[0]);
                doHelp(args);
        }
        return retCode;
    }

}
