package com.netapp.vee.security;

import java.io.File;
import java.io.FileReader;
import java.io.IOException;
import java.io.StringReader;
import java.math.BigInteger;
import java.net.InetAddress;
import java.security.InvalidKeyException;
import java.security.KeyPair;
import java.security.PublicKey;
import java.security.SecureRandom;
import java.security.SignatureException;
import java.security.cert.X509CRL;
import java.security.cert.X509Certificate;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.Date;
import java.util.List;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.bouncycastle.asn1.ASN1InputStream;
import org.bouncycastle.asn1.oiw.OIWObjectIdentifiers;
import org.bouncycastle.asn1.pkcs.PKCSObjectIdentifiers;
import org.bouncycastle.asn1.x500.X500Name;
import org.bouncycastle.asn1.x509.AlgorithmIdentifier;
import org.bouncycastle.asn1.x509.BasicConstraints;
import org.bouncycastle.asn1.x509.CRLReason;
import org.bouncycastle.asn1.x509.ExtendedKeyUsage;
import org.bouncycastle.asn1.x509.Extension;
import org.bouncycastle.asn1.x509.ExtensionsGenerator;
import org.bouncycastle.asn1.x509.GeneralName;
import org.bouncycastle.asn1.x509.GeneralNames;
import org.bouncycastle.asn1.x509.KeyPurposeId;
import org.bouncycastle.asn1.x509.KeyUsage;
import org.bouncycastle.asn1.x509.SubjectPublicKeyInfo;
import org.bouncycastle.asn1.x509.X509Extension;
import org.bouncycastle.cert.X509CRLHolder;
import org.bouncycastle.cert.X509CertificateHolder;
import org.bouncycastle.cert.X509ExtensionUtils;
import org.bouncycastle.cert.X509v2CRLBuilder;
import org.bouncycastle.cert.X509v3CertificateBuilder;
import org.bouncycastle.cert.jcajce.JcaX509CRLConverter;
import org.bouncycastle.cert.jcajce.JcaX509CertificateConverter;
import org.bouncycastle.crypto.params.AsymmetricKeyParameter;
import org.bouncycastle.crypto.util.PrivateKeyFactory;
import org.bouncycastle.openssl.PEMParser;
import org.bouncycastle.operator.ContentSigner;
import org.bouncycastle.operator.DefaultDigestAlgorithmIdentifierFinder;
import org.bouncycastle.operator.DefaultSignatureAlgorithmIdentifierFinder;
import org.bouncycastle.operator.DigestCalculator;
import org.bouncycastle.operator.bc.BcDigestCalculatorProvider;
import org.bouncycastle.operator.bc.BcRSAContentSignerBuilder;
import org.bouncycastle.operator.jcajce.JcaContentSignerBuilder;
import org.bouncycastle.pkcs.PKCS10CertificationRequest;
import org.bouncycastle.pkcs.PKCS10CertificationRequestBuilder;
import org.bouncycastle.pkcs.jcajce.JcaPKCS10CertificationRequestBuilder;
import org.bouncycastle.util.io.pem.PemObject;

/**
 * Certificate operations and utility methods.
 * 
 * @author cliftonj
 *
 */
public class CertUtils implements SecurityConstants {
    private static Log log = LogFactory.getLog(CertUtils.class);

	/**
	 * Debug logging - wrap Dev-ony code in this variable
	 */
	private static final boolean DEBUG = false;
	
	static {
        if (DEBUG) {
            // This is to make sure that we don't leave DEBUG enabled
            LogFactory.getLog(CertUtils.class).warn("WARNING: Developer (NON-PRODUCTION) debug enabled!", new RuntimeException());
        }	    
	}
	
    /**
     * No instances supported
     */
	private CertUtils() {
	    // empty
	}
	
    public static void logAndThrow(String message, Exception nested)
            throws CertUtilsException {
        CertUtilsException ex = new CertUtilsException(message, nested);
        log.error(ex.getMessage(), nested);
        throw ex;
    }
    
    /**
     * Test a certificate to see if it is self-signed.
     * 
     * @param cert Certificate under test.
     * @return <code>true</code> if certificate is self-signed
     * @throws CertUtilsException On error.
     */
	public static boolean isSelfSigned(X509Certificate cert)
	        throws CertUtilsException {
	    boolean result = true;
	    
	    try {
            cert.verify(cert.getPublicKey());
        } catch (InvalidKeyException e) {
            result = false;
        } catch (SignatureException e) {
            result = false;
        } catch (Exception e) {
            logAndThrow("Error checking certificate signature", e);
        }
	    
	    if (DEBUG) {
	        log.debug(cert + " ==> " + (result ? "" : "NOT ") + "SELF-SIGNED");
	    }
	    
	    return result;
	}
	
	/**
	 * Determine certificate certificate authority (CA) status.
	 * 
	 * @param cert Certificate under test.
	 * @return <code>true</code> if certificate is a CA.
	 * @throws CertUtilsException
	 */
	public static boolean isCA(X509Certificate cert)
	        throws CertUtilsException {
	    boolean result = false;
	    
	    try {
	        result = (cert.getBasicConstraints() == 0);
	    } catch (Exception e) {
	        logAndThrow("Error checking certificate CA status", e);
	    }
	    
	    return result;
	}
	
	/**
	 * Generate a self-signed certificate with CA role.
	 * 
	 * @param keyPair Public and private key for the new certificate.
	 * @param subjectDN String representation of Subject and Issuer fields.
	 * @return New self-signed certificate.
	 * @throws CertUtilsException On error.
	 */
	public static X509Certificate genSelfSignedCert(KeyPair keyPair, String subjectDN)
	        throws CertUtilsException {
	    return genSelfSignedCert(keyPair, subjectDN, true);
	}
	
	/**
	 * Generate a self-signed certificate.
	 * 
	 * @param keyPair Public and private key for the new certificate.
	 * @param subjectDN String representation of Subject and Issuer fields.
	 * @param isCA <code>true</code> if the signed certificate will be a certificate authority.
	 * @return New self-signed certificate.
	 * @throws CertUtilsException On error.
	 */
	public static X509Certificate genSelfSignedCert(KeyPair keyPair, String subjectDN, boolean isCA)
	        throws CertUtilsException {            
	    return genSelfSignedCert(keyPair, new X500Name(subjectDN), isCA);
	}
	
	/**
	 * Generate a self-signed certificate with CA role.
	 * 
	 * @param keyPair Certificate key pair.
	 * @param subjectDN Signed-certificate subject name.
	 * @return New self-signed certificate.
	 * @throws CertUtilsException On error.
	 */
	public static X509Certificate genSelfSignedCert(KeyPair keyPair, X500Name subjectDN)
	        throws CertUtilsException {
	    return genCASignedCert(subjectDN, keyPair.getPublic(), subjectDN, keyPair, true);
	}
	
	/**
	 * Generate a self-signed certificate.
	 * 
	 * @param keyPair Certificate key pair.
	 * @param subjectDN Signed-certificate subject name.
	 * @param isCA <code>true</code> if the signed certificate will be a certificate authority.
	 * @return New self-signed certificate.
	 * @throws CertUtilsException On error.
	 */
	public static X509Certificate genSelfSignedCert(KeyPair keyPair, X500Name subjectDN, boolean isCA)
	        throws CertUtilsException {
	    return genCASignedCert(subjectDN, keyPair.getPublic(), subjectDN, keyPair, isCA);
	}
	
	/**
	 * Process certificate signing request (CSR).
	 * 
	 * @param csr Certificate signing request to be processed.
	 * @param issuerDN Signing certificate subject name.
	 * @param issuerKeys Signing certificate keys.
	 * @return Signed certificate.
	 * @throws CertUtilsException On error.
	 */
	public static X509Certificate signCSR(PKCS10CertificationRequest csr, X500Name issuerDN, KeyPair issuerKeys)
	        throws CertUtilsException {
        AsymmetricKeyParameter issuerPrivKey = null;
        SubjectPublicKeyInfo issuerPubKey = null;
	    
        try {
            issuerPrivKey = PrivateKeyFactory.createKey(issuerKeys.getPrivate().getEncoded());
            issuerPubKey = SubjectPublicKeyInfo.getInstance(new ASN1InputStream(issuerKeys.getPublic().getEncoded()).readObject());
        }
        catch (IOException e) {
            logAndThrow("Sign CSR failed", e);
        }
	    
        return genCASignedCert(csr.getSubject(),
                               csr.getSubjectPublicKeyInfo(),
                               issuerDN,
                               issuerPrivKey,
                               issuerPubKey,
                               false);
	}

	/**
	 * 
	 * Sign the given certificate with the given issuer.
	 * 
	 * @param subjectDN Signed certificate subject name.
	 * @param certPubKey Signed certificate public key.
	 * @param issuerDN Issuer certificate subject name.
	 * @param issuerKeys Issuer certificate keys.
	 * @param isCA <code>true</code> if the signed certificate will be a certificate authority.
	 * @return Signed certificate.
	 * @throws CertUtilsException Signing error.
	 */
	public static X509Certificate genCASignedCert(X500Name subjectDN,
	                                              PublicKey certPubKey,
	                                              X500Name issuerDN,
	                                              KeyPair issuerKeys,
	                                              boolean isCA)
	        throws CertUtilsException {
        SubjectPublicKeyInfo subjectPubKey = null;
        AsymmetricKeyParameter issuerPrivKey = null;
        SubjectPublicKeyInfo issuerPubKey = null;
        
        try {
            subjectPubKey = SubjectPublicKeyInfo.getInstance(new ASN1InputStream(certPubKey.getEncoded()).readObject());
            issuerPrivKey = PrivateKeyFactory.createKey(issuerKeys.getPrivate().getEncoded());
            issuerPubKey = SubjectPublicKeyInfo.getInstance(new ASN1InputStream(issuerKeys.getPublic().getEncoded()).readObject());
        }
        catch (IOException e) {
            logAndThrow("Generate CA signed cert failed", e);
        }
        
        return genCASignedCert(subjectDN, subjectPubKey, issuerDN, issuerPrivKey, issuerPubKey, isCA);
	    
	}

	/**
	 * Sign the given certificate with the given issuer.
	 * 
	 * @param subjectDN Signed certificate subject name.
	 * @param certPubKey Signed certificate public key.
	 * @param issuerDN Issuer certificate subject name.
	 * @param issuerPrivKey Issuer certificate private key.
	 * @param issuerPubKey Issuer certificate public key.
	 * @param isCA <code>true</code> if the signed certificate will be a certificate authority.
	 * @return Signed certificate.
	 * @throws CertUtilsException Signing error.
	 */
	public static X509Certificate genCASignedCert(X500Name subjectDN,
	                                              SubjectPublicKeyInfo certPubKey,
	                                              X500Name issuerDN,
	                                              AsymmetricKeyParameter issuerPrivKey,
	                                              SubjectPublicKeyInfo issuerPubKey,
	                                              boolean isCA)
	        throws CertUtilsException {
	    
	    X509Certificate ret = null;
	    
        // TODO Break these params out for more control
        int randSn = new SecureRandom().nextInt();  // Get some random s/n
        BigInteger serial = BigInteger.valueOf(Math.abs(randSn));
        Date notBefore = new Date(); // Valid from now
        Date notAfter = new Date(notBefore.getTime() + DEFAULT_VALIDITY_MILLIS);
            
 	    try {
            AlgorithmIdentifier sigAlg = new DefaultSignatureAlgorithmIdentifierFinder().find(DEFAULT_SIGNATURE_ALGORITHM);
            AlgorithmIdentifier digAlg = new DefaultDigestAlgorithmIdentifierFinder().find(sigAlg);
            
            X509v3CertificateBuilder certBldr = new X509v3CertificateBuilder(issuerDN,
                                                                             serial,
                                                                             notBefore,
                                                                             notAfter, 
                                                                             subjectDN,
                                                                             certPubKey);
            // Add V3 certificate extensions
            DigestCalculator digCalc = new BcDigestCalculatorProvider().get(new AlgorithmIdentifier(OIWObjectIdentifiers.idSHA1));
            X509ExtensionUtils x509ExtensionUtils = new X509ExtensionUtils(digCalc);
            // Subject Key Identifier
            certBldr.addExtension(Extension.subjectKeyIdentifier, false,
                                  x509ExtensionUtils.createSubjectKeyIdentifier(certPubKey));
 
            // Authority Key Identifier
            certBldr.addExtension(Extension.authorityKeyIdentifier, false,
                                  x509ExtensionUtils.createAuthorityKeyIdentifier(issuerPubKey));
            
            // Usage extensions
            BasicConstraints bc = null;
            if (isCA) {
                bc = new BasicConstraints(0);
            } else {
                bc = new BasicConstraints(false);
            }
            certBldr.addExtension(Extension.basicConstraints, true, bc);
            KeyUsage usage = null;
            if (isCA) {
                usage = new KeyUsage(KeyUsage.digitalSignature |
                                     KeyUsage.keyCertSign |
                                     KeyUsage.cRLSign |
                                     KeyUsage.keyAgreement |
                                     KeyUsage.keyEncipherment);
            } else {
                usage = new KeyUsage(KeyUsage.digitalSignature |
                                     KeyUsage.keyEncipherment);
            }
            certBldr.addExtension(Extension.keyUsage, true, usage);
            
            // Subject Alternative Name
            // @TODO Resolve the whole SubjAlt thing with VMware
            //GeneralNames subjAltName = new GeneralNames(new GeneralName(GeneralName.dNSName, subjectDN.toString().substring(3)));
            String myIp = InetAddress.getLocalHost().getHostAddress();
            log.debug("Using IP for SubjAltName:" + myIp);
            GeneralNames subjAltName = new GeneralNames(new GeneralName(GeneralName.iPAddress, myIp));
            certBldr.addExtension(Extension.subjectAlternativeName, true, subjAltName);
            
            //Extended Key Usage
            KeyPurposeId[] extKeyUsage = new KeyPurposeId[2];
            extKeyUsage[0] = KeyPurposeId.id_kp_serverAuth; // TLS Web Server Authentication (1.3.6.1.5.5.7.3.1)
            extKeyUsage[1] = KeyPurposeId.id_kp_clientAuth; // TLS Web Client Authentication (1.3.6.1.5.5.7.3.2)
            certBldr.addExtension(Extension.extendedKeyUsage, false, new ExtendedKeyUsage(extKeyUsage));
 

            // Sign using issuer's private key
            ContentSigner signer = new BcRSAContentSignerBuilder(sigAlg,digAlg).build(issuerPrivKey);
            X509CertificateHolder holder = certBldr.build(signer);
            ret = new JcaX509CertificateConverter().setProvider(DEFAULT_PROVIDER_NAME).getCertificate(holder);
        }
        catch (Exception e) {
            logAndThrow("Generate CA signed cert failed", e);
        }
	    
	    return ret;	    
	}
	
	/**
	 * Generate a certificate signing request (CSR).
	 * 
	 * @param keyPair Key pair for requested certificate.
	 * @param subjectDN Certificate subject distinguished name.
	 * @param isCA <code>true</code> if this will be a certificate authority certificate.
	 * @return Generated CSR.
	 * @throws CertUtilsException On error.
	 */
	public static PKCS10CertificationRequest genCSR(KeyPair keyPair, X500Name subjectDN, boolean isCA)
	        throws CertUtilsException {
	    PKCS10CertificationRequest ret = null;
	    
	    try {
            AsymmetricKeyParameter privateKey = PrivateKeyFactory.createKey(keyPair.getPrivate().getEncoded());
            AlgorithmIdentifier sigAlg = new DefaultSignatureAlgorithmIdentifierFinder().find(DEFAULT_SIGNATURE_ALGORITHM);
            AlgorithmIdentifier digAlg = new DefaultDigestAlgorithmIdentifierFinder().find(sigAlg);
            
            ContentSigner signer = new BcRSAContentSignerBuilder(sigAlg, digAlg).build(privateKey);

            PKCS10CertificationRequestBuilder csrBuilder =
                    new JcaPKCS10CertificationRequestBuilder(subjectDN, keyPair.getPublic());
            ExtensionsGenerator extensionsGenerator = new ExtensionsGenerator();
            
            // Usage extensions
            extensionsGenerator.addExtension(X509Extension.basicConstraints, true, new BasicConstraints(isCA));
            KeyUsage usage = null;
            if (isCA) {
                usage = new KeyUsage(KeyUsage.digitalSignature |
                                     KeyUsage.keyCertSign |
                                     KeyUsage.cRLSign |
                                     KeyUsage.keyAgreement |
                                     KeyUsage.keyEncipherment);
            } else {
                usage = new KeyUsage(KeyUsage.digitalSignature |
                                     KeyUsage.keyEncipherment);
            }
            
            extensionsGenerator.addExtension(X509Extension.keyUsage, true, usage);
            
            // Subject Alternative Name
            // @TODO Resolve the whole SubjAltName thing with VMware
            //GeneralNames subjAltName = new GeneralNames(new GeneralName(GeneralName.dNSName, subjectDN.toString().substring(3)));
            String myIp = InetAddress.getLocalHost().getHostAddress();
            log.debug("Using IP for SubjAltName:" + myIp);
            GeneralNames subjAltName = new GeneralNames(new GeneralName(GeneralName.iPAddress, myIp));
            extensionsGenerator.addExtension(Extension.subjectAlternativeName, true, subjAltName);
            
            //Extended Key Usage
            KeyPurposeId[] extKeyUsage = new KeyPurposeId[2];
            extKeyUsage[0] = KeyPurposeId.id_kp_serverAuth; // TLS Web Server Authentication (1.3.6.1.5.5.7.3.1)
            extKeyUsage[1] = KeyPurposeId.id_kp_clientAuth; // TLS Web Client Authentication (1.3.6.1.5.5.7.3.2)
            extensionsGenerator.addExtension(Extension.extendedKeyUsage, false, new ExtendedKeyUsage(extKeyUsage));

            csrBuilder.addAttribute(PKCSObjectIdentifiers.pkcs_9_at_extensionRequest,
                extensionsGenerator.generate());
            
            ret = csrBuilder.build(signer);
        }
        catch (Exception e) {
            logAndThrow("CSR creation failed", e);
        }
	    
	    return ret;
	}
	
	/**
	 * Import X509 certificate from PEM format file.
	 * 
	 * @param certPath Path to certificate PEM file.
	 * @return Imported certificate(s).
	 * @throws CertUtilsException On error.
	 */
	public static X509Certificate[] importPEMCertificateFile(String certPath)
	        throws CertUtilsException {
	    X509Certificate[] ret = null;
	    // Expecting PEM style file format
        File pemFile = new File(certPath);
        FileReader reader = null;
        ArrayList<X509Certificate> list = new ArrayList<X509Certificate>();
        
        try {
            reader = new FileReader(pemFile);
            PEMParser pemParser = new PEMParser(reader);
            X509CertificateHolder certHolder = null;
            while (pemParser.ready()) {
                certHolder = (X509CertificateHolder)pemParser.readObject();
                list.add((X509Certificate) new JcaX509CertificateConverter().setProvider(DEFAULT_PROVIDER_NAME).getCertificate(certHolder));
            }
            ret = list.toArray(new X509Certificate[list.size()]);
        } catch (Exception e) {
            logAndThrow("PEM X509 certificate import failed.", e);
        }

	    return ret;
	}
	
	/**
	 * Import X509 certificate from PEM format file.
	 * This method expects the certificate BEGIN/END markers.
	 * 
	 * @param pemData String containing PEM certificates.
	 * @return Imported certificate(s).
	 * @throws CertUtilsException On error.
	 */
	public static X509Certificate[] importPEMCertificate(String pemData)
	        throws CertUtilsException {
	    X509Certificate[] ret = null;
        ArrayList<X509Certificate> list = new ArrayList<X509Certificate>();
        StringReader reader = null;
        
        try {
            reader = new StringReader(pemData);
            PEMParser pemParser = new PEMParser(reader);
            X509CertificateHolder certHolder = null;
            while (pemParser.ready()) {
                certHolder = (X509CertificateHolder)pemParser.readObject();
                if (certHolder == null) {
                    break;
                }
                list.add((X509Certificate) new JcaX509CertificateConverter().setProvider(DEFAULT_PROVIDER_NAME).getCertificate(certHolder));
            }
            ret = list.toArray(new X509Certificate[list.size()]);
        } catch (Exception e) {
            logAndThrow("PEM X509 certificate import failed.", e);
        }

	    return ret;	    
	}
	
	/**
	 * Export certificates to a PEM file.
	 * 
	 * @param certChain Chain of certificates to export.
	 * @param certPath File path for PEM file.
	 * @throws CertUtilsException Export error.
	 */
	public static void exportCertificateToPEMFile(X509Certificate[] certChain, String certPath)
	        throws CertUtilsException {
	    PemObject[] pObjs = new PemObject[certChain.length];
	    
	    try {
    	    for (int i = 0; i < certChain.length; i++) {
    	        pObjs[i] = new PemObject(PEMType.CERTIFICATE.getText(), certChain[i].getEncoded());
    	    }
    	    
    	    SecurityUtils.exportToPEM(pObjs, certPath);
        } catch (Exception e) {
            logAndThrow("PEM X509 certificate export failed.", e);
        }
	}
	
	/**
	 * Export certificates to a list of PEM strings.
	 * 
	 * @param chain Chain of certificates.
	 * @return List of PEM strings
	 * @throws CertUtilsException On error.
	 */
    public static List<String> toCertList(X509Certificate ... chain) throws CertUtilsException {
        ArrayList<String> ret = new ArrayList<String>();
        X509Certificate[] certArray = new X509Certificate[1];
        
        for (X509Certificate cert : chain) {
            certArray[0] = cert;
            ret.add(CertUtils.exportCertificateToPEM(certArray));
        }
        
        return ret;
    }
    
	/**
	 * Export certificates to a PEM string.
	 * 
	 * @param certChain Chain of certificates to export.
	 * @return PEM data string.
	 * @throws CertUtilsException Export error.
	 */
	public static String exportCertificateToPEM(X509Certificate[] certChain)
	        throws CertUtilsException {
	    PemObject[] pObjs = new PemObject[certChain.length];
	    String ret = null;
	    
	    try {
    	    for (int i = 0; i < certChain.length; i++) {
    	        pObjs[i] = new PemObject(PEMType.CERTIFICATE.getText(), certChain[i].getEncoded());
    	    }
    	    
    	    ret = SecurityUtils.exportToPEM(pObjs);
        } catch (Exception e) {
            logAndThrow("PEM X509 certificate export failed.", e);
        }
	    
	    return ret;
	}
	
	/**
	 * Import X509 certificate signing request (CSR) from PEM format file.
	 * 
	 * @param csrPath Path to CSR PEM file.
	 * @return Imported CSR.
	 * @throws CertUtilsException On error.
	 */
	public static PKCS10CertificationRequest importPEMCSRFile(String csrPath)
	        throws CertUtilsException {
	    PKCS10CertificationRequest ret = null;
	    // Expecting PEM style file format
        File pemFile = new File(csrPath);
        FileReader reader = null;
        
        try {
            reader = new FileReader(pemFile);
            PEMParser pemParser = new PEMParser(reader);
            while (pemParser.ready() && ret == null) {
                ret = (PKCS10CertificationRequest)pemParser.readObject();
                if (ret == null) {
                    break;
                }
            }
        } catch (Exception e) {
            logAndThrow("PEM X509 CSR import failed.", e);
        }

	    return ret;
	}
	
	/**
	 * Import X509 certificate signing request (CSR) from PEM format file.
	 * This method expects the CSR BEGIN/END markers.
	 * 
	 * @param pemData String containing PEM CSR.
	 * @return Imported CSR.
	 * @throws CertUtilsException On error.
	 */
	public static PKCS10CertificationRequest importPEMCSR(String pemData)
	        throws CertUtilsException {
	    PKCS10CertificationRequest ret = null;
        StringReader reader = null;
        
        try {
            reader = new StringReader(pemData);
            PEMParser pemParser = new PEMParser(reader);
            while (pemParser.ready() && ret == null) {
                ret = (PKCS10CertificationRequest)pemParser.readObject();
                if (ret == null) {
                    break;
                }
            }
        } catch (Exception e) {
            logAndThrow("PEM X509 CSR import failed.", e);
        }

	    return ret;	    
	}
	
	/**
	 * Export certificate signing request (CSR) to a PEM file.
	 * 
	 * @param csr CSR to export.
	 * @param certPath File path for PEM file.
	 * @throws CertUtilsException Export error.
	 */
	public static void exportCSRToPEMFile(PKCS10CertificationRequest csr, String certPath)
	        throws CertUtilsException {
	    PemObject[] pObjs = new PemObject[1];
	    
	    try {
    	    pObjs[0] = new PemObject(PEMType.CSR.getText(), csr.getEncoded());    	    
    	    SecurityUtils.exportToPEM(pObjs, certPath);
        } catch (Exception e) {
            logAndThrow("PEM X509 CSR export failed.", e);
        }
	}
	
	/**
	 * Export certificate signing request (CSR) to a PEM string.
	 * 
	 * @param csr CSR to export.
	 * @return PEM data string.
	 * @throws CertUtilsException Export error.
	 */
	public static String exportCSRToPEM(PKCS10CertificationRequest csr)
	        throws CertUtilsException {
	    PemObject[] pObjs = new PemObject[1];
	    String ret = null;
	    
	    try {
    	    pObjs[0] = new PemObject(PEMType.CSR.getText(), csr.getEncoded());
    	    
    	    ret = SecurityUtils.exportToPEM(pObjs);
        } catch (Exception e) {
            logAndThrow("PEM X509 CSR export failed.", e);
        }
	    
	    return ret;
	}
	
	/**
	 * Import X509 certificate revocation lists (CRL) from PEM format file.
	 * 
	 * @param csrPath Path to CRL PEM file.
	 * @return Imported CRL list.
	 * @throws CertUtilsException On error.
	 */
	public static Collection<X509CRL> importPEMCRLFile(String crlPath)
	        throws CertUtilsException {
	    ArrayList<X509CRL> ret = new ArrayList<X509CRL>();
	    // Expecting PEM style file format
        File pemFile = new File(crlPath);
        FileReader reader = null;
        
        try {
            reader = new FileReader(pemFile);
            PEMParser pemParser = new PEMParser(reader);
            X509CRLHolder holder = null;
            JcaX509CRLConverter conv = new JcaX509CRLConverter().setProvider(DEFAULT_PROVIDER_NAME);
            while (pemParser.ready()) {
                holder = (X509CRLHolder)pemParser.readObject();
                if (holder == null) {
                    break;
                }
                ret.add(conv.getCRL(holder));
            }
        } catch (Exception e) {
            logAndThrow("PEM X509 CRL import failed.", e);
        }

	    return ret;
	}
	
	/**
	 * Import X509 certificate revocation lists (CRL) from PEM format file.
	 * This method expects the CSR BEGIN/END markers.
	 * 
	 * @param pemData String containing PEM CRL.
	 * @return Imported CRL list.
	 * 
	 * @throws CertUtilsException On error.
	 */
	public static Collection<X509CRL> importPEMCRL(String pemData)
	        throws CertUtilsException {
	    ArrayList<X509CRL> ret = new ArrayList<X509CRL>();
        StringReader reader = null;
        
        try {
            reader = new StringReader(pemData);
            PEMParser pemParser = new PEMParser(reader);
            X509CRLHolder holder = null;
            JcaX509CRLConverter conv = new JcaX509CRLConverter().setProvider(DEFAULT_PROVIDER_NAME);
            while (pemParser.ready()) {
                holder = (X509CRLHolder)pemParser.readObject();
                if (holder == null) {
                    break;
                }
                ret.add(conv.getCRL(holder));
            }
        } catch (Exception e) {
            logAndThrow("PEM X509 CRL import failed.", e);
        }

	    return ret;	    
	}
	
	/**
	 * Export certificate revocation lists (CRL) to a PEM file.
	 * 
	 * @param csr CRLs to export.
	 * @param certPath File path for PEM file.
	 * @throws CertUtilsException Export error.
	 */
	public static void exportCRLToPEMFile(Collection<X509CRL> crls, String crlPath)
	        throws CertUtilsException {
	    PemObject[] pObjs = new PemObject[crls.size()];
	    
	    try {
	        int i = 0;
	        for (X509CRL crl : crls) {
	            pObjs[i] = new PemObject(PEMType.CRL.getText(), crl.getEncoded());
	            i++;
	        }
    	    SecurityUtils.exportToPEM(pObjs, crlPath);
        } catch (Exception e) {
            logAndThrow("PEM X509 CRL export failed.", e);
        }
	}
	
	/**
	 * Export list of CRLs to list of PEM strings.
	 * 
	 * @param crls CRL list.
	 * @return List of PEM CRL strings.
	 * @throws CertUtilsException On error.
	 */
    public static List<String> toCRLList(X509CRL ... crls) throws CertUtilsException {
        ArrayList<String> ret = new ArrayList<String>();
        ArrayList<X509CRL> crlCol = new ArrayList<X509CRL>();
        
        if (crls != null && crls.length > 0 && crls[0] != null) {
            for (X509CRL crl : crls) {
                crlCol.add(crl);
                ret.add(CertUtils.exportCRLToPEM(crlCol));
                crlCol.clear();
            }
        }
        
        return ret;
    }
    
	/**
	 * Export certificate revocation lists (CRL) to a PEM string.
	 * 
	 * @param csr CSR to export.
	 * @return PEM data string.
	 * @throws CertUtilsException Export error.
	 */
	public static String exportCRLToPEM(Collection<X509CRL> crls)
	        throws CertUtilsException {
	    PemObject[] pObjs = new PemObject[crls.size()];
	    String ret = null;
	    
	    try {
	        int i = 0;
	        for (X509CRL crl : crls) {
	            pObjs[i] = new PemObject(PEMType.CRL.getText(), crl.getEncoded());
	            i++;
	        }
    	    
    	    ret = SecurityUtils.exportToPEM(pObjs);
        } catch (Exception e) {
            logAndThrow("PEM X509 CRL export failed.", e);
        }
	    
	    return ret;
	}
    
    /**
     * Generate a CRL using own certificate.
     * 
     * @param sn Serial number of revoked signed certificate.
     * @return CRL
     * @throws CertUtilsException 
     */
    public static X509CRL genCRL(X509Certificate caCert,
                          KeyPair caKeys,
                          BigInteger revokedSn)
                                  throws CertUtilsException {
        X509CRL ret = null;
        
        Date now = new Date();
        try {
            X509v2CRLBuilder crlGen = new X509v2CRLBuilder(new X500Name(caCert.getSubjectDN().getName()), now);
            
            Date nextUpdate = new Date(now.getTime() + MULT_DAY);  // Update daily
            crlGen.setNextUpdate(nextUpdate);
            crlGen.addCRLEntry(revokedSn, now, CRLReason.privilegeWithdrawn);
            SubjectPublicKeyInfo caPubKeyInfo = SubjectPublicKeyInfo.getInstance(new ASN1InputStream(caKeys.getPublic().getEncoded()).readObject());
            
            DigestCalculator digCalc = new BcDigestCalculatorProvider().get(new AlgorithmIdentifier(OIWObjectIdentifiers.idSHA1));
            X509ExtensionUtils x509ExtensionUtils = new X509ExtensionUtils(digCalc);
            // Authority Key Identifier
            crlGen.addExtension(Extension.authorityKeyIdentifier, false,
                                x509ExtensionUtils.createAuthorityKeyIdentifier(caPubKeyInfo));
            
            X509CRLHolder crl = crlGen.build(new JcaContentSignerBuilder("SHA256withRSAEncryption").setProvider(DEFAULT_PROVIDER_NAME).build(caKeys.getPrivate()));
            ret = new JcaX509CRLConverter().setProvider(DEFAULT_PROVIDER_NAME).getCRL(crl);
        } catch (Exception e) {
            logAndThrow("X509 CRL creation failed.", e);
        }
        
        return ret;
    }
    
    /**
     * Convert a X509 certificate chain into a printable string.
     * 
     * @param certs X509 certificate chain.
     * @return String representation.
     */
    public static String dumpCertChain(X509Certificate[] certs) {
        if (certs != null) {
            return dumpCertChain(Arrays.asList(certs));
        } else {
            return "null certs";
        }
    }
    
    /**
     * Convert a X509 certificate chain into a printable string.
     * 
     * @param certs X509 certificate chain.
     * @return String representation.
     */
    public static String dumpCertChain(List<X509Certificate> certs) {
        StringBuffer buf = new StringBuffer();
        
        if (certs == null || certs.size() == 0) {
            buf.append("<EMPTY CHAIN>");
        } else {
            buf.append("[\n");
            boolean first = true;
            for (X509Certificate cert : certs) {
                if (first) {
                    first = false;
                } else {
                    buf.append(",\n");
                }
                buf.append(cert);
            }
            buf.append(']');
        }
        
        return buf.toString();
    }
}
