package com.netapp.vee.security;

import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.security.KeyStore;
import java.security.KeyStoreException;
import java.security.NoSuchAlgorithmException;
import java.security.UnrecoverableKeyException;
import java.security.cert.CertificateException;
import java.util.Enumeration;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

/**
 * KeyStore operations and utility methods.
 * 
 * @author cliftonj
 *
 */
public class KeyStoreMgr implements SecurityConstants {
    private static Log log = LogFactory.getLog(KeyStoreMgr.class);

	/**
	 * Debug logging - wrap Dev-ony code in this variable
	 */
	private static final boolean DEBUG = false;
	
	static {
        if (DEBUG) {
            // This is to make sure that we don't leave DEBUG enabled
            LogFactory.getLog(KeyStoreMgr.class).warn("WARNING: Developer (NON-PRODUCTION) debug enabled!", new RuntimeException());
        }
	}
	
    /**
     * No instances supported
     */
    private KeyStoreMgr() {
        // empty
    }
    
    public static void logAndThrow(String message, Exception nested)
            throws KeyStoreMgrException {
        KeyStoreMgrException ex = new KeyStoreMgrException(message, nested);
        log.error(ex.getMessage(), nested);
        throw ex;
    }
    
    /**
     * Create a new default key store.
     * 
     * @return A new KeyStore instance or <code>null</code> on error.
     * @throws KeyStoreException 
     */
    public static KeyStore getNewKeyStore()
            throws KeyStoreMgrException {
        KeyStore ret = null;
        try {
            ret = KeyStore.getInstance(DEFAULT_KEY_STORE_TYPE);
            ret.load(null, null);
        }
        catch (Exception e) {
            logAndThrow("Load key store create failed", e);
        }
        
        return ret;
    }
    
    /**
     * Load a key store from a file.
     * 
     * @param ksPath Path to key store file.
     * @param ksPassword Key store password.
     * @return KeyStore instance loaded from file.
     * @throws NoSuchAlgorithmException
     * @throws CertificateException
     * @throws IOException
     * @throws KeyStoreException
     */
    public static KeyStore loadKeyStoreFile(String ksPath, String ksPassword) 
            throws KeyStoreMgrException {
        KeyStore ret = null;
        FileInputStream is = null;
        
        try {
            ret = getNewKeyStore();
            is = new FileInputStream(ksPath);
            
            ret.load(is, ksPassword.toCharArray());
        }
        catch (Exception e) {
            logAndThrow("Load key store file failed", e);
        } finally {
            try {
                if (is != null) {
                    is.close();
                }
            } catch (IOException ioe) {
                // ignore
            }
        }
        
        return ret;
    }
    
    /**
     * Store a key store to a file.
     * 
     * @param ks KeyStore to save to file.
     * @param ksPath File path for key store.
     * @param ksPassword Key store password.
     * @throws KeyStoreException
     * @throws NoSuchAlgorithmException
     * @throws CertificateException
     * @throws IOException
     */
    public static void saveKeyStoreFile(KeyStore ks, String ksPath, String ksPassword)
            throws KeyStoreMgrException {
        FileOutputStream os = null;
        
        try {
            os = new FileOutputStream(ksPath);
            
            ks.store(os, ksPassword.toCharArray());
        }
        catch (Exception e) {
            logAndThrow("Save key store file failed", e);
        } finally {
            try {
                if (os != null) os.close();
            } catch (IOException ioe) {
                // ignore
            }

        }
    }
    
    /**
     * Find an alias based on the given pattern.
     * 
     * @param ks KeyStore for alias.
     * @param pattern Alias pattern.
     * @return Alias based on pattern and trailing numeral.
     * @throws KeyStoreException
     */
    public static String findUnusedAlias(KeyStore ks, String pattern)
            throws KeyStoreMgrException {
        String retAlias = pattern;
        int count = 0;
        
        // Sanity check
        if (pattern == null || pattern.length() == 0) {
            logAndThrow("Invalid alias pattern", null);
        }
        
        try {
            while (ks.containsAlias(retAlias)) {
                retAlias = pattern + ++count;
            }
        }
        catch (Exception e) {
            logAndThrow("find unused alias failed", e);
        }

        return retAlias;
    }
    
    /**
     * Dump the key store information for debugging.
     * 
     * @param ks Key store to dump.
     * @param privateKeys <code>true</code> to include private keys and associated certificate.
     * @param ksPassword Key store password.
     * @param trustCerts <code>true</code> to include trust certificates.
     * @return String with key store data.
     * @throws KeyStoreMgrException On error
     */
    public static String dumpKeyStore(KeyStore ks, String ksPassword, boolean privateKeys, boolean trustCerts)
            throws KeyStoreMgrException {
        StringBuffer privBuf = new StringBuffer();
        StringBuffer certBuf = new StringBuffer();
        KeyStore.ProtectionParameter pwParm = new KeyStore.PasswordProtection(ksPassword.toCharArray());
        
        privBuf.append("*** PRIVATE KEY DATA ***").append("\n\n");
        certBuf.append("*** CERTIFICATE DATA ***").append("\n\n");
        
        String alias = null;
        Enumeration<String> aliases = null;
        
        try {
            aliases = ks.aliases();
        }
        catch (Exception e1) {
                logAndThrow("Error dumping key private key entry", e1);
        }
        while (aliases.hasMoreElements()) {
            alias = aliases.nextElement();
            KeyStore.Entry entry = null;
            try {
                if (ks.entryInstanceOf(alias, KeyStore.TrustedCertificateEntry.class)) {
                    if (trustCerts) {
                        entry = ks.getEntry(alias, null);
                        certBuf.append("alias=").append(alias).append(", entry=").append(entry).append("\n\n-----\n\n");
                    }
                } else if (privateKeys) {
                    entry = ks.getEntry(alias, pwParm); 
                    privBuf.append("alias=").append(alias).append(", entry=").append(entry).append("\n\n-----\n\n");
                }
            }
            catch (Exception e) {
                logAndThrow("Error dumping key private key entry", e);
            }
        }
        
        privBuf.append("=====").append("\n\n").append(certBuf);
        
        return privBuf.toString();
    }
}
