package com.netapp.vee.security;

import java.io.File;
import java.io.FileReader;
import java.io.IOException;
import java.security.KeyPair;
import java.security.KeyPairGenerator;
import java.security.SecureRandom;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.bouncycastle.openssl.PEMDecryptorProvider;
import org.bouncycastle.openssl.PEMEncryptedKeyPair;
import org.bouncycastle.openssl.PEMKeyPair;
import org.bouncycastle.openssl.PEMParser;
import org.bouncycastle.openssl.jcajce.JcaPEMKeyConverter;
import org.bouncycastle.openssl.jcajce.JcePEMDecryptorProviderBuilder;

/**
 * KeyPair operations and utility meothds.
 *  
 * @author cliftonj
 *
 */
public class KeyUtils implements SecurityConstants {
    private static Log log = LogFactory.getLog(KeyUtils.class);
    
    /**
     * No instances supported
     */
    private KeyUtils() {
        // empty
    }
    
    public static void logAndThrow(String message, Exception nested)
            throws KeyUtilsException {
        KeyUtilsException ex = new KeyUtilsException(message, nested);
        log.error(ex.getMessage(), nested);
        throw ex;
    }
    
    /**
     * Generate a public/private key pair.
     * 
     * @return New key pair.
     * @throws KeyUtilsException Generation error.
     */
    public static KeyPair genKeyPair()
            throws KeyUtilsException {
        KeyPairGenerator kpGen = null;
        
        try {
            kpGen = KeyPairGenerator.getInstance(DEFAULT_KEY_PAIR_TYPE, DEFAULT_PROVIDER_NAME);
            SecureRandom randGen = SecureRandom.getInstance(DEFAULT_RANDOM_NUMBER_GEN);
            
            kpGen.initialize(DEFAULT_KEY_SIZE, randGen);
        }
        catch (Exception e) {
            logAndThrow("Generate key pair failed", e);
        }
        
        return kpGen.generateKeyPair();
    }
    
    /**
     * Import PEM key file.
     * 
     * @param kpPath Path to PEM format key file.
     * @param kpPassword Private key password.
     * @return Imported KeyPair.
     * @throws KeyUtilsException On error.
     */
    public static KeyPair importPEMKeyPair(String kpPath, String kpPassword)
            throws KeyUtilsException {
        KeyPair ret = null;
        // Expecting PEM style file format
        File kpFile = new File(kpPath);
        FileReader reader = null;
        
        try {
            reader = new FileReader(kpFile);
            PEMParser pemParser = new PEMParser(reader);
            Object kpObj = pemParser.readObject();
            JcaPEMKeyConverter converter = new JcaPEMKeyConverter().setProvider(DEFAULT_PROVIDER_NAME);
            
            if (kpObj instanceof PEMEncryptedKeyPair) {
                log.debug("Decrypting imported key pair");
                // Decrypt the PEMKeyPair as needed.
                PEMDecryptorProvider decryptorProv = new JcePEMDecryptorProviderBuilder().build(kpPassword.toCharArray());
                kpObj = ((PEMEncryptedKeyPair)kpObj).decryptKeyPair(decryptorProv);
            }
            ret = converter.getKeyPair((PEMKeyPair)kpObj);
         }
        catch (Exception e) {
            logAndThrow("PEM Key pair import failed", e);
        } finally {
            try {
                if (reader != null) {
                    reader.close();
                }
            } catch (IOException ioe) {
                // ignore
            }
        }
        
        return ret;
    }
}
