package com.netapp.vee.security;

import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.ByteArrayOutputStream;
import java.io.Closeable;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.Flushable;
import java.io.IOException;
import java.io.StringWriter;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.List;
import java.util.Map;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.bouncycastle.openssl.PEMWriter;
import org.bouncycastle.util.io.pem.PemObject;

/**
 * Misc. security utility functions that don't fit into
 * a more specific location.
 * 
 * @author cliftonj
 *
 */
public class SecurityUtils implements SecurityConstants {
    private static Log log = LogFactory.getLog(SecurityUtils.class);
    
    /**
     * No instances supported
     */
    private SecurityUtils() {
        // empty
    }
    
    /**
     * Export PEM object(s) to a PEM file.
     * 
     * @param pObjs PEM object chain.
     * @param pemPath File path name for exported objects.
     * @throws IOException On file I/O error.
     */
    public static void exportToPEM(PemObject[] pObjs, String pemPath)
            throws IOException {
        String pemStr = exportToPEM(pObjs);
        FileWriter fw = new FileWriter(pemPath);
        fw.write(pemStr);
        fw.close();
    }

    /**
     * Export PEM object(s) to PEM string.
     * 
     * @param pObjs PEM object chain.
     * @return String containing PEM data.
     * @throws IOException On string I/O error.
     */
    public static String exportToPEM(PemObject[] pObjs)
            throws IOException {
        StringWriter sw = new StringWriter();
        PEMWriter pw = new PEMWriter(sw);
        
        for (PemObject pObj : pObjs) {
            pw.writeObject(pObj);
        }
        pw.close();
        sw.close();
        
        return sw.toString();
    }
    
    public static boolean isLinux() {
        return System.getProperty("os.name").equalsIgnoreCase("linux");
    }
    
    public static boolean isReal(Collection c) {
        return (c != null && c.size() > 0);
    }

    public static boolean isReal(Map c) {
        return (c != null && c.size() > 0);
    }

    public static boolean isWindows() {
        return System.getProperty("os.name").toUpperCase().contains("WIN");
    }

    public static String join(Collection<String> value, String delimiter) {
        if (value == null || delimiter == null) {
            return "";
        }

        StringBuilder retVal = new StringBuilder();
        for (String s : value) {
            retVal.append(s);
            retVal.append(delimiter);
        }
        if (retVal.length() > delimiter.length()) {
            retVal.setLength(retVal.length() - delimiter.length());
        }
        return retVal.toString();
    }

    public static String join(String[] value, String delimiter) {
        if (value != null) {
            return join(Arrays.asList(value), delimiter);
        }
        return "";
    }

    public static List<String> getLinesSafe(String filename) throws Exception {
        try {
            return getLines(filename);
        } catch (Exception e) {
            return new ArrayList<String>();
        }
    }

    public static List<String> getLines(String filename) throws Exception {
        List<String> retVal = new ArrayList<String>();
        FileReader fr = new FileReader(filename);
        BufferedReader br = new BufferedReader(fr);
        String line;
        while ((line = br.readLine()) != null) {
            retVal.add(line);
        }
        br.close();
        fr.close();
        return retVal;
    }

    public static void writeListToFile(String filename, List<String> contents, boolean append) throws IOException {
        BufferedWriter bw = null;
        try {
            bw = new BufferedWriter(new FileWriter(filename, append));
            for (String string : contents) {
                bw.write(string + System.getProperty("line.separator"));
            }
            bw.flush();
            bw.close();
        }
        finally {
            if (bw != null) {
                bw.close();
            }
        }
    }
    
    /**
     * Add a file separator
     */
    public static String addFS(String pathPossiblyWithFileSeparator) {
        if (pathPossiblyWithFileSeparator != null) {
            if (!pathPossiblyWithFileSeparator.endsWith(java.io.File.separator)) {
                pathPossiblyWithFileSeparator += java.io.File.separator;
            }
        }
        return pathPossiblyWithFileSeparator;
    }

    /**
     * Add a file separator
     * @return
     */
    public static String addFS2(String... pathsPossiblyWithFileSeparator) {
        String temp ="";
        for(String pathPossiblyWithFileSeparator : pathsPossiblyWithFileSeparator){
            if (pathPossiblyWithFileSeparator != null) {
                if (!pathPossiblyWithFileSeparator.endsWith(java.io.File.separator)) {
                    pathPossiblyWithFileSeparator += java.io.File.separator;
                    temp = temp + pathPossiblyWithFileSeparator;
                }
            }
        }
        return temp;
    }

    public static boolean isReal(String incoming) {
        if (incoming == null || incoming.length() == 0 || incoming.trim().length() == 0) {
            return false;
        }
        return true;
    }
    
    public static void writeFile(String filename, String contents) throws IOException {
        writeFile(filename, contents, false);
    }

    public static void writeFile(String filename, String contents, boolean append) throws IOException {
        BufferedWriter bw = null;
        try {
            bw = new BufferedWriter(new FileWriter(filename, append));
            bw.write(contents);
            bw.flush();
        } finally {
            safeClose(bw);
        }
    }

    public static void writeFileBinary(String filename, byte[] contents, boolean append) throws IOException {
        FileOutputStream bw = null;
        try {
            bw = new FileOutputStream(filename, append);
            bw.write(contents);
            bw.flush();
            bw.close();
        } finally {
            safeClose(bw);
        }
    }

    public static void safeClose(Closeable stream) {
        if (stream != null) {
            try {
                stream.close();
            } catch (IOException e) {
                // ignore
            }
        }
    }

    public static void safeFlush(Flushable stream) {
        if (stream != null) {
            try {
                stream.flush();
            } catch (IOException e) {
                // ignore
            }
        }
    }

    public static byte[] readFileBinary(String filename) throws IOException {
        FileInputStream fis = new FileInputStream(filename);
        ByteArrayOutputStream baos = new ByteArrayOutputStream();
        try {
            byte[] bytes = new byte[16384];
            int length;
            while ((length = fis.read(bytes)) != -1) {
                baos.write(bytes, 0, length);
            }
            return baos.toByteArray();
        } finally {
            safeClose(fis);
        }
    }

    public static void writeFileSafe(String filename, String contents) {
        try {
            writeFile(filename, contents);
        } catch (IOException e) {
            // whatever
        }
    }

    public static List<String> splitIntoList(String incoming, String delimiter) {
        List<String> retVal = new ArrayList<String>();
        if (incoming == null || delimiter == null) {
            return retVal;            
        }
        String[] asArray = incoming.split(delimiter);
        for (String curr : asArray) {
            retVal.add(curr);
        }
        return retVal;
    }

}
